#include "stdafx.h"

extern "C" {
#include "sergey_rsf.h"
}
#include "cubeData.hpp"
#include "HaleDipFltOp.h"
//#include "cg_solv.h"
                   
///Dip field, the first dimension is 1 or 2, denotes 2D or 3D vector notation. (the first z component can be inferred since z^2+x^2+y^2 = 1.0f
///then follows axes indicating the 2D/3D cube. Always assume dipping downward, i.e all z component should be positive.
using namespace Yang;

int HaleDipFltOp::buildDipFilter(float* dipfield, const hypercube& shapeDip) {
  ndim = shapeDip.get_n(1)+1;
  VERIFY(ndim==2 || ndim==3);
  this->datShape  = shapeDip.return_compatible(2,2+ndim-1);
  this->modShape  = this->datShape;
  //save the dipfield
  long n123 = shapeDip.get_n123();
  mDip.resize(n123,0.f);
  myMemcpy(&(mDip[0]),dipfield, n123);
  return 0;
}

//Hale's one-pass method, has some problem?
int HaleDipFltOp::ATAfltH(float* out, float* in, long nd) {
//  PRV(nd); PRV(datShape.get_n123());
  VERIFY(nd == datShape.get_n123());
  int n1,n2,n3 = 1;
  stdVecInt ns = datShape.get_ns();
  n1=ns[0];n2=ns[1]; //PRVEC(ns,2);
  if (this->ndim == 2) {
    float u2i,u1i,a11,a12,a22,fa,fb,f1,f2,g1,g2,ga,gb;
    for (int i2=1; i2<n2; i2++) {
      for (int i1=1; i1<n1; i1++) {
        u2i = mDip[IND2(i1,i2,n1)]; //u_x
        u1i = sqrt(1.f - u2i*u2i);  //u_z

        a11 = 1.f - u1i*u1i;
        a22 = 1.f - u2i*u2i;
        a12 =      -u1i*u2i;

        fa  = in[IND2(i1  ,i2,n1)]  - in[IND2(i1-1,i2-1,n1)];
        fb  = in[IND2(i1-1,i2,n1)]  - in[IND2(i1  ,i2-1,n1)];
        f1  = 0.5f*(fa-fb); f2  = 0.5f*(fa+fb);

        g1  = a11*f1+a12*f2;
        g2  = a12*f1+a22*f2;
        ga  = 0.5f*(g1+g2); gb  = 0.5f*(g1-g2);
        out[IND2(i1   ,i2  ,n1)]   = ga;  //???? += or = ??
        out[IND2(i1-1 ,i2-1,n1)]  -= ga;
        out[IND2(i1-1 ,i2  ,n1)]  -= gb;
        out[IND2(i1   ,i2-1,n1)]  += gb;
      }
    }
  } else if (this->ndim == 3) {  //!!!! NOT FINISHED !!!!
    n3=ns[2];
  } else VERIFY(0);
  return 0;
}

int HaleDipFltOp::ATAflt(float* out, float* in, long nd) {
  VERIFY(nd == datShape.get_n123());
  int n1,n2,n3 = 1;
  stdVecInt ns = datShape.get_ns();
  n1=ns[0];n2=ns[1]; //PRVEC(ns,2);

  //add the damping here
  if (fabs(eps)<epsilon) {
  } else { //add some damping to this operator from Dx ==> (D+eps*I)x
    for (long i=0; i<nd; i++) {
      out[i] += in[i]*(1.f+eps);
    }
  }

  if (this->ndim == 2) {
    float u2i,u1i,a11,a12,a22,f1,f2,g1,g2;
    for (int i2=1; i2<n2; i2++) {
      for (int i1=1; i1<n1; i1++) {
        u2i = mDip[IND2(i1,i2,n1)]; //u_x
        u1i = sqrt(1.f - u2i*u2i);  //u_z

        a11 = 1.f - u1i*u1i;
        a22 = 1.f - u2i*u2i;
        a12 =      -u1i*u2i;

        f1  = in[IND2(i1  ,i2,n1)]  - in[IND2(i1-1,i2   ,n1)];
        f2  = in[IND2(i1  ,i2,n1)]  - in[IND2(i1  ,i2-1 ,n1)];
        g1  = a11*f1+a12*f2;
        g2  = a12*f1+a22*f2;
        out[IND2(i1   ,i2  ,n1)]  +=  (g1+g2);  
        out[IND2(i1-1 ,i2  ,n1)]  -=  g1;
        out[IND2(i1   ,i2-1,n1)]  -=  g2;
      }
    }
  } else if (this->ndim == 3) {  
    n3=ns[2];
    long ind;
    float u3i,u2i,u1i,a11,a12,a13,a23,a22,a33,f1,f2,f3,g1,g2,g3;
    for (int i3=1; i3<n3; i3++) {
      for (int i2=1; i2<n2; i2++) {
        for (int i1=1; i1<n1; i1++) {
          ind = IND3(i1,i2,i3,n1,n2)*2;
          u2i = mDip[ind]   ;       //u_x
          u3i = mDip[ind+1] ;       //u_y
          u1i = sqrt(1.f-u2i*u2i-u3i*u3i);  //u_z
  
          a11 = 1.f - u1i*u1i;
          a22 = 1.f - u2i*u2i;
          a33 = 1.f - u3i*u3i;
          a12 =      -u1i*u2i;
          a23 =      -u2i*u3i;
          a13 =      -u1i*u3i;

          f1  = in[IND3(i1,i2, i3, n1,n2)]- in[IND3(i1-1,i2,i3,n1,n2)];
          f2  = in[IND3(i1,i2,i3,n1,n2)]  - in[IND3(i1,i2-1,i3,n1,n2)];
          f3  = in[IND3(i1,i2,i3,n1,n2)]  - in[IND3(i1,i2,i3-1,n1,n2)];

          g1  = a11*f1+a12*f2+a13*f3;
          g2  = a12*f1+a22*f2+a23*f3;
          g3  = a13*f1+a23*f2+a33*f3;
          out[IND3(i1   ,i2   ,i3   ,n1,n2)]  += (g1+g2+g3);
          out[IND3(i1-1 ,i2   ,i3   ,n1,n2)]  -= g1;
          out[IND3(i1   ,i2-1 ,i3   ,n1,n2)]  -= g2;
          out[IND3(i1   ,i2   ,i3-1 ,n1,n2)]  -= g3;
        }
      }
    }
  } else VERIFY(0);
  return 0;
}

//Symmetrical operator
int HaleDipFltOp::oper(bool bAdj, bool bAdd, float* mod, float* dat) {
  //PRV(this->repeat);
  long nd = datShape.get_n123();
  VERIFY(bAdd == false); //Does NOT support bAdd=false;
  sf_adjnull(bAdj,bAdd,nd,nd,mod,dat);
  
  float *in, *out;
  if (!bAdj) { in = mod; out = dat;}
  else { in = dat; out = mod;}
  
  valarray<float> tt;
  float *t_in,*t_out;
  t_in = in; t_out = out;
  if (repeat>1) {
    tt.resize(nd);
    tt = 0.0f;
    valarray_assign(tt,in);
    t_in = va_ptr(tt);
    t_out= out;
  }
  
  for (int irep=0; irep<repeat; irep++) {
    if (this->fltType == HDFT_dirLap) {
      ATAflt    (t_out,t_in,nd);
    } else if (this->fltType == HDFT_aniDif1) {
      aniDifFlt1(t_out,t_in,nd);
    }
    std::swap(t_in,t_out);
  }
  //Now the result stores in t_in
  if (t_in != out) { //do the copy
    myMemcpy(out,t_in,nd);
  }
  return 0;
}

//apply anisotropic diffusion filter, use only simple finite difference gradient calculation.
int HaleDipFltOp::aniDifFlt1(float* out, float* in, long nd) {
  //Set up the matrix on the left (oper), and set up the right hand side, then throw it into to the CG solver.
  VERIFY(nd == datShape.get_n123());
  int n1,n2,n3 = 1;
  stdVecInt ns = datShape.get_ns();
  n1=ns[0];n2=ns[1]; //PRVEC(ns,2);
  
  float eps   = this->eps;
  float alpha = this->alpha;

  if (fabs(eps)<epsilon) {
    myMemcpy(out,in,nd);
  } else { //add some damping to this operator from Dx ==> (D+eps*I)x
    for (long i=0; i<nd; i++) {
      out[i] = in[i]*(1.f+eps);
    }
  }

  if (this->ndim == 2) {
    float u2i,u1i,a11,a12,a22,f1,f2,g1,g2;
    for (int i2=1; i2<n2; i2++) {
      for (int i1=1; i1<n1; i1++) {
        u2i = mDip[IND2(i1,i2,n1)]; //u_x
        u1i = sqrt(1.f - u2i*u2i);  //u_z

        a11 = 1.f - u1i*u1i;
        a22 = 1.f - u2i*u2i;
        a12 =     - u1i*u2i;  //D matrix (when lamda_u=1, lamda_v=0) == A matrix 

        f1  = in[IND2(i1  ,i2,n1)]  - in[IND2(i1-1,i2   ,n1)];
        f2  = in[IND2(i1  ,i2,n1)]  - in[IND2(i1  ,i2-1 ,n1)];
        g1  = a11*f1+a12*f2;
        g2  = a12*f1+a22*f2;
        g1 *= alpha; g2 *= alpha;
        out[IND2(i1   ,i2  ,n1)]  +=  (g1+g2);  
        out[IND2(i1-1 ,i2  ,n1)]  -=  g1;
        out[IND2(i1   ,i2-1,n1)]  -=  g2;
      }
    }
  } else if (this->ndim == 3) {  
    n3=ns[2];
    long ind;
    float u3i,u2i,u1i,a11,a12,a13,a23,a22,a33,f1,f2,f3,g1,g2,g3;
    for (int i3=1; i3<n3; i3++) {
      for (int i2=1; i2<n2; i2++) {
        for (int i1=1; i1<n1; i1++) {
          ind = IND3(i1,i2,i3,n1,n2)*2;
          u2i = mDip[ind]   ;       //u_x
          u3i = mDip[ind+1] ;       //u_y
          u1i = sqrt(1.f-u2i*u2i-u3i*u3i);  //u_z
  
          a11 = 1.f - u1i*u1i;
          a22 = 1.f - u2i*u2i;
          a33 = 1.f - u3i*u3i;
          a12 =     - u1i*u2i;
          a23 =     - u2i*u3i;
          a13 =     - u1i*u3i;

          f1  = in[IND3(i1,i2,i3,n1,n2)]  - in[IND3(i1-1,i2,i3,n1,n2)];
          f2  = in[IND3(i1,i2,i3,n1,n2)]  - in[IND3(i1,i2-1,i3,n1,n2)];
          f3  = in[IND3(i1,i2,i3,n1,n2)]  - in[IND3(i1,i2,i3-1,n1,n2)];

          g1  = a11*f1+a12*f2+a13*f3;
          g2  = a12*f1+a22*f2+a23*f3;
          g3  = a13*f1+a23*f2+a33*f3;
          g1 *= alpha; g2 *= alpha; g3 *= alpha;
          out[IND3(i1   ,i2   ,i3   ,n1,n2)]  += (g1+g2+g3);
          out[IND3(i1-1 ,i2   ,i3   ,n1,n2)]  -= g1;
          out[IND3(i1   ,i2-1 ,i3   ,n1,n2)]  -= g2;
          out[IND3(i1   ,i2   ,i3-1 ,n1,n2)]  -= g3;
        }
      }
    }
  } else VERIFY(0);
  return 0;
}

