#ifndef __HALEDIPFLTOP_H_INCL__
#define __HALEDIPFLTOP_H_INCL__

#include "op.h"
/**
 * Dip filter Dave Hale's first method (Directional Laplacian)
 *                ^  Y axis
 *                |
 *                |
 *                |
 *                |
 *                ---------- > X axis
 *               /
 *              /
 *             /
 *            /
 *          <  Z axis 
 * @author   zyang03
 */
namespace Yang {

  enum HaleDipFltType {
    HDFT_dirLap,  //Directional Laplacian
    HDFT_aniDif1 //Anisotropic diffusion
  };

  class HaleDipFltOp:public op {
public:
  // Constructor
  HaleDipFltOp();

  // Destructor
  virtual ~HaleDipFltOp();

  // Copy constructor
  // TODO: Uncomment the copy constructor when you need it.
  //HaleDipFltOp(const HaleDipFltOp& src);

  // Assignment operator
  // TODO: Uncomment the assignment operator when you need it.
  //HaleDipFltOp& operator=(const HaleDipFltOp& src);

  int buildDipFilter(float* dipfield, const hypercube& shapeDip);
  void setSmoothStrength(float alpha_, HaleDipFltType type, float eps=0.0f, int repeat_=1)  {
    this->alpha = alpha_; this->fltType = type; this->eps = eps; this->repeat = repeat_;
    //PRV(this->repeat); PRV(repeat_);
    VERIFY(repeat_>0);
    }
  //FLT using directional laplacian
  int ATAflt  (float* out, float* in, long nd);
  int ATAfltH (float* out, float* in, long nd);

  int aniDifFlt1(float* out, float* in, long nd);

  virtual int oper(bool bAdj, bool bAdd, float* mod, float* dat);
public:
  int   ndim;
  valarray<float> mDip;
  HaleDipFltType fltType;
  float alpha;
  float eps; //for damping, to stablize CG inversion
  int repeat;
};

// Constructor implementation
inline HaleDipFltOp::HaleDipFltOp() : op()
{
}

// Destructor implementation
inline HaleDipFltOp::~HaleDipFltOp()
{
}

int invFltCG    (Yang::op& fltOp, float* out, float* in, long nd, float* x0, int maxIter, float* eps, float fTol);
int invFltCG_prec (Yang::op& fltOp,Yang::op& S, float* out, float* in, long nd, float* x0, int maxIter, float* eps, float fTol);

}



#endif // __HALEDIPFLTOP_H_INCL__
