#include "stdafx.h"
#include "cubeData.hpp"
#include "LUsolver.h"
#include <math.h>
#define TINY 1.0e-12

using namespace Yang;
//forward/backward substitution
//mat is a nxn matrix that contains the LU decomposition, b is the nx1 matrix
//NOTICE: matrix is stored by row first, NOT column first. means that the elment m(r,c) is at r row, c column, the linear memeory address is r*ncol+c;
int lubksb(cubeData<float>& a,int n, int *indx, float* b)
{
	int i,ii=-1,ip,j;
	float sum;

	for (i=0;i<n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii>=0)	{
			for (j=ii;j<=i-1;j++) {sum -= a(i,j)*b[j];};
		}		
		else {
      if (fabs(sum)>TINY) { ii=i; }
    }
		b[i]=sum;
	}
	for (i=n-1;i>=0;i--) {
		sum=b[i];
		for (j=i+1;j<n;j++) {sum -= a(i,j)*b[j];}
		b[i]=sum/a(i,i);
	}
  return 0;
}

//d could be +1 or -1.
int ludcmp(cubeData<float>& a, int n, int *indx, float& d)
{
	int i,imax,j,k;
	float big,dum,sum,temp;
	stdVecFloat vv;

	vv.resize(n);
	d=1.0f;

	for (i=0;i<n;i++) {
		big=0.0;
		for (j=0;j<n;j++)
			if ((temp=fabs(a(i,j))) > big) big=temp;
		if (fabsf(big)<1e-12) {errPrt("Singular matrix in routine ludcmp\n");}
		vv[i]=1.0/big;
	}
	for (j=0;j<n;j++) {
		for (i=0;i<j;i++) {
			sum=a(i,j);
			for (k=0;k<i;k++) {sum -= a(i,k)*a(k,j);}
			a(i,j)=sum;
		}
		big=0.0;
		for (i=j;i<n;i++) {
			sum=a(i,j);
			for (k=0;k<j;k++) {
				sum -= a(i,k)*a(k,j);
      }
			a(i,j)=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=0;k<n;k++) {
				dum=a(imax,k);
				a(imax,k)=a(j,k);
				a(j,k)=dum;
			}
			d = -(d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (fabsf(a(j,j))<TINY) {a(j,j)=TINY;}
		if (j != n-1) {
			dum=1.0/(a(j,j));
			for (i=j+1;i<n;i++) {a(i,j) *= dum;}
		}
	}
	return 0;
}

//solve Ax = b, result is stored in b
int luSolve(cubeData<float>& a, int n, float* b) {
	stdVecInt indx(n);
	float d; d=0.0f;
	ludcmp(a, n, &indx[0], d);
	lubksb(a,n,  &indx[0], b);
	return 0;
}

#undef TINY
