#include "stdafx.h"
#include "rayTracer.h"
#include "RayTomoOp.h"
#include "waveOpParams.h"
#include "misc.h"

#include "myDebug.h"

using namespace Yang;
//Compute the raypath based on the information, suppose the angle increases when rotate counter-clockwise, because +z is downward, and +x is upward. 0 deg corresponds to +z axis.
void RayTomoOp::traceOrginPtInfoIs(const vector<fwk::degree>& vAng, const float* coords, bool bUpward, const fwk::degree* dip) {
  //const hypercube& vShp = acqPar_->vmodShape();
  //const vecAxis& axes = vShp.get_axes();
  //int ndim = axes.size();
  vInciAngle_ = vAng;
  //first find rays originating from the subsurf point at dips, 
  float fdip = NULL==dip? 0.0f:*dip;
  vector<fwk::degree> vecAngles(vAng.size());
  if (bUpward) { //should shoot upward
    fdip += 180.0f;
  } else { //shoot downward
  }
  for (int ig=0; ig<vAng.size(); ig++) {
    vecAngles[ig] = vAng[ig] + fdip;
  }
  rayTracer_->rayPath(rays_,coords,vecAngles); //this would save the raypaths to rays_ arrays.
  //PRV(*dip);PRVEC(coords,2); 
  //for (int ig=0; ig<vAng.size(); ig++) {PRV(vecAngles[ig]); prtStdVec(rays_[ig]);}
  return;
}

void RayTomoOp::acqCoverageMaskForCurSubsurfPt(vector<float>& mask, const vector<fwk::degree>& vADCIGang) {
  //For each angle in vADCIGang, the + and - angle needs to be measured in order to determine both src and receiver locations.
  //int ng = rays_.size();
  int ngADCIG = vADCIGang.size(); mask.resize(ngADCIG);
  int nRayAngle = vInciAngle_.size();
  float tol = 0.2f; 
  int igS,igR; float gamma;
  //assume the vADCIGang is in ascending order!
  for (int ig=0;ig<ngADCIG;ig++) {
    gamma = vADCIGang[ig];
    igS = findFirst(vInciAngle_.begin(),vInciAngle_.end(),gamma,tol) - vInciAngle_.begin();
    igR = findFirst(vInciAngle_.begin(),vInciAngle_.end(),-gamma,tol) - vInciAngle_.begin();
    if (igS >= nRayAngle || igR >= nRayAngle) {
      prtStdVec(vInciAngle_);
      PRT_LOC; errPrt("The corresponding ray_pair cannot be found for ADCIG angle at %f, check your initial call with traceOrginPtInfoIs!!\n",gamma);
    }
    //PRV(rays_.size()); PRV2(igS,igR);
    //Then use the acqusitionPar to determine
    mask[ig] = acqPar_->bSrcRecvLocCovered(rays_[igS],rays_[igR]);
  }
}



