#include "stdafx.h"
#include "op.h"
#include "cg_solv.h"

using namespace sf_yang;
using namespace Yang;
using namespace std;
// Implement the simple Conjugate gradient solver, based on the algo descrip on wikipedia, which is consistent
// with Sergey's CG tutoral SEP-report
//Build valarray objects to facilitate the calculation on arrays


//The actual number of iterations is the size of 'err' minus -1
int cg_solv::solve(float* mod, const float* mod0, const float* dat, op& L, 
            vector<float>& err, int maxIter, float fTol, bool bReport) {
  valarray<float> b;     //init input data
  valarray<float> r0,r_k; //residual                          
  valarray<float> p_k,p_k1,g; //search direct
  valarray<float> x0,x_k; //search direct
  valarray<float> tmp_d,tmp_m;

  long modsize,datsize;
  L.getSize(modsize,datsize);
  x0.resize(modsize); x_k.resize(modsize); p_k.resize(modsize);
  b.resize(datsize);  r0.resize(datsize); r_k.resize(datsize); 
  tmp_m.resize(modsize); tmp_d.resize(datsize);

  //Copy the input into valarray
  if (NULL!=mod0) valarray_assign(x0,mod);
  if (NULL!=dat)  valarray_assign(b,dat);

  /// CG loop
  float alpha,beta; float gdg,gdg1,f_r,f_r0;

  err.clear();
  //initial residual
  L.oper(false,false,&(x0[0]),&(tmp_d[0]));
  r0 = b - tmp_d;
  L.oper(true,false,&(g[0]),&(r0[0]));    //search direct is the gradient
  p_k = g;
  x_k = x0; r_k = r0;
  gdg = gdg1 = (g*g).sum();
  f_r0 = (r0*r0).sum(); err.push_back(f_r0);
  bool bRestart = false;
  if (bReport) {
    prt("%-4s %-12s  %-12s ", "iter", "error", "grad\n" ); //the first iter
    prt("\r %-4d %-8.3g ", 0,f_r0);
  }
  for (int iter=0; iter< maxIter; iter++) {
    if (bRestart) { //restart the cg process to counter numerical error thing
      L.oper(false,false,&(x_k[0]),&(tmp_d[0]));
      r_k = b - tmp_d;
      L.oper(true,false,&(g[0]),&(r_k[0]));    //search direct is the gradient
      p_k = g;
      bRestart =false;
      gdg   = (g*g).sum();
    } else {
      gdg   = gdg1; //use the one already calculated in the previous step
    }

    L.oper(false,false,&(p_k[0]),&(tmp_d[0])); //remap dm into data space;
    alpha = gdg / ((tmp_d*tmp_d).sum());

    //update the model & residual
    x_k += alpha*p_k;
    r_k -= alpha*tmp_d;

    f_r = ((r_k*r_k).sum());
    err.push_back(f_r);

    if (bReport) { //print the solver's internal information
      //this->report();
      prt("\r %-4d %-8.3g  %-8.3g ", iter,f_r,gdg );
    }
    if (f_r < f_r0*fTol ||  gdg< epsilon ) { break; }
         
    //calc the gradient, correspond to the rk+1 in the CG algo
    L.oper(true,false,&(g[0]),&(r_k[0]));
    gdg1  = (g*g).sum(); 
    beta  = gdg1/gdg;
    p_k   = g + beta*p_k;
  }
  //Now save the result into array mod
  valarray_assignr(mod,x_k);
  return 0;
}

//Make SURE that L operator needs to be symmetric posi-defi !
int cg_solv::solveSym(float* mod, const float* mod0, const float* dat, op& L, float* eps,
                      vector<float>& err, int maxIter, float fTol, bool bReport) {
  valarray<float> b;     //init input data
  valarray<float> r0,r_k; //residual                          
  valarray<float> p_k; //search direct
  valarray<float> x0,x_k; //updated solution vector
  valarray<float> tmpArr;
  
  long modsize,datsize;
  L.getSize(modsize,datsize); 
  VERIFY(modsize==datsize);
  x0.resize(modsize); x_k.resize(modsize); p_k.resize(modsize);
  b.resize(datsize);  r0.resize(datsize); r_k.resize(datsize); 
  tmpArr.resize(modsize);

  //Copy the input into valarray
  if (NULL!=mod0) valarray_assign(x0,mod0);
  VERIFY(dat); 
  valarray_assign(b,dat);

  bool bDamping = (eps!=NULL);
  /// CG loop
  float alpha,beta; float gdg,gdg1,f_r,f_r0;

  err.clear();
  //initial residual
  L.oper(false,false,&(x0[0]),&(tmpArr[0]));
  if (bDamping) {tmpArr += x0*(*eps);}

  r0 = b - tmpArr;
  p_k = r0;
  x_k = x0; r_k = r0;
  
  //PRVEC(x0,2); PRVEC(tmpArr,2); PRVEC(b,2);
  f_r0 = (r0*r0).sum(); err.push_back(f_r0);
  bool bRestart = false;

  if (bReport) {
    prt("%-4s %-12s  %-12s ", "iter", "error", "grad\n" ); //the first iter
    prt("\r%-4d %-8.3g \n", 0,f_r0);
  }
  for (int iter=0; iter< maxIter; iter++) {
    if (bRestart) { //restart the cg process to counter numerical error thing
      L.oper(false,false,&(x_k[0]),&(tmpArr[0]));
      if (bDamping) {tmpArr += x_k*(*eps);}
      r_k = b - tmpArr;
      p_k = r_k;
    }
    
    gdg = (r_k*r_k).sum();
    L.oper(false,false,&(p_k[0]),&(tmpArr[0])); //remap dm into data space;
    if (bDamping) {tmpArr += p_k*(*eps);}
    alpha = gdg / ((tmpArr*p_k).sum());
    //update the model & residual
    x_k += alpha*p_k;
    r_k -= alpha*tmpArr;

    f_r = ((r_k*r_k).sum());
    err.push_back(f_r);
    if (bReport) {
      prt("%-4d %-8.3g  %-8.3g \n", iter,f_r,gdg );
    }
    if (f_r < f_r0*fTol) {
      break; 
    }
    
    //calc the gradient, correspond to the rk+1 in the CG algo
    gdg1  = (r_k*r_k).sum(); 
    beta  = gdg1/gdg;
    p_k   = r_k + beta*p_k;
  }
  //Now save the result into array mod
  valarray_assignr(mod,x_k);
  return 0;
}

//I followed the peusdo code on wikipedia. in which M^-1 = S*S^{T}. S is the precondition operator that
//The epslion damping is NOT implemented yet!!!!!! ?????
int cg_solv::solveSym_prec(float* mod, const float* mod0, const float* dat, Yang::op& L, Yang::op& S, float* eps,
            vector<float>& err, int maxIter, float fTol, bool bReport) {
  valarray<float> b;     //init input data
  valarray<float> r0,r_k; //residual                          
  valarray<float> p_k; //search direct
  valarray<float> x0,x_k; //updated solution vector
  valarray<float> z0,z_k; //corresponding preconditioned model vector
  valarray<float> tmpArr;
  
  long modsize,datsize;
  L.getSize(modsize,datsize); 
  VERIFY(modsize==datsize);
  x0.resize(modsize); x_k.resize(modsize); p_k.resize(modsize); z0.resize(modsize); z_k.resize(modsize);
  b.resize(datsize);  r0.resize(datsize); r_k.resize(datsize);  
  tmpArr.resize(modsize);

  //Copy the input into valarray
  if (NULL!=mod0) valarray_assign(x0,mod0);
  VERIFY(dat); 
  valarray_assign(b,dat);

  bool bDamping = (eps!=NULL);
  /// CG loop
  float alpha,beta; float gdg,gdg1,f_r,f_r0;

  err.clear();
  //initial residual
  L.oper(false,false,&(x0[0]),&(tmpArr[0]));
  r0 = b - tmpArr;
  S.oper_AAT(va_ptr(r0),va_ptr(z0),datsize);
  p_k = z0;
  x_k = x0; r_k = r0; z_k = z0; //to bootstrap the loop
  //PRVEC(x0,2); PRVEC(tmpArr,2); PRVEC(b,2);
  f_r0 = (r0*r0).sum(); err.push_back(f_r0);
  bool bRestart = false;

  if (bReport) {
    prt("%-4s %-12s  %-12s ", "iter", "error", "grad\n" ); //the first iter
    prt("\r%-4d %-8.3g \n", 0,f_r0);
  }
  for (int iter=0; iter< maxIter; iter++) {
    if (bRestart) { //restart the cg process to counter numerical error thing
      L.oper(false,false,&(x_k[0]),&(tmpArr[0]));
      //if (bDamping) {tmpArr += x_k*(*eps);}
      r_k = b - tmpArr;
      S.oper_AAT(va_ptr(r_k),va_ptr(z_k),datsize);
      p_k = z_k;
    }
    
    gdg = (r_k*z_k).sum();
    L.oper(false,false,&(p_k[0]),&(tmpArr[0])); //remap dm into data space;
    //if (bDamping) {tmpArr += p_k*(*eps);}
    alpha = gdg / ((tmpArr*p_k).sum());
    //update the model & residual
    x_k += alpha*p_k;
    r_k -= alpha*tmpArr;

    f_r = ((r_k*r_k).sum());
    err.push_back(f_r);
    if (bReport) {
      prt("%-4d %-8.3g  %-8.3g \n", iter,f_r,gdg );
    }
    if (f_r < f_r0*fTol) {
      break; 
    }
    
    //calc the gradient, correspond to the rk+1 in the CG algo
    S.oper_AAT(va_ptr(r_k),va_ptr(z_k),datsize);  //Z_k+1 = M^{-1}r_{k+1}
    gdg1  = (r_k*z_k).sum(); 
    beta  = gdg1/gdg;
    p_k   = z_k + beta*p_k;
  }
  //Now save the result into array mod
  valarray_assignr(mod,x_k);
  return 0;
}

//Do a few precon, and then without precon
int cg_solv::solveSym_precmix(float* mod, const float* mod0, const float* dat, Yang::op& L, Yang::op& S, 
                    float* eps,vector<float>& err, int niterPrec, int maxIter, float fTol, bool bReport) { 
  long modsize = 0, datsize=0;
  L.getSize(modsize ,datsize);
  //First do with preconditioning
  valarray<float> mod1(modsize);
  solveSym_prec (va_ptr(mod1), mod0, dat, L, S, eps,
                err, niterPrec, fTol,bReport);
  stdVecFloat err2;
  solveSym      (mod, va_ptr(mod1), dat, L, eps,
                err2, maxIter-niterPrec, fTol,bReport);
  err.insert(err.end(),err2.begin(),err2.end());
  return 0;
}
