#include "stdafx.h"

//#include "srcWav.h"

#include "modelingFD2d_op.h"
#include "SR_manager.h"
#include "srcWav.h"
#include "interp.h"

using namespace Yang;


//Dummy function virtual
void SR_managerFD::init(waveOpParams* params)
{
   //SR_manager::getParam(par);
   m_pars = params;

   //Interp the wavelet to the desired sampling rate
   //m_pSrcWavelet->domain().info("src wavelet");
   
   //bool bTake2ndDeriv = m_pars->bTake2ndTimeDerivForRTM;
	
   //m_pSrcWavelet->InterpWavelet(m_pars->timeParam.dtComp,bTake2ndDeriv); //take 2nd derivative of wavelet
   //Take 2nd deriv on the data does not work as well as take 2nd deriv on the wfld. Although in theory they are the same.
   m_pSrcWavelet->InterpWavelet(m_pars->timeParam.dtComp,false); //not take 2nd derivative of wavelet   
}

//Just Read param do not do anything else
void SR_managerFD2d::init(waveOpParams* params)
{
   SR_managerFD::init(params);
   
  const hypercube& velGrid = params->velModelShape;
  velGrid.info("velGrid");
  int nx = velGrid.get_n(2);

	//Verify the actual number of shots
	int nMaxShots = ((nx-1)-params->mShot1_ix)/(params->mShotIncX) + 1;
	VERIFY (nMaxShots >= params->m_nShots)
	//Build the src location vector, here the coordinates is physical float number
	mVecSuperShotLoc.resize(params->m_nShots);  
	mVecShotLocPerSupershot.resize(1); //Here only one shot per supershot
  
  //LOG(mm2);
	for(int i=0; i<params->m_nShots; i++)
	 {

		mVecShotLocPerSupershot[0].sz = velGrid.get_coord(params->mShot1_iz,1);
		mVecShotLocPerSupershot[0].sx = velGrid.get_coord(params->mShot1_ix+params->mShotIncX*i,2);
		mVecSuperShotLoc[i] = mVecShotLocPerSupershot;
	 }
	
   //If modeling or AdjRTM, the seisRec is output, thus the shape can be determined from the modeling params
	//Verify receiver geometry, it is the CALLER's responsibility that the geometry described in params is the same as in seisRecord data file.
	//Need to build rec location vector at the begining of every shot !!!
	int nMaxTraces = (nx-1-params->mTrace1ix)/params->mTraceIncX + 1;
	
  //LOG(mm3);
	int trace1iz = params->mTrace1iz;
	VERIFY(trace1iz>=0);
	int nz = velGrid.get_n(1);
  //LOG(mm4);
	VERIFY(trace1iz+params->m_nPad < nz);
	if (params->m_nTraces> nMaxTraces)
	{
		 params->m_nTraces = nMaxTraces;
		 seperr("Due to Geometry, The total No. of Receivers cannot be greater than %d!\n",
				nMaxTraces); 
	}
	//irregular receiver not handled yet.
}

hypercube SR_managerFD::calcModelingDataSize(waveOpParams* pars) {
	  waveOpParams& params = *pars;
      //calc the size of seismic record
    float dtRec = params.timeParam.dtRecord;
   
    int n; float o,d;
    string label;
    o = 0.0; d = dtRec; n = params.timeParam.nRec;
    axis ax_t(n,o,d,"time","sec");
  
    hypercube& modelShape = params.modelGridShape;
 
    //Get the axis info on x direction of the vel model
    axis ax_tmp(modelShape.get_axis(2));
    d = ax_tmp.d * params.mTraceIncX;
    o = ax_tmp.o + params.mTrace1ix*ax_tmp.d; 
    n = params.m_nTraces;
    axis ax_x(n,o,d,"x","m");
  
    //Shot axis, this should represent the shot location
    n = params.m_nShots; 
    o = ax_tmp.o + params.mShot1_ix*ax_tmp.d;
    d = params.mShotIncX * ax_tmp.d;
    axis ax_shot(n,o,d,"shotX");
 
    std::vector<axis> axes;
    axes.push_back(ax_t); axes.push_back(ax_x); axes.push_back(ax_shot);
	  hypercube shapeSeisRec(axes);
	  return shapeSeisRec;
}

//Only works for transmmition case
void SR_managerFD2d::getVelModelMask(cubeData<float>& modelMask) {
   
   VERIFY(modelMask.domain().sameSize(this->m_pars->modelGridShape));
   modelMask.zero();
   //Based on the acquisition geometry, for 2D just define the four corners of the velocity.
   int ix1,ix2,iz1,iz2;
   ix1 = m_pars->mTrace1ix; ix2 = m_pars->mTrace1ix+m_pars->mTraceIncX*(m_pars->m_nTraces-1); 
   iz1 = m_pars->mShot1_iz; iz2 = m_pars->mTrace1iz;
   if (ix1>ix2) std::swap(ix1,ix2);
   if (iz1>iz2) std::swap(iz1,iz2);
   
   //prt("ix,iz %d,%d : %d,%d\n", ix1,ix2,iz1,iz2);
   //then set the modelMask.
   for (int i=ix1; i<ix2+1; i++) 
   {
		 for (int j=iz1;j<iz2+1;j++)
		 {
			modelMask(i,j) = 1.0;
		 }
   }
   return;
}

//Exert the point Src in the modeling program, iShot the shot index, "it" is the integer index of modeling time.[2]
void SR_managerFD2d::exertSrc( int iShot, int it, cubeData<ETYPE>& wavfield)
{
   if (it == -1) //This happens at the first timestep of modeling
   {
	  return;
   }
   //based on iShot,it and the wavelet, set the src's impact to the modeled wavefield
   //The wavefield, size is different from velocity cube
   cubeData<ETYPE>& pu = wavfield;

   //get the src wavelet
   cubeData<float> * pWavlet = m_pSrcWavelet->getWavelet();
   //assume it is 1d basically
   int len = pWavlet->domain().get_n(1);

   float dt = m_pars->timeParam.dtComp;
   
   //Should calculate this only once and store it.
   calcShotOffset(mVecShotOffsetPerSupershot,iShot);
   int shot_ix,shot_iz;
#define SRC_SPATIAL_EXTENT 0
   //float max_r = sqrtf(2.0*SQ(SRC_SPATIAL_EXTENT));
   //For each subshot in the supershot
   for (int shotsub=0; shotsub < int(mVecShotOffsetPerSupershot.size()); shotsub++) 
   {
	  //calculate the shot position
	  shot_ix = mVecShotOffsetPerSupershot[shotsub].sx + m_pars->offx;
	  shot_iz = mVecShotOffsetPerSupershot[shotsub].sz + m_pars->offz;
	  //Calculate the time, shot delay and shot amplitude that varies for
	  //each subshot needs to be considered. this case. 
	  if (it < len) 
	  {
		 float * p_src = pWavlet->pval;
		 float rr;
		 //Here I did not do the boundary check for array access.
		 //Spread the point src and tapering by the distance, to avoid the spatial aliasing.
		 for ( int j=shot_ix-SRC_SPATIAL_EXTENT;j<=shot_ix+SRC_SPATIAL_EXTENT;j++)
		 {
			for(int i=shot_iz-SRC_SPATIAL_EXTENT;i<=shot_iz+SRC_SPATIAL_EXTENT;i++)
			{
			   rr= sqrtf(SQ(j-shot_ix)+SQ(i-shot_iz));
			   //pu[j][i]=pu[j][i]+p_src[it]*.5*(cos(rr/max_r*pi)+1.0)*exp(-0.25*rr);
			   pu(j,i) += (p_src[it]*exp(-SQ(rr)))*dt;
			}
		 }
		 //pu[shot_ix][shot_iz] += 10*p_src[it];
	  }
   }
   return;
}

//Exert the point Src in the modeling program, iShot the shot index
//prepare the injected source, it is a cubeslice with the size nt by vel_nz. izVec gives the injection point's 
//Z coordinate for different ix.
void SR_managerFD2d::genSrcSlice( int iShot, cubeData<float>& srcSliceCube, cubeData<int>& izVec)
{
   //based on iShot,it and the wavelet, set the src's impact to the modeled wavefield
   //The wavefield, size is different from velocity cube
   const hypercube& velShape = m_pars->modelGridShape;
   int vel_nx; vel_nx = velShape.get_n(2);
   int vel_nz; vel_nz = velShape.get_n(1);
   int nt = m_pars->timeParam.Nt;
   hypercube sliceShape(nt,vel_nx); srcSliceCube.alloc(sliceShape); 
   hypercube shapeIzVec(vel_nx); izVec.alloc(shapeIzVec);

   srcSliceCube.zero();
   cubeData<ETYPE>& pu = srcSliceCube;
   
   izVec = 0;
   
   //get the src wavelet
   cubeData<float> * pWavlet = m_pSrcWavelet->getWavelet();
   //assume it is 1d basically
   int len = pWavlet->domain().get_n(1);

   //Should calculate this only once and store it.
   calcShotOffset(mVecShotOffsetPerSupershot,iShot);

   int shot_ix,shot_iz;
   float * p_src = pWavlet->pval;
   
   //For each subshot in the supershot
   for (int shotsub=0; shotsub < int(mVecShotOffsetPerSupershot.size()); shotsub++) 
   {
	  //calculate the shot position
	  shot_ix = mVecShotOffsetPerSupershot[shotsub].sx;
	  shot_iz = mVecShotOffsetPerSupershot[shotsub].sz;
    assert(shot_ix>=0 && shot_ix<vel_nx);
    assert(shot_iz>=0 && shot_iz<vel_nz);
    
	  //Calculate the time, shot delay and shot amplitude that varies for
	  //each subshot needs to be considered. this case. 
	  //Here I did not do the boundary check for array access.
	  for (int it=0; it < len; it++) {
      pu(shot_ix,it) += p_src[it];
	  }
	  izVec[shot_ix] = shot_iz;
   }
	
   //preserve the value of integration over dtCOmp
   srcSliceCube.scale(m_pars->timeParam.dtComp*1000);
   return;
}

//Exert the point Src in the modeling program, iShot the shot index
//prepare the injected receiver data, it is a cubeslice with the size nt by vel_nz. izVec gives the injection point's 
//Z coordinate for different ix.
void SR_managerFD2d::genRecvSlice( int iShot, cubeData<float>& recvSliceCube, cubeData<int>& recvIzVec)
{
   //prepare the injected source, it is a cubeslice with the size nt by vel_nz.
   const hypercube& velShape = m_pars->modelGridShape;
   int vel_nx; vel_nx = velShape.get_n(2);
   int nt = m_pars->timeParam.Nt;
   hypercube sliceShape(nt,vel_nx); recvSliceCube.alloc(sliceShape);
   hypercube shapeIzVec(vel_nx); recvIzVec.alloc(shapeIzVec); //notice it is integer elements.

   //based on iShot,it and the wavelet, set the src's impact to the modeled wavefield
   //The wavefield, size is different from velocity cube
   cubeData<ETYPE>& pu = recvSliceCube;
   recvSliceCube = 0.0f; recvIzVec = 0;
   
	//int ndim = m_pars->mShapeWavfield.get_ndim();
	//we need to know the receiver locations and the seismic signal
	
	 cubeData<float>& pRec = mInterpSeisCube;
   int trace1ix,trace1iz;
   const hypercube& shape = mInterpSeisCube.domain();
   const hypercube& shapeModel = m_pars->modelGridShape;
   axis axis_x1 = shape.get_axis(2);
   axis axis_x2 = shapeModel.get_axis(2);

   //Calculate the time and location of injection
   //Here we try to align the two grids, one is interped Receiver record, the other is
   //propagation grid, which the P field sits on.
   //Since the receiver interval is already interpolated as a subset of the vel grid
   //we just need to make the begining is correct, assume receive iz never changes

	
   assert(axis_x1.d == axis_x2.d);
   float d = axis_x1.d;
	
   trace1ix = lroundf((axis_x1.o-axis_x2.o)/d);
   assert(equalf(axis_x1.o,axis_x2.o+trace1ix*d));
   
	int nt2 = mInterpSeisCube.domain().get_n(1);   
	VERIFY(nt2 == recvSliceCube.domain().get_n(1));

	trace1iz = m_pars->mTrace1iz;
	recvIzVec = trace1iz;
	
	for (int ix=0; ix < axis_x1.n; ix++) {
			for (int it=0; it<nt; it++) //assume shape of InterpSeisCube is smaller than the velModel
			{
				pu(trace1ix+ix,it) = pRec(ix,it);
			}
	}
	return;
}

//Record the seismogram to file
//Need an dmRsf object for file output and a cubeData for actual memory storage [1]
//Need upgrade to handle irregular geometry.
void SR_managerFD2d::record(int iShot,int it, cubeData<ETYPE>& wavfield)
{
   //Always record at iz subscript = mTrace1iz, already 0 based
   //The wavefield, size is different from velocity cube
   cubeData<ETYPE>& pp    = wavfield;

   cubeData<float>& pRec  = mSeisCube;

   int iix;
   int   jtRec = m_pars->timeParam.jRec;
   if( it%jtRec == 0) //Record
   {
      int offx,offz;
	   offx = m_pars->offx;
      offz = m_pars->offz;
	  
      int tIndex = it/jtRec;
      //only do Land Recording Geometry by now, receiver does not move
      {
         for(iix=0;iix<mSeisCube.domain().get_n(2);iix++)
         {
            pRec(iix,tIndex) = pp(m_pars->mTrace1ix+iix*m_pars->mTraceIncX+offx,m_pars->mTrace1iz+offz);
         }
      }
      //else
      //   seperr("bRollAlong==True, not implemented yet!!");
   }
   return;
}

//For GPU modeling
void SR_managerFD2d::recordFromSnapshot(int iShot, cubeData<float>& snapshotCube) {
  int iix;
  //int jtRec = m_pars->timeParam.jRec;
  //Loop through the snapShotCube to grab the seismic record
  cubeData<float>& pRec = mSeisCube;
	
	float * pWf = snapshotCube.dataPtr();
	float * pp;
	int offx,offz;
	offx = m_pars->offx;
  offz = m_pars->offz;
	int nt = snapshotCube.domain().get_n(3);
	int nx = snapshotCube.domain().get_n(2);
	int nz = snapshotCube.domain().get_n(1);

	VERIFY(nt == mSeisCube.domain().get_n(1));
	
	for (int it=0; it < nt; it++ )
	  //only do Land Recording Geometry by now, receiver does not move
	  {
		 long long offset = it * nx * nz;
		 pp = pWf + offset;
		 for(iix=0;iix<mSeisCube.domain().get_n(2);iix++)
		 {
       pRec(iix,it) = pp[(m_pars->mTrace1ix+iix*m_pars->mTraceIncX+offx)*nz + m_pars->mTrace1iz+offz];
		 }
	  }
	  //else
	  //   seperr("bRollAlong==True, not implemented yet!!");
	  //mSeisCube.dump("dump1");
}

//For GPU modeling, cubeBuffer is an array of ntRec * nTrc
void SR_managerFD2d::recordFromGpuBuffer(int iShot, cubeData<float>& cubeBuffer) {

	//put cubeBuffer's content into the seismic record cube
	int ntRec = cubeBuffer.domain().get_n(1);
	assert( ntRec == mSeisCube.domain().get_n(1));
	int nTrc = cubeBuffer.domain().get_n(2);
	assert( nTrc == mSeisCube.domain().get_n(2));
	
	VERIFY(cubeBuffer.sameSize(mSeisCube));
	long long n = cubeBuffer.domain().get_n123();
	myMemcpy<float>(mSeisCube.dataPtr(),cubeBuffer.dataPtr(),n);
	return;
}

//For GPU modeling, cubeBuffer is an array of ntRec * nTrc
void SR_managerFD2d::genRecGeometryInfo(int iShot, int& nRecv, stdVecInt& iz_vec, stdVecInt& ix_vec, cubeData<float>& cubeBuffer) {

	//In this simple land acuqistion case, the size is easy to determin
	cubeBuffer.alloc(mSeisCube.domain());
	int nTrc = cubeBuffer.domain().get_n(2);
	nRecv = nTrc;
	
	iz_vec.resize(nTrc); 	ix_vec.resize(nTrc);
	for (int i=0; i<nTrc; i++) {
		ix_vec[i] = m_pars->mTrace1ix+i*m_pars->mTraceIncX;
		iz_vec[i] = m_pars->mTrace1iz;
	}
	return;
}

//Interpolate the receiver record to the sampling rate of computation
//For each shot, receiver record should be in memory.
void SR_managerFD2d::interpRecvRec(int ishot)
{
  //Associate with an incore dmRsf object
  
  //Do the interpolation
  m_pars->pRsfSeisRec->seq_grab(&mSeisCube);
  mInterpSeisCube.dealloc();
  
   float dtComp = m_pars->timeParam.dtComp;
   int Nt = m_pars->timeParam.Nt;
   
   //from it=0 => it=m_Nt,
   //NtRec*m_Nt is supposed to be the last coordinate value of the SeisRec cube
   
   //Spatial interpolation is necessary as well
   float o,d;
   o = mSeisCube.domain().get_o(2);
   d = m_pars->mShapeWavfield.get_d(2);
   //Notice that the x-direction it might not be the same size as of velocity grid  
      
   //assume the receiver sits exactly on the vel model grid
   seplib::interp::getInstance()->interpolate(mInterpSeisCube,mSeisCube,NULL,NULL,&dtComp,
                                     NULL,&o,&d,
                                     NULL,NULL,NULL);
   //after the size of mInterpSeisCube is decided, lets check the actual size of it
   VERIFY(mInterpSeisCube.domain().get_n(1) == Nt); // interped 0,1,...,Nt
}

void SR_managerFD2d::injectRecv(int ishot, int it, cubeData<ETYPE>& wavfield) //[2]
{
   if (it > m_pars->timeParam.Nt-1) //happens at the start of RTM
   { return; }
   
   switch(m_pars->mImposeRecvType) 
	  {
   case AsBoundary:
      //!!! !!! IMPLEMENT
      seperr("modelingFD_op::AsBoundary");
      break;
   case AsForce:
      //Same manner as exert Src
      injectRecvAsForce(ishot,it,wavfield);
      break;
   default:
      seperr("unknown ImposeRecvType");
   }
}

//calculate the actual Corresponding index in velocity cube of the physical shot location
void SR_managerFD2d::calcShotOffset(vector<point2dI>& vecShotLoc,int iShot)
{
	  //Get the index in vel grid
	  const hypercube& velGrid = m_pars->modelGridShape;
	  
	  // Implement Super shot gather HERE !!!!!
	  int numSubshots = mVecSuperShotLoc[iShot].size();
	  vecShotLoc.resize(numSubshots);
      //calculate the shot position
	  int shotix,shotiz;
	  for (int subshot=0; subshot < numSubshots ; subshot++) {
		 shotiz = velGrid.get_offset(mVecSuperShotLoc[iShot][subshot].sz,1);
		 shotix = velGrid.get_offset(mVecSuperShotLoc[iShot][subshot].sx,2);
		 vecShotLoc[subshot] = point2dI(shotiz,shotix);
	  }
}

void SR_managerFD2d::injectRecvAsForce(int iShot, int it, cubeData<ETYPE>& wavfield) //[2]
{
  //based on iShot,it and the wavelet, set the src's impact to the modeled wavefield
  //The wavefield, size is different from velocity cube
  cubeData<ETYPE>& pu = wavfield;
  
  //get the receiver record
  cubeData<float>& pRec = mInterpSeisCube;
  int trace1ix,trace1iz;
  const hypercube& shape = mInterpSeisCube.domain();
  const hypercube& shapeWfld = m_pars->mShapeWavfield;
  axis axis_x1 = shape.get_axis(2);
  axis axis_x2 = shapeWfld.get_axis(2);
  
  //Calculate the time and location of injection
  //Here we try to align the two grids, one is interped Receiver record, the other is
  //propagation grid, which the P field sits on.
  //Since the receiver interval is already interpolated as a subset of the vel grid
  //we just need to make the begining is correct, assume receive iz never changes
  trace1iz = m_pars->mTrace1iz + m_pars->offz;
    
  assert(axis_x1.d == axis_x2.d);
  float d = axis_x1.d;
  trace1ix = int((axis_x1.o-axis_x2.o)/d + 0.5);
  assert(equalf(axis_x1.o,axis_x2.o+trace1ix*d));
  
  float dt = m_pars->timeParam.dtComp;
  if (it < shape.get_n(1)) 
  {
    for (int j=0; j<axis_x1.n; j++) //assume shape of InterpSeisCube is smaller than the velModel
    {
       pu(trace1ix+j,trace1iz) += pRec(j,it)*dt;
    }
  }
  else //it should always be within range.
    assert(0);
  return;
}

//For modeling and adjModeling
//Init and alloc memories for a new shot
void SR_managerFD::initPerShot(int ishot, action_type ac)
{
   int ndim = m_pars->mShapeWavfield.get_ndim();
   //Notice here if changed to irregular geometry, the shape will be
   //more difficult to decide
   mSeisCube.alloc(m_pars->pRsfSeisRec->domain().return_compatible(ndim));
}

//Out put seis record to file, in modeling
void SR_managerFD::finishOneShot(int iShot)
{
	/*
	if (m_pars->bTake2ndTimeDerivForAdjRTM) {
		cubeData<float> tempC;
		mSeisCube.copyTo(tempC);
		timeDeriv2nd(mSeisCube,tempC,true);
	}
	*/
	m_pars->pRsfSeisRec->seq_update(&mSeisCube);
   mSeisCube.zero();
}
