#ifndef SR_MANGER_H
#define SR_MANGER_H 1

// #include "waveOps.h" !Cause recursive inclusion
#include "dmRsf.hpp"
#include "cubeData.hpp"
#include "point3d.hpp"
#include "waveOpParams.h"

namespace Yang {
   //Just declare the classes
   class modelingFD_op;
   class srcWav;
   struct waveOpParams;
   
   //Managing the source and receiver option
   class SR_manager
{
   public: //members
      cubeData<float> mSeisCube; //This cube stores just 1 shot of seismic record
      cubeData<float> mInterpSeisCube; //1 shot interpolated receiver cube
      
   public: //methods
      SR_manager()  {}
      virtual ~SR_manager() {
      }
      virtual void init(waveOpParams* params) = 0;
};

   class SR_managerFD : public SR_manager
{
   public:
      //data member
      srcWav * m_pSrcWavelet;
      waveOpParams* m_pars;
      
   protected:
      
   public: //methods
      //Associate the waveOps Object with SR_manager
      SR_managerFD(srcWav* pSrcWav): SR_manager(),
		 m_pSrcWavelet(pSrcWav) {
      }

      virtual void init(waveOpParams* params);
      void initPerShot(int ishot, action_type ac);
      //Exert the Src in to pWavfield
      virtual void exertSrc(int iShot, int it, cubeData<ETYPE>& wf) = 0;
      virtual void record(int iShot, int it, cubeData<ETYPE>& wf) = 0;

      void finishOneShot(int iShot);

	  virtual void interpRecvRec(int ishot) = 0;
      virtual void injectRecv(int ishot,int it, cubeData<ETYPE>& wavfield) = 0;
	  
	  hypercube calcModelingDataSize(waveOpParams* pars);
	  
	  virtual void getVelModelMask(cubeData<float>& modelMask) {}
	  
	  //For GPU modeling
	  virtual void recordFromSnapshot(int iShot, cubeData<float>& snapshotCube) = 0;
	  virtual void genSrcSlice( int iShot, cubeData<float>& srcSliceCube, cubeData<int>& izVec) = 0;
	  virtual void genRecvSlice( int iShot, cubeData<float>& srcSliceCube, cubeData<int>& izVec) = 0;

   protected:
      void injectRecvAsForce(int iShot, int it);
      
};

  class SR_managerFD2d : public SR_managerFD
{
  public: //data member
     //int mShot1_ix;
     //int mShot1_iz;
     //int mShotIncX;
     //int mTrace1ix,mTrace1iz,mTraceIncX,m_nTraces;
	 
  protected:
	 
	 //physical location / location on the simulation grid
	 
	 //current shot location on the simulation Grid.
	 vector<point2dF> mVecShotLocPerSupershot;
	 vector<point2dI> mVecShotOffsetPerSupershot;
	 
	 vector<point2dI> mVecRecvOffsetPerShot;
	 vector<point2dF> mVecRecvLocPerShot; //refresh every shot	 

	 vector< vector<point2dF> > mVecSuperShotLoc;
	 
  public: //method
     SR_managerFD2d(srcWav* pSrcWav): SR_managerFD(pSrcWav) {}
	 virtual ~SR_managerFD2d() {
	 }
	 
	 virtual void exertSrc(int iShot, int it, cubeData<ETYPE>& wf);
     virtual void record(int iShot, int it, cubeData<ETYPE>& wf);

     virtual void init(waveOpParams* params);
	 
     inline void calcShotOffset(vector<point2dI>& vecShotOffset, int iShot);
    
     virtual void injectRecv(int ishot,int it, cubeData<ETYPE>& wavfield);
	 
	 //For GPU modeling 2D
	 virtual void genSrcSlice( int iShot, cubeData<float>& srcSliceCube, cubeData<int>& izVec);
	 virtual void genRecvSlice( int iShot, cubeData<float>& srcSliceCube, cubeData<int>& izVec);
	 //For GPU 2d Modeling
	 virtual void recordFromSnapshot(int iShot, cubeData<float>& snapshotCube);
	 void recordFromGpuBuffer(int iShot, cubeData<float>& cubeBuffer);
	 void genRecGeometryInfo(int iShot, int& nRecv, stdVecInt& iz_vec, stdVecInt& ix_vec, cubeData<float>& cubeBuffer);

	 virtual void getVelModelMask(cubeData<float>& modelMask);
   
  protected:
	 virtual void interpRecvRec(int ishot);
     void injectRecvAsForce(int iShot, int it, cubeData<ETYPE>& wf);
	 
};


}
#endif
