#include "stdafx.h"

#include "velFD2d.h"
#include "modelingFD2d_op.h"
#include "boundaryFD.h"

using namespace Yang;
using namespace std;

void boundaryFD::initOperator(waveOpParams* par)
{
   m_pars = par;
   bFreeSurf = par->bFreeSurfaceOnTop;
   dt = par->timeParam.dtComp;
   m_bdType = par->m_bdType;
}

void boundaryFD2d::initOperator(waveOpParams* par)
{
   boundaryFD::initOperator(par);
      
  //This part should goes to per shot, since each shot the domain changes
  const hypercube& shapeVelModel = par->modelGridShape;
    
   nx = shapeVelModel.get_n(2);
   nz = shapeVelModel.get_n(1);
   dx = shapeVelModel.get_d(2); dz = shapeVelModel.get_d(1);
   dt = m_pars->timeParam.dtComp;
   
   dtDdx = dt/dx;
   dtDdz = dt/dz;
   
   offz = m_pars->offz;
   offx = m_pars->offx;
   
   //Get the velocity field
   vcub = &(par->pVel->get2dVelPtr());
}

void boundaryFD2dAbsorb::initOperator(waveOpParams* par) {
   boundaryFD2d::initOperator(par);
}

//Boundary applys to wfg.pp_cb[1]
void boundaryFD2dAbsorb::update(wavFieldGroup& wfg)
{
   //get the pointers to the wavefiled;
   cubeData<ETYPE> *pppDummy, *ppp1, *ppp2;
   wfg.getPtrs(&ppp1,&ppp2,&pppDummy);
   cubeData<ETYPE> &pp1=*ppp1, &pp2=*ppp2, &ppDummy=*pppDummy;
   
   const cubeData<float>& v = *(this->vcub);
   int ix,iz;
   //float **v is the velocity field
   ETYPE delp;
   //apply the repective boundary condition
	#define ABSORB_COEF_1D 1.2
   if(m_bdType == abs_ord1)
   {
	  //OpenMP Section Parrallelization?
	  #pragma omp parallel private(ix,iz,delp) 
	  {
		 //use 1 way wave equation for the boundary region
		 #pragma omp sections
		 {
		 //left and right
		 #pragma omp section
		 {
		 for(ix=offx-1;ix>=0;ix--)
		 {
			for(iz=offz;iz<offz+nz;iz++)
			{
			   delp = ABSORB_COEF_1D*v(0,iz-offz)*dtDdx;
			   pp2(ix,iz) = (1.-delp)/(1.+delp)*(pp1(ix,iz)-pp2(ix+1,iz)) + pp1(ix+1,iz);
			}
		 }
		 }
		 #pragma omp section
		 {
		 for(ix=nx+offx;ix<nx+2*offx;ix++)
		 {
			for(iz=offz;iz<offz+nz;iz++)
			{
			   delp = ABSORB_COEF_1D*v(nx-1,iz-offz)*dtDdx;
			   pp2(ix,iz) = (1.-delp)/(1.+delp)*(pp1(ix,iz)-pp2(ix-1,iz)) + pp1(ix-1,iz);
			}
		 }
		 }
		 
		 #pragma omp section
		 {
		 //bottom
		 for(ix=offx;ix<nx+offx;ix++)
		 {
			for(iz=offz+nz;iz<2*offz+nz;iz++)
			{
				  delp = ABSORB_COEF_1D*v(ix-offx,nz-1)*dtDdz;
				  pp2(ix,iz) = (1.-delp)/(1.+delp)*(pp1(ix,iz)-pp2(ix,iz-1)) + pp1(ix,iz-1);
			}
		 }
		 }
		 #pragma omp section
		 {
		 if (!bFreeSurf) //Do the top
		 {
			for(ix=offx;ix<nx+offx;ix++)
			{
			   for(iz=offz-1;iz>=0;iz--)
			   {
					 delp = ABSORB_COEF_1D*v(ix-offx,0)*dtDdz;
					 pp2(ix,iz) = (1.-delp)/(1.+delp)*(pp1(ix,iz)-pp2(ix,iz+1)) + pp1(ix,iz+1);
			   }
			}
		 }
		 else  //Free surface b.c, just mirror the wavefiled at the top iz=offz
		 {
			for(ix=offx;ix<nx+offx;ix++)
			{
			   for(iz=offz-1;iz>=0;iz--)
			   {
				  //pp2(ix,iz) = pp2[ix,2*offz-iz];
				  pp2(ix,iz) = 0.0f;
			   }
			}
		 }
		 }
		 }
	  } //end omp parallel
   }
   else if (abs_ord1par == m_bdType)
   {
	  //OpenMP Section Parrallelization, not as good as the previous one
	  #pragma omp parallel private(ix,iz,delp) 
	  {
		 //use 1 way wave equation for the boundary region
		 #pragma omp sections
		 {
		 //left and right
		 #pragma omp section
		 {
		 for(ix=offx-1;ix>=0;ix--)
		 {
			for(iz=offz;iz<offz+nz;iz++)
			{
			   delp = 1.2*v(0,iz-offz)*dtDdx;
			   pp2(ix,iz) = (1.-delp)*(pp1(ix,iz)) + delp * pp1(ix+1,iz);
			}
		 }
		 }
		 #pragma omp section
		 {
		 for(ix=nx+offx;ix<nx+2*offx;ix++)
		 {
			for(iz=offz;iz<offz+nz;iz++)
			{
			   delp = 1.2*v(nx-1,iz-offz)*dtDdx;
			   pp2(ix,iz) = (1.-delp)*(pp1(ix,iz)) + delp * pp1(ix-1,iz);
			}
		 }
		 }
		 
		 #pragma omp section
		 {
		 //bottom
		 for(ix=offx;ix<nx+offx;ix++)
		 {
			for(iz=offz+nz;iz<2*offz+nz;iz++)
			{
				  delp = 1.2*v(ix-offx,nz-1)*dtDdz;
				  pp2(ix,iz) = (1.-delp)*(pp1(ix,iz)) + delp* pp1(ix,iz-1);
			}
		 }
		 }
		 #pragma omp section
		 {
		 if (!bFreeSurf) //Do the top
		 {
			for(ix=offx;ix<nx+offx;ix++)
			{
			   for(iz=offz-1;iz>=0;iz--)
			   {
					 delp = 1.2*v(ix-offx,0)*dtDdz;
					 pp2(ix,iz) = (1.-delp)*(pp1(ix,iz)) + delp*pp1(ix,iz+1);
			   }
			}
		 }
		 else  //Free surface b.c, just mirror the wavefiled at the top iz=offz
		 {
			for(ix=offx;ix<nx+offx;ix++)
			{
			   for(iz=offz-1;iz>=0;iz--)
			   {
				  pp2(ix,iz) = pp2(ix,2*offz-iz);
			   }
			}
		 }
		 }
		 }
	  } //end omp parallel
   }
   else if (abs_ord2 == m_bdType || abs_null == m_bdType)
   {
   }
   else if (abs_Dave == m_bdType)
   {
	  //use Dave Hale's first order way wave equation for the boundary region, basically it
	  //estimates the wave propagation direction at each point of the boundary first 
	  //Here offx/offz cannot be zero !!!
	  //left and right
	  float in_vs,vs;
	  float v_val,vdt;
	  ETYPE invDt,invDz,invDx;
	  ETYPE dpdz,dpdx,dpdz2,dpdx2,dpdt,dpdt2;
	  float beta,gamma,cosa;
	  invDt = 1.0/dt;
	  invDz = 1.0/dz; invDx = 1.0/dx;
	  
	  //#pragma omp parallel private(ix,iz,v_val,vdt,vs,in_vs,dpdz,dpdx,dpdz2,dpdx2,dpdt,dpdt2,beta,gamma,cosa) 
	  {
	  //#pragma omp sections 
	  {
	  //#pragma omp section
	  { //Left boundary
	  for(ix=offx;ix>=1;ix--)
      {
         for(iz=offz;iz<offz+nz;iz++)
         {
  			v_val = v(0,iz-offz);
  			vdt = v_val*dt;
            vs = SQ(v_val);
			in_vs = 1.0f/vs;
			dpdz = (pp1(ix,iz+1) - pp1(ix,iz-1))*0.5*invDz;
			//dpdt = (pp2(ix,iz) - pp0(ix,iz))*0.5*invDt;
			dpdt = (pp2(ix,iz) - pp1(ix,iz))*invDt;
			dpdz2 = SQ(dpdz);
			dpdt2 = SQ(dpdt);
			if (in_vs*dpdt2>dpdz2)
			{
			  cosa = sqrt(1.0 - (dpdz2)*vs/dpdt2);
			}
			else cosa=0.0f;
			
			beta = dx/vdt*cosa;
			gamma = (1.0f-beta)/(1.0f+beta);
            pp2(ix-1,iz) = gamma*(pp2(ix,iz)-pp1(ix-1,iz)) + pp1(ix,iz);
         }
      }
	  }
	  
	  //#pragma omp section
	  {//right boundary
      for(ix=nx+offx-1;ix<nx+2*offx-1;ix++)
      {
         for(iz=offz;iz<offz+nz;iz++)
         {
			v_val = v(nx-1,iz-offz);
			vdt = v_val*dt;
            vs = SQ(v_val);
			in_vs = 1.0f/vs;
			dpdz = (pp1(ix,iz+1) - pp1(ix,iz-1))*0.5*invDz;
			
			//dpdt = (pp2(ix,iz) - pp0(ix,iz))*0.5*invDt;
			dpdt = (pp2(ix,iz) - pp1(ix,iz))*invDt;
			dpdz2 = SQ(dpdz);
			dpdt2 = SQ(dpdt);
			if (in_vs*dpdt2>dpdz2)
			{
			  cosa = sqrt(1.0 - (dpdz2)*vs/dpdt2);
			}
			else cosa=0.0f;
			
			beta = dx/vdt*cosa;
			gamma = (1.0f-beta)/(1.0f+beta);
            pp2(ix+1,iz) = gamma*(pp2(ix,iz)-pp1(ix+1,iz)) + pp1(ix,iz);
         }
      }
	  }
	  
	  //#pragma omp section
	  {
      //bottom
      for(ix=offx;ix<nx+offx;ix++)
      {
         for(iz=offz+nz-1;iz<2*offz+nz-1;iz++)
         {
			v_val = v(ix-offx,nz-1);
			vdt = v_val*dt;
      vs = SQ(v_val);
			in_vs = 1.0f/vs;
			dpdx = (pp1(ix+1,iz) - pp1(ix-1,iz))*0.5*invDx;
			//dpdt = (pp2(ix,iz) - pp0(ix,iz))*0.5*invDt;
			dpdt = (pp2(ix,iz) - pp1(ix,iz))*invDt;
			dpdx2 = SQ(dpdx);
			dpdt2 = SQ(dpdt);
			if (in_vs*dpdt2>dpdx2)
			{
			  cosa = sqrt(1.0 - (dpdx2)*vs/dpdt2);
			}
			else cosa=0.0f;
			
			beta = dz/vdt*cosa;
			gamma = (1.0f-beta)/(1.0f+beta);
            pp2(ix,iz+1) = gamma*(pp2(ix,iz)-pp1(ix,iz+1)) + pp1(ix,iz);
         }
      }
	  }
	  //#pragma omp section
	  {
      if (!bFreeSurf) //Do the top
      {
         for(ix=offx;ix<nx+offx;ix++)
         {
            for(iz=offz;iz>=1;iz--)
            {
			   v_val = v(ix-offx,0);
			   vdt = v_val*dt;
			   vs = SQ(v_val);
			   in_vs = 1.0f/vs;
			   dpdx = (pp1(ix+1,iz) - pp1(ix-1,iz))*0.5*invDx;
			   //dpdt = (pp2(ix,iz) - pp0(ix,iz))*0.5*invDt;
				dpdt = (pp2(ix,iz) - pp1(ix,iz))*invDt;
			   dpdx2 = SQ(dpdx);
			   dpdt2 = SQ(dpdt);
			   if (in_vs*dpdt2>dpdx2)
			   {
				 cosa = sqrt(1.0 - (dpdx2)*vs/dpdt2);
			   }
			   else cosa=0.0f;
			   
			   beta = dz/vdt*cosa;
			   gamma = (1.0f-beta)/(1.0f+beta);
			   pp2(ix,iz-1) = gamma*(pp2(ix,iz)-pp1(ix,iz-1)) + pp1(ix,iz);			   
            }
         }
      }
      else  //Free surface b.c, just mirror the wavefiled at the top iz=offz
      {
         for(ix=offx;ix<nx+offx;ix++)
         {
            for(iz=offz-1;iz>=0;iz--)
            {
               //pp2(ix,iz) = pp2[ix,2*offz-iz];
					pp2(ix,iz) = 0;
            }
         }
      }
	  }
	  } //end omp sections
	  } //end omp parallel
   }
   else {
	  seperr("unknown abs boundary type");
   }
}
