#ifndef CUBE_ALGO_H
#define CUBE_ALGO_H

#include "decl_fwd.h"
#include "cubeData.hpp"

namespace Yang {

//calc X(z)/A(z)=Y(z) 1D
//formula a0*y(n)=x(n)-(a1*y(n-1)+a2*y(n-2)+...+ak*y(n-k))
void polydiv(cubeData<float>& arrayY, const cubeData<float>& arrayX, const cubeData<float>& arrayA, int resultLen);

void crossCorr1d(cubeData<float>& out, const cubeData<float>& in1, const cubeData<float>& in2,
                 int startLag, int endLag, hypercube* pOutShape=NULL);
void crossCorr1dGeneral(cubeData<float>& out, const cubeData<float>& in1, const cubeData<float>& in2,
                        int startLag, int endLag, hypercube* pShapeOut=NULL);
void filter1D(cubeDataF& Io, const cubeDataF& Ii, const cubeDataF& filter, bool extend = false);

void timeDeriv2nd(cubeData<float>& resultCube, const cubeData<float>& trcCube, bool bScaleBy_dt);
void timeDeriv1st(cubeData<float>& resultCube, const cubeData<float>& trcCube, bool bScaleBy_dt);

void deriv2nd(cubeData<float>& resultCube, const cubeData<float>& inCube, int iAxis, bool bScaleBy_d);
void deriv1st(cubeData<float>& resultCube, const cubeData<float>& inCube, int iAxis, bool bScaleBy_d);

void spatialXderiv2nd(cubeData<float>& resultCube, const cubeData<float>& trcCube, bool bScaleBy_dx);
void spatialXderiv1st(cubeData<float>& resultCube, const cubeData<float>& trcCube, bool bScaleBy_dx);

void scaleSlice(cubeData<float>& outTrcCube, const cubeData<float>& trcCube, const cubeData<float>& scalercube);
void multiplySlice(cubeData<float>& outCube, const cubeData<float>& inCube, const cubeData<float>& sliceCube);
//More general case.
void multiplySubCube(cubeData<float>& outCube, const cubeData<float>& inCube, const cubeData<float>& sliceCube, int iAxisBeg, int iAxisEnd);

void scaleTrc(cubeData<float>& trcCube, const cubeData<float>& scalerCube);
void scaleTrc(cubeData<float>& outTrcCube, const cubeData<float>& trcCube, const cubeData<float>& scalerCube);

void shiftTrc(cubeData<float>& trcCube, const cubeData<int>& shiftCube);

void stackTrc(cubeData<float>& resultTrc, const cubeData<float>& trcCube);
void spreadTrc(cubeData<float>& resultCube, const cubeData<float>& inputTrc);

//Sum along time axis for each Trc
void sumAlongTimeTrc(cubeData<float>& resultCube, const cubeData<float>& trcCube);

///Sacle by 1/V(x)^3
void scaleVelGradient(cubeData<float>& inOut, const cubeData<float>& velCube);
void scaleSlowGradient(cubeData<float>& inOut, const cubeData<float>& velCube);
//Find the position of maximum value at each trc.
void findMaxValTimePerTrc(cubeData<float>& out, const cubeData<float>& C_cube);
void staticShiftPerTrc(cubeData<float>& out, const cubeData<float>& in, const cubeData<float>& shiftCube);
void staticShiftPerTrcArb(cubeData<float>& out, const cubeData<float>& in, const cubeData<float>& shiftCube);
void staticShiftPerTrcArbSinc(cubeData<float>& out, const cubeData<float>& in, const cubeData<float>& shiftCube);

hypercube shapeAfterStackAlongAxis(int iAxis, const hypercube& inShape);
void multAndStackAlongAxis2(cubeData<float>& outCube, const cubeData<float>& inCube, cubeData<float>* pMultiplyCube = NULL);
void multAndStackAlongAxis(cubeData<float>& outCube, const cubeData<float>& inCube, const PosInt& iAxis, cubeData<float>* pMultiplyCube = NULL);

//Common operators
void causalInt(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);

//Input regular data, and an array of irregular coord.
//output an array that contains the sampled value at these coord, use linear interpolation inside. if bExtend, use edge value for outliers
//if !bExtend, set the interped to be zero
void irregularSamp1d(cubeData<float>& out, const cubeData<float>& in, const cubeData<float>& coord, bool bAdj, bool bExtend=false, bool bAdd=false);
void linInterp1d(float* out, const stdVecFloat& outCoord, const float* in, const stdVecFloat& inCoord, bool bExtend = true);
void linInterp1dOper(float* dat, const stdVecFloat& datCoord, float* mod, const stdVecFloat& modCoord, bool bAdj, bool bAdd, bool bExtend=true);

//General opertators
//zero the output for operators.
template<class T>
inline void addNull(bool bAdj, bool bAdd, int nm, int nd, T* mod, T* dat) {
  if (!bAdd) { 
    if (!bAdj) {
      memset((void*)dat,0,sizeof(T)*nd);
    } else {
      memset((void*)mod,0,sizeof(T)*nm);
    }
  }
}

inline void addNull(cubeDataF& mod, cubeDataF& dat, bool bAdj, bool bAdd) {
  if (!bAdd) { 
    if (!bAdj) {
      dat.zero();
    } else {
      mod.zero();
    }
  }
}

//declare operator function pointer
typedef void (*opFunc)(cubeData<float>&, cubeData<float>&, bool, bool);


} //end namespace
#endif

