#ifndef CUBE_ALGO2_H
#define CUBE_ALGO2_H

#include "cubeData.hpp"
#include "decl_fwd.h"

namespace Yang {

///Invert a 2X2 matrix [ in[0]  in[1] ]  ==> (out[0]  out[1] )
///[ in[2]  in[3] ]  ==> (out[2]	 out[3] )
///Function returns the conditional number
inline float invert2X2(float *out, const float *in, float& determ) {
   determ = in[0]*in[3] - in[1]*in[2];
   float tmp = 1.0f/determ;
   out[0] = tmp * in[3];
   out[1] = tmp * -in[1];
   out[2] = tmp * -in[2];
   out[3] = tmp * in[0];
   //Calculate the conditional number, which is an indication of the linear dependency of the matrix's columns.
   tmp = in[0] + in[3];
   float cond, tmp2;
   tmp2 = tmp*tmp - 4 * determ;
   if (tmp2<= 0.0f) {
	  cond = 1.0f; //eigenvalues are complex conjugate compair.
   }
   else {
	  cond = (tmp+sqrtf(tmp2)) / (tmp-sqrtf(tmp2));
	  if (fabsf(cond) < 1.0f) cond = 1.0f / cond;
	  cond = fabsf(cond);
   }
   return cond;
}

  void sqrtCube(cubeData<float>& out, const cubeData<float>& input);
  
  float clipCubeVal(cubeData<float>& out, float fmax, float fmin,
                       bool bClipMax=true, bool bClipMin=true);
  //Elem by Elem divide, but be more careful about zeros.
  void buildThresholdMask(cubeData<float>& Mask, const cubeData<float>& input, float vThres , bool bUseAbsVal = true, bool bMarkHighValue=true);
  void divideWithCaveat(cubeData<float>& out, const cubeData<float>& in, cubeData<float>& divident,
                      float qt_mask, float qt_eps, bool bUseMask, cubeData<float>* pMask, bool bMaskInput = false);
  
  //Like Transpose, but more flexible.
  void shuffleAxis(cubeData<float>& out, const cubeData<float>& in,
                   const stdVecInt& reshape, const stdVecInt& axMapping,bool bStuffDummyAxis=false ,bool bOutMemAllocInside=true);

  void duplicateAlongAxis(cubeData<float>& out, const cubeData<float>& in, int iAxis);
  
  void deriv1stAtIndex(cubeData<float>& resultCube, const cubeData<float>& inCube, int iAxis, int index, bool bScaleBy_d);
  void deriv2ndAtIndex(cubeData<float>& resultCube, const cubeData<float>& inCube, int iAxis, int index, bool bScaleBy_d);
  
  void getGaussfunc(cubeDataF& gaussfunc,const axis& ax,float fhalfMaxWidth, float b=0.0f, int order=0);
  //maxRatio=0.5 mean half maximum width, 0.1 means 10th maximum width
  void getGaussfuncWithMaxWidth(const PosFloat& maxRatio, cubeDataF& gaussfunc,const axis& ax,float fMaxWidth, float b=0.0f, int order=0);

  void dotProdAlongAxis(cubeDataF& resultCube, const cubeDataF& inCube, const cubeDataF& dot2, int iAxis);
  
  ///Input cube and outcube must be <=3 dimension. Must has the same grid spacing. But the grid.n grid.o might be different.
  ///This function will first calculate the overlap of domain's of Input and output, and carve only that poriton to outcube.
  ///Note that The uncovered part in outcube will be kept original if bClearOutputFirst=false;
  template<typename T>
  void cropAlign(cubeData<T>& outCube, const cubeData<T>& inCube, bool bAdd=false, bool bClearOutputFirst=false) {
    FUNC_TRACE_BEG(cropAlign);
    if (bClearOutputFirst) {
      assert(!bAdd); //logically should not happen
      outCube.zero();
    }
    
  	const hypercube& shape1 = outCube.domain();
  	const hypercube& shape2 = inCube.domain();
  	int ndim = shape1.get_ndim();
  	assert(ndim == shape2.get_ndim());
  	assert(ndim <= 3);
  	vecAxis axes1,axes2;
  	axes1 = shape1.get_axes(); axes2 = shape2.get_axes();
  	
  	int n1,n2,n3; n2=1; n3=1;
  	float d1,d2,d3;
  	float o1beg,o2beg,o3beg;
  	int o1beg1n,o2beg1n,o3beg1n; int o1beg2n,o2beg2n,o3beg2n; //The start location for each cube
  	o2beg1n = o2beg2n = 0; o3beg1n = o3beg2n = 0;
  	//Calculate the range of overlap for axis 1
  	alignAxis(o1beg,n1,d1,axes1[0],axes2[0]); 
  	o1beg1n = lroundf((o1beg-axes1[0].o)/d1); o1beg2n = lroundf((o1beg-axes2[0].o)/d1);
  	if (ndim>1) {
  	   alignAxis(o2beg,n2,d2,axes1[1],axes2[1]);
  	   o2beg1n = lroundf((o2beg-axes1[1].o)/d2); o2beg2n = lroundf((o2beg-axes2[1].o)/d2);
  	}
  	if (ndim>2) {
  		alignAxis(o3beg,n3,d3,axes1[2],axes2[2]);
  		o3beg1n = lroundf((o3beg-axes1[2].o)/d3); o3beg2n = lroundf((o3beg-axes2[2].o)/d3);
  	 }
  
  	//PRV(o1beg1n);PRV(o1beg2n);PRV(o1beg);PRV(n1);PRV(d1);PRV(o2beg1n);PRV(o2beg2n);PRV(o2beg);PRV(n2);PRV(d2);
  	//PRV(o3beg1n);PRV(o3beg2n);PRV(o3beg);PRV(n3);PRV(d3);
  	
  	//Copy the values
  	T * __restrict p1; const T* __restrict p2;
         T *pOff1; const T* pOff2;
  	p1 = outCube.dataPtr(); p2 = inCube.dataPtr();
  	int off3_1,off3_2, off2_1,off2_2;
    
  	for (int i3=0; i3<n3; i3++) {
  	   off3_1 = o3beg1n+i3; off3_2 = o3beg2n+i3;
  	   for (int i2=0; i2<n2; i2++) {
  		  off2_1 = off3_1*axes1[1].n + o2beg1n+i2;
  		  off2_2 = off3_2*axes2[1].n + o2beg2n+i2;
        pOff1=p1+off2_1*axes1[0].n+o1beg1n;
        pOff2=p2+off2_2*axes2[0].n+o1beg2n;
        if (!bAdd) {
          memcpy(pOff1,pOff2,sizeof(T)*n1);
        } else {
          for (int i1=0; i1<n1; i1++) {
            *(pOff1+i1) += *(pOff2+i1);
          }
        }
  	   }
  	}
  	FUNC_TRACE_END()
  }
  int rectSmooth (cubeData<float>& cubIn, cubeData<float>& cubOut, int rect1, int rect2, int rect3, int nrepeat=1, bool bAbsval=false);
  void RtoC (      cubeData<cmplxf>& cubC, const cubeData<float>& cubR, const cubeData<float>& cubI);
  void CtoR (const cubeData<cmplxf>& cubC, cubeData<float>& cubR,       cubeData<float>& cubI);
  void cubeAbs (const cubeData<float>& cubi, cubeData<float>& o); //Take the absolute value

} //end namespace

#endif

