#ifndef CUBEDATA_HPP
#define CUBEDATA_HPP

//#include "cubePtr2d.hpp"
//#include "cubePtr3d.hpp"
#include "dmRsf.hpp"

namespace Yang {

//The in-memory dmRsf Object. It could be a purely memory object without attaching to any IO files.
//But it can be associated with INPUT/OUTPUT .H files.
   
template <typename T>
class cubeData: public dmRsf<T> 
{
public:
   //data members
   T *pval;    //The data pointer that contains the value array
   //cubePtr2d<T> mCubePtr2d;
   //cubePtr3d<T> mCubePtr3d;
   enum {max_dim = 8};
protected:
   //n variables for fast indexes
   int ns[max_dim];
   bool mbProxy; //if mbProxy, then it does not allocate memory itself, but just point to an address that is already been addressed
                 //Use it with cautious! Only throught explicit interface, can the user get a proxy object
 
   //tmp variable for random access
   cubeData<T>* m_pCubeDataTmp;
   long long    m_grabndCnter;
   
private:
   //ban the implicit copy and assignment
   cubeData( const cubeData<T>& other) : dmRsf<T>(string(),dm::OUT,string()) {seperr("error !!!! into cubeData copy constructor\n");}
   
   inline void init_ns() {
     for (int i=0; i<max_dim; i++) ns[i]=1;
     ns[0] = 0;
   }
public:
   cubeData(const string& itag, dm::typeIO _tIO, const string& szInfo, bool bNoDummy=false) : dmRsf<T>(itag,_tIO,szInfo,bNoDummy), pval(NULL), mbProxy(false)
   { init_ns();}
   cubeData() : dmRsf<T>(string(),dm::INOUT,string()), pval(NULL), mbProxy(false) {init_ns();}
   cubeData(const hypercube& shape) : dmRsf<T>(string(),dm::INOUT,string()), pval(NULL), mbProxy(false){
     init_ns();
     alloc(shape);
   }
   
	 //return the old shape.
  void reshape(const hypercube & newShape, hypercube* p_oldShape = NULL) {
	   FUNC_TRACE_BEG(cubeData::reshape)
	   VERIFY(this->dom.get_n123() == newShape.get_n123());
	   if (p_oldShape) {
	     *p_oldShape = this->dom;
	   }
	   
	   this->dom = newShape;
	   //Need to rebuild the pointers
		//int ndim = this->dom.get_ndim();
		// if(ndim == 2) //build 2d Ptr
		// {
		//	mCubePtr2d.gen2dPtr(this->dom,pval);
		// }
		// if(ndim ==3 ) //build 3d Ptr
		// {
		//	mCubePtr3d.gen3dPtr(&this->dom,pval);
		// }
		//Set the n1,n2... parameters
         stdVecInt vns = newShape.get_ns();
         VERIFY(vns.size() <= max_dim);
         std::copy(vns.begin(),vns.end(),&(this->ns[0]));
		FUNC_TRACE_END();
  }

   //assign a constant to cube
   cubeData& operator = (T val) {
	  FUNC_TRACE_BEG(cubeData::op = )
	  long long n = this->domain().get_n123();
	  long long i;
	  for(i=0;i<n;i++) {
		 pval[i] = val;
	  }
	  FUNC_TRACE_END(*this);
   }
   
//assign a cube to another, size has to agree
cubeData& operator = (const cubeData<T>& other) {
  FUNC_TRACE_BEG(cubeData::op = other cube);
  if (this != &other) {
    if (!(this->domain().sameSize(other.domain()))) {
          VERIFY(this->domain().get_n123() == other.domain().get_n123());
    }
    
    const T* pIn = other.dataPtr();
    if (this->pval != pIn) { //if point to the same memory, then do nothing, otherwise
          long long n = this->domain().get_n123();
          myMemcpy<T>(this->pval, pIn, n);
    }
  } //end if not assign to itself
  FUNC_TRACE_END(*this);
}
   
   virtual ~cubeData() {
	  if (NULL != pval) {
		 cubeData<T>::unInit();
	  }
   }
   
  T* dataPtr() {return pval;}

	 //srite(argument) historical reason not put const at the begining.
	 const T* dataPtr() const {return pval;}
	 
	 long long size() const {
	  return this->dom.get_n123();
	 }
   
   //explicit copy/assign: alloc with the newshape and copy content
   void copyTo(cubeData<T>& target) const {
    FUNC_TRACE_BEG(cubeData::copyTo)
	  long long n = this->domain().get_n123();
	  assert(0!=n);	  
	  target.alloc(this->domain());
	  //copy memory content
	  myMemcpy<T>(target.pval, this->pval, n);
	  FUNC_TRACE_END()
   }
   
  //Same as CopyTo, but only alloc memory and set shape
  void allocMemoryTo(cubeData<T>& target) const{
    assert(0!=this->domain().get_n123());	  
    target.alloc(this->domain());
  }
   
   //same as alloc function
   virtual void init(const hypercube* pDom, bool bRemoveDummyAxis = REMOVE_DUMMY_AXIS_DEF) {
   LOG("ff1");
	  FUNC_TRACE_BEG(cubeData::init)
	  if (this->m_bDummy) { //purely in-memory object
		 this->alloc(*pDom);
	  }
	  else {
		 
		 //deAllocate memory
		 dealloc();

		 //Do necessary IO work for file headers
		 dmRsf<T>::init(pDom,bRemoveDummyAxis);

		 _alloc(this->dom);

		 //About content
		 if(NULL==pDom)	//read from file
		 {
			dmRsf<T>::read_all(*this);
		 }
	  }
	  //if output, do nothing
	  FUNC_TRACE_END()
   }
   
   virtual void init(const hypercube* pDom, int num_axes) {
	  FUNC_TRACE_BEG("cubeData::init2");
	  if (this->m_bDummy) { //purely in-memory object
		 this->alloc(*pDom);
	  }
	  else {
    //deAllocate memory
    dealloc();
    
    //Do necessary IO work for file headers
    dmRsf<T>::init(pDom,num_axes);
    
    _alloc(this->dom);
    
    //About content
    if(NULL==pDom)	//read from file
    {
      dmRsf<T>::read_all(*this);
    }
	  }
	  //if output, do nothing
	  FUNC_TRACE_END();
   }
	 	    
   virtual void unInit()
   {
	  //if it is output, then write all to file
	  if((!this->m_bDummy) && this->isOutput()) {
		 dmRsf<T>::write_all(*this);
	  }
	  dmRsf<T>::unInit();
    dealloc();
   }
 
   //Dump to file with the tag specified
   void dump(string tag, bool bNoDummy=false) const
   {
	  FUNC_TRACE_BEG(cubeData::dump);
    if (!this->dom.isEmpty()) {
      dmRsf_oc<T> file(tag,dm::OUT,"",bNoDummy);
      file.init(&this->domain());
      file.seq_updateAll(this);
      file.unInit();
    }
	  FUNC_TRACE_END()
   }
   
	 //Load from file with the tag specified, if n_axis_desired=0, means let the hfile header info define the number of dimensions.
   void load(const string& tag, int n_axis_desired = NUM_AXIS_DEFAULT)
   {
	  FUNC_TRACE_BEG(cubeData::load);
		assert(this->isDummy()); //This cube cannot be associated a tag already.
	
		dmRsf_oc<T> file(tag,dm::IN,"",true);
    
    file.init(NULL,n_axis_desired);
    
		this->alloc(file.domain());
		file.seq_grabAll(this);
		file.unInit();
	  FUNC_TRACE_END()
   }
	 
   void alloc( const hypercube& shape_in) 
   {
     
     //FUNC_TRACE_BEG(cubeData::alloc)
     VERIFY(!mbProxy);
	   this->resetCounter();
      //if the newshape is same as original shape, then no need to reallocate memory again
      if (this->domain().sameSize(shape_in)) { 
            //dbPrt("warn:cubeData allocated more than once.\n");
			this->dom = shape_in;
            zero();
      }
      else {
		 //if it is being re-allocated with a different size and it is associated with a real file, then it 
		 //might be problematic 
		 if(!(this->domain().isEmpty()) && !(this->m_bDummy)) {
			dbPrt("!!!You tried reallocate the cube %s, it is problematic because it is associated with a file tag=%s",this->m_szInfo.c_str(),this->tag.c_str());
		 }
		 
     dealloc();
		 this->dom = shape_in;
		 _alloc(this->dom);
	  }
	  //FUNC_TRACE_END()
   }

  //Init to zero
  void zero()
  {
    long long n = this->dom.get_n123();
    assert(n>0); assert(NULL != pval);
    memset((void *)pval,0,sizeof(T)*n);
  }
    void dealloc()
   {
     if(!mbProxy) {
        SAFE_DELETE_ARRAY(this->pval);
     } else {
       this->pval = NULL;
     }
     this->dom.deallocate();
   }

    T maxval() const
    {     
     int length = this->dom.get_n123();   // establish size of array
     T max = pval[0];       // start with max = first element

     for(int i = 1; i<length; i++)
     {
          if(pval[i] > max)
                max = pval[i];
     }
     return max;             
    }
   
   T minval() const
   {
     int length = this->dom.get_n123();   // establish size of array
     T min = pval[0];       // start with max = first element

     for(int i = 1; i<length; i++)
     {
          if(pval[i] < min)
                min= pval[i];
     }
     return min;             
   }
   
   T maxvalAbs() const {     
	  long long length = this->dom.get_n123();   // establish size of array
     T max = std::abs(pval[0]);       // start with max = first element
	 T tabs;
     for(long long i = 1; i<length; i++)
     {
       tabs = std::abs(pval[i]);
       if(tabs> max) {
         max = tabs;
       }
     }
     return max;             
   }
   
  T minvalAbs() const {     
    long long length = this->dom.get_n123();   // establish size of array
    T min = std::abs(pval[0]);       // start with max = first element
    T tabs;
    for(long long i = 1; i<length; i++)
    {
       tabs = std::abs(pval[i]);
        if(tabs < min)
                min = tabs;
    }
    return min;             
  }
   
  inline long long index2d (int ix, int iz) const
  {
    //assert(this->dom.get_ndim() == 2);
    assert(ns[0]>0);
    assert(iz<ns[0]);assert(iz>=0);assert(ix<ns[1]);assert(ix>=0);
    return (ix*ns[0]+iz);
  }
  inline long long index3d (int iy, int ix, int iz) const 
  {
    //assert(this->dom.get_ndim() == 3);
    assert(iz<ns[0]);assert(ix<ns[1]);assert(iy<ns[2]);
    return ((iy*ns[1]+ix)*ns[0])+iz;
  }
  inline long long index4d (int i4, int iy, int ix, int iz) const
  {
    //assert(this->dom.get_ndim() == 4);
    assert(iz<ns[0]);assert(ix<ns[1]);assert(iy<ns[2]);assert(i4<ns[3]);
    return (((i4*ns[2]+iy)*ns[1]+ix)*ns[0]+iz);
  }
  inline long long index5d (int i5, int i4, int iy, int ix, int iz) const
  {
    assert(iz<ns[0]);assert(ix<ns[1]);assert(iy<ns[2]);assert(i4<ns[3]);assert(i5<ns[4]);
    return ((((i5*ns[3]+i4)*ns[2]+iy)*ns[1]+ix)*ns[0]+iz);
  }
  inline long long index6d (int i6, int i5, int i4, int iy, int ix, int iz) const
  {
    return (((((i6*ns[4]+i5)*ns[3]+i4)*ns[2]+iy)*ns[1]+ix)*ns[0]+iz);
  }
  inline long long index7d (int i7, int i6, int i5, int i4, int iy, int ix, int iz) const {
    return ((((((i7*ns[5]+i6)*ns[4]+i5)*ns[3]+i4)*ns[2]+iy)*ns[1]+ix)*ns[0]+iz);
  }
  inline long long index8d (int i8, int i7, int i6, int i5, int i4, int iy, int ix, int iz) const {
    return (((((((i8*ns[6]+i7)*ns[5]+i6)*ns[4]+i5)*ns[3]+i4)*ns[2]+iy)*ns[1]+ix)*ns[0]+iz);
  }

  //Reload the [] operator, just in 1d sense
  inline T& operator[] (long long index)
  {
    assert(index < this->dom.get_n123());
    return pval[index];
  }
  
  inline const T& operator[] (long long index) const
  {
    assert(index < this->dom.get_n123());
    return pval[index];
  }

  inline T& operator() (int iz)
  {
    assert(iz<ns[0]);
    return pval[iz];
  }
  inline const T& operator() (int iz) const
  {
    assert(iz<ns[0]);
    return pval[iz];
  }
  
  inline T& operator() (int ix, int iz)
  {
    return pval[this->index2d(ix,iz)];
  }
  inline const T& operator() (int ix, int iz) const
  {
    return pval[this->index2d(ix,iz)];
  }
  inline T& operator() (int iy, int ix, int iz)
  {
  return pval[this->index3d(iy, ix,iz)];
  }
  inline const T& operator() (int iy, int ix, int iz) const 
  {
  return pval[this->index3d(iy, ix,iz)];
  }   
  inline T& operator() (int i4, int iy, int ix, int iz)
  {
  return pval[this->index4d(i4,iy,ix,iz)];
  }
  inline const T& operator() (int i4, int iy, int ix, int iz) const
  {
  return pval[this->index4d(i4,iy,ix,iz)];
  }
  inline T& operator() (int i5, int i4, int iy, int ix, int iz) {
  return pval[this->index5d(i5,i4,iy,ix,iz)];
  }
  inline T& operator() (int i6, int i5, int i4, int iy, int ix, int iz) {
  return pval[this->index6d(i6,i5,i4,iy,ix,iz)];
  }
  inline T& operator() (int i7, int i6, int i5, int i4, int iy, int ix, int iz) {
  return pval[this->index7d(i7,i6,i5,i4,iy,ix,iz)];
  }
  inline T& operator() (int i8, int i7, int i6, int i5, int i4, int iy, int ix, int iz) {
  return pval[this->index8d(i8,i7,i6,i5,i4,iy,ix,iz)];
  }
   
   //Do elem by elem multiply with cube in, and self cube will be updated
   void reciprocal();
   void multiply(const cubeData<T>& in);
   void multiply(const cubeData<T>& in1, const cubeData<T>& in2);
   void divide(const cubeData<T>& in);
   void divide(const cubeData<T>& in1, const cubeData<T>& in2);
   void divideByMask(const cubeData<T>& in1, const cubeData<float>& mask);
   void divideByMask(const cubeData<T>& in1, const cubeData<T>& in2,const cubeData<float>& mask);
   //calculate Z=X+a*Y or X=X+a*Y, can reuse code
   void scaleAdd(const cubeData<T>& X, const cubeData<T>& Y, T a);
   void scaleAdd(const cubeData<T>& Y, T a);
   
   //calc Z=Z+Y or Z=X+Y
   void add(const cubeData<T>& X, const cubeData<T>& Y);
   void add(const cubeData<T>& Y);
   void add(const T& scalar);
   
   void subtract(const cubeData<T>& X, const cubeData<T>& Y);
   void subtract(const cubeData<T>& Y);
   
   //calc Z=a*X or X=a*x
   void scale(const cubeData<T>& X, T a);
   void scale(const T scale);
	
	  
  void sqrt();
  
  //return RMS value of the cube
  T RMS() const;
  void normalize ();
  //return RMS value tickling out the values close to zero ( smaller than scale*maxAbsVal)
  T RMS_NZ(float scale) const;
  T sum() const {
    T sumV = T(0); //Make zero of type T
    long long i;
    for (i=0;i<this->dom.get_n123();i++) {
      sumV += pval[i];
    }
    return sumV;
  }
  
  void info(const std::string& sz) const{
   cerr<<"\n cubeInfo:"<<sz<<". ";
   this->dom.info("domain");
   cerr<<"\nRMS:"<<RMS()<<",min/max"<<minval()<<','<<maxval()<<endl;
  }
   
   //calculate the prod of all elements in cube
   T  prod() const {
     T prodV = T(1); //Make zero of type T
     long long i;
     for (i=0;i=this->dom.get_n123();i++) {
       prodV *= pval[i];
     }
     return prodV;
   }
   
   
   template <typename boolFunc>
   long long count(boolFunc bf) const;
   
  //content access interface
  virtual void seq_grab(cubeData<T> * cube);
  virtual void seq_update(const cubeData<T> * cube); 
  virtual void seq_grab   (T* pArr, size_t n);
  virtual void seq_update (const T* pArr, size_t n);

  //Random access
  //Grab a window of the object
  virtual void grabWindow(cubeData<T> *cube,  const windParam& wind);
  //update a window of the data object 
  virtual void updateWindow(const cubeData<T>* cube,const windParam& wind);

  virtual void grab3d(cubeData<T> * cube,  const windParam& wind) ;
  virtual void grab2d(cubeData<T> * cube,  const windParam& wind) ;
  virtual void grab1d(cubeData<T> * cube,  const windParam& wind) ;
  virtual void grabXd(cubeData<T> * cube,  const windParam& wind, int ndim);
  virtual void grabSeveralXd(cubeData<T> * cube,  const windParam& wind, int ndim);
  
  virtual void update1d(const cubeData<T> * cube,   const windParam& wind);
  virtual void update2d(const cubeData<T> * cube,   const windParam& wind);
  virtual void update3d(const cubeData<T> * cube,   const windParam& wind);
  virtual void updateXd(const cubeData<T> * cube,   const windParam& wind, int ndim);
  virtual void updateSeveralXd(const cubeData<T> * cube,  const windParam& wind, int ndim);
    
  //set proxy
  void becomeProxyCube(const hypercube& shape);
  void setProxyXd(cubeData<T>* cube, const windParam& wind, int ndim);
  void proxyPtrIs(T* ptr);
  
protected:
   
  //Internal alloc function, actually alloc memory
  void _alloc( const hypercube& shape_in) {
    pval = NULL;
    long long n = shape_in.get_n123();
    assert(n>0);
    assert(n<MAX_MEMORY_ALLOC);
    pval = new T[n];
    zero();
         //build the cubePtr2d,3d
  //         if(ndim == 2) //build 2d Ptr
  //         {
  //            mCubePtr2d.gen2dPtr(this->dom,pval);
  //         }
  //         if(ndim ==3 ) //build 3d Ptr
  //         {
  //            mCubePtr3d.gen3dPtr(&this->dom,pval);
  //         }
         //Set the n1,n2... parameters
    assert(this->dom.get_ndim() <= 5);
    stdVecInt vns = shape_in.get_ns();
    VERIFY(vns.size() <= max_dim);
    std::copy(vns.begin(),vns.end(),&(this->ns[0]));
  }
   
  private:
     //Internel use for content access
   void grabUpdateXd(const cubeData<T>* cube, const windParam&wind, bool bGrab, int ndim);
   void grabUpdateSeveralXd(const cubeData<T>* cube, const windParam&wind, bool bGrab, int ndim);
   void traverse(long long offset, int idim);
   void GrabUpdateWnd(bool bgrab, const cubeData<T>* cube, const windParam& wind);

   //forbid implict assignment operators and copy constructors
}; //end of class definition

#include "cubeinc.h"

template <typename T>
T cubeData<T>::RMS () const
{
   T val = ::Yang::dotProd(*this,*this);
   return std::sqrt(val/(this->domain().get_n123()));
}

template <typename T>
void cubeData<T>::sqrt () 
{
  long long i=0, n123 = this->domain().get_n123();
  for(i=0; i<n123; i++) {
    this->pval[i] = std::sqrt(this->pval[i]);
  }
}

class absValGreaterThanFunClass {
public:
   float fRef;
   absValGreaterThanFunClass (float fRef_) {
	  fRef = fRef_;
   }
   bool operator() (float fin) {
	  if (fabs(fin) > fRef)
		return true;
	  else
		 return false;
   }
};

template <typename T>
template <typename boolFunc>
long long cubeData<T>::count(boolFunc bf) const {
   long long cnt = 0;
   long long index = 0;
   for (index=0; index < this->domain().get_n123(); index++) {
	  if (bf( (*this)[index] )) {
		 cnt ++;
	  }
   }
   return cnt;
}

template <typename T>
T cubeData<T>::RMS_NZ(float scale = 1e-5) const {
   T val = ::Yang::dotProd(*this,*this);
   float valCompare = this->maxvalAbs() * scale;
   
   absValGreaterThanFunClass funcObj(valCompare);
   float fCnt = float( this->count(funcObj));
   if (fCnt < 0.5f) { //All elements are zero
     return 0.0f;
   } else {
     return std::sqrt(val/fCnt);
   }
}

template <typename T>
void cubeData<T>::normalize ()
{
   this->scale(1.0f/(this->RMS()));
}

//calc X=X.*Y, or Z=X.*Y;
template <typename T>
void cubeData<T>::multiply(const cubeData<T>& in)
{
   ::Yang::multiply(*this,*this,in);
}

template <typename T>
void cubeData<T>::reciprocal()
{
   ::Yang::reciprocal(*this,*this);
}

template <typename T>
void cubeData<T>::multiply(const cubeData<T>& in1, const cubeData<T>& in2)
{
   ::Yang::multiply(*this,in1,in2);
}

//calc Z=Z./Y, or Z=X.*Y;
template <typename T>
void cubeData<T>::divide(const cubeData<T>& in)
{
   ::Yang::divide(*this,*this,in);
}
template <typename T>
void cubeData<T>::divide(const cubeData<T>& in1, const cubeData<T>& in2)
{
   ::Yang::divide(*this,in1,in2);
}

template <typename T>
void cubeData<T>::divideByMask(const cubeData<T>& in1, const cubeData<T>& in2,const cubeData<float>& mask) {
   ::Yang::divideByMask(*this,in1,in2,mask);
}
template <typename T>
void cubeData<T>::divideByMask(const cubeData<T>& in, const cubeData<float>& mask) {
   ::Yang::divideByMask(*this,*this,in,mask);
}

//calculate Z=X+a*Y or X=X+a*Y, can reuse code
template<typename T>
void cubeData<T>::scaleAdd(const cubeData<T>& X, const cubeData<T>& Y, T a) {
   ::Yang::scaleAdd(*this,X,Y,a);
}

template<typename T>
void cubeData<T>::scaleAdd(const cubeData<T>& Y, T a) {
   ::Yang::scaleAdd(*this,*this,Y,a);
}

//calc Z=Z-Y or Z=X-Y
template<typename T>
void cubeData<T>::subtract(const cubeData<T>& Y) {
   ::Yang::scaleAdd(*this,*this,Y,-1.0f);
}

template<typename T>
void cubeData<T>::subtract(const cubeData<T>& X, const cubeData<T>& Y) {
   ::Yang::scaleAdd(*this,X,Y,-1.0f);
}

//calc Z=Z+Y or Z=X+Y
template <typename T>
void cubeData<T>::add(const cubeData<T>& Y)
{
   ::Yang::add(*this,*this,Y);
}

template<typename T>
void cubeData<T>::add(const cubeData<T>& X, const cubeData<T>& Y) {
   ::Yang::add(*this,X,Y);
}

template<typename T>
void cubeData<T>::add(const T& scalar) {
  ::Yang::AddByNumber(*this,*this, scalar);
}

//calc Z=a*X or X=a*x
template<typename T>
void cubeData<T>::scale(const cubeData<T>& X, T a) {
   Yang::scaleByNumber(*this,X,a);
}

template<typename T>
void cubeData<T>::scale(T a) {
   this->scale(*this, a);
}

template <typename T>
void cubeData<T>::seq_grab(cubeData<T> * cube) //Only used when as an input class
{
  //do the size check
  rsf<T> :: seq_grab(cube);
  long long n = cube->domain().get_n123();
  memcpy((void*)(cube->pval),(void*)(this->pval+(this->m_seqCounter)-n), n*sizeof(T));
}
template <typename T>
void cubeData<T>::seq_grab(T* pArr, size_t n) {
   //do the size check
   rsf<T> :: seq_grab(pArr,n);
   memcpy((void*)(pArr),(void*)(this->pval+(this->m_seqCounter)-n), n*sizeof(T));
}

template <typename T>
void cubeData<T>::seq_update(const cubeData<T> * cube) //
{
	assert(!this->mbProxy);
	rsf<T> :: seq_update(cube);
	long long n=cube->domain().get_n123();
	memcpy((void*)(this->pval+(this->m_seqCounter)-n),(void*)(cube->pval),n*sizeof(T));
}
template <typename T>
void cubeData<T>::seq_update(const T* pArr, size_t n) {
  assert(!this->mbProxy);
	rsf<T> :: seq_update(pArr,n);
	memcpy((void*)(this->pval+(this->m_seqCounter)-n),(void*)(pArr),n*sizeof(T));
}

//Get a trace, n have to be compatible with domain size(e.g. assume a full trace/2Dslice/Cube grab, nw=dimension max, jw=1, same for 2d,3d but not for nd
//the index out-of-bound check is left to the gradUpdateXd function
template<typename T>
void cubeData<T>::grab1d(cubeData<T> * cube,  const windParam& wind) {
   rsf<T>::grab1d(cube,wind);                                
   grabUpdateXd(cube,wind,true,1);
}
template<typename T>
void cubeData<T>::grab2d(cubeData<T> * cube,  const windParam& wind) {
   rsf<T>::grab2d(cube,wind);
   grabUpdateXd(cube,wind,true,2);
}
template<typename T>
void cubeData<T>::grab3d(cubeData<T> * cube,  const windParam& wind) {
   rsf<T>::grab3d(cube,wind);                                
   grabUpdateXd(cube,wind,true,3);
}
template<typename T>
void cubeData<T>::grabXd(cubeData<T> * cube,  const windParam& wind, int ndim) {
   rsf<T>::grabXd(cube,wind,ndim);
   grabUpdateXd(cube,wind,true,ndim);
}
template<typename T>
void cubeData<T>::grabSeveralXd(cubeData<T> * cube,  const windParam& wind, int ndim) {
   rsf<T>::grabSeveralXd(cube,wind,ndim);
   grabUpdateSeveralXd(cube,wind,true,ndim);
}

//For update memory in blocks
template<typename T>
void cubeData<T>::updateSeveralXd(const cubeData<T> * cube,  const windParam& wind, int ndim) {
   rsf<T>::updateSeveralXd(cube,wind,ndim);
   grabUpdateSeveralXd(cube,wind,false,ndim);
}

template<typename T>
void cubeData<T>::updateXd(const cubeData<T> * cube,  const windParam& wind, int ndim) {
   rsf<T>::updateXd(cube,wind,ndim);
   grabUpdateXd(cube,wind,false,ndim);
}

template<typename T>
void cubeData<T>::update1d(const cubeData<T> * cube,  const windParam& wind) {
   rsf<T>::update1d(cube,wind);
   grabUpdateXd(cube,wind,false,1);
}

template<typename T>
void cubeData<T>::update2d(const cubeData<T> * cube,  const windParam& wind) {
   rsf<T>::update2d(cube,wind);
   grabUpdateXd(cube,wind,false,2);
}
template<typename T>
void cubeData<T>::update3d(const cubeData<T> * cube,  const windParam& wind) {
   rsf<T>::update3d(cube,wind);
   grabUpdateXd(cube,wind,false,3);
}

//The following TWO functions are not thread safe, and need testing as well
template<typename T>
void cubeData<T>::updateWindow(const cubeData<T>* cube,const windParam& wind)
{
	//check size
	rsf<T>::updateWindow(cube,wind);
	this->GrabUpdateWnd(false,cube,wind);
}

template<typename T>
void cubeData<T>::grabWindow(cubeData<T> *cube,  const windParam& wind)
{
	rsf<T>::updateWindow(cube,wind);
  //printWindParam(this->mWindParamsTmp,"windpar");
	this->GrabUpdateWnd(true,cube,this->mWindParamsTmp);
}

//Internal function
template<typename T>
void cubeData<T>::GrabUpdateWnd(bool bgrab, const cubeData<T>* cube, const windParam& wind) {
   long long offset = 0;
   int ndim = this->dom.get_ndim();
   
   this->m_bGrab    = bgrab;
   this->m_pWindPar = &wind;
   
   m_pCubeDataTmp = (cubeData<T>*) cube;
   m_grabndCnter = 0;
   //Use recursive to cover all the case
   //Calculate the offset in memory, starts with the highest dimension
   traverse(offset,ndim);
   assert(m_grabndCnter == cube->domain().get_n123());
}

template<typename T>
void cubeData<T>::traverse(long long offset, int idim) {
   long long noffset;
   if(idim>1) {
      noffset = offset + (*(this->m_pWindPar))[idim-1].f;
      for(int k=0; k<(*(this->m_pWindPar))[idim-1].n;k++) {
         traverse(noffset*(this->dom.get_n(idim-1)),idim-1);
         noffset += (*(this->m_pWindPar))[idim-1].j;
      }
   }
   else { //Already come to the first axis, copy the memory here
      T* pdest; T* psrc;
      noffset = offset + (*(this->m_pWindPar))[0].f;
      psrc = this->pval+noffset;
      pdest = this->m_pCubeDataTmp->pval+m_grabndCnter;
      size_t shiftDest = 0, shiftSrc = 0;
      
      for(int k=0; k<(*this->m_pWindPar)[0].n;k++) {
         if (this->m_bGrab) { //Copy from src to dest
            *(pdest+shiftDest) = *(psrc+shiftSrc);
         }
         else { //Update, //copy from dest to src
            *(psrc+shiftSrc) = *(pdest+shiftDest);
         }
         shiftSrc += (*this->m_pWindPar)[0].j;
         shiftDest ++;
         m_grabndCnter++;
      } 
   }
}

template<typename T>
void cubeData<T>::grabUpdateXd(const cubeData<T>* cube, const windParam&wind, bool bGrab, int ndim) {
   long long offset = 0;
   int ndimL = this->domain().get_ndim();
   assert(ndim == cube->domain().get_ndim());
  
   windParam windd(wind);
   //check whether the window parameter is out of bound
   rsf<T>::checkWindowParams_par(cube,windd);
   
   //Calculate the offset in memory, starts with the highest dimension
   stdVecInt ns = this->domain().get_ns();
   offset = wind[ndimL-1].f;
   for(int i=ndimL-2;i>=ndim;i--) {
      offset *= ns[i];
      offset += wind[i].f;
   }
   for(int i=ndim-1; i>=0; i--) {
	  offset *= ns[i];
   }
   
   assert(wind[0].j==1);
   
   size_t elemCnt = 1;
   for(int i=0;i<ndim; i++)
      elemCnt *= ns[i];
   //de cast the const operator
   cubeData<T>* dcube = (cubeData<T>*) cube;
   //do the memory copy
   if(bGrab) {
      myMemcpy(dcube->pval,this->pval+offset,elemCnt);
   }
   else //update
      myMemcpy(this->pval+offset,dcube->pval,elemCnt);
   return;
}

//This is the only way to get a cube this is just proxy
template<typename T>
void cubeData<T>::becomeProxyCube(const hypercube& shape) {
  if (!this->mbProxy) {
    VERIFY(this->dom.isEmpty());
    this->mbProxy = true;
  }
  this->dom = shape;
  
  stdVecInt vns = shape.get_ns();
  VERIFY(vns.size() <= max_dim);
  std::copy(vns.begin(),vns.end(),&(this->ns[0]));
  this->pval = NULL;
}

template<typename T>
void cubeData<T>::setProxyXd(cubeData<T>* cube, const windParam& windPar, int ndim) {

  VERIFY(cube->mbProxy);
  //check size and update window params.
  VERIFY(cube->domain().get_ndim()==ndim);
  assert(this->domain().compatible(cube->domain()));
  windParam wind(windPar);

  rsf<T>::checkWindowParams_par(cube,wind);
  

  long long offset = 0;
  int ndimL = this->domain().get_ndim();
  assert(ndim == cube->domain().get_ndim());
   
  //Calculate the offset in memory, starts with the highest dimension
  stdVecInt ns = this->domain().get_ns();
  offset = wind[ndimL-1].f;
  for(int i=ndimL-2;i>=ndim;i--) {
    offset *= ns[i];
    offset += wind[i].f;
  }
  for(int i=ndim-1; i>=0; i--) {
    offset *= ns[i];
  }
   
   //Just assign the pointer.
   cube->pval = this->pval+offset;
   return;
}

//The user should make sure the size of cube-> memory is compatible.
template<typename T>
void cubeData<T>::proxyPtrIs(T* ptr) {
  VERIFY(this->mbProxy);
  //Just assign the pointer.
  this->pval = ptr;
  return;
}

//Grab several Xd slice consequtively, needs to be continous in memory
//ndim+1 = cube.dimension in this case.
template<typename T>
void cubeData<T>::grabUpdateSeveralXd(const cubeData<T>* cube, const windParam&wind, bool bGrab, int ndim) {
   long long offset = 0;
   int ndimL = this->domain().get_ndim();
   assert(ndim+1 == cube->domain().get_ndim());
   
   //Calculate the offset in memory, starts with the highest dimension
   stdVecInt ns = this->domain().get_ns();
   offset = wind[ndimL-1].f;
   for(int i=ndimL-2;i>=ndim;i--) {
      offset *= ns[i];
      offset += wind[i].f;
   }
   for(int i=ndim-1; i>=0; i--) {
	  offset *= ns[i];
   }
   
   assert(wind[0].j==1);
   
   size_t elemCnt = 1;
   for(int i=0;i<ndim; i++)
      elemCnt *= ns[i];
   //only difference is now there are mutliple slices
   elemCnt *= wind[ndim].n;
   
   //de cast the const operator
   cubeData<T>* dcube = (cubeData<T>*) cube;
   //do the memory copy
   if(bGrab)
      myMemcpy(dcube->pval,this->pval+offset,elemCnt);
   else //update
      myMemcpy(this->pval+offset,dcube->pval,elemCnt);
   return;
}

//applys to both valarray and stdvec
template< class Tv>
void dumpStdVec(const std::vector<Tv>& vec, string tag, bool bNoDummy=false) {
  cubeData<Tv> errc;
  errc.alloc(hypercube(vec.size()));
  myMemcpy(errc.pval,&vec[0],vec.size());
  errc.dump(tag,bNoDummy);
}
template< class Tv>
void dumpStdArray(const std::valarray<Tv>& vec, string tag, bool bNoDummy=false) {
  cubeData<Tv> errc;
  errc.alloc(hypercube(vec.size()));
  myMemcpy(errc.pval,&vec[0],vec.size());
  errc.dump(tag,bNoDummy);
}

}//end of namespace definition
#endif
