#ifndef CUBE_PTR_2D_H
#define CUBE_PTR_2D_H

#include "hypercube.h"
#include "cubePtr.hpp"

namespace Yang {
      
template <typename T>
class cubeData;

template <typename T>
class cubePtr2d : public cubePtr
//This can be extended to a windowed memory of a cubeData object.
//Assume the array is 2D, then we can expose a 2D array pointer to make easy indexing
{
   
   friend class cubeData<T>;   
private://members
   T ** ptr2d;
   int m_n1,m_n2;

   
public:
	cubePtr2d() : ptr2d(NULL),m_n1(0),m_n2(0) {}
   //Get a 2D slice from the pval (with shape pshpae), window start at fw(array), window size nw(nw1,nw2) 
   //if nw==NULL, nw1,nw2 == pshape->n1,n2;
   //if fw==NULL, fw(:) = 0;
   
   ~cubePtr2d() {deallocate();}
	void deallocate()
	{ 
	   SAFE_DELETE_ARRAY(ptr2d)
   } 
   
private:
   void gen2dPtr(const hypercube& shape, T* pval, stdvecInt* fw = NULL, stdvecInt* nw = NULL)
	{ 
        //First deallocate the previous one
        deallocate();
        
        //from here ptr 'pval' is already allocated. we need to allocate the ptr2d array
        int n1,n2;
        n2 = shape.get_n(2);
        n1 = shape.get_n(1);

        //Locate the starting address
        long long offset;
        stdvecInt ns = shape.get_ns();

        stdvecInt *pfw,*pnw;
        stdvecInt nfw,nnw;
        pfw = fw;
        pnw = nw;

        if (nw==NULL)
        {
           nnw.push_back(n1);
           nnw.push_back(n2);
           pnw = &nnw;
        }

        if (fw==NULL)
        {
           //Set fw=all 0
           nfw.assign(ns.size(),0);
           pfw = &nfw;
        }  

        assert( pnw->size() == 2);
        m_n1 = (*pnw)[0];
        m_n2 = (*pnw)[1];
        offset = calcOffset(ns,*pfw);
        assert(checkBound(ns,*pfw,*pnw));
        
        ptr2d = new T*[m_n2];
        int i;
        for (i=0; i<m_n2; i++)
        {
            ptr2d[i] = (pval+offset+i*n1);
        }
      }


}; //end of class definition

}//end of namespace defi
#endif

