#ifndef CUBE_PTR_3D_H
#define CUBE_PTR_3D_H

#include "hypercube.h"
#include "cubePtr.hpp"

namespace Yang {

template <typename T>
class cubeData;
   
template <typename T>
class cubePtr3d : public cubePtr
//This can be extended to a windowed memory of a cubeData object.
//Assume the array is 2D, then we can expose a 3D array pointer to make easy indexing
{
   friend class cubeData<T>;
private:
   //members
   T *** ptr3d;
   int m_n1,m_n2,m_n3;
   
public:
	cubePtr3d() : ptr3d(NULL),m_n1(0),m_n2(0), m_n3(0) {}
   ~cubePtr3d() {deallocate();}

private:
   void gen3dPtr(hypercube* pshape, T* pval, stdvecInt* fw = NULL, stdvecInt* nw = NULL)
	 { 
        //First deallocate the previous one
        deallocate();
        
        //from here ptr 'pval' is already allocated. we need to allocate the ptr3d array
        hypercube & shape = *pshape;
        int n1,n2,n3;
        n3 = shape.get_n(3);
        n2 = shape.get_n(2);
        n1 = shape.get_n(1);

        //Locate the starting address
        long long offset;
        stdvecInt ns = shape.get_ns();

        stdvecInt *pfw,*pnw;
        stdvecInt nfw,nnw;
        pfw = fw;
        pnw = nw;

        if (nw==NULL)
        {
           nnw.push_back(n1);
           nnw.push_back(n2);
           nnw.push_back(n3);
           pnw = &nnw;
        }

        if (fw==NULL)
        {
           //Set fw=all 0
           nfw.assign(ns.size(),0);
           pfw = &nfw;
        }  

        assert( pnw->size() == 3);
        m_n1 = (*pnw)[0];
        m_n2 = (*pnw)[1];
        m_n3 = (*pnw)[2];

        offset = calcOffset(ns,*pfw);
        assert(checkBound(ns,*pfw,*pnw));
        
        ptr3d = new T**[m_n3];
        long long i,j;
        for (i=0; i<m_n3; i++)
        {
            ptr3d[i] = new T*[m_n2];
            for (j=0; j< m_n2; j++)
            {
               ptr3d[i][j] = (pval+offset+(i*n2+j)*n1);
            }
        }
    }

	void deallocate()
	{
	   long i;
	   for (i=0;i<m_n3;i++)
	   {
	      SAFE_DELETE_ARRAY(ptr3d[i])
	   }
	   SAFE_DELETE_ARRAY(ptr3d)
	   m_n3 = m_n2 = m_n1 = 0;
   }
}; //end of class definition

}//end of namespace defi
#endif


