//This file is used only by cubeData.hpp, need not be guarded by included header

//Define a copy cubeData function for float/double conversion
template<typename T1,typename T2>
void copyCubeData(cubeData<T2> & destCube, cubeData<T1> & srcCube){
   //First check size
   if (!srcCube.domain().sameSize(destCube.domain())) {
      seperr("the size of two cubes are not equal in func copyCubeData\n");
   }
   //Copy the element one by one
   long long n = srcCube.domain().get_n123();
   long long index;
   for(index=0;index<n;index++) {
      destCube[index] = srcCube[index];
   }
}

template<typename T>
void copyCubeData(cubeData<T> & destCube, cubeData<T>& srcCube){
   //First check size
   if (!srcCube.domain().sameSize(destCube.domain())) {
      seperr("the size of two cubes are not equal in func copyCubeData\n");
   }
   //Copy the element one by one
   long long n = srcCube.domain().get_n123();
   myMemcpy<T>(destCube.pval,srcCube.pval,n);
}


//Do elem by elem multiply
//This function out = multiply(in1,in2);
template<typename T>
void multiply(cubeData<T>& out, const cubeData<T>& in1, const cubeData<T>& in2)
{
   long long n = in1.domain().get_n123();
   assert(in2.domain().sameSize(in1.domain()));
   VERIFY( n== in1.domain().get_n123())
   long long index;
   const T *p1;
   T *p2,*pdest;
   pdest = out.pval; p1 = in1.pval; p2=in2.pval;
   
   for(index=0;index<n;index++){
      *(pdest+index) = (*(p1+index)) * (*(p2+index));
   }
}

template<typename T1, typename T2> //works for some cases, for example cmplxf*float=cmplxf
void multiply(cubeData<T1>& out, const cubeData<T1>& in1, const cubeData<T2>& in2)
{
   long long n = in1.domain().get_n123();
   assert(in2.domain().sameSize(in1.domain()));
   VERIFY( n== in1.domain().get_n123())
   long long index;
   const T2 *p2;
   T1 *p1,*pdest;
   pdest = out.pval; p1 = in1.pval; p2=in2.pval;
   
   for(index=0;index<n;index++){
      *(pdest+index) = (*(p1+index)) * (*(p2+index));
   }
}


//Reciprocal
template<typename T>
void reciprocal(cubeData<T>& out, const cubeData<T>& in)
{
   long long n = in.domain().get_n123();
   assert(out.domain().sameSize(in.domain()));
   long long index;
   const T *p1;
   T *pdest;
   pdest = out.pval; p1 = in.pval;
   for(index=0;index<n;index++){
      *(pdest+index) = 1.0f/(*(p1+index));
   }
}

//This function out = (in1 ./ in2);
template<typename T>
void divide(cubeData<T>& out, const cubeData<T>& in1, const cubeData<T>& in2)
{
   //Make sure there is no divide by zero first
   VERIFY(in2.minvalAbs() > epsilon);
   
   long long n = in1.domain().get_n123();
   assert(in2.domain().sameSize(in1.domain()));
   VERIFY( n== in1.domain().get_n123())
   long long index;
   const T * __restrict p1;
   T * __restrict p2,* __restrict pdest;

   pdest = out.pval; p1 = in1.pval; p2=in2.pval;
   
   //#pragma omp parallel for
   for(index=0;index<n;index++){
      *(pdest+index) = (*(p1+index)) / (*(p2+index));
   }
}

//Perform division only on where mask=1
template<typename T>
void divideByMask(cubeData<T>& out, const cubeData<T>& in1, const cubeData<T>& in2,const cubeData<float>& mask)
{
   //Make sure there is no divide by zero first
   assert(mask.sameSize(out));
   long long n = in1.domain().get_n123();
   assert(in2.domain().sameSize(in1.domain()));
   VERIFY( n== in1.domain().get_n123())
   long long index;
   const T * __restrict p1;
   T * __restrict p2,* __restrict pdest;
   const float* pMask = mask.dataPtr();
   pdest = out.pval; p1 = in1.pval; p2=in2.pval;
   
   for(index=0;index<n;index++){
	  if (equalf(*pMask,1.0f)) {
		 *(pdest+index) = (*(p1+index)) / (*(p2+index));
	  }
	  pMask++;
   }
}

//Do Adding
//This function out = in1 + in2;
template<typename T>
void add(cubeData<T>& out, const cubeData<T>& in1, const cubeData<T>& in2)
{
   long long n = in1.domain().get_n123();
   assert(in2.domain().sameSize(in1.domain()));
   VERIFY( n== in1.domain().get_n123())
   long long index;
   const T *p1;
   T *p2,*pdest;
   pdest = out.pval; p1 = in1.pval; p2=in2.pval;
   
   for(index=0;index<n;index++) {
      *(pdest+index) = (*(p1+index)) + (*(p2+index));
   }
}

template<typename T>
void scaleByNumber(cubeData<T>& out, const cubeData<T>& in, const T& scale)
{
   long long n = in.domain().get_n123();
   assert(out.domain().sameSize(in.domain()));
   
   long long index;
   const T *p1;
   T *pdest;
   pdest = out.dataPtr(); p1 = in.dataPtr();
   
   for(index=0;index<n;index++) {
      *(pdest+index) = *(p1+index) * scale;
   }
}

template<typename T>
void AddByNumber(cubeData<T>& out, const cubeData<T>& in, const T& operand)
{
   long long n = in.domain().get_n123();
   assert(out.domain().sameSize(in.domain()));
   
   long long index;
   const T *p1;
   T *pdest;
   pdest = out.dataPtr(); p1 = in.dataPtr();
   
   for(index=0;index<n;index++) {
      *(pdest+index) = *(p1+index) + operand;
   }
}

template<typename T>
double dotProd(const cubeData<T>& in1, const cubeData<T>& in2)
{
  double result = 0.0;
  long long i,n;
  n = in1.domain().get_n123();
  assert(n == in2.domain().get_n123());
  const T *p1,*p2;
  p1 = in1.dataPtr();
  p2 = in2.dataPtr();
  
  //May use OMP reduce here
  for(i=0;i<n;i++) {
    result += double(p1[i]*p2[i]);
  }
  return result;
}

//calc Z=X+a*Y
template<typename T>
void scaleAdd(cubeData<T>& out, const cubeData<T>& X, const cubeData<T>& Y, T a) {
   long long n = X.domain().get_n123();
   assert( n==out.domain().get_n123());
   
   long long index;
   const T *p1,*p2;
   T *pdest;
   pdest = out.dataPtr(); p1 = X.dataPtr(); p2 = Y.dataPtr();
   
   for(index=0;index<n;index++) {
      *(pdest+index) = *(p1+index) + *(p2+index)*a;
   }
}

