#ifndef __CUDAMODELING2D_H_INCL__
#define __CUDAMODELING2D_H_INCL__

// TODO: your code here
#define GPU_STENCIL_WIDTH 4	//We only have 8th order implementation currently.

//For 2D image gathers, o,d are integers for grid point coordinates, not physical coordinates
struct cube_c{
   float *p;	//Point to the memory chunk, this should always be valid
   float *p2;	//Pointer that might be used as second cube.
   bool bUse_p2;
   int	nframe;
   int d;
   int o;
   int nz;
   int nx;
};

inline void init_cube_c(cube_c *pcCube) {
   pcCube->p = NULL;
   pcCube->p2= NULL;
   pcCube->bUse_p2= false;
   pcCube->nframe = 0;
   pcCube->d  = 0;
   pcCube->o  = 0;
   pcCube->nz = 0;
   pcCube->nx = 0;
}

inline bool sameSize_cube_c(cube_c* p1, cube_c* p2) {
  if (p1->nz != p2->nz) return false;
  if (p1->nx != p2->nx) return false;
  if (p1->nframe != p2->nframe) return false;
  return true;
}

extern "C" inline void copy_cube_c_info(cube_c* pDest, cube_c* pSrc) {
   //pDest->p = pSrc->p; !Only copy info, not copying memory pointer
   pDest->nframe = pSrc->nframe;
   pDest->d  = pSrc->d;
   pDest->o  = pSrc->o;
   pDest->nz = pSrc->nz;
   pDest->nx = pSrc->nx;
}

//Simply 2d nonlinear modeling
//The dampCoefArray will be values starts from large negative absorbing coef to almost zero. dampArrayLength
//should be larger than the minumum distance from four boundaries of any points in the domain.
extern "C" void gpu2dModeling_init(int *wfld_j, int *record_j, int *gpu_id);
extern "C" bool gpu2dModeling_setVelocity(float *cpu_space, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int nRecv, int ntRec, int* ix_recv, int* iz_recv);
extern "C" void gpu2d_wave_prop(int *iz_cpu, int iz_cpu_len, float *source_cpu, float *wfld, float *recordBuffer);
extern "C" void gpuModeling2d_uninit();

//RTM
extern "C" void gpu2dRTM_init(int *wfld_j, int *gpu_id);
extern "C" bool gpu2dRTM_setVelocity(float *cpu_space, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
                                     float *dampCoefArray, int *dampArrayLength, int* npad, int *jImaging, cube_c* pImageGathers,
                                     float *srcIntense=NULL, float* recvIntense=NULL );

extern "C" void gpu2dRTM_imaging(int *src_iz_cpu, int src_iz_len, float *src_cpu,		//source inject
								  int *rec_iz_cpu, int rec_iz_len, float *rec_cpu, 	 	//receiver inject
								  float *wfld_src, float *wfld_recv, //for wavfield snapshot
								  float *cpu_image,bool bTake2ndDeriv=false,
                  bool bDoUpDownSeparation=false, float* cpu_image_refl=NULL, float* cpu_image_trans=NULL);

extern "C" void gpuRTM2d_uninit();

//adjRTM
extern "C" void gpu2dAdjRTM_init(int *wfld_j, int* record_j, int *gpu_id);
extern "C" void gpu2dAdjRTM_uninit();

extern "C" bool gpu2dAdjRTM_setVelocityAndRefl(float* vel_cpu, float* refl_cpu, cube_c* img_off_gather_cpu, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int *npad, int *jImaging, int nRecv, int ntRec, int* ix_recv, int* iz_recv);
extern "C" void gpu2dAdjRTM(int *iz_cpu, int iz_cpu_len, float *source_cpu, float *wfldSrc, float *wfldScatter, 
                            float *recordBuffer);
//extern "C" void gpuAdjRTM2d_uninit();

///WEM Tomograph operator
//First Adjoint opeartor from deltaI --> slowness/velocity perturbation
extern "C" void gpu2dTomAdj_init(int *wfld_j, int *gpu_id);
extern "C" void gpu2dTomAdj_uninit();
extern "C" bool gpu2dTomAdj_setVelocity(float *vel_cpu, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int *jImaging, 
										float* pImg_cpu, cube_c* img_off_gather_cpu);
extern "C" void gpu2dTomAdj(int *src_iz_cpu, int src_iz_len, float *src_cpu,		//source inject
								  int *rec_iz_cpu, int rec_iz_len, float *rec_cpu, 	 	//receiver inject
								  float *wfld_src, float *wfld_recv,float *wfld_src2, float *wfld_recv2,
								  float* dSs_cpu, float* dSr_cpu, bool bTake4thDeriv=false); //slowness pertubation for output

//Then Fwd operator from deltaS --> delta I
extern "C" void gpu2dTomFwd_init(int *wfld_j, int *gpu_id);
extern "C" void gpu2dTomFwd_uninit();
extern "C" bool gpu2dTomFwd_setVelocity(float *vel_cpu, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int *jImaging, 
										  cube_c* img_off_gather_cpu);
extern "C" void gpu2dTomFwd(int *src_iz_cpu, int src_iz_len, float *src_cpu,		//source inject
								  int *rec_iz_cpu, int rec_iz_len, float *rec_cpu, 	 	//receiver inject
								  int prestack_imaging_kernel_type,
								  float *wfld_src, float *wfld_recv,float *wfld_src2, float *wfld_recv2,
								  float *pDeltaS, //Input velocity/slowness perturbation
								  float* pImgS_cpu,float *pImgR_cpu, bool bTake4thDeriv=false); //offset gather image pertubation (delta(I(x,h)))

#endif // __CUDAMODELING2D_H_INCL__

