#include "stdafx.h"
#include "decl_fwd.h"
#include "cubeData.hpp"
#include "igrad2op.h"
#include "cubeAlgo.h"
#include "cubeAlgo2.h"
#include "dipEstimator.h"
#include "point2d.hpp"
#include "dmRsfDump.h"

namespace Yang {
  class VAind{  //array indexing object
  public:
    VAind(int n1, int n2=1, int n3=1, int n4=1, int n5=1) { //n5 is not needed to store..
      ns_[0]=n1; ns_[1]=n2; ns_[2]=n3; ns_[3]=n4;
    }
    inline int operator() (int i1, int i2) {
      return i2*ns_[0]+i1;
    }
    inline int operator() (int i1, int i2, int i3) {
      return ((i3*ns_[1]+i2)*ns_[0])+i1;
    }
    inline int operator() (int i1, int i2, int i3, int i4) {
      return (((i4*ns_[2]+i3)*ns_[1]+i2)*ns_[0])+i1;
    }
    inline int operator() (int i1, int i2, int i3, int i4, int i5) {
      return ((((i5*ns_[3]+i4)*ns_[2]+i3)*ns_[1]+i2)*ns_[0])+i1;
    }
  private:
    int ns_[4];
  };
}

// TODO: your implementation here
using namespace Yang;
using namespace std;

void StructTensorParam::readParam(const param_func* par) {
  gs_sm_1 = par->get_float(string(prefix_+"f_gsm1_STensor"));
  gs_sm_2 = par->get_float(prefix_+"f_gsm2_STensor");
  maskThresRatio = par->get_float(prefix_+"fThrSTensorAmpRatio",1e-2);
  return;
}

void smoothAlongAxis1(cubeDataF& Io, const cubeDataF& Ii, const cubeDataF& filter){
  //first need to align two axis.
  const axis& ax_f = filter.domain().get_axis(1);
  const axis& ax_1 = Ii.domain().get_axis(1);
  int n2 = Ii.domain().get_n(2), n1 = ax_1.n;
  
  Io.zero(); float f; int off;
  VERIFY(equalf(ax_f.d,ax_1.d));
  int ibegf = ax_f.get_offset(0.0f); //align the filter to the ind=0 lag.
  for (int i2=0; i2<n2; i2++) {
    for (int i1=0; i1<n1; i1++) {
      //for each element in Ii, multiply the filter and add the product onto Io.
      f = Ii(i2,i1);
      for (off = max(-ibegf,-i1); off<min(ax_f.n-ibegf,n1-i1) ;off++) {
        Io(i2,i1+off) += f*filter[ibegf+off];
      }
    }
  }
}

void smoothAlongAxis2(cubeDataF& Io, const cubeDataF& Ii, const cubeDataF& filter) {
  //first need to align two axis.
  const axis& ax_f = filter.domain().get_axis(1);
  const axis& ax_2 = Ii.domain().get_axis(2);
  int n1 = Ii.domain().get_n(1), n2 = ax_2.n;
  
  Io.zero(); float f; int off;
  VERIFY(equalf(ax_f.d,ax_2.d));
  int ibegf = ax_f.get_offset(0.0f); //align the filter to the ind=0 lag.
  for (int i1=0; i1<n1; i1++) {
    for (int i2=0; i2<n2; i2++) {
      //for each element in Ii, multiply the filter and add the product onto Io.
      f = Ii(i2,i1);
      for (off = max(-ibegf,-i2); off<min(ax_f.n-ibegf,n2-i2) ;off++) {
        Io(i2+off,i1) += f*filter[ibegf+off];
      }
    }
  }
}

//construct structure tensor
void dipEstimator::inputImgIs(const cubeDataF& cub) {
  float rNorm = 1.0f/cub.RMS(); //normalization ratio
  
  const hypercube& shp = cub.domain();
  int ndim = shp.get_ndim();
  VERIFY (ndim == 2); //currently only do 2D
  //cerr << "The dimension of the input is " << ndim << ".\n";

  //Out put will be a (2D or 3D) vector at each grid point. and that is
  //put to the FIRST dimension
  //assume the cube is (z,x,y)
  const axis& ax_x = shp.get_axis(2);
  const axis& ax_z = shp.get_axis(1);
 
  //PRV(sizeof(long)); PRV(sizeof(long long)); PRV(sizeof(size_t)); PRV(sizeof(dipEstimator));
  PRV(sizeof(cubeDataF)); PRV(sizeof(hypercube)); PRV(sizeof(axis));
  //Read in parameters
  float gsm1 = par_->gs_sm_1, gsm2 = par_->gs_sm_2;
  
  ///Read the actually cube
  int n1,n2,n3; int i1,i2,i3;
  n1 = ax_z.n; n2=ax_x.n; n3=1;
  int n4 = 0;
  if (ndim == 2) { n4 = 3; }//I11,I12,I22
  else  {n4=6; } //I11,I12,I13,I22,I23,I33
  
  hypercube shpm(n1*n2*n3*n4), shpg(n1*n2*n3*ndim);
  
  matrix_.resize(shpm.get_n123()); //the elments of the structure tensor matrix.
  cubeDataF g(shpg),Izz(shp),Ixx(shp),Izx(shp),Izzs(shp),Ixxs(shp),Izxs(shp),Itt(shp);

  int n123 = n1 * n2 * n3;
  VAind indIn(n1,n2,n3), indOut(n1,n2,n3,ndim), indMat(n1,n2,n3,n4);

  if (ndim==2) {
    VERIFY(n3==1);
    //Apply gradient operator and then generate g*transp(g)
    //allocate graident field g, and other three fields Ixx,Iyy,Ixy
    igrad2op gradOp(n1,n2,ax_z.d,ax_x.d);
    gradOp.oper(false,false,const_cast<cubeDataF&>(cub),g); //find g.
    g.scale(rNorm); //normalize the amplitude.
    i3 = 0; //fills in the structure tensor matrix
    for (i1=0; i1<n1; i1++) {
      for (i2=0; i2<n2; i2++) {
        Izz[indIn(i1,i2)] = SQ(g[indOut(i1,i2,i3,0)]);
        Ixx[indIn(i1,i2)] = SQ(g[indOut(i1,i2,i3,1)]);
        Izx[indIn(i1,i2)] = g[indOut(i1,i2,i3,0)] * g[indOut(i1,i2,i3,1)];
      }
    }
    //Then smooth the field with Gaussian kernels, 2-D gaussian smoothing can be implementated as a 2-pass 1D gaussian smoothing.
    cubeDataF gaussX,gaussZ;
    axis ax_gausZ; int hfgZ = int(2.577f*0.5f*gsm1/ax_z.d+0.5f); ax_gausZ.set(2*hfgZ+1,-hfgZ*ax_z.d,ax_z.d); //the width of the gaussian is truncated at 1/100 maximum amplitude
    axis ax_gausX; int hfgX = int(2.577f*0.5f*gsm2/ax_x.d+0.5f); ax_gausX.set(2*hfgX+1,-hfgX*ax_x.d,ax_x.d); 
    getGaussfunc(gaussZ, ax_gausZ, gsm1, 0.f, 0);
    getGaussfunc(gaussX, ax_gausX, gsm2, 0.f, 0);
    smoothAlongAxis1(Itt,Izz,gaussZ); smoothAlongAxis2(Izzs,Itt,gaussX);
    smoothAlongAxis1(Itt,Ixx,gaussZ); smoothAlongAxis2(Ixxs,Itt,gaussX);
    smoothAlongAxis1(Itt,Izx,gaussZ); smoothAlongAxis2(Izxs,Itt,gaussX);
    
    matrix_=0.0f;
    //put the elements into the matrix
    for (i1=0; i1<n1; i1++) {
      for (i2=0; i2<n2; i2++) {
        matrix_[indMat(i1,i2,0)] = Izzs[indIn(i1,i2)];
        matrix_[indMat(i1,i2,1)] = Ixxs[indIn(i1,i2)];
        matrix_[indMat(i1,i2,2)] = Izxs[indIn(i1,i2)];
      }
    }
    
    //build mask the takes out very small values, which the dip cannot be confidently estimated.
    mask_.alloc(shp); cubeDataF mask2(shp);
    float thresRatio = this->par_->maskThresRatio;
    linearity_.alloc(shp); 
    
    //we mostly deal with horizontal dips.
    buildThresholdMask(mask_, Izzs, Izzs.maxval()*thresRatio , true);
    //buildThresholdMask(mask2, Ixxs, Ixxs.maxval()*thresRatio , true);
    //should use OR; mask_.add(mask2);
    dmRsf_ocDump<float> IssDump("dump24","Iss"); IssDump.init(&shp);
    IssDump.dumpFrame(Izzs); IssDump.dumpFrame(Ixxs); IssDump.dumpFrame(Izxs); IssDump.dumpFrame(mask_);
    
    eigenVals_.resize(n123*2);  eigenVec2_.resize(n123*2); //eigenVec2_.resize(n123);
    dipAngleX_->alloc(shp);
    vecAxis axes2 = shp.get_axes(); axes2.push_back(axis(2)); hypercube eVecShp(axes2);
    eigenVec1_.alloc(axes2);
    
    //Then find the eigen values for the struct-tensor matrix.
    float a,b,d; float T,D;
    float l1,l2,det; point2dF v1,v2; float fmaxzx = Izxs.maxvalAbs();
    for (i2=0; i2<n2; i2++) {
      for (i1=0; i1<n1; i1++) {
        if (mask_(i2,i1)>0.5f) { //only compute where there is some amplitude
          a = matrix_[indMat(i1,i2,0)]; d = matrix_[indMat(i1,i2,1)]; b = matrix_[indMat(i1,i2,2)];
          T = a+d;    D = a*d - b*b;
          
          det = T*T*0.25f - D; //gauranteed to >=0.
          l1  = T/2 + sqrtf(det); //this is the bigger eigenvalue, the corresponding eigenvector corresponds to the gradient direction
          l2  = T/2 - sqrtf(det); //this is the smaller eigenvalue.
          
          //infer the eigen vectors
          if (fabs(b)>fmaxzx*0.01f) {  //if b is big enough
            v1 = point2dF(l1-d,b); v2 = point2dF(l2-d,b);
          } else { //use the other way to calculate eigen vector, for better numerical stability.
            if (a > d) {
              v1 = point2dF(l1-d,b); v2 = point2dF(-b,l1-d); //v2 should be perpendicular to v1, because matrix is symmetric
            } else if (a<d) {
              v1 = point2dF(b,l1-a); v2 = point2dF(l1-a,-b);
            }
          }
          //normalize the vectors.
          v1.normalize(); if (v1.x<0.0f) v1 = v1*-1.f;
          v2.normalize(); if (v2.x<0.0f) v2 = v2*-1.f;
          
          //put the eigen vals and eigen vecs into the stored arrays.
          eigenVals_[indMat(i1,i2,0)] = l1;   eigenVals_[indMat(i1,i2,1)] = l2;
          eigenVec1_[indMat(i1,i2,0)] = v1.x; eigenVec1_[indMat(i1,i2,1)] = v1.y;
          eigenVec2_[indMat(i1,i2,0)] = v2.x; eigenVec2_[indMat(i1,i2,1)] = v2.y;
          
          //calculate dip info and linearity factors
          linearity_[indIn(i1,i2)]      = (l1-l2)/l1;
          (*dipAngleX_)[indIn(i1,i2)]   = rad2deg(float(atan2(v1.sx,v1.sz))); //dip meansures the angle between normal and vertical directions.
        }
      }
    } //end for i1,i2
  } else { //ndim=3
    VERIFY(0);
  } 
  //PRV(sizeof(int);PRV(sizeof(float)); PRV(sizeof(degree)); PRV(sizeof(PosInt)); PRV(sizeof(PosFloat));
  return;
}

const cubeDataF& dipEstimator::estNormal() { //corresponding to the biggest eigenvalue.
  return eigenVec1_;
}
const cubeDataF& dipEstimator::linearityFactor() {
  return linearity_;
}
const cubeDataF& dipEstimator::dipEstMask() {
  return mask_;
}
