#ifndef __DIPESTIMATOR_H_INCL__
#define __DIPESTIMATOR_H_INCL__
#include "param_func.h"

#include "stdafx.h"
#include "cubeData.hpp"

namespace Yang {
/**
 * TODO: Use structure tensors to measure dips
 * 
 * @author   yang
 */
  
using boost::shared_ptr;

struct StructTensorParam {
public:
  StructTensorParam(const string& prefix = ""): prefix_(prefix){}
  void readParam(const param_func* par);
public:
  PosFloat gs_sm_1,gs_sm_2; //Smoothing width of Ixx,Izz,and Izx field.
  NnegFloat maskThresRatio; //mask threshold for zero entries in the structure tensor matrix.
  string prefix_;
};

class dipEstimator {
public:
  // Constructor
  dipEstimator(const shared_ptr<const StructTensorParam>& par): 
    par_(par), dipAngleX_(new cubeDataF()),dipAngleY_(new cubeDataF()) {}
  void inputImgIs(const cubeDataF& cub);
  
  const cubeDataF& estNormal(); //in radians.
  const cubeDataF& linearityFactor();
  const cubeDataF& dipEstMask();
  const cubeDataF& dipAngleX_deg() {return (*dipAngleX_);}
  const cubeDataF& dipAngleY_deg() {return (*dipAngleY_);}
  
  // Destructor
  virtual ~dipEstimator();
  
private:
  shared_ptr<const StructTensorParam> par_;
  //mask out places where the signal is zero(pattern is uniform constant) thus no way to measure local dip
  cubeDataF mask_,linearity_; //normals are in radians.
  std::valarray<float> matrix_; //structure tensor matrix
  std::valarray<float> eigenVals_,eigenVec2_,eigenVec3_;
  cubeDataF eigenVec1_;
  shared_ptr<cubeDataF> dipAngleX_, dipAngleY_; //in radians.
};

// Destructor implementation
inline dipEstimator::~dipEstimator()
{
}

// TODO: Uncomment the copy constructor when you need it.
//inline dipEstimator::dipEstimator(const dipEstimator& src)
//{
//   // TODO: copy
//}

// TODO: Uncomment the assignment operator when you need it.
//inline dipEstimator& dipEstimator::operator=(const dipEstimator& rhs)
//{
//   if (this == &rhs) {
//      return *this;
//   }
//
//   // TODO: assignment
//
//   return *this;
//}

} //end namespace

#endif // __DIPESTIMATOR_H_INCL__
