#ifndef DM_H
#define DM_H

#include "axis.h"

#define REMOVE_DUMMY_AXIS_DEF false

//data IO manager, can be used as Input or Output, but not both at the same time.
namespace Yang {
   
//detect whether a tag is given by the cmd line / par files, Note this is dif from tag_exist in auxin.c
inline bool tagExist(const string& tag) {
	  bool ret = true;
	  char tmp[1024];
	  if (0==getch(tag.c_str(),"s",tmp)) {	//grab failed, no such tag
		 if(tag.compare("out")!=0 && tag.compare("in")!=0) {
             ret = false;
       }
	  }
	  return ret;
   }
   
class dm {
public:
   enum typeIO {
	  IN,
	  OUT,
	  INOUT
   };
   
protected: //members
   string    tag; //tag string
   string    title; //title in the .H file
   typeIO	 tIO;
   
   //The object will be dummy(e.g without the actual read/write capability) when
   //1. tag str is ""  2. tag str cannot be found in the cmd line
   bool		 m_bDummy;
   string	 m_szInfo; //use for identification(for debugging output)
   unsigned int esize;
public:
    //If bNoDummy=true, then this object is guaranteed to be not dummy
   dm(const string& _tag, typeIO _tIO, const string& szInfo, bool bNoDummy);
   
   bool canOutput() {return !(this->tIO==IN); }
   bool canInput()  {return !(this->tIO==OUT); }
   bool isOutput() {return (this->tIO==OUT); }
   bool isInput()  {return (this->tIO==IN); }
   int  getElemSize() {return this->esize;}
   int  isDummy() {return this->m_bDummy;}
   
   virtual ~dm() {}
   //virtual void unInit() = 0; //means save to FILE
   //virtual void init() = 0; //means load from FILE
   const string& getTag() const {return tag;}
   
   int      get_int(const char *arg) const;
   int      get_int(const char *arg, int def) const;
   float    get_float(const char *arg) const;
   float    get_float(const char *arg, float def) const;
   std::string get_string(const char* arg) const;
   std::string get_string(const char* arg, const std::string& def) const;
   
   int getNumDataAxes(const char* sztag, bool bRemoveDummyAxis);
   
   void put_int(const char *arg, int val) const;
   void	put_float(const char *arg, float val) const;
   void put_string(const char *arg, const std::string& val) const;


protected:
   axis     read_axis(int iax); //iax is 1 based
   void     write_axis(int iax, axis a);
};
}

#endif
