#include "stdafx.h"
#include "cubeData.hpp"
#include "gVec.h"

/**
 * Example of a member function that may be performed on 
 * this object.
 */
 
using namespace Yang;

//Basic clone functions
void gVecComponent::copyToVec(gVecBase* pVec) const {
  //first cast to gVecComponent	   
  gVecComponent* pVecComponent = dynamic_cast<gVecComponent*>(pVec);
  assert(pVecComponent);
  this->cubeData<float>::copyTo(*pVecComponent);
}

void gVecComponent::allocMemoryToVec(gVecBase* pVec) const{
   //first cast to gVecComponent	   
   gVecComponent* pVecComponent = dynamic_cast<gVecComponent*>(pVec);
   assert(pVecComponent);
   this->cubeData<float>::allocMemoryTo(*pVecComponent);
}

//calculate Z=X+a*Y or X=X+a*Y, can reuse code
void gVecComponent::scaleAddVec(const gVecBase* pXVec, const gVecBase* pYVec, float a) {
   const gVecComponent *p1,*p2;
   p1 = dynamic_cast<const gVecComponent*>(pXVec);
   p2 = dynamic_cast<const gVecComponent*>(pYVec);
   this->cubeData<float>::scaleAdd(*(p1->getDataCube()),*(p2->getDataCube()),a);
}

void gVecComponent::scaleAddVec(const gVecBase* pYVec, float a){
   this->scaleAddVec(this,pYVec,a);
}

//calc Z=X+Y or X=X+Y
void gVecComponent::addVec(const gVecBase* pXVec, const gVecBase* pYVec) {
   const gVecComponent *p1,*p2;
   p1 = dynamic_cast<const gVecComponent*>(pXVec);
   p2 = dynamic_cast<const gVecComponent*>(pYVec);
   this->cubeData<float>::add(*(p1->getDataCube()),*(p2->getDataCube()));
}

void gVecComponent::addVec(const gVecBase* pYVec) {
   this->addVec(this,pYVec);
}

void gVecComponent::subVec(const gVecBase* pXVec, const gVecBase* pYVec) {
   const gVecComponent *p1,*p2;
   p1 = dynamic_cast<const gVecComponent*>(pXVec);
   p2 = dynamic_cast<const gVecComponent*>(pYVec);
   this->cubeData<float>::subtract(*(p1->getDataCube()),*(p2->getDataCube()));
}

void gVecComponent::subVec(const gVecBase* pYVec) {
   this->subVec(this,pYVec);
}

//calc Z=a*X or X=a*x
void gVecComponent::scaleVec(const gVecBase* pXVec, float a) {
   //first cast to gVecComponent	   
   const gVecComponent* pVecComponent = dynamic_cast<const gVecComponent*>(pXVec);
   assert(pVecComponent);
   this->cubeData<float>::scale(*(pVecComponent->getDataCube()),a);
}

void gVecComponent::scaleVec(float a) {
   this->scaleVec(this,a);
}

void gVecComponent::MemberFunction() {
	// TODO: Component method implementation
}

/**
 * Example of a numeric property getter
 */
int gVecComponent::get_NumberProperty() {
	return _numberPropertyValue;
}

/**
 * Example of a numeric property setter
 */
void gVecComponent::set_NumberProperty(int value) {
	_numberPropertyValue = value;
}

/**
 * A component has no children, so return 0
 */
int gVecComponent::get_ChildCount() {
	return 0;
}

/**
 * A component has no children, so return null
 */
gVecIter gVecComponent::Children() {
	return gVecIter();
}

/**
 * AddChild has no implementation in the component class
 */
int gVecComponent::AddChild(gVecBase* child) {
	return 0;
}

/**
 * RemoveChild has no implementation in the component class
 */
int gVecComponent::RemoveChild(gVecBase* child) {
	return 0;
}



