
/// 2-D finite-difference gradient
///Yang Zhang
///Modified from RSF program below
#include "stdafx.h"
#include "decl_fwd.h"
#include "cubeData.hpp"
#include "cubeAlgo.h"
//#include "op.h"
#include "igrad2op.h"

using namespace Yang;
int igrad2op::oper(bool adj,bool add, cubeDataF& mod, cubeDataF& dat)
// linear operator, dat[n1*n2*2] is the gradient of mod[n1*n2] 
{
  int i1,i2,i;
  uint n12 = this->n1_*this->n2_;
  VERIFY(mod.domain().get_n123() == n12);
  VERIFY(dat.domain().get_n123() == n12*2);

  //Take out the memory ptr of mod and dat, then
  float *p = &mod[0]; float *r = &dat[0];
  int nm = mod.size(), nd = dat.size();
  addNull(adj,add,nm,nd,p,r);

  uint ind1,ind2;

  for (i2=0; i2 < n2_-1; i2++) {
    for (i1=0; i1 < n1_-1; i1++) {
      ind1 = IND2(i1,i2,n1_);
      ind2 = ind1 + n12;
      if (adj) {
        p[ind1+1]   += r[ind1]*kd1_;
        p[ind1+n1_] += r[ind2]*kd2_;
        p[ind1]     -= (r[ind1]*kd1_ + r[ind2]*kd2_);
      } else {
        r[ind1]     += (p[ind1+1]  - p[ind1])  *kd1_;  //divide by the grid size d1
        r[ind2]     += (p[ind1+n1_] - p[ind1]) *kd2_;
      }
    }
  }
  return 0;
}

int igrad3op::oper(bool adj,bool add, cubeDataF& mod, cubeDataF& dat)
// linear operator, dat[n1*n2*n3*3] is the gradient of mod[n1*n2*n3]
{
  int i1,i2,i3,n123;

  n123 = n1_*n2_*n3_;
  VERIFY(mod.domain().get_n123() == n123);
  VERIFY(dat.domain().get_n123() == n123*3);

  //Take out the memory ptr of mod and dat, then
  float *p = &mod[0]; float *r = &dat[0];
  addNull(adj,add,n123,n123*3,p,r);

  uint ind1,ind2,ind3;
  uint n12 = n1_*n2_;
  for (i3=0; i3 < n3_-1; i3++) {
  for (i2=0; i2 < n2_-1; i2++) {
  for (i1=0; i1 < n1_-1; i1++) {
    ind1 = (i3*n2_+i2)*n1_+i1;
    ind2 = ind1 + n123;
    ind3 = ind2 + n123;
    if (adj) {
      p[ind1+1]   += r[ind1];
      p[ind1+n1_]  += r[ind2];
      p[ind1+n12] += r[ind3];
      p[ind1]     -= (r[ind1] + r[ind2] + r[ind3]);
    } else {
      r[ind1]    += (p[ind1+1]    - p[ind1]);
      r[ind2]    += (p[ind1+n1_]  - p[ind1]);
      r[ind3]    += (p[ind1+n12]  - p[ind1]);
    }
  }}}
  return 0;
}

