#include "stdafx.h"
#include "cubeData.hpp"
#include "param_func.h"
#include "misc.h"

using namespace seplib;
using namespace Yang;
using namespace std;

//Read an axis parameter from cmd line, the n,o,ds should be n-suffix,o-suffix,d-suffix.
void Yang::read_axis_from_par(axis& ax, const param_func* pars, const string& suffix,int *ndef,float *odef,float *ddef) {
  string tag;
  tag = "n"+suffix; if (ndef) {ax.n = pars->get_int(tag,*ndef);} else {ax.n = pars->get_int(tag);}
  tag = "o"+suffix; if (odef) {ax.o = pars->get_float(tag,*odef);} else {ax.o = pars->get_float(tag);}
  tag = "d"+suffix; if (ddef) {ax.d = pars->get_float(tag,*ddef);} else {ax.d = pars->get_float(tag);}
  return;
}

//Non-max suppresion, currently handle only 1D.
int Yang::nmSup(const cubeDataF& _in, cubeDataF& _out, const NMSup_Param* par) {
  bool b2D = !(par->wnd2_ == 1);
  VERIFY(_out.sameSize(_in));
  const hypercube& shp = _in.domain();
  VERIFY(shp.get_ndim()<3);
  int n1,n2; n1 = shp.get_n(1); n2 = shp.get_n(2);
  int wnd1 = par->wnd1_, wnd2 = par->wnd2_;
  int i1,i2;
  
  int hfwnd1 = (wnd1+1)/2;
  if (!b2D) {
    float fmax; int imax; int ihead,itail;
    for (i2=0; i2<n2; i2++) {
      //do the trace beginning,the edge case.
      const float* pmax = std::max_element(&_in(i2,0),&_in(i2,hfwnd1));
      const float* p    = &_in(i2,0);
      imax = pmax - &_in(i2,0); fmax = *pmax;
      //slide through the first half window
      itail=hfwnd1-1;
      for (i1=0; i1<=wnd1-hfwnd1; i1++) { //suppress the non-maximum values.
        _out(i2,i1) = 0;
        if (p[itail]>fmax) {fmax=p[itail], imax=itail;} //see if need update fmax
        else {
          if (i1==imax) _out(i2,i1)=p[i1]; //this is loca maximum
        }
        itail++;
      }
      
      //handle the center of the trace.
      ihead=0; itail=wnd1;
      //then slide through the trace to the end.
      for (i1=wnd1-hfwnd1+1;itail<n1;i1++) {
        _out(i2,i1) = 0;
        //ihead is leaving, and itail is entering. first update the max value within the window.
        if (imax==ihead) { //needs to track a new maxima
          pmax = std::max_element(p+ihead+1,p+itail+1);
          imax = pmax - p; fmax = *pmax;
        } else { //deal with itail entering the scope
          if (p[itail]>fmax) {fmax=p[itail], imax=itail;} //see if need update fmax
        }
        if (i1==imax) _out(i2,i1)=p[i1]; //this is local maximum
        ihead++;itail++;
      }
      
      //handle the end of trace,itail is fixed at n1 now.
      for ( ; i1<n1; i1++,ihead++) { 
        _out(i2,i1) = 0;
        //ihead is leaving, update the local maximum
        if (imax==ihead) { //needs to track a new maxima
          pmax = std::max_element(p+ihead+1,p+n1);
          imax = pmax - p; fmax = *pmax;
        }
        if (i1==imax) _out(i2,i1)=p[i1]; //this is loca maximum, mark it.
      }
    } //for each trace.
  } else { //2D case.
    VERIFY(0); //2D not implemented yet.
  }
  return 0;
}

