#include "stdafx.h"

#include "velFD2d.h"
#include "modelingFD2d_op.h"

using namespace Yang;

void modelingFD2d_op::initOperator(waveOpParams* params, action_type ac_type)
{
   modelingFD_op::initOperator(params,ac_type);
   axis ax_x,ax_z;
   
   //should be done every shot
   //verbPrt("mtime=%f\n",mTime);
   ax_z = params->modelGridShape.get_axis(1);
   ax_x = params->modelGridShape.get_axis(2);
   m_dz = ax_z.d;  m_dx = ax_x.d;
   //verbPrt("mtime=%f\n",mTime);
}

//Calculate Spatial Derivatives times dt/(dx^2)
void modelingFD2d_op::calcSpatialDeriv( cubeData<ETYPE> *in, cubeData<ETYPE> * out ) 
{
   //input is a 2D memory array of size(nx,nz), e.a. same as wavefield output is a 2D memory array
   //of size(nx,nz) Verify the array size first 
   assert(out->domain().sameSize(m_pars->mShapeWavfield));
   assert(in->domain().sameSize(m_pars->mShapeWavfield));
   
   // Apply the stencil along each direction 
   int ix,iz,iSt;
   
   ETYPE dtDdx2,dtDdz2,tmp; //(dt/dx)^2
   float dtComp = m_pars->timeParam.dtComp;
   tmp = (dtComp/m_dx); dtDdx2 = tmp * tmp;
   tmp = (dtComp/m_dz); dtDdz2 = tmp * tmp;

   ETYPE v2,wsm1,wsm2;
   assert(int(stencil.size()-1) == m_pars->stencilWidth);
   const double* sten = &(stencil[0]);
   
   //access velocity and wavefield array here !!!
   cubeData<ETYPE>& input = *in;
   cubeData<ETYPE>& output = *out;
   out->zero();
   
   const cubeData<float>&v = m_pars->pVel->get2dVelPtr();

   //Use 2nd order along the boundary
   //Use mStendWid th order in the middle 
   //Remember the Wavefield is of size(nx,nz), same as the velocity field; 
   //After calculation, the valid region shrink to (nx-2,nz-2) 


   //Do the middle first
   int offx = this->mOffsetPandModel[1];
   int offz = this->mOffsetPandModel[0];
   //PRV(offx);
   //PRV(offz);
   int nx,nz;
   nz = m_pars->mShapeWavfield.get_n(1);
   nx = m_pars->mShapeWavfield.get_n(2);
   
   int mStenWid = m_pars->stencilWidth;
   float * __restrict ps1;
   float * __restrict ps2;
   
#define CHUNKSIZE

   //should use openMP here !! openMP
   #pragma omp parallel private(iz,tmp,v2,wsm1,wsm2,iSt,ps1,ps2)
   {
   //#pragma omp for schedule(dynamic,CHUNKSIZE)
   #pragma omp for 
   for (ix=mStenWid; ix<nx-mStenWid; ix++)
   {
     ps1 = &input(ix,0); ps2 = &output(ix,0);
      for (iz=mStenWid;iz<nz-mStenWid; iz++)
      {
         tmp = v(ix-offx,iz-offz);  //velocity square
         v2 = tmp*tmp;
         
         //Along z direction, calculate the spatial derivative
         wsm1 = ps1[iz]*sten[0];
		 wsm2 = wsm1;
		 //unable to expand the loop will cause performance degrade.
         for (iSt=1; iSt <= mStenWid; iSt++)
         {
            wsm1 += ( ps1[iz+iSt] + ps1[iz-iSt] ) * sten[iSt]; //z direction
            wsm2 +=  ( input(ix+iSt,iz) + input(ix-iSt,iz) ) * sten[iSt]; //x direction
         }
         ps2[iz] = (wsm1*dtDdz2 + wsm2*dtDdx2) * v2;
      }
   }
   
   
   //assume mStenWid == offx == offz, we don't need to do the stencil on boundary.
   
   /*
   //Here, the number 1 and offx,offz is still related !! ??	
   //Then do the peripheral layer with only 2nd order
   //do the left and right
   #pragma omp for
   for (ix=1; ix<nx-1; ix++)
   {
      if(ix>=mStenWid && ix<nx-mStenWid) {
         continue;
      }
      for (iz=1; iz<nz-1; iz++)
      {	 
         assert(output[ix][iz]==0);
         tmp = v[ix-offx][iz-offz];  //velocity square
         v2 = tmp*tmp;
         
         //Along z direction, calculate the spatial derivative
         wsm = 0;
         wsm += input[ix][iz]* (-2.0);
         wsm += ( input[ix][iz+1] + input[ix][iz-1] ) * 1.0;
         output[ix][iz] = wsm * v2 * dtDdz2;

         //Along x direction
         wsm = 0;
         wsm += input[ix][iz]* (-2.0);
         wsm += ( input[ix+1][iz] + input[ix-1][iz] ) * 1.0;
         output[ix][iz] += wsm * v2 * dtDdx2;
      }
   }
   
   #pragma omp for
   //2nd order, boundary part. Do the top and bottom
   for (ix=mStenWid; ix<nx-mStenWid; ix++)
   {_
      for (iz=1; iz<nz-1; iz++)
      {
		 if(iz>=mStenWid && iz<nz-mStenWid) {
			 continue;
		 }
         assert(output[ix][iz]==0);
         tmp = v[ix-offx][iz-offz];  //velocity square
         v2 = tmp*tmp;
         
         //Along z direction, calculate the spatial derivative
         wsm = 0;
         wsm += input[ix][iz]* (-2.0);
         wsm += ( input[ix][iz+1] + input[ix][iz-1] ) * 1.0;
         output[ix][iz] = wsm * v2 * dtDdz2;

         //Along x direction
         wsm = 0;
         wsm += input[ix][iz]* (-2.0);
         wsm += ( input[ix+1][iz] + input[ix-1][iz] ) * 1.0;
         output[ix][iz] += wsm * v2 * dtDdx2;
      }
   }
   */
   
   } //end OMP parallel
   
}

void modelingFD2d_op::propagate_2t(const cubeData<ETYPE>& p0cube, const cubeData<ETYPE>& p1cube, cubeData<ETYPE>& p2cube) //second order in time
{

   //get the 2D pointer for three cubeData
   const cubeData<ETYPE> &pp1=p1cube, &pp0=p0cube;
   cubeData<ETYPE>& pp2=p2cube;
     
   int nx,nz;
   nz = m_pars->mShapeWavfield.get_n(1);
   nx = m_pars->mShapeWavfield.get_n(2);
   
   int offx = m_pars->offx;
   int offz = m_pars->offz;
   
   assert(p2cube.domain().sameSize(m_pars->mShapeWavfield));
   
   
   //Calculate the sptial deriv times (dt/dx)^2
   //Apply the stencil along each direction 
   int ix,iz,iSt;
   
   ETYPE dtDdx2,dtDdz2,tmp; //(dt/dx)^2
   float dtComp = m_pars->timeParam.dtComp;
   tmp = (dtComp/m_dx); dtDdx2 = tmp * tmp;
   tmp = (dtComp/m_dz); dtDdz2 = tmp * tmp;

   ETYPE v2,wsm1,wsm2;
   assert(int(stencil.size()-1) == m_pars->stencilWidth);
   const double* sten = &(stencil[0]);
   
   //access velocity and wavefield array here !!!

   const cubeData<float> &v = m_pars->pVel->get2dVelPtr();

   //Use mStendWid th order in the middle 
   //Remember the Wavefield is of size(nx,nz), same as the velocity field; 
   //After calculation, the valid region shrink to (nx-2*mStenWid,nz-2*mStenWid) 


   int mStenWid = m_pars->stencilWidth;
   const float * __restrict ps1;
   float * __restrict ps2;
   
#define CHUNKSIZE

   //should use openMP here !! openMP
   //#pragma omp parallel private(iz,tmp,v2,wsm1,wsm2,iSt,ps1,ps2)
   {
   //#pragma omp for schedule(dynamic,CHUNKSIZE)
   //#pragma omp for 
   for (ix=mStenWid; ix<nx-mStenWid; ix++)
   {
     ps1 = &pp1(ix,0); ps2 = &pp2(ix,0);
      for (iz=mStenWid;iz<nz-mStenWid; iz++)
      {
         tmp = v(ix-offx,iz-offz);  //velocity square
         v2 = tmp*tmp;
         
         //Along z direction, calculate the spatial derivative
         wsm1 = ps1[iz]*sten[0];
		 wsm2 = wsm1;
		 //unable to expand the loop will cause performance degrade.
         for (iSt=1; iSt <= mStenWid; iSt++)
         {
            wsm1 += ( ps1[iz+iSt] + ps1[iz-iSt] ) * sten[iSt]; //z direction
            wsm2 +=  (pp1(ix+iSt,iz) + pp1(ix-iSt,iz) ) * sten[iSt]; //x direction
         }
         ps2[iz] = (2.0*pp1(ix,iz)-pp0(ix,iz) + (wsm1*dtDdz2 + wsm2*dtDdx2) * v2);
      }
   }
   }
}

void modelingFD_op::propagate(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube)
{
   if (m_t_accuracy == order4)
      this->propagate_4t(p0cube,p1cube,p2cube);
   else if (m_t_accuracy == order2)
      this->propagate_2t(p0cube,p1cube,p2cube);
   else
      seperr("!error, m_t_accuracy");
}

//4th order in time, add a correction term
void modelingFD2d_op::propagate_4t(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube) 
{
   assert( m_t_accuracy == modelingFD_op::order4);

   int ix,iz, nz,nx;
   nz = m_pars->mShapeWavfield.get_n(1);
   nx = m_pars->mShapeWavfield.get_n(2);

   const cubeData<ETYPE> &pp0 = p0cube, &pp1=p1cube;
   cubeData<ETYPE>& pp2=p2cube;

   assert(dpDdx2_cb.sameSize(p0cube));

   //Calculate the sptial deriv times (dt/dx)^2, save to dpDdx4_cb
   ETYPE dtDdx2,dtDdz2,tmp; //(dt/dx)^2
   float dtComp = m_pars->timeParam.dtComp;
   tmp = (dtComp/m_dx); dtDdx2 = tmp * tmp;
   tmp = (dtComp/m_dz); dtDdz2 = tmp * tmp;

   ETYPE v2,wsm1,wsm2;
   assert(int(stencil.size()-1) == m_pars->stencilWidth);
   const double* sten = &(stencil[0]);
   
   //access velocity and wavefield array here !!!

   const cubeData<float>& v = m_pars->pVel->get2dVelPtr();

   //Use mStendWid th order in the middle 
   //Remember the Wavefield is of size(nx,nz), same as the velocity field; 
   int mStenWid = m_pars->stencilWidth;
   const float * __restrict ps1;
   float * __restrict ps2;
   int iSt; int offx,offz;
   offx = m_pars->offx; offz = m_pars->offz;
   //should use openMP here !! openMP
   //#pragma omp parallel private(iz,tmp,v2,wsm1,wsm2,iSt,ps1,ps2)
   {
   //#pragma omp for schedule(dynamic,CHUNKSIZE)
   //#pragma omp for 
   for (ix=mStenWid; ix<nx-mStenWid; ix++)
   {
	  ps1 = &pp1(ix,0); ps2 = &pp2(ix,0);
      for (iz=mStenWid;iz<nz-mStenWid; iz++)
      {
         tmp = v(ix-offx,iz-offz);  //velocity square
         v2 = tmp*tmp;
         
         //Along z direction, calculate the spatial derivative
         wsm1 = ps1[iz]*sten[0];
		 wsm2 = wsm1;
		 //unable to expand the loop will cause performance degrade.
         for (iSt=1; iSt <= mStenWid; iSt++)
         {
            wsm1 += (ps1[iz+iSt] + ps1[iz-iSt] ) * sten[iSt]; //z direction
            wsm2 += (pp1(ix+iSt,iz) + pp1(ix-iSt,iz) ) * sten[iSt]; //x direction
         }
         tmp =  (wsm1*dtDdz2 + wsm2*dtDdx2) * v2;
         dpDdx2_cb(ix,iz) = tmp;
         ps2[iz] = (2.0*pp1(ix,iz)-pp0(ix,iz)) + tmp;
      }
   }
   }
   

   //Then calculate the spatial deriv again to enable 4th order in time 
   //update 4th order term, shrink the updated region at the boundary
      //should use openMP here !! openMP
   //#pragma omp parallel private(iz,tmp,v2,wsm1,wsm2,iSt,ps1,ps2)
   {
   //#pragma omp for schedule(dynamic,CHUNKSIZE)
   //#pragma omp for 
   for (ix=2*mStenWid; ix<nx-2*mStenWid; ix++)
   {
	  ps1 = &(dpDdx2_cb(ix,0)); ps2 = &pp2(ix,0);
      for (iz=2*mStenWid;iz<nz-2*mStenWid; iz++)
      {
         tmp = v(ix-offx,iz-offz);  //velocity square
         v2 = tmp*tmp;
         
         //Along z direction, calculate the spatial derivative
         wsm1 = ps1[iz]*sten[0];
		 wsm2 = wsm1;
		 //unable to expand the loop will cause performance degrade.
         for (iSt=1; iSt <= mStenWid; iSt++)
         {
            wsm1 += ( ps1[iz+iSt] + ps1[iz-iSt] ) * sten[iSt]; //z direction
            wsm2 +=  ( pp1(ix+iSt,iz) + pp1(ix-iSt,iz) ) * sten[iSt]; //x direction
         }
         tmp =  (wsm1*dtDdz2 + wsm2*dtDdx2) * v2;
         //4th order correction
         ps2[iz] +=  (1.0/12.0f)*tmp;
      }
   }
   }
}


