#include "stdafx.h"

#include "waveOps.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "SR_manager.h"
#include "boundaryFD.h"
#include "paddingFD.h"
#include "cubeAlgo2.h"

#include "interp.h"

using namespace Yang;

//For each shot, it might be super shot or simple point shot. 
void modelingFD_op::modelingPerShot(int ishot) {
   //initize has been done in initPerShot
   
   //At time step it, calcuate the wavefield at time coord it
   // p0 -> it,it+1,it+2
	for (int it=0; it < m_pars->timeParam.Nt; it++) //simulate from it=0->m_Nt, record at 0, jRec, 2*jRec,...
		//Doing wave propagation for each time step, here the 
		//upperbound of the for loop cannot change, 
		//otherwise the size of snapshot and seisRec will be changing
		{
       	    //Exert Src at it+1
		    m_pSR_manager->exertSrc(ishot,it,*(wfGp0.pp_cb[1])); //update the source at time t+1, not t or t-1
		    //do Sponge layer at the padding region at it+1
		    m_pPaddingFD->updateEx(wfGp0);
            m_pBoundaryFD->update(wfGp0);
   
			//it is possible src_recv geometry is an irregular one, let SR_manager to handle this
            //record at time = it
			m_pSR_manager->record(ishot,it,*(wfGp0.pp_cb[0]));
			
			//take snapshot, this part should always combined with propagator
			this->takeSnapshot(ishot,it,*(wfGp0.pp_cb[0]),m_pars->pSrcSnapshot);
            
            this->propagate(*wfGp0.pp_cb[0],*wfGp0.pp_cb[1],*wfGp0.pp_cb[2]);
            //switch the array pointers, advance the time stamp, now it+1,it+2,it+3
            wfGp0.switchWavfieldPtrs();

	     }
    //Output one shot record to file, here Recv geometry might be irregular
    m_pSR_manager->finishOneShot(ishot);
}

//Do the wave-equation modeling, non-linear, output to pRsfSeisRec
void modelingFD_op::modeling()
{
   verbPrt("start modeling");
   VERIFY(this->mAction == Yang::acModeling);
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++) {
      m_pSR_manager->initPerShot(ishot,Yang::acModeling);
	  
	  verbPrt("\r modeling ishot=%d...", ishot);
      initPerShot(ishot);
	  modelingPerShot(ishot);
   }
}

void modelingFDgen_op::getSnapshotShapePerShot(hypercube& shapeSnapshot) {
   //Get the snapshot cube
   vector<axis> shapeSnapshotAxes = (m_pars->mShapeWavfield).get_axes();
   float o = 0.0; float d = m_pars->timeParam.dtSnap; int n = m_pars->timeParam.nSnap;
   axis axisT(n,o,d,"time","sec");
   shapeSnapshotAxes.push_back(axisT);
   shapeSnapshot.set_axes(shapeSnapshotAxes);   
}

void modelingFDgen_op::getShotAxisForSnap(axis& ax_shot) {
    axis ax_x = m_pars->velModelShape.get_axis(2);
    float d = ax_x.d; float o = ax_x.o;
    ax_shot.d = (m_pars->j_shot_snap)*(m_pars->mShotIncX)*d; ax_shot.n = m_pars->n_shot_snap;
    ax_shot.o = o+(m_pars->mShot1_ix)*d + (m_pars->f_shot_snap)*ax_shot.d;
}

bool modelingFDgen_op::shouldRecordSnap(int ishot, int* iIndexOnAxis) {
   bool b1 = (((ishot-m_pars->f_shot_snap) / m_pars->j_shot_snap) < m_pars->n_shot_snap);
   bool b2 = ((ishot-m_pars->f_shot_snap)%m_pars->j_shot_snap == 0);
   bool bRecordSnap = b1 && b2;
   if (bRecordSnap && NULL!=iIndexOnAxis) {
      *iIndexOnAxis = (ishot-m_pars->f_shot_snap) / m_pars->j_shot_snap;
   }
   return bRecordSnap;
}

//Forward Progagation of Src Wavefield
void modelingFD_op::fwdSrcPropagation(int iShot)
{
   //Doing wave propagation for each time step, here the upperbound of the for loop 
   //cannot change, otherwise the size of snapshot and seisRec will be changing.
   //Here we try to calculate wavefield at time it, given the slices from it-2, and it-1
   for (int it=0; it < m_pars->timeParam.Nt; it++)
	  {
         //Exert Src, irregular shot locations.
		 m_pSR_manager->exertSrc(iShot,it,*(wfGp0.pp_cb[0])); //update the source at time t+1, not t or t-1

		 //do Sponge layer at the padding region at it+1
		 m_pPaddingFD->updateEx(wfGp0);
		 //do boundary conditon
		 m_pBoundaryFD->update(wfGp0);
	   
		 //take snapshot
		 this->takeSnapshot(iShot,it,*(wfGp0.pp_cb[0]),m_pars->pSrcSnapshot);
		 //Take checkpt
		 this->takeWavCheckpt(iShot,it,wfGp0);
         
         //propagate to time it+2
         this->propagate(*wfGp0.pp_cb[0],*wfGp0.pp_cb[1],*wfGp0.pp_cb[2]);
         //switch the array pointers, advance the time stamp
         wfGp0.switchWavfieldPtrs();
	  }
}

//Back projection of receiver wavefield, while back projecting the wavefield, do correlation
//with the src propagating wavefield
//The operator take seisRecord as input, output a Image
void modelingFD_op::RTM() {

	//should reset the input data stream, this should be the user's responsibility

	//Init the memories, done by initOperator
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++)  { //For each shot, it might be super 
	  prt("\r Start iShot %d...",ishot);
	  
      //shot or simple point shot. 
	  this->initPerShot(ishot);
	  m_pSR_manager->initPerShot(ishot,Yang::acMigration);

	  //Src Receiver Geometry, and seis-record-data are in the object SR_manager
	  this->fwdSrcPropagation(ishot);
	  
	  //Need to interpolate the Receiver record first
      m_pSR_manager->interpRecvRec(ishot);
	  //m_pSR_manager->mInterpSeisCube.dump("dump5");
	  
	  //clear the wave field to zero 
	  wfGp0.reset();
	  dpDdx2_cb.zero();
	  //at time step it, calcuate the wavefield at time coord "it-2", using the wavefield from it and it-1
      //p0->it,it-1,it-2
      for (int it=m_pars->timeParam.Nt-1; it >= 0; it--)  
      //Doing wave propagation for each time step, backward in time, here the 
	  //upperbound of the for loop cannot change, otherwise the size of snapshot and seisRec will be changing.
	  {
		 //inject the Recv wavefield at it-1
		 m_pSR_manager->injectRecv(ishot,it,*(wfGp0.pp_cb[1])); //update wavefiled at time t-1, not t or t+1

		 //do Sponge layer at the padding region
		 m_pPaddingFD->updateEx(wfGp0);
		 //do boundary conditon
		 m_pBoundaryFD->update(wfGp0);
	   

		 //take snapshot, might be useful for debugging
		 this->takeRecvWavfieldSnapshot(ishot,it,*(wfGp0.pp_cb[0]),m_pars->pRecvSnapshot);
		 
		 //CrossCorrelate the current wavefield and the src wavefield, and accumulate it to a wavefield cube
		 //pp_cb[0] as the wavefile at time it;
         if ( 0 == it%m_pars->timeParam.jImaging) {
            if (!(m_pars->bTake2ndTimeDerivForRTM)) {
               this->correlateSrcRecv(ishot,it,*wfGp0.pp_cb[0]);
            } else {
               //Correlate with 2nd deriv of wavefield at time it
               this->correlateSrcRecv(ishot,it,dpDdx2_cb);
            }
         }
         
         //Propagate the receiver wavefield for one step, basically applying the finite difference stencils
         this->propagate(*wfGp0.pp_cb[0],*wfGp0.pp_cb[1],*wfGp0.pp_cb[2]);
         this->calcSpatialDeriv(wfGp0.pp_cb[1],&dpDdx2_cb);
         //switch the array pointers, advance the time stamp
	     wfGp0.switchWavfieldPtrs();
	  }
      //Image has the size of current vel model, need to be assembled to a final image later.
      this->genPerShotImage(ishot);
   } //end loop iShot

} //end of func backProjRecv

void modelingFDgen_op::cropWfldToVelGrid(cubeData<float>& velOutCube, cubeData<float>& wfldInCube) {
   //There could be extra axes than z,x. thus
   int ndim = m_pars->modelGridShape.get_ndim();
   int nAxes = wfldInCube.domain().get_ndim();
   
   windParam window(nAxes,windowAxis());
   window[0].f = m_pars->offz;
   window[1].f = m_pars->offx;
   window[0].n = m_pars->modelGridShape.get_n(1);
   window[1].n = m_pars->modelGridShape.get_n(2);
   
   if (ndim > 2) {//3D
	  window[2].f = m_pars->offy;
	  window[2].n = m_pars->modelGridShape.get_n(3);
   }
   wfldInCube.grabWindow(&velOutCube,window);
}

//Operator, this could output postStack image or prestack image I(x,h).
//Notice, this case, we assume in/output are wfld size.
void modelingFDgen_op::migration_op(dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube) {
	FUNC_TRACE_BEG(modelingFDgen_op::migration_op);
  //Init the memories, done by initOperator
  m_pars->pRsfSeisRec = pRsfSeisRec;
	m_pars->pRsfSeisRec->resetCounter();
	bool bDoOffsetGather = m_pars->mOffImgParams.bDoOffsetGather;
  if (bDoOffsetGather) {
    m_pars->pFinalOffImage = pImageCube;
  } else {
    m_pars->pFinalImage = pImageCube;
  }
  
	//Sanitary check on input/output
  this->checkInOutSize();
										 
  //Do RTM
  this->RTM();

  //m_pars->pRsfSeisRec->resetCounter(); this->RTM();

  //Now the result should be at pImageCube
  //If sparse sample the final image, it equals to apply a 0/1 mask on the image.
  if (m_pars->pImageMask) {
    pImageCube->multiply(*(m_pars->pImageMask));
  }
  FUNC_TRACE_END();
}

//The modeling Operator's interface as an operator
//input model: velocity cube
//output data: pRsfSeisRec
void modelingFDgen_op::modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec)
{
  FUNC_TRACE_BEG(modelingFDgen_op::modeling_op);
  if (pVelCube == NULL) { //assume the velocity is already in place
    ;
  } else {
    VERIFY(m_pars->velModelShape.sameSize(pVelCube->domain()));
    m_pars->pVel->setVelocity(*pVelCube);
  }
  m_pars->pRsfSeisRec = pRsfSeisRec;

  checkInOutSize();
  
  modeling();
  FUNC_TRACE_END();
}

//Not using it right now
void modelingFDgen_op::bornModeling_op(cubeData<float>* pReflCube, dmRsf<float>* pRsfSeisRec) {
  FUNC_TRACE_BEG(modelingFDgen_op::bornModeling_op);
	
	//sanitary check on input
	if (m_pars->mOffImgParams.bDoOffsetGather) {
		m_pars->pFinalOffImage = pReflCube;
		m_pars->pFinalOffImage->resetCounter();
	} else {
		m_pars->pFinalImage= pReflCube;
		m_pars->pFinalImage->resetCounter();
	}
  //set the output
	m_pars->pRsfSeisRec = pRsfSeisRec;

  checkInOutSize();
	this->adjRTM();
	FUNC_TRACE_END();
}

//DeltaIh could be zero-offset or multiple offset
void modelingFDgen_op::TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS) {
	FUNC_TRACE_BEG(TomAdj_op);
  //see if do offset-image or zero-offset image
  bool bOffset = m_pars->mOffImgParams.bDoOffsetGather;
  pDeltaIh->resetCounter();
  //Set input/output
  if (bOffset) {
   m_pars->pFinalOffImage = pDeltaIh;
  } else {
   m_pars->pFinalImage = pDeltaIh;
  }
  m_pars->pFinalDeltaS = pDeltaS;
  this->checkInOutSize();
  this->TomAdj();
	FUNC_TRACE_END();
}

void modelingFDgen_op::TomFwd_op (cubeData<float>* pDeltaS, cubeData<float>* pDeltaI) {
	FUNC_TRACE_BEG(TomFwd_op);
	//Set In/out
	pDeltaS->resetCounter();
	m_pars->pFinalDeltaS = pDeltaS;
  
	//set output
	bool bDoOffsetGather = m_pars->mOffImgParams.bDoOffsetGather;
	if (bDoOffsetGather) {
    m_pars->pFinalOffImage = pDeltaI;
	} else {
    m_pars->pFinalImage = pDeltaI;
	}
  checkInOutSize();
	this->TomFwd();
	FUNC_TRACE_END();
}

void modelingFDgen_op::TomAdj()
{
   //!!!! Need revisement
   VERIFY(this->mAction == Yang::acTomAdj);
   /***
   gpu2dTomAdj_init(&(m_pars->timeParam.jSnap),&m_gpuID);
   //calculate deltaS for each shot
   prt("\r Start GPU tomo Adj (vel backProj) operator \n"
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++) {
	  prt("\r Start iShot = %d", ishot)
	  this->initPerShot(ishot);
      m_pSR_manager->initPerShot(ishot,Yang::acTomAdj);
	  this->TomAdj_perShot(ishot);
	  this->genPerShotImage(ishot);
   }
   //The data and snap shot should be ready by now.
   gpu2dTomAdj_uninit();
   **/
}

void modelingFDgen_op::TomFwd() {
   VERIFY(this->mAction == Yang::acTomFwd);
   /** !!!!! Need revisement
   verbPrt("\n Start CPU WETomo Forward ...");
   
   gpu2dTomFwd_init(&(m_pars->timeParam.jSnap),&m_gpuID);

   prt("\n");
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++)  { //For each shot, it might be super shot
	  prt("\r Start iShot %d...",ishot);
	  this->initPerShot(ishot);
      //shot or simple point shot. 
	  m_pSR_manager->initPerShot(ishot,Yang::acTomFwd);
	  
	  //Need to interpolate the Receiver record first
      m_pSR_manager->interpRecvRec(ishot);
	  //m_pSR_manager->mInterpSeisCube is still not the size of vel_nx*vel_nz
 
	  this->TomFwd_perShot(ishot);
	  
	  //Image has the size of current vel model, need to be assembled to a final image later.
	  //Save the RTM images to m_pars structures.
	  this->genPerShotImage(ishot);
   } //end loop iShot
   //The final image is now at m_pars->pFinalImage, same size as wavefield.
   
   gpu2dTomFwd_uninit();
   **/
}

//Do the AdjRTM, i.e. Born modeling
void modelingFD_op::adjRTM()
{
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++) {
      m_pSR_manager->initPerShot(ishot,Yang::acBornModeling);
      initPerShot(ishot);
	  adjRTM_perShot(ishot);
   }
}

//This operator does not apply second time derivative
//i.e. The equation is " d(xr,xs,w) = sum_{x}{fs(w)G(x,xs,w)G(xr,x)I(x)}", no "w^2" term.
void modelingFD_op::adjRTM_perShot(int ishot)
{	  
	  //Align refl model to the wavefield, can use interpolation as a substitute
   	  bool bDoOffImg = m_pars->mOffImgParams.bDoOffsetGather; 
	  if (!bDoOffImg) {
         mReflCubePerShot.alloc(m_pars->mShapeWavfield);
		 alignCropToSliceShapeGrid(mReflCubePerShot,*(m_pars->pFinalImage),m_pars->mShapeWavfield);
	  } else {
         hypercube offImgShape;
         get_prestack_img_cube_shape(offImgShape, m_pars->mOffImgParams, m_pars->mShapeWavfield);
         mReflOffCubePerShot.alloc(offImgShape);
		 alignCropToSliceShapeGrid(mReflOffCubePerShot,*(m_pars->pFinalOffImage),m_pars->mShapeWavfield);
	  }
	  
      dpDdx2_cb.zero();
	  //Here, propagate the src wavefield, correlate with the reflectivity model, sum to another upcoming
	  //wavefield, and propagate that upcoming wavefield simultaneously
	  
	  //calculate the source(down going) wavefield at time it+2, given the wav slice at it, and it+1
	  //calculate the scatter(up going) wavefield at time it+2, given the slice at it and it+1.
      for (int it=0; it < m_pars->timeParam.Nt; it++) //simulate from it=0->m_Nt, record at 0, jRec, 2*jRec,...
         //Doing wave propagation for each time step, here the 
         //upperbound of the for loop cannot change, 
         //otherwise the size of snapshot and seisRec will be changing
         {
            //Exert Src at time it+1
            m_pSR_manager->exertSrc(ishot,it,*(wfGp0.pp_cb[1])); //update the source at time t+1, not t			
            
			//then scatter to the second wavefield
			//First correlate the wavefield with ReflModel, and inject the corrlation as a src, remember it should
			//be injected to the slice with time it+1
			if (0 == it % m_pars->timeParam.jImaging)
			{
			   if (!bDoOffImg) { //use zero-offset image

                  //wfGp0.pp_cb[0]->copyTo(mCorrCube);
                  //In this case, the 2nd derivative of the background wavefield is injected, to get rid of that low freq artifact, since second derive in time is the same as
                  //Second derive in spaec.
                  dpDdx2_cb.copyTo(mCorrCube);
				  mCorrCube.multiply(mReflCubePerShot);
				  //mCorrCube.scale(m_pars->timeParam.dtImaging);
				  wfGp1.pp_cb[1]->add(mCorrCube);
			   }
			   else { //use multi-offset image
                  VERIFY(0);
				  //!!!!!!! To do
			   }
			}
			
            //do Sponge layer at the padding region at time it+1
            m_pPaddingFD->updateEx(wfGp0);
            
            //do boundary conditon, in born modeling, cannot apply boundary condition on wavefield, because the growing DC component in wfld
            m_pBoundaryFD->update(wfGp0);

            //calculate at time it+2
            this->propagate(*wfGp0.pp_cb[0],*wfGp0.pp_cb[1],*wfGp0.pp_cb[2]);
            //should record the 2nd derivative of wfl for use in next iteration
            this->calcSpatialDeriv(wfGp0.pp_cb[1],&dpDdx2_cb);

			m_pPaddingFD->updateEx(wfGp1);
			m_pBoundaryFD->update(wfGp1);
            //Propagate the wavefield to it+2
			this->propagate(*wfGp1.pp_cb[0],*wfGp1.pp_cb[1],*wfGp1.pp_cb[2]);			
            
            //remember we record the wavefield at time it., now is it,it+1,it+2
            m_pSR_manager->record(ishot,it,*(wfGp1.pp_cb[0])); //record the second wavefield here

            this->takeSnapshot(ishot,it,*(wfGp0.pp_cb[0]),m_pars->pSrcSnapshot);
            this->takeSnapshot(ishot,it,*(wfGp1.pp_cb[0]),m_pars->pSrcSnapshot2);
            
			if (it%100==0)
			{
			   verbPrt("finish it=%d\n",it);
			}
            //switch the array pointers, now p0 -> it+1,it+2,it+3
            wfGp0.switchWavfieldPtrs();
            wfGp1.switchWavfieldPtrs();
         }
	   //Output one shot record to file, here Recv geometry might be irregular
      m_pSR_manager->finishOneShot(ishot);
}

void modelingFD_op::imposeRecv(wavFieldGroup& wfg)
{
   VERIFY(0);
   //if impose the receiver as src, then it is fine to use the same propagator as in modeling
   switch(m_pars->mImposeRecvType) 
   {
   case AsForce:
      //m_pSR_manager->injectRecv(
      break;
   case AsBoundary:
      //this->propagateRecvAsBoundary();
      seperr("this->propagateRecvAsBoundary not implemented yet");
      break;
   default:
      seperr("unknown mImposeRecvType");
   }
}

void modelingFD_op::takeSnapshot(int ishot, int it, cubeData<ETYPE>& wavfield,
                                 dmRsf<float>* pSnapShot ){ //[1]
   if( it % m_pars->timeParam.jSnap == 0 && shouldRecordSnap(ishot)) //Record
   {
      //int tIndex = it/m_jSnap;
      //Record the snapshot wavefield at pp_cb[1], pp_cb is of ETYPE
      #ifndef USE_DB_PRECISION
		//update the snapshot object for that particular shot index
      pSnapShot->seq_update(&wavfield);
      #else //Copy the wavefield in pp_cb[1] to SnapCube(float type)
      //seperr("!!!!Save wavefield snapshot of double precison not implemented yet");
      copyCubeData<double,float>(wavfield,mSnapCube);
      pSnapShot->seq_update(&mSnapCube);
      #endif
   }
}


inline void modelingFD_op::takeWavCheckpt(int ishot, int it, wavFieldGroup& wfg) { //[1]
   if( it % m_pars->timeParam.jImaging == 0) //Record
   {
      //wfg.pp_cb[0] corresponds to time it
      //Record the snapshot wavefield at pp_cb[1], pp_cb is of ETYPE
      m_pWavfieldCheckptPerShot->seq_update(wfg.pp_cb[0]);
   }
}


//Get the position of a specific shot and time in a wavefiled snapshot data obj
//return false, if the currently it does not correspond with a snap shot time point

bool modelingFD_op::getWindowInSnapshot(windParam& window, int ishot, int it)
{
   bool ret = true;
   int indexOnShotAxis;
   if( it % m_pars->timeParam.jSnap == 0 && shouldRecordSnap(ishot,&indexOnShotAxis))
   {
      int dimModel = m_pars->mShapeWavfield.get_ndim();
      hypercube shapeSnapshot;
      VERIFY( NULL != m_pars->pSrcSnapshot);
      int ndim = m_pars->pSrcSnapshot->domain().get_ndim();
      VERIFY(ndim == (dimModel+2));
      assert(dimModel<=3);
     
	  window.clear();
	  window.resize(ndim,windowAxis());
      
      int tIndex = it/m_pars->timeParam.jSnap;      
      //set the last and second last axis
      window[ndim-1].f=indexOnShotAxis;
      window[ndim-1].n=1;
      window[ndim-2].f=tIndex;
      window[ndim-2].n=1;
	}
	else ret = false;
	return ret;
}

bool modelingFD_op::getWindowInWavfieldCheckpt(windParam& window, int it)
{
   bool ret = true;
   const hypercube& mShapeWavfield = m_pars->mShapeWavfield;
   if( it % m_pars->timeParam.jImaging == 0)
   {
      int dimModel = mShapeWavfield.get_ndim();
	  window.clear();
	  window.resize(dimModel+1,windowAxis());
      
      int tIndex = it/(m_pars->timeParam.jImaging);      
      //set the last axis
      window[dimModel].f=tIndex;
      window[dimModel].n=1;
	}
	else ret = false;
	return ret;
}

//be caveat that now the time decreases [1]
void modelingFD_op::takeRecvWavfieldSnapshot(int ishot,int it, cubeData<ETYPE>& wfd, dmRsf<float>* pRecvSnapshot) 
{  
	windParam window;
    
   if(getWindowInSnapshot(window,ishot,it)) //Record
   {
      int dimModel = m_pars->modelGridShape.get_ndim();
      //Use random access of the RSF data, because we are in reverse order
      //Record the snapshot wavefield at pp_cb[1], pp_cb is of ETYPE
	  //m_pRecvSnapshot->domain().info("Recv Snapshot");
   #ifndef USE_DB_PRECISION
      if(dimModel==2)
         pRecvSnapshot->update2d(&wfd,window);
      if(dimModel==3)
         pRecvSnapshot->update3d(&wfd,window);
   #else //Copy the wavefield in pp_cb[1] to SnapCube(float type)
         seperr("!!!!Save wavefield snapshot of double precison not implemented yet");
      //copyCubeData<double,float>(*(pp_cb[1]),mSnapCube);
      //m_pRsfSnapShot->seq_update(&mSnapCube)
   #endif
   }
}

//Only support 2D case, subsurface offset case
void modelingFD2d_op::corrWithOffset(cubeData<float>& offCubeOut, const cubeData<float>& wf1,
									 const cubeData<float>& wf2)
{
   //!!!! To do
}

//Could be correlate with a offset parameter. e.g. f1(x-h)*f2(x+h), cpu implementation
void modelingFD_op::correlateSrcRecv(int ishot, int it, const cubeData<ETYPE>& recvwfld)
{
   //decide which Snap frame to do cross correlation, subject to snapshot recorded 
   //since now we iterate the src snapshot from reverse time order, use random access for
   //src wavefield snapshot
	windParam window;
   if(getWindowInWavfieldCheckpt(window,it)) //correlate
   {
	  bool bDoOffGather = m_pars->mOffImgParams.bDoOffsetGather;
      int dimModel = m_pars->mShapeWavfield.get_ndim();
      m_pWavfieldCheckptPerShot->grabXd(&mSrcWavfieldCube,window,dimModel);
      
      //do the crosscorrelation
      mSrcWavfieldCube.multiply(recvwfld);
	  if (it%100 == 0) {
		 //wfGp0.pp_cb[1]->dump("dump2");
		 //mSrcWavfieldCube.dump("dump3");
	   }
      //Sum the result to a generated image
      mRTM_imageCubePerShot.add(mSrcWavfieldCube);
	  
	  if (bDoOffGather) { //Do subsurface offset imaging
		 corrWithOffset(mRTM_offsetImgCubePerShot,mSrcWavfieldCube,*(wfGp0.pp_cb[1]));
	  }
   }
   return;
}

void modelingFDgen_op::genPerShotImage(int ishot)
{
  FUNC_TRACE_BEG(modelingFDgen_op::genPerShotImage);
  int dimModel = m_pars->modelGridShape.get_ndim();
  int ndim = dimModel + 1; //the extra dimension given by the shots.
  windParam window(ndim,windowAxis());
  //set the last axis
  bool bRecordIndividualShot = shouldRecordSnap(ishot);
  int irec_shot = (ishot-m_pars->f_shot_snap)/m_pars->j_shot_snap;
  window[ndim-1].f = irec_shot;
  window[ndim-1].n=1;
  
  //Need to change here. Notice that the image obtained from each shot
  //does not necessarily have same sizes;
  cubeData<float> imageFinalSize(m_pars->finalImageShape);
  
  bool bDoOffset = m_pars->mOffImgParams.bDoOffsetGather;
  
  if (!bDoOffset) {
    if(this->mAction == Yang::acMigration) {
      //Add to the final Image
      cropAlign(*m_pars->pFinalImage,mRTM_imageCubePerShot,true);
      if(bRecordIndividualShot) {
        //imageFinalSize.info("imgFinal");
        cropAlign(imageFinalSize,mRTM_imageCubePerShot,false);
        //m_pars->pImages->updateXd(&mRTM_imageCubePerShot,window,dimModel);
        SAFE_PTR_CALL(m_pars->pImages,updateXd(&imageFinalSize,window,dimModel));
      }
    }
    if (this->mAction == Yang::acTomFwd) {
      //Add to the final Image
      cropAlign(*m_pars->pFinalImage,mRTM_imageCubePerShot,true);
      if (bRecordIndividualShot) {
        cropAlign(imageFinalSize,mRTM_imageCubePerShot,false);
        //m_pars->pImages->updateXd(&mRTM_imageCubePerShot,window,dimModel);
        SAFE_PTR_CALL(m_pars->pImages,updateXd(&imageFinalSize,window,dimModel));
      }
    }
  } else {
    int dimOffImagePerShot = mRTM_offsetImgCubePerShot.domain().get_ndim();
    int ndimOff = dimOffImagePerShot+1;
    windParam window2(ndimOff,windowAxis());
    window2[ndimOff-1].f = irec_shot; window2[ndimOff-1].n = 1;
  
    if (mAction == Yang::acMigration || mAction == Yang::acTomFwd) {
      if (bRecordIndividualShot) {
        //Lets just forget about per shot image gathers!!
        //m_pars->pOffImages->updateXd(&mRTM_offsetImgCubePerShot,window2,dimOffImagePerShot);
        //SAFE_PTR_CALL(m_pars->pOffImages,updateXd(&mRTM_offsetImgCubePerShot,window2,dimOffImagePerShot));
      }
      //m_pars->pFinalOffImage->info("pFinalOffImage");
      //mRTM_offsetImgCubePerShot.info("p2Cube");
      cropAlign(*m_pars->pFinalOffImage,mRTM_offsetImgCubePerShot,true);
    }
  }	
  
  if (this->mAction == Yang::acTomAdj) {
    cropAlign(*m_pars->pFinalDeltaS,mRTM_imageCubePerShot,true);
    if (bRecordIndividualShot) {
      cropAlign(imageFinalSize,mRTM_imageCubePerShot,false);
      //m_pars->pDeltaSs->updateXd(&mRTM_imageCubePerShot,window,dimModel);
      SAFE_PTR_CALL(m_pars->pDeltaSs,updateXd(&imageFinalSize,window,dimModel));
    }
  }
  FUNC_TRACE_END();
}

