#include "stdafx.h"

#include "waveOps.h"
#include "velFD2d.h"
#include "SR_manager.h"
#include "boundaryFD.h"
#include "paddingFD.h"

#include "interp.h"

using namespace Yang;

void modelingFD_op::setStencil(int _stenWid)
{
   stencil.clear();
   double sum=0;
   //Different Coefficient based on different stenWid
   switch (_stenWid)
   {
   case 1:
     stencil.push_back(1.000000000000);
     break;
   case 2: //4th order
	 stencil.push_back(1.333333333333333);
	 stencil.push_back(-0.083333333333333);
	 break;
   case 4: //8th order
	  stencil.push_back( 1.600000000000001);
	  stencil.push_back(-0.200000000000000);
	  stencil.push_back( 0.025396825396825);
	  stencil.push_back(-0.001785714285714);
	  break;
   case 5: //2*5=10 th order
	 stencil.push_back( 1.666666666666667);
	 stencil.push_back(-0.238095238095238);
	 stencil.push_back( 0.039682539682540);
	 stencil.push_back(-0.004960317460317);
	 stencil.push_back( 0.000317460317460);
	 break;
   default:
	  seperr("In correct stencil width %d !!\n", _stenWid);
	  break;
   }
	sum = -2.0*std::accumulate(stencil.begin(), stencil.end(),0.0);
	stdVecDb::iterator it = stencil.begin();
	stencil.insert(it,1,sum);
}

void modelingFDgen_op::initOperatorEx(waveOpParams* params, waveOps::action_type ac_type) {
  FUNC_TRACE_BEG(modelingFDgen_op::initOperatorEx);
   this->initOperator(params, ac_type);
   this->m_pSR_manager->init(params);
   this->m_pPaddingFD->initOperatorEx(params);
   this->m_pBoundaryFD->initOperator(params);
  FUNC_TRACE_END();
}

//Init at the beginning of doing action
void modelingFDgen_op::initOperator(waveOpParams* params, action_type ac_type)
{
   FUNC_TRACE_BEG(modelingFDgen_op::initOperator)
   //First call the base class function, base class func is virtual=0
   //waveOps::initOperator(params,ac_type);
   mAction = ac_type;
   m_pars = params;

   VERIFY(m_bAssembled);
   
   
   //Choose 2nd/4th order in time
   bool bUse4thOrder = params->bUse4thOrder;
   if (bUse4thOrder) {
	  m_t_accuracy = order4;
   }
   else {
	  m_t_accuracy = order2;
   }
  
   LOG(mx0);
  //determine the entire image size based on the entire velocity model size.
  m_pars->finalImageShape = m_pars->velModelShape;
  
	//Parameters should be received through function interface not par files
	//This implementation is a big mistake !!!!! should have 2 functions
	//initModeling; initMigration_op(); in which send the paramters through the parameters.
	//when do RTM modeling, sent the input(Refl Model) and the output(seis-record) 
	//as parameters.

   //Remember Here the SnapShot obj is rsf within each shot, but might be different in size between 
   //each shot. should use std::vector<dmRsf*> to manage this, but does not necessarily involve 
   //sep3d here. 
   //Decide incore or outcore for the seis-record data obj
   
   //No matter which case, the Src wavefield SnapShot is generated through calculation
   
   //here we do not use the checkpoint tech for storing the snapshots.
   //append 2 extra axis, one for snapshot time, one for shots
	
   //allocate the wavefield snapshot object
   //To create the data object for seismic record
	
	//First need to determine the wavefield shape for creating the snapshot ptrs. //quan yi zhi ji
	this->setWavfieldShape(0);
  //LOG(mx05);
	//Then Check if the input velocity cube has the correct size, this should not be 
	//a part of operator
	if (!isModeling()) {
	   VERIFY(m_pars->velModelShape.sameSize(m_pars->pVel->domain()));
	}
  
  
  //allocate snapshot records. this is the only allocation happenned inside the operator
  ///Allocate wfld snapshots
	axis ax_shots;
	this->getShotAxisForSnap(ax_shots);
	ax_shots.print_info("ax_shot");
  hypercube shapeSnapshot, shapeSnapshotPershot;
  this->getSnapshotShapePerShot(shapeSnapshotPershot);
  vecAxis aaxx = shapeSnapshotPershot.get_axes();
  aaxx.push_back(ax_shots); shapeSnapshot.set_axes(aaxx);
	//shapeSnapshot.info("shapeSnap");
    
  //LOG(mx1);
  //allocate memory for snapshots.
  //m_pars->pSrcSnapshot->init(&shapeSnapshot);
  //PRV(m_pars->pSrcSnapshot); PRV(m_pars->pSrcSnapshot2); PRV(m_pars->pRecvSnapshot);
  
  SAFE_PTR_CALL(m_pars->pSrcSnapshot,init(&shapeSnapshot));
  if (this->isAdjRTM()) {
	  //m_pars->pSrcSnapshot2->init(&shapeSnapshot);
    SAFE_PTR_CALL(m_pars->pSrcSnapshot2,init(&shapeSnapshot));
  }
   //For RTM debug, Recv snap shot
  if(this->isRTM()) {
	  SAFE_PTR_CALL(m_pars->pRecvSnapshot,init(&shapeSnapshot));
  }
	if (mAction==acTomAdj || mAction==acTomFwd) {
		SAFE_PTR_CALL(m_pars->pRecvSnapshot,init(&shapeSnapshot));
		SAFE_PTR_CALL(m_pars->pSrcSnapshot2,init(&shapeSnapshot));
		SAFE_PTR_CALL(m_pars->pRecvSnapshot2,init(&shapeSnapshot));
	}
  LOG(mx2);
  FUNC_TRACE_END();
}

//Now the operator does not allocate the ptr memory on m_pars, it all rely on the Operator interface I/O
//to provide no-temporary memory it needs.
//Need call this operator before call the operation xxx_op()
void modelingFDgen_op::checkInOutSize() {
  FUNC_TRACE_BEG(modelingFDgen_op::checkInOutSize);
  LOG(xx0);
	//Reset the counter of d_obs for all actions other than modeling and RTM where d_obs
	//is either input or output
	if (!(isModeling()||isAdjRTM()||isRTM())) {
    assert(m_pars->pRsfSeisRec);
		m_pars->pRsfSeisRec->resetCounter();
	}
	
	//MAIN OBJECT: Init output object size and check the input data size
	//LOG(xx1);
	if (this->isAdjRTM() || this->isModeling()) {
		//set the domain for SeisRec and output domain info to file
	  //REMEMBER HERE m_pSR_manager has not INITIALIZED YET!!! THUS we need to supply params explicitly
	  assert(m_pars->pRsfSeisRec);
		hypercube seisRecShape = m_pSR_manager->calcModelingDataSize(m_pars);
    //seisRecShape.info("Ss"); m_pars->pRsfSeisRec->domain().info("dd");
		VERIFY(seisRecShape.sameShape(m_pars->pRsfSeisRec->domain()));
	}
  
	//LOG(xx2);
  
  axis ax_shots; this->getShotAxisForSnap(ax_shots);
  bool bDoImageGather = m_pars->mOffImgParams.bDoOffsetGather;

  const hypercube& imgShape = m_pars->finalImageShape;
  hypercube imgShapeOff;
  const FDModelingTimeParams& timePar = m_pars->timeParam;
	vector<axis> axes2 = imgShape.get_axes();
	axes2.push_back(ax_shots);
	hypercube RTM_imagesShape(axes2), RTM_OffimagesShape;
  
	if (bDoImageGather) {
		get_prestack_img_cube_shape(imgShapeOff,m_pars->mOffImgParams,
                                imgShape, timePar.dtComp);
    axes2 = imgShapeOff.get_axes();
    axes2.push_back(ax_shots);
    RTM_OffimagesShape.set_axes(axes2);
	}
  //LOG(xx3);
  
  //Check input
  if (mAction==acBornModeling && mAction==acTomAdj) {
    if (bDoImageGather) {
      VERIFY(m_pars->pFinalOffImage->domain().sameShape(imgShapeOff));
    } else {
      VERIFY(m_pars->pFinalImage->domain().sameShape(imgShape));
    }
  }
  if (mAction==acTomFwd) {
    VERIFY(m_pars->pFinalDeltaS->domain().sameShape(imgShape));
  }
  
  //Check output
	//LOG(xx6);
	if(this->isRTM() || mAction==acTomFwd) {
    if (!bDoImageGather) {
      //First Init the zero-offset image
      VERIFY(m_pars->pFinalImage->domain().sameShape(imgShape));
      //Also Init the pImages in m_pars, HERE we assume that the aperture is the same for every shot
      //m_pars->pImages->init(&RTM_imagesShape); still alloc memory here?
      SAFE_PTR_CALL(m_pars->pImages,init(&RTM_imagesShape));
    } else {
      VERIFY(m_pars->pFinalOffImage->domain().sameShape(imgShapeOff));
      //m_pars->pOffImages->init(&RTM_OffimagesShape);
      SAFE_PTR_CALL(m_pars->pOffImages,init(&RTM_OffimagesShape));
    }
  }
  
  LOG(xx7);
	///Allocate for Tom Adj, currently the deltaS computed from the engine 
  ///is still of imgShape, not vel shape, but in usual, they should be the same
	if(mAction==acTomAdj) {
	   VERIFY(m_pars->pFinalDeltaS->domain().sameShape(imgShape));
     SAFE_PTR_CALL(m_pars->pDeltaSs,init(&RTM_imagesShape));
	}
	FUNC_TRACE_END();
}

void modelingFD_op::initOperator(waveOpParams* params, action_type ac_type) {
  FUNC_TRACE_BEG(modelingFD_op::initOperator)
	modelingFDgen_op::initOperator(params,ac_type);
	//Get and set stencil Width
  this->setStencil(params->stencilWidth); 
	
	//Actual size is decided at the beginning of each shot
  m_pWavfieldCheckptPerShot = auto_ptr< cubeData<float> >(new cubeData<float>());
	
	FUNC_TRACE_END()
}

void modelingFD_op::unInitOperator()
{
   modelingFDgen_op::unInitOperator();
   delete(m_pWavfieldCheckptPerShot.release());
   //Let the caller to decide when to uninit those cubes in m_pars
   /*
   m_pars->pSrcSnapshot->unInit();
   
   //For RTM debug, Recv snap shot
   if(this->isRTM() || this->isAdjRTM()) {
		 m_pars->pRecvSnapshot->unInit();
   }
   //Allocate the memory saving final image
   if(this->isRTM()) {
      m_pars->pFinalImage.unInit();
	  m_pars->pImages->unInit();
   } 
   */
}

void modelingFDgen_op::setWavfieldShape(int iShot) {
	mOffsetPandModel.clear();
	vector<axis> axes;
	axis ax;
	int ndim = m_pars->modelGridShape.get_ndim();
	int stenWid = m_pars->stencilWidth;
	for (int i=1;i<=ndim;i++) //get_axis is 1 based.
	{
	  mOffsetPandModel.push_back(stenWid);
	  ax = m_pars->modelGridShape.get_axis(i);
	  ax.n = ax.n + 2*stenWid;
	  ax.o = ax.o- stenWid*ax.d;
	  axes.push_back(ax);
	}
	//Current case wavfield is just mStencil layer thicker than velocity
	m_pars->mShapeWavfield.set_axes(axes);
}

//If need to crop the background velocity to fit into the current RTM aperture,
//should do it here.
void modelingFDgen_op::initPerShot(int iShot) {
  FUNC_TRACE_BEG(modelingFDgen_op::initPerShot);

  //calculate the size of the Wavefield, (nz+2*stenWid,nx+2*stenWid,ny+2*stenWid)
	this->setWavfieldShape(iShot);
  //Then we need decide the image shape
  vecAxis axes1 = m_pars->mShapeWavfield.get_axes();
  vecAxis axes2; axis ax;
  int off[3];
  off[0] = m_pars->offz; off[1] = m_pars->offx; off[2] = m_pars->offy;
  for (size_t iax=0; iax<axes1.size(); iax++) {
    ax = axes1[iax];
    //tick out the boundary and the padding layer
    ax.o = ax.get_coord(off[iax]+m_pars->m_nPad);
    ax.n = ax.n - 2*(off[iax]+m_pars->m_nPad);
    axes2.push_back(ax);
  }
  m_pars->mImageShape.set_axes(axes2);
  m_pars->modelGridShape = m_pars->pVel->domain();
  FUNC_TRACE_END();
}

//Init at the beginging of every shot
void modelingFD_op::initPerShot(int iShot)
{
  FUNC_TRACE_BEG(modelingFD_op::initPerShot);

  modelingFDgen_op::initPerShot(iShot);
   
   const hypercube& mShapeWavfield = m_pars->mShapeWavfield;
   //Init the memory object for modeling and src/recv propagation
   //The function uses only mShapeWavfield param, assuing shapeWavfield ==  shapeSnapShot

   //Allocate memory for //HISTORY
#ifdef USE_DB_PRECISION
   mSnapCube.alloc(mShapeWavfield);
#else
  //don't do anything, since we have pp_cb[] to save wavefield; 
#endif
   
   dpDdx2_cb.alloc (mShapeWavfield); //cubeData obj will be used in propagate_4t and adjModeling function   
   
   //Allocate wavefield slices (t-1,t,t+1), e.t.c
   //Create wavefield array, currently limited to 3 time slices.
   wfGp0.alloc(mShapeWavfield);
   wfGp0.reset();
   
   bool bDoImageGather = m_pars->mOffImgParams.bDoOffsetGather;
   
   //Create the WavCheckPt object
   if (isRTM()) {
   
      hypercube shapeWavfieldCheckPt;
      //here we do not use the checkpoint tech, just store the snapshots.
      //append 2 extra axis, one for snapshot time, one for shots
      float o,d;
      int n;
      
      const FDModelingTimeParams& tpar = m_pars->timeParam;
      
      o = 0.0; d = tpar.dtImaging; n = tpar.nImaging;
      axis ax_t(n,o,d,"time","sec");  
      vector<axis> axes = mShapeWavfield.get_axes();
      axes.push_back(ax_t);
      shapeWavfieldCheckPt.set_axes(axes);
      m_pWavfieldCheckptPerShot->alloc(shapeWavfieldCheckPt);
      
      //Alloc other memory objects, for zero offset only
      mSrcWavfieldCube.alloc(m_pars->mShapeWavfield);
      mRTM_imageCubePerShot.alloc(m_pars->mShapeWavfield);
	  
	  //For prestack image
	  if (bDoImageGather) {
		 hypercube wfldShapeOff;
		 get_prestack_img_cube_shape(wfldShapeOff,m_pars->mOffImgParams,
										  m_pars->mShapeWavfield, tpar.dtComp);
		 mRTM_offsetImgCubePerShot.alloc(wfldShapeOff);
	  }
   }
   
   if(isAdjRTM()) {
	  wfGp1.alloc(m_pars->mShapeWavfield);
	  wfGp1.reset();
	  mCorrCube.alloc(m_pars->mShapeWavfield);
   }
  FUNC_TRACE_END();
}


void modelingFD_op::unInit(){
  FUNC_TRACE_BEG(modelingFD_op::unInit);
   wfGp0.dealloc();
   mSnapCube.dealloc();
   dpDdx2_cb.dealloc();
  FUNC_TRACE_END();
}

//assume inReflModel has the size of velocity, and its size is smaller than the wavefield
//also assume they have same grid spacing. Then we can align by simply creating window.
//This works best with inRefl bigger than outcube, cause you cannot control extend=0/1 in interp function
void modelingFDgen_op::alignCropToSliceShapeGrid(cubeData<float>& outReflCube, cubeData<float>& inReflModel, hypercube& sliceShape)
{  
  FUNC_TRACE_BEG(modelingFDgen_op::alignCropToSliceShapeGrid)
   VERIFY(outReflCube.domain().compatible(sliceShape));
   int ndimModel = sliceShape.get_ndim();
   hypercube outShape = outReflCube.domain();
   int ndim = outShape.get_ndim();
   assert(ndim == inReflModel.domain().get_ndim());
   vector<axis> axes = sliceShape.get_axes();
   VERIFY(ndimModel  <=3 );
   // Use interpolate to align grid.
   float o1,d1,o2,d2,o3,d3;
   int n1,n2,n3;
   axes[0].get_nod(n1,o1,d1);
   axes[1].get_nod(n2,o2,d2);
   
   //cerr<<ndim<<sep<<ndimModel<<n1<<sep<<n2<<sep<<n3<<sep<<endl;

   seplib::interpType interpMethod = seplib::nearest;
   cubeData<float> sliceIn, sliceOut;
   VERIFY(ndimModel==2);
   if (ndimModel == 3)
   {  axes[2].get_nod(n3,o3,d3);
      seplib::interp::getInstance()->interpolate(outReflCube,inReflModel,&n1,&o1,&d1,
                                     &n2,&o2,&d2,
                                     &n3,&o3,&d3, interpMethod);
   }
   if (ndimModel == 2) {
      //cerr<<endl<<"In Interp"<<n1<<','<<o1<<','<<d1<<','<<n2<<','<<o2<<','<<d2<<','<<endl;
      seplib::interp::getInstance()->interpolate(outReflCube,inReflModel,&n1,&o1,&d1,
                                   &n2,&o2,&d2,
                                   NULL,NULL,NULL,interpMethod);
   }
   FUNC_TRACE_END();
}

