#include "stdafx.h"
#include "myFFT.h"

using namespace Yang;

FFT1d::~FFT1d() {
  //free the memory
  int n_threads = nth_;
  for (int i_thread=0; i_thread<n_threads; i_thread++) {
    fftwf_free(mRealPtArray[i_thread]);
    fftwf_free(mCmplxPtArray[i_thread]);
    //Destroy the FFT plans
    fftwf_destroy_plan(mFFTplanArray[i_thread]);
  }
}

FFT1d::FFT1d(int N_in, bool bForward, FFT_IOtype iotype): N_(N_in), bForward_(bForward), ioType_(iotype){
  nth_ = 1; //g_NUM_THREADS
  int n_threads = nth_;
  int i_thread=0;
  
  mFFTplanArray.resize(n_threads);
  mCmplxPtArray.resize(n_threads); mRealPtArray.resize(n_threads);
  int N = N_in;
  int FFT_sign = bForward? FFTW_FORWARD:FFTW_BACKWARD;
  n_cmplx_ = (N%2==1 ? (N-1)/2+1:N/2+1); //rounded down
  if (iotype == fft_r2c) {
    //prepare for doing fft 1d
    for (i_thread=0; i_thread<n_threads; i_thread++) {
      mCmplxPtArray[i_thread]  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n_cmplx_); //these dimensions are after padding
      mRealPtArray[i_thread] = (float*)fftwf_malloc(sizeof(float)*N);
      memset(mCmplxPtArray[i_thread],0,n_cmplx_*sizeof(fftwf_complex)); memset(mRealPtArray[i_thread],0,N*sizeof(float));
      mFFTplanArray[i_thread]   = fftwf_plan_dft_r2c_1d(N,mRealPtArray[i_thread],mCmplxPtArray[i_thread],FFTW_MEASURE);
    }
  } else if (iotype == fft_c2r) {
    //prepare for doing fft 1d
    for (i_thread=0; i_thread<n_threads; i_thread++) {
      mCmplxPtArray[i_thread]  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n_cmplx_); //these dimensions are after padding
      mRealPtArray[i_thread] = (float*)fftwf_malloc(sizeof(float)*N);
      memset(mCmplxPtArray[i_thread],0,n_cmplx_*sizeof(fftwf_complex)); memset(mRealPtArray[i_thread],0,N*sizeof(float));
      mFFTplanArray[i_thread]   = fftwf_plan_dft_c2r_1d(N,mCmplxPtArray[i_thread],mRealPtArray[i_thread],FFTW_MEASURE);
    }
  } else {
    VERIFY(0);
  }
}

void FFT1d::tf(const float* pr, fftwf_complex* pc, long long nTraces) {
  VERIFY(ioType_ == fft_r2c);
  //Do FFT for pobs;
  { //omp section
  int i_thread = 0; //Get i_thread using omp function.
  long long i;
  for (i=i_thread; i<nTraces; i+=nth_) {
    memset(mRealPtArray[i_thread],0,N_*sizeof(float));
    myMemcpy(mRealPtArray[i_thread],pr+N_*i,N_);
    fftwf_execute(mFFTplanArray[i_thread]);
    myMemcpy(pc+n_cmplx_*i,mCmplxPtArray[i_thread],n_cmplx_);
  }
  } //end omp parallel
}
 
void FFT1d::tf(const fftwf_complex* pc, float* pr, long long nTraces) {
  VERIFY(ioType_ == fft_c2r);
  double f1N = 1.0/double(N_);
  
  //Do FFT for pobs;
  { //omp section
  int i_thread = 0; //Get i_thread using omp function.
  long long i; int j;
  for (i=i_thread; i<nTraces; i+=nth_) {
    memset(mCmplxPtArray[i_thread],0,n_cmplx_*sizeof(fftwf_complex));
    myMemcpy(mCmplxPtArray[i_thread],pc+n_cmplx_*i,n_cmplx_);
    fftwf_execute(mFFTplanArray[i_thread]);
    for (j=0; j<N_; j++) {mRealPtArray[i_thread][j] *= f1N;} //because it is backward FFT, multiply 1/N to normalize
    myMemcpy(pr+N_*i,mRealPtArray[i_thread],N_);
  }
  } //end omp parallel
}

  
void FFT1d::tf(const fftwf_complex* pc1, const fftwf_complex* pc2, long long nTraces) {
  VERIFY(0);
}
