#include "stdafx.h"
#include "decl_fwd.h"
#include "waveOpMPI.h"
#include "my_mpi.h"
#include "cubeData.hpp"
#include "my_timer.h"

extern "C" {
  #include "mpi_sep.h"
}

#undef DEBUG
#include "myDebug.h"


using std::string;
using namespace mpi;
using namespace Yang;

mpi::environment::environment(int* p_argc, char*** p_argv) {
  MPI_Init(p_argc,p_argv);
}

int mpi::slaveModeMPI::on() {
  string strCmd;
  mpi::communicator world;

  Yang::cubeDataF data;

  while(1) {
    const MPI_Status& stat = world.recv(0,MPI_ANY_TAG,strCmd);
    if (stat.MPI_TAG == EXITAG) {
      break; //exit the loop
    } else if (stat.MPI_TAG == WRKTAG) {
      //receive a cmd line and do the wrk
      if (world.rank()==2) {
        //LOG("From slaveModeMPI::on");PRV(strCmd);
      }
      VERIFY(system(strCmd.c_str())==0); usleep(int(1e5)); //puase for a while.
      //if (world.rank()==2) LOG("after executing cmd in slaveMode::on");
      world.barrier();
    } else if (stat.MPI_TAG == IDLETAG) { //for the master to test that if slaves are Idle.
      if (world.rank()==2) {
        LOG("Recv IDLETAG");
      } //do nothing.
    } else if (stat.MPI_TAG == REDUCETAG) {
      //Load the file name, assume its size is not too big, and do an MPI_Reduce
      string fn = strCmd;
      //BroadCast the size of the file, number of elements per blk and nblk, MPI cannot handle very large msg size.
      // N = per_blk*nblk+nleft;
      unsigned per_blk, nblk, nleft; int ne[3];
      exeMPI(MPI_Bcast(ne,3,MPI_UNSIGNED,0,MPI_COMM_WORLD));
      //Load the file
      data.load(fn);
      per_blk = ne[0]; nblk = ne[1]; nleft = ne[2];
      long long count = per_blk*nblk+nleft;
      if (count!=data.domain().get_n123()) {
        SEEV2(count,data.domain().get_n123());
        VERIFY(0);
      }
      float* pSend = const_cast<float*>(data.dataPtr());
      for (unsigned int iblk=0; iblk<nblk; iblk++) {
        exeMPI(MPI_Reduce(pSend,NULL,per_blk,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
        pSend += per_blk;
      }
      //the last remaining block
      exeMPI(MPI_Reduce(pSend,NULL,nleft,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
    } else {
      VERIFY(0);
    }
  }
  return 0;
}

//for root process, due to the constrain of MPI_Reduce, din and dout cannot be the same.
int mpi::ReduceCube_MPI(cubeDataF& dout, PosInt iAxis_per_blk, const branchFileName& bf_in) {
  mpi::communicator world;
  int nrank = world.size();
  string strCmd;
  //BroadCast the size of the file, number of elements
  long long count;
  const hypercube& shp = dout.domain();
  
  count = dout.domain().get_n123();
  for (int ir=1;ir<nrank;ir++) {
    strCmd = bf_in.fn_at(ir);
    world.send(ir,REDUCETAG,strCmd);
  }
  LOG("Begin MPI_BCast in Reduce");
  unsigned int ne[3];
  int ndim = shp.get_ndim(); VERIFY(iAxis_per_blk<=ndim);

  if (0) { //we can choose the blk based on the model shape
    ne[0] = shp.get_n23(1,iAxis_per_blk);
    ne[1] = 1; if (iAxis_per_blk<ndim) {ne[1] = shp.get_n23(iAxis_per_blk+1,ndim);}
    ne[2] = 0;
    assert(count == ne[0]*ne[1]);
  } else { //or just use 1M blocks as kind of optimum
    ne[0] = 8*1024; //8K is the magic number for cees-cluster
    ne[1] = count/ne[0];
    ne[2] = count % ne[0];
  }
  VERIFY( (ne[0]*ne[1]+ne[2]) == count);

  if (ne[0]<4e3 || ne[0]>65e3) { //if the msg size is not between 4k~64k, then it might hurt the performance 
    prt("blk size in ReduceCube_MPI might not be optimal, size is %d Bytes\n", ne[0]*4);
  }

  exeMPI(MPI_Bcast(ne,3,MPI_UNSIGNED,0,MPI_COMM_WORLD));
  LOG("Finish MPI_BCast in Reduce");
  Yang::timer clk;
  
  float* pRecv = dout.dataPtr(); dout.zero();
  std::valarray<float> arrSend(ne[0]); arrSend = 0.f; //dummy buffer

  for (unsigned int iblk=0; iblk<ne[1]; iblk++) {
    exeMPI(MPI_Reduce(va_ptr(arrSend),pRecv,ne[0],MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));
    pRecv += ne[0];
  }
  exeMPI(MPI_Reduce(va_ptr(arrSend),pRecv,ne[2],MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD));

  clk.check("reduce finish");
  return 0;
}

communicator::communicator() {
  exeMPI(MPI_Comm_rank(MPI_COMM_WORLD,&irank_));
  exeMPI(MPI_Comm_size(MPI_COMM_WORLD,&nrank_));
  
  int membersize = 0;
  //pack the string size
  exeMPI(MPI_Pack_size(1,MPI_INT,MPI_COMM_WORLD,&membersize));
  bufSize_ = membersize;
  exeMPI(MPI_Pack_size(sz_size,MPI_CHAR,MPI_COMM_WORLD,&membersize));
  bufSize_ += membersize;
  buffer_ = new char[bufSize_];
}

int communicator::send(int dst,int tag,const string& val) {
  //send using string pack
  szPack_.length_ = val.length()+1;
  VERIFY(szPack_.length_<=sz_size);
  strcpy(szPack_.sz_,val.c_str());
  
  int pos = 0;
  //exeMPI(MPI_Pack(&szPack_.length_,1,MPI_INT,buffer_,bufSize_,&pos,MPI_COMM_WORLD));
  //exeMPI(MPI_Pack(&szPack_.sz_,1024,MPI_CHAR,buffer_,bufSize_,&pos,MPI_COMM_WORLD));
  //exeMPI(MPI_Send(buffer_,pos,MPI_PACKED,dst,tag, MPI_COMM_WORLD));
  exeMPI(MPI_Send(szPack_.sz_,szPack_.length_,MPI_CHAR,dst,tag, MPI_COMM_WORLD));
  return 0;
}

const MPI_Status& communicator::recv(int src,int tag, string& val) {
  int pos, msgsize;
  //exeMPI(MPI_Recv(buffer_,bufSize_,MPI_PACKED,src,tag,MPI_COMM_WORLD,&status));
  //exeMPI(MPI_Get_count(&status,MPI_PACKED,&msgsize));
  //exeMPI(MPI_Unpack(buffer_,msgsize,&pos,&szPack_.length_,1,MPI_INT,MPI_COMM_WORLD));
  //exeMPI(MPI_Unpack(buffer_,msgsize,&pos,szPack_.sz_,1024,MPI_CHAR,MPI_COMM_WORLD));
  exeMPI(MPI_Recv(buffer_,bufSize_,MPI_CHAR,src,tag,MPI_COMM_WORLD,&MPI_stat_));
  //Now put it back to string.
  strcpy(szPack_.sz_,buffer_);

  val = szPack_.sz_;
  return MPI_stat_;
}

