#include "stdafx.h"
#include "param_func.h"
#include "param_clib.h"
#include "solverNonlinearSearch.h"
#include "nonLinSolver.param.h"

using namespace Yang;
using namespace std;

//Read nonLinSolver Parameters
void Yang::read_nonLinSolver_par(solverParam* pSolvPar, param_func* pars) {
  solverParam& solvPar = *pSolvPar;
  bool bDoSmooth = pars->get_bool("bDoSmooth",false);
  
  solvPar.maxIter      = pars->get_int("maxNonLinIter",5);
  const int* nspl = NULL;
  if (bDoSmooth) {
    int smTypeIndex = pars->get_int("smType",0);
    switch (smTypeIndex) {
    case 0:
    case 1:
      if (smTypeIndex==0) {solvPar.eSmooth = solverParam::sm_spline;}
      else {
        solvPar.eSmooth = solverParam::sm_rect; 
        prt("!Use seplib Rect Smooth!\n");
      }
      //0 means no smoothing on the model
      solvPar.nSpl[0]  = pars->get_int("n1spl",0);
      solvPar.nSpl[1]  = pars->get_int("n2spl",0);
      solvPar.nSpl[2]  = pars->get_int("n3spl",0);
      nspl = solvPar.nSpl;
      solvPar.nSpl_end[0]  = pars->get_int("n1sple",  nspl[0]);
      solvPar.nSpl_end[1]  = pars->get_int("n2sple",  nspl[1]);
      solvPar.nSpl_end[2]  = pars->get_int("n3sple",  nspl[2]);
      
      VERIFYC(nspl[0]>0);  VERIFYC(nspl[1]>0);
      solvPar.fsSpl[0] = pars->get_float("n1splScale",powf(float(solvPar.nSpl_end[0])/float(nspl[0]),1.0f/float(solvPar.maxIter)));
      solvPar.fsSpl[1] = pars->get_float("n2splScale",powf(float(solvPar.nSpl_end[1])/float(nspl[1]),1.0f/float(solvPar.maxIter)));
      if (nspl[2]!=0) {
        solvPar.fsSpl[2] = pars->get_float("n3splScale",powf(float(solvPar.nSpl_end[2])/float(nspl[2]),1.0f/float(solvPar.maxIter)));
      } else {solvPar.fsSpl[2]=1.0f;}
      PRVEC(solvPar.fsSpl,3);
      
      break;
    case 2:
      solvPar.eSmooth = solverParam::sm_alongDip; 
      VERIFY(0);
      break;
    default:
      VERIFY(0);
    }
    solvPar.repeat  = pars->get_int("repeat",1);
  }
  
  solvPar.bSmoothSearchDir = bDoSmooth;
  solvPar.bChangeSmoothParamWithIteration = pars->get_bool("bVarySmoothing",false);
  float fmaxVal        = pars->get_float("fmaxVel");//,3200.0f); //max allowed vel
  float fminVal        = pars->get_float("fminVel");//,1500.0f); //min allowed vel
  //PRV(fmaxVel); PRV(fminVel);
  solvPar.fmaxValAllow = 1.0f/fminVal; //slowness
  solvPar.fminValAllow = 1.0f/fmaxVal; //1.0f/fmaxVel;
  solvPar.errTol       = pars->get_float("errorAllowed",1e-4);
  solvPar.fMaxPerturbScale = pars->get_float("maxPerturbScale",0.05); //5 percent of the slowness RMS value
  solvPar.bDecreaseMaxStepLengthWithIteration = pars->get_bool("bVaryStepLength",false);
  solvPar.fbase1       = pars->get_float("maxPerturbBase",0.2f);
  solvPar.fdecay       = pars->get_float("maxPerturbDecay",0.8f);
  
  solvPar.bRecalcObjFuncEachIter = pars->get_bool("bRecalcObjFuncEachIter",false);
  
  //if resume the inversion from some intermediate chkpt
  static pair<int,solverParam::checkptResumeStat> rsm_match_table[] = {make_pair(0,solverParam::rsm_none),make_pair(1,solverParam::rsm_grad)};
  solvPar.rsmStat_ = findEnumItemFromPar(pars,rsm_match_table,NELEM(rsm_match_table),"resumeChkptStatus",0);

  //Gradient Mask
  bool bMaskGrad = pars->get_bool("bMaskGradient",false);
  solvPar.bMaskGradient = bMaskGrad;
  bool b3d = pars->get_bool(csz_b3D,false);
  if (bMaskGrad) { //load gradmask
    if (!b3d) {
      solvPar.gradMask.load("gradMask",2);
    } else {
      solvPar.gradMask.load("gradMask",3); //as 3D cube.
    }
    //solvPar.gradMask.info("gradMask");
  }
}

