#include "stdafx.h"

#include <fftw3.h>
#include <iostream>
#include <complex>

#include <boost/shared_array.hpp>
#include <boost/shared_ptr.hpp>

#include "commonOps.h"
#include "sep_params.h"
#include "cubeData.hpp"
#include "stackTraceC.h"
#include "tagHack.h"
#include "off2ang3d.h"
#include "cubeAlgo.h"
#include "cubeAlgo2.h"

#include "omp.h"

#include "dmRsfDump.h"
using namespace Yang;
using namespace std;


//In all cases, N intput, N output. There will be redundancy in r2c and c2r cases.
enum FFTtype {
   r2c, //forward
   c2r, //inverse, is Adjoint as well.
   c2c
};


static char szCmd[1024];
static char szToolFileName1[512],szToolFileName2[512];
static char szToolFileTag1[512],szToolFileTag2[512];
static string szFT3dTool("/net/server/homes/sep/yang/Research/codelib/bin/YFt3d.x");

void outputShape_od_AsString(const hypercube& shape,string& szAxis) {
  char sz[512];
  axis ax;
  szAxis.clear();
  for (int i=1;i<1+shape.get_ndim();i++) {
    ax = shape.get_axis(i);
    sprintf(sz," o%1d=%f d%1d=%f label%1d=\"%s\" unit%1d=%s \\\n", i,ax.o,i,ax.d,i,ax.label.c_str(),i,ax.unit.c_str());
    szAxis = szAxis + sz;
  }
  return;
}
#include "point2d.hpp"
/// find baycentric weights of a point in the triangles (the coordinates is in (x,y))
void bayctricCoord(float coord[3], const point2dF& xy, float x1, float x2, float x3, float y1, float y2, float y3){
  float detTr = 1.0f/((y2-y3)*(x1-x3) + (x3-x2)*(y1-y3));
  coord[0] = ((y2-y3)*(xy.x-x3)+(x3-x2)*(xy.y-y3)) * detTr;
  coord[1] = ((y3-y1)*(xy.x-x3)+(x1-x3)*(xy.y-y3)) * detTr;
  coord[2] = 1-coord[0] - coord[1];
  assert(max(max(fabs(coord[0]),fabs(coord[1])),fabs(coord[2])) < 1.0f);
}

///Here both inputs are filenames, files are (z,x,y,hx,hy) axis, and would be transformed to (FT) (kz,kmx,kmy,hx,hy) then transp (hx,hy,kz,kmx,kmy)
///then (MAP) -> (gamma,azim,kz,kmx,kmy) -> (kz,kmx,kmy,gamma,azim) ->IFT (z,mx,my,gamma,azim) -> (z,gamma,azim,mx,my)
void Yang::off2ang3d_file(param_off2ang3d& par, string& sz_inFn, string& sz_outFn) {
  //Do FFT on z,x,y axis
  stdVecInt ns = par.shapeOff.get_ns(); //Should be 5 Dimension (hx,hy,kz,kmx,kmy)
  //Then we know what "n1/n2/n3" to put in the cmd line.
  char szPar[256]; sprintf(szPar," n1=%d n2=%d n3=%d sign1=1 sign2=1 sign3=1 ", ns[2],ns[3],ns[4]);
  sprintf(szToolFileName1,"sepTemp%d.H", rand()%65536);
  sprintf(szCmd,"Rtoc <%s | %s %s >%s \n", sz_inFn.c_str(), szFT3dTool.c_str() , szPar, szToolFileName1); //need to convert to complx number before FFT.
  prt("Executing cmd: %s\n", szCmd);
  int sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  
  //transpose it
  sprintf(szToolFileName2,"septemp%d.H", rand()%65536);
  sprintf(szCmd,"Transp <%s plane=12 reshape=3,5 >%s ", szToolFileName1, szToolFileName2);
  prt("Executing cmd: %s\n", szCmd);
  sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  sprintf(szCmd,"Rm %s\n", szToolFileName1); VERIFY(system(szCmd)==0); //clean up temp file1

  //Now the output is in szToolFileName2, associate it with tag szToolFileTag2, read it back
  sprintf(szToolFileTag2,"sepTag%d.H", rand()%65536);
  dmRsf_oc<cmplxf> dm_hxy_kzxy(szToolFileName2,dm::IN,"hxy_kzxy cube",true);
  dm_hxy_kzxy.init(NULL);
  VERIFY(dm_hxy_kzxy.domain().sameShape(par.shapeOff));
  
  //Init the output 
  sprintf(szToolFileName1,"sep_angles_kzxy_temp%d.H", rand()%65536);
  sprintf(szToolFileTag1,"sepTag%d.H", rand()%65536);
  dmRsf_oc<cmplxf> dm_angles_kzxy(szToolFileName1,dm::OUT,"gamma_azim_kzxy cube",true);
  dm_angles_kzxy.init(&par.shapeAng);

  //Do the hx,hy->gamma,azim mapping
  Yang::off2ang3d(par, dm_hxy_kzxy, dm_angles_kzxy);
  
  //close the file handle
  dm_hxy_kzxy.unInit(); dm_angles_kzxy.unInit();
  
  //Now we need to do FFT again, the src file is the szToolFileName1
  //(gamma,azim,kz,kmx,kmy) -> (z,mx,my,gamma,azim) -> (z,gamma,azim,mx,my)
  stdVecInt nzxy = par.shape_zxy.get_ns(); 
  int randcode = rand()%65536;
  char szfilet[512],szfilet2[512]; sprintf(szfilet,"sepTemp%d.H", randcode); sprintf(szfilet2,"sepTemp%d_2.H", randcode);
  
  sprintf(szCmd,"Transp <%s plane=12 reshape=2,5 >%s ",
          szToolFileName1, szfilet);
  prt("Executing cmd: %s\n", szCmd); sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000);
  sprintf(szCmd,"%s <%s  sign1=-1 sign2=-1 sign3=-1  >%s ",
          szFT3dTool.c_str(), szfilet, szfilet2);
  prt("Executing cmd: %s\n", szCmd); sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000);
  
  //the (z,mx,my,gamma,azim) cube might have wrong axis info for the first three axes, therefore re-enforce that
  string szAxis;
  outputShape_od_AsString(par.shape_zxy,szAxis);
  sprintf(szCmd, "echo %s >> %s", szAxis.c_str(), szfilet2); 
  PRV(szCmd);
  VERIFY(system(szCmd)==0); 

  //The pipe gets unreliable once the file size is bigger than 2GB.
  sprintf(szCmd,"<%s Window3d squeeze=0 n1=%d n2=%d n3=%d  >%s", 
          szfilet2,nzxy[0],nzxy[1],nzxy[2],szfilet);
  prt("Executing cmd: %s\n", szCmd); sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  sprintf(szCmd,"<%s Transp plane=23 reshape=1,3,5 >%s", 
          szfilet,szfilet2);
  prt("Executing cmd: %s\n", szCmd); sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  sprintf(szCmd,"<%s Real >%s", 
          szfilet2,szfilet);
  prt("Executing cmd: %s\n", szCmd); sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  
  //Balance the image amplitude by multiplying 1/cos(gamma), for Oneway propagator.
  cubeData<float> imgin,acos_cube; imgin.load(string(szfilet),5);
  axis ax_angle = imgin.domain().get_axis(2);
  deg2radianAxis(ax_angle,ax_angle,true);  VERIFY(fabs(ax_angle.get_coord(ax_angle.n-1)) > PI/18.0f); //the max angle should be bigger than 10deg.
  acos_cube.alloc(hypercube(ax_angle.n));
  for (int i=0; i<ax_angle.n; i++) {
    float ag = ax_angle.get_coord(i);
    acos_cube(i) = 1.0f/(SQ(cosf(ag)));
  }
  //multiply the second axis with acos_cube
  multiplySubCube(imgin,imgin,acos_cube,2,2);
  imgin.dump(sz_outFn,true);
  verbPrt(1,"gamma_balancing file %s to %s \n", szfilet,sz_outFn.c_str());
  
  sprintf(szCmd,"Rm %s; Rm %s; Rm %s; Rm %s", szToolFileName1, szToolFileName2, szfilet, szfilet2);  //clean up temp file1
  //system(szCmd);
  return ;
}

///Here both inputs are filenames, (z,gamma,azim,mx,my) -> (z,mx,my,gamma,azim)-> then (FT) (kz,kmx,kmy,gamma,azim) --> (gamma,azim,kz,kmx,kmy) (MAP)-> (khx,khy,kz,kmx,kmy) 
/// -> (hx_pad,hy_pad,kz,kmx,ky) -> (kz,kmx,kmy,hx_pad,hy_pad) -> (z,x,y,hx_pad,hy_pad) -> (z,x,y,hx,hy) axis
void Yang::ang2off3d_file(param_off2ang3d& par, string& sz_inFn, string& sz_outFn) {
  //Then we know what "n1/n2/n3" to put in the cmd line.
  stdVecInt ns = par.shapeOff.get_ns(); //Should be 5 Dimension (hx,hy,kz,kmx,kmy)

  sprintf(szToolFileName1,"sepTemp%d.H", rand()%65536);
  //Transpose -> FFT -> Transp --> (gamma,azim,kz,kmx,kmy)
  char szfilet[512]; sprintf(szfilet,"sepTemp%d.H", rand()%65536);
  sprintf(szCmd,"Transp <%s plane=23 reshape=1,3,5 | Rtoc | %s sign1=1 sign2=1 sign3=1 n1=%d n2=%d n3=%d >%s",
          sz_inFn.c_str(), szFT3dTool.c_str(), ns[2],ns[3],ns[4], szfilet);
  prt("Executing cmd: %s", szCmd);
  int sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.

  sprintf(szCmd," Transp <%s plane=12 reshape=3,5 >%s \n", 
          szfilet, szToolFileName1);
  prt("Executing cmd: %s", szCmd);
  sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  
  //Do the mapping
  sprintf(szToolFileName2,"sep_hxy%d.H", rand()%65536);
  dmRsf_oc<cmplxf> dm_angles_kzxy(szToolFileName1,dm::IN,"gamma_azim_kzxy cube",true);
  dmRsf_oc<cmplxf> dm_hxy_kzxy(szToolFileName2,dm::OUT,"hxypad_kzxy cube",true);

  hypercube shapeOffpad = par.shapeOff;
  axis ax_hxpad = shapeOffpad.get_axis(1); ax_hxpad.n = par.n_hx_pad;
  axis ax_hypad = shapeOffpad.get_axis(2); ax_hypad.n = par.n_hy_pad;
  shapeOffpad.set_axis(1,ax_hxpad);  shapeOffpad.set_axis(2,ax_hypad);

  dm_angles_kzxy.init(NULL); dm_hxy_kzxy.init(&shapeOffpad);
  //Do the hx,hy->gamma,azim mapping
  bool bDoFtSmoothing = par.bDoFTsmoothing;
  Yang::ang2off3d(par, dm_angles_kzxy,dm_hxy_kzxy,bDoFtSmoothing);
  
  dm_hxy_kzxy.unInit(); dm_angles_kzxy.unInit(); usleep(100000);
  
  //transpose it to (kz,kmx,kmy,hx,hy) ->(FT) (z,mx,my,hx,hy), window it at last
  stdVecInt nzxy = par.shape_zxy.get_ns();
  char szfilet2[512]; sprintf(szfilet2,"sepTemp%d.H", rand()%65536);
  sprintf(szCmd,"Transp <%s plane=12 reshape=2,5 | %s sign1=-1 sign2=-1 sign3=-1 >%s ", szToolFileName2, szFT3dTool.c_str(), szfilet2);
  prt("Executing cmd: %s\n", szCmd);
  sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  
  string szAxis;
  outputShape_od_AsString(par.shape_zxy,szAxis);
  sprintf(szCmd, "echo %s >> %s", szAxis.c_str(), szfilet2); 
  PRV(szCmd);
  VERIFY(system(szCmd)==0); 

  sprintf(szCmd,"<%s Window3d n1=%d n2=%d n3=%d n4=%d n5=%d squeeze=0 | Real >%s ", szfilet2, nzxy[0], nzxy[1], nzxy[2],ns[0],ns[1], sz_outFn.c_str());
  prt("Executing cmd: %s", szCmd);
  sysret = system(szCmd); VERIFY(sysret == 0); usleep(100000); //wait to make sure the cmd is finished.
  
  sprintf(szCmd,"Rm %s; Rm %s; Rm %s; Rm %s ", szToolFileName1, szToolFileName2, szfilet, szfilet2); //clean up temp file1
  //sysret = system(szCmd);
  return ;
}

template <class T> //T could be float and double
inline int floor0(T value)  {
  return int(value<0.0? std::ceil(value):std::floor(value));
}

template <class T> //T could be float and double
inline int ceil0(T value)  {
  return int(value<0.0? std::floor(value):std::ceil(value));
}

//  public-domain code by Darel Rex Finley, 2007, http://alienryderflex.com/polygon_fill/
//  here the IMAGE_TOP<IMAGE_BOT, graphics convention
void fill_polygon(float* polyX, float* polyY, int IMAGE_TOP, int IMAGE_BOT, int IMAGE_LEFT, int IMAGE_RIGHT) {
  #define MAX_POLY_CORNERS 4  //we are only dealing with quadra-laterals.
  int  nodes, nodeX[MAX_POLY_CORNERS], pixelY, i, j, swap ;
  const int polyCorners = 4;
  //  Loop through the rows of the image.
  for (pixelY=IMAGE_TOP; pixelY<IMAGE_BOT; pixelY++) {
  
    //  Build a list of nodes, since all input polygons are complex, at most you will have 2 nodes.
    nodes=0; j=polyCorners-1;
    for (i=0; i<polyCorners; i++) { //if the horizontal line y=pixelY cross the edge[i].
      if ((polyY[i]<(double) pixelY && polyY[j]>=(double) pixelY)
      ||  (polyY[j]<(double) pixelY && polyY[i]>=(double) pixelY) ) {
        nodeX[nodes++]=(int) (polyX[i]+(pixelY-polyY[i])/(polyY[j]-polyY[i])
        *(polyX[j]-polyX[i]));  //find the crossing points
      }
      j=i; //step to next edge.
    }
    
    VERIFY(nodes==2 || nodes==0); //Find one segment or not at all.
    //  Sort the nodes, via a simple Bubble sort.
    i=0;
    while (i<nodes-1) {
      if (nodeX[i]>nodeX[i+1]) {
        swap=nodeX[i]; nodeX[i]=nodeX[i+1]; nodeX[i+1]=swap; if (i) i--; 
      }
      else {
        i++; 
      }
    }
  
    //  Fill the pixels between node pairs.
    for (i=0; i<nodes; i+=2) { //for each segments
      if   (nodeX[i  ]>=IMAGE_RIGHT) break;
      if   (nodeX[i+1]> IMAGE_LEFT ) {
        if (nodeX[i  ]< IMAGE_LEFT ) nodeX[i  ]=IMAGE_LEFT ;
        if (nodeX[i+1]> IMAGE_RIGHT) nodeX[i+1]=IMAGE_RIGHT;
        for (j=nodeX[i]; j<nodeX[i+1]; j++) {
          //fillPixel(j,pixelY); //!!!!
        }
      }
    }
  }
}

int iceil(float x) {
  return lroundf(std::ceil(x));
}
int ifloor(float x) {
  return lroundf(std::floor(x));
}

//Utility functions
#define MAX_POLY_CORNERS 4  //we are only dealing with quadra-laterals.
//  public-domain code by Darel Rex Finley, 2007, http://alienryderflex.com/polygon_fill/
//  here the IMAGE_TOP<IMAGE_BOT, graphics convention, has to be convex quadralaterals.
//  image_top,image_bot,left,right forms a BOUNDING BOX for this quadralaterals.

int fill_quadralaterl_c(float* polyX, float* polyY, int IMAGE_TOP, int IMAGE_BOT, int IMAGE_LEFT, int IMAGE_RIGHT, 
                       const cmplxf& valc, cubeData<cmplxf>& cube, int polyCorners = 4) { //default quadrilateral, can do traingle as well.
  int i,j;
  float swap; int pixelY;
  int   nodes;
  float nodeX[MAX_POLY_CORNERS];
  int ret = 0;
  //  Loop through the rows of the image.
  for (pixelY=IMAGE_TOP; pixelY<IMAGE_BOT; pixelY++) {
  
    //  Build a list of nodes, since all input polygons are complex, at most you will have 2 nodes.
    nodes=0; j = polyCorners-1;
    for (i=0; i<polyCorners; i++) { //if the horizontal line y=pixelY cross the edge[i].
      if ((polyY[i]<(double) pixelY && polyY[j]>=(double) pixelY)
      ||  (polyY[j]<(double) pixelY && polyY[i]>=(double) pixelY) ) {
        nodeX[nodes++]= float (polyX[i]+(pixelY-polyY[i])/(polyY[j]-polyY[i])
        *(polyX[j]-polyX[i]));  //find the crossing points
      }
      j=i; //step to next edge.
    }
    
    //Sort the nodes.
    i=0;
    if (nodes==2) {
      if (nodeX[i]>nodeX[i+1]) {
          swap=nodeX[i]; nodeX[i]=nodeX[i+1]; nodeX[i+1]=swap; 
      }
      if(lroundf(nodeX[0])<0) {
        PRVEC(nodeX,2);
        PRVEC(polyX,4); PRVEC(polyY,4); PRV4(IMAGE_TOP,IMAGE_BOT,IMAGE_LEFT,IMAGE_RIGHT);
      }
      VERIFY(lroundf(nodeX[0])>=0);
    } else {
      if (nodes!=0) { //find one segment or nothing, there is no third possibility.
        float mix,miy,maxx,maxy;
        mix = *std::min_element(polyX,polyX+4);  maxx = *std::max_element(polyX,polyX+4);
        miy = *std::min_element(polyY,polyY+4);  maxy = *std::max_element(polyY,polyY+4);
        //if ( mix<7 || miy<7 || maxx>9 || maxy >9)  {
          PRV2(nodes,pixelY); PRVEC(nodeX,2); 
          PRVEC(polyX,4); PRVEC(polyY,4); PRV4(IMAGE_TOP,IMAGE_BOT,IMAGE_LEFT,IMAGE_RIGHT);
          prt("!!!!! Node neq 0, should exit now!!\n");
        //}
        nodes = 0; //ignore it this time.
        //return -1;
      }
    }
    
    if (pixelY<0) {VERIFY(nodes==0);}
    //  Fill the pixels between node pairs.
    for (i=0; i<nodes; i+=2) { //for each segments
      if   (nodeX[i  ]>=IMAGE_RIGHT) break;
      if   (nodeX[i+1]> IMAGE_LEFT ) {
        if   (nodeX[i  ]< IMAGE_LEFT ) nodeX[i  ]=IMAGE_LEFT-0.5f ;
        if   (nodeX[i+1]> IMAGE_RIGHT) nodeX[i+1]=IMAGE_RIGHT-0.5f;
        for (j=iceil(nodeX[i]); j<iceil(nodeX[i+1]); j++) {
          cube(pixelY,j) = valc; //fillPixel(j,pixelY); 
        }
      }
    }
  } //end for pixelY
  return ret;
}

//convert one azim gather for every gamma
inline void map_gamma_azim_2_khxy(float* khx, float* khy, const axis& ax_gamma_rad, const float& cos_azim, const float& sin_azim, const float& k_mx, const float& k_my, const float& kz, const bool& sign_khy) {
  //Convert (kmx,kmy) to (kmx',kmy'), notice the sign is reversed from Tisserant report
  float k_mx_azim,k_my_azim;
  k_mx_azim=    k_mx*cos_azim + k_my*sin_azim;
  k_my_azim=   -k_mx*sin_azim + k_my*cos_azim;
  //      inverse is ??? The inverse is not consistent, the '-' should be transposed !!
  //      k_mx= k_mx_azim*cos_azim - k_my_azim*sin_azim
  //      k_my= k_mx_azim*sin_azim + k_my_azim*cos_azim
  //notice differrent than my notes because change in sign in beta(the azimuth angle), these below terms are for amplitude balancing purpose, I think.
  
  //float fconst0; 
  float khx_azim,khy_azim_copl;
  
  //fconst0 = -k_my_azim*k_mx_azim/(SQ(kkz_den) + SQ(k_my_azim)); //the const in eq.(7), the coef between khy' and khx'
  if (fabs(k_my_azim)<epsilon) {k_my_azim=(k_my_azim>=0.f? epsilon:-epsilon);}
  float ftmp0,gamma,tan_gamma;
  bool kz_pos = (kz>-epsilon); //kz==0 case might cause some trouble, therefore epsilon

  //PRV2(cos_azim,sin_azim);
  const float cos_delta = k_my_azim/sqrt(SQ(k_my_azim) + SQ(kz));
  for (int i_gamma=0; i_gamma<ax_gamma_rad.n; i_gamma++) {
    gamma = ax_gamma_rad.get_coord(i_gamma); tan_gamma=tan(gamma);
    //IMPORTANT part, calculate khx' acutally! WHY negative sign???? should correspond to eq(6) directly.
    ftmp0 = sqrt(SQ(kz) + SQ(k_my_azim));
    //khx_azim= -tan_gamma * (kkz>0 ? ftmp0:-ftmp0);
    khx_azim= tan_gamma * ( kz_pos ? ftmp0:-ftmp0);
    //why there can be both choices for the sign of khy'
    khy_azim_copl = tan_gamma*k_mx_azim*cos_delta;
    khy_azim_copl= sign_khy ? (khy_azim_copl):(-khy_azim_copl);

    //We know had khx' and khy', we need to rotate them back to khx and khy
    khx[i_gamma]  = khx_azim*cos_azim - khy_azim_copl*sin_azim;
    khy[i_gamma]  = khx_azim*sin_azim + khy_azim_copl*cos_azim;
    //PRV4(khx_azim,khy_azim_copl,khx[i_gamma],khy[i_gamma]);
  }
}

inline float area_triangle(const float& x1, const float& x2, const float& x3, const float& y1, const float& y2, const float& y3) {
  return .5*fabs( x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2) );
}

//pts must be ordered in CW/CCW ordering, then the area can be divided as traingle(1,2,3)+triangle(1,3,4)
inline float area_quadrilateral(const float& x1, const float& x2, const float& x3, const float& x4, const float& y1, const float& y2, const float& y3, const float& y4) {
  return  area_triangle(x1,x2,x3,y1,y2,y3) + 
          area_triangle(x1,x3,x4,y1,y3,y4);
}

/// input is of format (hx,hy,kz,kmx,kmy), output is of (gamma,azim,kz,kmx,kmy)
void Yang::off2ang3d(param_off2ang3d& par, dmRsf<cmplxf>& dm_in, dmRsf<cmplxf>& dm_out) {
  //////////////////////////Initialization part/////////////////////////////////
  int i_k_mx,i_k_my,i_k_z,i_k_hx,i_k_hy;
                     
  bool apply_jac_both=  ( par.apply_jac_gamma &&  par.apply_jac_azim);
  bool apply_jac_none=  (!par.apply_jac_gamma && !par.apply_jac_azim);

  vecAxis axes_in  = par.shapeOff.get_axes();
  int n_kz,n_kmx,n_kmy,n_hx,n_hy,n_hx_pad,n_hy_pad,n_khx,n_khy, n_khxy;
  n_kz  =axes_in[2].n;n_kmx =axes_in[3].n;n_kmy =axes_in[4].n;
  n_hx  =axes_in[0].n;n_hx_pad=par.n_hx_pad;  n_khx=n_hx_pad;
  n_hy  =axes_in[1].n;n_hy_pad=par.n_hy_pad;  n_khy=n_hy_pad; n_khxy = n_khx*n_khy;
  
  //Ask n_hxy_pad to be even, if not 1
  if(n_hx_pad%2 != 0 || (n_hy_pad%2 != 0 && n_hy_pad!=1)) {
    errPrt("!! n_hx_pad/n_hy_pad must be even, cur val=%d, %d\n",n_hx_pad,n_hy_pad);
  }
  
  axis ax_hx = axes_in[0], ax_hy = axes_in[1];
  float hx_o = ax_hx.o, hy_o = ax_hy.o;
  float d_kz=2*pi*axes_in[2].d, d_kmx=2*pi*axes_in[3].d, d_kmy=2*pi*axes_in[4].d;
  float d_khx=2.*pi/(ax_hx.d*float(n_hx_pad)), d_khy=2.*pi/(ax_hy.d*float(n_hy_pad));
  //PRV2(n_kz,d_kz);
  //VERIFY(fabs(d_kz*n_kz-2*pi/dz)<0.01*2*pi/dz); VERIFY(fabs(d_kmx*n_kmx-2*pi/dx)<0.01*2*pi/dx); VERIFY(fabs(d_kmy*n_kmy-2*pi/dy)<0.01*2*pi/dy);
  //float kz_o=0.f, kmx_o=0.f, kmy_o=0.f;

  axis ax_azim_rad, ax_gamma_rad; bool bFromDegToRadian = true;
  deg2radianAxis(ax_azim_rad,   par.ax_azim,  bFromDegToRadian);
  deg2radianAxis(ax_gamma_rad,  par.ax_gamma, bFromDegToRadian);
  ax_gamma_rad.print_info("ax_gamma_rad");

  int n_threads = g_NUM_THREADS;

  if(n_kz % n_threads != 0)  errPrt("!!n_kz be multiple of n_threads, cur val:%d,%d ",n_kz,n_threads);
  int n_bytes_hx          = n_hx*n_threads*8;
  int n_bytes_gamma_azim  = ax_azim_rad.n*ax_gamma_rad.n*n_threads*8;

  std::valarray<float> k_z,k_mx,k_my,k_hx,k_hy;
  //Allocate the axis variable arrays.
  k_z.resize(n_kz); k_mx.resize(n_kmx); k_my.resize(n_kmy); k_hx.resize(n_khx); k_hy.resize(n_khy);
  boost::shared_array<cubeData<cmplxf> > hx_hy(new cubeData<cmplxf>[n_threads]), gamma_azim(new cubeData<cmplxf>[n_threads]); //one array for each thread
  for (int i=0;i<n_threads; i++) {hx_hy[i].alloc(hypercube(n_hx_pad,n_hy_pad));}
  for (int i=0;i<n_threads; i++) {gamma_azim[i].alloc(hypercube(ax_gamma_rad.n,ax_azim_rad.n));}
  PRV2(ax_gamma_rad.n,ax_azim_rad.n);
  gamma_azim[0].domain().info("gamma_azim");
  
  axis ax_kz = axes_in[2], ax_kmx = axes_in[3], ax_kmy = axes_in[4];
  //fill in the array values, convert the k axis from [0->N] to [0,N/2]U[-N/2,-1]
  for (i_k_z=0; i_k_z < n_kz; i_k_z++) {
    if(i_k_z <= float(n_kz/2)) k_z[i_k_z] =  ax_kz.o+d_kz*i_k_z;
    else     k_z[i_k_z] = ax_kz.o+d_kz*(i_k_z-n_kz); //!!! Notice it was NEGATIVE: -ax_kz.o in the original code, same for the other axes
  }
  //d_kz*n_kz = 2*pi*kz_d*n_kz should be exactly 2*pi, same applies to all following axes kmx and kmy
  for (i_k_mx=0;i_k_mx<n_kmx; i_k_mx++) {
       if(i_k_mx <= float(n_kmx/2)) {
          k_mx[i_k_mx] =  ax_kmx.o+d_kmx*i_k_mx;
       }  else {
          k_mx[i_k_mx] =  ax_kmx.o+d_kmx*(i_k_mx-n_kmx);
       }
  }
  for (i_k_my=0;i_k_my<n_kmy; i_k_my++) {
       if(i_k_my <= float(n_kmy/2)) {
          k_my[i_k_my] =  ax_kmy.o+d_kmy*i_k_my;
       }  else {
          k_my[i_k_my] =  ax_kmy.o+d_kmy*(i_k_my-n_kmy);
       }
  }

  float  khx_o = 0.f, khy_o = 0.f;
  for (i_k_hx=0; i_k_hx<n_khx; i_k_hx++) {
       if(i_k_hx <= float(n_khx/2)) {
          k_hx[i_k_hx] = khx_o+d_khx*i_k_hx;
       }  else {
          k_hx[i_k_hx] = khx_o+d_khx*(i_k_hx-n_khx);
       }
  }

  //call reorder_fft_wavenumber(FFT_FORW,n_khx,k_hx)
  //khx_0=k_hx(1)

  for (i_k_hy=0; i_k_hy<n_khy; i_k_hy++) {
       if(i_k_hy <= float(n_khy/2)) {
          k_hy[i_k_hy] =  khy_o+d_khy* i_k_hy;
       }  else {
          k_hy[i_k_hy] = khy_o+d_khy*(i_k_hy-n_khy);
       }
  }
  
  PRV(par.debug1);
  bool sign_khy     = par.bSign_khy;
  int number_aliases= par.number_aliases;
  if(par.debug1) {
    PRV8(number_aliases,n_kz,n_kmx,n_kmy,n_hx,n_hy,n_hx_pad,n_hy_pad);
    PRV4(n_bytes_hx,n_bytes_gamma_azim,par.apply_jac_gamma,par.apply_jac_azim);
    //PRVEC(k_hx,k_hx.size());    PRVEC(k_hy,k_hy.size());
    //PRVEC(k_mx,k_mx.size());  PRVEC(k_my,k_my.size());
    //PRVEC(k_z,k_z.size());
  }

  ////////////////////////  Computation part Start ////////////////////
  bool in_bounds = true;
  float inv_d_khx=1.f/d_khx, inv_d_khy=1.f/d_khy; int i_kmx,i_kmy,i_kz,i_hy,i_azim,i_gamma,i_thread;
  float azim,gamma,kkz,kkz_den,sin_azim,cos_azim, k_mx_azim,k_my_azim,dkmx_dazim,dkmy_dazim;
  float sin_delta=0.f,cos_delta=0.f,tan_alpha=0.f;//jac_azim_norm=0.f;
  float tan_gamma, tan_gamma_jac;
  float ftmp0;
  
  //prepare for doing fft 2d
  fftwf_complex* fftwout2d[n_threads]; fftwf_complex*  fftwin2d[n_threads]; //pointer array
  fftwf_plan   FFTplan[n_threads]; 
  for (i_thread=0; i_thread<n_threads; i_thread++) {
    fftwout2d[i_thread] = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n_khxy); //these dimensions are after padding
    fftwin2d[i_thread]  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n_khxy);
    memset(fftwout2d[i_thread],0,n_khxy*sizeof(fftwf_complex)); memset(fftwin2d[i_thread],0,n_khxy*sizeof(fftwf_complex));
    FFTplan[i_thread]   = fftwf_plan_dft_2d(n_khy,n_khx,fftwin2d[i_thread],fftwout2d[i_thread],FFTW_BACKWARD,FFTW_MEASURE); //because we use different fft conventions, that is why it is backward(sign=+1).
  }
  VERIFY(sizeof(cmplxf)==sizeof(fftwf_complex));
  
  //We precompute the phase shift caused by the non zero hx_o and hy_o
  cubeData<cmplxf> phsh_hxy(hypercube(n_khx,n_khy));
  for (int i_khy=0; i_khy<n_khy; i_khy++){
    float offkhy = i_khy*d_khy*hy_o;
    for (int i_khx=0;i_khx<n_khx; i_khx++) { 
      phsh_hxy(i_khy   ,i_khx) = exp(cmplxf(0,i_khx*d_khx*hx_o+offkhy)); 
    }
  }

  //For inbound test of the angles<->khxy mapping.  
  int hf_nkhx = n_khx/2, hf_nkhy = n_khy/2;
  int n_khxl = -n_khx*number_aliases  - hf_nkhx; int n_khxh = (n_khx*(number_aliases+1)  -1 - hf_nkhx);
  int n_khyl = -n_khy*number_aliases  - hf_nkhy; int n_khyh = (n_khy*(number_aliases+1)  -1 - hf_nkhy);

  //For each kmy,kmx and kz, we are operating on a plane of (hx,hy), want to reach (khx,khy) then finally (gamma,beta)
  //n_kmy=1;n_kmx=1;
  for (i_kmy=0;i_kmy<n_kmy; i_kmy++) {
  for (i_kmx=0;i_kmx<n_kmx; i_kmx++) {
    //discard i_kz=0 piece
    for (i_hy=0; i_hy<n_hy; i_hy++) {
      //read nhx elements of data, due to the padding on hx and hy axis, we cannot read the 2d portion at once
      dm_in.seq_grab(&(hx_hy[0](i_hy,0)), n_hx);
    }
    gamma_azim[0]  = cmplxf(0.f,0.f);
    dm_out.seq_update(&(gamma_azim[0]));

    //now i_kz>0 case, actually we only need to compute half the kz spectrum [0,pi], due to the symmetry of fourier transform.
    for (i_kz=1; i_kz <n_kz; i_kz+=n_threads) {
      for (i_thread=0; i_thread<n_threads; i_thread++) {
        hx_hy[i_thread]       = cmplxf(0.f,0.f);  //initialize all to zero
        gamma_azim[i_thread]  = cmplxf(0.f,0.f);
        //Read in this (hx,hy) section
        for (i_hy=0; i_hy<n_hy; i_hy++) {
          //read nhx elements of data, due to the padding on hx and hy axis, we cannot read the 2d portion at once
          dm_in.seq_grab(&(hx_hy[i_thread](i_hy,0)), n_hx);
        }
      }
      
      for (i_thread=0; i_thread<n_threads; i_thread++) {
        kkz=k_z[i_kz+i_thread];
        //cannot have k_z equal to zero
        if (i_kz+i_thread == 0) kkz_den=k_z[1];  else kkz_den=kkz;
        
        //PRVEC((hx_hy[i_thread]),n_hx_pad*n_hy_pad); //test
        //doing 2D FT (hx,hy)->(khx,khy) !!!!!
        memcpy(fftwin2d[i_thread],hx_hy[i_thread].pval,n_khxy*sizeof(fftwf_complex)); //fill in the data
        fftwf_execute(FFTplan[i_thread]);
        memcpy(hx_hy[i_thread].pval,fftwout2d[i_thread],n_khxy*sizeof(fftwf_complex)); //retrieve the data
        //PRVEC(((cmplxf*)fftwin2d[i_thread]),n_hx_pad*n_hy_pad);  PRVEC((hx_hy[i_thread]),n_hx_pad*n_hy_pad); //test
        
        //We shift back the phase because non-zero hx_o and hy_o ,this is definitely better than adjusting the phase later during interpolation
        for (int i_khy=0; i_khy<n_khy; i_khy++){
          for (int i_khx=0;i_khx<n_khx; i_khx++) { 
            hx_hy[i_thread](i_khy   ,i_khx) *= phsh_hxy(i_khy,i_khx);
          }
        }
        
        for (i_azim=0; i_azim<ax_azim_rad.n; i_azim++){ //here azim is the beta angle
          azim = ax_azim_rad.get_coord(i_azim);
          sin_azim=sin(azim); cos_azim=cos(azim);
          //Convert (kmx,kmy) to (kmx',kmy'), notice the sign is reversed from Tisserant report
          k_mx_azim=    k_mx[i_kmx]*cos_azim + k_my[i_kmy]*sin_azim;
          k_my_azim=   -k_mx[i_kmx]*sin_azim + k_my[i_kmy]*cos_azim;
          //      inverse is ??? The inverse is not consistent, the '-' should be transposed !!
          //      k_mx= k_mx_azim*cos_azim - k_my_azim*sin_azim
          //      k_my= k_mx_azim*sin_azim + k_my_azim*cos_azim
          //notice differrent than my notes because change in sign in beta(the azimuth angle), these below terms are for amplitude balancing purpose, I think.
          dkmx_dazim= - k_mx[i_kmx]*sin_azim + k_my[i_kmy]*cos_azim;
          dkmy_dazim= - k_mx[i_kmx]*cos_azim - k_my[i_kmy]*sin_azim;
  
          //PRV4(sin_azim,cos_azim,k_mx_azim,k_my_azim); PRV(kkz);
          
          float fconst0,jac,jac_samp,khx_azim,khy_azim_copl,khx_copl,khy_copl;
          float khx_copl_int,w_khx_copl,khy_copl_int,w_khy_copl;
          int i_khx_copl,i_khy_copl,i_khx_copl_p1,i_khy_copl_p1;
          float inv_cos_gamma_sq,dkhx_dazim,dkhy_dazim,dkhx_dgamma,dkhy_dgamma;
          if(par.create_coplanar_ang) { //normal case, calculate both azim and gamma
          //tag_output=trim(off2ang3d_par%tag_coplanar_ang)
            
            fconst0 = -k_my_azim*k_mx_azim/(SQ(kkz_den) + SQ(k_my_azim)); //the const in eq.(7), the coef between khy' and khx'
            //This block is for generate amplitude compensated gathers, not about the kinematics.
            if(!apply_jac_none) { 
              sin_delta=k_my_azim/sqrt(SQ(k_my_azim) + SQ(kkz_den)); cos_delta=kkz/sqrt(SQ(k_my_azim) + SQ(kkz_den)); //delta correspond to eq.(2)
              //alpha is the inline dip angle defined in app (B-5).
              tan_alpha=k_mx_azim/kkz_den;  //jac_azim_norm = ax_azim_rad.d/sqrt(d_khx**2 + d_khy**2);
            }
      
            for (i_gamma=0; i_gamma<ax_gamma_rad.n; i_gamma++) {
              gamma = ax_gamma_rad.get_coord(i_gamma); tan_gamma=tan(gamma);
              tan_gamma_jac = 0.5f*( abs(tan(gamma-0.5f*ax_gamma_rad.d))+abs(tan(gamma+0.5*ax_gamma_rad.d)) ); //?what for? seems to calc a absolute local average of tan_gamma.
              //PRV2(tan_gamma,tan_gamma_jac);
              if(apply_jac_none)  { jac=1.f; 
              }  else if (apply_jac_both) {
                inv_cos_gamma_sq = (1.f+SQ(tan_gamma_jac));
                dkhx_dazim  =tan_gamma_jac*sin_delta*dkmy_dazim;
                dkhy_dazim  =tan_gamma_jac*(sin_delta*dkmx_dazim - tan_alpha*(CB(cos_delta))*dkmy_dazim);
                dkhx_dgamma =kkz_den*inv_cos_gamma_sq/cos_delta;
                dkhy_dgamma =kkz_den*inv_cos_gamma_sq*tan_alpha*sin_delta;
    
                jac_samp=(dkhx_dgamma*dkhy_dazim + dkhy_dgamma*dkhx_dazim)*(ax_azim_rad.d/(d_khx*d_khy));
                jac= max (.05f,jac_samp);
              }  else if (par.apply_jac_azim) {
                dkhx_dazim=tan_gamma_jac*sin_delta*dkmy_dazim;
                dkhy_dazim=tan_gamma_jac*(sin_delta*dkmx_dazim - tan_alpha*(CB(cos_delta))*dkmy_dazim);
      
                jac_samp=sqrt(SQ(dkhx_dazim*ax_azim_rad.d/d_khx) + SQ(dkhy_dazim*ax_azim_rad.d/d_khy) );
                jac=max(.05f,jac_samp);
                //jac=max(1.,sqrt(dkhx_dazim**2+dkhy_dazim**2)*jac_azim_norm)
                //jac=sqrt(dkhx_dazim**2 + dkhy_dazim**2);
              }  else if (par.apply_jac_gamma) {
                inv_cos_gamma_sq=(1.+SQ(tan_gamma_jac)); //why it is not tan gamma rather, 1/cos(gamma)
                dkhx_dgamma=kkz_den*inv_cos_gamma_sq/cos_delta;   //these two are only for energy scaling term
                dkhy_dgamma=kkz_den*inv_cos_gamma_sq*tan_alpha*sin_delta;
                jac=sqrt(SQ(dkhx_dgamma) + SQ(dkhy_dgamma));
              }  else {VERIFY(0);}
    
              if(par.ang_2d) {
                khx_azim= tan_gamma*kkz;
                //khx_azim= tan_gamma*(sqrt(kkz**2));
              } else {
                //IMPORTANT part, calculate khx' acutally! WHY negative sign???? should correspond to eq(6) directly.
                ftmp0 = sqrt(SQ(kkz) + SQ(k_my_azim));
                //khx_azim= -tan_gamma * (kkz>0 ? ftmp0:-ftmp0);
                khx_azim= tan_gamma * (kkz>0 ? ftmp0:-ftmp0);
              }
              //why there can be both choices for the sign of khy'
              khy_azim_copl= sign_khy ? (-khx_azim*fconst0):(khx_azim*fconst0);
  
              //We know had khx' and khy', we need to rotate them back to khx and khy
              khx_copl= khx_azim*cos_azim - khy_azim_copl*sin_azim;
              khy_copl= khx_azim*sin_azim + khy_azim_copl*cos_azim;
              //PRV4(khx_azim,khy_azim_copl,khx_copl,khy_copl); PRV(ftmp0);
              //Here we just do bilinear interpolation, find the 4 corner grid pts and the related weights,
              //notice khx and khy can be Negative value on the k axis.
              khx_copl_int=(khx_copl-khx_o)*inv_d_khx;
              i_khx_copl=int(floor(khx_copl_int)); //round down
              w_khx_copl=khx_copl_int-i_khx_copl;
    
              khy_copl_int=(khy_copl-khy_o)*inv_d_khy;
              i_khy_copl=int(floor(khy_copl_int)); //round down
              w_khy_copl=khy_copl_int-i_khy_copl;
              //PRV4(khx_copl_int,i_khx_copl,khy_copl_int,i_khy_copl);
              //PRV4(n_khxl,n_khxh,n_khyl,n_khyh);
              //seems number_aliases=0 is OK. however since khx_o, khy_o=0. It is possible that i_khx_copl would be negative, this should not be considered out of bounds.
  //            if((i_khx_copl > -n_khx*number_aliases -1) && (i_khx_copl <= (n_khx*(number_aliases+1)  -1) ) &&
  //               (i_khy_copl > -n_khy*number_aliases -1) && (i_khy_copl <= (n_khy*(number_aliases+1)) -1) ) {
              if((i_khx_copl >= n_khxl) && (i_khx_copl <= n_khxh ) &&
                 (i_khy_copl >= n_khyl) && (i_khy_copl <= n_khyh ) ) {
                in_bounds = true;
                i_khx_copl    =((number_aliases+1)*n_khx+i_khx_copl) % n_khx;
                i_khx_copl_p1 =((number_aliases+1)*n_khx+i_khx_copl+1) % n_khx;
                i_khy_copl    =((number_aliases+1)*n_khy+i_khy_copl) % n_khy;
                i_khy_copl_p1 =((number_aliases+1)*n_khy+i_khy_copl+1) % n_khy;
              } else {
                  in_bounds = false;
              }
    
              //debug msg prt
              if((i_kmy == 0) && (i_kmx == 0) && (i_kz == 0)) {
              //if((i_kmy == n_kmy/4 -1) && (i_kmx == n_kmx/4 -1) && (i_kz == n_kz/4 -1)) {
  //            prt("*****\n");
  //            PRV4(azim*(180/pi),gamma*(180/pi),jac_samp,jac);
  //            prt("*****\n");
              }
              
              //bilinear interpolation for the specific azim,gamma location, the weights are checked.
              if(in_bounds) {
                //notice our () indexing is slow dimension first, this is in reverse of fortran's ordering
                gamma_azim[i_thread](i_azim,i_gamma) = 
                  (hx_hy[i_thread](i_khy_copl   ,i_khx_copl)    *(1.f-w_khx_copl)*(1.f-w_khy_copl) +
                   hx_hy[i_thread](i_khy_copl_p1,i_khx_copl)    *(1.f-w_khx_copl)*(  w_khy_copl) +
                   hx_hy[i_thread](i_khy_copl   ,i_khx_copl_p1) *(   w_khx_copl)*(1.f-w_khy_copl) +
                   hx_hy[i_thread](i_khy_copl_p1,i_khx_copl_p1) *(   w_khx_copl)*(  w_khy_copl)) * jac;
                  //hx_hy[i_thread]((lroundf(khy_copl_int)+n_khy)%n_khy,(lroundf(khx_copl_int)+n_khx)%n_khx) *  //nearest neighbor interpolation
                  //exp(cmplxf(0,khx_copl*hx_o+khy_copl*hy_o)) ; //this is consistent with the sign=1 as the forward fourier transform
                //cerr << jac << sep << i_khy_copl << sep << i_khy_copl_p1 << sep << i_khx_copl << sep << i_khx_copl_p1 << sep << gamma_azim[i_thread](i_azim,i_gamma) << w_khx_copl << sep << w_khy_copl << endl;
                //PRVEC((hx_hy[i_thread]),n_hx_pad*n_hy_pad);
                if (abs(gamma_azim[i_thread](i_azim,i_gamma)) > 1e10) {
                  PRV2(exp(cmplxf(0,khx_copl*hx_o+khy_copl*hy_o)),jac);
                  PRV4(hx_hy[i_thread](i_khx_copl   ,i_khy_copl),hx_hy[i_thread](i_khx_copl   ,i_khy_copl_p1),
                       hx_hy[i_thread](i_khx_copl_p1   ,i_khy_copl),hx_hy[i_thread](i_khx_copl_p1   ,i_khy_copl_p1));
                  errPrt("gamma_azim val is huge!!\n");
                }
                //if our hx and hy axis does not start from zero, (but the Ft3d program will think it starts from zero) then we should re-adjust this phase shift back.
              }
            } //for igamma
        } else { //if (!par.create_coplanar_ang)
          VERIFY(0);
//          //tag_output=trim(off2ang3d_par%tag_coplanar_stack)
//          if(par.create_stack) {
//            gamma_azim[i_thread](i_azim,0)=cmplxf(0.f,0.f);
//            for (int i_khy=0; i_khy<n_khy; i_khy++) {
//              for (int i_khx=0; i_khx<n_khx; i_khx++) {
//                gamma_azim[i_thread](i_azim,0) += hx_hy[i_thread](i_khy,i_khx) *
//                  exp(cmplxf(0,k_hx[i_khx]*hx_o + k_hy[i_khy]*hy_o));
//              }
//            }
//          } else { // if par.create_stack
//            gamma_azim[i_thread](i_azim,0)=cmplxf(0.f,0.f);
//            for (int i_khy=0; i_khy<n_khy; i_khy++) {
//              if(!apply_jac_none) {
//                sin_delta=k_my_azim/sqrt(SQ(k_my_azim) + SQ(kkz_den));
//                cos_delta=abs(kkz_den)/sqrt(SQ(k_my_azim) + SQ(kkz_den));
//                tan_alpha=k_mx_azim/kkz_den;
//              }
//              fconst0 =-k_my_azim*k_mx_azim/(SQ(kkz_den) + SQ(k_my_azim));
//
//              for (int i_khx=0; i_khx < n_khx; i_khx++) {
//                khx_azim      = k_hx[i_khx];
//                khy_azim_copl = sign_khy*khx_azim*fconst0;
//                if(apply_jac_none) {
//                  jac=1;
//                } else {
//                  tan_gamma_jac=(khx_azim*cos_delta)/kkz_den;
//                  if(par.apply_jac_azim) {
//                    dkhx_dazim=tan_gamma_jac * sin_delta*dkmy_dazim;
//                    dkhy_dazim=tan_gamma_jac * (sin_delta*dkmx_dazim - tan_alpha*CB(cos_delta)*dkmy_dazim);
//                    jac_samp= sqrt(SQ(dkhx_dazim*ax_azim_rad.d/d_khx) + SQ(dkhy_dazim*ax_azim_rad.d/d_khy));
//                    jac= max(.05f,jac_samp);
//                    //jac=sqrt(SQ(dkhx_dazim) + SQ(dkhy_dazim))
//                  } else if (par.apply_jac_gamma) {
//                      seperr("Not possible to apply_jac_gamma in this case");
//                  } else if (apply_jac_both) {
//                      seperr("Not possible to apply_jac_both in this case");
//                  } else VERIFY(0);
//                }
//
//                khx_copl= khx_azim*cos_azim - khy_azim_copl*sin_azim;
//                khy_copl= khx_azim*sin_azim + khy_azim_copl*cos_azim;
//
//                khx_copl_int=(khx_copl-khx_o)*inv_d_khx;
//                i_khx_copl=int(khx_copl_int-.5f);
//                w_khx_copl=khx_copl_int-i_khx_copl;
//
//                khy_copl_int=(khy_copl-khy_o)*inv_d_khy;
//                i_khy_copl=int(khy_copl_int-.5f);
//                w_khy_copl=khy_copl_int-i_khy_copl;
//
//                if((i_khx_copl > -n_khx*number_aliases-1) &&
//                   (i_khx_copl <= n_khx*(number_aliases+1)-1)  &&
//                   (i_khy_copl >  -n_khy*number_aliases-1   ) &&
//                   (i_khy_copl <= n_khy*(number_aliases+1)-1) ) {
//                  in_bounds = true;
//                  i_khx_copl   =((number_aliases+1)*n_khx+i_khx_copl)%n_khx;
//                  i_khx_copl_p1=((number_aliases+1)*n_khx+i_khx_copl+1)%n_khx;
//
//                  i_khy_copl    = ((number_aliases+1)*n_khy+i_khy_copl)  %  n_khy;
//                  i_khy_copl_p1 = ((number_aliases+1)*n_khy+i_khy_copl+1)%  n_khy;
//                } else {
//                  in_bounds= false;
//                }
//                //Debug
//                if((i_kmy == 0)&&(i_kmx == n_kmx/2)&&(i_azim == (ax_azim_rad.n+1)/2)) {
//                  //  write(0,*)'***'
//                  //  write(0,*)'azim=',azim
//                  //  write(0,*)'in_bounds=',in_bounds
//                  //  write(0,*)'kkz_den=',kkz_den
//                  //  write(0,*)'kmy=',k_my(i_kmy)
//                  //  write(0,*)'kmx=',k_mx(i_kmx)
//                  //  write(0,*)'sin_delta=',sin_delta
//                  //  write(0,*)'cos_delta=',cos_delta
//                  //  write(0,*)'tan_alpha=',tan_alpha
//                  //  write(0,*)'tan_gamma_jac=',tan_gamma_jac
//                  //  write(0,*)'jac_samp=',jac_samp
//                  //  write(0,*)'jac=',jac
//                  //  write(0,*)'***'
//                }
//
//                if(in_bounds) {
//                  gamma_azim[i_thread](i_azim,0) +=
//   (hx_hy[i_thread](i_khy_copl    ,i_khx_copl)    *(1.f-w_khx_copl)*(1.f-w_khy_copl) +
//    hx_hy[i_thread](i_khy_copl_p1 ,i_khx_copl)    *(1.f-w_khx_copl)*(  w_khy_copl) +
//    hx_hy[i_thread](i_khy_copl    ,i_khx_copl_p1) *(   w_khx_copl)*(1.f-w_khy_copl) +
//    hx_hy[i_thread](i_khy_copl_p1 ,i_khx_copl_p1) *(   w_khx_copl)*(  w_khy_copl)) *
//                    exp(cmplxf(0,khx_copl*hx_o+khy_copl*hy_o)) * jac; //??? multiply hx_o not hx_d ???
//                }
//              } //end for i_khx
//            } //end for i_khy
//          } //endif par.create_stack
          } //endif (par.create_coplanar_ang)
        } //for i_azim
      } //for i_thread
      
      //gamma_azim[0].domain().info("gamma_azim cube");
      //LOGC("before seq_update"); PRV(n_threads);
      //PRVEC(gamma_azim[0],gamma_azim[0].size());
      //write the output to (gamma,azim,kz,kmx,kmy)
      for (i_thread=0; i_thread<n_threads; i_thread++) {
        dm_out.seq_update(&(gamma_azim[i_thread]));
      }
      //LOGC("after seq_update");
    }
  }} //for i_kz //for i_kmx //for i_kmy
  
  LOGC("Finish off3ang3d main loop");
  //cleaned up fftw
  for (i_thread=0; i_thread<n_threads; i_thread++) {
    fftwf_free(fftwout2d[i_thread]);  fftwf_free(fftwin2d[i_thread]);
    fftwf_destroy_plan(FFTplan[i_thread]);
  }
  LOGC("finish off3ang3d function");
  return;
}

/// input is of format (gamma,azim,kz,kmx,kmy), output is of (hx,hy,kz,kmx,kmy)
/// !!!! Not Finished yet !!!!!
void Yang::ang2off3d(param_off2ang3d& par, dmRsf<cmplxf>& dm_in, dmRsf<cmplxf>& dm_out, bool bDoFTsmoothing) {
  //////////////////////////Initialization part/////////////////////////////////
  int i_k_mx,i_k_my,i_k_z,i_k_hx,i_k_hy;
                     
  //bool apply_jac_both=  ( par.apply_jac_gamma &&  par.apply_jac_azim);
  //bool apply_jac_none=  (!par.apply_jac_gamma && !par.apply_jac_azim);

  vecAxis axes_in  = par.shapeAng.get_axes();
  int n_kz,n_kmx,n_kmy,n_hx_pad,n_hy_pad,n_khx,n_khy, n_khxy;
  int n_azim,n_gamma;
  n_kz    = axes_in[2].n;  
  n_kmx   = axes_in[3].n;   
  n_kmy   = axes_in[4].n;
  n_hx_pad= par.n_hx_pad;  n_khx=n_hx_pad;   n_hy_pad=par.n_hy_pad;  n_khy=n_hy_pad; n_khxy = n_khx*n_khy;
  n_gamma  =axes_in[0].n; n_azim =axes_in[1].n;
  axis ax_gamma = axes_in[0], ax_azim = axes_in[1];

  bool bTrue2Dmapping = ! (n_azim==1 && (equalf(ax_azim.get_coord(0),0.0f))); //otherwise it is just a 1D mapping.
  
  //Ask n_hxy_pad to be even, if not 1
  if(n_hx_pad%2 != 0 || (n_hy_pad%2 != 0 && n_hy_pad!=1)) {
    errPrt("!! n_hx_pad/n_hy_pad must be even, cur val=%d, %d\n",n_hx_pad,n_hy_pad);
  }
  
  vecAxis axes_off = par.shapeOff.get_axes();
  axis ax_hx = axes_off[0], ax_hy = axes_off[1];
  int n_hx,n_hy; n_hx = ax_hx.n; n_hy = ax_hy.n;
  float d_khx=2.*pi/(ax_hx.d*float(n_hx_pad)), d_khy=2.*pi/(ax_hy.d*float(n_hy_pad));
  
  float d_kz=2*PI*axes_in[2].d, d_kmx=2*pi*axes_in[3].d, d_kmy=2*pi*axes_in[4].d;
  //VERIFY(fabs(d_kz*n_kz-2*pi/dz)<0.01*2*pi/dz); VERIFY(fabs(d_kmx*n_kmx-2*pi/dx)<0.01*2*pi/dx); VERIFY(fabs(d_kmy*n_kmy-2*pi/dy)<0.01*2*pi/dy);
  //float kz_o=0.f, kmx_o=0.f, kmy_o=0.f;

  axis ax_azim_rad, ax_gamma_rad; bool bFromDegToRadian = true;
  VERIFY(ax_gamma.same_axis(par.ax_gamma));
  deg2radianAxis(ax_azim_rad,   par.ax_azim,  bFromDegToRadian);
  deg2radianAxis(ax_gamma_rad,  par.ax_gamma, bFromDegToRadian);
  ax_gamma_rad.print_info("ax_gamma_rad");

  int n_threads = 1; VERIFY(n_threads==1); //The smooth function is not thread-safe yet..

  if(n_kz % n_threads != 0)  errPrt("!!n_kz be multiple of n_threads, cur val:%d,%d ",n_kz,n_threads);
  int n_bytes_gamma_azim  = ax_azim_rad.n*ax_gamma_rad.n*n_threads*8;
  int n_bytes_hx          = n_hx*n_threads*8;
  
  std::valarray<float> k_z,k_mx,k_my,k_hx,k_hy;
  //Allocate the axis variable arrays.
  k_z.resize(n_kz); k_mx.resize(n_kmx); k_my.resize(n_kmy); k_hx.resize(n_khx); k_hy.resize(n_khy);

  boost::shared_array<cubeData<cmplxf> > hx_hy(new cubeData<cmplxf>[n_threads]), hx_hy2(new cubeData<cmplxf>[n_threads]), gamma_azim(new cubeData<cmplxf>[n_threads]); //one array for each thread
  boost::shared_array<cubeData<float> > cubReal(new cubeData<float>[n_threads]), cubRealsm(new cubeData<float>[n_threads]);
  boost::shared_array<cubeData<float> > cubImg(new cubeData<float>[n_threads]),  cubImgsm(new cubeData<float>[n_threads]);
  
  hypercube shape_khxy(n_hx_pad,n_hy_pad);
  for (int i=0;i<n_threads; i++) {
    hx_hy[i].alloc(shape_khxy);  hx_hy2[i].alloc(shape_khxy); 
    cubReal[i].alloc(shape_khxy);cubRealsm[i].alloc(shape_khxy);cubImg[i].alloc(shape_khxy);cubImgsm[i].alloc(shape_khxy);
  }
  
  for (int i=0;i<n_threads; i++) {gamma_azim[i].alloc(hypercube(ax_gamma_rad.n,ax_azim_rad.n));}
  PRV2(ax_gamma_rad.n,ax_azim_rad.n);
  //gamma_azim[0].domain().info("gamma_azim");
  
  axis ax_kz = axes_in[2], ax_kmx = axes_in[3], ax_kmy = axes_in[4];
  //fill in the array values, convert the k axis from [0->N] to [0,N/2-1]U[-N/2,-1], i.e. {0,1,2,3,4,5,6,7}->{0,1,2,3,-4,-3,-2,-1}, it is relative convinient because 
  //the regular FFT assumes starts from 0.
  for (i_k_z=0; i_k_z < n_kz; i_k_z++) {
    if(i_k_z <= float(n_kz/2)) k_z[i_k_z] =  ax_kz.o+d_kz*i_k_z;
    else     k_z[i_k_z] = ax_kz.o+d_kz*(i_k_z-n_kz); //!!! Notice it was NEGATIVE: -ax_kz.o in the original code, same for the other axes
  }
  //d_kz*n_kz = 2*pi*kz_d*n_kz should be exactly 2*pi, same applies to all following axes kmx and kmy
  for (i_k_mx=0;i_k_mx<n_kmx; i_k_mx++) {
       if(i_k_mx < (n_kmx/2)) {
          k_mx[i_k_mx] =  ax_kmx.o+d_kmx*i_k_mx;
       }  else {
          k_mx[i_k_mx] =  ax_kmx.o+d_kmx*(i_k_mx-n_kmx);
       }
  }
  
  //avoid the n_kmy=1 case.
  if (n_kmy==1) { 
    k_my[0] = ax_kmy.o;
  } else {
    for (i_k_my=0;i_k_my<n_kmy; i_k_my++) {
       if(i_k_my < (n_kmy/2)) {
          k_my[i_k_my] =  ax_kmy.o+d_kmy*i_k_my;
       }  else {
          k_my[i_k_my] =  ax_kmy.o+d_kmy*(i_k_my-n_kmy);
       }
    }
  }

  float  khx_o = 0.f, khy_o = 0.f;
  for (i_k_hx=0; i_k_hx<n_khx; i_k_hx++) {
       if(i_k_hx < (n_khx/2)) {
          k_hx[i_k_hx] = khx_o+d_khx*i_k_hx;
       }  else {
          k_hx[i_k_hx] = khx_o+d_khx*(i_k_hx-n_khx);
       }
  }
  
  if (n_khy==1) { 
    k_hy[0] = khy_o;
  } else {
    for (i_k_hy=0; i_k_hy<n_khy; i_k_hy++) {
      if(i_k_hy < (n_khy/2)) {
        k_hy[i_k_hy] =  khy_o+d_khy* i_k_hy;
      }  else {
        k_hy[i_k_hy] = khy_o+d_khy*(i_k_hy-n_khy);
      }
    }
  }
  
  //PRV(par.debug1);
  bool sign_khy     = par.bSign_khy;
  int number_aliases= par.number_aliases;
  if(par.debug1) {
    PRV8(number_aliases,n_kz,n_kmx,n_kmy,n_hx,n_hy,n_hx_pad,n_hy_pad);
    PRV4(n_bytes_hx,n_bytes_gamma_azim,par.apply_jac_gamma,par.apply_jac_azim);
    //PRVEC(k_hx,k_hx.size());    PRVEC(k_hy,k_hy.size());
    //PRVEC(k_mx,k_mx.size());  PRVEC(k_my,k_my.size());
    //PRVEC(k_z,k_z.size());
  }

  ////////////////////////  Computation part Start, from angle gather to offset gather ////////////////////
  //we tend to implitly do fftshift2d to place the zero freq at the center. therefore in the converted domain, we flip the sign every other pt to undo such effect.
  cubeData<float> bFlipCub(shape_khxy); bFlipCub = 1.0f; //cmplxf(1.0f,0.f);
  for (int ikhy = 0; ikhy<n_khy; ikhy++) {
    for (int ikhx = 0; ikhx<n_khx; ikhx++) {
      if ((ikhy+ikhx)%2 != 0) {
        bFlipCub(ikhy,ikhx) = -1.0f; //cmplxf(-1.0f,0.f);
      }
    }
  }

  float inv_d_khx=1.f/d_khx, inv_d_khy=1.f/d_khy; int i_kmx,i_kmy,i_kz,i_azim,i_gamma;
  PRV2(inv_d_khx,inv_d_khy); 
  //float k_mx_azim,k_my_azim,dkmx_dazim,dkmy_dazim;
  //float sin_delta=0.f,cos_delta=0.f,tan_alpha=0.f;//jac_azim_norm=0.f;
  //float gamma,tan_gamma, tan_gamma_jac;
  
  //prepare for doing fft 2d
  fftwf_complex* fftwout2d[n_threads]; fftwf_complex*  fftwin2d[n_threads]; //pointer array
  fftwf_plan   FFTplan[n_threads];
  for (int i=0; i<n_threads; i++) {
    fftwout2d[i] = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n_khxy); //these dimensions are after padding
    fftwin2d[i]  = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*n_khxy);
    memset(fftwout2d[i],0,n_khxy*sizeof(fftwf_complex)); memset(fftwin2d[i],0,n_khxy*sizeof(fftwf_complex));
    FFTplan[i]   = fftwf_plan_dft_2d(n_khy,n_khx,fftwin2d[i],fftwout2d[i],FFTW_FORWARD,FFTW_MEASURE); //because we use different fft conventions, that is why it is Forward(sign=-1).
  }
  VERIFY(sizeof(cmplxf)==sizeof(fftwf_complex));
  
  //We precompute the phase shift caused by the non zero hx_o and hy_o, this should be applied to the slices before fourier transform.
  float hy_o = ax_hy.o, hx_o = ax_hx.o;
  cubeData<cmplxf> phsh_hxy(shape_khxy);
  for (int i_khy=0; i_khy<n_khy; i_khy++){
    float offkhy = -i_khy*d_khy*hy_o;
    for (int i_khx=0;i_khx<n_khx; i_khx++) { 
      phsh_hxy(i_khy   ,i_khx) = exp(cmplxf(0,-i_khx*d_khx*hx_o+offkhy)); //notice this is the conjugate of what was used in the ang2off direction.
    }
  }
  
  //For each kmy,kmx and kz, we are operating on a plane of (hx,hy), first read (gamma,beta) then (khx,khy) finally (hx,hy);
  //n_kmy=1;n_kmx=1; //for test
  gamma_khxy_mapping_method mapping_method = par.m_mapping_method;
  PRV(mapping_method);
  bool bCalcMappingCorners = (mapping_method == gkm_polyfill || mapping_method == gkm_ap_NN);
  
  //For inbound test of the angles<->khxy mapping.  
  int hf_nkhx = n_khx/2, hf_nkhy = n_khy/2;
  int n_khxl,n_khxh,n_khyl,n_khyh;
  if (hf_nkhy!=0) {
    n_khxl = -n_khx*number_aliases  - hf_nkhx; n_khxh = (n_khx*(number_aliases+1)  -1 - hf_nkhx);
    n_khyl = -n_khy*number_aliases  - hf_nkhy; n_khyh = (n_khy*(number_aliases+1)  -1 - hf_nkhy);
  } else { //if n_khy is == 1
    n_khxl = -n_khx*number_aliases  - hf_nkhx; n_khxh = (n_khx*(number_aliases+1)  -1 - hf_nkhx);
    n_khyl = -1; n_khyh = 1;
  }
  
  float khx_corner_min,khx_corner_max,khy_corner_min,khy_corner_max;
  khx_corner_min = (-hf_nkhx-0.4f)*d_khx+khx_o, khx_corner_max = (hf_nkhx-.4f)*d_khx+khx_o; //not 0.5, just to safeguard some engineering surplus.
  
  if (hf_nkhy!=0) {
    khy_corner_min = (-hf_nkhy-0.4f)*d_khy+khy_o, khy_corner_max = (hf_nkhy-.4f)*d_khy+khy_o;
  } else { //if khy axis is dummy
    khy_corner_min = -(0.5f)*d_khy+khy_o, khy_corner_max = (0.5f)*d_khy+khy_o;
  }
  
  axis ax_azim_rad_endpt, ax_gamma_rad_endpt;
  ax_azim_rad_endpt = ax_azim_rad; ax_gamma_rad_endpt = ax_gamma_rad;
  ax_azim_rad_endpt.n   = ax_azim_rad.n+1;  ax_azim_rad_endpt.o   = ax_azim_rad.o   - .5f*ax_azim_rad.d;
  ax_gamma_rad_endpt.n  = ax_gamma_rad.n+1; ax_gamma_rad_endpt.o  = ax_gamma_rad.o  - .5f*ax_gamma_rad.d;
  
  hypercube shape_angles(n_gamma,n_azim);
  cubeData<float> mapped_khx[n_threads], mapped_khy[n_threads];
  cubeData<bool>  map_bInbound[n_threads];
  cubeData<float> mapped_khx_corner[n_threads],mapped_khy_corner[n_threads],mapped_areas[n_threads];
  hypercube s1(ax_gamma_rad_endpt.n,ax_azim_rad_endpt.n);
  for (int i=0; i<n_threads; i++) {
    mapped_khx[i].alloc(shape_angles);  mapped_khy[i].alloc(shape_angles);
    map_bInbound[i].alloc(shape_angles);
    mapped_khx_corner[i].alloc(s1);     mapped_khy_corner[i].alloc(s1);
    mapped_areas[i].alloc(shape_angles);
  }
  
  
  //Check that the gamma axis will sample gamma=0.
  float gamma0loc = ax_gamma.get_offsetf(0.0f); int i_gamma0loc = int(gamma0loc+0.5f); VERIFY(gamma0loc-i_gamma0loc<0.01f);
  //then for ax_gamma_rad_endpt axis, gamma0loc -> gamma0loc+1 will cross gamma=0.
  
  LOG("enter the kx/kmx/kmy main loop");
  dm_in.domain().info("dm_in"); 
  //PRV2(n_gamma,n_azim); //PRV(dm_in.getSeqCounter());
  cubeData<cmplxf> cubeIn,cubeOut;
  cubeIn.alloc(dm_in.domain().return_compatible(1,4));
  cubeOut.alloc(dm_out.domain().return_compatible(1,4));
  int ith = 0; windParam windkxyz[n_threads];
  for (int i=0; i<n_threads; i++) {
    windkxyz[i].resize(4);
    windkxyz[i][3].n = 1; windkxyz[i][2].n = 1;
  }
  
  //dmRsf_ocDump<float> mapAreaDump("dump7","mapArea"); mapAreaDump.init(&mapped_areas.domain());
  VERIFY(par.create_coplanar_ang);
  //PRV4(n_kmx,n_kmy,n_kz,n_threads);
  
  //build a hamming window to damp the (khx,khy) plane, the window is flat in the center, then rolls down with a hamming window from kh_low to kh_high, then stays constant with further bigger frequency.
  cubeData<cmplxf> dampingKhxy(shape_khxy);
  float kh_low,kh_high;
  kh_high = hypot(hf_nkhx*d_khx,hf_nkhy*d_khy);
  kh_low  = 0.6f * kh_high;
  
  float kh_det = PI/(kh_high-kh_low);
  float dist,disty,distx, wval;
  for (int i_khy=0; i_khy<n_khy; i_khy++){
    disty = (i_khy-hf_nkhy)*d_khy; 
    for (int i_khx=0;i_khx<n_khx; i_khx++) { 
      distx = (i_khx-hf_nkhx)*d_khx;
      dist = hypot(disty,distx);
      if (dist <= kh_low) {
        wval = 1.0f;
      } else if (dist <= kh_high) {
        wval = 0.54 - 0.46 * cosf((kh_high-dist)*kh_det);
      } else { //dist > kh_high
        wval = 0.54f-0.46f;
      }
      dampingKhxy(i_khy   ,i_khx) = cmplxf(wval,0.f); //notice this is the conjugate of what was used in the ang2off direction.
    }
  }

  
  //Loop over each my,mx and kz, try to parallelize along the i_kmx axis.
  for (i_kmy=0;i_kmy<n_kmy; i_kmy++) {
  //First read in the angle gathers for the entire (kx,kz,gamma,azim) volume, rather than accessing them in small pieces.
  dm_in.seq_grab(&cubeIn);
  
  //the openmp should run in this level.
  //#pragma omp parallel default(shared) private(i_kmx,i_kz,i_azim,i_gamma,ith)
  {
  //ith = omp_get_thread_num(); VERIFY(ith<n_threads);
  ith = 0;
  float azim,kkz,kkz_den,sin_azim,cos_azim;
  int   i_khx_copl,i_khy_copl;
  bool in_bounds = true;
  for (i_kmx=ith;i_kmx<n_kmx; i_kmx+=n_threads) {
    windkxyz[ith][3].f = i_kmx;
    //skip the kz=0 case
    for (i_kz=1; i_kz <n_kz ; i_kz++) { //kz=0 cause us numerical trouble, thus I skip this point and assume it is always zero.
      windkxyz[ith][2].f = i_kz;

      hx_hy[ith] = cmplxf(0.f,0.f);  //initialize the output to zero
      //Read in this (gamma,azim) section
      cubeIn.grab2d(&gamma_azim[ith], windkxyz[ith]);
      
      //Then do the mapping from (gamma,azim) to (khx,khy). The problem is the inadequate sampling of (gamma,azim) plane, and the different mapping Jacobians at different (gamma locations). 
        kkz=k_z[i_kz]; kkz_den = kkz;
        //cannot have k_z equal to zero, might cause some singularity problem
        //if (i_kz+i_thread == 0) kkz_den=k_z[1];  else kkz_den=kkz;
        
        //Find all the (gamma,azim)->(khx,khy) mappings first
        //First map the coordinates.      
        mapped_khx[ith].zero(); mapped_khy[ith].zero(); map_bInbound[ith] = false;
        for (i_azim=0; i_azim<ax_azim_rad.n; i_azim++) { //here azim is the beta angle
          azim = ax_azim_rad.get_coord(i_azim);
          sin_azim=sin(azim); cos_azim=cos(azim);
          map_gamma_azim_2_khxy(&mapped_khx[ith](i_azim,0), &mapped_khy[ith](i_azim,0), ax_gamma_rad, cos_azim, sin_azim, k_mx[i_kmx], k_my[i_kmy], kkz, sign_khy);
        }
        //After mapping, turn the coordinates into pixel range, i.e. divide by d
        mapped_khx[ith].add(-khx_o); mapped_khx[ith].scale(inv_d_khx);
        mapped_khy[ith].add(-khy_o); mapped_khy[ith].scale(inv_d_khy);
        
        //Plant the values based on the mappings, NN case
        for (i_azim=0; i_azim<ax_azim_rad.n; i_azim++) { //here azim is the beta angle
          azim = ax_azim_rad.get_coord(i_azim);
          sin_azim=sin(azim); cos_azim=cos(azim);
          //
           
          for (i_gamma=0;i_gamma<ax_gamma_rad.n; i_gamma++) {
            //We now had khx' and khy', we need to rotate them back to khx and khy
            //khx_copl= mapped_khx(i_azim,i_gamma); khy_copl= mapped_khy(i_azim,i_gamma);
            //check if mapping within our desired domain, convert to indexes
            //khx_copl_int= (khx_copl-khx_o)*inv_d_khx;   i_khx_copl  = lroundf(khx_copl_int);
            //khy_copl_int= (khy_copl-khy_o)*inv_d_khy;   i_khy_copl  = lroundf(khy_copl_int); //round to the nearest neighbor          
            i_khx_copl= ceil0(mapped_khx[ith](i_azim,i_gamma)); i_khy_copl= ceil0(mapped_khy[ith](i_azim,i_gamma));
            if((i_khx_copl >= n_khxl) && (i_khx_copl <= n_khxh ) &&
               (i_khy_copl >= n_khyl) && (i_khy_copl <= n_khyh ) ) {
              in_bounds = true;
              i_khx_copl    =((number_aliases+1)*n_khx+i_khx_copl) % n_khx;
              i_khy_copl    =((number_aliases+1)*n_khy+i_khy_copl) % n_khy;
            } else {
              in_bounds = false;
            }
            map_bInbound[ith](i_azim,i_gamma) = in_bounds; //Set the boolean mask.
            
            if (in_bounds && mapping_method == gkm_NN) {
              hx_hy[ith](i_khy_copl, i_khx_copl) += gamma_azim[ith](i_azim,i_gamma); //nearest neighbor interpolation, cause in this case, bilinear interpolation does not make much sense now.
            } //end if inbounds. 
            //debug msg prt
            if((i_kmy == 0) && (i_kmx == 0) && (i_kz == 1)) {
            //if((i_kmy == n_kmy/4 -1) && (i_kmx == n_kmx/4 -1) && (i_kz == n_kz/4 -1)) {
  //            prt("*****\n");
  //            PRV4(azim*(180/pi),gamma*(180/pi),jac_samp,jac);
  //            prt("*****\n");
            }            
          } //for igamma
        } //for i_azim
        
        //print out the info
        //PRVEC(mapped_khx, n_gamma*n_azim);  PRVEC(mapped_khy, n_gamma*n_azim);
        //PRVEC(map_bInbound, n_gamma*n_azim);
        //ax_azim_rad_endpt.print_info("ax_azim_rad_endpt");
        //if we use polyfill then we need to map the corner coordinates.
        if (bCalcMappingCorners) {
          mapped_khx_corner[ith].zero(); mapped_khy_corner[ith].zero();
          for (i_azim=0; i_azim<ax_azim_rad_endpt.n; i_azim++) { //here azim is the beta angle
            azim = ax_azim_rad_endpt.get_coord(i_azim);
            sin_azim=sin(azim); cos_azim=cos(azim);
            map_gamma_azim_2_khxy(&mapped_khx_corner[ith](i_azim,0), &mapped_khy_corner[ith](i_azim,0), ax_gamma_rad_endpt, cos_azim, sin_azim, k_mx[i_kmx], k_mx[i_kmy], kkz, sign_khy);
            for (i_gamma=0;i_gamma<ax_gamma_rad_endpt.n;i_gamma++) {
              //clip the corner pts if out of bounds
              float& rkhx_copl= mapped_khx_corner[ith](i_azim,i_gamma); 
              float& rkhy_copl= mapped_khy_corner[ith](i_azim,i_gamma);
              if (rkhx_copl < khx_corner_min) rkhx_copl = khx_corner_min;
              if (rkhx_copl > khx_corner_max) rkhx_copl = khx_corner_max;
              if (rkhy_copl < khy_corner_min) rkhy_copl = khy_corner_min;
              if (rkhy_copl > khy_corner_max) rkhy_copl = khy_corner_max;
            }
          }
          //convert the values now to pixel coordinates.
          mapped_khx_corner[ith].add(-khx_o); mapped_khx_corner[ith].scale(inv_d_khx);
          mapped_khy_corner[ith].add(-khy_o); mapped_khy_corner[ith].scale(inv_d_khy);
          if (ax_azim_rad_endpt.n==2) { //single azimuth, should assume no khy resolution at all, should not allow khy mapping distortion
            myMemset(&mapped_khy_corner[ith](0,0),-0.25f, ax_gamma_rad_endpt.n);
            myMemset(&mapped_khy_corner[ith](1,0), 0.25f, ax_gamma_rad_endpt.n);
          }
        }
        //calculate the areas for each mapping pts
        mapped_areas[ith].zero(); float f_khx_khy_0 = 0.0f;
        //PRVEC(mapped_khx_corner, (n_gamma+1)*(n_azim+1));  PRVEC(mapped_khy_corner, (n_gamma+1)*(n_azim+1));
        if (mapping_method == gkm_ap_NN) {
          const float* khx_corner_top; const float* khx_corner_bot; const float* khy_corner_top; const float* khy_corner_bot;
          for (i_azim=0; i_azim<ax_azim_rad.n; i_azim++) {
            khx_corner_top = &mapped_khx_corner[ith](i_azim  ,0); khy_corner_top = &mapped_khy_corner[ith](i_azim   ,0);
            khx_corner_bot = &mapped_khx_corner[ith](i_azim+1,0); khy_corner_bot = &mapped_khy_corner[ith](i_azim+1 ,0);
            for (int i=0; i<ax_gamma_rad.n; i++) {
              if (map_bInbound[ith](i_azim,i)) {
                if (i != i_gamma0loc) { //gamma = 0 case is processed separately
                  mapped_areas[ith](i_azim,i) = area_quadrilateral(khx_corner_top[i],khx_corner_bot[i],khx_corner_bot[i+1],khx_corner_top[i+1],
                                                                khy_corner_top[i],khy_corner_bot[i],khy_corner_bot[i+1],khy_corner_top[i+1]);
                //assert(mapped_areas(i_azim,i)>100*epsilon); //the area should not be zero, but it can be very close to zero, when kz,kmx,kmy are very small.
//              if(mapped_areas(i_azim,i)<100*epsilon) {
//                PRV2(tan(ax_gamma_rad_endpt.get_coord(i_azim)),tan(ax_gamma_rad_endpt.get_coord(i_azim+1)));
//
//                errPrt("map area is close to zero,%f,%f,%f,%f ; %f,%f,%f,%f\n", khx_corner_top[i],khx_corner_bot[i],khx_corner_bot[i+1],khx_corner_top[i+1],
//                                                              khy_corner_top[i],khy_corner_bot[i],khy_corner_bot[i+1],khy_corner_top[i+1]);
//              }
                } else { //gamma = 0; calc the area of the two triangle fans. the third point is the center (khx=khy=0);
                  mapped_areas[ith](i_azim,i) = area_triangle(khx_corner_top[i],khx_corner_bot[i],f_khx_khy_0,     khy_corner_top[i],khy_corner_bot[i],f_khx_khy_0) +
                                           area_triangle(khx_corner_bot[i+1],khx_corner_top[i+1],f_khx_khy_0, khy_corner_bot[i+1],khy_corner_top[i+1],f_khx_khy_0);
                }
                //Then set the value at the coordinates.
                i_khx_copl = lroundf(mapped_khx[ith](i_azim,i)); i_khx_copl=lroundf((i_khx_copl+n_khx))%n_khx;
                i_khy_copl = lroundf(mapped_khy[ith](i_azim,i)); i_khy_copl=lroundf((i_khy_copl+n_khy))%n_khy; 
                hx_hy[ith](i_khy_copl,i_khx_copl) += gamma_azim[ith](i_azim,i) * mapped_areas[ith](i_azim,i);
              }
            }
          }
        } //end if mapping== gkm_ap_NN
        
        int ret = 0;
        if (mapping_method == gkm_polyfill) {
          float polyX[5],polyY[5];
          int ileft,iright,itop,ibot;
          const float* khx_corner_top; const float* khx_corner_bot; const float* khy_corner_top; const float* khy_corner_bot;
          for (i_azim=0; i_azim<ax_azim_rad.n; i_azim++) {
            khx_corner_top = &mapped_khx_corner[ith](i_azim  ,0); khy_corner_top = &mapped_khy_corner[ith](i_azim   ,0);
            khx_corner_bot = &mapped_khx_corner[ith](i_azim+1,0); khy_corner_bot = &mapped_khy_corner[ith](i_azim+1 ,0);
            for (int i=0; i<ax_gamma_rad.n; i++) {
              if (map_bInbound[ith](i_azim,i)) {
                if (i != i_gamma0loc) { //at gamma=0 case, should be processed separately
                  //fill in the polygons defined by the corner points.
                  polyX[0] = khx_corner_top[i];   polyX[1] = khx_corner_top[i+1]; 
                  polyX[2] = khx_corner_bot[i+1]; polyX[3] = khx_corner_bot[i]; 
                  polyY[0] = khy_corner_top[i];   polyY[1] = khy_corner_top[i+1]; 
                  polyY[2] = khy_corner_bot[i+1]; polyY[3] = khy_corner_bot[i]; 
                  //shift the [-N/2,N/2) coord to [0,N) coord, because the c++ arrays khxhy are 0 based. Think a easy example, simple 8 pt case from (-4,-3,-2,-1,0,1,2,3) to (0,1,2,3,4,5,6,7). But indeed, khxky still place the zero frequency in the center, that is why it needs to multiply bFlipCub in order to shift the zero frequency back to position 0.
                  for (int ii=0;ii<4; ii++) { polyX[ii]+= float(hf_nkhx); polyY[ii]+= float(hf_nkhy);}
                  //also find the bounding box of the quadlateral.
                  ileft = int(floor (*std::min_element(polyX,polyX+4)));  ileft = max(ileft,0);
                  iright= int(ceil  (*std::max_element(polyX,polyX+4)));  iright= min(iright,n_khx);
                  itop  = int(floor (*std::min_element(polyY,polyY+4)));  itop  = max(itop,0);
                  ibot  = int(ceil  (*std::max_element(polyY,polyY+4)));  ibot  = min(ibot,n_khy);
                  ret = fill_quadralaterl_c(polyX,polyY, itop,ibot,ileft,iright,
                                      gamma_azim[ith](i_azim,i),hx_hy[ith]);
                } else {
                  //fill in the two triangles defined by the corner points.
                  polyX[0] = khx_corner_top[i];   polyX[1] = khx_corner_bot[i]; 
                  polyY[0] = khy_corner_top[i];   polyY[1] = khy_corner_bot[i]; 
                  polyX[2] = 0.0f; 
                  polyY[2]=0.0f;
                  polyX[3] = khx_corner_top[i+1]; polyX[4] = khx_corner_bot[i+1]; 
                  polyY[3] = khy_corner_top[i+1]; polyY[4] = khy_corner_bot[i+1]; 
                  //shift to 0 based, think a easy example, simple 8 pt k-domain case from (-4,-3,-2,-1,0,1,2,3) to (0,1,2,3,4,5,6,7)
                  for (int ii=0;ii<5; ii++) { polyX[ii]+= float(hf_nkhx); polyY[ii]+= float(hf_nkhy);}
                  //also find the bounding box of the quadlateral.
                  ileft = int(floor (*std::min_element(polyX,polyX+5)));  ileft = max(ileft,0);
                  iright= int(ceil  (*std::max_element(polyX,polyX+5)));  iright= min(iright,n_khx);
                  itop  = int(floor (*std::min_element(polyY,polyY+5)));  itop  = max(itop,0);
                  ibot  = int(ceil  (*std::max_element(polyY,polyY+5)));  ibot  = min(ibot,n_khy);
                  
                  //fill the first triangle
                  ret = fill_quadralaterl_c(polyX,polyY, itop,ibot,ileft,iright,
                                      gamma_azim[ith](i_azim,i),hx_hy[ith],3);
                  //fill the second triangle
                  ret = fill_quadralaterl_c(polyX+2,polyY+2, itop,ibot,ileft,iright,
                                      gamma_azim[ith](i_azim,i),hx_hy[ith],3);
                }
                if (ret == -1) { //error reporting.
                  PRV2(i_azim,i);
                  PRV2(ax_azim_rad_endpt.get_coord(i_azim),ax_azim_rad_endpt.get_coord(i_azim+1));
                  PRV2(ax_gamma_rad_endpt.get_coord(i),ax_gamma_rad_endpt.get_coord(i+1));
                  PRV4(k_mx[i_kmx], k_my[i_kmy], kkz, 1);
                  errPrt("error!!! for the mapping \n");
                }
              }
            } // end for igamma
          } //end for iazim
        } //end if mapping_method = fillpolygon
        
        cubeData<int>khxy_fold(hypercube(n_khx,n_khy)); khxy_fold = 0;
        if (mapping_method == gkm_triTexMapping) {          
          float polyX[5],polyY[5]; //only deal with triangles
          int ileft,iright,itop,ibot; float fleft,fright;
          if (bTrue2Dmapping) {
            VERIFY(0); //not finished yet!!!!
            const float* khx_corner_top; const float* khx_corner_bot; const float* khy_corner_top; const float* khy_corner_bot;
            for (i_azim=0; i_azim<ax_azim_rad.n; i_azim++) {
              khx_corner_top = &mapped_khx_corner[ith](i_azim  ,0); khy_corner_top = &mapped_khy_corner[ith](i_azim   ,0);
              khx_corner_bot = &mapped_khx_corner[ith](i_azim+1,0); khy_corner_bot = &mapped_khy_corner[ith](i_azim+1 ,0);
              for (int i=0; i<ax_gamma_rad.n; i++) {
                if (map_bInbound[ith](i_azim,i)) {
                  //fill in the two triangles defined by the corner points.
                  polyX[0] = khx_corner_top[i];   polyX[1] = khx_corner_bot[i]; 
                  polyY[0] = khy_corner_top[i];   polyY[1] = khy_corner_bot[i]; 
                  polyX[2] = 0.0f;  polyY[2] = 0.0f;
                  polyX[3] = khx_corner_top[i+1]; polyX[4] = khx_corner_bot[i+1]; 
                  polyY[3] = khy_corner_top[i+1]; polyY[4] = khy_corner_bot[i+1]; 
                  //shift to 0 based, think a easy example, simple 8 pt k-domain case from (-4,-3,-2,-1,0,1,2,3) to (0,1,2,3,4,5,6,7)
                  for (int ii=0;ii<5; ii++) { polyX[ii]+= float(hf_nkhx); polyY[ii]+= float(hf_nkhy);}
                  //also find the bounding box of the quadlateral.
                  ileft = int(floor (*std::min_element(polyX,polyX+3)));  ileft = max(ileft,0);
                  iright= int(ceil  (*std::max_element(polyX,polyX+3)));  iright= min(iright,n_khx);
                  itop  = int(floor (*std::min_element(polyY,polyY+3)));  itop  = max(itop,0);
                  ibot  = int(ceil  (*std::max_element(polyY,polyY+3)));  ibot  = min(ibot,n_khy);
                  
                  //fill the first triangle
                  ret = fill_quadralaterl_c(polyX,polyY, itop,ibot,ileft,iright,
                                      gamma_azim[ith](i_azim,i),hx_hy[ith],3);
                  //fill the second triangle
                  ret = fill_quadralaterl_c(polyX+2,polyY+2, itop,ibot,ileft,iright,
                                      gamma_azim[ith](i_azim,i),hx_hy[ith],3);
                  if (ret == -1) { //error reporting.
                    PRV2(i_azim,i);
                    PRV2(ax_azim_rad_endpt.get_coord(i_azim),ax_azim_rad_endpt.get_coord(i_azim+1));
                    PRV2(ax_gamma_rad_endpt.get_coord(i),ax_gamma_rad_endpt.get_coord(i+1));
                    PRV4(k_mx[i_kmx], k_my[i_kmy], kkz, 1);
                    errPrt("error!!! for the mapping \n");
                  }
                }
              } // end for igamma
            } //end for iazim
          } else { //else if only 1D mappings, just do 1D interpolation of the values.
            const float* khxs; float wt,gap;
            i_azim = 0; 
            //for (i_azim=0; i_azim<ax_azim_rad.n; i_azim++) {
              khxs = &mapped_khx[ith](i_azim,0);
              for (int i=0; i<n_gamma-1; i++) { //there are only n_gamma-1 intervals
                if (map_bInbound[ith](i_azim,i) && map_bInbound[ith](i_azim,i+1)) {
                  //find the texture coordinates for each khx in the interval.
                  if (khxs[i+1]>khxs[i]) {
                    fleft = khxs[i]; fright = khxs[i+1];
                  } else {
                    fleft = khxs[i+1]; fright = khxs[i];
                  }
                  gap = fright - fleft;
                  ileft = iceil(fleft); iright = ifloor(fright);
                  ileft += hf_nkhx; iright += hf_nkhx; //since c++ array hx_hy are 0 based, we need to place the [-N/2,N/2) domain over [0,N] index.
                  
                  for (int ii=ileft; ii<=iright; ii++) {
                    wt = fabs((float(ii-hf_nkhx)-khxs[i]))/gap;
                    //interpolate the values.
                    hx_hy[ith](0,ii) = wt*gamma_azim[ith](0,i+1)  + (1-wt)*gamma_azim[ith](0,i);
                    khxy_fold(0,ii) += 1;
                    if(khxy_fold(0,ii)!=1 && ii != hf_nkhx) {
                      LOGC('error, fold is bigger than ONE for a non-zero khx!!!');
                      PRV2(i,ii); PRV4(fleft,fright,ileft,iright); PRV2(hf_nkhx,wt);
                      //errPrt("");
                    }
                  }
                }
              }
          //}
          } //end if bTrue2Dmapping
        } //end if mapping_method = triangle_texture_mappings
                
        //not doing smoothing
        if (bDoFTsmoothing) {
          //break hx_hy[i_thread] into real and imaginary part.
          CtoR(hx_hy[ith],cubReal[ith],cubImg[ith]);
          //hx_hy[i_thread].domain().info("hx_hy cubeinfo");
          //Do some smoothing before we commit to fft. //the smoothing should not be too much, otherwise it kills the signals on far hx,hy positions.
          int rect1,rect2,rect3,nrepeat;
          if (mapping_method == gkm_polyfill) {
            rect1=2, rect2=2, rect3=1, nrepeat=1;
          } else if (mapping_method == gkm_NN || mapping_method==gkm_ap_NN) { //need stronger smoothing.
            rect1=2, rect2=2, rect3=1, nrepeat=2;
          }
          if (n_khy==1) { //single khy then no need to smooth on khy axis.
            rect2 = 1;
          }
          
          rectSmooth(cubReal[ith], cubRealsm[ith], rect1, rect2, rect3, nrepeat);
          rectSmooth(cubImg[ith],  cubImgsm[ith],  rect1, rect2, rect3, nrepeat);
          RtoC(hx_hy[ith],cubRealsm[ith],cubImgsm[ith]);
        } else { // do nothing.
        }
        
        hx_hy[ith].multiply(dampingKhxy);//first window the result with a hamming window, to avoid ringing in the khxy domain.
        hx_hy[ith].multiply(phsh_hxy); //shift the phase of khx,khy so that the first point on the transformed hx axis is o_hx rather than 0.
  
        //Finally FFT back from (khx,khy) to (hx,hy)
        memcpy(fftwin2d[ith],hx_hy[ith].pval,n_khxy*sizeof(fftwf_complex)); //fill in the data
        fftwf_execute(FFTplan[ith]);
        memcpy(hx_hy[ith].pval,fftwout2d[ith],n_khxy*sizeof(fftwf_complex)); //retrieve the data
        
        //we shift the domain to center 0 at the center for easy polygon filling process, now we need to shift back. (-4,-3,-2,-1,0,1,2,3) --> (0,1,2,3,4,-3,-2,-1), like fft shift.
        //In the transformed domain, it means multiple -1 every other point.
        if (mapping_method == gkm_polyfill || mapping_method==gkm_triTexMapping) {
          multiply(hx_hy[ith],hx_hy[ith],bFlipCub);
        }
      
      //gamma_azim[0].domain().info("gamma_azim cube");
      //LOGC("before seq_update"); PRV(n_threads);
      //PRVEC(gamma_azim[0],gamma_azim[0].size());
      //write the output to (gamma,azim,kz,kmx,kmy)
      cubeOut.update2d(&(hx_hy[ith]),windkxyz[ith]);
      //LOGC("after seq_update");
      //mapAreaDump.dumpFrame(mapped_areas);
    } //for i_kz
  } //for i_kmx
  } //omp
  //update each iy slice
  dm_out.seq_update(&cubeOut);
  } //for i_kmy
  
  //LOGC(Finish off3ang3d main loop);
  //cleaned up fftw
  for (int i=0; i<n_threads; i++) {
    fftwf_free(fftwout2d[i]);  fftwf_free(fftwin2d[i]);
    fftwf_destroy_plan(FFTplan[i]);
  }
  LOGC("finish ang2off3d function");
  return;
}

