#include "stdafx.h"

#include "cubeData.hpp"
#include "op.h"

using namespace Yang;
using namespace std;
//need set zero if bAdd == false

//General implementation for this oper interface
void op::oper(bool bAdj, bool bAdd, int nmod, int ndat, float* mod, float* dat) {
  VERIFY(ndat == DatShape().get_n123());
  VERIFY(nmod == ModShape().get_n123());
  
  cubeDataF modc,datc;
  modc.becomeProxyCube(this->ModShape()); datc.becomeProxyCube(this->DatShape());
  modc.proxyPtrIs(mod);  datc.proxyPtrIs(dat);
  this->oper(modc,datc,bAdj,bAdd);
}

////not finished yet
bool op::dot_test(bool verbose) {
   cubeData<float> mod1,mod2,dat1,dat2;
   
   VERIFY(!modShape.isEmpty());
   VERIFY(!datShape.isEmpty());

   mod1.alloc(modShape);
   dat2.alloc(datShape);
   
   mod1.allocMemoryTo(mod2);
   dat2.allocMemoryTo(dat1);
   
   //Generate random numbers for mod
   long long nMod = modShape.get_n123();
   long long nDat = datShape.get_n123();
   
   //generate from 0~1 uniform random numbers
   srand(time(NULL));
   long long i;
   for (i=0; i<nMod; i++) {
     mod1[i] = float(rand())/float(RAND_MAX);
   }
   for (i=0; i<nDat; i++) {
     dat2[i] = float(rand())/float(RAND_MAX);
   }
   
   verbPrt("\nINITIAL DATA in DotTest\n");
   
   //do forward
   bool bAdj =  false;
   this->oper(mod1,dat1,bAdj,false);
   float dot1=float(dotProd(dat1,dat2));
   //do adj
   this->oper(mod2,dat2,!bAdj,false);
   double dot2=dotProd(mod2,mod1);

   verbPrt("Dot(add=false) %g %g \n",dot1,dot2);
    
   float dot3=1.0f;    
   /*
   forward(true,mod1,dat1);
   adjoint(true,mod2,dat2);

   double dot3=mod1->dot(mod2);
   double dot4=dat1->dot(dat2);
   verbPrt("Dot(add=true) %g %g \n",dot4,dot3);

   if(fabs( (dot3-dot4)/dot3) >.0001) {
     verbPrt("Failed add=true dot product  %g (%g %g) \n",fabs( (dot3-dot4)/dot3),dot3,dot4);
     return false;
   }
   */
   
   if(fabs(dot1) < 1e-12 || fabs(dot3) < 1e-12) {
     verbPrt("Dot product suspiciously small\n");
     return false;
   }
   
   if(fabs( (dot1-dot2)/dot1) >.0001) {
     verbPrt("Failed add=false dot product %g %g\n",dot1,dot2);
     return false;
  }
  
   return true;
}


//Test linearity of the operator, verify 1.0*Lm_1+1.0*Lm_2 == L(m_1+Lm_2), m_1 and m_2 will have same norm.
bool op::linearity_test(bool bAdj, bool verbose) {
  bool bVerbose_init = gVerb;
  gVerb = verbose;
  
  cubeData<float> mod1,mod2,mod12, dat1,dat2, dat12;
  
  VERIFY(!modShape.isEmpty());
  VERIFY(!datShape.isEmpty());
  
  //Generate random numbers for mod
  long long nMod = modShape.get_n123();
  long long nDat = datShape.get_n123();
  
  bool ret = true;
  
  mod1.alloc(modShape); mod2.alloc(modShape); mod12.alloc(modShape);
  dat1.alloc(datShape); dat2.alloc(datShape); dat12.alloc(datShape);
  float rms,drms;
  
  if (!bAdj) { //testing fwd operator
    //generate from 0~1 uniform random numbers
    srand(time(NULL));
    long long i;
    for (i=0; i<nMod; i++) {
      mod1[i] = float(rand())/float(RAND_MAX);
      mod2[i] = float(rand())/float(RAND_MAX);
    }
    
    verbPrt("Initializing models in linearity_test_fwd\n");
    
    //do forward
    this->oper(false,false,nMod,nDat,mod1.pval,dat1.pval);
    this->oper(false,false,nMod,nDat,mod2.pval,dat2.pval);
    
    mod12.add(mod1,mod2);
    this->oper(false,false,nMod,nDat,mod12.pval,dat12.pval);
    
    verbPrt("After forward opeartor has been finished\n");
    
    //comparing the difference between dat1+dat2 and dat12
    dat1.add(dat2);
    rms = dat1.RMS();
    
    dat1.subtract(dat12);
    drms = dat1.RMS();
    
    prt("drms/rms:%e,%e \n",drms,rms);
    VERIFY(rms > 1e-6);
    float ratio = drms/rms;
    if (ratio>1e-3) {
      verbPrt("Linearity test FAILED, drms/rms=%e\n", ratio);
      ret = false;
    } else { //the error is small
      verbPrt("Linearity test SUCCESS, drms/rms=%e\n", ratio);
    }
  } else { //Test adjoint operator
    //generate from 0~1 uniform random numbers
    srand(time(NULL));
    long long i;

    for (i=0; i<nDat; i++) { dat1[i] = float(rand())/float(RAND_MAX); dat2[i] = float(rand())/float(RAND_MAX);}

    verbPrt("Initializing dats in linearity_test_adj\n");

    //do adjoints
    this->oper(true,false,nMod,nDat,mod1.pval,dat1.pval);
    this->oper(true,false,nMod,nDat,mod2.pval,dat2.pval);

    dat12.add(dat1,dat2);
    this->oper(true,false,nMod,nDat,mod12.pval,dat12.pval);

    verbPrt("After adjoint opeartor has been finished\n");

    //comparing the difference between mod1+mod2 and mod12
    mod1.add(dat2);
    rms = mod1.RMS();

    mod1.subtract(mod12);
    drms = mod1.RMS();

    float ratio = drms/rms;
    prt("drms/rms:%e,%e \n",drms,rms);
    if (ratio>1e-3) {
      verbPrt("Linearity test adjoint FAILED, drms/rms=%e\n", ratio);
      ret = false;
    } else { 
      verbPrt("Linearity test adjoint SUCCESS, drms/rms=%e\n", ratio);
    }
  }
  
  gVerb = bVerbose_init; //restore the Verb status  
  return ret;
}

//bool op::dot_test(bool verbose)
//{
//   gVec *mod1,*mod2,*dat1,*dat2;
//   
//   mod1=domain->clone_vec();
//   mod2=mod1->clone_vec();
//   dat1=range->clone_vec();
//   dat2=dat1->clone_vec();
//   mod1->random();
//   dat2->random();
//   
//   
//
//
//  // fprintf(stderr,"INITIAL DATA\n");
////   mod1->info("Input mod",1);
////   dat2->info("Input dat",2);
//
//   forward(false,mod1,dat1);
//   double dot1=dat1->dot(dat2);
//   adjoint(false,mod2,dat2);
//   double dot2=mod2->dot(mod1);
//
//   verbPrt("Dot(add=false) %g %g \n",dot1,dot2);
//
//
//
//    
//   forward(true,mod1,dat1);
//   adjoint(true,mod2,dat2);
//
//
////  mod2->info("OUT mod",1);
// // dat1->info("OUR dat",2);
//   
//
//   double dot3=mod1->dot(mod2);
//   double dot4=dat1->dot(dat2);
//   verbPrt("Dot(add=true) %g %g \n",dot4,dot3);
//   
//   delete dat1; delete dat2; delete mod1; delete mod2;
//   
//   if(fabs(dot1) < 1e-12 || fabs(dot3) < 1e-12) {
//     verbPrt("Dot product suspiciously small\n");
//     
//     return false;
//   }
//   
//   if(fabs( (dot1-dot2)/dot1) >.0001) {
//     verbPrt("Failed add=false dot product %g %g\n",dot1,dot2);
//     return false;
//  }
//  
//  if(fabs( (dot3-dot4)/dot3) >.0001) {
//     verbPrt("Failed add=true dot product  %g (%g %g) \n",fabs( (dot3-dot4)/dot3),dot3,dot4);
//     return false;
//  }
//   return true;


