
#include "stdafx.h"
#include "commonOps.h"
#include "opSpline.h"
#include "cubeAlgo.h"

using namespace Yang;

// From Xukai Shen, Reference See Biondo's Thesis
// several spline interpolation function which does either d=Am (forward) or m=A'd (adjoint)
// different functions have different input/output precision(float vs double), body of code is the same except for definition of variable precision
// it does 1D spline interpolation each time it's called, but can operate on different axis, 
// to perform 2D interpolation, you need to call the function twice with different idim,  
// 2D interpolation has been extensively tested and works fine (both forward and adjoint), n3=1, idim<2 in this case, operate on ONE dimension every time.
// 3D is not tested, will be soon  

// mod:   model, can be up to 3d, always 1d representation, the coefficent of the B spline bases. these are like control points.
//        And generally does not sits on data surface.
// dat:       data, can be up to 3d, always 1d representation 
// n1,n2,n3:  model dimension
// idim:      (0~2) is the dimension where spline is operated on, 0 being the fastest axis
// dmod:      spacing on 'idim' axis
// ndat,ddat: number of samples and spacing for data on 'idim' axis 

//Smoothing is achieved by adj and Fwd pair to the input data.

int Bspline1dFwd(float *mod, float *dat, int n1, int n2, int n3, int idim,
                 float dmod, int ndat, float ddat, bool bSkipThisAxis=false);
int Bspline1dAdj(float *mod, float *dat, int n1, int n2, int n3, int idim, 
                 float dmod, int ndat, float ddat, bool bSkipThisAxis=false);

//adjoint with float type. Given dat points, return the model points.
int Bspline1dAdj(float *mod, float *dat, int n1, int n2, int n3, int idim, float dmod, int ndat, float ddat, bool bSkipThisAxis)
{
   int i1,i2,i3;
   int imod,idat;
   float x,alpha, beta;
   float coef1,coef2,coef3,coef4;
   //int idat1,idat2,idat3,idat4;
   int idx,idx1,idx2,idx3,idx4;
   int imod1,imod2,imod3,imod4;
   int n1dat,n2dat,n3dat;
   int flag1,flag2,flag3,flag4;
   
   idx = idx1 = idx2 = idx3 = idx4 = 0;
   imod = idat = 0;

   //int fat=2;
   n1dat=n1;
   n2dat=n2;
   n3dat=n3;
   
   //Make sure the dmod and data don't go too off.
   float dmod_t = 0.f;
   if (idim==0) dmod_t = (ndat)*ddat / float(n1);
   else if (idim==1) dmod_t = (ndat)*ddat / float(n2);
   else if (idim==2) dmod_t = (ndat)*ddat / float(n3);
   else {VERIFY(0);}
   //PRV(dmod_t); PRV(dmod); PRV(idim);
   VERIFY( dmod_t < 1.2*dmod && dmod_t > 0.8*dmod);
   
   //Simply copy the memory
   if (bSkipThisAxis){
     //cerr<<"\n n123"<<n1<<','<<n2<<sep<<n3<<sep<<ndat<<endl;
     VERIFY(n1==ndat || n2==ndat ||n3==ndat);
     //Copy from dat to mod
     myMemcpy(mod,dat,n1*n2*n3);
     return 0;
   }
   
   //cerr << "adjparm:" << n1 << n2 << n3 << sep <<  idim << sep << dmod << sep << ndat << sep << ddat << sep << bSkipThisAxis << endl;
   if (idim==0)
      n1dat=ndat;
   else if(idim==1)
      n2dat=ndat;
   else if (idim==2)
      n3dat=ndat;
   
   for (i3=0;i3<n3dat;i3++)
      for ( i2=0;i2<n2dat;i2++)
         for (i1=0;i1<n1dat;i1++)
         {
            flag1=0; 
            flag2=0;
            flag3=0;
            flag4=0;
            if (idim==0)
               idat=i1;
            else if(idim==1)
               idat=i2;
            else if (idim==2)
               idat=i3;
 
            x=idat*ddat;
            imod=int(x/dmod);
            alpha=(x-imod*dmod)/dmod;
            beta=1.-alpha;
            coef1=CB(2.-(1.+alpha))/6.0f;
            coef2=(4.-6.*SQ(alpha)+3.f*CB(alpha))/6.;
            coef3=(4.-6.*SQ(beta )+3.f*CB(beta ))/6.;
            coef4=CB(2.-(2.-alpha))/6.;
    
            if (idim==0)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n1) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n1) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n1) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n1) )
                  flag4=1;

               idx1=imod1+i2*n1+i3*n1*n2;
               idx2=imod2+i2*n1+i3*n1*n2;
               idx3=imod3+i2*n1+i3*n1*n2;
               idx4=imod4+i2*n1+i3*n1*n2;
            }
            else if(idim==1)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n2) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n2) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n2) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n2) )
                  flag4=1;

               idx1=i1+imod1*n1+i3*n1*n2;
               idx2=i1+imod2*n1+i3*n1*n2;
               idx3=i1+imod3*n1+i3*n1*n2;
               idx4=i1+imod4*n1+i3*n1*n2;
            }
            else if (idim==2)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n3) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n3) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n3) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n3) )
                  flag4=1;

               idx1=i1+i2*n1+imod1*n1*n2;
               idx2=i1+i2*n1+imod2*n1*n2;
               idx3=i1+i2*n1+imod3*n1*n2;
               idx4=i1+i2*n1+imod4*n1*n2;
            }
            idx=i1+i2*n1dat+i3*n1dat*n2dat;
            if (flag1==1)
               mod[idx1]+=coef1*dat[idx];
            if (flag2==1)
               mod[idx2]+=coef2*dat[idx];
            if (flag3==1) 
               mod[idx3]+=coef3*dat[idx];
            if (flag4==1) 
               mod[idx4]+=coef4*dat[idx]; 
         }  
   return 0;
}

/// forward spline interp, from mod -> dat
int Bspline1dFwd(float *mod, float *dat, int n1, int n2, int n3, int idim,
                 float dmod, int ndat, float ddat, bool bSkipThisAxis)
{
   int i1,i2,i3;
   int imod,idat;
   float x,alpha, beta;
   float coef1,coef2,coef3,coef4;
   //int idat1,idat2,idat3,idat4;
   int imod1,imod2,imod3,imod4;
   int idx,idx1,idx2,idx3,idx4;
   int n1dat,n2dat,n3dat;
   
   idx = idx1 = idx2 = idx3 = idx4 = 0;
   imod = idat = 0;

   //Simply copy the memory
   if (bSkipThisAxis){
     VERIFY(n1==ndat || n2==ndat ||n3==ndat);
     //Copy from mod to dat
     myMemcpy(dat,mod,n1*n2*n3);
     return 0;
   }

   n1dat=n1;
   n2dat=n2;
   n3dat=n3;
   if (idim==0)
      n1dat=ndat;
   else if(idim==1)
      n2dat=ndat;
   else if (idim==2)
      n3dat=ndat;
   
   for (i3=0;i3<n3dat;i3++)
      for ( i2=0;i2<n2dat;i2++)
         for (i1=0;i1<n1dat;i1++)
         {
            if (idim==0)
               idat=i1;
            else if(idim==1)
               idat=i2;
            else if (idim==2)
               idat=i3;
 
            x=idat*ddat;
            imod=int(x/dmod);
            alpha=(x-imod*dmod)/dmod;
            beta=1.-alpha;
            coef1=pow( (2.-(1.+alpha)),3)/6.;
            coef2=(4.-6.*(pow(alpha,2))+3.*(pow(alpha,3)))/6.;
            coef3=(4.-6.*(pow(beta,2))+3.*(pow(beta,3)))/6.;
            coef4=pow((2.-(2.-alpha)),3)/6.;

            if (idim==0)
            {
               imod1=MIN(MAX(imod-1,0),n1-1);
               imod2=MIN(MAX(imod  ,0),n1-1);
               imod3=MIN(MAX(imod+1,0),n1-1);
               imod4=MIN(MAX(imod+2,0),n1-1);
              
               idx1=imod1+i2*n1+i3*n1*n2;
               idx2=imod2+i2*n1+i3*n1*n2;
               idx3=imod3+i2*n1+i3*n1*n2;
               idx4=imod4+i2*n1+i3*n1*n2;
            }
            else if(idim==1)
            {
               //fprintf(stderr,"idim==1 in forward, i1=%d, imod=%d\n",i1,imod);
               imod1=MIN(MAX(imod-1,0),n2-1);
               imod2=MIN(MAX(imod  ,0),n2-1);
               imod3=MIN(MAX(imod+1,0),n2-1);
               imod4=MIN(MAX(imod+2,0),n2-1);
               
               idx1=i1+imod1*n1+i3*n1*n2;
               idx2=i1+imod2*n1+i3*n1*n2;
               idx3=i1+imod3*n1+i3*n1*n2;
               idx4=i1+imod4*n1+i3*n1*n2;
            }
            else if (idim==2)
            {
               imod1=MIN(MAX(imod-1,0),n3-1);
               imod2=MIN(MAX(imod  ,0),n3-1);
               imod3=MIN(MAX(imod+1,0),n3-1);
               imod4=MIN(MAX(imod+2,0),n3-1);
               
               idx1=i1+i2*n1+imod1*n1*n2;
               idx2=i1+i2*n1+imod2*n1*n2;
               idx3=i1+i2*n1+imod3*n1*n2;
               idx4=i1+i2*n1+imod4*n1*n2;
            }
            idx=i1+i2*n1dat+i3*n1dat*n2dat;
            dat[idx]+=coef1*mod[idx1]+coef2*mod[idx2]+coef3*mod[idx3]+coef4*mod[idx4];
         }  
   return 0;
}

//only the ns in _modShape is used.
void opSpline::init(const hypercube &_modShape, const hypercube& _datShape) {
  this->datShape = _datShape;
  //Here need to adjust modShape.d such that modShape and datShape cover similar range.
  //Now the d's
  vecAxis axes; axis ax,ax_d;
  int ndim = _datShape.get_ndim();
  mSkipAxis.resize(ndim);
  for (int idim=0; idim<ndim; idim++) {
    ax_d = _datShape.get_axis(idim+1);
    ax =   _modShape.get_axis(idim+1);
    VERIFY(ax.n!=1);
    if (ax.n==0) { //Do no smoothing
      ax.n = ax_d.n;
      ax.d = ax_d.d;
      mSkipAxis[idim] = true;
    } else {
      ax.d = ((ax_d.n-1)*ax_d.d)/(ax.n-1);
      mSkipAxis[idim] = false;
    }
    ax.o = ax_d.o;
    axes.push_back(ax);
  }
  this->modShape.set_axes(axes);
  //modShape.info("modshape IN opSpline::init");
}

void opSpline::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  //LOG(opSpline::oper);
  addNull(mod,dat,bAdj,bAdd);
  
  //dat.info("dat in opSpline");
  //PRV(bAdj); PRV(bAdd);
  const hypercube & mShape = mod.domain();
  const hypercube & dShape = dat.domain();
  VERIFY(mShape.sameSize(this->modShape));
  VERIFY(datShape.sameSize(this->datShape));

  //mShape.info("mShape in opSpline::oper");
  int ndim = mShape.get_ndim();
  assert(ndim == dShape.get_ndim());
  assert(ndim <= 3);
  
  float dmod, ddat; int ndat;
  int nm[3], nd[3];
  nm[0]=1, nm[1]=1, nm[2]=1;
  nd[0]=1, nd[1]=1, nd[2]=1;
  for (int idim=0; idim<ndim; idim++) {
    nm[idim] = mShape.get_n(idim+1); //get_n is one based
    nd[idim] = dShape.get_n(idim+1);
  }
  //PRVEC(nm,3); PRVEC(nd,3);
  //mShape.info("mshape"); dShape.info("dshape");
  
  float * pmod = mod.dataPtr();
  float * pdat = dat.dataPtr();
  
  //need intermediate buffers
  float *p1[4];
  p1[0] = p1[1] = p1[2] = p1[3] = NULL;
  
  int nnmod[3][3]; //mod size when doing each dimension.
  //@@@ Big Lesson Learned, here, If use icc compiler, CANNOT ASSUME nnmod[3][3]==nnmod[9] !!!
  for (int i=0; i<3; i++) {
    for (int j=0; j<3; j++) {
      nnmod[i][j] = 1;
    }
  }
  
  //PRVEC(nnmod[1],3);
  
  int np11,np12;
  np11 = np12 = 0;
  if (ndim==1) {
    //no need intermediate buffers
    if (!bAdj) {
      p1[0] = pmod;
      p1[1] = pdat;
    } else {
      p1[1] = pmod; p1[0] = pdat;
    }
    nnmod[0][0] = nm[0];
  } else if (ndim==2) {
    if (!bAdj) { // need an intermediate buffer that is the size nd1*nm2
      p1[0] = pmod;
      np11 = nd[0]*nm[1];
      p1[1] = (float*)malloc(np11*sizeof(float));
      p1[2] = pdat;
      nnmod[0][0] = nm[0]; nnmod[0][1] = nm[1];
      nnmod[1][0] = nd[0]; nnmod[1][1] = nm[1];
    } else { //need an intermediate buf of size nm1*nd2
      p1[0] = pdat; 
      np11=nm[0]*nd[1]; p1[1] = (float*)malloc(np11*sizeof(float));
      p1[2] = pmod;
      nnmod[0][0] = nm[0]; nnmod[0][1] = nd[1];
      nnmod[1][0] = nm[0]; nnmod[1][1] = nm[1];
      //PRVEC(nnmod[1],3);
    }
  } else if (ndim==3) {
    if (!bAdj) { //need TWO interm buffer
      p1[0] = pmod; 
      np11 = nd[0]*nm[1]*nm[2]; p1[1] = (float*)malloc(np11*sizeof(float));
      np12 = nd[0]*nd[1]*nm[2]; p1[2] = (float*)malloc(np12*sizeof(float));
      p1[3] = pdat;
      nnmod[0][0] = nm[0]; nnmod[0][1] = nm[1]; nnmod[0][2] = nm[2];
      nnmod[1][0] = nd[0]; nnmod[1][1] = nm[1]; nnmod[1][2] = nm[2];
      nnmod[2][0] = nd[0]; nnmod[2][1] = nd[1]; nnmod[2][2] = nm[2];
    } else {
      p1[0] = pdat; 
      np11 = nm[0]*nd[1]*nd[2]; p1[1] = (float*)malloc(np11*sizeof(float));
      np12 = nm[0]*nm[1]*nd[2]; p1[2] = (float*)malloc(np12*sizeof(float));
      p1[3] = pmod;
      nnmod[0][0] = nm[0]; nnmod[0][1] = nd[1]; nnmod[0][2] = nd[2];
      nnmod[1][0] = nm[0]; nnmod[1][1] = nm[1]; nnmod[1][2] = nd[2];
      nnmod[2][0] = nm[0]; nnmod[2][1] = nm[1]; nnmod[2][2] = nm[2];
    }
  }
  
  //zero the intermediate buffer
  if (np11>0) {
    memset(p1[1],0,np11*sizeof(float));
  }
  if (np12>0) {
    memset(p1[2],0,np12*sizeof(float));
  }

  //PRVEC(nnmod[0],3); PRVEC(nnmod[1],3); PRVEC(nnmod[2],3); PRVEC(mSkipAxis,ndim);
  
  //Do each axis
  for (int idim=0; idim<ndim; idim++) {
    dmod = mShape.get_d(idim+1);
    ddat = dShape.get_d(idim+1);
    ndat = dShape.get_n(idim+1);
    //PRV(dmod); PRV(ddat); PRV(ndat);
    //std::cerr<<endl<<"nnmod"<<nnmod[idim][0]<<","<<nnmod[idim][1]<<","<<nnmod[idim][2]<<endl;
    if (!bAdj) {
      Bspline1dFwd(p1[idim],p1[idim+1],nnmod[idim][0],nnmod[idim][1],nnmod[idim][2],idim,
                   dmod,ndat,ddat,mSkipAxis[idim]);
    } else {
      Bspline1dAdj(p1[idim+1],p1[idim],nnmod[idim][0],nnmod[idim][1],nnmod[idim][2],idim,
                   dmod,ndat,ddat,mSkipAxis[idim]);
      //LOGC(in Bspline1dAdj); arrayStat("p1", p1[idim], nnmod[idim][0]*nnmod[idim][1]/nnmod[idim][idim]*ndat);
      //arrayStat("p1n", p1[idim+1], nnmod[idim][0]*nnmod[idim][1]);
    }
  }

//if (bAdj) {
//  cerr << "p1[2]=" << p1[2] <<" : " << " pmod = " << pmod << endl;
//  cerr << "p1[0]=" << p1[0] <<" : " << " pdat = " << pdat << endl;
//}
  
  //free memory
  if (ndim>=2) {
    free(p1[1]);
  }
  if (ndim>=3) {
    free(p1[2]);
  }
  return;
}


////////////////////////////////////// Yang's encapsulation ////////////////////////
//Input a cube, use the preset modelShape to perform adj/fwd pair to do smoothing.
void opSpline::smooth(cubeData<float>& input, cubeData<float>& output, int repeat){
   //Read in the desired mod shape
  assert( input.domain().sameSize(datShape));
  assert(output.sameSize(input));

  int ndim = datShape.get_ndim();
  assert(ndim == modShape.get_ndim());
  
  //Do adj and fwd
  cubeData<float> modCube;
  modCube.alloc(modShape);
  
  assert(repeat>=0); PRV(repeat);
  if (repeat==0) {
    //LOG(repeat eq 0);
    output = input;
  }
  else if (repeat==1) {
    //LOG(repeat eq 1);
    this->oper(modCube, input,true,false);
    this->oper(modCube,output,false,false);
  } else { //repeat multiple times
    cubeDataF tmpCube;
    input.copyTo(tmpCube);
    for (int ir=0; ir<repeat-1; ir++) {
      this->oper(modCube,tmpCube,true,false);
      this->oper(modCube,tmpCube,false,false);
    }
    this->oper(modCube,tmpCube,true,false);
    this->oper(modCube,output,false,false);
  }

  //input.info("input");output.info("output");modCube.info("modCube");
  return;
}

int opSpline::setSmoothScale(const stdVecInt &nNodesVec, const hypercube& shape_input) {
  hypercube modShape(nNodesVec);
  hypercube datShape(shape_input);
  this->init(modShape, datShape);
  return 0;
}

///////////////////////// for seplib smoother /////////////////////
int  smoother_seplib::setSmoothScale(const stdVecInt &nNodesVec, const hypercube& shape_input) {
  this->ns_input       = shape_input.get_ns();
  VERIFY(ns_input.size() <= 3); //seplib Smooth cannot handle over 3 dimensions.
  VERIFY(ns_input.size() == nNodesVec.size());
  this->nsHfWids.clear();
  //Find the smoothing half width based on the nNodesVec
  int nHfWid = 1, n;
  for (size_t i=0; i<ns_input.size(); i++) {
    n = nNodesVec[i];
    VERIFY(n>=0 && n<= ns_input[i] && n!=1);
    if (n==0) { //means no smoothing
      nHfWid = 1;
    } else {
      nHfWid = MAX(1, lroundf(float(ns_input[i]) / float(n)));
      nHfWid = MIN(nHfWid, int(ns_input[i]+1)/2);
    }
    this->nsHfWids.push_back(nHfWid);
  }
  PRVEC(this->nsHfWids,ns_input.size());
  return 0;
}

//#include "misc.h"
#include "cubeAlgo2.h"
void smoother_seplib::smooth(cubeData<float> &in, cubeData<float> &out, int repeat) {
  stdVecInt ns = in.domain().get_ns();
  VERIFY(std::equal(ns_input.begin(),ns_input.end(),ns.begin()));
  
  int ndim = ns.size();
  int nwids[3] = {1,1,1};
  for (int i=0; i<ndim; i++) nwids[i] = nsHfWids[i];
  ////Invoke the SEPlib commands to do the smoothing
  //char sz[256];
  //sprintf(sz,"repeat=%d rect1=%d rect2=%d rect3=%d ", repeat, nwids[0], nwids[1], nwids[2]);
  //seplib::callSepTool(out, in, "Smooth", string(sz));
  //PRVEC(nwids,3);
  Yang::rectSmooth(in,out,nwids[0],nwids[1],nwids[2],repeat);
  //in.dump("dump3"); out.dump("dump4"); exit(0);
  return;
}

