#include "stdafx.h"

#include "modelingFD2d_op.h"
#include "paddingFD.h"
#include "velFD2d.h"

using namespace Yang;

//get params for the chain
void paddingFD::initOperatorEx(waveOpParams* par)
{
   if (NULL!=m_pNextPaddingObj) {
		 m_pNextPaddingObj->initOperator(par);
   }
   this->initOperator(par);
}

void paddingFD::updateEx(wavFieldGroup& wfg) {
   if (NULL!=m_pNextPaddingObj) {
		 m_pNextPaddingObj->update(wfg);
   }
   this->update(wfg);
}

void paddingFD::initOperator(waveOpParams* par)
{
   //padding::getParam(par);
   m_pars = par;
   m_npad = par->m_nPad;
   bFreeSurf = par->bFreeSurfaceOnTop;
}

void paddingFD2d::initOperator(waveOpParams* par)
{
   paddingFD::initOperator(par);
   
   //This part should goes to per shot, since each shot the domain changes
   const hypercube& shapeVelModel = m_pars->modelGridShape;
    
   nx = shapeVelModel.get_n(2);
   nz = shapeVelModel.get_n(1);
   dx = shapeVelModel.get_d(2); dz = shapeVelModel.get_d(1);
   dt = m_pars->timeParam.dtComp;
   
   dtDdx = dt/dx;
   dtDdz = dt/dz;
   
   offz = m_pars->offz;
   offx = m_pars->offx;
   
   //Get the velocity field
   vcub = &(par->pVel->get2dVelPtr());
}

//init the sponge layer mult coef
void paddingFD2dSponge1::initOperator(waveOpParams* par)
{
   
   paddingFD2d::initOperator(par);
   //Init the multiplier used in updating padded region, size is the same as 
   //velocity cube shape
   m_mult_cb.alloc((m_pars->modelGridShape));

   //Init the content of the multiplier
   int npad = m_npad;
   
   m_mult_cb.zero();
   cubeData<ETYPE>& pm = this->m_mult_cb;
   
   const cubeData<float>& v = *vcub;
   int ix,iz;   
   if (!bFreeSurf)
   {
      //calc the top sponge
      for(ix=0;ix<nx; ix++)
      {
         for(iz=0; iz<npad; iz++)
         {
            pm(ix,iz) = ETYPE(npad-iz) / ETYPE(npad)*mSpongeCoef*
                        (v(ix,iz)*dtDdz);
         }
      }
   }  //else do nothing

   //bottom sponge
   for(ix=0;ix<nx; ix++)
   {
      for(iz=nz-npad; iz<nz; iz++)
      {
         pm(ix,iz) = ETYPE(iz-(nz-npad)+1) / ETYPE(npad)*mSpongeCoef*
                     (v(ix,iz)*dtDdz);
      }
   }

   //left and right sponge
   for(ix=0;ix<npad; ix++)
   {
      for(iz=npad; iz<nz-npad; iz++)
      {
         pm(ix,iz) = ETYPE(npad-ix) / ETYPE(npad)*mSpongeCoef*
                     (v(ix,iz)*dtDdx);
      }
   }
   
   for(ix=nx-npad;ix<nx; ix++)
   {
      for(iz=npad; iz<nz-npad; iz++)
      {
         pm(ix,iz) = ETYPE(ix-(nx-npad)+1) / ETYPE(npad)*mSpongeCoef*
                     (v(ix,iz)*dtDdx);
      }
   }
   return;
}

//Using sponge absorption for the padding layer
// Adjust the wave field in the absorbing boundaries using the
// Israeli-Orszag (J. Comp Phys 1981) method to remove waves coming
// back from the boundaries.  :-( not comprehensible
// According to John Etgen's code
//update the wavefield at padded region at each time step, Apply on pp_cb[1]
void paddingFD2dSponge1::update(wavFieldGroup& wfg)
{
   //Get all necessary parameters
   int npad = m_npad;

   const cubeData<ETYPE>& pm = m_mult_cb;
   int ix,iz;

   //get the pointers to the wavefiled
   cubeData<ETYPE> *ppp2dummy, *ppp1, *ppp2;
   wfg.getPtrs(&ppp1,&ppp2,&ppp2dummy);
   cubeData<ETYPE> &pp2dummy=*ppp2dummy, &pp1=*ppp1, &pp2=*ppp2;   
   const cubeData<float>& v=*vcub;


   //the velocity field is v;
 
   //Update the 4 sides starting from the edge of the wavefield, ignore the corners
   //points from [0,offx] & [0,offz] will be
   //overwritten by boundary updates.
   
   int ivx,ivz;
   //First the left and right
   for(ix=offx; ix<npad+offx; ix++)
   {
	  ivx = ix - offx;
      for(iz=0+offz+npad; iz<nz+offz-npad; iz++)
      {
		 ivz = iz-offz;
         pp2(ix,iz) = pp2(ix,iz) - pm(ivx,ivz)*
               (pp2(ix,iz)-pp1(ix,iz)+v(ivx,ivz)*dtDdx*(pp1(ix-1,iz)-pp1(ix,iz)));
      }
   }
   
   for(ix=nx-npad+offx; ix<nx+offx; ix++)
   {
     ivx = ix - offx;
     for(iz=0+offz+npad; iz<nz+offz-npad; iz++)
     {
       ivz = iz - offz;
       pp2(ix,iz) = pp2(ix,iz) - pm(ivx,ivz)*
               (pp2(ix,iz)-pp1(ix,iz)+v(ivx,ivz)*dtDdx*(pp1(ix+1,iz)-pp1(ix,iz)));
      }
   }
   
   //Bottom
   for(ix=0+offx; ix<nx+offx; ix++)
   {
     ivx = ix - offx;
     for(iz=nz-npad+offz; iz<nz+offz; iz++)
     {
       ivz = iz - offz;
       pp2(ix,iz) += ( -pm(ivx,ivz)*
               (pp2(ix,iz)-pp1(ix,iz)+v(ivx,ivz)*dtDdz*(pp1(ix,iz+1)-pp1(ix,iz))) );
         
      }
   }
   if (!bFreeSurf)  //Add the top region
   {
      for(ix=0+offx; ix<nx+offx; ix++)
      {
        ivx = ix - offx;
        for(iz=0+offz; iz<npad+offz; iz++)
        {
          ivz = iz - offz;
          pp2(ix,iz) += (-pm(ivx,ivz)*
                  (pp2(ix,iz)-pp1(ix,iz)+v(ivx,ivz)*dtDdz*(pp1(ix,iz-1)-pp1(ix,iz))) );
            
         }
      } 
   }
}
