#ifndef PADDING_FD_H
#define PADDING_FD_H 1

#include "modelingFD2d_op.h"
#include "sep_params.h"

namespace Yang {

   //calculate the wavefield in padding layer and boundary.
   class padding
    {
   public: //methods
	  virtual ~padding() {
	  }
   };

   //use decorator pattern here, therefore we can apply different padding layers in a chain.
   class paddingFD :  public padding
    {
      public: //members
      waveOpParams* m_pars;
      
      int m_npad;      
      float dt; //computation time step
      bool  bFreeSurf;

      paddingFD* m_pNextPaddingObj; //NULL if there is no next
	  
   public: //methods	//pNextPaddingObj is NULL if not intend to concatenate
      paddingFD(paddingFD* pNextPaddingObj): 
         m_pNextPaddingObj(pNextPaddingObj) {}
	  
	  virtual ~paddingFD() {
	  }
	 //External functions that outside should call
     void initOperatorEx(waveOpParams *par);
     void updateEx(wavFieldGroup& wfg);
   protected:
     //Internal funcs, should not call from outside
     virtual void update(wavFieldGroup& wfg) = 0;
	 virtual void initOperator(waveOpParams *par);

   };

   class paddingFD2d :  public paddingFD
    {
      public: //members
         int nx,nz;
         ETYPE dz,dx;
         ETYPE dtDdz,dtDdx;
         int offx,offz; //The offset between vel cube and wavefield cube
         const cubeData<float>* vcub;
      public: //methods
      paddingFD2d(paddingFD* pNextPaddingObj) : paddingFD(pNextPaddingObj)
         { nx=nz=0; dt=dz=dx=0; dtDdz=dtDdx=0; bFreeSurf=false; vcub=NULL;}
      virtual ~paddingFD2d() {
	  }
	  
     virtual void update(wavFieldGroup& wfg) {}
	  virtual void initOperator(waveOpParams *par);
   }; //The paddingFD2d object can also be used as a dummy object that does not do any damping on padding layers.

   // Adjust the wave field in the absorbing boundaries using the
   // Israeli-Orszag (J. Comp Phys 1981) method to remove waves coming
   // back from the boundaries.  :-( not comprehensible
   class paddingFD2dSponge1: public paddingFD2d
   {
      public: //data members
      cubeData<ETYPE> m_mult_cb; //multiplier cube used in updating the sponge layer

      public: //methods
	  paddingFD2dSponge1(paddingFD* pNextPaddingObj,float spongeCoef) : paddingFD2d(pNextPaddingObj),mSpongeCoef(spongeCoef) {}
	  virtual ~paddingFD2dSponge1() {
	  }
	  
      virtual void update(wavFieldGroup& wfg);
      virtual void initOperator(waveOpParams *par);
   protected:
      float mSpongeCoef;
   };
}
#endif
