#include "stdafx.h"
#include "sep_params.h"

#include "modelingFD2d_op.h"
#include "paddingFD_damping.h"
#include "velFD2d.h"

using namespace Yang;

//init damping mult coef
void paddingFD2dDamping::initOperator(waveOpParams* par)
{
   paddingFD2d::initOperator(par);
   
   //Init the multiplier used in updating padded region, size is the same as 
   //Vel Model Shape.
   m_mult_cb.alloc((par->modelGridShape));
   
   //Init the content of the multiplier
   int npad = m_npad;
   
   m_mult_cb.zero();
   cubeData<ETYPE>& pm = this->m_mult_cb;
   
   stdVecFloat dampCoef;
   m_pDampingFunc->getCoefEx(dampCoef,npad,mDampingExtent);
   int ix,iz;
   //Notice that the corners are overwritten twice because of the overlap of the padding regions.
   if (!bFreeSurf)
   {
      //calc the top sponge
      for(ix=0;ix<nx; ix++)
      {
         for(iz=0; iz<npad; iz++)
         {
            pm(ix,iz) = dampCoef[npad-1-iz];
         }
      }
   }  //else do nothing

   //bottom sponge
   for(ix=0;ix<nx; ix++)
   {
      for(iz=nz-npad; iz<nz; iz++)
      {
         pm(ix,iz) = dampCoef[iz-(nz-npad)];
      }
   }

   //left and right sponge
   float tmp;
   for(ix=0;ix<npad; ix++)
   {
      for(iz=0; iz<nz; iz++)
      {
        tmp = pm(ix,iz);
  		 //Here handle the corner case
  		 if(!equalf(tmp,0.0f) && tmp < dampCoef[npad-1-ix]) //keep pm value
  		 {
  		 }
  		 else {
  			   pm(ix,iz) = dampCoef[npad-1-ix];
  		 }
      }
   }
   for(ix=nx-npad;ix<nx; ix++)
   {
      for(iz=0; iz<nz; iz++)
      {
		 tmp = pm(ix,iz);
		 //Here handle the corner case
		 if(!equalf(tmp,0.0f) && tmp < dampCoef[ix-(nx-npad)]) //keep pm value
		 {
		 }
		 else {
			pm(ix,iz) = dampCoef[ix-(nx-npad)];
		 }
      }
   }
   m_mult_cb.dump("dump1");
   return;
}

//Simple dampping wavefield
void paddingFD2dDamping::update(wavFieldGroup& wfg)
{
   //Get all necessary parameters
   int npad = m_npad;

   cubeData<ETYPE>& pm = m_mult_cb;
   int ix,iz;

   //get the pointers to the wavefiled
   cubeData<ETYPE> *ppp1, *ppp2, *ppp2dummy;
   wfg.getPtrs(&ppp1,&ppp2,&ppp2dummy);
   cubeData<ETYPE> &pp1=*ppp1, &pp2=*ppp2, &pp2dummy=*ppp2dummy;
   //the velocity field is v;
 
   int ivx,ivz;
   //Update the 4 sides
   //First the left and right
   //actuall memory cube pm is not necessary, the damping coef can be found through address
   //the damping function itself.
   for(ix=0+offx; ix<npad+offx; ix++)
   {
	  ivx = ix - offx;
      for(iz=0+offz; iz<nz+offz; iz++)
      {
		 ivz = iz - offz;
         pp2(ix,iz) = pp2(ix,iz) * pm(ivx,ivz);
         assert(pm(ivx,ivz)>0.1f);
      }
   }
   
   for(ix=nx-npad+offx; ix<nx+offx; ix++)
   {
	  ivx = ix - offx;
      for(iz=0+offz; iz<nz+offz; iz++)
      {
		 ivz = iz - offz;
         pp2(ix,iz) = pp2(ix,iz) * pm(ivx,ivz);
		 		 assert(pm(ivx,ivz)>0.1f); 
      }
   }
   
   //Bottom   
   for(ix=0+offx; ix<nx+offx; ix++)
   {
	  ivx = ix - offx;
      for(iz=nz-npad+offz; iz<nz+offz; iz++)
      {
		 ivz = iz - offz;
         pp2(ix,iz) = pp2(ix,iz) * pm(ivx,ivz);
         assert(pm(ivx,ivz)>0.1f);
      }
   }
   
   if (!bFreeSurf)  //Add the top region
   {
      for(ix=0+offx; ix<nx+offx; ix++)
      {
		 ivx = ix - offx;
         for(iz=0+offz; iz<npad+offz; iz++)
         {
           ivz = iz - offz;
           pp2(ix,iz) = pp2(ix,iz) * pm(ivx,ivz);
           assert(pm(ivx,ivz)>0.1f);
         }
      } 
   }
}


