#ifndef PARAM_FUNC_H
#define PARAM_FUNC_H 1
#include<string>
#include<stdbool.h>
#include<utility>

class param_func {
  public:
  
    param_func();
    int get_int(const char *arg)  const {
      return this->get_int(string(arg));
    }
    int get_int(const char *arg, int def) const {
      return this->get_int(string(arg),def);
    }
    virtual int get_int(const std::string& arg) const = 0;
    virtual int get_int(const std::string& arg, int def) const = 0;
    
    float get_float(const char *arg)            const {
      return this->get_float(string(arg));
    }
    float get_float(const char *arg, float def) const {
      return this->get_float(string(arg),def);
    }
    virtual float get_float(const std::string& arg)             const = 0;
    virtual float get_float(const std::string& arg, float def)  const = 0;

    bool get_bool(const char *arg)  const {
      return this->get_bool(string(arg));
    }
    bool get_bool(const char *arg, bool def) const {
      return this->get_bool(string(arg),def);
    }
    virtual bool get_bool(const std::string& arg) const = 0;
    virtual bool get_bool(const std::string& arg, bool def) const = 0;
    
    std::string get_string(const char *arg)     const {
      return this->get_string(string(arg));
    }
    std::string get_string(const char *arg, const char *def) const {
      return this->get_string(string(arg),string(def));
    }
    virtual std::string get_string(const std::string& arg)                  const = 0;
    virtual std::string get_string(const std::string& arg, std::string def) const = 0;
    
    int *get_ints(const char* arg, int nmax, int* p_nRead) const {
      return this->get_ints(string(arg),nmax,p_nRead);
    }
    int *get_ints(const char* arg, int nmax, int *def, int* p_nRead)const{
      return this->get_ints(string(arg),nmax,def,p_nRead);
    }
    virtual int *get_ints(const std::string& arg, int nmax, int* p_nRead)          const = 0;
    virtual int *get_ints(const std::string& arg, int nmax, int *def,int* p_nRead) const = 0;
    virtual float *get_floats(const string&arg, int nmax,float *def, int* p_nRead) const = 0;
    
    virtual bool is_input_file(const std::string& tag)const{if(tag=="");return false;}
    virtual bool is_inout_file(const std::string& tag)const{if(tag=="");return false;}
    virtual void error(const std::string& aa) const {if(aa=="") return;}
    virtual ~param_func() {}

protected:
  static int s_init_once_;
 };

#define NELEM(array) (sizeof(array)/sizeof(array[0]))
 
template <class T>
inline T findEnumItemFromPar(const param_func* pars, std::pair<int,T>* matchTable,int nItems, const string& parTag,int imenuDef) {
  int i=0; T ret;
  int imenu = pars->get_int(parTag,imenuDef);
  for (i=0;i<nItems;i++) {
    if (matchTable[i].first == imenu) {
      ret = matchTable[i].second;
      break;
    }
  }
  if (i==nItems) {//cannot find match
    errPrt("!!Param Item %s cannot be found, make sure it is a valid entry!\n", parTag.c_str());
  }
  return ret;
}

//global param reader 
extern param_func* gParamReader;

#endif

