#ifndef POINT_2D_H
#define POINT_2D_H 1

#include "stdafx.h"



// borrow and adapted from terathon.com Graphics library

namespace Yang {
template <typename T>
class point2d
{
public:
  union {
    T x;
    T sz;
  };
  union {
    T y;
    T sx;
  };
  
  //T x,y;
public:
        point2d() {}
        
        point2d( T r, T s)
        {
            x = r;
            y = s;
        }
		
		//Copy constructor
		point2d(const point2d& other)
		{
		   (*this)=other;
		}
		
		//assign operator
		point2d& operator = (const point2d& other)
		{
		   x = other.x;
		   y = other.y;
		   return (*this);
		}
		
        point2d& set(const T& r, const T& s)
        {
            x = r;
            y = s;
            return (*this);
        }
        
		//This operation might be at risk, it implictly assumes the memory layout of members x,y
        T& operator [](long k)
        {
            return ((&x)[k]);
        }
        
        const T& operator [](long k) const
        {
            return ((&x)[k]);
        }
        
        float module() const {
          return sqrtf(SQ(x)+SQ(y));
        }
        
        void normalize() {
          float fmod = this->module();
          VERIFY(fmod>10*epsilon);
          fmod = 1.0f/fmod;
          this->x *= fmod;
          this->y *= fmod;
        }
        
        point2d& operator +=(const point2d& v)
        {
            x += v.x;
            y += v.y;
            return (*this);
        }
        
        point2d& operator -=(const point2d& v)
        {
            x -= v.x;
            y -= v.y;
            return (*this);
        }
        
		
        point2d& operator *=(const T& t)
        {
            x *= t;
            y *= t;
            return (*this);
        }
        
		//reload the *= operator
        point2d& operator *=(const point2d& v)
        {
            x *= v.x;
            y *= v.y;
            return (*this);
        }
        
		//unitary negative operator
        point2d operator -(void) const
        {
            return (point2d(-x, -y));
        }
        
        point2d operator +(const point2d& v) const
        {
            return (point2d(x + v.x, y + v.y));
        }
        
        point2d operator -(const point2d& v) const
        {
            return (point2d(x - v.x, y - v.y));
        }
        
        point2d operator *(const T& t) const
        {
            return (point2d(x * t, y * t));
        }
        
		
        T dotProd(const point2d& v) const
        {
            return (x * v.x + y * v.y);
        }
        
        point2d operator *(const point2d& v) const
        {
            return (point2d(x * v.x, y * v.y));
        }
        
        bool operator ==(const point2d& v) const
        {
            return (equalf(x,v.x) && equalf(y,v.y));
        }
        
        bool operator !=(const point2d& v) const
        {
            return !(*this == v);
        }
};

typedef point2d<float> point2dF;
typedef point2d<int>point2dI;

//function that convert float pt type point2d to Int type, using integer rounding
//And report if truncation happens
inline void point2dFloatToInt( point2d<int>& dest, point2d<float>& src)
{
   dest.x = int(src.x+0.5f);
   dest.y = int(src.y+0.5f);
   //Verify the truncation error
   if(!equalf(src.x,float(dest.x)))
   {  prt("Not equal from point2d.x Float to Int");}
   if(!equalf(src.y,float(dest.y)))
   {  prt("Not equal from point2d.y Float to Int");}
}

inline void point2dIntToFloat( point2d<float>& dest, point2d<int>& src)
{
   dest.x = float(src.x);
   dest.y = float(src.y);
}

//Rotate a point
inline void rotate(point2d<float>& pt, float angle)
{
   float s = sinf(angle);
   float c = cosf(angle);
   float nx = c*pt.x - s*pt.y;
   float ny = s*pt.x + c*pt.y;
   pt.x = nx; pt.y = ny;
}

inline void normalize(point2d<float>& pt)
{
   pt *= (1.0f/sqrtf(SQ(pt.x) + SQ(pt.y)));
}

inline float magnitude(const point2d<float>& v)
{
    return (sqrtf(v.x * v.x + v.y * v.y));
}

inline float squaredMag(const point2d<float>& v)
{
    return (v.x * v.x + v.y * v.y);
}

} //end namespace
#endif

