#ifndef POINT_3D_H
#define POINT_3D_H

#include "point2d.hpp"

/*
   borrow and adapted from terathon.com Graphics library
*/

namespace Yang {
template <typename T>
class point3d
{
public:
   union {
	  T x;
	  T sz;
   };
   
   union {
	  T y;
	  T sx;
   };
   
   union {
	  T z;
	  T sy;
   };        

   point3d() {}
        
        point3d(T r, T s, T t)
        {
            x = r;
            y = s;
            z = t;
        }
        
        point3d(const point2d<T>& v, T u)
        {
            x = v.x;
            y = v.y;
            z = u;
        }
        
		//Copy constructor
		point3d(const point3d& other)
		{
		   (*this)=other;
		}
		
		//assign operator
		point3d& operator = (const point3d& other)
		{
		   x = other.x;
		   y = other.y;
		   z = other.z;
		   return (*this);
		}

        point3d& set(T r, T s, T t)
        {
            x = r;
            y = s;
            z = t;
            return (*this);
        }
        
        point3d& set(const point2d<T>& v, T u)
        {
            x = v.x;
            y = v.y;
            z = u;
            return (*this);
        }

		//Use at risk
        T& operator [](long k)
        {
            return ((&x)[k]);
        }
        
        const T& operator [](long k) const
        {
            return ((&x)[k]);
        }
        
		        
        point3d& operator +=(const point3d<T>& v)
        {
            x += v.x;
            y += v.y;
            z += v.z;
            return (*this);
        }
        
        point3d& operator +=(const point2d<T>& v)
        {
            x += v.x;
            y += v.y;
            return (*this);
        }
        
        point3d& operator -=(const point3d<T>& v)
        {
            x -= v.x;
            y -= v.y;
            z -= v.z;
            return (*this);
        }
        
        point3d& operator -=(const point2d<T>& v)
        {
            x -= v.x;
            y -= v.y;
            return (*this);
        }
        
        point3d& operator *=(T t)
        {
            x *= t;
            y *= t;
            z *= t;
            return (*this);
        }
                
        point3d& operator *= (const point3d<T>& v)
        {
            x *= v.x;
            y *= v.y;
            z *= v.z;
            return (*this);
        }
        
        point3d operator -(void) const
        {
            return (point3d(-x, -y, -z));
        }
        
        point3d operator +(const point3d<T>& v) const
        {
            return (point3d(x + v.x, y + v.y, z + v.z));
        }
        
        point3d operator +(const point2d<T>& v) const
        {
            return (point3d(x + v.x, y + v.y, z));
        }
        
        point3d operator -(const point3d<T>& v) const
        {
            return (point3d(x - v.x, y - v.y, z - v.z));
        }
        
        point3d operator -(const point2d<T>& v) const
        {
            return (point3d(x - v.x, y - v.y, z));
        }
        
        point3d operator *(T t) const
        {
            return (point3d(x * t, y * t, z * t));
        }
                
                
        point3d operator &(const point3d& v) const
        {
            return (point3d(x * v.x, y * v.y, z * v.z));
        }
        
        bool operator ==(const point3d& v) const
        {
            return (equalf(x,v.x) && equalf(y,v.y) && equalf(z,v.z));
        }
        
        bool operator !=(const point3d& v) const
        {
            return !(*this == v);
        }
		
		point3d crossProd(const point3d<T>& v) const
		{
		   T r, s, t;
		   r = y * v.z - z * v.y;
		   s = z * v.x - x * v.z;
		   t = x * v.y - y * v.x;
		   return point3d<T>(r,s,t);
		}

		T dotProd(const point3d<T>& v1, const point3d<T>& v2) const
		{
			return (v1.x*v2.x + v1.y*v2.y + v1.z*v2.z);
		}
};


typedef point3d<float> point3dF;
typedef point3d<int>	point3dI;

//function that convert float pt type point2d to Int type, using integer rounding
//And report if truncation happens
inline void point3dFloatToInt( point3dI& dest, point3dF& src)
{
   dest.x = int(src.x+0.5f);
   dest.y = int(src.y+0.5f);
   dest.z = int(src.z+0.5f);
   
   //Verify the truncation error
   if(!equalf(src.x,float(dest.x)))
	  prt("Not equal from point3d.x Float to Int");
   if(!equalf(src.y,float(dest.y)))
	  prt("Not equal from point3d.y Float to Int");
   if(!equalf(src.z,float(dest.z)))
	  prt("Not equal from point3d.z Float to Int");
}

inline void point3dIntToFloat( point3dF& dest, point3dI& src)
{
   dest.x = float(src.x);
   dest.y = float(src.y);
   dest.z = float(src.z);
}

inline float magnitude(const point3dF& v)
{
    return (sqrtf(v.x * v.x + v.y * v.y + v.z * v.z));
}

inline float SquaredMag(const point3dF& v)
{
    return (v.x * v.x + v.y * v.y + v.z * v.z);
}

inline void normalize(point3dF& u)
{
   u *= sqrtf(u.x*u.x + u.y * u.y + u.z * u.z);
}

inline void rotateAboutX(point3dF& u, float angle)
{
   float s = sinf(angle);
   float c = cosf(angle);
    
   float ny = c * u.y - s * u.z;
   float nz = c * u.z + s * u.y;
   
   u.y = ny;
   u.z = nz;
}

inline void rotateAboutY(point3dF& u, float angle)
{
    float s = sinf(angle);
    float c = cosf(angle);
    
    float nx = c * u.x + s * u.z;
    float nz = c * u.z - s * u.x;
    
    u.x = nx;
    u.z = nz;
}

inline void rotateAboutZ(point3dF& u, float angle)
{
    float s = sinf(angle);
    float c = cosf(angle);
    
    float nx = c * u.x - s * u.y;
    float ny = c * u.y + s * u.x;
    
    u.x = nx;
    u.y = ny;
}

inline void rotateAboutAxis(point3dF& u, float angle, const point3dF& axis)
{
    float s = sinf(angle);
    float c = cosf(angle);
    float k = 1.0f - c;
    
    float nx = u.x * (c + k * axis.x * axis.x) + u.y * (k * axis.x * axis.y - s * axis.z)
            + u.z * (k * axis.x * axis.z + s * axis.y);
    float ny = u.x * (k * axis.x * axis.y + s * axis.z) + u.y * (c + k * axis.y * axis.y)
            + u.z * (k * axis.y * axis.z - s * axis.x);
    float nz = u.x * (k * axis.x * axis.z - s * axis.y) + u.y * (k * axis.y * axis.z + s * axis.x)
            + u.z * (c + k * axis.z * axis.z);
    
    u.x = nx;
    u.y = ny;
    u.z = nz;
}

}

#endif
