
#undef NDEBUG
#include <assert.h>

#include "stdafx.h"
#include "cubeData.hpp"
#include "rayTracer.h"

#undef DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;

int rayTracerZ1d::rayPath(vector<stdVecFloat>& rays, const float* coords, const vector<fwk::degree>& vAng) const {
  int iOz,iOx,iOy; //the icoords of the origin of rays. 
  iOy = 0;
  const vecAxis& axes=vel_->domain().get_axes();
  assert(vel_.get());
  iOz = axes[0].get_offset(coords[0]);  iOx = axes[1].get_offset(coords[1]);
  if (ndim_>2) iOy = axes[2].get_offset(coords[2]);
  int ng = vAng.size(); rays.resize(ng); float gamma; bool bUpward;
  int nz = axes[0].n, nx = axes[1].n;
  float dz=axes[0].d;
  //Just do 1D trace.
  const float* pv = &(*vel_)(iOy,iOx,0);
  int iiz,iip; float sinOverV_c = 0.0f; //The constant remains through all the layers.
  float theta1,theta2,xcoord; float tmp;
  float xmin,xmax; axes[1].getRange(&xmin,&xmax);
  PRV2(xmin,xmax);
  for (int ig=0; ig<ng; ig++) {
    gamma = vAng[ig]; if (gamma > 180.0f) gamma -= 360.0f;
    if (gamma>-90.0f && gamma <90.0f) bUpward = false; else bUpward = true;
    sinOverV_c = sinf(deg2rad(gamma))/pv[iOz]; theta1 = deg2rad(gamma);
    bool bFullRefract = false;
    if (bUpward) {
      theta1 = PI - theta1;
      //Init zeros to the ray arrays.
      rays[ig].resize((iOz+1)*ndim_); rays[ig].assign((iOz+1)*ndim_,0.0f);
      rays[ig][0] = axes[0].get_coord(iOz); rays[ig][1] = axes[1].get_coord(iOx); 
      if (ndim_>2) rays[ig][2] = iOy; //push in the origin point
      //trace from bot to top
      for (iiz=iOz-1; iiz>=0; iiz--) {
        iip = iOz - iiz;
        tmp = pv[iiz]*sinOverV_c;
        rays[ig][iip*ndim_+0] = axes[0].get_coord(iiz); //z coord
        if (fabs(tmp)<1.0f) { //normal case, apply the Snell law.
          theta2 = asinf(tmp);
        } else { //refraction angle is 90 degree.
          bFullRefract = true;
          //just fill one last point and leave.
          if (tmp>0)  rays[ig][iip*ndim_+1] = axes[1].get_coord(nx-1); //x coord
          else        rays[ig][iip*ndim_+1] = axes[1].get_coord(0); //x coord
          
          break;
        }
        //record the coordinates
        xcoord = rays[ig][(iip-1)*ndim_+1] + 0.5f*dz*(tanf(theta1)+tanf(theta2));
        rays[ig][iip*ndim_+1] = xcoord; //x coordinates.
        theta1 = theta2;
        //check if the x coord is out of domain, if so just stop
        if (xcoord<xmin || xcoord>xmax) break; //go to search for the next angle.
      }
    } else { //trace downwards
      rays[ig].resize((nz-iOz)*ndim_);  rays[ig].assign((nz-iOz)*ndim_,0.0f);
      for (iiz=iOz+1; iiz<nz; iiz++) {
        iip = iiz - iOz;
      }
      VERIFY(0);
    }
  } //for each angle.
  return 0;
}

