#ifndef RAY_TRACER_H
#define RAY_TRACER_H 1

#include "decl_fwd.h"
#include "cubeData.hpp"
#include "stdafx.h"

using boost::shared_ptr;
namespace Yang {
//Simple raytracer, give a subsurface point, and a dip angle, it will shoot rays with a specific angle, and it will produce a bunch or ray pairs gives its path and give you the angle of coverage, roughly the surface location of the src/recv location.
//1D depth ray tracing
class rayTracer{
public:
  void velIs(const shared_ptr<const cubeDataF>& vel) {
    this->vel_ = vel;
    ndim_ = this->vel_->domain().get_ndim();
    VERIFY(ndim_<=3);
    if (ndim_==3) {
      b3d_ = true;
    } else {
      b3d_ = false;
    }
  }
  const shared_ptr<const cubeDataF>& vel() const{ return this->vel_;}
  
  //give a point in vel, and dip, trace the raypath for different angles specified by the program. Rays are a list of stdVectors, each stdVector in the list corresponds to the (float) coordinates of the ray at that specific angle. 
  //The angle is defined as along the +z axis, i.e. from -90 to +90, shooting downward, from 90 to 270, shooting upward
  virtual int rayPath(vector<stdVecFloat>& rays, const float* coords, const vector<fwk::degree>& vAng) const = 0;
protected:
  shared_ptr<const cubeDataF> vel_; //The velocity field
  bool b3d_; int ndim_;
  
};

class rayTracerZ1d: public rayTracer {
public:
  virtual int rayPath(vector<stdVecFloat>& rays, const float* coord, const vector<fwk::degree>& vAng) const; //get raypath from tracer at a model point.
};

} //end namespace Yang
#endif

