#ifndef __RSF_H_INCL__
#define __RSF_H_INCL__

/**
 * TODO: Add class description
 * 
 * @author   yang
 */

#include "hypercube.h"

namespace Yang {
  //Windowing parameters
   struct windowAxis
      {
         int n; //length
         int f; //frame start index
         int j; //jump interval
         windowAxis():n(0),f(0),j(1) {}
         windowAxis(int _n, int _f=0, int _j=1) { set(_n,_f,_j);}
         void set(int _n,int _f, int _j) {
            n = _n;
            f = _f;
            j = _j;
         }
		 void print() const{
			prt("\n f = %d, j = %d, n = %d", f,j,n);
		 }
      };
   typedef vector<windowAxis> windParam;
   inline void printWindParam(const windParam& wind, std::string sz = "") {
	  int numAxis = wind.size();
	  prt("\n window param %s has %d Axes", sz.c_str(), numAxis);
	  for (int i=0; i<numAxis; i++) {
		 wind[i].print(); prt(": axis %d", i+1);
	  }
   }

   //forward declare the cubeData class, which is a child of rsf
template<typename T>
class cubeData;

//Interface classes for block grab and update
template<typename T>
class rsf {

protected: //data members

   hypercube dom;
   //tmp data member for grab/update functions
   bool m_bGrab;
   windParam   mWindParamsTmp;
   const windParam*  m_pWindPar;
   //For read/write access
   long long m_seqCounter;

public:
	// Constructor
	rsf():m_seqCounter(0){}
	// Destructor
	virtual ~rsf() {}
	virtual void resetCounter() {
	   m_seqCounter = 0;
	}
	
	long long getSeqCounter() {
		return m_seqCounter;
	}
	const hypercube& domain() const{ return dom; }
	
   //Random access of the data, here the incore and outcore objects differs
   //Grab a 1d,2d,3d slice from the RSF data object, assume a memory copy or read from file, memory are pre-allocated by the caller

   //Random read
   virtual void grab3d(cubeData<T> * cube,  const windParam& wind)
   {
      VERIFY(cube->domain().get_ndim()==3);
      VERIFY(this->domain().compatible(cube->domain()))
   }
   virtual void grab2d(cubeData<T> * cube,  const windParam& wind)
   {
      VERIFY(cube->domain().get_ndim()==2);
      VERIFY(this->domain().compatible(cube->domain()))
   }
   virtual void grab1d(cubeData<T> * cube,  const windParam& wind)
   {
      //Just do some size checking
      VERIFY(cube->domain().get_ndim()==1);
      VERIFY(this->domain().compatible(cube->domain()))
   }
   virtual void grabXd(cubeData<T> * cube,  const windParam& wind, int ndim)
   {
      //Just do some size checking
      VERIFY(cube->domain().get_ndim()==ndim);
      VERIFY(this->domain().compatible(cube->domain()))
   }
   virtual void grabSeveralXd(cubeData<T> * cube,  const windParam& wind, int ndim)
   {
      //Just do some size checking
      VERIFY(cube->domain().get_ndim()==ndim+1);
      VERIFY(this->domain().compatible(cube->domain().return_compatible(ndim)))
      
      //check whether the window parameter is out of bound
      windParam windd(wind);
      checkWindowParams_par(cube,windd);
   }	  
   	  
   virtual void grabWindow(cubeData<T> * cube,  const windParam& wind) {
      checkWindowParams(cube,wind);
   }
  
   //Random write 
   virtual void update1d(const cubeData<T> * cube,  const windParam& wind)
   {
      VERIFY(cube->domain().get_ndim()==1);
	  VERIFY(this->domain().compatible(cube->domain()))
      checkWindowParams(cube,wind);
   }
   virtual void update2d(const cubeData<T> * cube,  const windParam& wind)
   {
      VERIFY(cube->domain().get_ndim()==2);
	  VERIFY(this->domain().compatible(cube->domain()))
      checkWindowParams(cube,wind);
   }
   virtual void update3d(const cubeData<T> * cube,  const windParam& wind)
   {
      VERIFY(cube->domain().get_ndim()==3);
	  VERIFY(this->domain().compatible(cube->domain()))
      checkWindowParams(cube,wind);
   }
   virtual void updateXd(const cubeData<T> * cube,  const windParam& wind, int ndim)
   {
      VERIFY(cube->domain().get_ndim()==ndim);
	  VERIFY(this->domain().compatible(cube->domain()))
      checkWindowParams(cube,wind);
   }
   virtual void updateSeveralXd(const cubeData<T> * cube,  const windParam& wind, int ndim)
   {
      VERIFY(cube->domain().get_ndim()==ndim+1);
      VERIFY(this->domain().compatible(cube->domain().return_compatible(ndim)))

      //check whether the window parameter is out of bound
      windParam windd(wind);
      checkWindowParams_par(cube,windd);
   }

   virtual void updateWindow(const cubeData<T> * cube,  const windParam& wind){
      checkWindowParams(cube,wind);
   }
   
   //Sequentially read-write functions //Do size checking here
   virtual void seq_grab(cubeData<T> * cube) {
     if(this->dom.compatible(cube->domain())) {
  		 long long n = cube->domain().get_n123();
  		 //Update the counter at first hand
  		 m_seqCounter += n;
  		 if(m_seqCounter > dom.get_n123()) {
         PRT_LOC; errPrt("Seq grab out of boundary in dmRsf::seq_grab! %lld/%lld/%lld \n", m_seqCounter,n,dom.get_n123());
  		 }
     } else {
       PRT_LOC; seperr("Size incompatible in rsf::seq_grab\n");
     }
   }

   virtual void seq_grab(T* pArr, size_t n) {
		 //Update the counter at first hand
		 m_seqCounter += n;
		 if(m_seqCounter > dom.get_n123()) {
			 PRT_LOC; errPrt("Seq grab out of boundary in dmRsf::seq_grab! %lld/%lld/%lld \n", m_seqCounter,n,dom.get_n123());
		 }
   }
   
   virtual void seq_update(const cubeData<T> * cube){
     if(this->dom.compatible(cube->domain())) {
  		 long long n = cube->domain().get_n123();
  		 m_seqCounter += n;
  		 if(m_seqCounter > dom.get_n123())  {
         PRT_LOC; errPrt("Seq grab out of boundary in dmRsf::seq_grab! %lld/%lld/%lld \n", m_seqCounter,n,dom.get_n123());
       }
     }
     else {
       PRT_LOC; seperr("Size incompatible in rsf::seq_update\n");
     }
   }
   
    virtual void seq_update(const T* pArr, size_t n){
		  m_seqCounter += n;
		  if(m_seqCounter > dom.get_n123())
		  {
        PRT_LOC; errPrt("Seq grab out of boundary in dmRsf::seq_grab! %lld/%lld/%lld \n", m_seqCounter,n,dom.get_n123());
		  }
    }
    
   bool sameSize(const rsf<T>& rsfObj) const
   {
	  return this->domain().sameSize(rsfObj.domain());
   }
   
   bool compatible(const rsf<T>& rsfObj) const
   {
	  return this->domain().compatible(rsfObj.domain());
   }
   
protected: 
   //does not allow copy and assig
	// Copy constructor
	// TODO: Uncomment the copy constructor when you need it.
	//rsf(const rsf& src);

	// Assignment operator
	// TODO: Uncomment the assignment operator when you need it.
	//rsf& operator=(const rsf& src);
	
  //This function uses a internal variable mWindParamsTmp, which makes it unparallelizable.
	void checkWindowParams(const cubeData<T> *cube,const windParam& wind) {
      mWindParamsTmp = wind;
      //Make sure the size of cube, is consistent with wind, and this->dom
      //First fill the default values in wind param
      int ndim = dom.get_ndim();
      VERIFY(uint(ndim) == mWindParamsTmp.size());
      VERIFY(cube->domain().get_ndim() <= ndim);
      int i,n;
      
      long long N = 1;
      windowAxis ax_w;
      for(i=0;i<ndim;i++) {
         ax_w = mWindParamsTmp[i];
         n = dom.get_n(i+1);
         if (0==ax_w.n) //means get the full dimension of the target cube
         {
            ax_w.n = int((n-ax_w.f-1)/ax_w.j)+1;
            //VERIFY(ax_w.n ==  cube->domain().get_n(i+1));
         }
         //boundary check at each dimension.
         VERIFY(ax_w.f + (ax_w.n-1)*ax_w.j < n);
         mWindParamsTmp[i] = ax_w;
         N *= ax_w.n;
      }
      VERIFY(N == cube->domain().get_n123());
   }
   
   //can be parallized, but it will modify the input wind param.
   void checkWindowParams_par(const cubeData<T> *cube, windParam& wind) {
      //Make sure the size of cube, is consistent with wind, and this->dom
      //First fill the default values in wind param
      int ndim = dom.get_ndim();
      VERIFY(uint(ndim) == wind.size());
      VERIFY(cube->domain().get_ndim() <= ndim);
      int i,n;
      
      long long N = 1;
      windowAxis ax_w;
      for(i=0;i<ndim;i++) {
         ax_w = wind[i];
         n = dom.get_n(i+1);
         if (0==ax_w.n) //means get the full dimension of the target cube
         {
            ax_w.n = int((n-ax_w.f-1)/ax_w.j)+1;
            //VERIFY(ax_w.n ==  cube->domain().get_n(i+1));
         }
         //PRV4(ax_w.f,ax_w.n,ax_w.j,n);
         //boundary check at each dimension.
         VERIFY(ax_w.f + (ax_w.n-1)*ax_w.j < n);
         wind[i] = ax_w;
         N *= ax_w.n;
      }
      VERIFY(N == cube->domain().get_n123());
   }
};
}
#endif // __RSF_H_INCL__
