#include "stdafx.h"
#include "cubeData.hpp"
#include "misc.h"
#include "tagHack.h"
#include <sstream>
//Use seplib tools, in the form of command line calling

#undef DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace seplib;

//Use randomized tag and filename as temporary intermediate results.
static string szToolFileTag1,szToolFileTag2;
static char szToolFileName1[1024], szToolFileName2[1024];
static char szCmd[8192]; //bib enough

void checkToolFileTag() {
   if(tagExist(szToolFileTag1)) {
      prt("Required tag %s already exists!\n",szToolFileTag1.c_str());
      raise(SIGSEGV);
   }
   if(tagExist(szToolFileTag2)) {
      prt("Required tag %s already exists!\n",szToolFileTag2.c_str());
      raise(SIGSEGV);
   }
}
/*
//Random generate filename to avoid collision
void seplib::genFileName() {
   FUNC_TRACE_BEG(genFileName)
   //Get current time
   time_t sec = time(NULL);
   srand(sec);
   int num = rand();
   sprintf(szToolFileName1,"tool-%d.H",num);
   sprintf(szToolFileName2,"tool-%d.H",rand());
   FUNC_TRACE_END()
}
*/

static std::string str_path("../tmp/");

//Use Transp to transpose dataset, OFF2ANG and ANG2OFF
//outCube is allocated inside the function
void seplib::callSepTool(cubeData<float>& outCube, const cubeData<float>& inCube, const string& toolName, const string& par, bool bAllocOutMemInside) {
   
   FUNC_TRACE_BEG(callSepTool);
   //Build the randomized tag and filename   
   srand(time(NULL));
   int rsuf = 0; std::ostringstream oss;
   oss.str(""); rsuf = rand()%65536;
   oss << str_path << "sepToolTemp" << rsuf <<".H";  
strcpy(szToolFileName1,oss.str().c_str());
   oss.str(""); oss<< "sepToolTag"<<rsuf; szToolFileTag1 = oss.str();
   
   oss.str(""); rsuf = rand()%65536;
   oss << str_path << "sepToolTemp" << rsuf <<".H"; 
strcpy(szToolFileName2,oss.str().c_str());
   oss.str(""); oss<< "sepToolTag"<<rsuf; szToolFileTag2 = oss.str();
   VERIFY(strcmp(szToolFileName1,szToolFileName2)!=0);
   
   bool bInput = false; bool bNoDummy = true;
   //First Dump the input to tag1
   //PRV(szToolFileTag1);PRV(szToolFileName1);
   inCube.dump(szToolFileName1,bNoDummy);  usleep(1e5); //sleep 0.1s safety
   //setup the cmd, the user can chain cmds by putting the after cmds in par string
   sprintf(szCmd,"%s <%s %s >%s \n", toolName.c_str(), szToolFileName1, par.c_str(), szToolFileName2);
   LOG("E0");
   prt("\nExecuting cmd: %s", szCmd);
   int sysret = system(szCmd); VERIFY(sysret == 0); usleep(1e5); //wait to make sure the cmd is finished.
   LOG("E1");
   //Now the output is in szToolFileName2, associate it with tag szToolFileTag2, read it back
   dmRsf_oc<float> tool2Cube(szToolFileName2,dm::IN,"result",bNoDummy); tool2Cube.init(NULL);
   //cubeDataF tool2Cube; tool2Cube.load(szToolFileName2,inCube.domain().get_ndim());
   //Now the data is in tool2Cube.
   LOG("E2");
   //If the user alloc the output memory, then need check if the memory size is consistent
   if (!bAllocOutMemInside) {
     VERIFY(outCube.sameSize(tool2Cube));
   } else { //alloc memory inside
     outCube.alloc(tool2Cube.domain());
   }
   LOG("E3");
   //read the values to outCube.
   tool2Cube.seq_grabAll(&outCube);
   LOG("E4");
   //Remove the intermediate files
   sprintf(szCmd, "Rm %s; Rm %s", szToolFileName1,szToolFileName2); 
   verbPrt("exec cmd:%s\n",szCmd); 
   //sysret = system(szCmd); VERIFY( 0 == sysret); 
   usleep(1e5);
   FUNC_TRACE_END()
}

// Need to make a function that does angle transform
// Does OFF2ANG transform, if binverse=1 then does ANG to OFF transform
// 2D transform, The axis arrangement is (z,hx,x) <--> (z,ax,x)
// cube_transp is the variable for intermediate transpose result.
//need to stuff dummy axis for OFF2ANG program, from z,hx,x --> z,hx,hy,x,y
void getShapeWithDummyAxis(hypercube& outShape, const hypercube& transpShape) {
   vecAxis AxesWithDummy = transpShape.get_axes();
   axis dummyAxis(1,0.0f,1.0f);
   VERIFY(AxesWithDummy.size() >= 2);
   AxesWithDummy.insert(AxesWithDummy.begin()+2,dummyAxis);
   AxesWithDummy.push_back(dummyAxis);
   outShape.set_axes(AxesWithDummy);
 }
void removeDummyAxisFromShape(hypercube& outShape, const hypercube& transpShapeWithDummyAxis) {
   
   vecAxis AxesWithDummy = transpShapeWithDummyAxis.get_axes();
   //now axis is z,ax,ay,x,y, get rid of 2 of them
   assert(AxesWithDummy[4].n ==1 && AxesWithDummy[2].n==1 );
   AxesWithDummy.pop_back(); AxesWithDummy.erase(AxesWithDummy.begin()+2);
   outShape.set_axes(AxesWithDummy);
}

void seplib::off2ang(cubeData<float>& outCube, cubeData<float>& inCube, bool bInverse, float a_h_max) {
   FUNC_TRACE_BEG(off2ang)
   hypercube shapeIn = inCube.domain();
   //Assume 2D
   VERIFY(shapeIn.get_ndim()==3);

   char tmpCmd[1024],tmpPar[1024];

   //Then transform to angle domain
   //need to stuff dummy axis for OFF2ANG program, from z,hx,x --> z,hx,hy,x,y
   hypercube ShapeWithDummyAxis;
   getShapeWithDummyAxis(ShapeWithDummyAxis,shapeIn);
   
   inCube.reshape(ShapeWithDummyAxis);
   inCube.resetCounter();
   if (!bInverse) { //From offset to angle
      sprintf(tmpCmd,"OFF2ANG.x"); sprintf(tmpPar,"amax=%.1f ",a_h_max);
      VERIFY(a_h_max >=0.0f && a_h_max < 90.0f);
   } else {
      sprintf(tmpCmd,"OFF2ANG.x"); sprintf(tmpPar,"hmax=%.1f inverse=1",a_h_max);
   }
   seplib::callSepTool(outCube,inCube,string(tmpCmd),string(tmpPar));
   
   //now axis is z,ax,ay,x,y, get rid of 2 of them
   hypercube transpShapeAng;
   removeDummyAxisFromShape(transpShapeAng,outCube.domain());

   //shape incube back
   inCube.reshape(shapeIn);
   outCube.reshape(transpShapeAng);
   
   if (bInverse) {
      //Reverse the axis because the bug of ANG2OFF inverse.
      seplib::callSepTool(outCube,outCube,
                          string("Reverse"),string(" which=1 opt=ignore esize=4 verb=1 ")); 
   }
   FUNC_TRACE_END()
}

