#include "stdafx.h"
#include "cubeData.hpp"
#include "cubeAlgo2.h"
#include <seplib.h>
#include <valarray>
#include <boost/shared_array.hpp>

///////////////////Copy from SEPlib smooth program seis/filter ///////////////////////

/* NOTE: C TYPE STORAGE  */
#define DATA(i,j) data[n1*(i)+(j)]
#define SMOO(i,j) smoo[n1*(i)+(j)]
#define SUM(i,j)  sum[n1*(i)+(j)]
#define TEMP(i,j,k) temp[n12*(reflect(b3,i))+n1*(j)+(k)]

#define TRIS_NMAX 8192


using namespace std;
using namespace Yang; 

int reflect(int n,int i)
{
  int j;
  j=i;
  if (n==1) return 0;
  while (j>n-1  ||  j<0) {
    if  (j>n-1) j= 2*n-1-j;
    if  (j<0)   j=  (-j); 
  } 
  return j;
}

void copy(int n,float *xx,float *yy)
{
	int i;

	for (i=0;i<n;i++) yy[i] = xx[i];

	return;
}

/*************************************************************************
*                          Subroutine boxconv                            *
**************************************************************************
**************************************************************************
*   inputs:       nx,  xx(i), i=1,nx      the data                       *
*                 nb                      the box length                 *
*   output:       yy(i),i=1,nx+nb-1       smoothed data                  *
***************************************************************************/
void boxconv(int nb,int nx,float *xx,float *yy, valarray<float>& bb)
{
	int i,ny;

	/*  real bb(nx+nb) */
	/*    float bb[TRIS_NMAX]; */

	if( nb<1 || nb>nx) seperr("ERROR: in boxconv");
	ny = nx+nb-1;
	for (i=0;i<ny;i++) bb[i]=0.;

	bb[0] = xx[0];
	for (i=1;i<nx;i++)
		bb[i] = bb[i-1] + xx[i];         /*   make B(Z) = X(Z)/(1-Z) */
	for (i=nx;i<ny;i++)
		bb[i] = bb[i-1];
	for (i=0;i<nb;i++)
		yy[i] = bb[i];
	for (i=nb;i<ny;i++)
		yy[i] = bb[i] - bb[i-nb];        /*   make Y(Z) = B(Z)*(1-Z**nb) */
	for (i=0;i<ny;i++)
		yy[i] = yy[i]/nb;
}

/*************************************************************************
*                          Subroutine tris                               *
**************************************************************************
*          tridiagonal simultaneous equations as in FGDP and IEI         *
**************************************************************************/

int tris(int n,float endl, float a, float b, float c, float endr,
    float *d, float *t,valarray<float>& e,valarray<float>& deni,valarray<float>& f)
{
	int i;

/*    float e[TRIS_NMAX], f[TRIS_NMAX], deni[TRIS_NMAX]; */
/*    if (n > TRIS_NMAX) seperr("ERROR: n > TRIS_NMAX, recompile needed"); */

	if( n == 1) { t[0] = d[0] / b; return 0; }
	e[0] = - a / endl;
	for (i=1;i<n-1;i++) {
		deni[i] = 1.f / ( b + c * e[i-1] );
		e[i] = - a * deni[i];
	}

	f[0] = d[0] / endl;
	for (i=1;i<n-1;i++)
		f[i] = (d[i] - c * f[i-1]) * deni[i];

	t[n-1] = (d[n-1] - c * f[n-2])/(endr + c * e[n-2]);
	for (i=n-2;i>=0;i--)
		t[i] = e[i] * t[i+1] + f[i];

	return 0;
}

/*************************************************************************
*                          Subroutine leaky                              *
**************************************************************************
*                  tridiagonal smoothing on 1-axis or 2-axis             *
**************************************************************************
*   distance,  input:  1. < distance < infinity                          *
*   uu[m1*n12]:        data in  is the vector (uu[m1*i], i=0,n12-1)      *
*   vv[m1*n12]:        data out is the vector (vv[m1*i], i=0,n12-1)      *
*************************************************************************/
void leaky(float distance,int m1,int n12,float *uu,float *vv, valarray<float>& vecin, valarray<float>& vecout,valarray<float>& e,valarray<float>& deni,valarray<float>& f)
{
	float a, b, dc, side;
	int i;
    
  /*  float vecin[TRIS_NMAX], vecout[TRIS_NMAX]; */
  /*  if (n12 > TRIS_NMAX) seperr("ERROR: n > TRIS_NMAX, recompile needed"); */

	a  = - (1.f-1.f/distance);        b = 1.+a*a;     dc = b+a+a;
	a = a/dc;       b = b/dc;       side = a + b;

	for (i=0;i<n12;i++)  vecin[i] = uu[i*m1];

	if( distance<=1. || n12==1) copy( n12, va_ptr(vecin), va_ptr(vecout));
	else tris( n12, side, a, b, a, side, va_ptr(vecin), va_ptr(vecout),e,deni,f);
  
	for (i=0;i<n12;i++)  vv[i*m1]=vecout[i];
	return;
}


/*************************************************************************
*                          Subroutine triangle                           *
**************************************************************************
*                        Convolve with triangle                          *
**************************************************************************
*                                                                        *
*  input:  nr    rectangle width (points) (Triangle base twice as wide.) *
*  input:  uu(m1,i2),i2=1,n12      is a vector of data.                  *
*  output: vv(m1,i2),i2=1,n12      may be on top of uu                   * 
*************************************************************************/
void triangle(int nr,int m1,int n12,float *uu,float *vv, valarray<float>&pp, valarray<float>& qq, valarray<float>& tt, valarray<float>& bb)
{
	int np,nq,i;

  /*  real pp(n12+nr-1), qq(n12+nr+nr-2), tt(n12) */
  /*  float pp[TRIS_NMAX], qq[TRIS_NMAX], tt[TRIS_NMAX]; */

	for (i=0;i<n12;i++)  qq[i] = uu[i*m1];

	if (n12 == 1) myMemcpy( va_ptr(tt), va_ptr(qq), n12);
	else {
		boxconv(nr, n12, va_ptr(qq), va_ptr(pp),bb);      np = nr+n12-1;
		boxconv( nr, np , va_ptr(pp), va_ptr(qq),bb);     nq = nr+np-1;
		for (i=0;i<n12;i++)  tt[i]  = qq[i+nr-1];
		for (i=0;i<nr-1;i++) tt[i]      +=qq[nr-i-2];      /* fold back near end */
		for (i=0;i<nr-1;i++) tt[n12-i-1]+=qq[n12+(nr-1)+i]; /* fold back far end */
	}
	for (i=0;i<n12;i++)  vv[i*m1]=tt[i];
	return;
}

//no more than 3 dimensions
int Yang::rectSmooth (cubeData<float>& cubIn, cubeData<float>& cubOut, int rect1, int rect2, int rect3, int nrepeat, bool bAbsval)
{
	int n1,n2,n3,n4,n5,n6,ntotal,n12,nread;
	//int tridiag,scaleup,diff1,diff2,agc;
	//int absval,sum3;
	int b3,maxrect;
	int j, i1,i2,i3,irepeat;
	//int maxmem, memcount;
	float dist1,dist2; //scale;
	float scale3;

  valarray<float> vecin,vecout;
  valarray<float> pp, qq, tt;
  valarray<float> e, f, deni;
  valarray<float> bb;
  valarray<float> data,smoo,sum,temp;

  hypercube domIn = cubIn.domain();
  VERIFY(domIn.sameSize(cubOut.domain()));
  int ndim = domIn.get_ndim();
  n1 = domIn.get_n(1);
  n2 = n3 = n4 = n5 = n6 = 1;
  if (ndim>=2) n2=domIn.get_n(2);
  if (ndim>=3) n3=domIn.get_n(3);
  if (ndim>=4) n4=domIn.get_n(4);
  if (ndim>=5) n5=domIn.get_n(5);
  if (ndim>=6) n6=domIn.get_n(6);

//if (getch("tridiag","d",&tridiag)==0) tridiag=1;
//if (getch("scaleup","d",&scaleup)==0) scaleup=0;
//if (getch("diff1","d",&diff1)==0) diff1=0;
//if (getch("diff2","d",&diff2)==0) diff2=0;
//if (getch("agc","d",&agc)==0) agc=0;
//if (getch("absval","d",&absval)==0) absval=agc;
//if (getch("sum3","d",&sum3)==0) sum3=0;
//if (getch("maxmem","d",&maxmem)==0) maxmem=200000000;

	ntotal=n3*n4*n5*n6; n12=n1*n2;

	//memcount = 3*n12*sizeof(float) + n12*sizeof(float)*(2*rect3-1);

	/* allocate working space */
	data.resize(n12);	smoo.resize(n12);
	sum.resize(n12);	temp.resize((2*rect3-1)*n12);

	dist1 = 1.f + .5f * rect1;
	dist2 = 1.f + .6f * rect2;

  /* output processing information */
	//prt("Data smoothed with: rect1","i",&rect1); putch("Data smoothed with: rect2","i",&rect2); prt("Data smoothed with: rect3","i",&rect3);

	b3= 2*rect3-1;
	scale3= 1.f/((float) rect3);

	maxrect=MAX(MAX(rect1,rect2),rect3);

	/* allocate more working space */
  int tridiag = 0;
  
	if (tridiag==0) {
		pp.resize(n12+maxrect  -1);
		qq.resize((n12+2*maxrect-2));
		tt.resize(n12);
		bb.resize(maxrect*(maxrect+n12));
	} else {
		vecin.resize(n12); e.resize(n12);
		vecout.resize(n12);f.resize(n12);
		deni.resize(n12);
	}

  cubIn.resetCounter(); cubOut.resetCounter();
  VERIFY(cubIn.pval != cubOut.pval); //the counter will be screwed up if use inplace.
  //PRV2(cubIn.getSeqCounter(),cubOut.getSeqCounter());
	/* initial temp load */
	nread=0;
	for( i3=0; i3<rect3; i3++ ) {
		if( nread < ntotal ) {    /* ntotal = # slices in data */
      //PRV4(i3,nread,ntotal,rect3);
      cubIn.seq_grab(va_ptr(data),n12); //read in the data;
			nread++;   /* Read in a slice if we're not to data end */
		} else {     /* Set data=temp if off end of data cube */
			for (i2=0;i2<n2;i2++)
				for (i1=0;i1<n1;i1++)
					DATA(i2,i1) = TEMP(n3,i2,i1);
		}
		for (i2=0;i2<n2;i2++) {
			for (i1=0;i1<n1;i1++) {
				TEMP(rect3-1+i3,i2,i1) = DATA(i2,i1);
				if( i3!=rect3-1) 
				  TEMP(rect3-2-i3,i2,i1)= DATA(i2,i1);
			}
		}
	} 

  int sum3 = 0;
	for( i3=0; i3<ntotal; i3++) {
		if( i3!=0 ) {
			/* shift everything along */
			for (j=0;j<b3-1;j++)
				for (i2=0;i2<n2;i2++)
					for (i1=0;i1<n1;i1++)
						TEMP(j,i2,i1) = TEMP(j+1,i2,i1);

			/* load next plane */
			if (nread <ntotal) { 
				cubIn.seq_grab(va_ptr(data),n12);
        nread++; 
			} else {
				for (i2=0;i2<n2;i2++)
					for (i1=0;i1<n1;i1++)
						DATA(i2,i1) = TEMP(nread,i2,i1);
			}
			for (i2=0;i2<n2;i2++)
				for (i1=0;i1<n1;i1++)
					TEMP(b3-1,i2,i1)=DATA(i2,i1);
		}

		/* smooth temp and add into data */
		for (i2=0;i2<n2;i2++)
			for (i1=0;i1<n1;i1++)
				DATA(i2,i1) = TEMP(rect3-1,i2,i1);

		for (j=0;j<rect3-1;j++)
			for (i2=0;i2<n2;i2++)
				for (i1=0;i1<n1;i1++)
					DATA(i2,i1) += ((float) (j+1))/((float) rect3)* 
						(TEMP(j,i2,i1)+TEMP(2*rect3-2-j,i2,i1));

		for (i2=0;i2<n2;i2++)
			for (i1=0;i1<n1;i1++) 
				DATA(i2,i1) *= scale3;

		for (i2=0;i2<n2;i2++)
			for (i1=0;i1<n1;i1++)
				if (bAbsval) SMOO(i2,i1) = fabs( DATA(i2,i1));
				else         SMOO(i2,i1) =      DATA(i2,i1) ;

		for (irepeat=0;irepeat<nrepeat;irepeat++) {
			if (rect1>1  &&  n1>1) {
				for (i2=0;i2<n2;i2++)
					if (tridiag==0) triangle(rect1,1,n1,va_ptr(smoo)+i2*n1,va_ptr(smoo)+i2*n1,pp,qq,tt,bb);
					else            leaky(dist1,1, n1, va_ptr(smoo)+i2*n1, va_ptr(smoo)+i2*n1, vecin, vecout,e, deni,f);
			}

			if (rect2>1  &&  n2>1) {
				for (i1=0;i1<n1;i1++)
					if (tridiag==0) triangle(rect2, n1,n2, va_ptr(smoo)+i1, va_ptr(smoo)+i1,pp,qq,tt,bb);
					else            leaky(   dist2, n1,n2, va_ptr(smoo)+i1, va_ptr(smoo)+i1, vecin, vecout,e, deni,f);
			}
		}

		if (sum3==0) {
//		if (agc > 0) {
//			for (i2=0;i2<n2;i2++)
//				for (i1=0;i1<n1;i1++)
//					if (SMOO(i2,i1) != 0)
//						SMOO(i2,i1) = DATA(i2,i1) / SMOO(i2,i1);
//		}
      cubOut.seq_update(va_ptr(smoo),n12);
		} else {
      for (i2=0;i2<n2;i2++)
				for (i1=0;i1<n1;i1++)
					SUM(i2,i1) += SMOO(i2,i1);
		}
	} //end for i3
	if (sum3>0) { //this branch not used
    cubOut.seq_update(va_ptr(smoo),n12); 
  }
	return 0;
}


