#include "stdafx.h"
#include "param_func.h"
#include "solverNonlinearSearch.h"
#include "cubeData.hpp"
#include "commonOps.h"
#include "opSpline.h"
#include "misc.h"
#include "cubeAlgo2.h"

// TODO: your implementation here
using namespace Yang;
using boost::shared_ptr;

static char szCmd[1024];
int nonLinSearchDirection::searchDirection(gVecComponent& dirOut,const gVecComponent& dir,bool forget){
  if (forget == true) { //if forget, then just use the current search direction
    dirOut = dir;
  } else {
    if (method_ == steepest) {
      beta_ = 0.0f;
    } else if (method_ == FletcherR) {
      beta_ = dotProd(dir,dir)/dotProd(*prevObjNegGradient_,*prevObjNegGradient_);
    } else if (method_ == PolakR) {
      dirOut.scaleAddVec(&dir,prevObjNegGradient_.get(),-1.0f);
      beta_ = (dotProd(dir,dirOut))/dotProd(*prevObjNegGradient_,*prevObjNegGradient_);
    } else { //Hesteness
      dirOut.scaleAddVec(&dir,prevObjNegGradient_.get(),-1.0f);
      beta_ = -dotProd(dir,dirOut)/dotProd(*prevSearchDir_,dirOut);
    }
    VERIFYC(!isnan(beta_));
    if (beta_<0.0f) {
      prt("beta=%f, in nonLinSearchDirection::searchDirection is negative!! Set it to zero\n",beta_);
      beta_ = 0.0f;
    } else if (beta_>1.0f) {
      prt("beta=%f, in nonLinSearchDirection::searchDirection too big! Set it to 1.0\n",beta_);
      beta_ = 1.0f;
    } else {
      prt("beta=%f, in nonLinSearchDirection::searchDirection. \n",beta_);
    }
    //After we get beta
    dirOut.scaleAddVec(&dir,prevSearchDir_.get(),beta_); //s(n) = dir + beta*s(n-1);
  }
  //record the models at this current step.
  *prevSearchDir_ = dirOut; *prevObjNegGradient_ = dir;
  return 0;
}

//given (x,y) 3 pts, fit parabola ax^2+b^x+c
void fitParabola(float coef[3], float x[3], float y[3])
{
   float x1,x2,x3;
   float y1,y2,y3;
   x1 = x[0]; x2=x[1]; x3=x[2];
   y1 = y[0]; y2=y[1]; y3=y[2];
   
   float denom = (x1-x2)*(x1-x3)*(x2-x3);
   coef[0] = (x3*(y2-y1) + x2*(y1-y3) + x1*(y3-y2)) / denom;
   coef[1] = (x3*x3*(y1-y2) + x2*x2*(y3-y1) + x1*x1*(y2-y3)) / denom;
   coef[2] = (x2*x3*(x2-x3)*y1 + x3*x1*(x3-x1)*y2 + x1*x2*(x1-x2)*y3)/denom;
}

// Constructor implementation
solverNonlinearSearch::solverNonlinearSearch(
  const shared_ptr<objFunc>& pObj, const gVecComponent* pInitModel, solverParam* pSolParam, const shared_ptr<nonLinSearchDirection>& directionPredict)
:m_pObjFunc(pObj), m_pSmoother(), m_pSolverPar(pSolParam)
{
  //init the filenames that might be used to store calculated data cubes.
  calDataFns[0] = "mvD0.H";
  calDataFns[1] = "mvD1.H";
  calDataFns[2] = "mvD2.H";

  m_apCurModel.reset( new gVecComponent());
  m_apTrialModel.reset( new gVecComponent());
  
  directionPredictor_ = directionPredict;
  this->init(pObj, pInitModel, pSolParam);
}

void solverNonlinearSearch::init(const shared_ptr<objFunc>& pObj, const gVecComponent* pInitModel, solverParam* pSolParam) {

  m_pObjFunc   = pObj;
  m_pSolverPar = pSolParam;

  //Init the cubeData ptr
  pInitModel->copyTo(*m_apCurModel);
  m_apCurModel->allocMemoryToVec(m_apTrialModel.get());
  
  if (m_pSolverPar->bSmoothSearchDir) {
    switch(m_pSolverPar->eSmooth) {
    case solverParam::sm_spline:
      m_pSmoother.reset(new opSpline());
      break;
    case solverParam::sm_rect:
      m_pSmoother.reset(new smoother_seplib());
      //m_pSolverPar->repeat *= 2; //use Smooth rect is kind of weak.
      //if (m_pSolverPar->repeat != 2) PRV(m_pSolverPar->repeat);
      break;
    case solverParam::sm_alongDip:
      m_pSmoother.reset();
    default:
      VERIFY(0);
   }
  }
    
  bool bSmooth = pSolParam->bSmoothSearchDir;
  if (bSmooth) {
    //create the smoothing operator
    const hypercube& datShape = pInitModel->domain();
    this->setSmoothParam(pSolParam->nSpl);
    
    //Set the scaling quantities, if we change the smoothing over iterations.
    stdVecInt ns_model = datShape.get_ns();
    assert(m_pSolverPar->maxIter != 0);
     
    for (size_t i=0; i<ns_model.size(); i++) {
      m_pSolverPar->curnspl_f[i] = float(m_pSolverPar->nSpl[i]); //init the current number of splines variables.
      VERIFYC(m_pSolverPar->fsSpl[i]>=0.99f);
      VERIFY( m_pSolverPar->nSpl_end[i] < ns_model[i] );
      
      //if (m_pSolverPar->nSpl[i]!=0) {
      //  if (m_pSolverPar->maxIter==1) { m_pSolverPar->fsSpl[i]=1.0f; }//does not matter what the value of fsSpl is
        //else {m_pSolverPar->fsSpl[i] = powf(float(m_pSolverPar->nSpl_end[i])/float(m_pSolverPar->nSpl[i]),1.0f/float(m_pSolverPar->maxIter-1));
      //  else {
          //m_pSolverPar->fsSpl[i] = (float(m_pSolverPar->nSpl_end[i]) - float(m_pSolverPar->nSpl[i])) /float(m_pSolverPar->maxIter-1);
      //  }
      //}
    }
  //PRV(m_pSolverPar->errTol);
  } //end if bSmooth
  PRVEC(m_pSolverPar->fsSpl,3);
  PRV(m_pSolverPar->bMaskGradient);
  if (m_pSolverPar->bMaskGradient) {
    VERIFY(m_pSolverPar->gradMask.domain().sameSize(m_apCurModel->domain()));
  }
}   

// Destructor implementation
solverNonlinearSearch::~solverNonlinearSearch() {
}

//Put in the smooth operator
void solverNonlinearSearch::setSmoothParam(int nSpl[3]) {
  FUNC_TRACE_BEG(solverNonlinearSearch::setSmoothParam);
  assert(m_pSolverPar->bSmoothSearchDir);
  //Get the model space shape
  const hypercube& modelSpaceShape = m_apCurModel->domain();
  int ndim = modelSpaceShape.get_ndim();
  assert(ndim <= 3);
  //now create the spline model shape, only ns matters
  vecAxis axes; axis ax,ax_d; ax.o = 0.0f;
  for (int idim=0; idim<ndim; idim++) {
    ax.n = nSpl[idim];
    axes.push_back(ax);
  }
  hypercube splModShape(axes); 
  m_pSmoother->setSmoothScale(splModShape.get_ns(), modelSpaceShape);
  FUNC_TRACE_END(); 
}

void solverNonlinearSearch::storeDataVec(int ind) {
  if (this->bDataInCore) {
    //in case we resume from a chkpt and without loading the calculated data.
    if (ind==0 && m_pObjFunc->getCalcData().domain().isEmpty()) {
    } else {
      m_pObjFunc->getCalcData().copyTo(calDataCubes[ind]);
    }
  } else {
    string dfn = m_pObjFunc->getCalcData_fn();
    sprintf(szCmd,"Mv %s %s", dfn.c_str(), calDataFns[ind].c_str());
    prt("Exec cmd:%s\n", szCmd);
    VERIFY(system(szCmd)==0);
  }
}

void solverNonlinearSearch::loadDataVec(int ind) {
  if (this->bDataInCore) {
    if (ind==0 && calDataCubes[0].domain().isEmpty()) {
    } else {
      m_pObjFunc->setCalcData(calDataCubes[ind]);
    }
  } else {
    m_pObjFunc->setCalcData_fn(calDataFns[ind]);
  }
}

//Maybe change cubeData into a gVec is better
//return true if there is improvement
bool solverNonlinearSearch::updateModel(const gVecComponent& searchDir, float maxPerturbScale, float curStepSize, float* pNewStepSize) {
  refStepSize_ = curStepSize;

  bool ret = true;
  VERIFY(searchDir.sameSizeVec(m_apCurModel.get()));
  
  //find the scale for search Dir first.
  gVecComponent sdirNorm; searchDir.copyToVec(&sdirNorm);
  sdirNorm.normalize();

  float maxval_searchDir = searchDir.maxvalAbs();
  if (maxval_searchDir < 1e-12) { //mask sure that searchDir is not too small
    searchDir.dump("dump2"); VERIFY(0);
  } 

  float fmax,fmin,fClipRat; //allowed slowness range
  fmax = m_pSolverPar->fmaxValAllow;
  fmin = m_pSolverPar->fminValAllow;
  //PRV2(fmax,fmin);

  int i;
  float stepsizes[3], objFuncVals[3], coefParabola[3];
  float stepRef;
  //the stepsize is 1/2,1,2 times of the allowed model perturbation given maxPerturbScale
  if (equalf(m_RMS_initModel,0.0f)) {
    stepRef = 0.1f; //Just set 0.1f, if does not know the what the scaling should be.
    prt("!error, m_RMS_initModel == 0\n");
  } else {
    stepRef = (maxPerturbScale*m_RMS_initModel); //The RMS of searchDir is already scale to 1
    //PRV2(maxPerturbScale,m_RMS_initModel);
    if(!(stepRef>0.0f)) {
      prt("stepRef close to zero!! %e \n", stepRef);
    }
  }
  
  //multiply stepRef LATER
  stepsizes[0] = 0.0;
  stepsizes[1] = 0.5*refStepSize_;
  stepsizes[2] = 1.0*refStepSize_;
  
  //perturb 2 step size respectively and evaluate the objective functions.
  //build Trial model
  objFuncVals[0] = m_curObjFuncValue;
  
  //Get the current data vector and store it
  storeDataVec(0);
         
  ocDumpF modDump("trial-ms","dump trial models");
  modDump.init(&(m_apCurModel->domain()));
  modDump.dumpFrame(*m_apCurModel); //dump the current model

  for (i=1; i<3; i++)
  {
    m_apTrialModel->scaleAddVec(m_apCurModel.get(),&sdirNorm,stepsizes[i]*stepRef);
    //Only this is not enough, need to clip the values in the new model
    fClipRat = clipCubeVal(*m_apTrialModel,fmax,fmin,true,true); 
    if (fClipRat>0.10f) {
      prt("Clip the trial solution by ratio %f, very big ratio !!!!! \n",fClipRat); exit(-1);
    }
    //evaluate obj function
    modDump.dumpFrame(*m_apTrialModel);
    objFuncVals[i] = m_pObjFunc->evaluate(*m_apTrialModel);

    //store the data so we don't recalculate again, if the data is out-of-core, we can not save it in memory.
    this->storeDataVec(i);
  }

  //optimal among current 3 pts
  float opt3ObjFunc; float opt3Stepsize;
  stdVecFloat vecfval(objFuncVals, objFuncVals+3);
  stdVecFloat vecfstep(stepsizes, stepsizes+3);
  
  stdVecFloat::iterator minIter = std::min_element(vecfval.begin(),vecfval.end());
  int optIndex = minIter - vecfval.begin(); //optIndex is 0~2.
  opt3ObjFunc = vecfval[optIndex];
  opt3Stepsize = vecfstep[optIndex];
  
  //Then fit a parabola among these 3 pts.
  fitParabola(coefParabola,stepsizes,objFuncVals);
  
  float optObjFunc, optStepsize;
  float optParaStepsize = -1.0f, optParaObjFunc = -1.0f;
  //decide the optimal point, according to the shape of the parabola
  //First sets the optimum as the current best.
  optObjFunc = opt3ObjFunc;
  optStepsize = opt3Stepsize;

  bool bTryOneMorePt = true;
  //if the parabola is upside-down or just a line
  if (equalf(coefParabola[0],0.0f) || coefParabola[0] < 0.0f) {
    prt("curvature of parabola is close to zero or pathological !! %f \n", coefParabola[0]);
    if (optIndex==0) { //no improvement at any points
      optParaStepsize = refStepSize_*0.125f; //try a smaller stepsize that is four-times smaller
    } else if (optIndex==1) { //impossible unless three obj func values are the same
      PRT_LOC; SEEVEC(vecfval,3); errPrt("Rare case!! something is wrong");
    } else { //optIndex==2, then double the stepsize, and try this point
      optParaStepsize = refStepSize_*2;
    }
  } else {//bottom-up normal parabola
    //Get the optimal pt, -b/2a
    optParaStepsize = -coefParabola[1]/(2.0f*coefParabola[0]);
    if (optParaStepsize < 0.0f) {//the optimum is negative, we should reduce the step length.
      optParaStepsize = refStepSize_*0.125f; //try a smaller stepsize that is four-times smaller
    } else if (optParaStepsize > 4.0*refStepSize_) { //if the step size is too big, don't go that far.
      optParaStepsize = 4.0*refStepSize_; 
    } else {//otherwise keep as is. 
    }
  }

  //Try the fourth point, see the relative postion with respect to the 3 anchor pts on the parabola, Test the obj func at this pt
  m_apTrialModel->scaleAddVec(m_apCurModel.get(),&sdirNorm,optParaStepsize*stepRef);
  fClipRat = clipCubeVal(*m_apTrialModel,fmax,fmin,true,true);
  //evaulate the obj function at that optimal pt
  optParaObjFunc = m_pObjFunc->evaluate(*m_apTrialModel);
  
  //if parabola bottom is even better, BE AWARE THAT optParaObjFunc could be NAN !
  if (optParaObjFunc < opt3ObjFunc)
  {
    VERIFY(!isnan(optParaObjFunc)); //should not be not-a-number
    optIndex = 3;
    optObjFunc = optParaObjFunc;
    optStepsize = optParaStepsize;
    refStepSize_ = fabsf(optParaStepsize*2); //the refStepSize_should not grow too fast
  } else { //the new trial pt is not better
    if (optIndex==1) { //keep refStepSize unchanged
    } else if (optIndex==2) {
      refStepSize_ *= 2.0f;
    } else { //optIndex==0 <=> no improvement, therefore reduce to its 1/8, since we just tried 1/4.
      refStepSize_ *= 0.125f;
    }
  }
  
  verbPrt("\nstep size: %f,%f,%f,%f, objFunc=%e,%e,%e,%e optimal size %f, funcVal %e\n", stepsizes[0],stepsizes[1],stepsizes[2],optParaStepsize,
          objFuncVals[0],objFuncVals[1],objFuncVals[2], optParaObjFunc, optStepsize, optObjFunc );

  //select a best point among these tested locations
  //if there is improvements
  if (!equalf(optStepsize,0.0f)) {
    if (optStepsize < 0.0f) {
      prt("\noptimum step size is negative !! %f \n", optStepsize);
      VERIFYC(optIndex == 3);
    }
    //update the model using this point
    m_curObjFuncValue = optObjFunc;
    m_apTrialModel->scaleAddVec(m_apCurModel.get(),&sdirNorm,optStepsize*stepRef);
    fClipRat = clipCubeVal(*m_apTrialModel,fmax,fmin,true,true);
    //switch the ptrs
    std::swap(m_apTrialModel,m_apCurModel);
  }
  else {//there is no improvement, search stalls
    ret = false;
  }

  //Set the calc data back to the obj function.
  if (optIndex!=3) {
    this->loadDataVec(optIndex);
  }
  
  if (pNewStepSize != NULL) { //update
    *pNewStepSize = refStepSize_;
  }
  return ret;
}
      
//The non-linear line search optimization work flow
//Given the constrain of the maximum iterations avaible and the relative error difference between iterations
int solverNonlinearSearch::solve(gVecComponent& finalModel, stdVecFloat& 
finalErrorVec, int& finalIter ,ocDumpF* pVecModels,ocDumpF* pVecGrads,ocDumpF* pVecDatas, bool bDataInCore){
  refStepSize_ = 1.0f;
  
  FUNC_TRACE_BEG(solverNonlinearSearch::solve);
  this->bDataInCore = bDataInCore;
  bool bRecordModels =  (pVecModels != NULL);
  bool bRecordGrads  =  (pVecGrads  != NULL);
  bool bRecordDatas  =  (pVecDatas  != NULL);
  
  solverParam* par  = m_pSolverPar;
  float stepSize1,stepSize2;  //keep track of perturbation amount for low-k and high-k iterations separately.
  stepSize1 = stepSize2 = 1.0f;
  float maxPerturb = m_pSolverPar->fMaxPerturbScale;

  int maxIter             = m_pSolverPar->maxIter;
  float errorAllowed      = m_pSolverPar->errTol;
  
  //Init the output params
  finalIter = 0;  finalErrorVec.clear();
   
  //Evaluate the initial model
  objFunc* pObjFunc = this->m_pObjFunc.get();

  //evaluate the initial model
  LOGC("initial evaluate obj");
  float fInitError;
  
  if (m_pSolverPar->rsmStat_ == solverParam::rsm_none) {
    fInitError = pObjFunc->evaluate(*m_apCurModel);
  } else { //resume from a point where the objfunc is given
    fInitError = gParamReader->get_float("rsm_OFval");
  }
  //LOGC("finish initial evaluate obj");
 
  finalErrorVec.push_back(fInitError);
  m_curObjFuncValue = fInitError;
  m_RMS_initModel = m_apCurModel->RMS_NZ(1e-6);
  PRV(m_RMS_initModel);
  
  PRT_LOC; verbPrt("Finish initial evalution of obj func, = %e \n ", fInitError);
  
  //cubeData<float>& dCalInit = pObjFunc->getCalcData();
  //dCalInit.dump("dump6");
  
  //init the dump variables
  const hypercube& shpM = m_apCurModel->domain();
  if (bRecordModels) {pVecModels->init(&shpM);}
  if (bRecordGrads)  {pVecGrads->init(&shpM);}
  bool bDataVecDumpHasBeenInit = false;

  int iter; bool forget;
  float fObjFuncOld, fObjFuncNew;
  gVecComponent searchDir, searchDirFinal, searchDir2;
  this->m_apCurModel->allocMemoryToVec(&searchDir);
  this->m_apCurModel->allocMemoryToVec(&searchDir2);
  this->m_apCurModel->allocMemoryToVec(&searchDirFinal);
  float df,dfmax; dfmax = 0.0f;
  
  for (iter=0; iter<maxIter+1; iter++) {
    if (iter>0 ) {
      //needs to reevaluate obj in order to calculate gradient, we can save this calculation by saving the calc data.
      //pObjFunc->evaluate(*m_apCurModel); //Here we don't need to reevaluate, because we save each of the calculated data during the nonlinear line search step.
      if (this->bDataInCore) {pObjFunc->getCalcData().dump("fImage");}
      pObjFunc->callBackAtIter(iter); //Call back func for objFunc to update info
    }
    
    //If data is out-of-core, it is inconvenient to save the data at each iterations, the final data is also recorded
    if (bRecordDatas) {
      if (iter==0 && m_pSolverPar->rsmStat_ != solverParam::rsm_none) {
        //This case, we did not compute the data in the first place, so do not save the data.
      } else {
        VERIFY(this->bDataInCore);
        if (!bDataVecDumpHasBeenInit) { //Init the data dump
          pVecDatas->init(&(pObjFunc->getCalcData().domain()));
          bDataVecDumpHasBeenInit = true;
        }
        pVecDatas->dumpFrame(pObjFunc->getCalcData());
      }
    }
    if (iter==maxIter) break;
    
    bool bNeedCalcSearchDir = (iter%2==0);
    //calculate the update direction, provide a chkpt scheme
    if (iter==0 && m_pSolverPar->rsmStat_ == solverParam::rsm_grad) {
        //just load the gradient result
        int nd = m_apCurModel->domain().get_ndim();
        PRV(nd); searchDir.load("rsmSrchdir",nd);
    } else if (bNeedCalcSearchDir) {//calculate search direction every other iterations
        pObjFunc->calcSearchDir(searchDir,*m_apCurModel);
    } else { //just reuse the search direction that is given by last time.
    }
    
    //Do normalization here, in case it blows up
    searchDir.normalize();
    //Might need smooth the gradient

    if (bNeedCalcSearchDir) {
      if (m_pSolverPar->bSmoothSearchDir) {
        this->smoothGradient(searchDir, searchDirFinal);
        searchDir.dump("srch"); searchDirFinal.dump("dump1"); 
        LOGC("dump the search directions"); cerr.flush();
        //YPAUSE;
        //m_pSmoother->ModShape().info("\nspl modshape"); m_pSmoother->DatShape().info("\nspl datshape");
      }
    } else {
      //otherwise use the searchDirection used in the previous iteration
      //searchDir.copyToVec(&searchDirFinal);
      rectSmooth(searchDir,searchDirFinal,7,15,1);
    }

    //Mask the gradient if necessary
    if (m_pSolverPar->bMaskGradient) {
      searchDirFinal.multiply(m_pSolverPar->gradMask);
    }
    if (bRecordGrads) {
      pVecGrads->dumpFrame(searchDir); pVecGrads->dumpFrame(searchDirFinal); //dump the raw gradient and smoothed one.
    }

    //searchDir2.dump("dump5"); searchDir.dump("dump6");
    if (m_pSolverPar->bRecalcObjFuncEachIter) { //If needs to recompute the objective function at the begining of each line-search process, otherwise use previous obtained value
      this->m_curObjFuncValue = pObjFunc->evaluate(*m_apCurModel); 
    }
    fObjFuncOld = this->m_curObjFuncValue;
    PRT_LOC; verbPrt("at iter = %d,", iter);
    
    //First calculate the actual search direction
    if (iter==0) {forget = true;}
    else { forget = false;}

    searchDirFinal.copyToVec(&searchDir2);
    directionPredictor_->searchDirection(searchDirFinal,searchDir2,forget);

    //update the model
    bool bImprove = false, bImprovePrev = true;
    //try twice, if not improve, then cannot descend.
    //searchDir2.dump("srchdir",true); //dump serach direction
    LOG("J8");
    float* pStepSize = NULL;

    if (bNeedCalcSearchDir) {
      pStepSize = &stepSize1;
    } else {
      pStepSize = &stepSize2;
    }
    for (int itry=0; itry<1; itry++) { //just try once
      bImprove = this->updateModel(searchDirFinal,maxPerturb, *pStepSize, pStepSize);
      if (bImprove) break;
    } 
    LOG("J9");

    //Now record the updated model
    if (bRecordModels) {pVecModels->dumpFrame(*m_apCurModel);}
    //Here save the two data points that we tried
    if (bRecordDatas) { //bRecordDatas
      if (!bDataVecDumpHasBeenInit) { //Init the shape first
        pVecDatas->init(&(pObjFunc->getCalcData().domain()));
        bDataVecDumpHasBeenInit = true;
      }
      //if (!calDataCubes[0].domain().isEmpty()) {pVecDatas->dumpFrame(calDataCubes[0]);} //trial point 0.0
      pVecDatas->dumpFrame(calDataCubes[1]); //trial point 0.5
      pVecDatas->dumpFrame(calDataCubes[2]); //trial point 1.0
    }

    LOG("J10");
    //Might need to adjust the smoothing operator and the maxStepLength
    //The formula is stepLenMax = (base1 + (1-base1)*(fdecay^(iter)))*fMaxPerturbScale
//    if (par->bDecreaseMaxStepLengthWithIteration) {
//      maxPerturbScale = (par->fbase1+(1.0f-par->fbase1)*powf(par->fdecay,float(iter+1)))*par->fMaxPerturbScale;
        //PRV(maxPerturbScale);
//    }
    //vary the smooth scale every other iteration; && iter%2==1
    if (par->bSmoothSearchDir && par->bChangeSmoothParamWithIteration && iter<maxIter) {
      //Need to refine the spline points.
      int nspls[3] = {0,0,0};
      stdVecInt ns_model = m_apCurModel->domain().get_ns();
      for (size_t i=0; i<ns_model.size(); i++) {
        par->curnspl_f[i] *= (par->fsSpl[i]);
        nspls[i] = MIN(par->nSpl_end[i],int(par->curnspl_f[i]));
      }
      this->setSmoothParam(nspls);
      PRVEC(nspls,3);
    }

    LOG("J11");
    fObjFuncNew = this->m_curObjFuncValue;
    finalErrorVec.push_back(fObjFuncNew);
    
    PRV2(bImprove,bImprovePrev);
    if (!bImprove & !bImprovePrev) {
      verbPrt("\nNo improvement in nonlinear solver!!!\n");break;
    };
    bImprovePrev = bImprove;

    m_apCurModel->dump("curModel"); //dump the current model
    
    //see how much the error has improved.
    VERIFYC(fObjFuncNew <= fObjFuncOld);
    df = fabsf(fObjFuncOld - fObjFuncNew);  dfmax = std::max(df,dfmax);
    if ( fabs(fObjFuncNew) < fabs(fInitError)*errorAllowed) {
      prt("\nThe current Obj func %f is small enough in terms of Initial Error %f, quit the optimization loop at iter=%d \n",
           fabs(fObjFuncNew), fabs(fInitError), iter);
      break;
    }

    if ( fabs(df) < dfmax * 1e-3) {
    //  verbPrt("\ndifference %f is small enough in terms of previous maximum residual reduction %f, quit the optimization loop at iter=%d \n",df, dfmax, iter); break;
    }
    //    if ( fabs(fObjFuncOld - fObjFuncNew) < fabs(fInitError) * errorAllowed) {
//      verbPrt("\ndifference %f is small enough in terms of Initial Error %f, quit the optimization loop at iter=%d \n",fabs(fObjFuncOld - fObjFuncNew), fabs(fInitError), iter); break;
//    }
//    if ( fabs(fObjFuncOld - fObjFuncNew) < fabs(fObjFuncNew) *errorAllowed) {
//      verbPrt("\ndifference %f is small enough in terms of current Error %f, quit the optimization loop at iter=%d \n", fabs(fObjFuncOld - fObjFuncNew), fabs(fObjFuncNew), iter); break; 
//    }
  } //End for iter
  
  finalIter = iter;
  this->m_apCurModel->copyToVec(&finalModel);
  
  //report statistics
  prt("nIter=%d, and the final Error Vec is: \n",finalIter);
  prtStdVec(finalErrorVec);
  
  ////no need to evaluate the final model, the final model and calculated data is been put  in place by updateModel() function.
  
  //clean up
  mCur_nspl.clear();
  int ret = 0;
  if (finalIter==0) {
    //if  && !bImprove) {}
    ret = -1; //can not descend at all
  }
  FUNC_TRACE_END(ret);
}

void solverNonlinearSearch::smoothGradient(const gVecComponent& inputSearchDir, gVecComponent& searchDir) {
  //assert the size match as well.
  CONST_CAST_REF(input,inputSearchDir,gVecComponent);
  m_pSmoother->smooth(input,searchDir,m_pSolverPar->repeat);
}

