#ifndef STDAFX_H
#define STDAFX_H 1

#include "stdafxc.h"

#include <iostream>
//#include <direct.h>
#include <vector>
#include <stdio.h>

//OpenMP support
#include <omp.h>
#include "myOMP.h"

#include <iostream>
#include <valarray>
#include <complex>
using std::cout;
using std::cerr;
using std::endl;
using std::vector;
using std::string;
using std::valarray;

#include <boost/shared_ptr.hpp>
namespace std{
  namespace tr1 = ::boost;
}
using   ::boost::shared_ptr;

typedef std::vector<float> stdVecFloat;
typedef std::vector<int> stdVecInt;
typedef std::vector<double> stdVecDb;
typedef std::vector<float> stdvecFloat;
typedef std::vector<int> stdvecInt;
typedef std::vector<double> stdvecDb;

//the path to store temparay files
extern std::string str_tmpPath;

//Macros

#ifndef SAFE_DELETE
#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#endif    
#ifndef SAFE_DELETE_ARRAY
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }
#endif    
#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p)=NULL; } }
#endif

#define ITER_SOLVER_TOL   (1e-6f)

typedef std::complex<float>   cmplxf;
typedef std::complex<double>  cmplxd;

template <typename T>
inline T* va_ptr(std::valarray<T>& va) {return &(va[0]);}

template <typename T>
void valarray_assign(valarray<T>& array, const T* array_c) {
  for (ulong i=0; i<array.size(); i++) {
    array[i] = array_c[i];
  }
}

template <typename T>
void valarray_assignr(T* array_c, const valarray<T>& array) {
  for (ulong i=0; i<array.size(); i++) {
    array_c[i]  = array[i];
  }
}

//Decide float number equal or not
template <typename T>
bool equal(const T& x, const T& y) 
{
   return (fabs(x-y)<epsilon);
}

bool equalf(const float& x,const float& y);

inline float rad2deg(float rad){
  static float ratio = 180.0f / PI;
  return rad*ratio;
}

inline float deg2rad(float angle){
  static float ratio = PI / 180.0f;
  return angle*ratio;
}

template<class T>
bool inRange(const T& val, const T& minv, const T& maxv) {
  return (maxv >= val && minv <= val);
}

//Taking SQUARES
template <typename T>
T SQ(const T& a) {return a*a;}
template <typename T>
T CB(const T& b) {return b*b*b;}

//Already defined in SEPINC headers
//#define MIN(a,b)  ( (a)<(b)? (a):(b) )
//#define MAX(a,b)  ( (a)>(b)? (a):(b) )

//uncomment this line to use double precision in compuation
//#define USE_DB_PRECISION
 
#ifndef USE_DB_PRECISION
   #define ETYPE float  // or double if you want
#else
   #define ETYPE double
#endif      

//For verbose message
/*#define VERB_PRT(args) {   			\
						if (gVerb)  \//verbose print
						   (printf("VERB: "), printf args); \ 
						}
*/

#define DEBUG
#include "myDebug.h"

#define VERB_PRV(x) \
   { if(gVerb) cerr<<"\n"<<__FILE__<<":"<<__LINE__<<"; "<< #x "= " << x;} 
//Print single variable 

#ifdef DEBUG
   #define DB_PRV(x) {cerr<<__FILE__<<":"<<__LINE__<<"; "<< #x "= " << x << "\n" ;}
#else
   #define DB_PRV(x) {}
#endif

//Define Modeling, RTM, adjRTM operations for GPU engine.
const char* const sep=",";

//For sep_err quick tracef, like verify macro in MFC
#define VERIFY(S) {   \
  if(!bool(S))   \
    {  \
      cerr <<"\n"<<__FILE__<<":"<<__LINE__<<"; "<<"!!(" <<  #S << ") Did not pass verify !!"; \
	    abort(); \
   }} 

//Casting
#define DYN_CAST_PTR(pTarg,pSrc,derClassName) 	derClassName* pTarg = dynamic_cast<derClassName*>(pSrc)
#define DYN_CAST_REF(Targ,Src,derClassName) 	derClassName& Targ = dynamic_cast<derClassName&>(Src)
#define STATIC_CAST_PTR(pTarg,pSrc,derClassName) 	derClassName* pTarg = static_cast<derClassName*>(pSrc)
#define STATIC_CAST_REF(Targ,Src,derClassName) 	derClassName& Targ = static_cast<derClassName&>(Src)
#define CONST_CAST_PTR(pTarg,pSrc,ClassName) 	ClassName*    pTarg = const_cast<ClassName*>(pSrc)
#define CONST_CAST_REF(Targ,Src,ClassName) 	  ClassName& Targ  = const_cast<ClassName&>(Src)

//ptr is the class obj ptr, func is the func name along with the parameters)
//If the ptr is NULL, then do not execute the function
#define SAFE_PTR_CALL(ptr,func) \
  if (NULL!=ptr) { \
    (ptr)->func; \
  }
  

// Maximum 2Gigabytes 
const unsigned int MAX_MEMORY_ALLOC = 2048000000; //~2G


//MY LIB
#include "axis.h"
#include "hypercube.h"

//Functions
//Some function wrappers, it might out of work if ElemCount is too big!!!!!
template<typename T>
inline void myMemcpy(T* dest,const T* src, size_t ElemCount) {
   T* pSrc = const_cast<T*>(src);
   memcpy((void*)(dest),(void*)(pSrc),sizeof(T)*ElemCount);
}
template<typename T>
inline void myMemset(T* dest,T val, size_t ElemCount) {
  long long i;
  for (i=0; i< ElemCount; i++) {
    *(dest+i) = val;
  }
}

template<typename T>
void prtStdVec(const std::vector<T>& vec) {
   int cnt=vec.size();
   int i;
   for (i=0; i<cnt; i++) {
     cerr << "'" << vec[i] << "'";
   }
   cerr << endl;
}

//Calc the statistics of a numeric array.
template<typename T>
void arrayStat(const char* arrName, const T* arr, long len) {
  long index=0, nz=0;
  float fsum, fsqr,f, fmax,fmin;
  fsum= fsqr=0.0f;
  fmax = fmin = arr[0];
  for (index=0; index<len; index++) {
    f     = arr[index];
    fsum += f;
    fsqr += f*f;
    if (fmax < f) fmax=f;
    if (fmin > f) fmin=f;
    if (f!=0.f) {
      nz++;
    }
  }
  float fmean =  fsum/float(len);
  prt("arr stat of %s:size %ld, min:%5.3g max:%5.3g mean=%5.3g, sqsum=%5.3g, Nonzeros=%10li\n", 
         arrName,len,fmin,fmax,fmean,fsqr,nz);
  return;
}

//The other includes
#include "my_types.h"

#endif //include protection macro

