#ifndef STD_AFX__C__H
#define STD_AFX__C__H

#include <stdlib.h>
extern "C" {
#include <seplib.h>
#include <sep3d.h>
}

#define NDEBUG //by default disable assertion
#include <assert.h>

#include <math.h>

#include <string.h>

//MACRO CONSTS
#define MAX_CHAR_BUF 4096

//constants
const double PI=3.14159265358979323846264338327950288419716939937510;
const float epsilon=1e-7f;

//array indexing
#define IND2(i1,i2,n1) ((i2)*(n1)+(i1))
#define IND3(i1,i2,i3,n1,n2) (((i3)*(n2)+(i2))*(n1)+(i1))

extern bool gVerb; //if verbose
extern int gVerbLevel; //if verbose
extern bool gIsLogging; //if enable logging


void verbPrt(const char* form, ...);
void verbPrt(int verbLevel, const char* form, ...);
void prt(const char* form,...);
void dbPrt(const char* format,...);
//print the error and exit
void errPrt(const char* form,...);

inline void arrayStat_c(const char* arrName, const float* arr, long len) {
  long index=0, nz=0;
  float fsum, fsqr,f, fmax,fmin;
  fsum= fsqr=0.0f;
  fmax = fmin = arr[0];
  for (index=0; index<len; index++) {
    f     = arr[index];
    fsum += f;
    fsqr += f*f;
    if (fmax < f) fmax=f;
    if (fmin > f) fmin=f;
    if (fabs(f) > epsilon) {
      nz++;
    }
  }
  float fmean =  fsum/float(len);
  prt("arr stat of %s:size %ld, min:%5.3g max:%5.3g mean=%5.3g, sqsum=%5.3g, Nonzeros=%10li\n", 
         arrName,len,fmin,fmax,fmean,fsqr,nz);
  return;
}

#define LOGC(word) \
	if (gIsLogging) \
		{fprintf(stderr,"\n%s:%d;  %s" ,__FILE__, __LINE__,#word);}


#define VERIFYC(S) {   \
  if(!(S))   \
   {  \
   fprintf(stderr,"\n%s:%d;!! %s Did not pass verify",__FILE__,__LINE__,#S); \
	  abort();\
	  exit(0); \
   }}
    
#endif
