#include "stdafx.h"
#include "cubeData.hpp"
#include "point3d.hpp"
#include "sep_params.h"
#include "cubeAlgo.h"
#include "cubeAlgo2.h"

using namespace Yang;
using namespace std;

   
//Apply Jon's Marine data gain t^2/t0 to the data

int main (int argc, char *argv[])
{
  sep_params* pars = new sep_params(argc,argv);
  cubeData<float> inc,outc,t0c;
  int nt,ntrc;
  
  //read the input
  inc.load("in");
  hypercube shpIn = inc.domain();
  nt    = shpIn.get_n(1);
  axis ax_t = shpIn.get_axis(1);
  ntrc  = shpIn.get_n23();
  
  t0c.alloc(hypercube(ntrc));
  
  bool verb = pars->get_bool("verb");
  float scThres = pars->get_float("thres_sc",10.0f);
  if (verb) {PRV2(nt,ntrc);}
  
  //First find all t0, calculate each trace's RMS, search for the first peak, and then take it as the t0
  int ir=0, nCompWindow = 10; //find the maximum that among the window of (2*10+1) samples
  hypercube shpnt(nt);
  cubeData<float> trcc(shpnt);
  
  float rmsnz,thres,it;
  stdVecFloat inCoord,outCoord, inVec, outVec;
  rmsnz = inc.RMS_NZ(1e-5);
  thres = rmsnz * scThres;
  PRV(thres);
  for (ir=0; ir<ntrc; ir++) {
    inc.seq_grab(&trcc);
    
    //search the maximum as t increases
    float fmax; int curmax;
    curmax = 0; fmax = 0.0f;
    bool bReachThres = false;
    for (it=0; it<nt; it++) {
      if (fmax < fabs(trcc(it))) {
        fmax = fabs(trcc(it));
        curmax = it;
        if (fmax > thres) {
          bReachThres = true;
          break;
        }
      }
    }
    if (bReachThres) {
      t0c(ir) = ax_t.get_coord(curmax);
      inCoord.push_back(float(ir));   inVec.push_back(t0c(ir));
    } else {
      t0c(ir) = -1.0f; //means could not find any meaningful value
      outCoord.push_back(float(ir));  outVec.push_back(0.0f);
    }
  }
  
  t0c.dump("t0_pre");
  
  VERIFY(inCoord.size() > ntrc/2);
  
  //Now do some interpolation of t0 arrays for places where it cannot find, here just use std input
  linInterp1d(&outVec[0],outCoord,&inVec[0],inCoord,true);
  
  //Now apply the gain
  int ind = 0;
  for (ind=0; ind<outCoord.size(); ind++) {
    t0c[int(outCoord[ind]+0.5f)] = outVec[ind];
  }
  
  outc.alloc(shpIn);
  float t0,t; int it0; float *pIn,*pOut;
  for (ir=0; ir<ntrc; ir++) {
    t0 = t0c[ir];
    pIn  = inc.pval+ir*nt;
    pOut = outc.pval+ir*nt;
    it0 = ax_t.get_offset(t0);
    //if smaller than t0, then apply the gain t
    for (int it=0; it<std::min(nt,it0+1); it++) {
      pOut[it] = ax_t.get_coord(it) * pIn[it];
    }
    //apply the gain t^2/t0 
    for (int it=it0;it<nt; it++) {
      t = ax_t.get_coord(it);
      pOut[it] = t*t/t0*pIn[it];
    }
  }
  
  outc.dump("out"); t0c.dump("t0"); 
  return 0;
}
