#include "stdafx.h"
#include "cubeData.hpp"
#include "point3d.hpp"
#include "myOMP.h"
#include "stackTraceC.h"
#include "../objFunc.h"
#include "commonOps.h"
#include "../solverNonlinearSearch.h"
#include "sep_params.h"

using namespace Yang;
using namespace std;

   
//Test the non linear line search of non-lin solver.
int main (int argc, char *argv[])
{
  initCallStackTrace();
  
  sep_params* pars = new sep_params(argc,argv);
  opMatrixMultiply op;
  cubeData<float> matr,dat;
  
  int m,n;
  m=2; n=2;
  hypercube shape(n,m);
  //matrix is nXm
  matr.alloc(shape);
  matr(0,0) = 2.0f;
  matr(0,1) = 1.5f;
  //matr(0,2) = 2.0f;
  matr(1,0) = 3.5f;
  matr(1,1) = 3.0f;
  //matr(1,2) = 2.0f;

  op.init(matr);
  
  dat.alloc(hypercube(n));
  dat(0) = 5.5;
  dat(1) = 4.5;
  
  objFunc* pObjFunc = new simpleObjFunc(&op,dat);

  gVecComponent initModelVec, finalModelVec;
  initModelVec.alloc(hypercube(m));
  initModelVec(0) = 2.0f;
  initModelVec(1) = 0.0f;
  initModelVec.allocMemoryToVec(&finalModelVec);
  
  solverNonlinearSearch nonLinSolver(pObjFunc, &initModelVec);
   
  stdVecFloat finalErrorVec;
  int finalIter;
  int maxIter = pars->get_int("maxNonLinIter",5);
  float errorAllowed = pars->get_float("errorAllowed",1e-6);
  float maxPerturbScale = pars->get_float("maxPerturbScale",0.1); //5 percent of the slowness RMS value
  
  vector<gVecComponent*> modelsVec;
  nonLinSolver.solve( finalModelVec, finalErrorVec, finalIter, 
              maxIter, errorAllowed, maxPerturbScale, &modelsVec);
  
  //Save result
  finalModelVec.dump("out");
  //finalModelVec.dump("out");
  finalModelVec.unInit();
  
  int iSize = finalErrorVec.size();
  cubeData<float> finalErrCube(iSize);
  myMemcpy(finalErrCube.pval,&finalErrorVec[0],iSize);
  finalErrCube.dump("error");
  
  //Dump the final data
  cubeData<float>& finalImg = pObjFunc->getCalcData();
  
  dumpVecModels(modelsVec,"models");
  deleteVecPointers(modelsVec);
  SAFE_DELETE(pObjFunc);
  return 0;
}
