#include "stdafx.h"
#include "sep_params.h"
#include "cubeData.hpp"

#include "commonOps.h"
#include "opSpline.h"
#include "boost/shared_ptr.hpp"

using namespace Yang;
using namespace std;

//Test for Bspline smoothing.
int main( int argc, char** argv){
  
   //Read params
   boost::shared_ptr<sep_params> pars(new sep_params(argc,argv));
   
   smoother_intf* pSmooth = NULL;
   switch (pars->get_int("method",0)) {
   case 0: //use seplib smooth function
     pSmooth = new smoother_seplib();
     break;
   case 1:
     pSmooth = new opSpline();
     break; //use B spline nodes;
   default:
     errPrt("Invalid method\n");
   }
   smoother_intf& smoother = *pSmooth;
   hypercube modShape;
   hypercube datShape;
   
   //Read in a .H file and smooth it
   cubeData<float> in("in",dm::IN,"input");
   in.init(NULL,true);
   datShape = in.domain();
   
   datShape.info("dat Shape");
   //Read in the desired mod shape
   int ndim = datShape.get_ndim();
   VERIFY(ndim<=3);
   int nm[3] = {0,0,0};
   int idim;
   for (idim=0; idim < ndim; idim++) {
     nm[idim] = datShape.get_n(idim+1);
   }
   
   nm[0] = pars->get_int("n1",nm[0]);
   nm[1] = pars->get_int("n2",nm[1]);
   nm[2] = pars->get_int("n3",nm[2]);
   int repeat = pars->get_int("repeat",1);
   
   //Now the d's
   vecAxis axes; axis ax,ax_d; ax.o = 0.0f;
   for (idim=0; idim<ndim; idim++) {
     ax_d = datShape.get_axis(idim+1);
     ax.n = nm[idim];
     ax.d = ((ax_d.n-1)*ax_d.d)/(ax.n-1);
     
     axes.push_back(ax);
   }

   modShape.set_axes(axes);
   
   smoother.setSmoothScale(modShape.get_ns(), datShape);
   
   LOGC(ts1);
   cubeData<float> outCube;
   outCube.init(&datShape);

   
   smoother.smooth(in,outCube,repeat);
   PRV(outCube.RMS());
   outCube.dump("out");
   return 0;
}


