#include "stdafx.h"
#include "sep_params.h"
#include "cubeData.hpp"
#include "stackTraceC.h"

#include <fftw3.h>
#include <iostream>
#include <complex>

using namespace Yang;
using namespace std;

enum FFTtype {
   r2c, //forward
   c2r, //inverse, is Adjoint as well.
   c2c
};

//In all cases, N intput, N output. There will be redundancy in r2c and c2r cases.
///Fourier transform using FFTW libary
int main(int argc, char* argv[]) {
    initCallStackTrace();
	//have to init par
	sep_params* pars = new sep_params(argc,argv);
    
    //First get the FFT parameters.
    
    string op = pars->get_string("op","r2c");
    FFTtype type = r2c;
    //See if use radian unit rather than degree.
    bool bRadian = pars->get_bool("bRadian",false);
    
    hypercube inShape,outShape;
    bool bRemoveEndDummyAxis = true;
    int N = pars->get_int("N",0); //output length
    bool bPad = pars->get_bool("bPad",false);
    
    //Need a SEP header file reader. But this reader does not modify the binaries.
    //Just use 
    sepHeaderReader headReader("in",dm::IN,"input");
    headReader.getShape(inShape,bRemoveEndDummyAxis);
    inShape.info("inShape");
    
    int nt = inShape.get_n(1);
    int npad = 0;
    //Pad to the nearest power of 2
    if(bPad) {
       int npow = int(ceilf(log2(nt)));
       npad = int(powf(2,npow));
       prt("npad=%d\n", npad);
    }
    //Finally N is determined
    N = MAX(nt,N); N=MAX(N,npad);
    //N is the actually length for FFT IN/OUTPUT
    
    //calc the output memory size
    int nTrc = inShape.get_n23();
    vecAxis outAxes = inShape.get_axes();
    axis ax_t,ax_f;
    
    ax_t = outAxes[0]; //it might be the freq axis, depends on the op direction
    VERIFY(equalf(ax_t.o,0.0f));
    
    ax_f.o = 0.0f;
    ax_f.n = N; //usually we just need half the spectrum for real signal, e.g from 0 to pi, 
                //but here we get the entire spectrum
    if (0==op.compare("r2c")) {
       // df is defined by Nyquist Frequency 2*f_nyquist = 1/(N*dt)
       if(!bRadian) {
          ax_f.d = 2.0f/(2.0f * N * ax_t.d);
          ax_f.label = "Hz";
       } else {
          ax_f.d = 2.0f/(2.0f * N * ax_t.d) * 2.0f *PI;
          ax_f.label = "rad/s";
       }
       
       outAxes[0] = ax_f;
       outShape.set_axes(outAxes);
       outShape.info("outShape");
       cubeData<float> in("in",dm::IN,"input");
       cubeData<std::complex<float> > out("output",dm::OUT,"output");
       in.init(NULL, bRemoveEndDummyAxis);
       out.init(&outShape);
       //Perform FFT on the First axis
       const float *pIn, *pi;
       std::complex<float>* pOut, *po;
       pIn = in.dataPtr(); pOut = out.dataPtr();
   
       
       //Basically FFTWf take float precision, default fftw takes double precision
       float*  fInTrc = (float*)malloc(sizeof(float)*N);
       
       fftwf_complex*  fOutTrc= (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex)*(N/2+1));
       fftwf_plan   FFTplan;
       FFTplan = fftwf_plan_dft_r2c_1d(N, fInTrc, fOutTrc, FFTW_MEASURE);
   
       VERIFY(nTrc > 0);
       for (long long i = 0; i<nTrc; i++) {
          pi = pIn + i*nt; po = pOut + i*N;
   
          //Fill in the data to FFT module
          for (int j=0; j<nt; j++) {
             fInTrc[j] = pi[j];
          }
          for (int j=nt; j<N; j++) {
             fInTrc[j] = 0.0f; //just pad zero
          }
          
          fftwf_execute(FFTplan);
          
          //put the output back to float
          for (int j=0; j<(N/2+1); j++) {
             po[j] = std::complex<float>(fOutTrc[j][0],fOutTrc[j][1]);
          }
          //From the symmetricity, P(j) = conj(P(N-j));
          for (int j=(N/2+1); j<N; j++) {
             po[j] = std::complex<float>(fOutTrc[N-j][0],-1.0f*fOutTrc[N-j][1]);
          }
       }
       
       //Save to output and cleaned up
       out.unInit();
       fftwf_destroy_plan(FFTplan);
       fftwf_free(fOutTrc);
       free(fInTrc);
    } else if (0==op.compare("c2r")) {
       if(!bRadian) {
          ax_f.d = 2.0f/(2.0f * N * ax_t.d);
       } else {
          ax_f.d = 2.0f/(2.0f * N * ax_t.d) * 2.0f *PI;
       }
       
       ax_f.label = "sec";
       outAxes[0] = ax_f;
       outShape.set_axes(outAxes);
       outShape.info("outShape");
       cubeData<std::complex<float> > in("in",dm::IN,"input");
       cubeData<float> out("output",dm::OUT,"output");
       in.init(NULL, bRemoveEndDummyAxis);
       out.init(&outShape);
       //Perform FFT on the First axis
       const std::complex<float>  *pIn, *pi;
       float *pOut, *po;
       pIn = in.dataPtr(); pOut = out.dataPtr();
   
       
       //Basically fftwf take float precision, default fftw takes double precision
       std::complex<float>*  fInTrc = (std::complex<float>*)malloc(sizeof(std::complex<float>)*(N/2+1));
       float*  fOutTrc= (float*)malloc(sizeof(float)*N);
       fftwf_plan   FFTplan;
       FFTplan = fftwf_plan_dft_c2r_1d(N, (float(*)[2])fInTrc, fOutTrc, FFTW_MEASURE);
   
       VERIFY(nTrc > 0);
       int nInput = MIN(N/2+1,nt); //Only need at most N/2+1 input
       for (long long i = 0; i<nTrc; i++) { 
          pi = pIn + i*nt; po = pOut + i*N;
   
          //Fill in the data to the input module. 
          for (int j=0; j<nInput; j++) {
             fInTrc[j] = pi[j];
          }
          for (int j=nInput; j<(N/2+1); j++) {
             fInTrc[j] = std::complex<float>(0.0f,0.0f); //just pad zero
          }
          
          fftwf_execute(FFTplan);
          
          //put the output back to float
          for (int j=0; j<N; j++) {
             po[j] = fOutTrc[j];
          }
       }
       
       //Save to output and cleaned up
       out.unInit();
       fftwf_destroy_plan(FFTplan);
       free(fOutTrc);
       free(fInTrc);
    } else if (0==op.compare("c2c")) {
       VERIFY(0);
    } else {
       seperr("\n unknown op type !");
    }
    return 0;
}



