#include "stdafx.h"
#include "sep_params.h"
#include "cubeData.hpp"
#include "myOMP.h"

#include "imaging.h"
using namespace Yang;
using namespace std;

#define MY_SEP_DOC \
 sep_add_doc_line("NAME");\
 sep_add_doc_line("    Off2Slope - Transform between suboffset image gathers and angle gather(tan(gamma))");\
 sep_add_doc_line("");\
 sep_add_doc_line("SYNOPSIS");\
 sep_add_doc_line("    Off2Slope < in.H o1 pars > out.H");\
 sep_add_doc_line("    Input is arranged as (z,hx,x) or (z,px,x), same as output");\
 sep_add_doc_line("    px=tan(gamma) ");\
 sep_add_doc_line("");\
 sep_add_doc_line("INPUT PARAMETERS");\
 sep_add_doc_line("    adj - bool");\
 sep_add_doc_line("                [0]: if 1, transform ang to h ");\
 sep_add_doc_line("");\
 sep_add_doc_line("    amax - float");\
 sep_add_doc_line("                 [-60deg]: maximum angle in degree, for forward transform, the program will convert it to slope automatically ");\
 sep_add_doc_line("");\
 sep_add_doc_line("    amin - float");\
 sep_add_doc_line("                [-amax]: minimum angle in degree");\
 sep_add_doc_line("");\
 sep_add_doc_line("    hmax - float");\
 sep_add_doc_line("                []; maximum offset, user need to specify, for adj transform"); \
 sep_add_doc_line("");\
 sep_add_doc_line("    hmin - float");\
 sep_add_doc_line("                [-hmin]: minimum offset");\
 sep_add_doc_line("");\
 sep_add_doc_line("    n - int");\
 sep_add_doc_line("                [in.n]: number of pts on the angle(offset) axis for fwd(adj) transform ");\
 sep_add_doc_line("");

//Do offset to slope(angle) tranform, similar as OFF2ANG
//the axis conf (z,hx,x) <-> (z,px,x)

int main( int argc, char** argv){
   //Read params
   auto_ptr<sep_params> pars(new sep_params(argc,argv));
   
   //in is the input data
   cubeDataF input("in",dm::IN,"model");
  	init_omp(2); 
   input.init(NULL,true);
   vecAxis inAxes = input.domain().get_axes();
   const int ndim = inAxes.size();
   vecAxis newInAxes(inAxes);
   if (ndim<3) { //if input 2D then put a dummy impulse there
     newInAxes.push_back(axis(1,0.f,1.f,"x","kft"));
     input.reshape(newInAxes);
   } else if (inAxes.size()>3) {
     //lump the rest of axis together
     axis ax_3; ax_3 = inAxes[2];
     ax_3.n = input.domain().get_n23(3,ndim);
     newInAxes.erase(newInAxes.begin()+2,newInAxes.end());
     newInAxes.push_back(ax_3);
   }
   //input should have z,hx,x,[y]
   VERIFY(newInAxes.size()==3);
   hypercube inShape(newInAxes);
	
	inShape.info("inShape new");
	input.reshape(inShape);

   axis ax_x = inShape.get_axis(3);
   axis ax_z = inShape.get_axis(1);
   axis ax_a,ax_h;
   //ax_p.label = string("p,tan(gamma)");
   ax_h.label = string("subOffset");
   ax_a.label = string("deg");
   vecAxis outAxes;
   outAxes.push_back(ax_z); outAxes.push_back(ax_x);
   hypercube outShape;
   
   //Get the params
   bool bAdj = pars->get_bool("adj",false);
   int n = pars->get_int("n",inShape.get_axis(2).n);
   VERIFY(n>1);

   if (!bAdj) { //Do off to ang
     LOG(Doing off 2 slope);
     float amax = pars->get_float("amax",60.0);
     float amin = pars->get_float("amin",-amax);
     ax_a.o = amin; ax_a.n = n; ax_a.d = (amax-amin)/(n-1);
   } else { //slope to off
     LOG(Doing slope 2 off);
     float hmax = pars->get_float("hmax");
     float hmin = pars->get_float("hmin",-hmax);
     VERIFY(hmin<hmax && hmax>0.0f);
     ax_h.o = hmin; ax_h.d = (hmax-hmin)/(n-1); ax_h.n = n;
   }
   
   //Do the transform
   outShape.info("output shape");
   bool bAntialias = pars->get_bool("b_AA",true);
   float fAntialias = pars->get_float("f_AA",1.0f);
   PRV(bAntialias); PRV(fAntialias);
   cubeDataF cbOut;
   if (!bAdj) {
     Yang::off2Ang2d_new(cbOut, input, false, slantstackYangAA, ax_a,
                       fAntialias);
   } else {
     Yang::off2Ang2d_new(cbOut, input, true, slantstackYangAA, ax_h,
                       fAntialias);
   }
   
   //now reshape the outcube.
   const hypercube& shpO = cbOut.domain();
   vecAxis newOutAxes;
   if (ndim>3) { //we need to split out the last axis of shp0
     newOutAxes.push_back(shpO.get_axis(1));
     newOutAxes.push_back(shpO.get_axis(2));
     newOutAxes.insert(newOutAxes.end(),inAxes.begin()+2,inAxes.end());
   }
   hypercube newOutShp(newOutAxes);
   newOutShp.info("new Out Shp");
   cbOut.reshape(newOutShp);
   cbOut.dump("out");
}

