#include "stdafx.h"
#include "stackTraceC.h"
#include "cubeData.hpp"
#include "off2ang3d.h"
#include "sep_params.h"
#include "off2ang3d.param.h"

/**
 * Main application entry point
 * 
 * Off2ang 3d tool, algorithm based on Biondi and Tisserant's SEP report, SEP113. etc.
 * Format:
   Off2ang3dB <in.H >out.H jac_gamma=y/n jac_beta=y/n [params]
   Other params pls see the read_param_off2ang3d function.
 */

using namespace Yang;
int main (int argc, char *argv[])
{
  // TODO: Add application code here
  initCallStackTrace();
  init_omp(100); //only use one thread.
  
  //have to init par
	sep_params* pars = new sep_params(argc,argv);
  string szin = pars->get_string("input","");
  string szout = pars->get_string("output","");
  cerr << "szin:szout; " << szin << " : " << szout << endl;

  //Read in/out tag
  bool bNotDummy = true;;
  dmRsf_oc<cmplxf> dmIn("in",Yang::dm::IN,"input",bNotDummy);  dmRsf_oc<cmplxf> dmOut("out",Yang::dm::OUT,"output",bNotDummy);
  dmRsf<cmplxf>& ddmIn = dmIn; dmRsf<cmplxf>& ddmOut=dmOut;
  dmIn.init(NULL);
  hypercube inShape = dmIn.domain();
  //read in params
  param_off2ang3d par_op;
  read_param_off2ang3d(&par_op, pars, inShape, hypercube(1)); //the 4th argument is dummy in this case (only in this case).
  //set the output shape
  dmOut.init(&par_op.shapeAng);

  //Apply the transform
  off2ang3d(par_op,dmIn,dmOut);
  
  //// Test
//cubeData<cmplxf> cube(hypercube(par_op.shapeOut.get_n(1),par_op.shapeOut.get_n(2)));
//for (int i=0; i<par_op.shapeOut.get_n(1)*par_op.shapeOut.get_n(2); i++) {
//  cube[i] = float(rand())/RAND_MAX;
//}
//ddmOut.seq_update(&cube);
  
  //finish
  ddmIn.unInit();
  ddmOut.unInit(); 
  delete pars;
  return 0;
}

