#include "stdafx.h"
#include "stackTraceC.h"
#include "cubeData.hpp"
#include "off2ang3d.h"
#include "sep_params.h"
#include "off2ang3d.param.h"

/**
 * Main application entry point
 * 
 * Off2ang 3d tool, algorithm based on Biondi and Tisserant's SEP report, SEP113. etc.
 * Format:
   Off2ang3dB input=in.H output=out.H jac_gamma=y/n jac_beta=y/n [params]
   Other params pls see the read_param_off2ang3d function.
   
  input/outupt should be in axis (z,x,y,hx,hy)/(z,gamma,azim,x,y)
 */

using namespace Yang;

void convert_axis_to_fourier(axis& ax, const axis& ax_in,int new_n) {
  char sztmp[512];
  sprintf(sztmp,"1/%s",ax_in.label.c_str()); ax.label  = sztmp;
  sprintf(sztmp,"1/%s",ax_in.unit.c_str());  ax.unit   = sztmp;
  ax.o = 0.0f; ax.n = new_n; ax.d = 1.f/(ax.n*ax_in.d);
}

int main (int argc, char *argv[])
{
  // TODO: Add application code here
  initCallStackTrace();
  init_omp(100); //only use one thread.
  
  srand(time(NULL));
  
  //have to init par
	sep_params* pars = new sep_params(argc,argv);
  string szin = pars->get_string("input","");
  string szout = pars->get_string("output","");
  cerr << "szin:szout; " << szin << " : " << szout << endl;

  //Read input/output tag
  bool bNotDummy = true;;
  dmRsf_oc<float> dmIn("input",Yang::dm::IN,"input",bNotDummy);  dmRsf_oc<float> dmOut("output",Yang::dm::OUT,"output",bNotDummy);
  dmIn.init(NULL); //acquire its shape
  hypercube inShape = dmIn.domain();
  dmIn.unInit();
  //read in params
  param_off2ang3d par_op;
  bool boff2ang = pars->get_bool("bforward", true);
  hypercube shapezxy;
  if (boff2ang) {
    shapezxy = inShape.return_compatible(1,3); 
  } else {
    int ia[] = {1,4,5};
    stdVecInt nv(ia,ia+3);
    shapezxy = inShape.return_arbAxesShape(nv);
  }
  shapezxy.info("shape zxy:");
  //also we need to read in the kx,ky,kz numbers, and turn the x,y,z axis in inShape to kx,ky,kz.
  int nkx = pars->get_int("nkx");
  int nky = pars->get_int("nky");
  int nkz = pars->get_int("nkz");
  axis ax_kx,ax_ky,ax_kz;
  if (boff2ang) {
    ax_kz = inShape.get_axis(1);
    ax_kx = inShape.get_axis(2);
    ax_ky = inShape.get_axis(3);
  } else {
    ax_kz = inShape.get_axis(1);
    ax_kx = inShape.get_axis(4);
    ax_ky = inShape.get_axis(5);
  }
  VERIFY(ax_kx.n<=nkx);
  VERIFY(ax_ky.n<=nky);
  convert_axis_to_fourier(ax_kz,ax_kz, nkz);
  convert_axis_to_fourier(ax_kx,ax_kx, nkx);
  convert_axis_to_fourier(ax_ky,ax_ky, nky);
  
  //inShape is either (hx,hy,kz,kx,ky) or (gamma,azim,kz,kx,ky)
  if (boff2ang) {
    axis ax_hx = inShape.get_axis(4); axis ax_hy = inShape.get_axis(5);
    inShape.set_axis(1,ax_hx);  inShape.set_axis(2,ax_hy);
  } else {
    axis ax_gamma = inShape.get_axis(2); axis ax_azim = inShape.get_axis(3);
    inShape.set_axis(1,ax_gamma);  inShape.set_axis(2,ax_azim);
  }
  inShape.set_axis(3,ax_kz);
  inShape.set_axis(4,ax_kx);
  inShape.set_axis(5,ax_ky);
    
  read_param_off2ang3d(&par_op, pars, inShape, shapezxy, boff2ang); //the 4th argument is dummy in this case (only in this case).
  //Apply the transform
  if (boff2ang) {
    off2ang3d_file(par_op,szin,szout);
  } else {
    ang2off3d_file(par_op,szin,szout);
  }
  
  //// Test
//cubeData<cmplxf> cube(hypercube(par_op.shapeOut.get_n(1),par_op.shapeOut.get_n(2)));
//for (int i=0; i<par_op.shapeOut.get_n(1)*par_op.shapeOut.get_n(2); i++) {
//  cube[i] = float(rand())/RAND_MAX;
//}
//ddmOut.seq_update(&cube);
  
  //finish
  delete pars;
  return 0;
}
  
