#include "stdafx.h"
#include "sep_params.h"
#include "cubeData.hpp"
#include "cubeAlgo.h"
#include "dampingFunc.h"
using namespace Yang;
using namespace std;

///Taper the input data set on the edge, assume the 3rd axis is the
///shot axis, and the shot axis gives the shot location.

int main( int argc, char** argv){
   
   //Read params
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   
   //in is the input data
   //out is the Tapered data
   cubeDataF input("in",dm::IN,"model");
   cubeDataF output("out",dm::OUT,"data");
   
   input.init(NULL);
   output.init(&(input.domain()));
   const hypercube& dataShape = input.domain();

   axis ax_shot = dataShape.get_axis(3);
   hypercube sliceShape = dataShape.return_compatible(1,2);
   
   cubeData<float> slice(sliceShape);
   cubeData<float> trcDampingCube;
   int nShot = ax_shot.n;
   cosDampingFunc dampFunc;
   stdVecFloat dampLeft, dampRight;
   float shotPos; axis ax_recv = dataShape.get_axis(2);
   trcDampingCube.alloc(ax_recv.n);
   float d = ax_recv.d; int n = ax_recv.n;

   //The length at which it will taper to zerol
   int taperLength = autoPars->get_int("taperLength",n);
   
   dampFunc.getCoefEx(dampRight,taperLength,1.0);
   if (taperLength<n) {
      dampRight.insert(dampRight.end(),n-taperLength,0.0f); //Pad zero if not long enough.
   }
   
   for (int i=0; i<nShot; i++) {
      shotPos = i*ax_shot.d + ax_shot.o;
      //the distance from two ends.
      float dis1,dis2; int iDis1,iDis2;
      dis1 = shotPos - ax_recv.o;
      dis2 = ax_recv.o + (ax_recv.n-1)*d;
      iDis1 = lroundf(dis1/d); iDis2 = lround(dis2/d);
      //Damp based on the distance, the further the more damping.
      int iShotPos = ax_recv.get_offset(shotPos);
      for (int j=0; j<iShotPos; j++) {
         trcDampingCube[j] = dampRight[iShotPos-j-1];
      }
      trcDampingCube[iShotPos] = 1.0f;
      for (int j=iShotPos+1; j<n; j++) {
         trcDampingCube[j] = dampRight[j-iShotPos-1];
      }
      input.seq_grab(&slice);
      scaleTrc(slice,trcDampingCube);
      output.seq_update(&slice);
   }
   input.unInit();
   output.unInit();
}

