///Dip estimation of a 2D or 3D cube.
///Use the structure-tensor method, which finds the eigen vector of the ST matrix
///Yang Zhang

#include "stdafx.h"
#include "eig3.h"

using namespace std;
int main(int argc, char* argv[])
{
  //input a 2D/3D cube
  sf_init(argc,argv);

  iRSF par(0);
  CUB inCub("in","i"); in.headin();

  inCub.report();
  //Check data type
  assert(sizeof(float) == inCub.esize);

  int ndim = inCub.nd;
  assert (ndim == 2 || ndim == 3 );
  cerr << "The dimension of the input is " << ndim << ".\n";

  //Out put will be a (2D or 3D) vector at each grid point. and that is
  //put to the FIRST dimension

  //Get axis information, notice the grid sampling for axes might be different
  //we should consider the highest possible dip given the grid sampling.
  //assume the cube is (z,x,y)
  sf_axis az,ax,ay, aD;
  aD.o = 1; aD.d = 1; aD.n = ndim;
  az = inCub.getax(0); ax = inCub.getax(1); ay.n = 1;

  CUB outCub("out","o"); outCub.setup(ndim+1,inCub.esize());
  outCub.putax(0,az); outCub.putax(1,ax);
  if (nidm == 2) {
      outCube.putax(2,aD);
  } else { //ndim == 3
    ay = inCub.getax(2);
    outCub.putax(2,ay);
    outCub.putax(3,aD);
  }
  outCub.headou();

  //Read in parameters
  float fSmth = par.get("fsmth",0.0f); //the half amplitude width of the gaussian for smth
  float fSmth1, fSmth2, fSmth3;
  fSmth1 = fSmth2 = fSmth3  = fSmth;
  bool bVerb = par.get("verb",false);

  ///Read the actually cube
  int n1,n2,n3; int i1,i2,i3;
  n1 = az.n; n2=ax.n; n3=ay.n;
  valarray<float> inArr(n1*n2*n3);
  int n4=1;
  if (ndim == 2) { n4 = 3; }//I11,I12,I22
  else  { n4=6; } //I11,I12,I13,I22,I23,I33
  valarray<float> outArr(n1*n2*n3*n4); 
  valarray<float> g,Izz,Ixx,Izx, Izzs, Ixxs, Izxs;
  
  int n123 = n1 * n2 * n3;
  g.resize(n123*ndim); Izz.resize(n123); Ixx.resize(n123); Izx.resize(n123);
  
  VAI indIn(n1,n2,n3), indOut(n1,n2,n3,ndim), indMat(n1,n2,n3,n4);

  //Read in the input cube, supposedly an image
  inCub >> inArr;
  //Apply gradient operator and then generate g*transp(g)
  
  if (ndim==2) {
    VERIFY(n3==1);
    //allocate graident field g, and other three fields Ixx,Iyy,Ixy
    igrad2op gradOp;
    gradOp.init(n1,n2);
    gradOp.op(false,false,inArr,g); //find g.
    i3 = 0;
    for (i1=0; i1<n1; i1++) {
      for (i2=0; i2<n2; i2++) {
        Izz = SQ(g[indOut(i1,i2,i3,0)]);
        Ixx = SQ(g[indOut(i1,i2,i3,1)]);
        Izx = g[indOut(i1,i2,i3,1)] * g[indOut(i1,i2,i3,0)];
      }
    }
    //Then smooth the field with Gaussian kernels.
    
    //Then find the eigen values for the matrix
    float a,b,c; a = 1.0f;
    i3 = 0; float l1,l2,det; point2d v1,v2;
    for (i1=0; i1<n1; i1++) {
      for (i2=0; i2<n2; i2++) {
        b = -(Izz[indIn(i1,i2,i3)] + Ixx[indIn(i1,i2,i3)]);
        c =  Izz[indIn(i1,i2,i3)]*Ixx[indIn(i1,i2,i3)] - SQ(Izx[indIn(i1,i2,i3)]);
        det = SQ(b) - 4.0f*a*c;
        if (det<0.0f) {PRV(det);det=0.0f;LOGC("!!det<0 in dip estimation");}
        else {
          det = sqrtf(det);
        }
        l1 = (-b+det)/(2.f*a); //this is the bigger eigenvalue, the corresponding eigenvector corresponds to the gradient direction
        l2 = (-b-det)/(2.f*a); //this is the smaller eigenvalue.
        //infer the eigen vectors
        v1.y = Izz[indIn(i1,i2,i3)]-l1; v1.x = -Izx[indIn(i1,i2,i3)];
        v1.normalize(); if (v1.x<0.0f) {v1.x *= -1.f; v1.y *= -1.f;}
        v2.y = Izz[indIn(i1,i2,i3)]-l2; v2.x = -Izx[indIn(i1,i2,i3)];
        v2.normalize(); if (v2.x<0.0f) {v2.x *= -1.f; v2.y *= -1.f;}
        
        //Now assemble the diffusion matrix
        l1 = l1/(l1+l2); l2 = 1.f - l1;
        outArr[indMat(i1,i2,i3,0)] = l1*v1.x*v1.x + l2*v2.x*v2.x;
        outArr[indMat(i1,i2,i3,1)] = l1*v1.x*v1.y + l2*v2.x*v2.y;
        outArr[indMat(i1,i2,i3,2)] = l2*v1.y*v1.y + l2*v2.y*v2.y;
      }
    } //end for i1
  } else { //ndim=3
    VERIFY(0);
  }
  
  outCub << outArr;
  return 0;
}

