#ifndef __DMCUBE_H_INCL__
#define __DMCUBE_H_INCL__

/**
 * TODO: Add class description
 * 
 * @author   yang
 */
#include "cubeData.hpp"
#include "dmRsf.hpp"

namespace Yang {

//Auxilary func, dump a cubeData to file, for debug inspection
//template <class T>
//void dumpCube(string tag,const cubeData<T> & cube);
template <class T>
void dumpCube(string tag,const cubeData<T> & cube)
{
   dmRsf<T> file(tag,false);
   file.getDomain(&cube.domain());
   file.saveAll(cube);
}

//As the incore, dmRsf class
class dmCube: public dmRsf<float> {
   public:
	  cubeData<float>   * m_pData;
	  
public:
	// Constructor
	dmCube(const string& itag, bool iInput, const string& szInfo);

	// Destructor
	virtual ~dmCube();
    
    //This is the general method that handles the initialization of a RSF object
	virtual void init(const hypercube* pDom)
    {
	  dmRsf<float>::init(pDom);
      //Allocate the memory
	   if (bInput) {
		 //read all the data into memory
		  assert(bInput);
		  assert(this->dom.get_ndim()!=0);
		  m_pData->alloc(this->dom);
		  this->read_all(*m_pData);
	   }
	   else {
		  m_pData->alloc(this->dom);
	   }
    }
    
    //marking the end of RSF object
	virtual void save() {
       dmRsf<float>::save();
       //write the entire content of m_pData
       this->write_all(*m_pData);
    }
	
    virtual void resetCounter() {m_pData->resetCounter();}
	
	//content access interface sequential
	//Allocate the memory using the full size of dmRsf, like a in-core version of rsf
    virtual void seq_grabAll(cubeData<float>& cube) {
      assert(bInput);
	  assert(this->dom.get_ndim()!=0);
	  cube.alloc(this->dom);
	  m_pData->seq_grab(&cube);
   }
   
   //Save the full size of cube to file and !!close the file!! like in-core version of rsf
   virtual void seq_updateAll(const cubeData<float>& cubeObj) 
   {
	  VERIFY((this->dom).sameSize(cubeObj.domain()));
      m_pData->seq_update(&cubeObj);
   }
   
   virtual void seq_grab(cubeData<float> * cube)  { m_pData->seq_grab(cube); }
   virtual void seq_update(const cubeData<float> * cube){ m_pData->seq_update(cube);} 

  //Random access, in these cases cube is alloacted by the caller
  virtual void grab3d(cubeData<float> * cube,  const windParam& wind) { m_pData->grab3d(cube,wind);}
  virtual void grab2d(cubeData<float> * cube,  const windParam& wind) { m_pData->grab2d(cube,wind);}
  virtual void grab1d(cubeData<float> * cube,  const windParam& wind) { m_pData->grab1d(cube,wind);}

  //Generalization
  virtual void grabXd(cubeData<float> * cube,  const windParam& wind, int idim) 
  { m_pData->grabXd(cube,wind, idim);}

  //Grab a window of the object
  virtual void grabWindow(cubeData<float> *cube,  const windParam& wind) { m_pData->grabWindow(cube,wind);}
  
  virtual void update1d(const cubeData<float> * cube,   const windParam& wind) { m_pData->update1d(cube,wind);}
  virtual void update2d(const cubeData<float> * cube,   const windParam& wind) { m_pData->update2d(cube,wind);}
  virtual void update3d(const cubeData<float> * cube,   const windParam& wind) { m_pData->update3d(cube,wind);}
  virtual void updateXd(const cubeData<float> * cube,   const windParam& wind, int idim) 
  { m_pData->updateXd(cube,wind,idim);}
  
  //update a window of the data object 
  virtual void updateWindow(const cubeData<float>* cube,const windParam& wind) { m_pData->updateWindow(cube,wind);}

private:
	// Copy constructor
	// TODO: Uncomment the copy constructor when you need it.
	dmCube(const dmCube& src);

	// Assignment operator
	// TODO: Uncomment the assignment operator when you need it.
	dmCube& operator=(const dmCube& src);
};

// Constructor implementation
inline dmCube::dmCube(const string& itag, bool iInput, const string& szInfo): dmRsf<float>(itag,iInput,szInfo)
{
	   //if is Input, load the values right away
	   m_pData = new cubeData<float>();
}

// Destructor implementation
inline dmCube::~dmCube()
{
   SAFE_DELETE(m_pData);
}

}
// TODO: Uncomment the copy constructor when you need it.
//inline dmCube::dmCube(const dmCube& src)
//{
//   // TODO: copy
//}

// TODO: Uncomment the assignment operator when you need it.
//inline dmCube& dmCube::operator=(const dmCube& rhs)
//{
//   if (this == &rhs) {
//      return *this;
//   }
//
//   // TODO: assignment
//
//   return *this;
//}

#endif // __DMCUBE_H_INCL__
