#include "stdafx.h"

#include "waveOpOneway.h"
#include <sstream>

#include "cubeAlgo2.h"
#include "velAna.param.h"
#include "misc.h"

#include "param_func.h"
#include "dmHeaderReader.h"

using boost::shared_ptr;
using namespace std;
using namespace Yang;

std::string sep_par_get_string(const char *arg) {
  FUNC_TRACE_BEG(sep_par_get_string);
  char x[1024];
  std::string ret;
  int iret=getch(arg,"s",x);
  if(iret==0) {
    prt("\n !!error getting par %s ", arg);
    VERIFY(0);
  } else {
    ret=x;
  }
  FUNC_TRACE_END(ret);
}

static string strCmd;
static char szCmd[4096]; //Should be enough space

waveOpOneway::waveOpOneway(const shared_ptr<waveOpParams>& waveOpPar, const shared_ptr<AcqPar>& acqPars,
             const shared_ptr<ImgPar>& imgPars,   const cubeDataF& vmod, param_func* pars):
  waveOps(waveOpPar,acqPars,imgPars,vmod) {
  string strT;
  if (NULL != pars) {
    strT = pars->get_string("wemig-1way");  strcpy(szWemig, strT.c_str());
    strT = pars->get_string("wewet-1way");  strcpy(szWemva, strT.c_str());      
  } else {
    //static char szWemig[1024] = "/net/server/homes/sep/yang/from-other/yaxun_DSO/Bin/wem3d.x";  //Migration op
    strcpy(szWemig,"/net/server/homes/sep/yang/from-other/yaxun_weo/bin/bwi-wem3d.x");  //Migration op
    //static char szWemva[1024] = "/net/server/homes/sep/yang/from-other/yaxun_DSO/Bin/wemva3d_baspix.x";  //Tomography op
    strcpy(szWemva,"/net/server/homes/sep/yang/from-other/yaxun_weo/bin/bwi-wet3d.x");  //Tomography op
  }
}

void waveOpOneway::initOperator(const waveOpParams* params, waveOps::action_type ac_type) {
   FUNC_TRACE_BEG(waveOpOneway::initOperator);
   FUNC_TRACE_END();
}

void waveOpOneway::setOneWayParam( OneWayParamTang* par) {
  FUNC_TRACE_BEG("OneWay::setOneWayParam");
  
  const hypercube& vmodShp = this->vmod().domain();
  //axis ax_z = vmodShp.get_axis(1);
  axis ax_x = vmodShp.get_axis(2);
  axis ax_y(1,0,1.f);
  
  shared_ptr<const AcqPar> acqPars = this->acqPars();
  
  if (acqPars->srType() == AcqPar::tagAlong ) {
    par->offset = 'y';  //input data is sorted by (shot,offset), like marine 
  } else {
    par->offset = 'n'; //Input data is sorted by (shot,receiver) like land-geometry.
  }
  
  PosInt ns_x,ds_x,nr_x,dr_x; NnegInt os_x; int or_x;
  acqPars->getSrcRecvInfoX(ns_x, os_x, ds_x, nr_x, or_x, dr_x);
  par->ngx = nr_x; par->nsx = ns_x;
  if (acqPars->srType() == AcqPar::tagAlong) {
    par->ogx = ax_x.get_coord(or_x)-ax_x.get_coord(os_x); //the coordinates with respect to the shot location
  } else {
    par->ogx = ax_x.get_coord(or_x); //the absolute coordinates 
  }

  par->dgx = ax_x.d * dr_x;
  par->osx = ax_x.get_coord(os_x); par->dsx = ax_x.d * ds_x;

  sprintf(par->szGeomPar," ngx=%d nsx=%d ogx=%f dgx=%f osx=%f dsx=%f offset=%c ",
          par->ngx,par->nsx,par->ogx,par->dgx,par->osx,par->dsx, par->offset);
  
  char szApp[256];
  bool b3D = acqPars->b3D();
  if (b3D) {
    ax_y = vmodShp.get_axis(3);

    PosInt ns_y,ds_y,nr_y,dr_y; NnegInt os_y; int or_y;
    acqPars->getSrcRecvInfoY(ns_y, os_y, ds_y, nr_y, or_y, dr_y);

    par->ngy = nr_y; par->nsy = ns_y;
    par->ogy = ax_y.get_coord(or_y); par->dgy = ax_y.d * dr_y;
    par->osy = ax_y.get_coord(os_y); par->dsy = ax_y.d * ds_y;
    sprintf(szApp," ngy=%d nsy=%d ogy=%f dgy=%f osy=%f dsy=%f ",
            par->ngy, par->nsy, par->ogy, par->dgy, par->osy, par->dsy);
    strcat(par->szGeomPar, szApp);
  }
  
  //set the offset image migration parameters
  const ImgPar* p = this->imgPars().get();
  VERIFY(NULL!=p);
  const ImgPar& imgPar = *p;
  bool bDoOffset = imgPar.bGenOffsetGather();
  par->nhx = NumSubOffset(1); par->ohx = 0.0f; par->dhx = 0.0f; float d=0.f;
  const offsetImagingParams& offImgPar = imgPar.offImgPars();
  if (bDoOffset) {
    d = ax_x.d; //dx
    par->ohx = offImgPar.ohx * d;
    par->dhx = offImgPar.dhx * d;
    par->nhx = offImgPar.nhx;
    sprintf(par->szOffImgPar, " ohx=%f dhx=%f nhx=%d ", par->ohx, par->dhx, par->nhx.value());
    if (b3D) {
      d = ax_y.d; //dy
      par->ohy = offImgPar.ohx * d;
      par->dhy = offImgPar.dhy * d;
      par->nhy = offImgPar.nhy;
      sprintf(szApp, " ohy=%f dhy=%f nhy=%d ", par->ohy, par->dhy, par->nhy.value());
      strcat(par->szOffImgPar, szApp);
    }
  } else { //not do offset
  }
  
  const OnewayOpPar* wavePar = this->waveOpPars()->onewayPar();
  //set the migration parameters  
//  sprintf(par->szMigPar,"pdip=%s ntpx=%d nws=%d oper=%s nxpad_beg=%d nxpad_end=%d offset=%c %s ",
//          par->pdip.c_str(),par->ntpx,par->nws,par->oper.c_str(),
//          par->nxpad_beg,par->nxpad_end, par->offset, m_pars->sz_additional_mig_par.c_str());
  sprintf(par->szMigPar," offset=%c %s ", par->offset, wavePar->sz_additional_mig_par.c_str());
  if (b3D) {
    //char szt[256];
    //sprintf(szt," nypad_beg=%d nypad_end=%d ", par->nypad_beg,par->nypad_end);
    //strcat(par->szMigPar,szt);
  }
  
  //set the image size
  const hypercube& finalImgShp = imgPar.finalImageShp();
  bool b_vel_img_sameShape = vmodShp.sameShape(finalImgShp);
  if (!b_vel_img_sameShape) {
    float xmin,xmax,zmin,zmax,ymin,ymax; ymin=ymax=0.0f;
    finalImgShp.get_axis(1).getRange(&zmin,&zmax);
    finalImgShp.get_axis(2).getRange(&xmin,&xmax);
    if (b3D)   {finalImgShp.get_axis(3).getRange(&ymin,&ymax);}
    char szt[512];
    sprintf(szt," image_zmin=%f image_zmax=%f image_xmin=%f image_xmax=%f image_ymin=%f image_ymax=%f ",
            zmin,zmax,xmin,xmax,ymin,ymax);
    strcat(par->szMigPar,szt);
  }
  
  //sprintf(par->szBSCpar," bsctyp=fourie "); //new code
  
  //get the filename corresponds to the source signature.
  //PRV2(wavePar->csouFileName,wavePar->crecFileName);
  par->csouFileName = wavePar->csouFileName; par->crecFileName = wavePar->crecFileName;
  
  //Make randomize name for other filenames
  //srand(time(NULL));
  int rsuf = 0; rsuf = rand()/65536;
  std::ostringstream oss;
  oss.str("");
  oss << str_tmpPath<< "seprrvmod" << rsuf <<".H"; par->vmodFileName = 
oss.str(); //oss << "rrvmodTag" << rsuf; par->vmodTag = oss.str();
  oss.str("");
  oss << str_tmpPath<< "seprrbvel" << rsuf <<".H"; par->bvelFileName = oss.str();
  //oss << "rrbvelTag" << rsuf; par->bslwTag   = oss.str();
  oss.str("");
  oss << str_tmpPath<< "seprrimag" << rsuf <<".H"; par->imagFileName = oss.str();
  //oss << "rrimagTag" << rsuf; par->imagTag = oss.str();
  oss.str("");
  oss << str_tmpPath<< "seprrdimg" << rsuf <<".H"; par->dimgFileName = oss.str();
  //oss << "rrdimgTag" << rsuf; par->dimgTag = oss.str();
  oss.str("");
  oss << str_tmpPath<< "seprrdslw" << rsuf <<".H"; par->dslwFileName = oss.str();
  //oss << "rrdslwTag" << rsuf; par->dslwTag = oss.str();

  oss.str("");
  oss << str_tmpPath<< "seprrsrIl" << rsuf <<".H"; par->srIlFileName  = oss.str(); //srIllumination
  oss.str("");
  oss << str_tmpPath<< "seprrtmp" << rsuf <<".H"; par->tmpFileName   = oss.str(); //temporary filename
  
  oss.str("");
  oss << str_tmpPath<< "log.oneway"<< rsuf; par->logfilename = oss.str();
  oss.str("");
  oss << str_tmpPath<< "seprrcrec" << rsuf <<".H"; par->tempRecFileName = oss.str();
  FUNC_TRACE_END();
}

//Handles the format of my velocity cube and Oneway code's input velocity model
//This works for both 2D/3D cases.
void waveOpOneway::convertVelModel(cubeData<float> &vmodYaxun, cubeData<float>& myVelCube, bool bFromMyvelToYaxunVel) {
  FUNC_TRACE_BEG(waveOpOneway::convertVelModel);
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first
  
  shared_ptr<const AcqPar> acqPars = this->acqPars();
  int shotOffZ = acqPars->shotDepth();
  VERIFY(acqPars->recvDepth() == shotOffZ);

  cubeData<float> vmodCube;
  axis ax_z,ax_znew;
  const hypercube& shpV = this->vmod().domain();
  
  ax_z = shpV.get_axis(1); ax_znew = ax_z.alter(shotOffZ);
  vecAxis axes = shpV.get_axes(); axes[0] = ax_znew;
  bool b3D = acqPars->b3D();
  
  if (bFromMyvelToYaxunVel) {
    //set size vmod and copy to m_pars->pVel
    hypercube velCropShape(axes); vmodCube.alloc(velCropShape);
    cropAlign(vmodCube,myVelCube,false);
    //Then transpose from (z,x)/(z,x,y) to (x,y,z), 2D/3D are the same
    int t1[2] = {1,3}; int t2[2] = {2,1};
    stdVecInt reshp(t1,t1+2),mapp(t2,t2+2);
    shuffleAxis(vmodYaxun,vmodCube,reshp,mapp,true,true);
  } else { //convert the otherway around
    //first transpose vmodYaxun 
    //convert (x,y,z) back to (z,x,/y);
    int t1[3] = {1,2,3}; int t2[3] = {2,3,1};
    stdVecInt reshp(t1,t1+3),mapp(t2,t2+3);
    shuffleAxis(vmodCube,vmodYaxun,reshp,mapp,true,true);
    if (!b3D) {
      //strip out the last axis y.
      hypercube shape2 = vmodCube.domain().return_compatible(1,2);
      vmodCube.reshape(shape2,NULL);
    }
    cropAlign(myVelCube,vmodCube,false);
  }
  //LOG('h2');
  //vmodYaxun.dump("dump1");
  FUNC_TRACE_END();
}


//This works only for 2D case, where my img is (z,x,hx) and yaxun's img is (hx,hy,x,y,z)
void waveOpOneway::convertImage(cubeData<float> &ImageYaxun, cubeData<float>& myImgCube, bool bFromMyImageToYaxun) {
  FUNC_TRACE_BEG(waveOpOneway::convertImage);
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first
  shared_ptr<const AcqPar> acqPars = this->acqPars();
  int               shotOffZ = acqPars->shotDepth();

  const ImgPar* imgpar = this->imgPars().get();
  
  cubeData<float> imgN;
  axis ax_z,ax_znew;
  ax_z = myImgCube.domain().get_axis(1); ax_znew = ax_z.alter(shotOffZ);
  vecAxis axes = myImgCube.domain().get_axes(); axes[0] = ax_znew;
  hypercube myImgShape(axes);
  bool bDoOffsetGather = imgpar->bGenOffsetGather();
  stdVecInt reshp,mapp;
  
  if (bFromMyImageToYaxun) {
    //Crop my image first
    imgN.alloc(myImgShape); 
    cropAlign(imgN,myImgCube,false);
    //flip axes from (z,x,hx,...hy,y) to (hx,hy,x,y,z) 
    reshp.resize(4); reshp[0]=1; reshp[1]=2; reshp[2]=4; reshp[3]=5;
    mapp.resize(4);  mapp[0] =4; mapp[1] =2; mapp[2] =1; mapp [3]=3;
    Yang::shuffleAxis(ImageYaxun,imgN,reshp,mapp,true,true);
  } else {
    //Arrange the image in axis (hx,hy,x,y,z) to (z,x) or (z,x,hx)
    reshp.resize(4); reshp[0]=2; reshp[1]=3; reshp[2]=4; reshp[3]=5;
    mapp.resize(4);  mapp[0] =3; mapp[1] =2; mapp[2] =4; mapp [3]=1;
    Yang::shuffleAxis(imgN,ImageYaxun,reshp,mapp,true,true);
    //Clip the last axis
    hypercube shape2;
    if (!bDoOffsetGather) {
      shape2 = imgN.domain().return_compatible(1,2);
    } else {
      shape2 = imgN.domain().return_compatible(1,3);
    }
    imgN.reshape(shape2);
    //Now crop imgN to pImageCube;
    cropAlign(myImgCube,imgN,false);
  }
  //imgN.info("imgN info in convertImage Oneway");
  FUNC_TRACE_END();
}

//Yaxun's image has the ordering of (hx,hy,x,y,z), my 3D ordering would be (z,x,y,hx,hy). Therefore Transp <myimg3d.H reshape=3,5 plane=12 | Transp reshape=2,3,5 plane=23 > yaxunimg3d.H
//This should work for both 2D/3D cases.
void waveOpOneway::convertImage(const string& fn_ImageYaxun, const string& fn_myImg, bool bFromMyImageToYaxun) {
  FUNC_TRACE_BEG("waveOpOneway::convertImage3d");
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  const AcqPar* acqPar = this->acqPars().get();
  const ImgPar* imgPar = this->imgPars().get();
  
  int shotOffZ = acqPar->shotDepth();
  
  cubeData<float> imgN;
  axis ax_z,ax_znew;
  const hypercube& shape0offimg = imgPar->finalImageShp();
  const hypercube& shape_offimg = imgPar->finalSubOffImageShp();
  
  ax_z = shape0offimg.get_axis(1); ax_znew = ax_z.alter(shotOffZ);
  //hypercube myImgShape(axes);
  bool bDoOffsetGather = imgPar->bGenOffsetGather();
  const hypercube* pMyImgCube = bDoOffsetGather? &shape_offimg:&shape0offimg;
  
  stdVecInt reshp,mapp;
  bool b3D = acqPar->b3D();
  int n_axis_img = 3;
  if (b3D) n_axis_img = 5;
  
  char tmpFname[256]; sprintf(tmpFname,"sep_ci%d.H", rand()%65536);
  //piping the Transp command will get hang for some reason!~~
  if (bFromMyImageToYaxun) {
    //Then just Window and then transp the images to comply with Yaxun's format
    sepHeaderReader myImgHeader(fn_myImg,dm::IN,"",true);
    hypercube shape2;
    myImgHeader.getShape(shape2,n_axis_img); 
    PRV(fn_myImg); shape2.info("shape2"); pMyImgCube->info("image shape");
    VERIFY(shape2.sameSize(*pMyImgCube));
    if (!b3D) {
      //flip axes from (z,x,hx,...hy,y) to (hx,hy,x,y,z) 
      //sprintf(szCmd, "Window3d squeeze=n f1=%d <%s | Transp reshape=2,4,5 plane=12 | Transp reshape=2,3,5 plane=23 >%s", shotOffZ, fn_myImg.c_str(), fn_ImageYaxun.c_str());
      sprintf(szCmd, "Window3d squeeze=n f1=%d <%s | Transp reshape=2,4,5 plane=12 >%s ", shotOffZ, fn_myImg.c_str(), tmpFname); VERIFY(system(szCmd)==0);
      sprintf(szCmd, "<%s Transp reshape=2,3,5 plane=23 >%s", tmpFname, fn_ImageYaxun.c_str()); VERIFY(system(szCmd)==0);
    } else {
      //flip axes from (z,x,y,hx,hy) to (hx,hy,x,y,z) 
      //sprintf(szCmd, "Window3d squeeze=n f1=%d <%s | Transp reshape=1,3,5 plane=12 | Transp reshape=3,5 plane=12 >%s", shotOffZ, fn_myImg.c_str(), fn_ImageYaxun.c_str());
      sprintf(szCmd, "Window3d squeeze=n f1=%d <%s | Transp reshape=1,3,5 plane=12 >%s ", shotOffZ, fn_myImg.c_str(),tmpFname); VERIFY(system(szCmd)==0);
      sprintf(szCmd, "<%s Transp reshape=3,5 plane=12 >%s", tmpFname, fn_ImageYaxun.c_str()); VERIFY(system(szCmd)==0);
    } //If no offset image, then all these dummy axes (hx,hy) are initally at the end, it can be processed in the same way and there is no need to differentiate.
  } else { //from Yaxun's image to mine
    //Arrange the image in axis (hx,hy,x,y,z) to (z,x) or (z,x,hx)
    if (!b3D) {
      //flip axes from (hx,hy,x,y,z) to (z,x,hx,...hy,y) and pad
      //sprintf(szCmd, " Window3d <%s squeeze=n | Transp reshape=2,3,4,5 plane=12 | Transp reshape=4,5 plane=12 | Pad beg1=%d >%s", fn_ImageYaxun.c_str(), shotOffZ, fn_myImg.c_str());
      sprintf(szCmd, " Window3d <%s squeeze=n | Transp reshape=2,3,4,5 plane=12 >%s ", fn_ImageYaxun.c_str(), tmpFname); VERIFY(system(szCmd)==0);
      sprintf(szCmd, " <%s Transp reshape=4,5 plane=12 | Pad beg1=%d >%s", tmpFname, shotOffZ, fn_myImg.c_str()); VERIFY(system(szCmd)==0);
    } else {
      //flip axes from (hx,hy,x,y,z) to (z,x,y,hx,hy) and pad z
      //sprintf(szCmd, " Window3d <%s squeeze=n | Transp reshape=2,5 plane=12 | Transp reshape=2,3,5 plane=12 | Pad beg1=%d >%s",     fn_ImageYaxun.c_str(), shotOffZ, fn_myImg.c_str());
      sprintf(szCmd, " Window3d <%s squeeze=n | Transp reshape=2,5 plane=12 >%s ", fn_ImageYaxun.c_str(), tmpFname); VERIFY(system(szCmd)==0);
      sprintf(szCmd, " <%s Transp reshape=2,3,5 plane=12 | Pad beg1=%d >%s", tmpFname,    shotOffZ, fn_myImg.c_str()); VERIFY(system(szCmd)==0);
    }
  } //end if bFromMyImageToYaxun
  prt("Executing Cmd in waveOpOneway convertImage 3D:%s \n", szCmd);
  VERIFY(system(szCmd)==0);
  FUNC_TRACE_END();
}


//No non-lin modeling for one-way equation
void waveOpOneway::modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec) {
  FUNC_TRACE_BEG(waveOpOneway::modeling_op);
  VERIFY(0);
  FUNC_TRACE_END()
}

//Invoke Yaxun's oneway program to perform tasks.
//Invoke wem3d.x, not finished yet..............
void waveOpOneway::bornModeling_op(const cubeData<float>* pReflCube, dmRsf<float>* pRsfSeisRec) {
  VERIFY(!pRsfSeisRec->isDummy() && pRsfSeisRec->isInput());
  //Get file name from its tag
  string dTag = pRsfSeisRec->getTag();
  string seis_fn = sep_par_get_string(dTag.c_str());
  this->bornModeling_op(pReflCube,seis_fn);  
  VERIFY(0);
}

void waveOpOneway::bornModeling_op(const cubeData<float>* pReflCube, const string& seis_fn) {
  OneWayParamTang param;
  this->setOneWayParam(&param);
  param.crecFileName = seis_fn;
  //Do the actual modeling
  this->bornModeling(param,pReflCube);
}

//Input changed to a tag string
void waveOpOneway::migration_op(const string& seis_fn, cubeData<float>* pImageCube) {
  FUNC_TRACE_BEG("waveOpOneway::migration_op2");
  //Get file name from its tag
  //string crecFileName = sep_par_get_string(dataTag.c_str());
  OneWayParamTang param;
  this->setOneWayParam(&param);
  param.crecFileName = seis_fn;
  
  //Do the actual migration
  this->migration(param,pImageCube);

  //If sparse sample the image
  const waveOpParams* opPar = this->waveOpPars().get();
  if (opPar->pImageMask) {
    pImageCube->multiply(*(opPar->pImageMask));
  }
  FUNC_TRACE_END();
}

void waveOpOneway::migration_op(dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube) {
  FUNC_TRACE_BEG("waveOpOneway::migration_op with rsf.seisrec");
  //Figure out the data filename, assume its data file never changes (valid for WEMVA but not for FWI, while one-way is not used in FWI anyway. It is still safe to do.
  VERIFY(!pRsfSeisRec->isDummy() && pRsfSeisRec->isInput());
  //Get file name from its tag
  string dTag = pRsfSeisRec->getTag();
  
  string seis_fn = sep_par_get_string(dTag.c_str());
  this->migration_op(seis_fn, pImageCube);
  FUNC_TRACE_END();
}

//Do born modeling
void waveOpOneway::bornModeling(OneWayParamTang& param, const cubeDataF* pRefl) {
  //const AcqPar* acqPar  = this->acqPars().get();
  //int   shotOffZ        = acqPar->shotDepth();
  
  cubeData<float> vmodCube2;
  //set size vmod and copy to m_pars->pVel, convert my velmodel to Yaxun
  cubeDataF& vel = const_cast< cubeDataF & > (this->vmod());
  this->convertVelModel(vmodCube2,vel,true);
  //Dump the vmod to file
  vmodCube2.dump(param.vmodFileName,true);  usleep(int(1e5)); //for IO safety

  //Convert the input image to Tang's format
  cubeDataF Image2;
  this->convertImage(Image2,const_cast<cubeDataF&>(*pRefl),true);
  Image2.dump(param.imagFileName,true); usleep(int(1e5));

  //LOG(Now the inputs are ready);
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgfwd memchk=n report=y %s %s >&%s", //redirect the output
          szWemig,param.csouFileName.c_str(),param.crecFileName.c_str(),param.vmodFileName.c_str(),
          param.imagFileName.c_str(),param.szMigPar,param.szOffImgPar, param.logfilename.c_str());
  PRV(szCmd);
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //LOG(finish invoking the wem3d program);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.vmodFileName.c_str(),param.imagFileName.c_str());
  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
}

//The depth of the source has to be on top of the model.
//First check the size is correct or not. And align the offset correspondingly.
//Check if the input pRsfSeisrec is of correct shape, then 
//if RsfSeisrec is a dmRsf_oc, then take its tag to find corresponding filename. 
//else the let pRsfSeisrec dump to "crec" tag.
//invoke the wem3d command, direct the output to file rrimag%
//use another dmRsf_oc object to read rrimag% file, and read in the content
//The axes of image is (z,x,hx)
void waveOpOneway::migration(OneWayParamTang& param, cubeData<float>* pImageCube) {
  //const AcqPar* acqPar  = this->acqPars().get(); int shotOffZ    = acqPar->shotDepth();
  
  cubeData<float> vmodCube2,vmodCube1;
  //set size vmod and copy to m_pars->pVel, convert my velmodel to Yaxun
  cubeDataF& vel = const_cast< cubeDataF & > (this->vmod());
  this->convertVelModel(vmodCube1,vel,true);
  pad2D_Tvmod(vmodCube2,vmodCube1);
  //Dump the vmod to file
  vmodCube2.dump(param.vmodFileName,true);  usleep(int(1e5)); //for IO safety

  //LOG(Now the inputs are ready);
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgadj memchk=n report=y %s %s >&%s", //redirect the output
          szWemig,param.csouFileName.c_str(),param.crecFileName.c_str(),param.vmodFileName.c_str(),
          param.imagFileName.c_str(),param.szMigPar,param.szOffImgPar, param.logfilename.c_str());
  PRV(szCmd);
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //LOG(finish invoking the wem3d program);

  //Need read the result image back
  cubeData<float> imgY;
  imgY.load(param.imagFileName);usleep(int(1e5));
  int nx=vel.domain().get_n(2); unpad2D_Timg(vmodCube2,imgY,nx);
  this->convertImage(vmodCube2,*pImageCube,false);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.vmodFileName.c_str(),param.imagFileName.c_str());
  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
}

//For 3D case
void waveOpOneway::migration3d_op(const string& fn_seisRec, const string& fn_img_off) {
  FUNC_TRACE_BEG(waveOpOneway::migration3d_op);
  //Figure out the data filename, assume its data file never changes (valid for WEMVA but not for FWI, while one-way is not used in FWI anyway, so it is safe to do this);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  param.crecFileName = fn_seisRec;
  
  //Do the actual migration
  //const AcqPar* acqPar  = this->acqPars().get();  int shotOffZ = acqPar->shotDepth();

  //Also need to dump the velocity model
  cubeData<float> vmodCube2; //Yaxun's vel model
  //Convert my velocity model to Yaxun's vel model
  //set size vmod and copy to m_pars->pVel, crop velocity model.
  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  this->convertVelModel(vmodCube2,vel, true);
  
  //we can just dump the new velocity model to the filename
  vmodCube2.dump(param.vmodFileName,true); 
  usleep(int(1e5)); //for IO safety

  //LOG(Now the inputs are ready);
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgadj memchk=n report=y %s %s >& %s", //redirect the output
          szWemig,param.csouFileName.c_str(),param.crecFileName.c_str(),param.vmodFileName.c_str(),
          param.imagFileName.c_str(),param.szMigPar,param.szOffImgPar, param.logfilename.c_str());
  prt("Executing cmd:%s. \n",szCmd);
  VERIFY(system(szCmd)==0); usleep(int(1e5));

  //LOG(finish invoking the wem3d program);
  //Need now convert the result image back to my own format.
  this->convertImage(param.imagFileName, fn_img_off, false);
  
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.vmodFileName.c_str(),param.imagFileName.c_str());
  //sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  
  
  //If sparse sample the subsurf offset image
  char szFnt[256]; sprintf(szFnt,"junk%d.H",rand()%256);
  const waveOpParams* wavePar = this->waveOpPars().get();
  if (!(wavePar->fn_imageMask.empty())) {
    sprintf(szCmd, " Add %s %s mode=prod >%s ",wavePar->fn_imageMask.c_str(),fn_img_off.c_str(), szFnt);
    PRV(szCmd); VERIFY(system(szCmd)==0);
    sprintf(szCmd, " cp %s %s", szFnt, fn_img_off.c_str());
  }
  FUNC_TRACE_END();
}

//Input Image perturbation and output slowness perturbation
void waveOpOneway::TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS){
  FUNC_TRACE_BEG(waveOpOneway::TomAdj_op);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  //const AcqPar* acqPar  = this->acqPars().get(); int shotOffZ          = acqPar->shotDepth();

  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  //Dump back ground velocity
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,vel,true);
  vmodCube2.dump(param.bvelFileName,true); usleep(int(1e5));
  
  //Convert the image perturbation
  cubeDataF Image2;
  this->convertImage(Image2,*pDeltaIh,true);
  Image2.dump(param.dimgFileName,true); usleep(int(1e5));
  
  const velAnaParam* vaPar = this->velAnaPars();
  VERIFY(NULL != vaPar);
  char cIllumBal = 'n';
  PRV(vaPar->bIllumBal_);
  if (vaPar && vaPar->bIllumBal_) {cIllumBal='y';}
  
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s dvel=%s dimg=%s adj=y %s %s mode=tomadj bIllumBal=%c srIl=%s memchk=n report=y >&%s",
          szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bvelFileName.c_str(),
          param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szOffImgPar,cIllumBal,param.srIlFileName.c_str(),param.logfilename.c_str());

  PRV(szCmd);
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //LOG(finish invoking the wemvaTom program);
  
  //Need read the result image back
  cubeData<float> velY;
  velY.load(param.dslwFileName,3); usleep(int(1e5));
  
  //Now convert shape from slowY to pDeltaS
  this->convertVelModel(velY,*pDeltaS,false);
  
  //the output of yaxun's program is vel, convert dvel -> ds by dividing -v^2;
  pDeltaS->divide(vel); pDeltaS->divide(vel); (*pDeltaS).scale(-1.0);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bvelFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

//Input Image perturbation and output slowness perturbation, 3D case.
void waveOpOneway::TomAdj3d_op (const string& fn_dImg_off, cubeData<float>& dslwCube){
  FUNC_TRACE_BEG("waveOpOneway::TomAdj3d_op");
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  //const AcqPar* acqPar  = this->acqPars().get(); int shotOffZ = acqPar->shotDepth();

  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  //Dump back ground slowness
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,vel,true);
  vmodCube2.dump(param.bvelFileName,true); usleep(int(1e5)); //!! Notice it is velocity now
  
  //Convert the image perturbation to Yaxun's format
  this->convertImage(param.dimgFileName,fn_dImg_off,true);
  
  //LOG(Now the inputs are ready to do Tom_adj);
  sprintf(szCmd," %s csou=%s crec=%s bvel=%s dvel=%s dimg=%s memchk=n report=y mode=tomadj %s %s >&%s",
          szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bvelFileName.c_str(),
          param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szOffImgPar,param.logfilename.c_str());
  PRV(szCmd);
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //Need to convert the result slowness model back, Keep in mind the newest version, 
  cubeDataF dslwYaxun;  dslwYaxun.load(param.dslwFileName,3);
  dslwCube.domain().info("dslwCube");
  this->convertVelModel(dslwYaxun,dslwCube,false);
  
  //the output is dv, not ds. therefore multiply -s^2 to convert dv to ds.
  dslwCube.divide(vel); dslwCube.divide(vel); dslwCube.scale(-1.0);
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bvelFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  //sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

void Yang::smoothIllumBalancing(cubeDataF& dvel, cubeDataF& illum, cubeDataF& tmp, int iShotZ) {
  //Smooth the illumination first, then divide
  tmp.alloc(illum.domain());
  const hypercube& shp = dvel.domain();
  const axis& ax_z = shp.get_axis(1);
  
  long n23 = shp.get_n23();
  int i,ind;
  rectSmooth(illum,tmp,2,5,1);
  illum.dump("dump5"); tmp.dump("dump6");
  
  //mute the very shallow that is close to the shots
  for (ind=0; ind<n23; ind++) {
    for (i=0; i<iShotZ+3; i++) {
      tmp(ind,i) = 0.0f;
    } 
  } 
  float fThr = tmp.maxval()*1e-2;
  for (ind=0; ind<n23; ind++) {
    for (i=0; i<ax_z.n; i++) {
      if (tmp(ind,i)>fThr) {
        dvel(ind,i) /= (tmp(ind,i));
      } else {
        dvel(ind,i) = 0.0f;
      }
    } 
  } 
  return;
}

//Input image and image perturb, then return the slowness perturbation, assume 2D case.
void waveOpOneway::TomoImit_op (const cubeData<float>* pRefl, const cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS){
  FUNC_TRACE_BEG(waveOpOneway::TomAdj_op);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  //const AcqPar* acqPar  = this->acqPars().get();  int shotOffZ          = acqPar->shotDepth();

  //ImgPar imgPar = new ImgPar(*(this->imgPars())); //make a copy
  //const hypercube& reflShp = pRefl->domain();
  //offsetImagingParams* offPar = new offsetImagingParams(imgPar.offImgPars());
  
  OneWayParamTang param1;
  this->setOneWayParam(&param1);
  
  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  //Dump back ground velocity
  cubeDataF vmodCube1,vmodCube2;
  this->convertVelModel(vmodCube1,vel,true);
  //Pad x, the first axis
  seplib::callSepTool(vmodCube2,vmodCube1,"Pad","beg1=50 end1=50 extend=1",true);
  vmodCube2.dump(param1.bvelFileName,true); usleep(int(1e5));
  
  //Convert the image perturbation
  cubeDataF Image2;
  this->convertImage(Image2,const_cast<cubeDataF&>(*pRefl),true);
  Image2.dump(param1.imagFileName,true); usleep(int(1e5));
  
  //do not need the subsurf imaging parameter, bcoz it is contained in file bimg
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s %s %s mode=imgfwd memchk=n report=y >&%s",
          szWemig,param1.csouFileName.c_str(),param1.tempRecFileName.c_str(),param1.bvelFileName.c_str(),param1.imagFileName.c_str(),
          param1.szGeomPar, param1.szMigPar, param1.logfilename.c_str());
  PRV(szCmd); int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //LOG(finish invoking the wemvaTom program);
  
  const velAnaParam* vaPar = this->velAnaPars();  VERIFY(NULL!=vaPar);
  char cIllumBal = 'n';
  if (vaPar && vaPar->bIllumBal_) {cIllumBal='y';}
  
  //Now use the synthesized data to perform the tomo operator
  this->convertImage(Image2,const_cast<cubeDataF&>(*pDeltaIh),true);
  Image2.dump(param1.dimgFileName,true); usleep(int(1e5));
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s dvel=%s dimg=%s adj=y %s mode=tomadj bIllumBal=%c srIl=%s memchk=n report=y 2>> %s",
          szWemva,param1.csouFileName.c_str(),param1.tempRecFileName.c_str(),param1.bvelFileName.c_str(),
          param1.dslwFileName.c_str(),param1.dimgFileName.c_str(),param1.szMigPar,cIllumBal,param1.srIlFileName.c_str(),param1.logfilename.c_str());
  PRV(szCmd); sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  
  //Need read the result image back
  cubeData<float> velY;
  vmodCube1.load(param1.dslwFileName,3); usleep(int(1e5));
  const axis& ax_x = vel.domain().get_axis(2);
  char pars[128]; sprintf(pars,"f1=50 n1=%d squeeze=0 ", ax_x.n);
  seplib::callSepTool(velY,vmodCube1,"Window3d",pars,true); //window out the padded region.
  //Now convert shape from slowY to pDeltaS
  this->convertVelModel(velY,*pDeltaS,false);
  
  //read the illumination back, and divide
  if (vaPar && vaPar->bIllumBal_) {
    cubeData<float> illuTang,illuY;
    
    sprintf(szCmd,"<%s Window3d f1=50 n1=%d squeeze=0 >%s", 
            param1.srIlFileName.c_str(),ax_x.n,param1.tmpFileName.c_str());
    VERIFY(0 == system(szCmd)); usleep(int(1e5));
    illuTang.load(param1.tmpFileName,3); 
    vel.allocMemoryTo(illuY);
    this->convertVelModel(illuTang,illuY,false); //convert from Tang to Yang format
    
    int shotOffZ = this->acqPars()->shotDepth();
    //Smooth the illumination and divide
    smoothIllumBalancing(*pDeltaS,illuY,illuTang,shotOffZ);
  }
  
  //the output of yaxun's program is vel, convert dvel -> ds by dividing -v^2;
  pDeltaS->divide(vel); pDeltaS->divide(vel); (*pDeltaS).scale(-1.0);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s;Rm %s",param1.bvelFileName.c_str(),param1.dimgFileName.c_str(),param1.dslwFileName.c_str(),param1.tempRecFileName.c_str());
  //sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

/*
void waveOpOneway::TomFwd_op (cubeData<float>* pDeltaS, cubeData<float>* pDeltaI){
  
  FUNC_TRACE_BEG(waveOpOneway::TomFwd_op);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);
  
  //Dump back ground slowness
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,*(m_pars->pVel),true);
  vmodCube2.reciprocal(); //convert to slowness
  addNewSepTag(param.bslwTag.c_str(),false,param.bvelFileName.c_str());
  vmodCube2.dump(param.bslwTag,true); usleep(int(1e5));
  
  //Convert the slowness perturbation
  cubeDataF slow2;
  this->convertVelModel(slow2,*pDeltaS,true);
  addNewSepTag(param.dslwTag.c_str(),false,param.dslwFileName.c_str());
  slow2.dump(param.dslwTag,true); usleep(int(1e5));
  
  LOG("Now the inputs are ready to do Tom_adj");
  sprintf(szCmd,"%s csou=%s crec=%s dvel=%s bvel=%s dimg=%s mode=tomfwd %s %s >&%s",
            szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bvelFileName.c_str(),
            param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szOffImgPar,param.logfilename.c_str());

  PRV(szCmd);
  
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  LOG("finish invoking the wemvaTom program");

  //Need read the result image back
  cubeData<float> imageY;
  readSepFileUsingFileName(imageY,param.dimgTag,param.dimgFileName); usleep(int(1e5));
  
  //Now convert shape from slowY to pDeltaS
  this->convertImage(imageY,*pDeltaI,false);
  
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bvelFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  verbPrt("Exec cmd:%s\n",szCmd);
  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}
*/

///// Not finished yet! ~~~!!
//void waveOpOneway::TomFwd3d_op (cubeData<float>* pDeltaS, cubeData<float>* pDeltaI){
//
//  FUNC_TRACE_BEG(waveOpOneway::TomFwd_op);
//  OneWayParamTang param;
//  this->setOneWayParam(&param);
//
//  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first
//  int shotOffZ = m_pars->mTrace1iz;
//  VERIFY(m_pars->mShot1_iz == shotOffZ);
//
//  //Dump back ground slowness
//  cubeDataF vmodCube2;
//  this->convertVelModel(vmodCube2,*(m_pars->pVel),true);
//  vmodCube2.reciprocal(); //convert to slowness
//  addNewSepTag(param.bslwTag.c_str(),false,param.bslwFileName.c_str());
//  vmodCube2.dump(param.bslwTag,true); usleep(int(1e5));
//
//  //Convert the slowness perturbation
//  cubeDataF slow2;
//  this->convertVelModel(slow2,*pDeltaS,true);
//  addNewSepTag(param.dslwTag.c_str(),false,param.dslwFileName.c_str());
//  slow2.dump(param.dslwTag,true); usleep(int(1e5));
//
//  LOG("Now the inputs are ready to do Tom_adj");
//  sprintf(szCmd,"%s csou=%s crec=%s bslw=%s dslw=%s dimg=%s adj=n %s %s %s >&%s",
//          szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bslwFileName.c_str(),
//          param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szBSCpar,param.szOffImgPar,param.logfilename.c_str());
//  PRV(szCmd);
//
//  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
//  LOG("finish invoking the wemvaTom program");
//
//  //Need read the result image back
//  cubeData<float> imageY;
//  readSepFileUsingFileName(imageY,param.dimgTag,param.dimgFileName); usleep(int(1e5));
//
//  //Now convert shape from slowY to pDeltaS
//  this->convertImage(imageY,*pDeltaI,false);
//
//  //Remove the intemediate files
//  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bslwFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
//  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
//  FUNC_TRACE_END();
//}

