#ifndef WAVE_OP_ONEWAY_H
#define WAVE_OP_ONEWAY_H 1

#include "waveOps.h"
#include "misc.h"
#include <sstream>

namespace Yang {
//Forward declaration of template class
//template <typename T>
//class dmRsf_oc;

struct OneWayParamTang;
struct Yang::OneWayParamTang{
  //geometry
  int ngx,nsx,ngy,nsy;
  float ogx,dgx,  osx,dsx;
  float ogy,dgy,  osy,dsy;
  
  //Notice this offset indicates if the data are gather in shot,receiver or CMP axes. It has NOTHING to do with
  //subsurface offset gather.
  char offset; //y or n, 
  char szGeomPar[1024];
  char szOffImgPar[256];
  char szMigPar[1024];
  //char szBSCpar[256]; //
  //int image_xmin, image_xmax;
  NumSubOffset nhx,nhy;  
  float dhx,dhy; float ohx,ohy;
 
  string csouFileName,crecFileName;
  //These files needs to be randomized, since the dimension is different from my convention
  string vmodFileName,imagFileName,bvelFileName,srIlFileName,dslwFileName,dimgFileName,tmpFileName;
  string tempRecFileName;
  string vmodTag,imagTag,bslwTag,dslwTag,dimgTag;
  string logfilename;
};

class waveOpOneway: public waveOps  //Define the class of wave propagation operator
{
public:
   
public://methods
  waveOpOneway(const shared_ptr<waveOpParams>& waveOpPar, const shared_ptr<AcqPar>& acqPars,
          const shared_ptr<ImgPar>& imgPars,   const cubeDataF& vmod, param_func* pars=NULL);
  virtual ~waveOpOneway() {}
    
  virtual void modeling_op(    cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec);
  virtual void bornModeling_op(const cubeDataF* pReflCube, dmRsf<float>* pRsfSeisRec); //Need to set acquisition and etc.
  virtual void bornModeling_op(const cubeDataF* pReflCube, const string& seis_fn);
  
  //in core & out of core
  virtual void migration_op(dmRsf<float>* pRsfSeisRec,  cubeData<float>* pImageCube);
  virtual void migration_op(const string& dataFile,     cubeData<float>* pImageCube); 
  virtual void migration3d_op(const string& fn_seisRec, const string& fn_img_off); //3d

  virtual void TomAdj_op   (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS);
  virtual void TomAdj3d_op (const string& fn_dImg_off, cubeData<float>& dslwCube);  //3d case
  
  virtual void TomoImit_op (const cubeData<float>* pRefl, const cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS);
  
protected:
    void setOneWayParam(OneWayParamTang* par);
    void convertImage(cubeData<float> &ImageYaxun, cubeData<float>& myImgCube, bool bFromMyImageToYaxun);
    void convertImage(const string& fn_ImageYaxun, const string& fn_myImg, bool bFromMyImageToYaxun); //out of core, method
    void convertVelModel(cubeData<float> &vmodYaxun, cubeData<float>& myVelCube, bool bFromMyvelToYaxunVel);

private:
  virtual void initOperator(const waveOpParams* params, waveOps::action_type ac_type);

  virtual void migration(OneWayParamTang& param, cubeDataF* pImageCube); //invoke Yaxun's program and handle the data format conversion.
  virtual void bornModeling(OneWayParamTang& param, const cubeDataF* pRefl);
protected:
  char szWemig[512],szWemva[512];
};

  inline void pad2D_Tvmod(cubeDataF& vTout, const cubeDataF& vTin) {
    //pad axis 1, which is x
    seplib::callSepTool(vTout,vTin,"Pad","beg1=50 end1=50 extend=1",true);
  }
  inline string pad2D_Tvmod_cmd(const string& voutFn, const string& vinFn) {
    //pad axis 1, which is x
    std::stringstream ss;
    ss << "Pad beg1=50 end1=50 extend=1 < " << vinFn << " > " << voutFn;
    return ss.str();
  }
  
  inline void unpad2D_Tvmod(cubeDataF& vTout, const cubeDataF& vTin, int nx) {
    char pars[128]; sprintf(pars,"f1=50 n1=%d squeeze=0 ", nx); //axis 1 is x
    seplib::callSepTool(vTout,vTin,"Window3d",pars,true); //window out the padded region.
  }
  inline void unpad2D_Timg(cubeDataF& iTout, const cubeDataF& iTin, int nx) {
    char pars[128]; sprintf(pars,"f3=50 n3=%d squeeze=0 ", nx); //axis 3 is x
    seplib::callSepTool(iTout,iTin,"Window3d",pars,true); //window out the padded region.
  }
  inline string pad2D_Timg_cmd(const string& ioutFn, const string& iinFn) {
    std::stringstream ss;
    ss << "Pad beg3=50 end3=50 extend=1 < " << iinFn << " > " << ioutFn;
    return ss.str();
  }

  void smoothIllumBalancing(cubeDataF& dvel, cubeDataF& illum, cubeDataF& tmp, int iShotZ);

} //end of namespace Yang
#endif

