#include "stdafx.h"

#include "decl_fwd.h"
#include "waveOpOneway.h"
#include <sstream>

#include "dmHeaderReader.h"
#include "cubeAlgo2.h"
#include "velAna.param.h"
#include "misc.h"

#include "waveOpMPI.h"

static string strCmd;
static char szCmd[4096];
using namespace Yang;
using namespace std;

branchFileName::branchFileName(const std::string& origFileName, const string& suf) {
  origFn_ = origFileName; suf_=suf;
  //find the extension . position
  extPos_ = origFn_.rfind('.'); 
  VERIFY(extPos_ != string::npos); VERIFY(origFn_.at(extPos_+1) != '/');
  return;
}

string branchFileName::fn_at(int irank) const {
  stringstream& ss = const_cast<stringstream&> (ss_);
  ss.str(""); ss <<"-" << suf_ << "-np" << irank;
  string new_fn(origFn_); new_fn.insert(extPos_,ss.str());
  return new_fn;
}

string branchFileName::stackCmd(int nrank) const {
  string cmd;
  cmd = "Add "; 
  for (int ir=1; ir<nrank; ir++) {
    cmd += this->fn_at(ir); cmd.append(" "); 
  }
  cmd.append(" > "); cmd += origFn_;
  return cmd;
}

string branchFileName::catCmd(int nrank, int iAxis) const {
  std::stringstream ss;
  ss << "Cat ";
  for (int ir=1; ir<nrank; ir++) {
    ss << " " << this->fn_at(ir) << " ";
  }
  ss << " axis=" << iAxis << " "; 
  ss << " > " << origFn_;
  return ss.str();
}

workScheduler1d::workScheduler1d(int njobs, int np, scheduleType sType): np_(np) {
  ns_.resize(np); os_.resize(np); ds_.resize(np);
  int ip=0;
  int n,o,d;
  VERIFY(njobs>=1); VERIFY(np>=1);
  if (sType == hopping) {
    for (ip=0; ip<np; ip++) {
      //the number of shots for this process, it handles shots ir,ir+nr,ir+2*nr,..., till shots nsx 
      n = int((njobs-(ip+1))/np)+1;
      o = ip;
      d = np;
      ns_[ip]=n; os_[ip]=o; ds_[ip]=d;
    }
  } else if (sType==chunk) {
    int nj_left = njobs;
    for (ip=0; ip<np; ip++) {
      //the number of shots for this process, it handles shots ir,ir+1,ir+nr-1, nr might vary.
      d = 1;
      n = nj_left/(np-ip) + (nj_left%(np-ip)==0 ? 0:1);
      o = njobs-nj_left;
      nj_left -= n;
      ns_[ip]=n; os_[ip]=o; ds_[ip]=d;
    }
    VERIFY(nj_left==0);
  } else {
    VERIFY(0);
  }
}

#ifndef USE_MPI

#else
#include "my_mpi.h"

int split_file_MPI(const string& fnIn, const string& sufOut, int iAxis, const workScheduler1d& ws, string datapath) {
  //The outfile will be in the same folder as in File
  int np    = ws.np(); //np is number of slaves.
  VERIFY(np>=1);

  //Read in the header info.
  sepHeaderReader hReader(fnIn, dm::IN, "", true);
  hypercube shp;
  hReader.getShape(shp);

  shp.info("shp");
  const axis& ax = shp.get_axis(iAxis);
  int n = ax.n;

  branchFileName bf(fnIn,sufOut);
  int npp,fpp;
  char szCmd[4096];

  for (int ir=1; ir<=np; ir++) {
    int n,o,d;
    ws.get_nod(ir-1,&n,&o,&d);
    npp = n;  fpp = o; 
    sprintf(szCmd,"Window3d <%s f%d=%d j%d=%d n%d=%d squeeze=n >%s", fnIn.c_str(), iAxis,fpp, iAxis,d, iAxis,npp, bf.fn_at(ir).c_str()); 
    if (!datapath.empty()) {
      char sz[1024];
      sprintf(sz," datapath=%s ", datapath.c_str());
      strcat(szCmd,sz);
    }
    PRV(szCmd);
    VERIFY(system(szCmd)==0);
  }
  return 0;
}

//simply duplicate file for the purpose of MPI
int dup_file_MPI(const string& fnIn, const string& sufOut,const workScheduler1d& ws, string datapath) {
  //The outfile will be in the same folder as in File
  int np    = ws.np(); //np is number of slaves.
  VERIFY(np>=1);

  branchFileName bf(fnIn,sufOut);
  int npp,fpp;
  char szCmd[4096];

  for (int ir=1; ir<=np; ir++) {
    if (!datapath.empty()) {
      sprintf(szCmd,"Cp %s %s datapath=%s ", fnIn.c_str(), bf.fn_at(ir).c_str(),datapath.c_str());  //copy binary
    } else { //not copying binary
      sprintf(szCmd,"cp %s %s ", fnIn.c_str(), bf.fn_at(ir).c_str());
    }
    PRV(szCmd);
    VERIFY(system(szCmd)==0);
  }
  return 0;
}

waveOpOnewayMPI::waveOpOnewayMPI(const shared_ptr<waveOpParams>& waveOpPar, const shared_ptr<AcqPar>& acqPars, const shared_ptr<ImgPar>& imgPars, 
const cubeDataF& vmod, param_func* pars, const shared_ptr<workScheduler1d>& ws): waveOpOneway(waveOpPar,acqPars,imgPars,vmod,pars), ws_(ws) { 
  strTmpDir_ = "/tmp/";
}

//Used for MPI case, write to szGeomPar for each process.
void waveOpOnewayMPI::setGeomPar(char* szGeom, const OneWayParamTang* par, int nrank, int irank) {
  FUNC_TRACE_BEG("OneWay::setGeomParam");
  
  const hypercube& vmodShp = this->vmod().domain();
  //axis ax_z = vmodShp.get_axis(1);
  axis ax_x = vmodShp.get_axis(2);
  axis ax_y(1,0,1.f);
  
  boost::shared_ptr<const AcqPar> acqPars = this->acqPars();
  
  int nproc = nrank-1;
  int nsx; nsx=1;
  float osx,dsx;
  //See how many processes are in total.
  if (nrank==1) { //emulating the no MPI case
    sprintf(szGeom," ngx=%d nsx=%d ogx=%f dgx=%f osx=%f dsx=%f offset=%c ",
          par->ngx,par->nsx,par->ogx,par->dgx,par->osx,par->dsx, par->offset);
  } else { //if MPI mode
    VERIFY(irank>0);
    int n,o,d;
    ws_->get_nod(irank-1,&n,&o,&d);
    nsx = n; //the number of shots for this process, it handles shots ir,ir+nr,ir+2*nr,..., till shots nsx
    osx = par->osx + par->dsx*o;
    dsx = par->dsx*d;
    sprintf(szGeom,"ngx=%d nsx=%d ogx=%f dgx=%f osx=%f dsx=%f offset=%c ",
          par->ngx,nsx,par->ogx,par->dgx,osx,dsx, par->offset);
  }
  
  //just break along x, not along Y;
  char szApp[256];
  bool b3D = acqPars->b3D();
  if (b3D) {
    sprintf(szApp," ngy=%d nsy=%d ogy=%f dgy=%f osy=%f dsy=%f ",
            par->ngy, par->nsy, par->ogy, par->dgy, par->osy, par->dsy);
    strcat(szGeom, szApp);
  }
  FUNC_TRACE_END();
}

void waveOpOnewayMPI::migration(OneWayParamTang& param, cubeData<float>* pImageCube) {
  //const AcqPar* acqPar  = this->acqPars().get(); int shotOffZ    = acqPar->shotDepth();

  cubeData<float> vmodCube2,vmodCube1;
  //set size vmod and copy to m_pars->pVel, convert my velmodel to Yaxun
  cubeDataF& vel = const_cast< cubeDataF & > (this->vmod());
  this->convertVelModel(vmodCube1,vel,true);
  pad2D_Tvmod(vmodCube2,vmodCube1);
  //Dump the vmod to file
  vmodCube2.dump(param.vmodFileName,true);  usleep(int(1e5)); //for IO safety

  LOGC("Now the inputs are ready for waveOpOnewayMPI::migration");
  
  string bf_suf = this->waveOpPars()->suf_MPI_;
  branchFileName bf_vmod(param.vmodFileName,bf_suf);
  
  //This does not work well
////Here we use MPI distribute the job to each slave process
//for (int ir=1; ir<world_.size(); ir++) {
//  //generate the commands for each process to execute its part
//  //distribute the velocity models to local disk.
//  //modify the filenames, find the occurance of the dot and add '-ip#' to it for each process.
//  create_cpy_one2all(strCmd, strTmpDir_.c_str(),param.vmodFileName.c_str(),bf_vmod.fn_at(ir).c_str());
//  //Now distribute the file
//  world_.send(ir,WRKTAG,strCmd);
//}
//world_.barrier();
  branchFileName bf_csou(param.csouFileName,bf_suf), bf_crec(param.crecFileName,bf_suf), bf_imag(param.imagFileName,bf_suf), 
bf_log(param.logfilename,bf_suf); //The form the cmd for the actual migration
  for (int ir=1; ir<world_.size(); ir++) {
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgadj memchk=n report=y %s %s >&%s", //redirect the output
            szWemig,bf_csou.fn_at(ir).c_str(),bf_crec.fn_at(ir).c_str(),param.vmodFileName.c_str(),
            bf_imag.fn_at(ir).c_str(),param.szMigPar,param.szOffImgPar, bf_log.fn_at(ir).c_str());
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd); //ask the children to do their work.
  }
  LOG("finish invoking the Mig wem3d program in MPI version");
  PRV(strCmd);
  world_.barrier(); //sync

  cubeData<float> imgY;
  //Then the partial images has been created for each process, now read all partial images and sum them up.
  if (0) {
    strCmd = bf_imag.stackCmd(world_.size()); PRV(strCmd);
    VERIFY(system(strCmd.c_str()) == 0);
    //Need read the result image back from param.imagFileName
    vmodCube2.load(param.imagFileName);usleep(int(1e5));
  } else { //Use MPI reduce to stack the final image
    sepHeaderReader hdr(bf_imag.fn_at(1),dm::IN,"partial img hdr",true);
    hypercube shp_partial_img;  hdr.getShape(shp_partial_img,5);
    vmodCube2.alloc(shp_partial_img);
    shp_partial_img.info("partial img hdr");
    mpi::ReduceCube_MPI(vmodCube2,PosInt(3),bf_imag); //stack thre results using MPI_Reduce, result in vmodCube2, divide blk at (hx,hy,x | y,z), essentially each z slice as a block 
  }
  int nx=vel.domain().get_n(2); unpad2D_Timg(imgY,vmodCube2,nx);
  this->convertImage(imgY,*pImageCube,false);
  LOG("F11");
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.vmodFileName.c_str(),param.imagFileName.c_str());
  //VERIFY(system(szCmd)==0); usleep(int(1e5));
  return;
}

//Do born modeling
void waveOpOnewayMPI::bornModeling(OneWayParamTang& param, const cubeDataF* pRefl) {
  //const AcqPar* acqPar  = this->acqPars().get();
  //int   shotOffZ        = acqPar->shotDepth();
  
  LOG("DD1");  
  cubeData<float> vmodCube1,vmodCube2;
  //set size vmod and copy to m_pars->pVel, convert my velmodel to Yaxun
  cubeDataF& vel = const_cast< cubeDataF & > (this->vmod());
  this->convertVelModel(vmodCube1,vel,true);
  pad2D_Tvmod(vmodCube2,vmodCube1);
  //Dump the vmod to file
  vmodCube2.dump(param.vmodFileName,true);  usleep(int(1e5)); //for IO safety
  PRV(param.vmodFileName);
  
  //Convert the input image to Tang's format, and pad the images.
  cubeDataF Image2;
  this->convertImage(Image2,const_cast<cubeDataF&>(*pRefl),true);
  string tmpFn = param.imagFileName+"-a.H";
  Image2.dump(tmpFn,true);
  strCmd = pad2D_Timg_cmd(param.imagFileName, tmpFn); //do the paddings.
  VERIFY(system(strCmd.c_str())==0);
  
  PRV2(tmpFn,strCmd);
  strCmd = "Rm "; strCmd += tmpFn; VERIFY(system(strCmd.c_str())==0); //clean the tmp file.
  LOG("DD3");
  
  //Now ask each process to simulate the data
  const waveOpParams* wPar = this->waveOpPars().get();
  PRV2(param.csouFileName,param.crecFileName); PRV(param.logfilename);
  branchFileName bf_csou(param.csouFileName,wPar->suf_MPI_), bf_crec(param.crecFileName,wPar->suf_MPI_), bf_log(param.logfilename,wPar->suf_MPI_); 
  char szGeom[512];
  for (int ir=1; ir<world_.size(); ir++) {
    //find the Geom for this process
    this->setGeomPar(szGeom, &param, world_.size(), ir);
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgfwd memchk=n report=y %s %s >&%s", //redirect the output
            szWemig,bf_csou.fn_at(ir).c_str(),bf_crec.fn_at(ir).c_str(),param.vmodFileName.c_str(),
            param.imagFileName.c_str(),param.szMigPar,szGeom,bf_log.fn_at(ir).c_str());
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd); //ask the children to do their work.
    PRV(strCmd);
  }
  world_.barrier(); //wait for other worknodes to finish.
  
  //now cat the crec files together
  strCmd = bf_crec.catCmd(world_.size(),4); //the 4th axis is the sx axis.
  PRV(strCmd);
  VERIFY(system(strCmd.c_str())==0);

  LOG("finish invoking the wem3d modeling program");
  //cleaned up.
  for (int ir=1; ir<world_.size(); ir++)  {
    sprintf(szCmd,"rm %s;", bf_log.fn_at(ir).c_str());
    //strCmd = szCmd; world_.send(ir,WRKTAG,strCmd); 
  }
  LOG("DD5");
  //Need to wait for all process to finish.
  
  //usleep(20e6);
  LOG("DD6");
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.vmodFileName.c_str(),param.imagFileName.c_str());
  //VERIFY(system(szCmd)==0);usleep(int(1e5));
  usleep(2e6);
}


//Input Image perturbation and output slowness perturbation
void waveOpOnewayMPI::TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS){
  FUNC_TRACE_BEG(waveOpOneway::TomAdj_op);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  //const AcqPar* acqPar  = this->acqPars().get(); int shotOffZ          = acqPar->shotDepth();

  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  //Dump back ground velocity, and pad.
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,vel,true);
  string tmpFn = param.bvelFileName+"-a.H";
  vmodCube2.dump(tmpFn,true); usleep(int(1e5));
  strCmd = pad2D_Tvmod_cmd(param.bvelFileName,tmpFn);
  VERIFY(system(strCmd.c_str())==0); usleep(int(1e5));
  
  //Convert the image perturbation
  cubeDataF Image2;
  this->convertImage(Image2,*pDeltaIh,true);
  Image2.dump(tmpFn,true); usleep(int(1e5));
  strCmd = pad2D_Timg_cmd(param.dimgFileName,tmpFn);
  VERIFY(system(strCmd.c_str())==0); usleep(int(1e5));
  
  const velAnaParam* vaPar = this->velAnaPars();
  VERIFY(NULL != vaPar);
  char cIllumBal = 'n';
  PRV(vaPar->bIllumBal_);
  if (vaPar && vaPar->bIllumBal_) {cIllumBal='y';}
  
  int ir=0;
  string bf_suf = this->waveOpPars()->suf_MPI_;
  branchFileName bf_csou(param.csouFileName,bf_suf),bf_crec(param.crecFileName,bf_suf),bf_log(param.logfilename,bf_suf),bf_dvel(param.dslwFileName,bf_suf),
    bf_srIl(param.srIlFileName,bf_suf);
  for (ir=1; ir<world_.size(); ir++) {
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s dvel=%s dimg=%s adj=y %s %s mode=tomadj bIllumBal=%c srIl=%s memchk=n report=y >&%s",
            szWemva,bf_csou.fn_at(ir).c_str(),bf_crec.fn_at(ir).c_str(),param.bvelFileName.c_str(),
            bf_dvel.fn_at(ir).c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szOffImgPar,cIllumBal,bf_srIl.fn_at(ir).c_str(),bf_log.fn_at(ir).c_str());
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd); //ask the children to do their work.
  }
  world_.barrier(); //sync
  PRV(strCmd);

  //cleaned up.
  for (ir=1; ir<world_.size(); ir++)  {
    sprintf(szCmd,"Rm %s;Rm %s", bf_log.fn_at(ir).c_str(), bf_srIl.fn_at(ir).c_str());
    //strCmd = szCmd; world_.send(ir,WRKTAG,strCmd); 
  }
  //might need to stack the srIl illumnation pattern.
  
  cubeData<float> velY;
  //Stack the output
  if (0) {
    strCmd = bf_dvel.stackCmd(world_.size());
    VERIFY(system(strCmd.c_str())==0); usleep(int(1e5));
    //Need read the result image back
    velY.load(param.dslwFileName,3); usleep(int(1e5));
  } else { //Use MPI reduce to stack the final slowness
    sepHeaderReader hdr(bf_dvel.fn_at(1),dm::IN,"vel perturbation",true);
    hypercube shp_dvel;  hdr.getShape(shp_dvel,3);
    velY.alloc(shp_dvel);
    mpi::ReduceCube_MPI(velY,PosInt(3),bf_dvel); //stack thre results using MPI_Reduce
  }
  
  //Now convert shape from slowY to pDeltaS, the cropping is included
  this->convertVelModel(velY,*pDeltaS,false);
  //the output of yaxun's program is vel, convert dvel -> ds by dividing -v^2;
  pDeltaS->divide(vel); pDeltaS->divide(vel); (*pDeltaS).scale(-1.0);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bvelFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  VERIFY(system(szCmd)==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

//Input image and image perturb, then return the slowness perturbation, assume 2D case.
void waveOpOnewayMPI::TomoImit_op (const cubeData<float>* pRefl, const cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS){
  FUNC_TRACE_BEG(waveOpOneway::TomAdj_op);

  OneWayParamTang param1;
  this->setOneWayParam(&param1);
  
  cubeDataF& vel = const_cast<cubeDataF&> (this->vmod());
  //Dump back ground velocity
  cubeDataF vmodCube1,vmodCube2;
  this->convertVelModel(vmodCube1,vel,true);
  //Pad x, the first axis
  seplib::callSepTool(vmodCube2,vmodCube1,"Pad","beg1=50 end1=50 extend=1",true);
  vmodCube2.dump(param1.bvelFileName,true); usleep(int(1e5));
  
  //Convert the image perturbation
  cubeDataF Image2;
  this->convertImage(Image2,const_cast<cubeDataF&>(*pRefl),true);
  Image2.dump(param1.imagFileName,true); usleep(int(1e5));
  
  string bf_suf = this->waveOpPars()->suf_MPI_;
  branchFileName bf_csou(param1.csouFileName,bf_suf), bf_tcrec(param1.tempRecFileName,bf_suf), bf_log(param1.logfilename,bf_suf);
  branchFileName bf_dvel(param1.dslwFileName,bf_suf),bf_srIl(param1.srIlFileName,bf_suf);

  char szGeom[512];
  for (int ir=1; ir<world_.size(); ir++) {
    //find the Geom for this process
    this->setGeomPar(szGeom, &param1, world_.size(), ir);
    //do not need the subsurf imaging parameter, bcoz it is contained in file bimg
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s %s %s mode=imgfwd memchk=n report=y >&%s",
          szWemig,bf_csou.fn_at(ir).c_str(),bf_tcrec.fn_at(ir).c_str(),param1.bvelFileName.c_str(),param1.imagFileName.c_str(),
          szGeom, param1.szMigPar, bf_log.fn_at(ir).c_str());
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd);
  }
  PRV(strCmd);
  world_.barrier();

  const velAnaParam* vaPar = this->velAnaPars();  VERIFY(NULL!=vaPar);
  char cIllumBal = 'n';
  if (vaPar && vaPar->bIllumBal_) {cIllumBal='y';}
  
  //Now use the synthesized data to perform the tomo operator
  this->convertImage(Image2,const_cast<cubeDataF&>(*pDeltaIh),true);
  Image2.dump(param1.dimgFileName,true); usleep(int(1e5));
  
  for (int ir=1; ir<world_.size(); ir++) {
    sprintf(szCmd,"%s csou=%s crec=%s bvel=%s dvel=%s dimg=%s adj=y %s mode=tomadj bIllumBal=%c srIl=%s memchk=n report=y 2>> %s",
            szWemva,bf_csou.fn_at(ir).c_str(),bf_tcrec.fn_at(ir).c_str(),param1.bvelFileName.c_str(),
            bf_dvel.fn_at(ir).c_str(),param1.dimgFileName.c_str(),param1.szMigPar,cIllumBal,bf_srIl.fn_at(ir).c_str(),bf_log.fn_at(ir).c_str());
    //Send the workers the cmd
    strCmd = szCmd;
    world_.send(ir,WRKTAG,strCmd);usleep(int(1e5));
  }
  PRV(strCmd);
  world_.barrier(); //sync
  
  cubeData<float> velY;
  //Stack the dslw files.
  if (0) {
    strCmd = bf_dvel.stackCmd(world_.size());
    VERIFY(system(strCmd.c_str())==0);
    velY.load(param1.dslwFileName,3); usleep(int(1e5));
  } else { //Use MPI reduce to stack the final slowness
    sepHeaderReader hdr(bf_dvel.fn_at(1),dm::IN,"vel perturbation",true);
    hypercube shp_dvel;  hdr.getShape(shp_dvel,3);
    velY.alloc(shp_dvel);
    mpi::ReduceCube_MPI(velY,PosInt(3),bf_dvel); //stack thre results using MPI_Reduce
  }
  const axis& ax_x = vel.domain().get_axis(2);
  //Now convert shape from slowY to pDeltaS, the cropping is included
  this->convertVelModel(velY,*pDeltaS,false);
  
  //read the illumination back, and divide
  if (vaPar && vaPar->bIllumBal_) {
    cubeData<float> illuTang,illuY;
    
    //stack the Illumination file
    strCmd = bf_srIl.stackCmd(world_.size());
    VERIFY(system(strCmd.c_str())==0);
    
    sprintf(szCmd,"<%s Window3d f1=50 n1=%d squeeze=0 >%s", 
            param1.srIlFileName.c_str(),ax_x.n,param1.tmpFileName.c_str());
    VERIFY(0 == system(szCmd)); usleep(int(1e5));
    illuTang.load(param1.tmpFileName,3); 
    vel.allocMemoryTo(illuY);
    this->convertVelModel(illuTang,illuY,false); //convert from Tang to Yang format
    
    int shotOffZ = this->acqPars()->shotDepth();
    //Smooth the illumination and divide
    smoothIllumBalancing(*pDeltaS,illuY,illuTang,shotOffZ);
  }
  
  //the output of yaxun's program is vel, convert dvel -> ds by dividing -v^2;
  pDeltaS->divide(vel); pDeltaS->divide(vel); (*pDeltaS).scale(-1.0);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param1.bvelFileName.c_str(),param1.dimgFileName.c_str(),param1.dslwFileName.c_str());
  //sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

#endif

