#ifndef WAVE_OPS_H
#define WAVE_OPS_H 1

#include "stdafx.h"
#include "hypercube.h"
#include "cubeData.hpp"
#include "dmRsf.hpp"
#include "waveOpParams.h"

namespace Yang {
//Aux function
void calcTimeParamsFor2wayModeling(FDModelingTimeParams* timePar, float maxFreq, float vmax, float vmin, const hypercube &modelGridShape);

using boost::shared_ptr;
//Forward declaration of template class
//template <typename T>
//class dmRsf_oc;
class waveOps
{
public:
  enum action_type {
  acModelingNL, //non Linear modeling modeling
  acMigration, acBornModeling,
  acTomFwd, //From deltaS -> delta I
  acTomAdj,  //From deltaI -> delta S
  acTomImit};

private:
  action_type mAction; 
  shared_ptr<waveOpParams>  waveOpPars_;
  const velAnaParam*        velAnaPars_;
  shared_ptr<AcqPar>        acqPars_;
  shared_ptr<ImgPar>        imgPars_;
  cubeDataF vmod_;
  
public://methods
  waveOps(const shared_ptr<waveOpParams>& waveOpPar, const shared_ptr<AcqPar>& acqPars,
          const shared_ptr<ImgPar>& imgPars,   const cubeDataF& vmod):
    waveOpPars_(waveOpPar),velAnaPars_(NULL),acqPars_(acqPars),imgPars_(imgPars) {
    vmod.copyTo(vmod_);
  }
  virtual ~waveOps() {}
  void velAnaParsIs(const velAnaParam* p) {this->velAnaPars_ = p;}
  const velAnaParam* velAnaPars() const {return this->velAnaPars_;}
  
  int waveOpParsIs(const shared_ptr<waveOpParams>& waveOpParIn) {waveOpPars_ = waveOpParIn; return 0;}
  shared_ptr<const waveOpParams> waveOpPars() const {return waveOpPars_;}
  int acqParsIs(const shared_ptr<AcqPar>& acqParIn) {acqPars_ = acqParIn; return 0;}
  shared_ptr<const AcqPar> acqPars() const {return acqPars_;}
  int ImgParsIs(const shared_ptr<ImgPar>& imgParIn) {imgPars_ = imgParIn; return 0;}
  shared_ptr<const ImgPar> imgPars() const {return imgPars_;}
  int vmodIs(const cubeDataF& vmod)   {vmod.copyTo(vmod_); return 0;}
  const cubeDataF& vmod() const       {return vmod_;}
  
  action_type action() {return mAction;}
  int actionIs( action_type act); //mutator
   
  //virtual void initOperator(waveOpParams* params, action_type ac_type) = 0;
  void initOperatorEx(const waveOpParams* params, action_type ac_type);
   
  virtual void modeling_op(    cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec) {}
  virtual void bornModeling_op(const cubeData<float>* pReflCube, dmRsf<float>* pRsfSeisRec) {} //Need to set acquisition and etc.
  virtual void bornModeling_op(const cubeData<float>* pReflCube, const string& seis_fn) {}

  //in core & out of core
  virtual void migration_op(dmRsf<float>* pRsfSeisRec,  cubeData<float>* pImageCube) = 0;
  virtual void migration_op(const string& dataFile,     cubeData<float>* pImageCube) {} 
  virtual void migration3d_op(const string& fn_seisRec, const string& fn_img_off) {} //3d
  virtual void TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS) {}
  virtual void TomAdj3d_op (const string& fn_dImg_off, cubeData<float>& dslwCube) {}  //3d case
  virtual void TomoImit_op (const cubeData<float>* pRefl, const cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS) {}
  
private:
  virtual void initOperator(const waveOpParams* params, action_type ac_type) = 0;
};

/*
//declaration
 class SR_managerFD;      //Src/Recv Information manager, collect the info of src/receiver geometry, src wavelet and how to exert source
 class boundaryFD;  //how to propagate wavefield by one step; which consists of FDscheme, wavefiled damping and absorption at the boundary.
 class paddingFD;

//General Finite difference modeling operator
class modelingFDgen_op : public waveOps 
{
public:
   //data memter
   //declare friend class which are the component of the modelingFD operator
   friend class SR_managerFD;
   friend class boundaryFD;
   friend class paddingFD;

   
   SR_managerFD * m_pSR_manager;
   boundaryFD  *m_pBoundaryFD;
   paddingFD   *m_pPaddingFD;

   stdvecInt   mOffsetPandModel; //becoz shape of wavefield is bigger than shape of vel model by 2*stencilWid
   
private:
  waveOpParams* m_pars; //all modeling params.
  
protected:
   enum temporal_accuracy
   {
      order2, order4
   };
   temporal_accuracy m_t_accuracy;
   impose_receiver_record_type mImposeRecvType;
      
   
   cubeData<float>  mS0cube; //hold initial slowness per shot
   cubeData<float>  mReflCubePerShot; //ReflModel aligned to mShapeWavfield, inner object
   cubeData<float>  mReflOffCubePerShot; //ReflModel prestack , inner object
   //temporary memories for every shot
   cubeData<float>  mRTM_imageCubePerShot, mRTM_imageCubePerShot2;
   cubeData<float>  mSrcIntensePerShot, mRecvIntensePerShot;
   //For prestack imaging I(x,h)=p1(x-h)*p2(x+h)
   cubeData<float>  mRTM_offsetImgCubePerShot, mRTM_offsetImgCubePerShot2;
   
   
protected:
      bool m_bAssembled;
public:
      modelingFDgen_op(): waveOps(), m_pSR_manager(NULL), m_pBoundaryFD(NULL),
      m_pPaddingFD(NULL), 
      m_bAssembled(false) {} 
   
   //Assemble all the component like SR_manager and boundary, etc. 
   void assemble(SR_managerFD * pSR, boundaryFD* pBC, paddingFD* pPD) {
       m_pSR_manager = pSR; m_pBoundaryFD = pBC;  m_pPaddingFD = pPD;
	   m_bAssembled = true;
   }
   virtual ~modelingFDgen_op(){
   }
   
   //init the operator using the wave propagation params. e.g. dtComp
   
   //called before each invoking each _op func
   void checkInOutSize();
   virtual void initOperatorEx(waveOpParams* params, action_type ac_type);
   void unInitOperatorEx() {}
   
   virtual void modeling() {VERIFY(0);}
   virtual void modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec);
   virtual void RTM() = 0;
   virtual void migration_op(dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube);
   virtual void migration_op(const string& dataTag, cubeData<float>* pImageCube) {}
   virtual void adjRTM() = 0;
   virtual void bornModeling_op(cubeData<float>* pReflCube, dmRsf<float>* pRsfSeisRec);

   virtual void TomAdj();
   virtual void TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS);
   virtual void TomFwd();
   virtual void TomFwd_op(cubeData<float>* pDeltaS, cubeData<float>* pDeltaI);

   virtual void initPerShot(int ishot);
   
protected:
   virtual void initOperator(waveOpParams* params, action_type ac_type);
   virtual void unInitOperator() {}
   
   void genPerShotImage(int ishot);
   //Slices could be 2D or 3D
   void alignCropToSliceShapeGrid(cubeData<float>& outReflCube, cubeData<float>& inReflModel, hypercube& sliceShape);
   void cropWfldToVelGrid(cubeData<float>& velOutCube, cubeData<float>& wfldInCube);
   
   
   void getSnapshotShapePerShot(hypercube& shapeSnapshot);
   void getShotAxisForSnap(axis& ax_shot);
   bool shouldRecordSnap(int ishot, int* iIndexOnAxis=NULL);
   void setWavfieldShape(int iShot);
};

//CPU FD modeling
class modelingFD_op: public modelingFDgen_op {
   
protected:
   stdVecDb  stencil;  //FD spatial stencil

   wavFieldGroup   wfGp0; //wavefield group 1, default
   wavFieldGroup   wfGp1; //for the adjRTM modeling
   cubeData<ETYPE>      dpDdx2_cb; //cube used for storing 2nd order in space for wfld
   auto_ptr<cubeData<ETYPE> > m_pWavfieldCheckptPerShot;
   
   //For a slice of src wavefield, these three are per shot
   cubeData<float>   mSrcWavfieldCube;
   cubeData<ETYPE>   mCorrCube;
   
   //Just for Db->float transition.not necessary
   cubeData<float>    mSnapCube;
   

public:
      modelingFD_op(): modelingFDgen_op() {}
	  virtual ~modelingFD_op(){
	  }
   
   //Get the parameters for wave propagation operator. e.g. dt,dx,freq(center freq of wavlet),
   virtual void initOperator(waveOpParams* params, action_type ac_type);
   virtual void unInitOperator();   
   
   virtual void modeling();
   virtual void RTM();
   virtual void adjRTM();
   //virtual void modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec);
   //virtual void migration_op (dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube);

   virtual void initPerShot(int iShot);
   void modelingPerShot(int ishot);
   void adjRTM_perShot(int ishot);

   void unInit();   
   void imposeRecv(wavFieldGroup& wfg);
   void takeRecvWavfieldSnapshot(int ishot, int it, cubeData<ETYPE>& wfd, dmRsf<float>* pRecvSnapShot);
   void takeWavCheckpt(int ishot, int it, wavFieldGroup&);
   void correlateSrcRecv(int iShot, int it, const cubeData<ETYPE>& recvwfld);

protected:
   bool getWindowInSnapshot(windParam& window, int ishot, int it);
	bool getWindowInWavfieldCheckpt(windParam& window, int it);   
   void fwdSrcPropagation(int iShot);
   void setStencil(int _stenWid);

   virtual void propagate(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube);
   virtual void propagate_2t(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube) = 0;
   virtual void propagate_4t(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube) = 0; //4th order in time
   virtual void calcSpatialDeriv( cubeData<ETYPE> *in, cubeData<ETYPE> * out) = 0;

   void takeSnapshot(int ishot, int it, cubeData<ETYPE>& wavfield, dmRsf<float>* pSnapShot);
   virtual void corrWithOffset(cubeData<float>& offCubeOut, const cubeData<float>& wf1,const cubeData<float>& wf2) = 0;
};

*/

} //end of namespace


#endif

