#include "stdafx.h"
#include "decl_fwd.h"

#undef NDEBUG
#include <assert.h>

#include "param_func.h"
#include "cubeData.hpp"
#include "imaging.h"
#include "point2d.hpp"
#include "cubeAlgo.h"

#include "ADCIG_Recon.h"
#include "util_seplib.h"

#include "dmRsfDump.h"
#include "misc.h"

using namespace Yang;
using namespace std;
using namespace seplib;
using boost::shared_ptr;


int Param_RMO_Measurer::readParam(const param_func* par, const hypercube& img0offShp) {
  fThrLinearity       = par->get_float("thrLinearity",0.2f);
  fThrMinimumAmpRatio = par->get_float("fThrMinimumAmpRatio",0.2f);
    
  thrDipAngleX        = fwk::degree(par->get_float("fThrDipAngleX",20.0f));
  thrDipAngleY        = fwk::degree(par->get_float("fThrDipAngleY",20.0f));
  
  thrEventEnergyRatio_= par->get_float("thrEventEnergyRatio");//,0.1f); //amplitude means sqrt(ratio).
  thrNbTraceCorrCoef_ = par->get_float("thrNbTraceCorrCoef"); //,0.3f);
  thrCCmaxvalRatio_  = par->get_float("thrCCmaxvalRatio");    //,0.25f); //Used when finding RMO with cross-correlation method (CC with the zero-offset image).
  
  axis ax_z = img0offShp.get_n(1), ax_x = img0offShp.get_n(2);
  
  fEventWndLength  = par->get_float("fEventWndLen"); //the width of the wavelets at the reflectors.
  fEventMatchingWndLen = par->get_float("fEventMatchingWndLen");//The scanning range when tries to find a match.
  dRMO_smWidth    = fwk::degree(par->get_float("degRMO_smWidth"));
  
  waterDepthAver_     = par->get_float("confidentVelDepthAver");
  fApproxVelErrRatio_ = par->get_float("approxVelErrRatio");
  
  static pair<int,matchingMethod> method_match_table[] = {make_pair(0,CrossCorr),make_pair(1,EnvMax),make_pair(2,NbCorr),make_pair(3,dipTracking),make_pair(4,Semblance)};
  this->method_ = findEnumItemFromPar(par,method_match_table,NELEM(method_match_table),"RMOMatchingMethod",0);

  static pair<int,bpRMOdiffMethodType> bpRMOmethod_match_table[] = {make_pair(0,subtractAngStk),make_pair(1,subtract0ang),make_pair(2,diffNeighbor),make_pair(3,diff0ang),make_pair(4,diffAngStack),make_pair(5,diffNearAngStack),make_pair(6,diffFullVaraince)};
  this->bpRMOdiffMethod_ = findEnumItemFromPar(par,bpRMOmethod_match_table,NELEM(bpRMOmethod_match_table),"bpRMOmethod",5);
  PRV2(this->method_,this->bpRMOdiffMethod_);
  
  //read the semblance axis from the cmd line.
  if (this->method_ == Semblance) {
    read_axis_from_par(ax_semb_, par, "Semb");
    sembGaussDerivWidth_ = par->get_float("SembGaussDerivWidth",ax_semb_.n*ax_semb_.d);
    VERIFY(lroundf(-ax_semb_.o/ax_semb_.d) == (ax_semb_.n-1)/2);
  }

  return 0;
}

int ADCIG_ReconFull::inputGatherIs(const cubeDataF& img_hx_t, const cubeDataF* offMask) {
  //Here input an offset gather, the program will reconstruct the corresponding angle gather with CSensing tech.
  const Param_o2a* par_o2a = this->o2a_param();
  VERIFY(!par_o2a->b3D_);
  //currently, no CSensing, just convert offset to angle gathers.
  ///convert between (z,x,hx) offset image and (z,angle,x) angel domain image.
  const hypercube& img_hx_t_shape = img_hx_t.domain();
  VERIFY(img_hx_t_shape.get_ndim()==3);
  
  //prt("In ADCIG_Recon::convertAngleOffImage,");PRV(par_o2a->off2angType_); 
  //don't need to transpose now, Assume 2D gather is in axis order z,hx,x
  cubeDataF& img_ax_t = *recon_adcigs_;
  
  //Then convert to angle domain
  off2Ang2d_new(img_ax_t,img_hx_t,false, par_o2a->off2angType_,par_o2a->ax_gamma()); 
  //LOG("c3");  img_ax_t.info("img_ax_t");
  
  if (par_o2a->bOnewayCompensate_) { //divide the angle gather by cos^2(gamma) to balance the amplitude.
    axis_deg ax_p(img_ax_t.domain().get_axis(2));
    hypercube shp1(ax_p);
    cubeData<float> cosr(shp1);
    float fang;
    for (int ia=0; ia<ax_p.n; ia++) {
      fang = deg2rad(ax_p.get_coord(ia));
      cosr(ia) = ((1.0/cosf(fang))); //do less high angle compensation
    }
    //cosr.dump("dump10"); cosr.info("cosr");
    multiplySubCube(img_ax_t,img_ax_t,cosr,2,2);
  }
  
  LOG("c4");
  //now the angle gather is in recon_adcigs_;
  return 0;
}

ADCIG_ReconStomp::ADCIG_ReconStomp(const shared_ptr<const Param_o2a>& par_o2a, bool bUseAngleDomain, param_func* pars): ADCIG_Recon(par_o2a), bUseAngleDomain_(bUseAngleDomain) {
  //also read the additional params
  if (pars) {
    this->strStompPar_ = pars->get_string("add_recon_par");
    PRV(this->strStompPar_);
  }
}

static std::string strRecon("/homes/sep/yang/Research/beamWET2/bin/Stomp-ADCIG-Recon.x");


//Use CSensing to reconstrcut the angle gathers.
int ADCIG_ReconStomp::inputGatherIs(const cubeDataF& img_hx_t, const cubeDataF* offMask) {
  //Here input an offset gather, the program will reconstruct the corresponding angle gather with CSensing tech.
  int sysret;
  
  VERIFY(offMask);
  
  const Param_o2a* par_o2a = this->o2a_param();
  VERIFY(!par_o2a->b3D_);

  const hypercube& img_hx_t_shape = img_hx_t.domain();
  VERIFY(img_hx_t_shape.get_ndim()==3);
  
  prt("In ADCIG_Recon::convertAngleOffImage,");PRV(par_o2a->off2angType_); 
  //First transpose, assume 2D tranpose from z,x,hx -> z,hx,x, thus plane=23
  char szPars[1024],szCmd[1024]; string f1;
  ostringstream oss; int rsuf; rsuf = rand()%65536;
  oss.str(""); oss << "sepToolTemp" << rsuf <<"-hxt.H";   f1 = oss.str();
  img_hx_t.dump(f1,true);
  
  cubeDataF& img_ax_t = *recon_adcigs_;
  //Then convert to angle domain, using the CSensing program that we have.
  char sz1[256];
  char c_bAngle = '1';  if (!bUseAngleDomain_) c_bAngle='0';
  const axis& ax_g = this->o2a_param()->ax_gamma();
  sprintf(sz1," oang=%2.1f dang=%2.1f nang=%d ", ax_g.o, ax_g.d, ax_g.n);
  //generate three random fileNames
  string f_msk,f2,f3,f_out,f_log;
  oss.str(""); oss << str_tmpPath << "sepToolTemp" << rsuf <<"-msk.H"; f_msk = oss.str();
  oss.str(""); oss << str_tmpPath << "sepToolTemp" << rsuf <<"-dat.H"; f2    = oss.str();
  oss.str(""); oss << str_tmpPath << "sepToolTemp" << rsuf <<"-mod.H"; f3    = oss.str();
  oss.str(""); oss << str_tmpPath << "sepToolTemp" << rsuf <<"-aig.H"; f_out = oss.str();
  oss.str(""); oss << str_tmpPath << "log." << rsuf; f_log = oss.str();
  
  VERIFY(offMask!=NULL); offMask->dump(f_msk,true); //write the offset mask to file
  VERIFY(offMask->domain().sameSize(img_hx_t_shape)); //subsurf msk is transposed
  sprintf(szPars, " bAngleDomain=%c %s %s subSampMask=%s datDump=%s modDump=%s ", 
          c_bAngle,sz1,strStompPar_.c_str(),f_msk.c_str(),f2.c_str(),f3.c_str()); //create three random file names.
  sprintf(szCmd, "%s <%s %s >%s 2>%s", strRecon.c_str(),f1.c_str(),szPars,f_out.c_str(),f_log.c_str());
  PRV(szCmd); sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));

  //Now load the reconstructed adcig back
  img_ax_t.load(f_out,3);
  
  if (par_o2a->bOnewayCompensate_) { //divide the angle gather by cos^2(gamma) to balance the amplitude.
    axis_deg ax_p(img_ax_t.domain().get_axis(2));
    hypercube shp1(ax_p);
    cubeData<float> cosr(shp1);
    float fang;
    for (int ia=0; ia<ax_p.n; ia++) {
      fang = deg2rad(ax_p.get_coord(ia));
      cosr(ia) = (SQ(1.0/cosf(fang)));
    }
    //cosr.dump("dump10"); cosr.info("cosr");
    multiplySubCube(img_ax_t,img_ax_t,cosr,2,2);
  }
  //LOG("c4");
  //now the angle gather is in recon_adcigs_;
  return 0;
}

