#ifndef __ADCIG_RECON_H_INCL__
#define __ADCIG_RECON_H_INCL__

#include "stdafx.h"
#include "cubeData.hpp"
#include "my_types.h"
#include "decl_fwd.h"
#include "point2d.hpp"
#include "imaging.h"

const std::string envbin = "/data/sep/zyang03/bins/Envelope";

using boost::shared_ptr;
using seplib::AGC_Param;

namespace Yang {

// TODO: your code here
//Measure the moveout from an adcig gather.
class Param_o2a {
public:
  Param_o2a(const axis_deg& ax_gamma, const axis& ax_h, enumOff2angType o2atype=slantstackYangAA, bool b3d=false, bool bOnewayCompensate=true ): off2angType_(o2atype),b3D_(b3d),bOnewayCompensate_(bOnewayCompensate),ax_gamma_(ax_gamma),ax_h_(ax_h) {
  }
  
public:
  enumOff2angType off2angType_;
  bool b3D_;
  bool bOnewayCompensate_;
  void ax_gamma_is(const axis& ax_g) { ax_gamma_ = axis_deg(ax_g);}
  const axis& ax_gamma() const {return ax_gamma_;}
  void ax_h_is(const axis& ax_h) { ax_h_ = ax_h;}
  const axis& ax_h() const {return ax_h_;}
  
private:
  axis_deg  ax_gamma_;
  axis      ax_h_;
};


class ADCIG_Recon {
public:
  ADCIG_Recon(const shared_ptr<const Param_o2a>& par_o2a): par_o2a_(par_o2a), recon_adcigs_(new cubeDataF()), img_hx_(new cubeDataF()) {}
  virtual int inputGatherIs(const cubeDataF& img_hx_t,const cubeDataF* offMask) = 0; //image offset mask
  const cubeDataF& recon_adcig() const {return *recon_adcigs_;}
  
  const Param_o2a* o2a_param() const {return par_o2a_.get();}
  
  virtual ~ADCIG_Recon() {}
protected:
  shared_ptr<cubeDataF>       recon_adcigs_,img_hx_;
private:
  shared_ptr<const Param_o2a> par_o2a_;
};

class ADCIG_ReconFull: public ADCIG_Recon {
public:
  ADCIG_ReconFull(const shared_ptr<const Param_o2a>& par_o2a): ADCIG_Recon(par_o2a){}
  virtual int inputGatherIs(const cubeDataF& img_hx_t,const cubeDataF* offMask=NULL); //image offset mask
};

class ADCIG_ReconStomp: public ADCIG_Recon {
public:
  ADCIG_ReconStomp(const shared_ptr<const Param_o2a>& par_o2a, bool bUseAngleDomain=true, param_func* pars=NULL);
  virtual int inputGatherIs(const cubeDataF& img_hx_t,const cubeDataF* offMask); //image offset mask
private:
  bool bUseAngleDomain_; //Use angle domain as compression
  string strStompPar_;
};

struct Param_RMO_Measurer{
public:
  int readParam(const param_func* par, const hypercube& img0Shp);
public:
  ///for selecting candidate points.
  float         fThrLinearity,fThrMinimumAmpRatio;
  fwk::degree   thrDipAngleX,thrDipAngleY;
  PosFloat      fEventWndLength,fEventMatchingWndLen;
  fwk::degree   dRMO_smWidth;
  
  float thrEventEnergyRatio_,thrNbTraceCorrCoef_,thrCCmaxvalRatio_; //for moveout neighboring traces search
  //for an estimation of the amount of moveout, use formula ratio rho*tan^2(gamma)
  float waterDepthAver_; //the depth above which the velocity is regarded as correct.
  float fApproxVelErrRatio_; //esitmated slowness error ratio.
  
  typedef enum {CrossCorr,EnvMax,NbCorr,dipTracking,Semblance} matchingMethod; matchingMethod method_;
  axis ax_semb_;
  float sembGaussDerivWidth_; //for Guassian derivative
  typedef enum __bpRMOdiffMethod{ //the way to back-project the RMO information.
    subtractAngStk, //  subtract angle-stack
    subtract0ang,   //  subtract zero-angle
    diffNeighbor,   //  differencing neighboring moveout, actually back-project the curvature (1,-2,1).
    diff0ang,        //  differencing zero-angle moveout, will also back-project to sum-negative difference of the zero-angle.
    diffAngStack,
    diffNearAngStack,
    diffFullVaraince
  } bpRMOdiffMethodType;
  bpRMOdiffMethodType bpRMOdiffMethod_;
  
};

class RMO_Measurer {
public:
  RMO_Measurer(const shared_ptr<dipEstimator>& est, const shared_ptr<const Param_RMO_Measurer>& par, const shared_ptr<const AGC_Param>& AGCpar);
  
  //set a rayTomo operator to help RMO measurement, you can opt not setting it.
  void RayTomoOpIs(const shared_ptr<RayTomoOp>& rayTomo) {rayTomoOp_ = rayTomo;}
  
  int inputADCIG_is(const cubeDataF& img_ax, const cubeDataF& img0off);
  const cubeDataF& RMO () const {return *RMOsm_;}
  const cubeDataF& MoveoutPts () const {return *candPts_;}
  const std::valarray<int>& candPtsSel () const {return candPtsSel_;}
  const cubeDataF& RMO_ptsMask()  const  {return finalMask_;}
  const cubeDataF& RMOmask()      const  {return *RMO_mask_;}
  const cubeDataF& ptWeights()    const  {return *ptWeights_;}
  
  shared_ptr<cubeDataF> RMOplotted() const; //return a cube that draws the RMO measured.
  
  const Param_RMO_Measurer* RMO_Params() const {return par_.get();}
  const AGC_Param*          AGC_Params() const {return AGCpar_.get();}
  
private:
  void measureMoveout(const std::vector<point2dI>& vecCandPts);
  void genMoveoutForSensKernel(const vector<point2dI>& vecCandPts);
  
private:
  const cubeDataF *img_ax_,*img0off_;
  shared_ptr<cubeDataF> RMO_,RMOsm_,RMO_mask_,candPts_; //coordinates of points where RMO is measured.
  shared_ptr<cubeDataF> ptWeights_; //data weight at each RMO point
  std::valarray<int> candPtsSel_; //not all candPts get selected.
  NnegInt   nRMO_pts_;
  
  cubeDataF   envelop_,envelopBal_,img0offBal_,envelopNMSup_; //energy envelope, and its balanced version.
  cubeDataF   env_nmsup_; //after non-maximum suppression.
  cubeDataF   envMask_,structureMask_,finalMask_; //mask for selecting RMO measure candidates.
  stdVecFloat POI_; //point of interest on the image.

  shared_ptr<dipEstimator>              dipEstimator_;
  shared_ptr<RayTomoOp>                 rayTomoOp_;  
  shared_ptr<const Param_RMO_Measurer>  par_;
  shared_ptr<RMOfinder>         RMOfinder_;
  shared_ptr<const AGC_Param>     AGCpar_;
};

} //end namespace
#endif // __ADCIG_RECON_H_INCL__

