#include "stdafx.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "waveOps.h"
#include "paddingFD.h"
#include "SR_manager.h"
#include "boundaryFD.h"
#include "dampingFunc.h"
#include "FWI_FD.h"

using namespace Yang;

//Forward modeling
void FWI_FD::calcData()
{
   //Init the operator, do modeling to create calculated data this time
   waveOpParams& params = *m_FDModelingParam;
   waveOp_->initOperatorEx(&params, Yang::acModeling);
   
   //simulate the calc data from the current slowness s
   //verbPrt("\r simulate data from the current velocity s\n");
   m_pFWIparams->dCalCube.alloc(m_pVelAnaParam->pDataObs->domain());
   waveOp_->modeling_op(NULL,&(m_pFWIparams->dCalCube));   
   //m_pFWIparams->dCalCube.info("Calculate data"); //Note that it might blow up due to locally too high velocity
}

void FWI_FD::calcCorrelation(){   
  //Do cross-correlation
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   cubeData<float>& d_calCube=  m_pFWIparams->dCalCube;
   
   //After getting the calculated data P_cal, calculate the cross-correlation with P_obs.
   //correlate the data with the observed data, there should be a way to make sure the  
   //geometry is exactly the same for modeling data and observed data
   //need a little more frame work code on it.
   crossCorr1d(C_cube,d_calCube,d_obsCube,m_pFWIparams->iStartLag,m_pFWIparams->iEndLag);
}

void FWI_FD_classic::flipGradient(cubeData<float>& gradCube) {

   //Because the magic -1 phenomenon, flip the sign
   //gradCube.scale(-1.0f);
}

void FWI_FD_YL::flipGradient(cubeData<float>& gradCube) {
   //Because the magic -1 phenomenon, might need to flip the sign
}

//Calculate Gradient, this is a virtual function
void FWI_FD_classic::calcBackProjData() {
   const cubeData<float>& dObsCube = m_pVelAnaParam->dObsCube;
   m_pFWIparams->backProjCube.alloc(dObsCube.domain());
   m_pFWIparams->dObs_shiftedCube.alloc(dObsCube.domain());
   //m_pFWIparams->dObsCube.domain().info(""); m_pFWIparams->dCalCube.domain().info("");
   //m_pFWIparams->backProjCube.domain().info("out");
   //Simple subtraction
   m_pFWIparams->dObs_shiftedCube.subtract(m_pFWIparams->dCalCube,dObsCube);
   //m_pFWIparams->dObs_shiftedCube.dump("dump7");
   //Take second derivative !!! Now we only back propagte the calculated data.
   timeDeriv2nd(m_pFWIparams->backProjCube, m_pFWIparams->dObs_shiftedCube, true);
   m_pFWIparams->backProjCube.scale(-1.0f);
   //timeDeriv2nd(m_pFWIparams->backProjCube, m_pFWIparams->dCalCube, true);
   //prt("finish calcBackProjData in FD_classic\n");
   return;
}


//virtual function
float FWI_FD_classic::calcObjFunc() {
   //Use the model given in m_pFWI, calculate the objective function ||d_obs - d_cal||^2 for all xs,xg,t
   this->calcData();
   //then calc the cost function
   m_pFWIparams->dObs_shiftedCube.alloc(m_pFWIparams->dCalCube.domain());
   m_pFWIparams->dObs_shiftedCube.subtract(m_pFWIparams->dCalCube,m_pVelAnaParam->dObsCube);
   float costFunc = dotProd<float>(m_pFWIparams->dObs_shiftedCube,m_pFWIparams->dObs_shiftedCube);
   prt("finish calc Obj Func in FD_classic, obj = %f\n", costFunc);
   return costFunc;
}

//YI Luo's Method 1991, only works for transmission, for its ability to track only one single event
//Can only be called after calcObjFunc
void FWI_FD_YL::calcBackProjData() {
   const cubeData<float>& dObsCube = m_pVelAnaParam->dObsCube;
   m_pFWIparams->backProjCube.alloc(dObsCube.domain());

   //Here calculate the correlation panel, along with observed data, calculate the correlation panel, we've done it when evaluating the objective function
   //Then the maximum correlation at each trace of the data is also computed during evaulating objfunc
   cubeData<float>& maxPosCube = m_pFWIparams -> deltaTauCube;
   //int ndim = m_pFWIparams->C_cube.domain().get_ndim();
  
   m_pFWIparams->dObs_shiftedCube.alloc(dObsCube.domain());
   //m_pFWIparams->dObs_shiftedCube.domain().info("");
   
   
   staticShiftPerTrc(m_pFWIparams->dObs_shiftedCube, dObsCube, maxPosCube);
   //Take 2nd derivative
   m_pFWIparams->dObs2_shiftedCube.alloc(dObsCube.domain());
   timeDeriv2nd(m_pFWIparams->dObs2_shiftedCube,m_pFWIparams->dObs_shiftedCube, true);
   
   //Calc the 1/E term in the paper here for each xs and xg.
   m_pFWIparams->E_cube.alloc((maxPosCube.domain()));
   cubeData<float> tmpCube(m_pFWIparams->dObs2_shiftedCube.domain());
   tmpCube.multiply( m_pFWIparams->dObs2_shiftedCube, m_pFWIparams->dCalCube);
   sumAlongTimeTrc(m_pFWIparams->E_cube,tmpCube);

   m_pFWIparams->E_cube.scale(-1.0f);
   
   //Now calc the backProj data, = delta(Tau) * 2 / E * dObs''(t+deltaTau);
   maxPosCube.divide(m_pFWIparams->E_cube);
   //maxPosCube.scale(2.0f);
   timeDeriv1st(m_pFWIparams->dObs_shiftedCube, m_pFWIparams->dObs2_shiftedCube, true);
   m_pFWIparams->dObs_shiftedCube.copyTo(m_pFWIparams->backProjCube);
   scaleTrc(m_pFWIparams->backProjCube, maxPosCube);
}

//using objective function, \int \tau^2 d\tau C^2(\tau+\deltaTau), i.e focusing on the origin of \tau axis.
//But unlike beamfocusing, there is no intermediate moveout functions.
//no window, trace by trace.
void FWI_FD_YL2::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,iGroupMax;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   //Actually no need to keep subgroup, but just for generality
   //int subGroupSize = m_pVelAnaParam->beamSize;
   int subGroupSize = 1;
   
   cubeData<float> C_subTrc,C1_subTrc,C2_subTrc, t_subTrc;
   cubeData<float> trcCube2;
   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   hypercube shapeCtrc = C_cube.domain().return_compatible(1,1);
   
   //create a trace of \tau^2.
   cubeData<float> cubeTau2(shapeCtrc);
   axis ax_tau = shapeCtrc.get_axis(1);
   for (int i=0; i<ax_tau.n; i++) {
     cubeTau2[i] = SQ(ax_tau.o+(ax_tau.d*i));
   }
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
     C_subTrc.alloc(shapeCtrc);
     C1_subTrc.alloc(shapeCtrc);
     C2_subTrc.alloc(shapeCtrc);
     t_subTrc.alloc(shapeCtrc);
	  
	  //Calculate the number of beams we need, this time we make non overlap beams.
	  VERIFY( subGroupSize == 1);
	  iGroupMax = nRecv;
	  
	  for(iGroup=0; iGroup<iGroupMax; iGroup++)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      //grab a trace.
      C_cubePerShot.grabXd(&C_subTrc,subGroupWindow,1);
      C1_cubePerShot.grabXd(&C1_subTrc,subGroupWindow,1);
      C2_cubePerShot.grabXd(&C2_subTrc,subGroupWindow,1);

      C1_subTrc.zero();C1_subTrc.multiply(C_subTrc,cubeTau2);
      
		  //put C1_subcube back to the large cube
		  //Here we need to add, rather than replace, therefore first take the original data out		
		  C_backProjPerShot.updateXd(&C1_subTrc,subGroupWindow,1);
	  } //end for iGroup
    
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);

    //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
  } //for each shot
  //since it is minization rather than optimization, here we flip the sign
  d_injectCube.scale(-1.0f);
  return;
}

void FWI_FD_YL::postCalcData() {
   //Here calculate the correlation panel, along with observed data, calculate the correlation panel
   this->calcCorrelation();
   //Then find the maximum correlation at each trace of the data
   cubeData<float>& maxPosCube = m_pFWIparams->deltaTauCube;
   int ndim = m_pFWIparams->C_cube.domain().get_ndim();
   maxPosCube.alloc(m_pFWIparams->C_cube.domain().return_compatible(2,ndim));
   findMaxValTimePerTrc(maxPosCube, m_pFWIparams->C_cube);
   //maxPosCube.dump("dump8");
}

float FWI_FD_YL::calcObjFunc() {
   //Use the model given in m_pFWI, calculate the objective function deltaTau^2 for all xs,xg
   this->calcData();
   this->postCalcData();
   
   //then calc the cost function
   //should Tick out the outliers in maxPosCube.!
   cubeData<float>& maxPosCube = m_pFWIparams->deltaTauCube;
   float costFunc = dotProd<float>(maxPosCube, maxPosCube);
   //prt("finish calc Obj Func in FWI_FD_YL, obj = %f\n", costFunc);
   if (costFunc>100.0f) { //abnormal case 
     maxPosCube.dump("dump10");
     m_pFWIparams->dCalCube.dump("dump11");
     m_FDModelingParam->pVel->dump("dump12");
   }
   return costFunc;
}

//It is acutally slowness gradient.
void FWI_FD::calcSlowGradient( cubeData<float> & slowGradCube) {
  waveOpParams& params = *(m_FDModelingParam);
  VERIFY(slowGradCube.sameSize(*(params.pVel)));
   
  //Init the operator, doing RTM this time
	waveOp_->initOperatorEx(&params, Yang::acMigration);
  //prt("finish the initialization of the RTM Operator");

  //Do RTM to calculate gradient
  slowGradCube.zero();
   
  cubeData<float>& d_injectCube = m_pFWIparams -> backProjCube;   
  d_injectCube.resetCounter();
  //slowGradcube has to be in correct size
  waveOp_->migration_op(&d_injectCube,&slowGradCube);
  
  //Should uninit d_injectCube here
  //scaleVelGradient(velGradCube, *(params.pVel));
  scaleSlowGradient(slowGradCube, *(params.pVel));
  //reason of wrong sign, I could not figure out so far   
  this->flipGradient(slowGradCube);   
  return;
}


//!Not work. The kinematics of the backproj data does not agree with dcal
//Yang's objective function. J = sum_{[C(\tau)g(\tau)]^2}, g(\tau) is the Gaussian Derivative.
void FWI_FD_Yang::postCalcData() {
   //Here calculate the correlation panel, along with observed data, calculate the correlation panel
   this->calcCorrelation();
}

//Backproj data is the correlation of Pobs(t+\tau) and g(\tau) over \tau axis.
void FWI_FD_Yang::calcBackProjData() {
  //along with observed data, calculate the correlation panel
  cubeData<float>& C_cube =  m_pFWIparams->C_cube;
  cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
    
  axis ax_tau = C_cube.domain().get_axis(1);
  
  int iShot,nShots;
  cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
  
  //Assert receiver size match
  //int nRecv = C_cube.domain().get_n(2);
  
  cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
  
  //get Gaussian Derviative Func
  cubeData<float> gauss;
  float fGaussinWidth = m_pFWIparams->widthGaussDeriv;
  getGaussfunc(gauss,ax_tau,fGaussinWidth,0.0f,1);
   
  gauss.dump("dump3");
  
  int dimPerShot = 2;
  cubeData<float> d_obsCubePerShot;
  cubeData<float> C_cubePerShot;
  
  windParam shotWindow;
  shotWindow.resize(dimPerShot+1); shotWindow[dimPerShot].n = 1;
  
  //shape is a subcube the big data cube
  hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
  hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
  
  waveOpParams& params = *(m_FDModelingParam);
  d_injectCube.alloc(d_obsCube.domain());
  nShots = params.m_nShots;
   
  cubeDataF wtCube(C_cube.domain().return_compatible(2,3)); //For each src,recv
  cubeDataF wtPershot; wtPershot.becomeProxyCube(wtCube.domain().return_axes(1,1));
  windParam windWt(2); windWt[1].n = 1;
  
  for(iShot=0;iShot< nShots; iShot++)
  {
    d_obsCubePerShot.alloc(shapeDataPerShot);
    //aux cubes, remember that finally a correlation cube is injected
    d_injectCubePerShot.alloc(shapeDataPerShot);
    d_injectCubePerShot2.alloc(shapeDataPerShot);
    
    C_cubePerShot.alloc(shapeCorrPerShot);
    
    //read in d_obsCubePerShot
    shotWindow[dimPerShot].f = iShot;
    
    d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
    C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
    
    windWt[1].f = iShot;
    wtCube.setProxyXd(&wtPershot, windWt, 1);
    //Calc C(\tau)g'(\tau) d\tau
    dotProdAlongAxis(wtPershot,C_cubePerShot,gauss,1);
    
    //We need to calculat INT{g(\tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
    int nt = d_obsCube.domain().get_n(1);
    int tStartLag = 0; int tEndLag = nt-1;
    
    scaleSlice(d_obsCubePerShot,d_obsCubePerShot,wtPershot);
    tStartLag = 0; tEndLag = nt - 1;
    crossCorr1dGeneral(d_injectCubePerShot,gauss,d_obsCubePerShot,tStartLag,tEndLag);
    
    //back project this data, (take 2nd derivative) because the born-modeling formula.
    timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);
    
    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
  } //for each shot
  //since it is minization rather than optimization, here we flip the sign
  d_injectCube.scale(-1.0f);
  
  wtCube.dump("wt");
  
  return;
}

float FWI_FD_Yang::calcObjFunc() {
  return 0.0f;
}

