#ifndef __FWI_FD_H_INCL__
#define __FWI_FD_H_INCL__

#include "velAna.h"

namespace Yang {

class FWI_FD : public velAna_FD {
public:
	// Constructor
	FWI_FD(velAnaParam* pVelAnaParam,
         const shared_ptr<waveOps>& pModel, FWI_Param * pPar):
	   velAna_FD(pVelAnaParam,pModel),m_pFWIparams(pPar){}

	// Destructor
	virtual ~FWI_FD();

	virtual void calcData();
	virtual void calcBackProjData() = 0;
	void calcSlowGradient(cubeData<float>& slowGradCube);
	
	//calculate objective function, used in solver
	virtual float calcObjFunc() = 0;
  
  virtual void postCalcData() {} //calculate things after calculating the data
protected:
	virtual void flipGradient(cubeData<float>& gradCube) {}
  void calcCorrelation(); //calculate the correlation between d_obs and d_cal
public:
  FWI_Param *m_pFWIparams;
};

// Destructor implementation
inline FWI_FD::~FWI_FD()
{
}

class FWI_FD_classic : public FWI_FD {
public:
   FWI_FD_classic(velAnaParam* pVelAnaParam,
				  const shared_ptr<waveOps>& pModel, FWI_Param * pPar): 
	  FWI_FD(pVelAnaParam,pModel,pPar) {}
   
   virtual void calcBackProjData();
   virtual float calcObjFunc();
protected:
	virtual void flipGradient(cubeData<float>& gradCube);
};

class FWI_FD_YL : public FWI_FD {
public:
      FWI_FD_YL(velAnaParam* pVelAnaParam,
                const shared_ptr<waveOps>& pModel, FWI_Param * pPar): 
	  FWI_FD(pVelAnaParam,pModel,pPar) {}

   virtual void calcBackProjData();
   virtual float calcObjFunc();
   virtual void postCalcData();
protected:
	virtual void flipGradient(cubeData<float>& gradCube);
};

//Same objective func as beamFocus13, but without the intermediate moveout params.
class FWI_FD_YL2 : public FWI_FD_YL {
public:
      FWI_FD_YL2(velAnaParam* pVelAnaParam,
                 const shared_ptr<waveOps>& pModel, FWI_Param * pPar): 
        FWI_FD_YL(pVelAnaParam,pModel,pPar) {}
      virtual void calcBackProjData();
protected:
};


//Yang's objective function. J = sum_{[C(\tau)g(\tau)]^2}, g(\tau) is the Gaussian Derivative.
class FWI_FD_Yang : public FWI_FD {
public:
      FWI_FD_Yang(velAnaParam* pVelAnaParam,
                  const shared_ptr<waveOps>& pModel, FWI_Param * pPar): 
      FWI_FD(pVelAnaParam,pModel,pPar) {}

   virtual void   calcBackProjData();
   virtual float  calcObjFunc();
   virtual void   postCalcData();
protected:
	 //virtual void flipGradient(cubeData<float>& gradCube);
};

}

#endif // __FWI_FD_H_INCL__
