#include "stdafx.h"

#include "dmRsfDump.h"
#include "cubeData.hpp"

#include "velFD2d.h"
#include "waveOps.h"
#include "dampingFunc.h"
#include "FWI_FD_DSO.h"

using namespace Yang;

void FWI_FD_DSO::flipGradient(cubeData<float>& gradCube) {
   //Because the magic -1 phenomenon, flip the sign
   gradCube.scale(-1.0f);
}

/**
void FWI_FD_DSO1::flipGradient(cubeData<float>& gradCube) {
   //Because the magic -1 phenomenon, flip the sign
   //gradCube.scale(-1.0f);
} 
**/ 


//DSO obj func, measure the square sum of the first derivative over x axis on the correlation panel
float FWI_FD_DSO::calcObjFunc() {

   //Use the model given in m_pFWI, calculate the objective function deltaTau^2 for all xs,xg
   this->calcData();

   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //Take the DSO, calculate the field of (dC/dx_g)
   cubeData<float> dCdxCube;
   bool bScaleBy_dx = true;
   dCdxCube.alloc(C_cube.domain());
   spatialXderiv1st(dCdxCube, C_cube, bScaleBy_dx);
   
   //Take the square sum of dC/dx_g
   float fObjFuncVal = 0.0f;
   
   float dx = C_cube.domain().get_d(2);
   fObjFuncVal += dotProd<float>(dCdxCube,dCdxCube) * dtau * dx;
   
   //debug
   dCdxCube.dump("dCdx");
   return fObjFuncVal;
}

void FWI_FD_DSO::postCalcData() {
  this->calcCorrelation();
}

//DSO in transmission case
void FWI_FD_DSO1::calcBackProjData() {

   //Assume the correlation is been calculated correctly
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube 	  =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   //After getting the calculated data P_cal, calculate the cross-correlation with P_obs.
   //we've done the correlation in function evaluation term, so it could be omitted
   float dt,dtau;
   dt = C_cube.domain().get_d(1); dtau = dt;
   
   //take the 2nd spatail derivative of C
   cubeData<float> Cx2_cube;
   Cx2_cube.alloc(C_cube.domain());   
   deriv2nd(Cx2_cube,C_cube,2,true);
   //spatialXderiv2nd(Cx2_cube,C_cube,true);
   Cx2_cube.dump("dCdx2");
   //Cx2_cube.dump("dump3");

   int iShot,nShots;
   
   cubeData<float>& d_injectCube = m_pFWIparams -> backProjCube;
   d_injectCube.alloc(d_obsCube.domain());
      
   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;

   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam shotWindow;
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   
   waveOpParams& params = *(m_FDModelingParam);
   nShots = params.m_nShots;

   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
     //aux cubes, remember that finally a correlation cube is injected
     d_injectCubePerShot.alloc(shapeDataPerShot);
     d_injectCubePerShot2.alloc(shapeDataPerShot);

     C_backProjPerShot.alloc(shapeCorrPerShot); C_backProjPerShot.zero();
	
     //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
     d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     Cx2_cube.grabXd(&C_backProjPerShot,shotWindow,dimPerShot);
      
	   //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	   //Notice that a small section of the correlation cube is truncated.
	   int nt = d_obsCube.domain().get_n(1);
	   //Assemble the back injected data by cross-correlate C2x with observed data cube dtau*C1(tau)*P(tau+t)
	   int tStartLag = 0, tEndLag = nt - 1;
	   crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);

     timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);
     //Put the data back to a large cube
     d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
   } //for each shot   
   return;
}


//DSO2 in transmission case, just change the 2nd time deriv to 1st time deriv
void FWI_FD_DSO2::calcBackProjData() {

   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube 	  =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   //After getting the calculated data P_cal, calculate the cross-correlation with P_obs.
   //we've done the correlation in function evaluation term, so it could be omitted
   float dt,dtau;
   dt = C_cube.domain().get_d(1); dtau = dt;
   
   //take the 2nd spatail derivative of C
   cubeData<float> Cx2_cube;
   Cx2_cube.alloc(C_cube.domain());   
   spatialXderiv2nd(Cx2_cube,C_cube,true);
   Cx2_cube.dump("dCdx2");
   //Cx2_cube.dump("dump3");

   int iShot,nShots;
   
   cubeData<float>& d_injectCube = m_pFWIparams -> backProjCube;
   d_injectCube.alloc(d_obsCube.domain());
      
   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;

   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam shotWindow;
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   
   waveOpParams& params = *(m_FDModelingParam);
   nShots = params.m_nShots;

   for(iShot=0;iShot< nShots; iShot++)
   {
      d_obsCubePerShot.alloc(shapeDataPerShot);
	  //aux cubes, remember that finally a correlation cube is injected
	  d_injectCubePerShot.alloc(shapeDataPerShot);
	  d_injectCubePerShot2.alloc(shapeDataPerShot);

	  C_backProjPerShot.alloc(shapeCorrPerShot); C_backProjPerShot.zero();
	
	  //read in d_obsCubePerShot
    shotWindow[dimPerShot].n = 1;
    shotWindow[dimPerShot].f = iShot;
      
	  d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
    Cx2_cube.grabXd(&C_backProjPerShot,shotWindow,dimPerShot);
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  //Assemble the back injected data by cross-correlate C2x with observed data cube dtau*C1(tau)*P(tau+t)
	  int tStartLag = 0, tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  timeDeriv1st(d_injectCubePerShot2,d_injectCubePerShot,true);
    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
   } //for each shot   
   return;
}

