#include "stdafx.h"

#include "dmRsfDump.h"
#include "cubeData.hpp"

#include "velFD2d.h"
#include "waveOps.h"
#include "dampingFunc.h"
#include "interp.h"
#include "FWI_FD_beamFocus.h"

using namespace Yang;

void FWI_FD_beamFocus::flipGradient(cubeData<float>& gradCube) {
   //Because the magic -1 phenomenon, flip the sign
   gradCube.scale(-1.0f);
}

void FWI_FD_beamFocusY::flipGradient(cubeData<float>& gradCube) {
   //Because the magic -1 phenomenon, flip the sign
   //gradCube.scale(-1.0f);
}

//calc the power of the stacked trace after moveout by a,b parameters
float FWI_FD_beamFocus::findStackPowerAfterMoveout(cubeData<float>& outCube, cubeData<float>& stackCube, const cubeData<float>& inCube, 
												   const cubeData<float>& fh, const cubeData<float>& gh, cubeData<float>& moveOutCube, float a, float b) {
   float ret = 0.0f;
   float dt = inCube.domain().get_d(1);
   assert(outCube.sameSize(inCube));
   //We need to shift input cube based on our picking parameters.
   moveOutCube.scale(fh,a);  moveOutCube.scaleAdd(gh,b);
   //Dump shiftCube should here.
   
   //Shift the Correlation Panel
   staticShiftPerTrcArbSinc(outCube, inCube, moveOutCube);
   //stacked trace
   float dh = inCube.domain().get_d(2);
   stackTrc(stackCube,outCube); stackCube.scale(dh);
   //calculate the power
   ret = dotProd<float>(stackCube, stackCube) * dt;
   return ret;
}


void FWI_FD_beamFocus::postCalcData() {
  this->calcCorrelation();
}

//Biondo's moveout method
float FWI_FD_beamFocus::calcObjFunc() {
   //Use the model given in m_pFWI, calculate the objective function deltaTau^2 for all xs,xg
   this->calcData();

   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1); dtau = dt;
   
   //stack the subarray for each location, need to do it one by one beam
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   //axis for local offset
   float o,d; int n;
   d = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-d);
   axis ax_offset(n,o,d,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   int dimPerShot = 2;
   cubeData<float> C_cubePerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);

   waveOpParams& params = *(m_FDModelingParam);
   int nShots = params.m_nShots;
   
   cubeData<float> C_subCube,C_shiftSubCube, trcCube;
   float fObjFuncVal = 0.0f;
   
   //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
	VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);

	//for each subgroup, Keep in mind there is overlop for each point
	int offsetBeam = subGroupSize / 2;
	cubeData<float> dThetaOverDu,subCubeShift, DJDUcube;
	dThetaOverDu.alloc(offsetShape);
	subCubeShift.alloc(offsetShape);
	hypercube shape = C_cube.domain().return_compatible(2,C_cube.domain().get_ndim());
	DJDUcube.alloc(shape);
	velAna::setMoveoutDeriv(dThetaOverDu,m_pVelAnaParam->moveOutType);

	 //Let the second pt move just 1dt. there fore
	 float du = dt/(dThetaOverDu[offsetBeam/2]);
	 subCubeShift.scale(dThetaOverDu,du);


   for(int iShot=0;iShot< nShots; iShot++)
   {

	  C_cubePerShot.alloc(shapeCorrPerShot);
	  //read in d_obsCubePerShot
      shotWindow[dimPerShot].n = 1;
      shotWindow[dimPerShot].f = iShot;
      
	  int nGroup,iGroup, nRecv;
	  nRecv = C_cubePerShot.domain().get_n23();
	  nGroup = nRecv - 2 * int(subGroupSize/2);

	  VERIFY(nGroup > 0);

      C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
	 
	  C_subCube.alloc(subCubeShape); C_shiftSubCube.alloc(subCubeShape);
      trcCube.alloc(hypercube(vector<axis> (1,(C_subCube.domain().get_axis(1)))));

	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {
         //subdivide the correlation panel C(tau;s0)
         //Read in C_subCube, C1_subCube, C2_subCube
         subGroupWindow[dimPerShot-1].n = subGroupSize;
         subGroupWindow[dimPerShot-1].f = iGroup;

         C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);

		 //sum the traces along the subarray offset
         trcCube.zero();
		 stackTrc(trcCube, C_subCube);
		 
		 //Take squares and integrate along tau axis
		 float stackPower = dotProd<float>(trcCube,trcCube);
		 fObjFuncVal += stackPower;
		 
		 /*
		 //Here let's use secant method to find the derivative DJ/DU at each point.
		 //Shift the trace first
		 staticShiftPerTrc(C_shiftSubCube, C_subCube, subCubeShift);
		 trcCube.zero(); stackTrc(trcCube, C_shiftSubCube);
		 //Take squares and integrate along tau axis
		 float stackPower2 = dotProd<float>(trcCube,trcCube);
		 float djdu = (stackPower2-stackPower) / du;
		 if (iGroup+offsetBeam==256) {
			C_shiftSubCube.dump("dump7");
			C_subCube.dump("dump6");
			trcCube.dump("dump5");
			prt("\n stackpower 1,2 =%f,%f",stackPower,stackPower2);
		 }
		 
		 DJDUcube(iShot,iGroup+offsetBeam) = djdu;
		 */
	  }
   } //for each shot
   
   //Dump DJDU cube
   //DJDUcube.dump("DJDU");
   //To make it negative, so that it becomes from maximization to minimization
   return -1.0*fObjFuncVal;
}

//Beam focus, Biondi 2010, SEP 140
void FWI_FD_beamFocus1::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube 	  =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   //cubeData<float>& d_calCube =  m_pFWIparams->dCalCube;
   
   //After getting the calculated data P_cal, calculate the cross-correlation with P_obs.
   //we've done the correlation in function evaluation term, so it could be omitted

   float dt,dtau;
   dt = C_cube.domain().get_d(1); dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);


   int iShot,nShots,iGroup,nGroup;
   
   cubeData<float>& d_injectCube = m_pFWIparams -> backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> trcCube,trcCube2;

   cubeData<float> dThetaOverDu;//Use to save array like D(theta)/D(u)

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;

   float termB2;
   cubeData<float> termB2cube, termHcube, termEcube; //a scalar for each xs,xg location.
   hypercube termHShape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termB2cube.alloc(termHShape); termHcube.alloc(termHShape); termEcube.alloc(termHShape);
   
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   dThetaOverDu.alloc(offsetShape);
   
   velAna::setMoveoutDeriv(dThetaOverDu,m_pVelAnaParam->moveOutType);
   //setDthetaOverDu(dThetaOverDu,Linear);
   //dThetaOverDu.dump("dump3");
   
   float denom, weight;
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;

   for(iShot=0;iShot< nShots; iShot++)
   {
      d_obsCubePerShot.alloc(shapeDataPerShot);
	  //aux cubes, remember that finally a correlation cube is injected
	  d_injectCubePerShot.alloc(shapeDataPerShot);
	  d_injectCubePerShot2.alloc(shapeDataPerShot);

	  C_cubePerShot.alloc(shapeCorrPerShot);
	  C_backProjPerShot.alloc(shapeCorrPerShot); C_backProjPerShot.zero();
	  C1_cubePerShot.alloc(shapeCorrPerShot);
	  C2_cubePerShot.alloc(shapeCorrPerShot);
	
	  //for dump	
	  cubeData<float> C_backProjPerShotDump1, C_backProjPerShotDump2, d_injectCubePerShotDump;
	  C_backProjPerShotDump1.alloc(C_cubePerShot.domain());	  C_backProjPerShotDump2.alloc(C_cubePerShot.domain());
	  d_injectCubePerShotDump.alloc(d_injectCubePerShot.domain());

      
	  //read in d_obsCubePerShot
    shotWindow[dimPerShot].n = 1;
    shotWindow[dimPerShot].f = iShot;
    
    d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
    C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
    C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
    C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
	  //d_obsCubePerShot.dump("dump3");
	  //C_cubePerShot.dump("dump4");
	  
      C_subCube.alloc(subCubeShape);
      C1_subCube.alloc(subCubeShape);
      C2_subCube.alloc(subCubeShape);
      t_subCube.alloc(subCubeShape);
	  
      trcCube.alloc(hypercube(vector<axis> (1,(C_subCube.domain().get_axis(1)))));
      trcCube2.alloc(hypercube(vector<axis> (1,(C_subCube.domain().get_axis(1)))));

	  //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  nGroup = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {
         //subdivide the correlation panel C(tau;s0)
         //Read in C_subCube, C1_subCube, C2_subCube
         subGroupWindow[dimPerShot-1].n = subGroupSize;
         subGroupWindow[dimPerShot-1].f = iGroup;

         C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
         C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
         C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		 //sum the traces along the subarray offset
         trcCube.zero(); stackTrc(trcCube, C_subCube); trcCube.scale(dh);
         trcCube2.zero();C1_subCube.copyTo(t_subCube);scaleTrc(t_subCube,dThetaOverDu);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
		 
		 //sum along the tau axis
		 termB2 = dotProd<float>(trcCube,trcCube2)*dtau;
		 //!! Just work in two D case
		 termB2cube(iShot, iGroup+offsetBeam) = termB2;
		 //term B2 is acutall DJ_local / Dv(x)
		 
		 //Now we have one scaler  for each subgroup location x_bar
		 //Calculate the denominator
		 //calculate the 2nd derivative C''(tau;s0)
		 //scale the C''(tau;s0) by D(theta)/D(u) along the local offset, twice
		 //C2_subCube.dump("dump2");
		 scaleTrc(C2_subCube,dThetaOverDu);scaleTrc(C2_subCube,dThetaOverDu); 
		 denom = dotProd<float>(C2_subCube,C_subCube) * dtau * dh;
		 //Dot prod the previous and the C(tau;s0)

		 //then we calculate the weight term
		 weight = termB2/denom;
		 termEcube(iShot, iGroup+offsetBeam) = denom;
		 termHcube(iShot, iGroup+offsetBeam) = weight;
		 
		 //scale the C'(tau;s0) by offset using the scaler we just obtained.
		 scaleTrc(C1_subCube, dThetaOverDu);
		 C1_subCube.scale(weight);
		 
		 //put C1_subcube back to the large cube
		 //Here we need to add, rather than replace, therefore first take the original data out

		 //dump the back projected data
		 if(iGroup==0) {
			C_backProjPerShotDump1.updateSeveralXd(&C1_subCube,subGroupWindow,1);
		 }
		 if (iGroup == nGroup/2) {
			C_backProjPerShotDump2.updateSeveralXd(&C1_subCube,subGroupWindow,1);
			C1_subCube.dump("dump1");  
		 }

		 C_backProjPerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		 C_subCube.add(C1_subCube);
		 C_backProjPerShot.updateSeveralXd(&C_subCube,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  
	  
	  //prt("RMS = %f \n", C_cubePerShot.RMS());
	  	  
	  //termEcube.dump("dump7");
	  //termWeightCube.dump("dump5");
	  
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  
	  //for dump
	  crossCorr1dGeneral(d_injectCubePerShotDump,C_backProjPerShotDump1,d_obsCubePerShot,tStartLag,tEndLag);
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShotDump,true);
	  d_injectCubePerShotDump.dump("bd-l");  C_backProjPerShotDump1.dump("dump2");
	  crossCorr1dGeneral(d_injectCubePerShotDump,C_backProjPerShotDump2,d_obsCubePerShot,tStartLag,tEndLag);
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShotDump,true);
	  d_injectCubePerShotDump.dump("bd-m");  C_backProjPerShotDump2.dump("dump4");
	  
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  //prt("in taper edge");
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  d_injectCubePerShot(i,j) *= 0.0f; //dampingCoef[subGroupSize-1-i];
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 d_injectCubePerShot(i,j) *= 0.0f; //dampingCoef[i-(nx-subGroupSize)];
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
   } //for each shot
   
   termHcube.dump("H");
   termB2cube.dump("DJDU");
   termEcube.dump("E");
   return;
}
 
//scan moveout parameters a and b, and save them to picked curvature cube
void FWI_FD_beamFocus::scanMoveoutParmas(const axis& ax_a_in, const axis& ax_b_in) {

   cubeData<float>& C_cube = m_pFWIparams->C_cube;
   cubeData<float>& aCube = m_pVelAnaParam->pickedCurvatureCubeSubsamp;
   cubeData<float>& bCube = m_pVelAnaParam->pickedSlopeCubeSubsamp;
   int subGroupSize = m_pVelAnaParam->beamSize;   
   VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
   //for each subgroup, Keep in mind there is overlop for each point
   int offsetBeam = subGroupSize / 2;
   int scanSubsampRate = subGroupSize/2; //picking is expensive, thus save some time.

   //Set the size of pickParam Cube
   vecAxis paramAxes = C_cube.domain().return_axes(3, C_cube.domain().get_ndim()); //One param for each x_s,x_g
   axis ax_x = C_cube.domain().get_axis(2);
   //Downsample x axis, remember starts from offsetBeam to ax.n-2*offsetBeam
   int n = (ax_x.n-2*offsetBeam-1)/scanSubsampRate + 1;
   axis ax_x_new(n,ax_x.o+offsetBeam*ax_x.d,ax_x.d*scanSubsampRate);
   paramAxes.insert(paramAxes.begin(),ax_x_new);
   hypercube pickedParamShape(paramAxes);
   aCube.alloc(pickedParamShape); bCube.alloc(pickedParamShape);
   
   int dimPerShot = 2;
   windParam shotWindow;   shotWindow.resize(dimPerShot+1);
   windParam subGroupWindow; subGroupWindow.resize(dimPerShot);
   cubeData<float> C_cubePerShot;
   //shape is a subcube the big data cube
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   
   //Then for each shot calculate the theoretic moveout Param based on :
   //1. the receiver/src location;  2. the constant true velocity and initial velocity.
   

   waveOpParams& params = *(m_FDModelingParam);
   int nShots = params.m_nShots;
   
   //Calculate the number of beams we need, every x location there is a beam except the points on the edge.

   //create a scanned moveout parameters panel, it has a 2D plane for each beam location x_s,x_g.
   axis ax_a, ax_b;
   //needs to !!! define the range of param a/b here !!!!!!!!
   ax_a = ax_a_in; ax_b = ax_b_in;
   
   vector<axis> axesBeam = C_cube.domain().return_axes(2,C_cube.domain().get_ndim());
   vector<axis> axesScan; axesScan.push_back(ax_a); axesScan.push_back(ax_b);
   hypercube shapeScanPanel(axesScan);
   axesScan.insert(axesScan.end(),axesBeam.begin(),axesBeam.end());
   hypercube shapeScanCube(axesScan);
   
   cubeData<float> scanPanel; scanPanel.alloc(shapeScanPanel);
   cubeData<float> scanCube; scanCube.alloc(shapeScanCube);
   
   
   //axis for local offset
   float o,dh;
   dh = C_cube.domain().get_d(2); n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   //Shape for subwindow cube of Correlation
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   hypercube trcShape = C_cube.domain().return_compatible(1,1);
   
   cubeData<float> fh,gh, moveoutCube;
   fh.alloc(offsetShape); gh.alloc(offsetShape); moveoutCube.alloc(offsetShape);
   velAna::setMoveoutDeriv(fh,Quadratic);
   velAna::setMoveoutDeriv(gh,Linear);
   
   cubeData<float> C_subCube, C_subCubeShift;
   cubeData<float> stackCube;
   float a,b, fStackPower;
   
   for(int iShot=0;iShot< nShots; iShot++) {
    verbPrt("\rstart scan param , ishot=%d",iShot);
	  C_cubePerShot.alloc(shapeCorrPerShot);
	  
	  int nRecv = C_cubePerShot.domain().get_n23();
	  int nGroup = nRecv - 2 * offsetBeam;
	  VERIFY(nGroup > 0);
	  
	  //Take the data from each shot
	  shotWindow[dimPerShot].n = 1;
	  shotWindow[dimPerShot].f = iShot;
	  C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
	  
    C_subCube.alloc(subCubeShape);    C_subCubeShift.alloc(subCubeShape);
	  stackCube.alloc(trcShape); 
	  
	  int n_a = shapeScanCube.get_n(1); int n_b = shapeScanCube.get_n(2);
	  //C_cubePerShot.dump("dump4");
    
	  //for each beam (receiver) location, get the windowed data first
	  for(int iGroup=0; iGroup<nGroup; iGroup+=scanSubsampRate) {
         subGroupWindow[dimPerShot-1].n = subGroupSize;
         subGroupWindow[dimPerShot-1].f = iGroup;
         C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		 //for each point on the param (a,b) plane, calculate the objective function 
		 //and Find the (a,b) corresponding to the maximum stack power in scanPanel
		 float fmaxPower = 0.0f; 
		 int max_ia = -1; int max_ib=-1;
		 for (int ib=0; ib<n_b; ib++) {
			for (int ia=0; ia<n_a; ia++)
			{
			   a = shapeScanCube.get_coord(ia,1) ; b = shapeScanCube.get_coord(ib,2);
			   fStackPower = findStackPowerAfterMoveout(C_subCubeShift, stackCube, C_subCube, fh,gh, 
															moveoutCube, a, b);
			   scanPanel(ib,ia) = fStackPower;
			   if (fmaxPower < fStackPower) {
				  max_ia = ia; max_ib = ib;
				  fmaxPower = fStackPower;
			   }
			}
		 }
		 assert(max_ia != -1 && max_ib != -1);
		 a = shapeScanCube.get_coord(max_ia,1); b = shapeScanCube.get_coord(max_ib,2);
     //prt("\n a=%f,b=%f",a,b);
		 aCube(iShot,(iGroup/scanSubsampRate)) = a;
		 bCube(iShot,(iGroup/scanSubsampRate)) = b;
		 
		 //put it into scanCube
		 int nScanCubeDim = scanCube.domain().get_ndim();
		 windParam scanPanelWindow; scanPanelWindow.resize(nScanCubeDim);
		 scanPanelWindow[nScanCubeDim-1].f = iShot;  scanPanelWindow[nScanCubeDim-1].n = 1;
		 scanPanelWindow[nScanCubeDim-2].f = iGroup+offsetBeam;  scanPanelWindow[nScanCubeDim-2].n = 1;
		 scanCube.updateXd(&scanPanel,scanPanelWindow,2);
	  }
    //C_cubePerShot.dump("dump4");
   } //for each shot
   
   //Linear interp the result
   cubeData<float>& aCubeInterp = m_pVelAnaParam->pickedCurvatureCube;
   cubeData<float>& bCubeInterp = m_pVelAnaParam->pickedSlopeCube;
   seplib::interp::getInstance()->interpolate(aCubeInterp,aCube,
                                              &ax_x.n,&ax_x.o,&ax_x.d,
                                              NULL,NULL,NULL,NULL,NULL,NULL,
                                              seplib::linear);
   seplib::interp::getInstance()->interpolate(bCubeInterp,bCube,
                                              &ax_x.n,&ax_x.o,&ax_x.d,
                                              NULL,NULL,NULL,NULL,NULL,NULL,
                                              seplib::linear);
   //bCube.dump("dump2");
   //bCubeInterp.dump("dump3");
   //dump scancube result;
   //scanCube.dump("scanCube");
}

//Beam focus using another objective function, \int \tau^2 d\tau C^2(\tau+\deltaTau) and only shift moveout param is used
//no window, trace by trace.
void FWI_FD_beamFocus13::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,iGroupMax;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   //Actually no need to keep subgroup, but just for generality
   //int subGroupSize = m_pVelAnaParam->beamSize;
   int subGroupSize = 1;
   
   cubeData<float> C_subTrc,C1_subTrc,C2_subTrc, t_subTrc;
   cubeData<float> trcCube2;
   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,F11;
   float A2;	float H1;
     
   cubeData<float> termA2cube;
   cubeData<float> EmatCube,FmatCube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termA2cube.alloc(termA2Shape);
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes);
   
   //axis for local offset
   hypercube shapeCtrc = C_cube.domain().return_compatible(1,1);
   
   //create a trace of \tau^2.
   cubeData<float> cubeTau2(shapeCtrc);
   axis ax_tau = shapeCtrc.get_axis(1);
   for (int i=0; i<ax_tau.n; i++) {
     cubeTau2[i] = SQ(ax_tau.o+(ax_tau.d*i));
   }
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
     C_subTrc.alloc(shapeCtrc);
     C1_subTrc.alloc(shapeCtrc);
     C2_subTrc.alloc(shapeCtrc);
     t_subTrc.alloc(shapeCtrc);
	  
	  dmRsf_ocDump<float> Adump("trcA","A trace"); dmRsf_ocDump<float> A1dump("trcA1","A1 trace"); 
	  Adump.init(&shapeCtrc);  A1dump.init(&shapeCtrc);  
	  
	  //Calculate the number of beams we need, this time we make non overlap beams.
	  VERIFY( subGroupSize == 1);
	  iGroupMax = nRecv;
	  
    //Calculate trace A first, because A trace is a global stack over all locations.
	  for(iGroup=0; iGroup<iGroupMax; iGroup++)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      //grab a trace.
      C_cubePerShot.grabXd(&C_subTrc,subGroupWindow,1);
      C1_cubePerShot.grabXd(&C1_subTrc,subGroupWindow,1);
      C2_cubePerShot.grabXd(&C2_subTrc,subGroupWindow,1);

		  //calc A term and E term.
		  //A is C*C1*tau^2 dtau
      t_subTrc.multiply(C_subTrc,C1_subTrc);
      //sum along the tau axis
      A2 = dotProd<float>(t_subTrc,cubeTau2) * dtau;
		 	A2 = A2*1e9; //balance the value amplitude.	 
		  //!! Just work in two D case
		  termA2cube(iShot, iGroup) = A2;
		  //term A2 is acutall DJ_local / Da, B2 is DJ_local / Db
		 
		  //Calculate E and F 2X2 matrix.
		  E11 = dotProd<float>(C2_subTrc,C_subTrc) * dtau;
      
		  EmatCube(0,iShot,iGroup) = E11;
		  EmatCube(1,iShot,iGroup) = 0.0f;
		  EmatCube(2,iShot,iGroup) = 0.0f;
		 
      F11 = 1.0f/E11;
		  FmatCube(0,iShot,iGroup) = F11;
		  FmatCube(1,iShot,iGroup) = 0.0f;
		  FmatCube(2,iShot,iGroup) = 0.0f;
		 
		  //then we calculate H1 H2 Term.
		  H1 = A2*F11;
      
		  C1_subTrc.scale(H1);
		  //put C1_subcube back to the large cube
		  //Here we need to add, rather than replace, therefore first take the original data out		
		  C_backProjPerShot.updateXd(&C1_subTrc,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		}
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  //save trc A,A1,B1
	  Adump.unInit(); A1dump.unInit(); 
  } //for each shot
  //since it is minization rather than optimization, here we flip the sign
  d_injectCube.scale(-1.0f);
  
	termA2cube.dump("A2");
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
  return;
}

//Beam focus using another objective function, every trc is aligned to the neighboring one that is
//close to the zero angle, or center trace.
// only shift moveout param is used
//no window, trace by trace.
void FWI_FD_beamFocus15::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,iGroupMax;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   //Actually no need to keep subgroup, but just for generality
   //int subGroupSize = m_pVelAnaParam->beamSize;
   int subGroupSize = 1;
   
   cubeData<float> C_subTrc,C1_subTrc,C2_subTrc, Cref_subTrc, t_subTrc;
   cubeData<float> trcCube2;
   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,F11;
   float A2;	float H1;
     
   cubeData<float> termA2cube;
   cubeData<float> EmatCube,FmatCube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termA2cube.alloc(termA2Shape);
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes);
   
   //axis for local offset
   hypercube shapeCtrc = C_cube.domain().return_compatible(1,1);
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   //Need to figure out shot location
   float fShotLoc; int iShotLocOnReceiverAxis;
   axis ax_shot = C_cube.domain().get_axis(3);
   axis ax_recv = C_cube.domain().get_axis(2);
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
     fShotLoc = ax_shot.get_coord(iShot);
     //Figure out what offset corresponds to on the receiver axis.
     iShotLocOnReceiverAxis = ax_recv.get_offset(fShotLoc);
     PRV(iShotLocOnReceiverAxis);
     
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
     C_subTrc.alloc(shapeCtrc);
     C1_subTrc.alloc(shapeCtrc);
     C2_subTrc.alloc(shapeCtrc);
     t_subTrc.alloc(shapeCtrc);
     Cref_subTrc.alloc(shapeCtrc);
	  
	  dmRsf_ocDump<float> Adump("trcA","A trace"); dmRsf_ocDump<float> A1dump("trcA1","A1 trace"); 
	  Adump.init(&shapeCtrc);  A1dump.init(&shapeCtrc);  
	  
	  //Calculate the number of beams we need, this time we make non overlap beams.
	  VERIFY( subGroupSize == 1);
	  iGroupMax = nRecv;
	  
    //Calculate trace A first, because A trace is a global stack over all locations.
	  for(iGroup=0; iGroup<iGroupMax; iGroup++)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      //grab a trace.
      C_cubePerShot.grabXd(&C_subTrc,subGroupWindow,1);
      C1_cubePerShot.grabXd(&C1_subTrc,subGroupWindow,1);
      C2_cubePerShot.grabXd(&C2_subTrc,subGroupWindow,1);

		  //calc A2 term and E term.
      bool bLeft = iGroup<iShotLocOnReceiverAxis;  //figure out where the trace sits.
      //grab the neighboring trace for this one
      if (bLeft) {
        subGroupWindow[dimPerShot-1].f = iGroup+1;
      } else if (iGroup==iShotLocOnReceiverAxis) { //do nothing
        subGroupWindow[dimPerShot-1].f = iGroup;
      } else {
        subGroupWindow[dimPerShot-1].f = iGroup-1;
      }
      C_cubePerShot.grabXd(&Cref_subTrc,subGroupWindow,1);
      //set the window param back
      subGroupWindow[dimPerShot-1].f = iGroup;
      
		  //A is the C'Cref d\tau
      A2 = dotProd<float>(Cref_subTrc,C1_subTrc) * dtau;
		 	A2 = A2*1e9; //balance the value amplitude.	 
		  //!! Just work in two D case
		  termA2cube(iShot, iGroup) = A2;
		  //term A2 is acutall DJ_local / Da, B2 is DJ_local / Db
		 
		  //Calculate E and F 2X2 matrix.
		  E11 = dotProd<float>(C2_subTrc,C_subTrc) * dtau;
      
		  EmatCube(0,iShot,iGroup) = E11;
		  EmatCube(1,iShot,iGroup) = 0.0f;
		  EmatCube(2,iShot,iGroup) = 0.0f;
		 
      F11 = 1.0f/E11;
		  FmatCube(0,iShot,iGroup) = F11;
		  FmatCube(1,iShot,iGroup) = 0.0f;
		  FmatCube(2,iShot,iGroup) = 0.0f;
		 
		  //then we calculate H1 H2 Term.
		  H1 = A2*F11;
      
		  C1_subTrc.scale(H1);
		  //put C1_subcube back to the large cube
		  //Here we need to add, rather than replace, therefore first take the original data out		
		  C_backProjPerShot.updateXd(&C1_subTrc,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		}
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  //save trc A,A1,B1
	  Adump.unInit(); A1dump.unInit(); 
  } //for each shot
   
	termA2cube.dump("A2");
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
  return;
}


