#ifndef  FWI_FD_BEAMFOCUS_H
#define  FWI_FD_BEAMFOCUS_H

#include "FWI_FD.h"

namespace Yang {
   
class FWI_FD_beamFocus : public FWI_FD {
public:
   FWI_FD_beamFocus(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge)
   :FWI_FD(pVelAnaParam, pFDModelingParam,pModel, pPar), m_bTaperEdge(bTaperEdge) {}
   
   virtual void calcBackProjData() = 0;
   virtual float calcObjFunc();
   virtual void postCalcData();

   void scanMoveoutParmas(const axis& ax_a_in, const axis& ax_b_in);
public:
	bool m_bTaperEdge;
	
protected:
    float findStackPowerAfterMoveout(cubeData<float>& outCube, cubeData<float>& stackCube, const cubeData<float>& inCube, 
												   const cubeData<float>& fh, const cubeData<float>& gh, cubeData<float>& moveOutCube, float a, float b);
	virtual void flipGradient(cubeData<float>& gradCube);
};

//Use single parameter fitting
class FWI_FD_beamFocus1 : public FWI_FD_beamFocus {
public:
   FWI_FD_beamFocus1(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   
   virtual void calcBackProjData();
   //virtual float calcObjFunc();
public:

protected:
};

//The global objective function changed to account for global shift
//Beam focus, single parameter fitting 2010. Use only slope
//Window overlap
class FWI_FD_beamFocus12 : public FWI_FD_beamFocus {
public:
   FWI_FD_beamFocus12(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   
   virtual void calcBackProjData();
   //virtual float calcObjFunc();
public:

protected:
};

class FWI_FD_beamFocus13 : public FWI_FD_beamFocus {
public:
   FWI_FD_beamFocus13(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   
   virtual void calcBackProjData();
   //virtual float calcObjFunc();
public:

protected:
};

class FWI_FD_beamFocus14 : public FWI_FD_beamFocus {
public:
   FWI_FD_beamFocus14(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   
   virtual void calcBackProjData();
   //virtual float calcObjFunc();
public:

protected:
};

class FWI_FD_beamFocus15 : public FWI_FD_beamFocus {
public:
   FWI_FD_beamFocus15(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   
   virtual void calcBackProjData();
   //virtual float calcObjFunc();
public:

protected:
};

//Use double parameter fitting
class FWI_FD_beamFocus2 : public FWI_FD_beamFocus {
public:
   FWI_FD_beamFocus2(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   virtual void calcBackProjData();
   //virtual float calcObjFunc();
protected:
};

//use double parameter fitting, but use shift and slope this time
//objective function is local flatness, old obj func
//aux obj function utilize two params, both SHIFT and slope.
class FWI_FD_beamFocus20: public FWI_FD_beamFocus{
public:
   FWI_FD_beamFocus20(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   virtual void calcBackProjData();
protected:
};

//use double parameter fitting, but use shift and slope this time
class FWI_FD_beamFocus21: public FWI_FD_beamFocus{
public:
   FWI_FD_beamFocus21(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   virtual void calcBackProjData();
   virtual float calcObjFunc();
protected:
};

//use double parameter fitting, but use shift and slope this time
class FWI_FD_beamFocus22: public FWI_FD_beamFocus{
public:
   FWI_FD_beamFocus22(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   virtual void calcBackProjData();
   //virtual float calcObjFunc()
protected:
};


//Use double parameter fitting, and measures the moveOutParam
class FWI_FD_beamFocusY : public FWI_FD_beamFocus{
public:
   FWI_FD_beamFocusY(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocus(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   virtual float calcObjFunc();
   virtual void calcBackProjData() = 0;
   virtual void pickMoveoutParmas();
   virtual void postCalcData();
public:
   float m_vTrue;
   float m_v0;
protected:
   virtual void flipGradient(cubeData<float>& gradCube);
};

class FWI_FD_beamFocusY1 : public FWI_FD_beamFocusY{
public:
   FWI_FD_beamFocusY1(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocusY(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}

   virtual void calcBackProjData();
   virtual void calc_dJdb();
protected:
};

//use single moveout param, but minimize the average shift
class FWI_FD_beamFocusY1sh : public FWI_FD_beamFocusY1{
public:
   FWI_FD_beamFocusY1sh(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocusY1(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}
   virtual void calc_dJdb();
protected:
};

class FWI_FD_beamFocusY2 : public FWI_FD_beamFocusY{
public:
   FWI_FD_beamFocusY2(velAnaParam* pVelAnaParam, waveOpParams* pFDModelingParam, 
                    modelingFDgen_op * pModel, FWI_Param * pPar, bool bTaperEdge):
     FWI_FD_beamFocusY(pVelAnaParam,pFDModelingParam,pModel,pPar,bTaperEdge)
   {}

   virtual void calcBackProjData();
protected:
};

}
#endif
