#include "stdafx.h"

#include "dmRsfDump.h"
#include "cubeData.hpp"

#include "velFD2d.h"
#include "waveOps.h"
#include "dampingFunc.h"
#include "interp.h"
#include "FWI_FD_beamFocus.h"

using namespace Yang;

//Beam focus, double parameter fitting 2010. Use both line parameter and curvature parameter to fit.
void FWI_FD_beamFocus2::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube 		=  m_pFWIparams->C_cube;
	 
	 velAnaParam* pVelAnaParam 	=	velAnaPar();
   cubeData<float>& d_obsCube = pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,nGroup;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> trcCube,trcCube2;

   cubeData<float> fhCube,ghCube; //For storing the moveOut function h^2 and h.
   cubeData<float> HfgCube; //Storing H1*f(h)+H2*g(h)

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,E12,E22, F11,F12,F22;
   float A2,B2;	float H1,H2;
     
   cubeData<float> termA2cube, termB2cube;
   cubeData<float> EmatCube,FmatCube, H12cube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termA2cube.alloc(termA2Shape);
   termB2cube.alloc(termA2Shape);
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes); //store H1 and H2
   H12cube.alloc(H12cubeShape);
   
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   fhCube.alloc(offsetShape);  ghCube.alloc(offsetShape);  HfgCube.alloc(offsetShape);
   
   velAna_FD::setMoveoutDeriv(fhCube,Quadratic);
   velAna_FD::setMoveoutDeriv(ghCube,Linear);
   //fhCube.dump("fh"); ghCube.dump("gh");
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
      C_subCube.alloc(subCubeShape);
      C1_subCube.alloc(subCubeShape);
      C2_subCube.alloc(subCubeShape);
      t_subCube.alloc(subCubeShape);
	  
	  hypercube corrTrcShape(vector<axis> (1,(C_subCube.domain().get_axis(1))));
      trcCube.alloc(corrTrcShape);  trcCube2.alloc(corrTrcShape);
	  dmRsf_ocDump<float> Adump("trcA","A trace"); dmRsf_ocDump<float> A1dump("trcA1","A1 trace"); dmRsf_ocDump<float> B1dump("trcB1","B1 trace");
	  Adump.init(&corrTrcShape);  A1dump.init(&corrTrcShape);  B1dump.init(&corrTrcShape);
	  dmRsf_ocDump<float> HfgDump("Hfg","Hfg");
	  HfgDump.init(&offsetShape);
	  dmRsf_ocDump<float> CsubDump("Csub",""), C1subDump("C1sub",""), C2subDump("C2sub","");
	  CsubDump.init(&subCubeShape); C1subDump.init(&subCubeShape); C2subDump.init(&subCubeShape);
	  
	  //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  nGroup = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {
         //subdivide the correlation panel C(tau;s0)
         //Read in C_subCube, C1_subCube, C2_subCube
         subGroupWindow[dimPerShot-1].n = subGroupSize;
         subGroupWindow[dimPerShot-1].f = iGroup;

         C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
         C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
         C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		 //Calculate A and A1, B1 trace and A2 B2 term.
		 //sum the traces along the subarray offset
         trcCube.zero(); stackTrc(trcCube, C_subCube); trcCube.scale(dh);
		 
         trcCube2.zero();C1_subCube.copyTo(t_subCube); scaleTrc(t_subCube,fhCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
		 //sum along the tau axis
		 A2 = dotProd<float>(trcCube,trcCube2) * dtau;
		 //prt("term H=%f \n", termH);
		 //Dump A(tau) and A1(tau);
		 if(iGroup==0 || iGroup == nGroup/2) {
			CsubDump.dumpFrame(C_subCube); C1subDump.dumpFrame(C1_subCube); C2subDump.dumpFrame(C2_subCube);
			Adump.dumpFrame(trcCube); A1dump.dumpFrame(trcCube2);
		 }
		 
		 trcCube2.zero();C1_subCube.copyTo(t_subCube);scaleTrc(t_subCube,ghCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
		 if(iGroup==0 || iGroup == nGroup/2) {
			  B1dump.dumpFrame(trcCube2);
		 }
		 
		 //sum along the tau axis
		 B2 = dotProd<float>(trcCube,trcCube2) * dtau;
		 
		 //!! Just work in two D case
		 termA2cube(iShot, iGroup+offsetBeam) = A2;
		 termB2cube(iShot, iGroup+offsetBeam) = B2; 
		 //term A2 is acutall DJ_local / Da, B2 is DJ_local / Db
		 
		 //Calculate E and F 2X2 matrix.
		 C2_subCube.copyTo(t_subCube);
		 scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,fhCube);
		 //Dot prod the previous and the C(tau;s0)
		 E11 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 C2_subCube.copyTo(t_subCube);
		 scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,ghCube);
		 //Dot prod the previous and the C(tau;s0)
		 E12 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 C2_subCube.copyTo(t_subCube);
		 scaleTrc(t_subCube,ghCube); scaleTrc(t_subCube,ghCube);
		 //Dot prod the previous and the C(tau;s0)
		 E22 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 
		 EmatCube(0,iShot,iGroup+offsetBeam) = E11;
		 EmatCube(1,iShot,iGroup+offsetBeam) = E12;
		 EmatCube(2,iShot,iGroup+offsetBeam) = E22;
		 
		 //solve E 2X2 matrix to get F
		 float tmpF[4]; float tmpE[4]; float determ, cond;
		 tmpE[0] = E11; tmpE[1]=tmpE[2]=E12; tmpE[3] = E22;
		 cond = invert2X2(tmpF,tmpE,determ);
		 if (cond > 1E3) {
			verbPrt("\n cond number is very big %f in beamFocus2!! E=%.3e,%.3e,%.3e,%.3e", cond, E11, E12, E12, E22);
		 }
		 F11 = tmpF[0]; F12 = tmpF[1]; F22 = tmpF[3];

		 FmatCube(0,iShot,iGroup+offsetBeam) = F11;
		 FmatCube(1,iShot,iGroup+offsetBeam) = F12;
		 FmatCube(2,iShot,iGroup+offsetBeam) = F22;
		 
		 //then we calculate H1 H2 Term.
		 H1 = A2*F11 + B2*F12; 	H2 = A2*F12 + B2*F22;
		 //prt("term weight H/E =%f \n", weight);
		 H12cube(0,iShot,iGroup+offsetBeam) = H1;
		 H12cube(1,iShot,iGroup+offsetBeam) = H2;
		 		 
		 //calculate the scale cube H1*f(h)+H2*g(h)
		 HfgCube.scale(fhCube,H1); HfgCube.scaleAdd(ghCube,H2);
		 if(iGroup==0 || iGroup == nGroup/2) {
				HfgDump.dumpFrame(HfgCube);
		 }
		 scaleTrc(C1_subCube, HfgCube);
		 //put C1_subcube back to the large cube
		 //Here we need to add, rather than replace, therefore first take the original data out		
		 C_backProjPerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		 C_subCube.add(C1_subCube);
		 C_backProjPerShot.updateSeveralXd(&C_subCube,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  //d_injectCubePerShot(i,j) *= dampingCoef[subGroupSize-1-i];
				  d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 //d_injectCubePerShot(i,j) *= dampingCoef[i-(nx-subGroupSize)];
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

      //Put the data back to a large cube
      d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  //save trc A,A1,B1
	  Adump.unInit(); A1dump.unInit(); B1dump.unInit();
	  HfgDump.unInit(); CsubDump.unInit(); C1subDump.unInit(); C2subDump.unInit();
      //Sum the gradient over all shots.
   } //for each shot
	
	termA2cube.dump("A2"); termB2cube.dump("B2");
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
	H12cube.dump("H12");
   return;
}

//objective function is local flatness, old obj func
//aux obj function utilize two params, both SHIFT and slope.
void FWI_FD_beamFocus20::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,nGroup;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> trcCube,trcCube2;

   cubeData<float> fhCube,ghCube; //For storing the moveOut function h^2 and h.
   cubeData<float> HfgCube; //Storing H1*f(h)+H2*g(h)

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,E12,E22, F11,F12,F22;
   float A2,B2;	float H1,H2;
     
   cubeData<float> termA2cube, termB2cube;
   cubeData<float> EmatCube,FmatCube, H12cube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termA2cube.alloc(termA2Shape);
   termB2cube.alloc(termA2Shape);
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes); //store H1 and H2
   H12cube.alloc(H12cubeShape);
   
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   fhCube.alloc(offsetShape);  ghCube.alloc(offsetShape);  HfgCube.alloc(offsetShape);
   
   //!Here we have the linear (slope) and shift.
   velAna_FD::setMoveoutDeriv(fhCube,Linear);
   velAna_FD::setMoveoutDeriv(ghCube,constant);
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
      C_subCube.alloc(subCubeShape);
      C1_subCube.alloc(subCubeShape);
      C2_subCube.alloc(subCubeShape);
      t_subCube.alloc(subCubeShape);
	  
	  hypercube corrTrcShape(vector<axis> (1,(C_subCube.domain().get_axis(1))));
      trcCube.alloc(corrTrcShape);  trcCube2.alloc(corrTrcShape);
	  dmRsf_ocDump<float> Adump("trcA","A trace"); dmRsf_ocDump<float> A1dump("trcA1","A1 trace"); dmRsf_ocDump<float> B1dump("trcB1","B1 trace");
	  Adump.init(&corrTrcShape);  A1dump.init(&corrTrcShape);  B1dump.init(&corrTrcShape);
	  dmRsf_ocDump<float> HfgDump("Hfg","Hfg");
	  HfgDump.init(&offsetShape);
	  dmRsf_ocDump<float> CsubDump("Csub",""), C1subDump("C1sub",""), C2subDump("C2sub","");
	  CsubDump.init(&subCubeShape); C1subDump.init(&subCubeShape); C2subDump.init(&subCubeShape);
	  
	  //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  nGroup = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {
         //subdivide the correlation panel C(tau;s0)
         //Read in C_subCube, C1_subCube, C2_subCube
         subGroupWindow[dimPerShot-1].n = subGroupSize;
         subGroupWindow[dimPerShot-1].f = iGroup;

         C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
         C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
         C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		 //Calculate A and A1, B1 trace and A2 B2 term.
		 //sum the traces along the subarray offset
         trcCube.zero(); stackTrc(trcCube, C_subCube); trcCube.scale(dh);
		 
         trcCube2.zero();C1_subCube.copyTo(t_subCube); scaleTrc(t_subCube,fhCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
		 //sum along the tau axis
		 A2 = dotProd<float>(trcCube,trcCube2) * dtau;
		 //prt("term H=%f \n", termH);
		 //Dump A(tau) and A1(tau);
		 if(iGroup==0 || iGroup == nGroup/2) {
			CsubDump.dumpFrame(C_subCube); C1subDump.dumpFrame(C1_subCube); C2subDump.dumpFrame(C2_subCube);
			Adump.dumpFrame(trcCube); A1dump.dumpFrame(trcCube2);
		 }
		 
		 trcCube2.zero();C1_subCube.copyTo(t_subCube);scaleTrc(t_subCube,ghCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
		 if(iGroup==0 || iGroup == nGroup/2) {
			  B1dump.dumpFrame(trcCube2);
		 }
		 
		 //sum along the tau axis
		 B2 = dotProd<float>(trcCube,trcCube2) * dtau;
		 
		 //!! Just work in two D case
		 termA2cube(iShot, iGroup+offsetBeam) = A2;
		 termB2cube(iShot, iGroup+offsetBeam) = B2; 
		 //term A2 is acutall DJ_local / Da, B2 is DJ_local / Db
		 
		 //Calculate E and F 2X2 matrix.
		 C2_subCube.copyTo(t_subCube);
		 scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,fhCube);
		 //Dot prod the previous and the C(tau;s0)
		 E11 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 C2_subCube.copyTo(t_subCube);
		 scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,ghCube);
		 //Dot prod the previous and the C(tau;s0)
		 E12 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 C2_subCube.copyTo(t_subCube);
		 scaleTrc(t_subCube,ghCube); scaleTrc(t_subCube,ghCube);
		 //Dot prod the previous and the C(tau;s0)
		 E22 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 
		 EmatCube(0,iShot,iGroup+offsetBeam) = E11;
		 EmatCube(1,iShot,iGroup+offsetBeam) = E12;
		 EmatCube(2,iShot,iGroup+offsetBeam) = E22;
		 
		 //solve E 2X2 matrix to get F
		 float tmpF[4]; float tmpE[4]; float determ, cond;
		 tmpE[0] = E11; tmpE[1]=tmpE[2]=E12; tmpE[3] = E22;
		 cond = invert2X2(tmpF,tmpE,determ);
		 if (cond > 1E3) {
			verbPrt("\n cond number is very big %f in beamFocus2!! E=%.3e,%.3e,%.3e,%.3e", cond, E11, E12, E12, E22);
		 }
		 F11 = tmpF[0]; F12 = tmpF[1]; F22 = tmpF[3];

		 FmatCube(0,iShot,iGroup+offsetBeam) = F11;
		 FmatCube(1,iShot,iGroup+offsetBeam) = F12;
		 FmatCube(2,iShot,iGroup+offsetBeam) = F22;
		 
     //The dJ/db(shift) i.e the B2 term should be zero. Therefore
     B2 = 0.0f;
		 //then we calculate H1 H2 Term.
		 H1 = A2*F11 + B2*F12; 	H2 = A2*F12 + B2*F22;
		 //prt("term weight H/E =%f \n", weight);
		 H12cube(0,iShot,iGroup+offsetBeam) = H1;
		 H12cube(1,iShot,iGroup+offsetBeam) = H2;
		 		 
		 //calculate the scale cube H1*f(h)+H2*g(h)
		 HfgCube.scale(fhCube,H1); HfgCube.scaleAdd(ghCube,H2);
		 if(iGroup==0 || iGroup == nGroup/2) {
				HfgDump.dumpFrame(HfgCube);
		 }
		 scaleTrc(C1_subCube, HfgCube);
		 //put C1_subcube back to the large cube
		 //Here we need to add, rather than replace, therefore first take the original data out		
		 C_backProjPerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		 C_subCube.add(C1_subCube);
		 C_backProjPerShot.updateSeveralXd(&C_subCube,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  //d_injectCubePerShot(i,j) *= dampingCoef[subGroupSize-1-i];
				  d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 //d_injectCubePerShot(i,j) *= dampingCoef[i-(nx-subGroupSize)];
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

      //Put the data back to a large cube
      d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  //save trc A,A1,B1
	  Adump.unInit(); A1dump.unInit(); B1dump.unInit();
	  HfgDump.unInit(); CsubDump.unInit(); C1subDump.unInit(); C2subDump.unInit();
      //Sum the gradient over all shots.
   } //for each shot
	
	termA2cube.dump("A2"); termB2cube.dump("B2");
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
	H12cube.dump("H12");
   return;
}

float FWI_FD_beamFocus21::calcObjFunc() {
  //!! Need imple
  return 1.0f;
}

//The global objective function changed to account for global shift
//Beam focus, single parameter fitting 2010. Use only shift
//Window overlap
//result it gives is the same as beamFocus21 and/or beamFocus22, So slope has a very small
//impact compared with shift itself
void FWI_FD_beamFocus12::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,iGroupMax;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> trcCube2,trcA;

   cubeData<float> fhCube; //For storing the moveOut function h^2 and h.
   cubeData<float> HfgCube; //Storing H1*f(h)+H2*g(h)

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,F11;
   float A2;	float H1;
     
   cubeData<float> termA2cube, termB2cube;
   cubeData<float> EmatCube,FmatCube, H12cube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termA2cube.alloc(termA2Shape);
   termB2cube.alloc(termA2Shape);
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes); //store H1 and H2
   H12cube.alloc(H12cubeShape);
   
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   fhCube.alloc(offsetShape);
   HfgCube.alloc(offsetShape);
   
   //Here change it to SHIFT only
   velAna_FD::setMoveoutDeriv(fhCube,constant);
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
     C_subCube.alloc(subCubeShape);
     C1_subCube.alloc(subCubeShape);
     C2_subCube.alloc(subCubeShape);
     t_subCube.alloc(subCubeShape);
	  
	   hypercube corrTrcShape(vector<axis> (1,(C_subCube.domain().get_axis(1))));
     trcCube2.alloc(corrTrcShape); trcA.alloc(corrTrcShape);
    
	  dmRsf_ocDump<float> Adump("trcA","A trace"); dmRsf_ocDump<float> A1dump("trcA1","A1 trace"); 
	  Adump.init(&corrTrcShape);  A1dump.init(&corrTrcShape);  
	  dmRsf_ocDump<float> HfgDump("Hfg","Hfg");
	  HfgDump.init(&offsetShape);
	  dmRsf_ocDump<float> CsubDump("Csub",""), C1subDump("C1sub",""), C2subDump("C2sub","");
	  CsubDump.init(&subCubeShape); C1subDump.init(&subCubeShape); C2subDump.init(&subCubeShape);
	  
	  //Calculate the number of beams we need, this time we make non overlap beams.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  iGroupMax = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
    //Calculate trace A first, because A trace is a global stack over all locations.
    trcA.zero(); stackTrc(trcA, C_cubePerShot);
	  for(iGroup=0; iGroup<iGroupMax; iGroup++)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
      C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
      C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		  //calc A1, B1 trace and A2 B2 term.
		  //sum the traces along the subarray offset
      trcCube2.zero();C1_subCube.copyTo(t_subCube); scaleTrc(t_subCube,fhCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
      //sum along the tau axis
      A2 = dotProd<float>(trcA,trcCube2) * dtau;
		  //prt("term H=%f \n", termH);
		  //Dump A(tau) and A1(tau);
		  if(iGroup==0 || fabs(float(iGroup - iGroupMax/2)/subGroupSize)<1.0f) {
        CsubDump.dumpFrame(C_subCube); C1subDump.dumpFrame(C1_subCube); C2subDump.dumpFrame(C2_subCube);
        Adump.dumpFrame(trcA); A1dump.dumpFrame(trcCube2);
      }
		 		 
		  //!! Just work in two D case
		  termA2cube(iShot, iGroup+offsetBeam) = A2;
		  termB2cube(iShot, iGroup+offsetBeam) = 0.0f; 
		  //term A2 is acutall DJ_local / Da, B2 is DJ_local / Db
		 
		  //Calculate E and F 2X2 matrix.
		  C2_subCube.copyTo(t_subCube);
		  scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,fhCube);
		  //Dot prod the previous and the C(tau;s0)
		  E11 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
      
		  EmatCube(0,iShot,iGroup+offsetBeam) = E11;
		  EmatCube(1,iShot,iGroup+offsetBeam) = 0.0f;
		  EmatCube(2,iShot,iGroup+offsetBeam) = 0.0f;
		 
      F11 = 1.0f/E11;
		  FmatCube(0,iShot,iGroup+offsetBeam) = F11;
		  FmatCube(1,iShot,iGroup+offsetBeam) = 0.0f;
		  FmatCube(2,iShot,iGroup+offsetBeam) = 0.0f;
		 
		  //then we calculate H1 H2 Term.
		  H1 = A2*F11;
		  //prt("term weight H/E =%f \n", weight);
		  H12cube(0,iShot,iGroup+offsetBeam) = H1;
		  H12cube(1,iShot,iGroup+offsetBeam) = 0.0f;
		 		 
		  //calculate the scale cube H1*f(h)+H2*g(h)
		  HfgCube.scale(fhCube,H1);
		  if(iGroup==0 || iGroup == iGroupMax/2) {
		 		HfgDump.dumpFrame(HfgCube);
		  }
		  scaleTrc(C1_subCube, HfgCube);
		  //put C1_subcube back to the large cube
		  //Here we need to add, rather than replace, therefore first take the original data out		
		  C_backProjPerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		  C_subCube.add(C1_subCube);
		  C_backProjPerShot.updateSeveralXd(&C_subCube,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  //d_injectCubePerShot(i,j) *= dampingCoef[subGroupSize-1-i];
				  d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 //d_injectCubePerShot(i,j) *= dampingCoef[i-(nx-subGroupSize)];
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  //save trc A,A1,B1
	  Adump.unInit(); A1dump.unInit(); 
	  HfgDump.unInit(); CsubDump.unInit(); C1subDump.unInit(); C2subDump.unInit();
  } //for each shot
	
	termA2cube.dump("A2"); termB2cube.dump("B2");
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
	H12cube.dump("H12");
  return;
}

//use beam flatness obj function
//auxilary obj function consider only shift params. Window overlap
//aux function now based on each receiver location, not on each beam.
void FWI_FD_beamFocus14::calcBackProjData() {
  
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   float dt,dtau; dt = C_cube.domain().get_d(1); dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,iGroupMax;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> trcCube2,trcA;

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,F11;
     
   cubeData<float> EmatCube,FmatCube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes); //store H1 and H2
   
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   cubeDataF trcScaleCube(offsetShape);
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
     C_subCube.alloc(subCubeShape);
     C1_subCube.alloc(subCubeShape);
     C2_subCube.alloc(subCubeShape);
     t_subCube.alloc(subCubeShape);
	  
	   hypercube corrTrcShape(vector<axis> (1,(C_subCube.domain().get_axis(1))));
     trcCube2.alloc(corrTrcShape); trcA.alloc(corrTrcShape);
    
	  dmRsf_ocDump<float> CsubDump("Csub",""), C1subDump("C1sub",""), C2subDump("C2sub","");
	  CsubDump.init(&subCubeShape); C1subDump.init(&subCubeShape); C2subDump.init(&subCubeShape);
	  
	  //Calculate the number of beams we need, this time we make non overlap beams.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  iGroupMax = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
	  for(iGroup=0; iGroup<iGroupMax; iGroup++)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
      C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
      C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		  //calc trc A
		  //sum the traces along the subarray offset
      trcA.zero(); stackTrc(trcA, C_subCube);
      
      //calc E term
      E11 = dotProd<float>(C_subCube,C2_subCube) * dtau;
		 
		  EmatCube(0,iShot,iGroup+offsetBeam) = E11;
		  EmatCube(1,iShot,iGroup+offsetBeam) = 0.0f;
		  EmatCube(2,iShot,iGroup+offsetBeam) = 0.0f;
		 
      F11 = 1.0f/E11;
		  FmatCube(0,iShot,iGroup+offsetBeam) = F11;
		  FmatCube(1,iShot,iGroup+offsetBeam) = 0.0f;
      FmatCube(2,iShot,iGroup+offsetBeam) = 0.0f;
      spreadTrc(t_subCube,trcA);
      t_subCube.multiply(C1_subCube);
      trcScaleCube.zero();
      sumAlongTimeTrc(trcScaleCube, t_subCube); trcScaleCube.scale(F11);
      scaleTrc(t_subCube,C1_subCube,trcScaleCube);
		  //put C1_subcube back to the large cube
		  //Here we need to add, rather than replace, therefore first take the original data out		
		  C_backProjPerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		  C_subCube.add(t_subCube);
		  C_backProjPerShot.updateSeveralXd(&C_subCube,subGroupWindow,1);
	  } //end for iGroup

	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
    
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
    
	  //Mute all the beam stack result on the edge
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  
	  CsubDump.unInit(); C1subDump.unInit(); C2subDump.unInit();
  } //for each shot
	
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
  return;
}

//The global objective function changed to account both shift and slope
//Beam focus, double parameter fitting 2010. Use both shift and slope parameter
//Window do not overlap
void FWI_FD_beamFocus21::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,iGroupMax;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> trcCube2,trcA;

   cubeData<float> fhCube,ghCube; //For storing the moveOut function h^2 and h.
   cubeData<float> HfgCube; //Storing H1*f(h)+H2*g(h)

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,E12,E22, F11,F12,F22;
   float A2,B2;	float H1,H2;
     
   cubeData<float> termA2cube, termB2cube;
   cubeData<float> EmatCube,FmatCube, H12cube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termA2cube.alloc(termA2Shape);
   termB2cube.alloc(termA2Shape);
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes); //store H1 and H2
   H12cube.alloc(H12cubeShape);
   
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   fhCube.alloc(offsetShape);ghCube.alloc(offsetShape);  HfgCube.alloc(offsetShape);
   
   //Here change it to SHIFT and SLOPE
   velAna_FD::setMoveoutDeriv(fhCube,constant);
   velAna_FD::setMoveoutDeriv(ghCube,Linear);
   //fhCube.dump("fh"); ghCube.dump("gh");
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
     C_subCube.alloc(subCubeShape);
     C1_subCube.alloc(subCubeShape);
     C2_subCube.alloc(subCubeShape);
     t_subCube.alloc(subCubeShape);
	  
	   hypercube corrTrcShape(vector<axis> (1,(C_subCube.domain().get_axis(1))));
     trcCube2.alloc(corrTrcShape); trcA.alloc(corrTrcShape);
    
	  dmRsf_ocDump<float> Adump("trcA","A trace"); dmRsf_ocDump<float> A1dump("trcA1","A1 trace"); dmRsf_ocDump<float> B1dump("trcB1","B1 trace");
	  Adump.init(&corrTrcShape);  A1dump.init(&corrTrcShape);  B1dump.init(&corrTrcShape);
	  dmRsf_ocDump<float> HfgDump("Hfg","Hfg");
	  HfgDump.init(&offsetShape);
	  dmRsf_ocDump<float> CsubDump("Csub",""), C1subDump("C1sub",""), C2subDump("C2sub","");
	  CsubDump.init(&subCubeShape); C1subDump.init(&subCubeShape); C2subDump.init(&subCubeShape);
	  
	  //Calculate the number of beams we need, this time we make non overlap beams.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  iGroupMax = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
    //Calculate trace A first, because A trace is a global stack over all locations.
    trcA.zero(); stackTrc(trcA, C_cubePerShot);
	  for(iGroup=0; iGroup<iGroupMax; iGroup+=subGroupSize)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
      C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
      C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		  //calc A1, B1 trace and A2 B2 term.
		  //sum the traces along the subarray offset
      trcCube2.zero();C1_subCube.copyTo(t_subCube); scaleTrc(t_subCube,fhCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
      //sum along the tau axis
      A2 = dotProd<float>(trcA,trcCube2) * dtau;
		  //prt("term H=%f \n", termH);
		  //Dump A(tau) and A1(tau);
		  if(iGroup==0 || fabs(float(iGroup - iGroupMax/2)/subGroupSize)<1.0f) {
        CsubDump.dumpFrame(C_subCube); C1subDump.dumpFrame(C1_subCube); C2subDump.dumpFrame(C2_subCube);
        Adump.dumpFrame(trcA); A1dump.dumpFrame(trcCube2);
      }
		 
      //B1 trace
      trcCube2.zero();C1_subCube.copyTo(t_subCube);scaleTrc(t_subCube,ghCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
      if(iGroup==0 || fabs(float(iGroup - iGroupMax/2)/subGroupSize)<1.0f) {
			  B1dump.dumpFrame(trcCube2);
		  }
		 
		  //sum along the tau axis
		  B2 = dotProd<float>(trcA,trcCube2) * dtau;
		  
		  //!! Just work in two D case
		  termA2cube(iShot, iGroup+offsetBeam) = A2;
		  termB2cube(iShot, iGroup+offsetBeam) = B2; 
		  //term A2 is acutall DJ_local / Da, B2 is DJ_local / Db
		 
		  //Calculate E and F 2X2 matrix.
		  C2_subCube.copyTo(t_subCube);
		  scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,fhCube);
		  //Dot prod the previous and the C(tau;s0)
		  E11 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		  C2_subCube.copyTo(t_subCube);
		  scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,ghCube);
		  //Dot prod the previous and the C(tau;s0)
		  E12 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		  C2_subCube.copyTo(t_subCube);
		  scaleTrc(t_subCube,ghCube); scaleTrc(t_subCube,ghCube);
		  //Dot prod the previous and the C(tau;s0)
		  E22 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 
		  EmatCube(0,iShot,iGroup+offsetBeam) = E11;
		  EmatCube(1,iShot,iGroup+offsetBeam) = E12;
		  EmatCube(2,iShot,iGroup+offsetBeam) = E22;
		 
		  //solve E 2X2 matrix to get F
		  float tmpF[4]; float tmpE[4]; float determ, cond;
		  tmpE[0] = E11; tmpE[1]=tmpE[2]=E12; tmpE[3] = E22;
		  cond = invert2X2(tmpF,tmpE,determ);
		  if (cond > 1E3) {
			 verbPrt("\n cond number is very big %f in beamFocus2!! E=%.3e,%.3e,%.3e,%.3e", cond, E11, E12, E12, E22);
		  }
		  F11 = tmpF[0]; F12 = tmpF[1]; F22 = tmpF[3];

		  FmatCube(0,iShot,iGroup+offsetBeam) = F11;
		  FmatCube(1,iShot,iGroup+offsetBeam) = F12;
		  FmatCube(2,iShot,iGroup+offsetBeam) = F22;
		 
		  //then we calculate H1 H2 Term.
		  H1 = A2*F11 + B2*F12; 	H2 = A2*F12 + B2*F22;
		  //prt("term weight H/E =%f \n", weight);
		  H12cube(0,iShot,iGroup+offsetBeam) = H1;
		  H12cube(1,iShot,iGroup+offsetBeam) = H2;
		 		 
		  //calculate the scale cube H1*f(h)+H2*g(h)
		  HfgCube.scale(fhCube,H1); HfgCube.scaleAdd(ghCube,H2);
		  if(iGroup==0 || fabs(float(iGroup - iGroupMax/2)/subGroupSize)<1.0f) {
		 		HfgDump.dumpFrame(HfgCube);
		  }
		  scaleTrc(C1_subCube, HfgCube);
		  //put C1_subcube back to the large cube
		  //Here we need to add, rather than replace, therefore first take the original data out		
		  C_backProjPerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		  C_subCube.add(C1_subCube);
		  C_backProjPerShot.updateSeveralXd(&C_subCube,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  //d_injectCubePerShot(i,j) *= dampingCoef[subGroupSize-1-i];
				  d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 //d_injectCubePerShot(i,j) *= dampingCoef[i-(nx-subGroupSize)];
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  //save trc A,A1,B1
	  Adump.unInit(); A1dump.unInit(); B1dump.unInit();
	  HfgDump.unInit(); CsubDump.unInit(); C1subDump.unInit(); C2subDump.unInit();
  } //for each shot
	
	termA2cube.dump("A2"); termB2cube.dump("B2");
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
	H12cube.dump("H12");
  return;
}

//Beam focus, double parameter fitting 2010. Use both shift and slope parameter
//Here the window would overlap with each other
void FWI_FD_beamFocus22::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);
   dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,iGroupMax;
   
   cubeData<float>& d_injectCube = m_pFWIparams->backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> trcCube2,trcA;

   cubeData<float> fhCube,ghCube; //For storing the moveOut function h^2 and h.
   cubeData<float> HfgCube; //Storing H1*f(h)+H2*g(h)

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,E12,E22, F11,F12,F22;
   float A2,B2;	float H1,H2;
     
   cubeData<float> termA2cube, termB2cube;
   cubeData<float> EmatCube,FmatCube, H12cube;
   hypercube termA2Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termA2cube.alloc(termA2Shape);
   termB2cube.alloc(termA2Shape);
   axis ax_3layers(3); //store E_11,E_12,E_22;
   vecAxis axes = termA2Shape.get_axes(); axes.push_back(ax_3layers);
   hypercube EmatShape(axes);
   EmatCube.alloc(EmatShape); FmatCube.alloc(EmatShape);
   axes = termA2Shape.get_axes(); axes.push_back(axis(2));
   hypercube H12cubeShape(axes); //store H1 and H2
   H12cube.alloc(H12cubeShape);
   
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   fhCube.alloc(offsetShape);ghCube.alloc(offsetShape);  HfgCube.alloc(offsetShape);
   
   //Here change it to SHIFT and SLOPE
   velAna_FD::setMoveoutDeriv(fhCube,constant);
   velAna_FD::setMoveoutDeriv(ghCube,Linear);
   //fhCube.dump("fh"); ghCube.dump("gh");
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   cubeData<float> C_backProjPerShot;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   for(iShot=0;iShot< nShots; iShot++)
   {
     d_obsCubePerShot.alloc(shapeDataPerShot);
	   //aux cubes, remember that finally a correlation cube is injected
	   d_injectCubePerShot.alloc(shapeDataPerShot);
	   d_injectCubePerShot2.alloc(shapeDataPerShot);

	   C_cubePerShot.alloc(shapeCorrPerShot);
	   C_backProjPerShot.alloc(shapeCorrPerShot);
	   C1_cubePerShot.alloc(shapeCorrPerShot);
	   C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	   //read in d_obsCubePerShot
     shotWindow[dimPerShot].n = 1;
     shotWindow[dimPerShot].f = iShot;
      
	   d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
     C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
     C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
     C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
     C_subCube.alloc(subCubeShape);
     C1_subCube.alloc(subCubeShape);
     C2_subCube.alloc(subCubeShape);
     t_subCube.alloc(subCubeShape);
	  
	   hypercube corrTrcShape(vector<axis> (1,(C_subCube.domain().get_axis(1))));
     trcCube2.alloc(corrTrcShape); trcA.alloc(corrTrcShape);
    
	  dmRsf_ocDump<float> Adump("trcA","A trace"); dmRsf_ocDump<float> A1dump("trcA1","A1 trace"); dmRsf_ocDump<float> B1dump("trcB1","B1 trace");
	  Adump.init(&corrTrcShape);  A1dump.init(&corrTrcShape);  B1dump.init(&corrTrcShape);
	  dmRsf_ocDump<float> HfgDump("Hfg","Hfg");
	  HfgDump.init(&offsetShape);
	  dmRsf_ocDump<float> CsubDump("Csub",""), C1subDump("C1sub",""), C2subDump("C2sub","");
	  CsubDump.init(&subCubeShape); C1subDump.init(&subCubeShape); C2subDump.init(&subCubeShape);
	  
	  //Calculate the number of beams we need, this time we make non overlap beams.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  iGroupMax = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
    //Calculate trace A first, because A trace is a global stack over all locations.
    trcA.zero(); stackTrc(trcA, C_cubePerShot);
	  for(iGroup=0; iGroup<iGroupMax; iGroup++)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
      C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
      C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		  //calc A1, B1 trace and A2 B2 term.
		  //sum the traces along the subarray offset
      trcCube2.zero();C1_subCube.copyTo(t_subCube); scaleTrc(t_subCube,fhCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
      //sum along the tau axis
      A2 = dotProd<float>(trcA,trcCube2) * dtau;
		  //prt("term H=%f \n", termH);
		  //Dump A(tau) and A1(tau);
		  if(iGroup==0 || iGroup == iGroupMax/2) {
        CsubDump.dumpFrame(C_subCube); C1subDump.dumpFrame(C1_subCube); C2subDump.dumpFrame(C2_subCube);
        Adump.dumpFrame(trcA); A1dump.dumpFrame(trcCube2);
      }
		 
      //B1 trace
      trcCube2.zero();C1_subCube.copyTo(t_subCube);scaleTrc(t_subCube,ghCube);stackTrc(trcCube2, t_subCube); trcCube2.scale(dh);
      if(iGroup==0 || iGroup == iGroupMax/2) {
			  B1dump.dumpFrame(trcCube2);
		  }
		 
		  //sum along the tau axis
		  B2 = dotProd<float>(trcA,trcCube2) * dtau;
		  
		  //!! Just work in two D case
		  termA2cube(iShot, iGroup+offsetBeam) = A2;
		  termB2cube(iShot, iGroup+offsetBeam) = B2; 
		  //term A2 is acutall DJ_local / Da, B2 is DJ_local / Db
		 
		  //Calculate E and F 2X2 matrix.
		  C2_subCube.copyTo(t_subCube);
		  scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,fhCube);
		  //Dot prod the previous and the C(tau;s0)
		  E11 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		  C2_subCube.copyTo(t_subCube);
		  scaleTrc(t_subCube,fhCube); scaleTrc(t_subCube,ghCube);
		  //Dot prod the previous and the C(tau;s0)
		  E12 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		  C2_subCube.copyTo(t_subCube);
		  scaleTrc(t_subCube,ghCube); scaleTrc(t_subCube,ghCube);
		  //Dot prod the previous and the C(tau;s0)
		  E22 = dotProd<float>(t_subCube,C_subCube) * dtau * dh;
		 
		  EmatCube(0,iShot,iGroup+offsetBeam) = E11;
		  EmatCube(1,iShot,iGroup+offsetBeam) = E12;
		  EmatCube(2,iShot,iGroup+offsetBeam) = E22;
		 
		  //solve E 2X2 matrix to get F
		  float tmpF[4]; float tmpE[4]; float determ, cond;
		  tmpE[0] = E11; tmpE[1]=tmpE[2]=E12; tmpE[3] = E22;
		  cond = invert2X2(tmpF,tmpE,determ);
		  if (cond > 1E3) {
			 verbPrt("\n cond number is very big %f in beamFocus2!! E=%.3e,%.3e,%.3e,%.3e", cond, E11, E12, E12, E22);
		  }
		  F11 = tmpF[0]; F12 = tmpF[1]; F22 = tmpF[3];

		  FmatCube(0,iShot,iGroup+offsetBeam) = F11;
		  FmatCube(1,iShot,iGroup+offsetBeam) = F12;
		  FmatCube(2,iShot,iGroup+offsetBeam) = F22;
		 
		  //then we calculate H1 H2 Term.
		  H1 = A2*F11 + B2*F12; 	H2 = A2*F12 + B2*F22;
		  //prt("term weight H/E =%f \n", weight);
		  H12cube(0,iShot,iGroup+offsetBeam) = H1;
		  H12cube(1,iShot,iGroup+offsetBeam) = H2;
		 		 
		  //calculate the scale cube H1*f(h)+H2*g(h)
		  HfgCube.scale(fhCube,H1); HfgCube.scaleAdd(ghCube,H2);
		  if(iGroup==0 || fabs(float(iGroup - iGroupMax/2)/subGroupSize)<1.0f) {
		 		HfgDump.dumpFrame(HfgCube);
		  }
		  scaleTrc(C1_subCube, HfgCube);
		  //put C1_subcube back to the large cube
		  //Here we need to add, rather than replace, therefore first take the original data out		
		  C_backProjPerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
		  C_subCube.add(C1_subCube);
		  C_backProjPerShot.updateSeveralXd(&C_subCube,subGroupWindow,1);
	  } //end for iGroup
      
	  //We need to calculat INT{C_cubePerShot(tau)d_obs(t+tau)(dtau)} and then convolve with Green's function g(x,-t;xr,0)
	  //Notice that a small section of the correlation cube is truncated.
	  int nt = d_obsCube.domain().get_n(1);
	  int tStartLag = 0; int tEndLag = nt-1;
	  
	  //Assemble the back injected data by cross-correlate C1 with observed data cube dtau*C1(tau)*P(tau+t)
	  tStartLag = 0; tEndLag = nt - 1;
	  crossCorr1dGeneral(d_injectCubePerShot,C_backProjPerShot,d_obsCubePerShot,tStartLag,tEndLag);
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  //d_injectCubePerShot(i,j) *= dampingCoef[subGroupSize-1-i];
				  d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 //d_injectCubePerShot(i,j) *= dampingCoef[i-(nx-subGroupSize)];
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
	  
	  //save trc A,A1,B1
	  Adump.unInit(); A1dump.unInit(); B1dump.unInit();
	  HfgDump.unInit(); CsubDump.unInit(); C1subDump.unInit(); C2subDump.unInit();
  } //for each shot
	
	termA2cube.dump("A2"); termB2cube.dump("B2");
	EmatCube.dump("Emat"); FmatCube.dump("Fmat");
	H12cube.dump("H12");
  return;
}


