#include "stdafx.h"

#include "cubeData.hpp"

#include "velFD2d.h"
#include "waveOps.h"
#include "dampingFunc.h"
#include "FWI_FD_beamFocus.h"
#include "misc.h"
#include "interp.h"

using namespace Yang;


//measure the moveout parameters and test the obj func with the moveout params.
void FWI_FD_beamFocusY2::calcBackProjData() {
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  	 m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   //After getting the calculated data P_cal, calculate the cross-correlation with P_obs.
   //we've done the correlation in function evaluation term, so it could be omitted
   float dt,dtau;
   dt = C_cube.domain().get_d(1);  dtau = dt;
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);

   int iShot,nShots,iGroup,nGroup;
   
   cubeData<float>& d_injectCube = m_pFWIparams -> backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> C_subCubeShift, C1_subCubeShift, C2_subCubeShift;
   cubeData<float> d_obsSubCube, d1_obsSubCube, d_obsSubCubeShift, d1_obsSubCubeShift;
   cubeData<float> dInjectSubCube1, dInjectSubCube2;
   
   // 6 single trace quantity
   cubeData<float> A0cube,A1cube,A2cube;     cubeData<float> B1cube,B2cube,B3cube;
   hypercube trcShape = C_cube.domain().return_compatible(1,1);
   A0cube.alloc(trcShape); A1cube.alloc(trcShape); A2cube.alloc(trcShape);
   B1cube.alloc(trcShape); B2cube.alloc(trcShape); B3cube.alloc(trcShape);
   cubeData<float> trcCube,trcCube2, AIcube;
   trcCube.alloc(trcShape);    trcCube2.alloc(trcShape); AIcube.alloc(trcShape); 
   
   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E11,E12,E22, F11,F12,F22;
   float I1,I2;
     
   cubeData<float> termI1cube, termI2cube; 
   hypercube termI1Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termI1cube.alloc(termI1Shape);
   termI2cube.alloc(termI1Shape);
  	
   cubeData<float> fhCube,ghCube,ffCube,ggCube,fgCube, IfgCube; //For storing the moveOut function h^2 and h.
   cubeData<float> shiftCube, shift1Cube, shift2Cube; //within each beam window.
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2); n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   //Shape for subwindow cube of Correlation
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   //Shape for subwindow cube of observed/backproj data
   vector<axis> axes_dSubCube = axesSubCube;
   axes_dSubCube[0] = d_obsCube.domain().get_axis(1);
   hypercube d_subCubeShape(axes_dSubCube);
   
   fhCube.alloc(offsetShape); ghCube.alloc(offsetShape); ffCube.alloc(offsetShape); ggCube.alloc(offsetShape); fgCube.alloc(offsetShape); IfgCube.alloc(offsetShape);
   shiftCube.alloc(offsetShape);  shift1Cube.alloc(offsetShape);  shift2Cube.alloc(offsetShape);
   
   velAna::setMoveoutDeriv(fhCube,Quadratic);
   velAna::setMoveoutDeriv(ghCube,Linear);
   ffCube.multiply(fhCube,fhCube);    fgCube.multiply(fhCube,ghCube);    ggCube.multiply(ghCube,ghCube); 
   
   
   //dThetaOverDu.dump("dump3");
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot, d1_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   
	  
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   int nt = d_obsCube.domain().get_n(1);
   int tStartLag = 0; int tEndLag = nt-1;

   for(iShot=0;iShot< nShots; iShot++)
   {
      d_obsCubePerShot.alloc(shapeDataPerShot);
      d1_obsCubePerShot.alloc(shapeDataPerShot);	  
	  //aux cubes, remember that finally a correlation cube is injected
	  d_injectCubePerShot.alloc(shapeDataPerShot);
	  d_injectCubePerShot2.alloc(shapeDataPerShot);

	  C_cubePerShot.alloc(shapeCorrPerShot);
	  C1_cubePerShot.alloc(shapeCorrPerShot);
	  C2_cubePerShot.alloc(shapeCorrPerShot);
	  
      
	  //read in d_obsCubePerShot
      shotWindow[dimPerShot].n = 1;
      shotWindow[dimPerShot].f = iShot;
      
	  d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
	  timeDeriv1st(d1_obsCubePerShot,d_obsCubePerShot,true);
	  
      C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
      C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
      C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
	  //d_obsCubePerShot.dump("dump3");
	  //C_cubePerShot.dump("dump4");
	  
      C_subCube.alloc(subCubeShape);    C_subCubeShift.alloc(subCubeShape);
      C1_subCube.alloc(subCubeShape);   C1_subCubeShift.alloc(subCubeShape);
      C2_subCube.alloc(subCubeShape);   C2_subCubeShift.alloc(subCubeShape);
      t_subCube.alloc(subCubeShape);    
	  
	  d_obsSubCube.alloc(d_subCubeShape); d_obsSubCubeShift.alloc(d_subCubeShape);
	  d1_obsSubCube.alloc(d_subCubeShape); d1_obsSubCubeShift.alloc(d_subCubeShape);	 
	  dInjectSubCube1.alloc(d_subCubeShape); dInjectSubCube2.alloc(d_subCubeShape);
	  
	  //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  nGroup = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  float H2 = SQ(offsetBeam*dh); float H3 = H2 * (offsetBeam*dh);
	  float H5 = H2 * H3;
	  
	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {
     //subdivide the correlation panel C(tau;s0)
     //Read in C_subCube, C1_subCube, C2_subCube
     subGroupWindow[dimPerShot-1].n = subGroupSize;
     subGroupWindow[dimPerShot-1].f = iGroup;

     C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
     C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
     C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);
		 d_obsCubePerShot.grabSeveralXd(&d_obsSubCube,subGroupWindow,1);
		 d1_obsCubePerShot.grabSeveralXd(&d1_obsSubCube,subGroupWindow,1);		 
		 
		 //We need to shift C,C1,C2 cubes based on our picking parameters.
		 //read the picking params first
		 float a = m_pVelAnaParam->pickedCurvatureCube(iShot,iGroup + offsetBeam);
		 float b = m_pVelAnaParam->pickedSlopeCube(iShot,iGroup + offsetBeam);
		 shift1Cube.scale(fhCube,a);  shift2Cube.scale(ghCube,b);
		 shiftCube.zero(); shiftCube.add(shift1Cube, shift2Cube);
		 //Dump shiftCube should here.
		 
		 //Shift the Correlation Panel
		 staticShiftPerTrcArb(C_subCubeShift, C_subCube, shiftCube);
		 staticShiftPerTrcArb(C1_subCubeShift, C1_subCube, shiftCube);
		 staticShiftPerTrcArb(C2_subCubeShift, C2_subCube, shiftCube);
		 //Shift the observed data
		 staticShiftPerTrcArb(d_obsSubCubeShift, d_obsSubCube, shiftCube);
		 staticShiftPerTrcArb(d1_obsSubCubeShift, d1_obsSubCube, shiftCube);
		 		 
		  
		 //Calculate \tau trace variables A0,A1,A2,B0,B1,B2
		 //sum the traces along the subarray offset, A0
         stackTrc(A0cube,C_subCubeShift); A0cube.scale(dh);
		 
         // A1, A2
		 C1_subCubeShift.copyTo(t_subCube); scaleTrc(t_subCube,fhCube);
		 stackTrc(A1cube, t_subCube); A1cube.scale(dh);
		 C1_subCubeShift.copyTo(t_subCube); scaleTrc(t_subCube,ghCube);
		 stackTrc(A2cube, t_subCube); A2cube.scale(dh);
		 //B1,B2,B3
		 C2_subCubeShift.copyTo(t_subCube); scaleTrc(t_subCube,ffCube);
		 stackTrc(B1cube, t_subCube); B1cube.scale(dh);
		 C2_subCubeShift.copyTo(t_subCube); scaleTrc(t_subCube,ggCube);
		 stackTrc(B2cube, t_subCube); B2cube.scale(dh);
		 C2_subCubeShift.copyTo(t_subCube); scaleTrc(t_subCube,fgCube);
		 stackTrc(B3cube, t_subCube); B3cube.scale(dh);
		 
		  
		 //Calculate E and F 2X2 matrix, using A0,A1,A2 and B1,B2,B3.
		 E11 = dotProd<float>(A1cube,A1cube); E11 += dotProd<float>(A0cube,B1cube); E11 *= dtau;
		 E22 = dotProd<float>(A2cube,A2cube); E22 += dotProd<float>(A0cube,B2cube); E22 *= dtau;
		 E12 = dotProd<float>(A1cube,A2cube); E12 += dotProd<float>(A0cube,B3cube); E12 *= dtau;

		 //solve E 2X2 matrix to get F
		 float tmpF[4]; float tmpE[4]; float determ, cond;
		 tmpE[0] = E11; tmpE[1]=tmpE[2]=E12; tmpE[3] = E22;
		 cond = invert2X2(tmpF,tmpE,determ);
 
     // if (cond > 1E3) { verbPrt("\n cond number is very big %f
     // in beamFocus2!! E=%f,%f,%f,%f", cond, E11, E12, E12,E22); }
		 
		 //verbPrt("\n E11,E12,E22=%f, %f, %f, cond=%f", E11, E12, E22, cond);
		 F11 = tmpF[0]; F12 = tmpF[1]; F22 = tmpF[3];		 
		 
		 //Then calculate I1 and I2. 
		 I1 = 0.4 * a * H5 * F11 + (2.0f/3.0f)*b*H3*F12;
		 I2 = 0.4 * a * H5 * F12 + (2.0f/3.0f)*b*H3*F22;
		 
		 
		 //!! Just work in two D case
		 //termA2cube(iShot, iGroup+offsetBeam) = A2;
		 //termB2cube(iShot, iGroup+offsetBeam) = B2; 
		 
		 //calculate the scale cube I1*f(h)+I2*g(h)
		 IfgCube.scale(fhCube,I1); IfgCube.scaleAdd(ghCube,I2);
	
		 //calculate the crossCorrelation of Pobs and Ax traces.
		 tStartLag = 0; tEndLag = nt - 1;
		 crossCorr1dGeneral(dInjectSubCube1,A0cube,d1_obsSubCubeShift,tStartLag,tEndLag);
		 scaleTrc(dInjectSubCube1, IfgCube);
		 
		 trcCube.scale(A1cube,I1); trcCube2.scale(A2cube,I2);
		 AIcube.zero(); AIcube.add(trcCube,trcCube2);
		 crossCorr1dGeneral(dInjectSubCube2,AIcube,d_obsSubCubeShift,tStartLag,tEndLag);
		 dInjectSubCube1.add(dInjectSubCube2);		 	 
		 
		 //Grab the inject cube, add and then put it back
		 d_injectCubePerShot.grabSeveralXd(&dInjectSubCube2,subGroupWindow,1);
		 dInjectSubCube1.add(dInjectSubCube2);
		 d_injectCubePerShot.updateSeveralXd(&dInjectSubCube1,subGroupWindow,1);
	  } //end for iGroup
      
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  //d_injectCubePerShot(i,j) *= dampingCoef[subGroupSize-1-i];
					d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 //d_injectCubePerShot(i,j) *= dampingCoef[i-(nx-subGroupSize)];
					   d_injectCubePerShot(i,j) *= 0.0f;
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true);

      //Put the data back to a large cube
      d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot);
      //Sum the gradient over all shots.
   } //for each shot
	
   //termA2cube.dump("A2");
   //termB2cube.dump("B2");
   return;
}

//Single parameter, picking
void FWI_FD_beamFocusY1::calcBackProjData() {
	 //First calculate the dJ over moveout params
	verbPrt("\r Start calculate dJdb");
	 this->calc_dJdb();
	verbPrt("\r finish calc dJdb");
	 m_pVelAnaParam->dJdbCube.dump("DJDU");
	 
   //along with observed data, calculate the correlation panel
   cubeData<float>& C_cube =  	 m_pFWIparams->C_cube;
   cubeData<float>& d_obsCube =  m_pVelAnaParam->dObsCube;
   
   float dt,dtau;
   dt = C_cube.domain().get_d(1);  dtau = dt;
   
   //d_calCube.dump("dump1");
   //C_cube.dump("dump1");
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube,true);
   timeDeriv2nd(C2_cube,C_cube,true);
   //C_cube.dump("dump3"); C1_cube.dump("dump4");
   
   int iShot,nShots,iGroup,nGroup;
   
   cubeData<float>& d_injectCube = m_pFWIparams -> backProjCube;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   int subGroupSize = m_pVelAnaParam->beamSize;
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube, t_subCube;
   cubeData<float> C_subCubeShift, C1_subCubeShift, C2_subCubeShift;
   cubeData<float> d_obsSubCube, d1_obsSubCube, d_obsSubCubeShift, d1_obsSubCubeShift;
   cubeData<float> dInjectSubCube1, dInjectSubCube2;
   
   // 6 single trace quantity
   cubeData<float> A0cube,A1cube,A2cube;     cubeData<float> B1cube,B2cube,B3cube;
   hypercube trcShape = C_cube.domain().return_compatible(1,1);
   A0cube.alloc(trcShape); A1cube.alloc(trcShape); A2cube.alloc(trcShape);
   B1cube.alloc(trcShape); B2cube.alloc(trcShape); B3cube.alloc(trcShape);
   cubeData<float> trcCube,trcCube2, AIcube;
   trcCube.alloc(trcShape);    trcCube2.alloc(trcShape); AIcube.alloc(trcShape); 
   
   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   
   float E22,F22;
   float I2;
     
   cubeData<float> termI1cube, termI2cube; 
   hypercube termI1Shape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim());
   termI1cube.alloc(termI1Shape);
   termI2cube.alloc(termI1Shape);
  	
   cubeData<float> ghCube,ggCube,IfgCube; //For storing the moveOut function h^2 and h.
   cubeData<float> shiftCube, shift1Cube, shift2Cube; //within each beam window.
   //axis for local offset
   float o,dh; int n;
   dh = C_cube.domain().get_d(2); n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-dh);
   axis ax_offset(n,o,dh,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   //Shape for subwindow cube of Correlation
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   //Shape for subwindow cube of observed/backproj data
   vector<axis> axes_dSubCube = axesSubCube;
   axes_dSubCube[0] = d_obsCube.domain().get_axis(1);
   hypercube d_subCubeShape(axes_dSubCube);
   
   ghCube.alloc(offsetShape); ggCube.alloc(offsetShape); IfgCube.alloc(offsetShape);
   shiftCube.alloc(offsetShape);  shift1Cube.alloc(offsetShape);  shift2Cube.alloc(offsetShape);
   
   velAna::setMoveoutDeriv(ghCube,Linear);
   ggCube.multiply(ghCube,ghCube); 
   
   //dThetaOverDu.dump("dump3");
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot, d1_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
	  
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);

   waveOpParams& params = *(m_FDModelingParam);
   d_injectCube.alloc(d_obsCube.domain());
   nShots = params.m_nShots;
   
   int nt = d_obsCube.domain().get_n(1);
   int tStartLag = 0; int tEndLag = nt-1;

   for(iShot=0;iShot< nShots; iShot++)
   {
		 verbPrt("\rStart calc backProj Data ishot=%d",iShot);
		 
      d_obsCubePerShot.alloc(shapeDataPerShot);
      d1_obsCubePerShot.alloc(shapeDataPerShot);	  
	  //aux cubes, remember that finally a correlation cube is injected
	  d_injectCubePerShot.alloc(shapeDataPerShot);
	  d_injectCubePerShot2.alloc(shapeDataPerShot);

	  C_cubePerShot.alloc(shapeCorrPerShot);
	  C1_cubePerShot.alloc(shapeCorrPerShot);
	  C2_cubePerShot.alloc(shapeCorrPerShot);
	  
	  //read in d_obsCubePerShot
      shotWindow[dimPerShot].n = 1;
      shotWindow[dimPerShot].f = iShot;
      
	  d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
	  timeDeriv1st(d1_obsCubePerShot,d_obsCubePerShot,true);
	  //d1_obsCubePerShot.dump("dump6");
	  
      C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
      C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
      C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
      C_subCube.alloc(subCubeShape);    C_subCubeShift.alloc(subCubeShape);
      C1_subCube.alloc(subCubeShape);   C1_subCubeShift.alloc(subCubeShape);
      C2_subCube.alloc(subCubeShape);   C2_subCubeShift.alloc(subCubeShape);
      t_subCube.alloc(subCubeShape);    
	  
	  d_obsSubCube.alloc(d_subCubeShape); d_obsSubCubeShift.alloc(d_subCubeShape);
	  d1_obsSubCube.alloc(d_subCubeShape); d1_obsSubCubeShift.alloc(d_subCubeShape);	 
	  dInjectSubCube1.alloc(d_subCubeShape); dInjectSubCube2.alloc(d_subCubeShape);
	  
	  //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
	  VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	  nGroup = nRecv - 2 * int(subGroupSize/2);
	  
	  //for each subgroup, Keep in mind there is overlop for each point
	  int offsetBeam = subGroupSize / 2;
	  
	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {
      //subdivide the correlation panel C(tau;s0)
      //Read in C_subCube, C1_subCube, C2_subCube
      subGroupWindow[dimPerShot-1].n = subGroupSize;
      subGroupWindow[dimPerShot-1].f = iGroup;
      
      C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
      C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
      C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);
      d_obsCubePerShot.grabSeveralXd(&d_obsSubCube,subGroupWindow,1);
      d1_obsCubePerShot.grabSeveralXd(&d1_obsSubCube,subGroupWindow,1);		 
		 
		 //We need to shift C,C1,C2 cubes based on our picking parameters.
		 //read the picking params first
		 //float a = m_pVelAnaParam->pickedCurvatureCube(iShot,iGroup + offsetBeam);
		 float b = m_pVelAnaParam->pickedSlopeCube(iShot,iGroup + offsetBeam);
		 float dJdb = m_pVelAnaParam->dJdbCube(iShot,iGroup+offsetBeam);
		 
		 shift1Cube.zero(); shift2Cube.scale(ghCube,b);
		 shiftCube.zero(); shiftCube.add(shift1Cube, shift2Cube);

     //Dump shiftCube should here.
		 //shiftCube.dump("dump5");
		 
		 //Shift the Correlation Panel
		 staticShiftPerTrcArb(C_subCubeShift, C_subCube, shiftCube);
		 staticShiftPerTrcArb(C1_subCubeShift, C1_subCube, shiftCube);
		 staticShiftPerTrcArb(C2_subCubeShift, C2_subCube, shiftCube);
		 //Shift the observed data
		 staticShiftPerTrcArb(d_obsSubCubeShift, d_obsSubCube, shiftCube);
		 staticShiftPerTrcArb(d1_obsSubCubeShift, d1_obsSubCube, shiftCube);
		 		 
		 //C_subCube.dump("dump6");
		 //C_subCubeShift.dump("dump7"); C1_subCubeShift.dump("dump8");
		 
		 //Calculate \tau trace variables A0,A1,A2,B0,B1,B2
		 //sum the traces along the subarray offset, A0
     trcCube.zero(); stackTrc(A0cube,C_subCubeShift); A0cube.scale(dh);
		 
     // A1, A2
		 //C1_subCube.copyTo(t_subCube); scaleTrc(t_subCube,fhCube);
		 //stackTrc(A1cube, t_subCube); A1cube.scale(dh);
		 C1_subCubeShift.copyTo(t_subCube); scaleTrc(t_subCube,ghCube);
		 stackTrc(A2cube, t_subCube); A2cube.scale(dh);
		 //B1,B2,B3
		 //C2_subCube.copyTo(t_subCube); scaleTrc(t_subCube,ffCube);
		 //stackTrc(B1cube, t_subCube); B1cube.scale(dh);
		 C2_subCubeShift.copyTo(t_subCube); scaleTrc(t_subCube,ggCube);
		 stackTrc(B2cube, t_subCube); B2cube.scale(dh);
		 //C2_subCube.copyTo(t_subCube); scaleTrc(t_subCube,fgCube);
		 //stackTrc(B3cube, t_subCube); B3cube.scale(dh);
		 
		 //Calculate E and F 2X2 matrix, using A0,A1,A2 and B1,B2,B3.
		 E22 = dotProd<float>(A2cube,A2cube); E22 += dotProd<float>(A0cube,B2cube); E22 *= dtau;

		 //Just F22 = inv(E22);
		 F22 = 1.0f/E22;
		 
		 //Then calculate I1 and I2. 
		 //I2 = (2.0f/3.0f)*b*H3*F22;
		 I2 = dJdb * F22;
		 
		 
		 //!! Just work in two D case
		 //termA2cube(iShot, iGroup+offsetBeam) = A2;
		 //termB2cube(iShot, iGroup+offsetBeam) = B2; 
		 
		 //First part of inject data		 
		 //calculate the scale cube I1*f(h)+I2*g(h)
		 IfgCube.zero(); IfgCube.scaleAdd(ghCube,I2);
		 //calculate the crossCorrelation of Pobs and Ax traces.
		 tStartLag = 0; tEndLag = nt - 1;
		 //dInjectSubCube1.zero(); dInjectSubCube1.dump("dump6");
		 crossCorr1dGeneral(dInjectSubCube1,A0cube,d1_obsSubCubeShift,tStartLag,tEndLag);
		 //d1_obsSubCubeShift.dump("dump5"); A0cube.dump("dump7"); dInjectSubCube1.dump("dump6");
		 
		 scaleTrc(dInjectSubCube1, IfgCube);
		 //dInjectSubCube1.dump("dump7");
		 
		 if (!(m_pVelAnaParam->bIgnoreBfY2ndBpjTerm)) { 
			 //second part of inject data, Let's ignore this one.		 
			 AIcube.scale(A2cube,I2);
			 crossCorr1dGeneral(dInjectSubCube2,AIcube,d_obsSubCubeShift,tStartLag,tEndLag);
			 //A2cube.dump("dump5"); d_obsSubCubeShift.dump("dump6");
			 //dInjectSubCube2.dump("dump7");
			 dInjectSubCube1.add(dInjectSubCube2);		 	 
			 //dInjectSubCube1.dump("dump5");
		 }
		 
		 //Grab the inject cube, add and then put it back
		 d_injectCubePerShot.grabSeveralXd(&dInjectSubCube2,subGroupWindow,1);
		 dInjectSubCube1.add(dInjectSubCube2);
		 d_injectCubePerShot.updateSeveralXd(&dInjectSubCube1,subGroupWindow,1);
	  } //end for iGroup
      
	  //d_injectCubePerShot.dump("dump6");
	  //Mute all the beam stack result on the edge
		if(m_bTaperEdge) {
		   cosDampingFunc dampFunc;
		   stdVecFloat dampingCoef;
		   dampFunc.getCoefEx(dampingCoef,subGroupSize,1.0);
			  for (int i=0;i<subGroupSize;i++) {
				 for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
				  //d_injectCubePerShot(i,j) *= dampingCoef[subGroupSize-1-i];
					d_injectCubePerShot(i,j) *= 0.0f;
			  }
			  int nx = d_injectCubePerShot.domain().get_n(2);
			  for (int i=nx-1;i>=nx-subGroupSize;i--) {
					for(int j=0; j< d_injectCubePerShot.domain().get_n(1); j++)
					 //d_injectCubePerShot(i,j) *= dampingCoef[i-(nx-subGroupSize)];
					 d_injectCubePerShot(i,j) *= 0.0f;
			  }
		   }
	  //back project this data, (take 2nd derivative) because the born-modeling formula, might be better to take 2nd derivative of the fwd modeling wavelet.
	  timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot,true); 
    //Put the data back to a large cube
    d_injectCube.updateXd(&d_injectCubePerShot2,shotWindow,dimPerShot); 
   } //for each shot
	
   //termA2cube.dump("A2");
   //termB2cube.dump("B2");
   return;
}

//Pick moveout parameters, i.e the curvature cube, and the slope cube for each beam.
void FWI_FD_beamFocusY::pickMoveoutParmas() {
  
	cubeData<float>& aCube = m_pVelAnaParam->pickedCurvatureCube;
	cubeData<float>& bCube = m_pVelAnaParam->pickedSlopeCube;
	int subGroupSize = m_pVelAnaParam->beamSize;   
	int offsetBeam = subGroupSize / 2;
	waveOpParams& params = *(m_FDModelingParam);
  int nShots = params.m_nShots;

	/**Hack the theoretic moveout params in the simple uniform velocity case
   cubeData<float>& C_cube = m_pFWIparams->C_cube;
   //Set the size of pickParam Cube
   hypercube pickedParamShape = C_cube.domain().return_compatible(2, C_cube.domain().get_ndim()); //One param for each x_s,x_g
   
   aCube.alloc(pickedParamShape); bCube.alloc(pickedParamShape);
   
   int dimPerShot = 2;
   windParam shotWindow;   shotWindow.resize(3);
   cubeData<float> C_cubePerShot;
   //shape is a subcube the big data cube
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   
   float deltaS = (1.0f/m_vTrue) - (1.0f/m_v0);

   //Then for each shot calculate the theoretic moveout Param based on :
   //1. the receiver/src location;  2. the constant true velocity and initial velocity.
   
   //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
   VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);

   //for each subgroup, Keep in mind there is overlop for each point
   int ixShot,izShot; float xShot,zShot;
   float xRecv, zRecv;
   for(int iShot=0;iShot< nShots; iShot++)
   {
	  C_cubePerShot.alloc(shapeCorrPerShot);
	  ixShot = params.mShot1_ix;
	  izShot = params.mShot1_iz;	  
	  const hypercube& shapeVel = params.modelGridShape;
	  zShot = shapeVel.get_coord(izShot,1);	  xShot = shapeVel.get_coord(ixShot,2);
	  
	  zRecv = shapeVel.get_coord(params.mTrace1iz,1);
	  
	  //Get the shot location and the receiver location
	  int nRecv = C_cubePerShot.domain().get_n23();
	  int nGroup = nRecv - 2 * int(subGroupSize/2);

	  VERIFY(nGroup > 0);

	  shotWindow[dimPerShot].n = 1;
	  shotWindow[dimPerShot].f = iShot;
      C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
	  float a,b;
	  for(int iGroup=0; iGroup<nGroup; iGroup++)
	  {
		 //Determin Receiver Location
		 xRecv = C_cubePerShot.domain().get_coord(iGroup+offsetBeam,2);
		 //Use m_v0 and m_vTrue to find the a,b
		 float H = zRecv - zShot; float h = xRecv - xShot;
		 float Hh = sqrtf(H*H + h*h);
		 b = h/Hh;
		 a = SQ(H) / (CB(Hh));
		 aCube(iShot,iGroup+offsetBeam) = a * deltaS;
		 bCube(iShot,iGroup+offsetBeam) = b * deltaS;
	  }
    
   } //for each shot
   **/
   
	 //m_pVelAnaParam->ax_b.print_info("ax_b");
   scanMoveoutParmas(m_pVelAnaParam->ax_a, m_pVelAnaParam->ax_b);
	 //Pad both ends with same value on the edge
	 for(int iShot=0;iShot< nShots; iShot++)
   {
		 float a,b;
		 int nGroup = m_pFWIparams->C_cube.domain().get_n(2) - 2*offsetBeam;
    for(int iGroup=0; iGroup<offsetBeam; iGroup++) {
      a = aCube(iShot,offsetBeam);
      b = bCube(iShot,offsetBeam);
      aCube(iShot,iGroup) = a; bCube(iShot,iGroup) = b;
      a = aCube(iShot,offsetBeam+nGroup-1);
      b = bCube(iShot,offsetBeam+nGroup-1);
      aCube(iShot,iGroup+nGroup+offsetBeam) = a; 
      bCube(iShot,iGroup+nGroup+offsetBeam) = b;
    }
	 }
	 return;
}

void FWI_FD_beamFocusY::postCalcData() {
   FWI_FD_beamFocus::postCalcData();
	 
	 verbPrt("\n Picking moveout Params");
   pickMoveoutParmas();
   verbPrt("\r Finished Picking moveout Params");
	 
   //The moveout Params picked are very blocky, interp it so it becomes smooth
   //It is best to fit a curve based on this blocky curve, currently just smooth it
   cubeData<float> aCube, bCube;
   m_pVelAnaParam->pickedCurvatureCube.copyTo(aCube);
   m_pVelAnaParam->pickedSlopeCube.copyTo(bCube);

   //Simple smoothing is not quite good
   seplib::callSepTool(m_pVelAnaParam->pickedCurvatureCube,aCube,"Smooth","rect1=10");
   seplib::callSepTool(m_pVelAnaParam->pickedSlopeCube,bCube,"Smooth","rect1=10");

   //Try downsampling and interpolation, not work well
   /**
   cubeData<float> a2Cube, b2Cube; //First downSample by 10
   int downSampleRate = 5;
   
   seplib::interpType interpMethod = seplib::nearest;
   float d_a = m_pVelAnaParam->pickedCurvatureCube.domain().get_d(1);
   float d_b = m_pVelAnaParam->pickedSlopeCube.domain().get_d(1);
   float d_aout = d_a * downSampleRate; float d_bout = d_b * downSampleRate;
   seplib::interp::getInstance()->interpolate(a2Cube,aCube,NULL,NULL,&d_aout,
                                     NULL,NULL,NULL,
                                     NULL,NULL,NULL,interpMethod);
   seplib::interp::getInstance()->interpolate(b2Cube,bCube,NULL,NULL,&d_bout,
                                     NULL,NULL,NULL,
                                     NULL,NULL,NULL,interpMethod);
   //Then interploate using sinc function
   
   a2Cube.dump("dump1"); b2Cube.dump("dump2");
   interpMethod = seplib::sinc;
   int lengthSinc = 40;
   seplib::interp::getInstance()->interpolate(m_pVelAnaParam->pickedCurvatureCube,a2Cube,
                                     NULL,NULL,&d_a,
                                     NULL,NULL,NULL,
                                     NULL,NULL,NULL,interpMethod,0,lengthSinc);
   seplib::interp::getInstance()->interpolate(m_pVelAnaParam->pickedSlopeCube,b2Cube,
                                     NULL,NULL,&d_b,
                                     NULL,NULL,NULL,
                                     NULL,NULL,NULL,interpMethod,0,lengthSinc);
   **/
	 m_pVelAnaParam->pickedCurvatureCubeSubsamp.dump("aCube0");
	 m_pVelAnaParam->pickedSlopeCubeSubsamp.dump("bCube0");
   m_pVelAnaParam->pickedCurvatureCube.dump("aCube");
   m_pVelAnaParam->pickedSlopeCube.dump("bCube");
}

//calculate the derivative of objective function over moveout_params
//dJdb = (2.0f/3.0f)*H3*b
void FWI_FD_beamFocusY1::calc_dJdb() {
	cubeData<float>& bCube = m_pVelAnaParam->pickedSlopeCube;
	cubeData<float>& dJdbCube = m_pVelAnaParam->dJdbCube;
	int subGroupSize = m_pVelAnaParam->beamSize;   
	int offsetBeam = subGroupSize / 2;
	float dh = m_pFWIparams->C_cube.domain().get_d(2);
	float H = dh * offsetBeam;
	float H3 = CB(H);
	
	bCube.copyTo(dJdbCube);
	dJdbCube.scale((2.0f/3.0f)*H3);
	return;
}


//calculate the derivative of objective function over moveout_params
//dJdb 	//after we have picked the parameters, we calculate the dJ/d(moveout param) using the
//formula Deriv(j) = \sum_i{(m_i-\bar(m))*C_{ij}}, m means the shift(moveout) amount.
void FWI_FD_beamFocusY1sh::calc_dJdb() {
	cubeData<float>& bCube = m_pVelAnaParam->pickedSlopeCube;
	cubeData<float>& dJdbCube = m_pVelAnaParam->dJdbCube;
	bCube.allocMemoryTo(dJdbCube);
	int subGroupSize = m_pVelAnaParam->beamSize;   
	int offsetBeam = subGroupSize / 2;
	waveOpParams& params = *(m_FDModelingParam);
	int nShots = params.m_nShots;
	//float dh = m_pFWIparams->C_cube.domain().get_d(2);

	int ndim = bCube.domain().get_ndim();
	assert(ndim == 2);
	hypercube shapePerShot = bCube.domain().return_compatible(1,ndim-1);
	cubeData<float> shiftCubePershot(shapePerShot), bCubePershot(shapePerShot);
	float shift_avg;
	bool bAdd = false, bAdj = false;
	bCube.resetCounter(); dJdbCube.resetCounter();
	cubeDataF maskCube(shapePerShot);
	
	for(int iShot=0;iShot< nShots; iShot++)
	{
		int nGroup = bCube.domain().get_n(1) - 2*offsetBeam;
		int iGroup;
		maskCube.zero();
		for(iGroup=0; iGroup<nGroup; iGroup++) {
			maskCube(iGroup+offsetBeam) = 1.0f;
		}

		bCube.seq_grab(&bCubePershot); bCubePershot.multiply(maskCube);
		//Do the causal integ
		causalInt(shiftCubePershot, bCubePershot, bAdj, bAdd); 
		shiftCubePershot.multiply(maskCube);
		//bCubePershot.dump("dump11");
		//shiftCubePershot.dump("dump12");
		//calculate average
		shift_avg = shiftCubePershot.sum() / nGroup;
		
		cubeDataF Cji(hypercube(nGroup,nGroup));
		for(int j=0; j<nGroup; j++) {
			for (int i=0; i<j; i++) {
				Cji(j,i) = float(j-1)/nGroup-1;
			}
			for (int i=j; i<nGroup; i++) {
				Cji(j,i) = float(j-1)/nGroup;
			}
		}
		
		float dShift, fsum;
		//Now calculate the terms.
		for(int j=0+offsetBeam; j<nGroup+offsetBeam; j++) {
			fsum = 0.0f;
			for (int i=0+offsetBeam; i<nGroup+offsetBeam; i++) {
				dShift = shiftCubePershot(i) -shift_avg;
				fsum += dShift * Cji(j,i);
			}
			dJdbCube(iShot,j) = fsum;
		}
	}
	return;
}

float FWI_FD_beamFocusY::calcObjFunc() {
   //Use the model given in m_pFWI, calculate the objective function deltaTau^2 for all xs,xg
   this->calcData();

   cubeData<float>& C_cube =  m_pFWIparams->C_cube;
   float dt,dtau;
   dt = C_cube.domain().get_d(1); dtau = dt;
   
   //Pick moveout parameters, i.e the curvature cube, and the slope cube for each beam.
   this->postCalcData();
     
   //After that, calculate the square sum of the moveout amounts.
   //stack the subarray for each location, need to do it one by one beam
   int subGroupSize = m_pVelAnaParam->beamSize;
   //axis for local offset
   float o,d; int n;
   d = C_cube.domain().get_d(2);
   n = subGroupSize; 	
   //for each subgroup, Keep in mind there is overlop for each point
   int offsetBeam = subGroupSize / 2;
   float beamAperture = offsetBeam * d;

   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-d);
   axis ax_offset(n,o,d,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   int dimPerShot = 2;
   
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);

   waveOpParams& params = *(m_FDModelingParam);
   int nShots = params.m_nShots;
   
   cubeData<float> C_subCube,C_shiftSubCube, trcCube;
   float fObjFuncVal = 0.0f;
   
   //Calculate the number of beams we need, every x location there is a beam except the points on the edge.
   VERIFY( subGroupSize >= 1 && subGroupSize % 2 ==1);
	

   float a,b;
   float H2 = SQ(beamAperture); float H3 = H2 * beamAperture;
   float H5 = H2 * H3;
   for(int iShot=0;iShot< nShots; iShot++)
   {
	  //should read in the picked parameter cube for each shot.
      
	  int nGroup,iGroup, nRecv;
	  nRecv = shapeCorrPerShot.get_n23();
	  nGroup = nRecv - 2 * offsetBeam;
	  VERIFY(nGroup > 0);

	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {
         //Get picked parameter for this location, works only for 2d case
		 a = m_pVelAnaParam->pickedCurvatureCube(iShot,iGroup+offsetBeam);
		 b = m_pVelAnaParam->pickedSlopeCube(iShot,iGroup+offsetBeam);
		 
		 fObjFuncVal += 0.2f*SQ(a)*H5 + (1.0f/3.0f)*SQ(b)*H3;
	  }
   } //for each shot
   
   return fObjFuncVal;
}

