#include "stdafx.h"
#include "cubeData.hpp"

#include "FWI_FD.h"
#include "WEMVA.h"
#include "FWI_objFunc.h"

#include "myOMP.h"
#include <cmath>

using namespace Yang;
using namespace std;

//model is velocity, simply wrap around the function interface
float FWI_objFunc::evaluate(const gVecComponent& model) {
  FUNC_TRACE_BEG(FWI_objFunc::evaluate);

  //set the model to the FWI object
  //convert slowness to velocity
  cubeData<float> vel;
  model.allocMemoryTo(vel);
  Yang::reciprocal(vel,model);
  m_pFWI->setVelocity(&vel); 
  float ret = this->m_pFWI->calcObjFunc();
  FUNC_TRACE_END(ret);
}


void  FWI_objFunc::calcSearchDir( gVecComponent &searchDir) {
  FUNC_TRACE_BEG(FWI_objFunc::calcSearchDir);

  //LOG(calc bpj data);
  //Here we have already calculated the data, since we evaluated the obj function already.
  m_pFWI->calcBackProjData();
  
  //LOG(calc slow gradient)
  //calc slowness gradient
  m_pFWI->calcSlowGradient(searchDir);

  //LOG(fin slow grad);
  //creak mask along the survey region,Here SR_manager will calculate the velModelMask.
  VERIFY(m_pFWI->m_pVelAnaParam->velModelMask.sameSize(searchDir));
  searchDir.multiply(m_pFWI->m_pVelAnaParam->velModelMask);
  FUNC_TRACE_END();
}

cubeData<float>& FWI_objFunc::getCalcData() {
   return m_pFWI->m_pFWIparams->dCalCube;
}


/////////////////// WEMVA objective function ///////////////////////////////////////////
//model is velocity, simply wrap around the function interface
float WEMVA_objFunc::evaluate(const gVecComponent& model) {
  FUNC_TRACE_BEG(WEMVA_objFunc::evaluate)
   //set the model to the FWI object
   //convert slowness to velocity
   cubeData<float> vel;
   LOG(mm1);
   model.allocMemoryTo(vel);
   Yang::reciprocal(vel,model);
   LOG(mm2);
   m_pWEMVA->setVelocity(&vel);
   
   //LOG(calcualte initial image);
   //First calculate the data
   m_pWEMVA->calcData();
   LOG(mm3);
   m_pWEMVA->postCalcData(); //calc the angle gather at this step
   LOG(mm4);
   //LOG(calc obj func in wemva_evaluate);
   float ret = this->m_pWEMVA->calcObjFunc();
   if (std::isnan(ret)) {
     prt("\n the cost func is nan ! ");
   }
  FUNC_TRACE_END(ret);
}

void  WEMVA_objFunc::calcSearchDir( gVecComponent &searchDir) {
  FUNC_TRACE_BEG(WEMVA_objFunc::calcSearchDir);
  //Here we have already calculated the data, since we evaluated the obj function already.
  m_pWEMVA->calcBackProjData();
  
  //calculate gradient this time
  //It is actually SLOWNESS gradient
  m_pWEMVA->calcSlowGradient(searchDir);
  
  //m_pWEMVA->m_pWEMVAparams->Img0_hx.dump("dump6");
  //searchDir.dump("dump7");
  FUNC_TRACE_END();
}

const cubeData<float>& WEMVA_objFunc::getCalcData() {
  //return the final angle domain image gathers.
  return m_pWEMVA->getCalcData();
}

int  WEMVA_objFunc::setCalcData(const cubeData<float>& data) {
  return m_pWEMVA->setCalcData(data);
}

//For WEMVA objFunc to update status in a iterated solver
void WEMVA_objFunc::callBackAtIter( int iIter) {
  //e.g. Change the Smoothing parameter for MoveoutParam method.
  m_pWEMVA->callBackAtIter(iIter);
  return;
}
