#ifndef __MOVEOUTCORROBJFUNC_H_INCL__
#define __MOVEOUTCORROBJFUNC_H_INCL__

#include "gVec.h"
#include "objFunc.h"
#include "modelingFD2d_op.h"

/**
The moveout of correlation objective function defined in 
(Biondo1,2010) report. 
 
The local one Defined by 
  
S(x_bar)B(x_bar)C(tau;s) 
 
s is the slowness model, keep in mind, use slowness NOT velocity
*/

namespace Yang {

   class FWI_FD;
   class WEMVA_FD;

   class FWI_objFunc:public objFunc {
public:
   FWI_FD* m_pFWI;
public:
	// Constructor
	FWI_objFunc(FWI_FD* pOp): m_pFWI(pOp)
  {}

	// Destructor
	virtual ~ FWI_objFunc();
	
	//objFunc interface
	virtual float evaluate(const gVecComponent& model);
	virtual void  calcSearchDir( gVecComponent &searchDir);
	virtual cubeData<float>& getCalcData();   
  virtual int   setCalcData(const cubeData<float>& data) {VERIFY(0);}
  
private:
	// Copy constructor
	// TODO: Uncomment the copy constructor when you need it.
	FWI_objFunc(const FWI_objFunc& src);

	// Assignment operator
	// TODO: Uncomment the assignment operator when you need it.
	FWI_objFunc& operator=(const FWI_objFunc& src);
};

// Destructor implementation
inline FWI_objFunc::~ FWI_objFunc()
{
}

//WEMVA objective function, for the solver
   class WEMVA_objFunc:public objFunc {
public:
   WEMVA_FD* m_pWEMVA;
public:
	// Constructor
	WEMVA_objFunc(WEMVA_FD* pOp): m_pWEMVA(pOp)
  {}

	// Destructor
	virtual ~ WEMVA_objFunc() {}
	
	//objFunc interface
	virtual float evaluate(const gVecComponent& model);
	virtual void  calcSearchDir( gVecComponent &searchDir);
	virtual const cubeData<float>& getCalcData();
  virtual int   setCalcData(const cubeData<float>& data);

  virtual void callBackAtIter( int iIter);
private:
	// Copy constructor
	// TODO: Uncomment the copy constructor when you need it.
	WEMVA_objFunc(const WEMVA_objFunc& src);
	// Assignment operator
	// TODO: Uncomment the assignment operator when you need it.
	WEMVA_objFunc& operator=(const WEMVA_objFunc& src);
};
}

#endif // __MOVEOUTCORROBJFUNC_H_INCL

