#include "stdafx.h"
#include "decl_fwd.h"

#define NDEBUG
#include <assert.h>

#include "imaging.h"
#include "point2d.hpp"
#include "cubeAlgo.h"
#include "cubeAlgo2.h"

#include "cubeData.hpp"
#include "ADCIG_Recon.h"
#include "util_seplib.h"
#include "dipEstimator.h"
#include "dmRsfDump.h"
#include "misc.h"
#include "RayTomoOp.h"

#include "RMOfinder.h"

#define DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;
using namespace seplib;
using boost::shared_ptr;

void GenPtListsFromMask(const cubeDataF& mask, std::vector<point2dI>& vecPts) {
  const hypercube& shp = mask.domain();
  VERIFY(shp.get_ndim()==2);
  vecPts.clear();
  int n1=shp.get_n(1), n2=shp.get_n(2);
  for (int i2=0; i2<n2; i2++) {
    for (int i1=0; i1<n1; i1++) {
      if (mask(i2,i1)>0.5f) {
        vecPts.push_back(point2dI(i1,i2));
      }
    }
  }
  return;
}

RMO_Measurer::RMO_Measurer(const shared_ptr<dipEstimator>& est, const shared_ptr<const Param_RMO_Measurer>& par, const shared_ptr<const AGC_Param>& AGCpar): 
  par_(par), RMO_(new cubeDataF()),RMOsm_(new cubeDataF()),RMO_mask_(new cubeDataF()),ptWeights_(new cubeDataF()), candPts_(new cubeDataF()), 
  dipEstimator_(est), AGCpar_(AGCpar), RMOfinder_(new RMOfinder(par)) {
  }

int RMO_Measurer::inputADCIG_is(const cubeDataF& _img_ax, const cubeDataF& _img_0off) {
  img_ax_ = &_img_ax; img0off_ = &_img_0off;
  const cubeDataF &img_ax = *img_ax_;
  const cubeDataF &img_0off = *img0off_;  
  
  LOG("f1");
  //Here input an angle gather, the program will measure its and store the moveout information in the RMO_ information.
  const hypercube& shp = img_ax.domain();
  VERIFY(shp.get_ndim() == 3);
  
  axis_deg ax_gamma = axis_deg(shp.get_axis(2)); 
  axis ax_z = shp.get_axis(1);  axis ax_x = shp.get_axis(3);
  
  const hypercube& shp0 = img_0off.domain();
  VERIFY(ax_z.same_axis(shp0.get_axis(1))); VERIFY(ax_x.same_axis(shp0.get_axis(2)));
  
  LOG("f2");
  ///first identify the interesting points in the zero-offset image, based on two criterial:
  ///1. local maximum of the envelop function
  seplib::callSepTool(envelop_,img_0off,envbin,"tc1=3",true);
  //then Do some AGC to balance the amplitude.
  const AGC_Param& agc_par = *AGCpar_; //AGC parameter
  seplib::AGC(envelop_, envelopBal_, &agc_par);
  
  img0offBal_.alloc(shp0);  seplib::AGC(img_0off, img0offBal_, &agc_par);
  img_ax.dump("angImg.H",true);
  envelop_.dump("dump3");envelopBal_.dump("dump4"); img_0off.dump("dump5");img0offBal_.dump("dump6");
  
  envelopBal_.normalize(); img0offBal_.normalize();
  ///do 1-D non maximum suppression, this should be the estimated duration of a single event.
  NMSup_Param nmsup_par; nmsup_par.wnd1_ = int(this->par_->fEventWndLength/ax_z.d+0.5f); nmsup_par.wnd2_=1; //!need get param from outside later.
  envelopNMSup_.alloc(shp0);  nmSup(envelopBal_, envelopNMSup_, &nmsup_par);  envelopNMSup_.dump("dump7");
  
  cubeDataF mask; envelopNMSup_.allocMemoryTo(mask); 
  structureMask_.alloc(shp0); envMask_.alloc(shp0); finalMask_.alloc(shp0);
  //remove too weak reflectors on the energy envelope
  mask.zero(); buildThresholdMask(mask,envelopNMSup_,this->par_->fThrMinimumAmpRatio*envelopNMSup_.maxvalAbs());  envMask_ = mask;
  LOG("f5");
  ///2. The linearity measure is above a certian threshold, use cullison's thesis.
  //build structure tensors using balanced 0off image
  dipEstimator_->inputImgIs(img0offBal_); //then the linearity gets measured
  
  LOG("f6");
  const cubeDataF& linearity = dipEstimator_->linearityFactor();
  const cubeDataF& dipDetectMask   = dipEstimator_->dipEstMask();
  const cubeDataF& dipAngle  = dipEstimator_->dipAngleX_deg();
  //linearity.dump("dump8"); dipAngle.dump("dump10");
  
  dmRsf_ocDump<float> maskDump("dump14","RMO_Measure_mask"); maskDump.init(&linearity.domain());
  
  //remove largedip & dipEnergyMask & low linearity.
  mask.zero(); buildThresholdMask(mask,linearity, this->par_->fThrLinearity); structureMask_ = mask;
  mask.zero(); buildThresholdMask(mask,dipAngle,  this->par_->thrDipAngleX,true,false); 
  maskDump.dumpFrame(structureMask_); maskDump.dumpFrame(linearity); maskDump.dumpFrame(mask);
  
  structureMask_.multiply(mask); //here keep all dip angles below threshold not those above.
  maskDump.dumpFrame(envMask_); maskDump.dumpFrame(dipDetectMask); dipAngle.dump("dump11"); 

  structureMask_.multiply(dipDetectMask);
  finalMask_.multiply(envMask_,structureMask_);
  maskDump.dumpFrame(finalMask_);
  
  vector<point2dI> vecCandPts;  GenPtListsFromMask(finalMask_, vecCandPts);

  this->measureMoveout(vecCandPts);
  //this->genMoveoutForSensKernel(vecCandPts);
  return 0;
}

void RMO_Measurer::genMoveoutForSensKernel(const vector<point2dI>& vecCandPts) {
  const cubeDataF &img_ax = *img_ax_; //const cubeDataF &img_0off = *img0off_;
  const hypercube& shp = img_ax.domain();
  axis_deg ax_gamma = axis_deg(shp.get_axis(2)); 
  axis ax_z = shp.get_axis(1);  axis ax_x = shp.get_axis(3);
  int ng=ax_gamma.n, nz=ax_z.n;

  ///Second, select some points and make a parabolic moveout
  int nPts = vecCandPts.size();
  
  PRV(nPts);
  stdVecFloat vecTanMoveout(ng); float fDeg;
  for (int ig=0; ig<ng; ig++) {
    fDeg = deg2rad(ax_gamma.get_coord(ig)); vecTanMoveout[ig] = SQ(tanf(fDeg));
  }

  vecAxis axes; axes.resize(2); axes[0]=ax_gamma; axes[1]=axis(nPts);
  hypercube shpRMO(axes),shpCandPts(2,nPts);  RMO_->alloc(shpRMO); RMO_mask_->alloc(shpRMO);RMOsm_->alloc(shpRMO); candPts_->alloc(shpCandPts);
  ptWeights_->alloc(hypercube(nPts)); *ptWeights_ = 1.0f;

  candPtsSel_.resize(nPts); candPtsSel_ = 0;
  int ipp=0;
  int iz,ix;
  PRV(ipp);
  for (int ip=0; ip<nPts; ip++) {
    iz = vecCandPts[ip].sz; ix = vecCandPts[ip].sx;
    if (ix == ax_x.n/2) {
      //PRV2(iz,ix);
      //Stores the coordinates
      (*candPts_)(ip,0) = ax_z.get_coord(iz);  (*candPts_)(ip,1) = ax_x.get_coord(ix);
      candPtsSel_[ip] = 1; ipp++;
      for (int ig=0; ig<ng; ig++) {
        (*RMO_mask_)(ip,ig)  = 1.0f;
        (*RMOsm_)(ip,ig)     = vecTanMoveout[ig];
        (*RMO_)(ip,ig)       = (*RMOsm_)(ip,ig);
      }
    } else {
    }
  }
  return;
}

void RMO_Measurer::measureMoveout(const vector<point2dI>& vecCandPts) {
  const cubeDataF &img_ax = *img_ax_; //const cubeDataF &img_0off = *img0off_;
  const hypercube& shp = img_ax.domain();
  axis_deg ax_gamma = axis_deg(shp.get_axis(2)); 
  axis ax_z = shp.get_axis(1);  axis ax_x = shp.get_axis(3);
  int ng=ax_gamma.n, nz=ax_z.n;

  ///Second, for each point, locate the moveout information.
  int nPts = vecCandPts.size();

  int iz,ix,ii;
  int iHfEventWndSize     = int(this->par_->fEventWndLength*0.5f/ax_z.d);
  //int iHfEventMatchingWnd = int(this->par_->fEventMatchingWndLen*0.5f/ax_z.d);
  
  axis ax1(2*iHfEventWndSize+1,ax_z.o-iHfEventWndSize*ax_z.d,ax_z.d);
  hypercube shpf(ax1);
  cubeDataF matchFlter(shpf), matchFltCCout,mFltOutNMSup; //store the cross-correlation result
  //matchFlter.alloc();
  //const cubeDataF& referImg = img_0off;
  
  vecAxis axes(2); axes[0] = ax_z; axes[1] = ax_gamma; hypercube shpSlice(axes);cubeDataF img_ax_slice; img_ax_slice.becomeProxyCube(shpSlice); 
  windParam sliceWnd(3); sliceWnd[2].n=1;
  ///do cross-correlation of the reference trace with traces at all angles, find the maximum and the measure the relative shift at each angle.
  cubeDataF measuredMoveoutPerPt,smMeasuredMoveoutPerPt; 
  axes.resize(1); axes[0] = ax_gamma; measuredMoveoutPerPt.alloc(hypercube(axes));                  measuredMoveoutPerPt.allocMemoryTo(smMeasuredMoveoutPerPt);
  
  //Use gaussian kernel smoothing for RMO smoothing
  int hfSmKer = int(2.577f*0.5f*this->par_->dRMO_smWidth/ax_gamma.d+0.5f); axis ax_smKernel(2*hfSmKer+1,-hfSmKer*ax_gamma.d,ax_gamma.d); //the width of the gaussian is truncated at 1/100 maximum amplitude
  cubeDataF smKernel; getGaussfunc(smKernel,ax_smKernel,this->par_->dRMO_smWidth, 0.f, 0); 
  //normalize the RMO_smoothFunc so that the flt coefs sum to 1.
  float rs = 1.0f/smKernel.sum(); smKernel.scale(rs);
  int istart,iend;
  
  dmRsf_ocDump<float> matchFlterDump("dump16","ss");  matchFlterDump.init(&matchFlter.domain());
  axes.resize(2); axes[0]=ax_gamma; axes[1]=axis(nPts);
  hypercube shpRMO(axes),shpCandPts(2,nPts);  RMO_->alloc(shpRMO); RMO_mask_->alloc(shpRMO);RMOsm_->alloc(shpRMO); candPts_->alloc(shpCandPts); 
  candPtsSel_.resize(nPts); candPtsSel_ = 0; //default all candi pts are not selected.
  ptWeights_->alloc(hypercube(nPts)); *ptWeights_ = 1.0f;

  std::valarray<int> angle_axis_mask(ng); std::valarray<float> angle_axis_RMO(ng);
  int ipp = 0; //some points are skipped. so ipp would be smaller than ip.
  //build moveout function to get an approximate moveout amount
  stdVecFloat vecTanMoveout(ng); float fDeg;
  for (int ig=0; ig<ng; ig++) {
    fDeg = deg2rad(ax_gamma.get_coord(ig)); vecTanMoveout[ig] = SQ(tanf(fDeg));
  }
  float vErrRatio = this->par_->fApproxVelErrRatio_;  float fWaterDepth = this->par_->waterDepthAver_;

  //we can use it make a mask for ADCIG moveout.
  bool bUseRayTomo = (rayTomoOp_.get()!=NULL);
  const cubeDataF& dipAngle = this->dipEstimator_->dipAngleX_deg();
  float fMaxMoveout = vecTanMoveout[ng-1]*vErrRatio*(ax_z.get_coord(ax_z.n-1)-fWaterDepth);
  prt("maximum amount of moveout: %f\n", fMaxMoveout);
  
  SEEV(vecCandPts.size());
  vector<fwk::degree> vAng(ng),rayTraceAng; vector<float> maskADCIGray;
  //rayTraceAng needs to cover all angles in vAng symmetrically.
  float amin,amax,aminn,amaxn; ax_gamma.getRange(&amin,&amax);

  //VERIFY(fabs(amin+amax) <= 0.1f); //otherwise need to modify the rayTraceAng.

  aminn = std::min(-fabs(amin),-fabs(amax));aminn = std::min(aminn,0.0f);
  amaxn = std::max(fabs(amin),fabs(amax));  amaxn = std::max(amaxn,0.0f);
  rayTraceAng.resize(lroundf((amaxn-aminn)/ax_gamma.d)+1); 
  for (int ig=0; ig<rayTraceAng.size(); ig++) rayTraceAng[ig] = fwk::degree(aminn+ig*ax_gamma.d);
  for (int ig=0; ig<ng; ig++) vAng[ig] = fwk::degree(ax_gamma.get_coord(ig));
  
  //set ADCIG to RMOfinder, let it figure out some pre-processing.
  RMOfinder_->inputGatherIs(img_ax);
  PRV(iHfEventWndSize);
  int ret;

  vecAxis axes3(2); axes3[0]=ax_z; axes3[1]=ax_x;
  hypercube shp3(axes3);
  cubeDataF wtsCube(shp3);

  for (int ip=0; ip<nPts; ip+=1) { //1094
    iz = vecCandPts[ip].sz; ix = vecCandPts[ip].sx;
    //For test 
    //if (!(ix%50>=7 && ix%50<=7)) continue;
    //PRV2(iz,ix);
    //Stores the coordinates
    (*candPts_)(ip,0) = ax_z.get_coord(iz);  (*candPts_)(ip,1) = ax_x.get_coord(ix);
    
    //if (ip>15) exit(0);
    //Use Ray Tomography to generate mask
    maskADCIGray.assign(ng,1.0f);
    fwk::degree dip;
    if (bUseRayTomo) {
      dip = fwk::degree(dipAngle(ix,iz));
      rayTomoOp_->traceOrginPtInfoIs(rayTraceAng,&(*candPts_)(ip,0),true,&dip);
      rayTomoOp_->acqCoverageMaskForCurSubsurfPt(maskADCIGray,vAng);
      PRVEC(&(*candPts_)(ip,0),2); PRVEC(maskADCIGray,ng);
    }
    
    //crop out the matching filter (windowed segment from the img0off)
    if (iz-iHfEventWndSize<0 || iz+iHfEventWndSize>=nz) {
      prt("The CandPt: (%f,%f) is skipped because it is too close to boundary!\n", ax_z.get_coord(iz),ax_x.get_coord(ix));
      continue; //skip the boundary case.
    }
    
    //LOG("i1");
    sliceWnd[2].f=ix;
    const_cast<cubeDataF&>(img_ax).setProxyXd(&img_ax_slice,sliceWnd,2);
    //LOG("i2");
    //measure the moveout at current cand pt.
    int iCoords[2] = {iz,ix}; 
    angle_axis_mask = 0; angle_axis_RMO = 0.0f;
    float rhoPicked;
    ret = RMOfinder_->RMOatPt(iCoords, img0offBal_, img_ax, img_ax_slice, vecTanMoveout,angle_axis_mask,angle_axis_RMO, (*ptWeights_)(ip),rhoPicked);
    if (ret < 0) continue; //skip this point and move to next.
    
    float ft = (*ptWeights_)(ip);
    for (ii=-2;ii<=2; ii++) {
    for (int jj=-2;jj<=2; jj++) {
      wtsCube(ix+ii,iz+jj) = ft * rhoPicked;
    }}
    
    //LOG("i3");
    
    //Now use ADCIGmask to maskout the unreliable measurements
    int igbp = ax_gamma.get_offset(50.0f);
    for (int ig=0; ig<ng; ig++) {
      if (maskADCIGray[ig]<0.5f) {angle_axis_mask[ig] = 0;}
      if (ig<igbp) { //mask out the high angles, cause there are not correct by semblance RMO.
        (*RMO_mask_)(ip,ig) = angle_axis_mask[ig];
        (*RMO_)(ip,ig) = angle_axis_RMO[ig];
      } else {
        angle_axis_mask[ig] = 0;
      }
    }
    
    //LOG("i4");
    //Now all the moveout has been measured. but there might be holes, should fill the holes and apply some smoothing, just use kernel smoothing.
    int ig0 = ax_gamma.get_offset(0.0f), igmir; float ftmp;
    (*RMO_)(ip,ig0)       = angle_axis_RMO[ig0]; //handle zero angle.
    
    for (int ig=0;ig<ig0; ig++) { //use symmetry to fill some holes first
      igmir = (2*ig0-ig);
      if (igmir < ng) {
        if (angle_axis_mask[ig]==0 && angle_axis_mask[igmir]==0) { // do nothing
        } else {
          if (angle_axis_mask[ig]==0 && angle_axis_mask[igmir]>0) {
            (*RMO_)(ip,ig)    = angle_axis_RMO[igmir];
            (*RMO_)(ip,igmir) = angle_axis_RMO[igmir];
          } else if (angle_axis_mask[ig]>0 && angle_axis_mask[igmir]==0) {
            (*RMO_)(ip,ig)    = angle_axis_RMO[ig];
            (*RMO_)(ip,igmir) = angle_axis_RMO[ig];
          } else {
            ftmp = (angle_axis_RMO[igmir]+angle_axis_RMO[ig]) * 0.5f; //average the moveout, because it should be symmetric
            (*RMO_)(ip,igmir) = ftmp; (*RMO_)(ip,ig) = ftmp;
          }
          angle_axis_mask[ig] = 1; angle_axis_mask[igmir] = 1;
        }
      }
    } //end for each ig
    
    //LOG("i5");
    //do the Gaussian kernel smoothing.
    float fDenom,fNum;
    for (int ig=0;ig<ng; ig++) {
      istart = std::max(0,ig - hfSmKer); iend = std::min(ig+hfSmKer,ng-1);
      //find available values in the range.
      fDenom=0.f; fNum=0.f;
      for (ii=istart; ii<=iend; ii++) {
        if (angle_axis_mask[ii]>0) {
          fDenom += (*RMO_)(ip,ii)*smKernel(ii-ig+hfSmKer);
          fNum   += smKernel(ii-ig+hfSmKer);
        }
      }
      if (fNum > 1e-3f) {//there is at least one hit
        (*RMOsm_)(ip,ig) = (fDenom/fNum); (*RMO_mask_)(ip,ig) = 1.0f;
      }
    }
    
    //again use ADCIGmask to maskout the unreliable measurements
    for (int ig=0; ig<ng; ig++) {
      if (maskADCIGray[ig]<0.5f) {
        (*RMOsm_)(ip,ig) = 0.0f;  (*RMO_mask_)(ip,ig) = 0.0f;
      }
    }
    //PRVEC(angle_axis_mask,ng); PRVEC(&(*RMO_)(ip,0),ng);PRVEC(&(*RMOsm_)(ip,0),ng); LOG("g7");
    //PRV(ax_x.get_coord(ix)); PRV(ip); LOG("g8");
    candPtsSel_[ip] = 1; ipp++;
  } //for each candidate pts.
  //LOG("f8");
  nRMO_pts_ = NnegInt(ipp);
  SEEV(nRMO_pts_); RMO_mask_->dump("dump20"); RMO_->dump("dump21"); RMOsm_->dump("dump22"); candPts_->dump("candPts.H",true);
  ///may fit parameters for the measured moveout.
  this->RMOplotted()->dump("rmo.H",true);
  wtsCube.dump("dJdrho.H",true);
  return;
}

shared_ptr<cubeDataF> RMO_Measurer::RMOplotted() const { //return a cube that draws the RMO measured.
  assert(img_ax_);
  shared_ptr<cubeDataF> plotRMO(new cubeDataF());
  img_ax_->allocMemoryTo(*plotRMO);
  const vecAxis& axes = img_ax_->domain().get_axes();
  const axis& ax_z = axes[0]; const axis& ax_gamma = axes[1];
  const axis& ax_x = axes[2]; float dz = ax_z.d;
  int npts = candPts_->domain().get_n23();
  int iz,ix,iiz; float fiz; 
  const std::valarray<int>& candSel = candPtsSel();
  float amp = 1.0f;
  stdVecFloat axz_coords(ax_z.n),datCoords(1);
  for (iz=0; iz<ax_z.n; iz++) axz_coords[iz] = float(iz);
  
  //LOG("R2");
  for (int ip=0; ip<npts; ip++) {
    amp = (*ptWeights_)(ip);
    iz = ax_z.get_offset((*candPts_)(ip,0)); ix = ax_x.get_offset((*candPts_)(ip,1));
    if (candSel[ip]>0) {
      for (int ig=0; ig<ax_gamma.n; ig++) {
        if ((*RMO_mask_)(ip,ig)>0.5f) {
          fiz = iz+(*RMOsm_)(ip,ig)/dz; iiz = lroundf(fiz);
          iiz = std::max(std::min(iiz,ax_z.n-1),0); // should (iiz >=0 && iiz<ax_z.n);
          if (fabs(fiz-iiz)<0.1f) {
            (*plotRMO)(ix,ig,iiz) = amp;
          } else {
            datCoords[0] = fiz;
            //LOG("S3");
            linInterp1dOper(&amp, datCoords, &(*plotRMO)(ix,ig,0), axz_coords, true,true,true);
            //LOG("S4");
          }
        }
      }
    }
  }
  return plotRMO;
}

class funcThresHoldingKeepLarge{
  funcThresHoldingKeepLarge(float fThres): fThres_(fThres) {}
  void operator() (float& i) {
    i = i>fThres_? i:0.0f;
  }
private:
  float fThres_;
};
class funcThresHoldingKeepSmall{
  funcThresHoldingKeepSmall(float fThres): fThres_(fThres) {}
  void operator() (float& i) {
    i = i<fThres_? i:0.0f;
  }
private:
  float fThres_;
};

