#include "stdafx.h"
#include "decl_fwd.h"

#undef NDEBUG
#include <assert.h>

#include "misc.h"
#include "util_seplib.h"
#include "dipEstimator.h"
#include "cubeData.hpp"
#include "cubeAlgo.h"
#include "cubeAlgo2.h"
#include "ADCIG_Recon.h"
#include "RMOfinder.h"

#include "commonOps.h"


#define DEBUG
#include "myDebug.h"

#define MIN_LOC_SEARCH_WND 3

using namespace Yang;
using namespace std;
using namespace seplib;
using boost::shared_ptr;

//Do some pre-computing
void RMOfinder::inputGatherIs(const cubeDataF& img_ax_t) {
  NMSup_Param nmsup_par;
  const axis& ax_z = img_ax_t.domain().get_axis(1);
  nmsup_par.wnd1_ = int(this->par_->fEventWndLength/ax_z.d+0.5f); nmsup_par.wnd2_=1; //!need get param from outside later.
  if (par_->method_ == Param_RMO_Measurer::CrossCorr) {
    //do nothing
  } else if (par_->method_ == Param_RMO_Measurer::EnvMax) { //compute the envelope
    img_ax_t.allocMemoryTo(envelopeZ_ADCIG_);
    seplib::callSepTool(envelopeZ_ADCIG_,img_ax_t,envbin,"tc1=3");
    //Then need to do non-minimum supression.
    envelopeZ_ADCIG_.allocMemoryTo(NMSupEnvelope_);
    nmSup(envelopeZ_ADCIG_,NMSupEnvelope_, &nmsup_par);
  } else if (par_->method_ == Param_RMO_Measurer::NbCorr) {
    //do nothing here
  } else if (par_->method_ == Param_RMO_Measurer::dipTracking) { //compute the dipTracking technique
    //should build dip field, and track the moveout using the dip field.
  } else if (par_->method_ == Param_RMO_Measurer::Semblance) {
    //compute the Z-envelope
    img_ax_t.allocMemoryTo(envelopeZ_ADCIG_);
    seplib::callSepTool(envelopeZ_ADCIG_,img_ax_t,envbin,"tc1=3");
    //Then need to do non-minimum supression.
    envelopeZ_ADCIG_.allocMemoryTo(NMSupEnvelope_);
    //nmSup(envelopeZ_ADCIG_,NMSupEnvelope_, &nmsup_par);

    vecAxis axes2 = img_ax_t.domain().get_axes();
    axis ax_deg = axes2[1];
    cubeDataF* img_ax_rad = NULL;
    cubeDataF& semb = this->semb_;
    axis ax_rad;  deg2radianAxis(ax_rad,ax_deg,true);
    axes2[1] = ax_rad; //convert axis to radian temporarily.
    img_ax_rad = const_cast<cubeDataF*>(&img_ax_t);

    //ax_rad.print_info("ax_rad"); ax_deg.print_info("ax_deg");
    img_ax_rad->reshape(axes2);
    //First Calculate the semblance panel
    axes2[1] = par_->ax_semb_; semb.alloc(axes2);
    float fAntiAlias=1.0f;
    calcMoveoutSemblance_omp(semb, *img_ax_rad, Yang::tangentSquare,fAntiAlias, true);
    semb.multiply(semb); semb.allocMemoryTo(envSemb_);
    rectSmooth(semb,envSemb_,4,1,1); //smooth with rect1=4
    envSemb_.dump("semb.H",true);
    //after computation, restore the deg axis
    axes2[1] = ax_deg; img_ax_rad->reshape(axes2);
    
    //Check the shape of Guassian Deriv    
    Yang::getGaussfunc(gaussDeriv_,par_->ax_semb_,par_->sembGaussDerivWidth_,0.0f,1);
    gaussDeriv_.dump("gs1.H",true);
  } else {
    {VERIFY(0);}
  }
}

void RMOfinder::locateMoveoutUsingNbCorrelation(int igamma, int ilag, valarray<int>& angle_axis_mask, valarray<float>& RMO_per_angle, const stdVecFloat& estMoveoutPerAngle, const cubeDataF& img_axt_slice) //does not work very well
{
  axis_deg ax_gamma = axis_deg(img_axt_slice.domain().get_axis(2));
  const axis& ax_z = img_axt_slice.domain().get_axis(1);
  int ng=ax_gamma.n, nz=ax_z.n;

  float fRMOMax = *std::max_element(estMoveoutPerAngle.begin(),estMoveoutPerAngle.end());
  int iMaxHfCorrSize = std::max(MIN_LOC_SEARCH_WND,int(1.0f/ax_z.d*fabs(fRMOMax)+0.5f));
  stdVecFloat corrs(2*iMaxHfCorrSize+1),energys(2*iMaxHfCorrSize+1); 

  float thrNbTraceCorrCoef  = par_->thrNbTraceCorrCoef_;  //=0.3f; 
  float thrEventEnergyRatio = par_->thrEventEnergyRatio_; //=0.1f;
  int iHfEventWndSize     = int(par_->fEventWndLength*0.5f/ax_z.d);
  int igg, iggLast = igamma; int iSearchLag, ilag1 = ilag, ii,imax;

  float tmp0,fFltEmax,fFltE;
  for (igg=igamma; igg>=0; igg--) {
    if (angle_axis_mask[igg]==0) {  //check if a match can be found.
      if (abs(iggLast-igg)>5) break;//cannot find consecutviely for 5 traces, should give up.
      iSearchLag = std::max(MIN_LOC_SEARCH_WND,int(0.5f+1.0f/ax_z.d*fabs(estMoveoutPerAngle[iggLast]-estMoveoutPerAngle[igg])));
      //PRV2(igg,iSearchLag);
      if ((iz_+ilag1+iSearchLag+iHfEventWndSize)>=nz || iz_+ilag1-iSearchLag-iHfEventWndSize<0) break; //will cause memory access out of range, therefore skip this.
      //Use the matchFlter provided to search for neighboring coordinates.
      for (int ioff=-iSearchLag; ioff<=iSearchLag; ioff++) { //only search the neihboring 11pts.
        corrs[ioff+iSearchLag] = 0.0f; energys[ioff+iSearchLag] = 0.0f;
        for (ii=-iHfEventWndSize; ii<=iHfEventWndSize; ii++) {
          //PRV(iz+ilag1+ioff+ii); prt("%d,%d,%d,%d.",iz,ilag1,ioff,ii);
          tmp0 = img_axt_slice(igg,iz_+ilag1+ioff+ii);
          corrs[ioff+iSearchLag] += matchFlter_(ii+iHfEventWndSize)*tmp0;
          energys[ioff+iSearchLag] += tmp0*tmp0;
        }
      }
      imax = (std::max_element(&corrs[0],&corrs[2*iSearchLag+1])) - &corrs[0];
      //check the value of corr, make sure it is good enough to be selected
      PRV(ax_gamma.get_coord(igg));PRV(corrs[imax]);PRV(imax); PRVEC(corrs,2*iSearchLag+1);
      if (imax!=0 && imax!=2*iSearchLag) {//the max should not be on either ends
        if (energys[imax]>fFltEmax*thrEventEnergyRatio && corrs[imax]>sqrtf(fFltE*energys[imax])*thrNbTraceCorrCoef) {//the energy should not be too small and the correlation coef should be higher.
          ilag1 += (imax-iSearchLag); //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = ilag1*ax_z.d;
        }
      }
    }
    //Now a match actually been found, load the matching filter on ig for the next round, otherwise the matching filter will keep the same.
    if (angle_axis_mask[igg] > 0) {
      for (ii=-iHfEventWndSize; ii<=iHfEventWndSize; ii++) {
        matchFlter_(ii+iHfEventWndSize) = img_axt_slice(igg,iz_+ilag1+ii);
      }
      iggLast = igg;  fFltE = dotProd(matchFlter_,matchFlter_);
      fFltEmax = std::max(fFltE,fFltEmax);
      PRVEC(matchFlter_,2*iHfEventWndSize+1); PRV(matchFlter_[iHfEventWndSize]);PRV(fFltE); PRV(ilag1);
    }
  } //end for igg
  //LOG("g5");
  ilag1 = ilag; //search the other direction on gamma axis.
  for (igg=igamma; igg<ng; igg++) {
    if (angle_axis_mask[igg] == 0) {//check if a match can be found.
      if (abs(iggLast-igg)>5) break; //cannot find consecutviely for 5 traces, should give up.
      iSearchLag = std::max(MIN_LOC_SEARCH_WND,int(0.5f+1.0f/ax_z.d*fabs(estMoveoutPerAngle[iggLast]-estMoveoutPerAngle[igg])));
      if ((iz_+ilag1+iSearchLag+iHfEventWndSize)>=nz || iz_+ilag1-iSearchLag-iHfEventWndSize<0) break; //will cause memory access out of range, therefore skip this.
      //Use the matchFlter provided to search for neighboring coordinates.
      for (int ioff=-iSearchLag; ioff<=iSearchLag; ioff++) { //only search the neihboring 11pts.
        corrs[ioff+iSearchLag] = 0.0f; energys[ioff+iSearchLag] = 0.0f;
        for (ii=-iHfEventWndSize; ii<=iHfEventWndSize; ii++) {
          tmp0 = img_axt_slice(igg,iz_+ilag1+ioff+ii);
          corrs[ioff+iSearchLag] += matchFlter_(ii+iHfEventWndSize)*tmp0;
          energys[ioff+iSearchLag] += tmp0*tmp0;
        }
      }
      imax = (std::max_element(&corrs[0],&corrs[2*iSearchLag])) - &corrs[0];
      PRV(ax_gamma.get_coord(igg));PRV(corrs[imax]);PRV(imax); PRVEC(corrs,2*iSearchLag+1);
      //check the value of corr, make sure it is good enough to be selected
      if (imax!=0 && imax!=2*iSearchLag) {//the max should not be on either ends
        if (energys[imax]>fFltEmax*thrEventEnergyRatio && corrs[imax]>sqrtf(fFltE*energys[imax])*thrNbTraceCorrCoef) {//the energy should not be too small and the correlation coef should be higher.
          ilag1 += (imax-iSearchLag); //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = ilag1*ax_z.d;
        }
      }
    }
    //Now a match actually been found, load the matching filter on ig for the next round, otherwise the matching filter will keep the same.
    if (angle_axis_mask[igg]>0) {
      for (ii=-iHfEventWndSize; ii<=iHfEventWndSize; ii++) {
        matchFlter_(ii+iHfEventWndSize) = img_axt_slice(igg,iz_+ilag1+ii);
      }
      iggLast = igg;  fFltE = dotProd(matchFlter_,matchFlter_);
      fFltEmax = std::max(fFltE,fFltEmax);
      PRVEC(matchFlter_,2*iHfEventWndSize+1); PRV(matchFlter_[iHfEventWndSize]);PRV(fFltE); PRV(ilag1);
    }
  } // end for igg
}

void RMOfinder::locateMoveoutUsingCC(int igamma, int ilag, valarray<int>& angle_axis_mask, valarray<float>& RMO_per_angle, const stdVecFloat& estMoveoutPerAngle, const cubeDataF& img_axt_slice) {
  int igg,iggLast = igamma; int iSearchLag,ii;
  const axis& ax_z = img_axt_slice.domain().get_axis(1);
  int ng = estMoveoutPerAngle.size();
  int hfEventMatchWnd = maskCC_.domain().get_n(1)/2;
  int ilag1 = ilag, ind;
  for (igg=igamma-1; igg>=0; igg--) {
    if (abs(iggLast-igg)>5) break;//cannot find consecutviely for 5 traces, should give up.
    iSearchLag = std::max(MIN_LOC_SEARCH_WND,int(0.5f+1.0f/ax_z.d*fabs(estMoveoutPerAngle[iggLast]-estMoveoutPerAngle[igg])));
    for (ii=0; ii<iSearchLag; ii++) { //search from lag=0 to further lags.
      ind = ilag1+ii+hfEventMatchWnd;
      if (ind<2*hfEventMatchWnd+1 && ind>=0) {
        if (maskCC_(igg,ind)>0.5f) { //found
          ilag1 += ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = ilag1*ax_z.d;
          break;
        }
      }
      ind = ilag1-ii+hfEventMatchWnd;
      if (ind<2*hfEventMatchWnd+1 && ind>=0) {
        if (maskCC_(igg,ind)>0.5f) { //found
          ilag1 -= ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = ilag1*ax_z.d;
          break;
        }
      }
    }
    //mFltOutNMSup_;  matchFltCCout_;
  } //end for igg
  
  ilag1 = ilag; iggLast = igamma;
  for (igg=igamma+1; igg<ng; igg++) {
    if (abs(iggLast-igg)>5) break;//cannot find consecutviely for 5 traces, should give up.
    iSearchLag = std::max(MIN_LOC_SEARCH_WND,int(0.5f+1.0f/ax_z.d*fabs(estMoveoutPerAngle[iggLast]-estMoveoutPerAngle[igg])));
    for (ii=0; ii<iSearchLag; ii++) { //search from lag=0 to further lags.
      ind = ilag1+ii+hfEventMatchWnd;
      if (ind<2*hfEventMatchWnd+1 && ind>0) {
        if (maskCC_(igg,ind)>0.5f) { //found
          ilag1 += ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = ilag1*ax_z.d;
          break;
        }
      }
      ind = ilag1-ii+hfEventMatchWnd;
      if (ind<2*hfEventMatchWnd+1 && ind>0) {
        if (maskCC_(igg,ind)>0.5f) { //found
          ilag1 -= ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = ilag1*ax_z.d;
          break;
        }
      }
    }
    //mFltOutNMSup_;  matchFltCCout_;
  } //end for igg
}

//Look at the envelope of ADCIG, pick the peak and then choose it as moveout.
void RMOfinder::locateMoveoutUsingEnvelopMaximum(int igamma, int ilag, valarray<int>& angle_axis_mask, valarray<float>& RMO_per_angle, const stdVecFloat& estMoveoutPerAngle, const cubeDataF& img_axt_slice) {
  int igg,iggLast = igamma; int iSearchLag,ii;
  const axis& ax_z = img_axt_slice.domain().get_axis(1);
  int ng = img_axt_slice.domain().get_n(2);
  int shift_envpeak, d_envpeak_lag = 0;
  
  //The envelope of ADCIG is already in the envelopeZ_ADCIG_ cube, find the shift between Envelope peak and ilag.
  float fThr = NMSupEnvelope_.maxvalAbs()*0.1f;
  for (ii=0; ii<MIN_LOC_SEARCH_WND; ii++) {
    if (NMSupEnvelope_(ix_,igamma,iz_+ii) > fThr) {
      shift_envpeak = ii; break;
    }
    if (NMSupEnvelope_(ix_,igamma,iz_-ii) > fThr) {
      shift_envpeak = -ii; break;
    }
  }
  d_envpeak_lag = ilag - shift_envpeak;
    
  int ilag1 = shift_envpeak, ind; iggLast = igamma;
  for (igg=igamma-1; igg>=0; igg--) {
    if (abs(iggLast-igg)>5) break;//cannot find consecutviely for 5 traces, should give up.
    iSearchLag = std::max(MIN_LOC_SEARCH_WND,int(0.5f+1.0f/ax_z.d*fabs(estMoveoutPerAngle[iggLast]-estMoveoutPerAngle[igg])));
    for (ii=0; ii<iSearchLag; ii++) { //search from lag=0 to further lags.
      ind = ilag1+ii+iz_;
      if (ind<ax_z.n && ind>=0) {
        if (NMSupEnvelope_(ix_,igg,ind)>fThr) { //found
          ilag1 += ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = (ilag1+d_envpeak_lag)*ax_z.d;
          break;
        }
      }
      ind = ilag1-ii+iz_;
      if (ind<ax_z.n && ind>=0) {
        if (NMSupEnvelope_(ix_,igg,ind)>fThr) { //found
          ilag1 -= ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = (ilag1+d_envpeak_lag)*ax_z.d;
          break;
        }
      }
    }
  } //end for igg
  
  iggLast = igamma;
  for (igg=igamma+1; igg<ng; igg++) {
    if (abs(iggLast-igg)>5) break;//cannot find consecutviely for 5 traces, should give up.
    iSearchLag = std::max(MIN_LOC_SEARCH_WND,int(0.5f+1.0f/ax_z.d*fabsf(estMoveoutPerAngle[iggLast]-estMoveoutPerAngle[igg])));
    for (ii=0; ii<iSearchLag; ii++) {
      ind = ilag1+ii+iz_;
      if (ind<ax_z.n && ind>=0) {
        if (NMSupEnvelope_(ix_,igg,ind)>fThr) { //found
          ilag1 += ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = (ilag1+d_envpeak_lag)*ax_z.d;
          break;
        }
      }
      ind = ilag1-ii+iz_;
      if (ind<ax_z.n && ind>=0) {
        if (NMSupEnvelope_(ix_,igg,ind)>fThr) { //found
          ilag1 -= ii; iggLast = igg; //adjust the current lag.
          angle_axis_mask[igg]=1; RMO_per_angle[igg] = (ilag1+d_envpeak_lag)*ax_z.d;
          break;
        }
      }
    }
  } //end for igg
}

//Look at the envelope of ADCIG, pick the peak and then choose it as moveout.
void RMOfinder::locateMoveoutUsingSemblanceMaximum(int igamma, int ilag, valarray<int>& angle_axis_mask, valarray<float>& RMO_per_angle, const stdVecFloat& estMoveoutPerAngle, const cubeDataF& img_axt_slice, float& dJdrho, float& rhoPick) const{ 
  int igg,iggLast = igamma; int iSearchLag,ii;
  const axis& ax_z = img_axt_slice.domain().get_axis(1);
  int ng = img_axt_slice.domain().get_n(2);
  const axis& ax_deg = img_axt_slice.domain().get_axis(2);

  //find the maximum on the semblance panel
  const axis& ax_semb = envSemb_.domain().get_axis(2);
  stdVecFloat rhos(ax_semb.n);
  for (int i=0;i<ax_semb.n;i++) { rhos[i] = envSemb_(ix_,i,ilag+iz_); }

  int iRhoMax = std::max_element(rhos.begin(),rhos.end())-rhos.begin();
  //PRV2(iz_,ix_); PRV(ilag); PRVEC(rhos,ax_semb.n);
  float rhomax = ax_semb.get_coord(iRhoMax);
  float fdeg;

  //Now fill in the moveout function.
  for (igg=0; igg<ng; igg++) {
    fdeg = ax_deg.get_coord(igg);
    angle_axis_mask[igg]=1; RMO_per_angle[igg] = SQ(tanf(deg2rad(fdeg)))*rhomax;
  }
  //if (rhomax > 1e-6) {PRV(rhomax); PRV(RMO_per_angle[ng-1]);}
  
  //After we found rhomax, track the ADCIG along the moveout to remove those angles that are not predicted correctly or not presented with enough energy.
  float fRMOi = 0.0f;
  valarray<float> fE(ng); fE=0.0f;
  const int srchWnd = 2;
  int ioff, izz; float *start,*end;
  for (igg=0; igg<ng; igg++) {
    fRMOi = (RMO_per_angle[igg] / ax_z.d);
    izz = lroundf(iz_+fRMOi)+ilag;
    if (izz<srchWnd || izz+srchWnd>=ax_z.n) {continue;} //out of bound
    start = const_cast<float*>(&envelopeZ_ADCIG_(ix_,igg,izz-srchWnd));
    end   = const_cast<float*>(&envelopeZ_ADCIG_(ix_,igg,izz+srchWnd));
    fE[igg] = *(std::max_element(start,end+1));
  } //end for

  //Tick out values that are too small
  float fEmax = fE.max(), fEthres = fEmax * 0.33f;
  for (igg=0; igg<ng; igg++) {
    if (fE[igg]<fEthres) {angle_axis_mask[igg] = 0;}
  }
  for (igg=1; igg<ng-1; igg++) { //tick out the isolate points where its neighbors are not detected
    if (angle_axis_mask[igg]>0 && angle_axis_mask[igg-1]<1 && angle_axis_mask[igg+1]<1) { angle_axis_mask[igg] = 0;}
  }
  //LOG("K3");
  //calculate dJdrho, use Gaussian derivative.
  izz = iz_+ilag;
  cubeDataF gaussDeriv;
  int nSemb = ax_semb.n;
  float sum=0.0f,sum2=0.0f;
  for (int is=0; is<nSemb; is++) {
    sum   += gaussDeriv_(is)*envSemb_(ix_,is,izz);
    sum2  += envSemb_(ix_,is,izz);
  }
  dJdrho =  -sum/sum2;  //dotProd + scale(-1.0) equals to convolve the Guassian Deriv
  if (fabs(rhomax)<ax_semb.d) {
    dJdrho = fabs(dJdrho/ax_semb.d); //moveout is too small to count.
  } else {
    dJdrho /= rhomax; //divide the rhomax because it is already counted in the RMO computation
  }
  rhoPick = rhomax;
  //LOG("K4");
  return;
}

void RMOfinder::locateMoveoutUsingDipTracking(int igamma, int ilag, valarray<int>& angle_axis_mask, valarray<float>& RMO_per_angle, const stdVecFloat& estMoveoutPerAngle, const cubeDataF& img_axt_slice) {
  VERIFY(0);
  return;
}

int RMOfinder::RMOatPt(int* iCoords, const cubeDataF& img0, const cubeDataF& img_ax_t,const cubeDataF& img_ax_t_slice, const stdVecFloat& vecTan2Moveout,  //input
                       valarray<int>& angle_axis_mask, valarray<float>& RMO_per_angle, float& ptWeight,float& rhoPick)  //output 
{
  int iz = iCoords[0], ix = iCoords[1];
  ix_ = ix; iz_ = iz;
  ptWeight = 1.0f;

  int ret=0; int ii;
  //do local crosscorrelate within the ADCIG cube, and find the nearest match. set the corresponding slice.
  axis_deg ax_gamma = axis_deg(img_ax_t.domain().get_axis(2));
  const axis& ax_z = img_ax_t.domain().get_axis(1);
  const axis& ax_x = img_ax_t.domain().get_axis(3);
  int ng=ax_gamma.n;    //nz=ax_z.n;

  int iHfEventWndSize     = int(this->par_->fEventWndLength*0.5f/ax_z.d);
  int iHfEventMatchingWnd = int(this->par_->fEventMatchingWndLen*0.5f/ax_z.d);
  NMSup_Param nmsup_par;  nmsup_par.wnd1_ = int(2*iHfEventWndSize+1); nmsup_par.wnd2_=1;
  
  axis ax_flt(2*iHfEventWndSize+1,-iHfEventWndSize*ax_z.d+ax_z.o,ax_z.d);
  matchFlter_.alloc(hypercube(ax_flt));
  
  for (ii=-iHfEventWndSize; ii<=iHfEventWndSize; ii++) {
    matchFlter_(ii+iHfEventWndSize) = img0(ix,iz+ii);
  }

  int istartLag=iz-iHfEventMatchingWnd,iendLag=iz+iHfEventMatchingWnd;
  crossCorr1dGeneral(matchFltCCout_, matchFlter_, img_ax_t_slice,
                     istartLag,iendLag, NULL);
  mFltOutNMSup_.alloc(matchFltCCout_.domain());
  //find an anchor point to start measuring the neiboring moveout.
  nmSup(matchFltCCout_,mFltOutNMSup_,&nmsup_par);
  //Mask out small values.
  mFltOutNMSup_.allocMemoryTo(maskCC_);
  buildThresholdMask(maskCC_,mFltOutNMSup_,mFltOutNMSup_.maxval()*(this->par_->thrCCmaxvalRatio_),false);
  //locate the angle where the matched event are closest.
  //matchFlter_.dump("dump15");img_ax_t_slice.dump("dump16"); matchFltCCout_.dump("dump17"); mFltOutNMSup_.dump("dump18"); maskCC_.dump("dump19");
  
  //YPAUSE;
  //LOG("g3");
  int igamma,ilag,ig;
  //only search near angles that are smaller than 30deg, and from center lag to larger lags, stops while the first candidate is found.
  int igmin = ax_gamma.get_offset(0.0f), ngmaxSrch = ax_gamma.get_offset(20.0f) - igmin;
  bool bFound=false;
  for (int iil=0; iil<=5; iil++) {
    ilag = iHfEventMatchingWnd-iil;
    for (ig=0;ig<=ngmaxSrch;ig++) {
      igamma = igmin+ig;
      if (igamma<ng && maskCC_(igamma,ilag)>0.5f) {
        bFound = true;  break;
      }
      igamma = igmin-ig;
      if (igamma>=0 && maskCC_(igamma,ilag)>0.5f) {
        bFound = true;  break;
      }
    }
    if (bFound) break; //the cand has been found, otherwise search the other lag.
    ilag = iHfEventMatchingWnd+iil;
    for (ig=0;ig<=ngmaxSrch;ig++) {
      igamma = igmin+ig;
      if (igamma<ng && maskCC_(igamma,ilag)>0.5f) {
        bFound=true; break;
      }
      igamma = igmin-ig;
      if (igamma>=0 && maskCC_(igamma,ilag)>0.5f) {
        bFound = true;  break;
      }
    }
    if (bFound) break;
  } //end for iil
  
  if (!bFound) {
    //cannot find a good match, we should discard this point, and do the next point.
    prt("Cannot locate the moveout match at CandPt: (%f,%f)!\n", ax_z.get_coord(iz),ax_x.get_coord(ix));
    ret = -1; return ret; //search failed, return this function.
  }
  
  float vErrRatio = par_->fApproxVelErrRatio_;  float fWaterDepth   = par_->waterDepthAver_;
  //LOG("g4");
  angle_axis_mask = 0; //zero the angle_axis_mask array.
  //now locate the moveout at the neiboring angles from my ankor point, here we can use a small search window, because the neiboring events should not be too far from each other.

  //PRV(ax_gamma.get_coord(igamma));
  ilag -= iHfEventMatchingWnd;
  angle_axis_mask[igamma] = 1; RMO_per_angle[igamma] = ilag*ax_z.d;
  
  stdVecFloat estMoveoutPerAngle(ng); 
  float z = ax_z.get_coord(iz)-fWaterDepth;
  for (ig=0;ig<ng;ig++) {estMoveoutPerAngle[ig] = z*vErrRatio*vecTan2Moveout[ig];}
  
  if (par_->method_ == Param_RMO_Measurer::NbCorr) {
    locateMoveoutUsingNbCorrelation(igamma,ilag, angle_axis_mask,RMO_per_angle, estMoveoutPerAngle, img_ax_t_slice);
  } else if (par_->method_ == Param_RMO_Measurer::CrossCorr) { //simplest, just look at the maximum in the cross-correlation panel.
    locateMoveoutUsingCC(igamma,ilag, angle_axis_mask,RMO_per_angle, estMoveoutPerAngle, img_ax_t_slice);
    //PRVEC(RMO_per_angle,ng);PRVEC(angle_axis_mask,ng);
  } else if (par_->method_ == Param_RMO_Measurer::EnvMax) {
    locateMoveoutUsingEnvelopMaximum(igamma,ilag, angle_axis_mask,RMO_per_angle, estMoveoutPerAngle, img_ax_t_slice);
  } else if (par_->method_ == Param_RMO_Measurer::dipTracking) {
    VERIFY(0);
  } else if (par_->method_ == Param_RMO_Measurer::Semblance) {
    locateMoveoutUsingSemblanceMaximum(igamma,ilag, angle_axis_mask,RMO_per_angle, estMoveoutPerAngle, img_ax_t_slice, ptWeight,rhoPick);
  } else {
    VERIFY(0);
  }
  
  //do some clean up.
  int ig0 = ax_gamma.get_offset(0.0f); //put zero degree.
  VERIFY(equalf(float(ig0),ax_gamma.get_offsetf(0.0f)));
  int cnt0;
  cnt0 = 0;
  for (ig=0;ig<ng; ig++) { //if for the entire we cannot find RMO measured as zero, then it is a weird case, and I will disgard it.
    if (angle_axis_mask[ig]>0 && fabsf(RMO_per_angle[ig]) < 0.5f*ax_z.d) cnt0++;
  }
  //SEEV(cnt0); SEEVEC(&(*RMO_)(ip,0),ng);
  //if (cnt0<5) {prt("point at (%f,%f)is discarded due to cnt0=%d\n",ax_z.get_coord(iz),ax_x.get_coord(ix),cnt0); ret = -2;}//disgard this point
  if (angle_axis_mask.sum()<ng/4) ret = -3; //The toal number of detection is too few.
  return ret;
}

