#include "stdafx.h"
#include "cubeData.hpp"
#include "gVec.h"
#include "dampingFunc.h"
#include "waveOps.h"
#include "misc.h"
#include "off2ang3d.h"
#include "WEMVA.h"

#include "dmRsfDump.h"
#include "param_func.h"

using namespace Yang;
using namespace std;


//Forward modeling, basically do imaging
void WEMVA_FD::calcData() {
  FUNC_TRACE_BEG(WEMVA_FD);
  //Here the Data is actuall the image from the velocity.
  //Init the operator, do RTM to create the prestack image
  
  waveOps*            pModeling = this->waveOp().get();
  const waveOpParams* pFDParam  = pModeling->waveOpPars().get();
  pModeling->initOperatorEx(pFDParam, waveOps::acMigration);   
  
  //verbPrt("\r RTM Imaging from the current velocity in WEMVA module s\n");
  //Need to allocate Img0_hx here
  hypercube shapeOffImg;
  const ImgPar* imgPar = pModeling->imgPars().get();
  bool bDoOffset       = imgPar->bGenOffsetGather();
  bool b3D             = pModeling->acqPars()->b3D();
  
  if (bDoOffset) { //figure out the size of the subsurface offset image.
    shapeOffImg = imgPar->finalSubOffImageShp();
  }
  if (!b3D) { //if not doing 3D, then everything is in core.
    if (bDoOffset) {  //Do non-zero offset
      m_pWEMVAparams->Img0_hx.alloc(shapeOffImg);
    } else { //Only do zero offset
      m_pWEMVAparams->Img0_hx.alloc(pFDParam->finalImageShape);
    }
  }
  
  const velAnaParam* pVelAnaParam = this->velAnaPar();
  //pModeling->migration_op(&(m_pVelAnaParam->dObsCube),&(this->m_pWEMVAparams->Img0_hx));
  //To accomodate for Oneway, here change cubeData -> dmRsf_oc, because dObserved never changes.
  if (!pFDParam->bOneway()) {
    assert(pVelAnaParam->pDataObs);
    pModeling->migration_op(pVelAnaParam->pDataObs,&(this->m_pWEMVAparams->Img0_hx));
    VERIFY(!b3D); //Only implement 2-D case in FDTD, can be extended later.
  } else { //Use Yaxun's oneway propagator, Use the file tag as input
    VERIFY(!pVelAnaParam->dObsFilename.empty());
    if (!b3D) { //2D
      pModeling->migration_op(pVelAnaParam->dObsFilename,&(this->m_pWEMVAparams->Img0_hx));
    } else { //3D, I/O are filenames/tags both.
      VERIFY(!m_pWEMVAparams->fn_img_off.empty()); //the file name should be assigned by now.
      PRV2(pVelAnaParam->dObsFilename,m_pWEMVAparams->fn_img_off);
      pModeling->migration3d_op(pVelAnaParam->dObsFilename,m_pWEMVAparams->fn_img_off);
    }
  }
  LOGC("Finish calculating image gather in WEMVA_FD::calcData");
  
  //Put the zero offset back into Img0
  const hypercube& zeroOffImgShp = imgPar->finalImageShp();
  if (!b3D) { //if not doing 3D, then everything is in core.
    if (bDoOffset) {  //Do non-zero offset
      m_pWEMVAparams->Img0.alloc(zeroOffImgShp);
      windParam wind(3,windowAxis());
      axis ax_h = shapeOffImg.get_axis(3);
      //Apply the imaging mask to the image.
      if (m_pWEMVAparams->imgh0_mask.get()!=NULL) {
        m_pWEMVAparams->Img0_hx.info("m_pWEMVAparams->Img0_hx");
        multiplySlice(m_pWEMVAparams->Img0_hx,m_pWEMVAparams->Img0_hx,*m_pWEMVAparams->imgh0_mask);
      }
      wind[2].n = 1; wind[2].f = ax_h.get_offset(0.0f); 
      m_pWEMVAparams->Img0_hx.grab2d(&m_pWEMVAparams->Img0,wind);
      //Here we just crop out the region we don't need
      if (m_pWEMVAparams->croppedImg0_) {
        cropAlign(*(m_pWEMVAparams->croppedImg0_),m_pWEMVAparams->Img0,false,true);
        cropAlign(m_pWEMVAparams->Img0,*(m_pWEMVAparams->croppedImg0_),false,true);
      }
      //transpose the image and put it in hx_t cube.
      seplib::callSepTool(m_pWEMVAparams->Img0_hx_t,m_pWEMVAparams->Img0_hx,"Transp"," plane=23",true);
    } else {
      m_pWEMVAparams->Img0_hx.copyTo(m_pWEMVAparams->Img0);
    }
  } else { //3D case
    VERIFY(0);
  }
  onImageUpdated();
  this->m_pWEMVAparams->Img0_hx.dump("imgh.H",true); //LOG(finish RTM);
  FUNC_TRACE_END();
}

float WEMVA_FD::calcObjFunc() {
  float ret = -1234.0f; //default value
  switch(m_pWEMVAparams->objFuncType_) {
  case WEMVA_Param::cheatTrueModObjFunc:
    ret = cheatTrueModObjFunc();
    break;
  case WEMVA_Param::StackPowerObjFunc:
    ret = stackPowerObjFunc();
    break;
  case WEMVA_Param::NormalizedStackPowerOF:
    ret = NormStackPowerObjFunc();
    break;
  case WEMVA_Param::DSO_off_objFunc:
    ret = offDSO_ObjFunc();
    break;
  default: //dismiss the rest cases
    
    break;
  }
  return ret;
}

//cheat obj function, basically minimize model difference between true and current
float WEMVA_FD::cheatTrueModObjFunc() {
  VERIFY(m_pWEMVAparams->vmodTrue.get());
  cubeDataF dv;
  m_pWEMVAparams->vmodTrue->copyTo(dv);
  dv.subtract(this->waveOp()->vmod());
  return dotProd(dv,dv);
}

void WEMVA_FD::onImageUpdated() {
  this->m_bAngleGatherCalculated = false; //here reset the AngleGather has not been calculated yet.
}

const cubeDataF& WEMVA_FD::getCalcData() {
  bool b3D = this->waveOp()->acqPars()->b3D();
  if (!b3D) {
//    if (m_bCalcAngleGather) { //output angle gather
//      return m_pWEMVAparams->Img0_ax_t;
//    } else {
      return m_pWEMVAparams->Img0_hx;
//    }
  } else {
    VERIFY(0);
  }
}

int WEMVA_FD::setCalcData(const cubeDataF& data) {
//  if (m_bCalcAngleGather) { //output angle gather
//    m_pWEMVAparams->Img0_ax_t = data;
//  } else {
    m_pWEMVAparams->Img0_hx = data;
//  }
  bool b3D = this->waveOp()->acqPars()->b3D();
  VERIFY(!b3D);
  return 0;
}

string WEMVA_FD::getCalcData_fn() {
  bool b3D = this->waveOp()->acqPars()->b3D();
  if (b3D) {
    //if (m_bCalcAngleGather) { //output angle gather
    //  return m_pWEMVAparams->fn_img_ang;
    //} else {
      return m_pWEMVAparams->fn_img_off;
    //}
  } else { //not for 2D currently.
    VERIFY(0);
  }
}

int WEMVA_FD::setCalcData_fn(const string& img_fn) {
  LOGC("Enter WEMVA_FD::setCalcData_fn");
  //Mv the file to my current file name
  char szcmd[512];
  //if (m_bCalcAngleGather) { //output angle gather
  //  sprintf(szcmd,"Mv %s %s", img_fn.c_str(), m_pWEMVAparams->fn_img_ang.c_str());
  //} else {
    sprintf(szcmd,"Mv %s %s -f", img_fn.c_str(), m_pWEMVAparams->fn_img_off.c_str()); //w/out -f are both fine
  //}
  //Execute command.
  prt("Exec cmd in WEMVA_FD::setCalcData_fn:%s\n", szcmd);
  VERIFY(system(szcmd)==0);
  this->m_bAngleGatherCalculated = false;
  return 0;
}

//Generate image gather.
void WEMVA_FD::postCalcData() {
  //PRV(m_bCalcAngleGather);
  //PRV(this->waveOp().get());PRV(this->waveOp()->acqPars());
  bool b3D = this->waveOp()->acqPars()->b3D();
  //LOG("i1");
  //After then calculate the angle gathers if needed
  if (m_bCalcAngleGather) {
    if (!m_bAngleGatherCalculated) {
      if(!b3D) { //if do 2D only.
        this->convertAngleOffImage(m_pWEMVAparams->Img0_ax_t, m_pWEMVAparams->Img0_hx,true);
      } else { //Use biondo's program to generate 3D angle gathers, then we need to calculate from off2ang3d program.
        off2ang3d_file(*(m_pWEMVAparams->par_o2a3d),m_pWEMVAparams->fn_img_off,m_pWEMVAparams->fn_img_ang);
      }
      m_bAngleGatherCalculated = true; //already calc angle gathers.
    } else {
      prt("Warning:enter postCalcData with bAngleGatherCalculated is true\n");
    }
  }
  //LOG("i2");
  return;
}

//Maximize the stack power of objective function
//Lets assume the objective function is the angle stack power,
//ie the energy of the zero offset image.
float WEMVA_FD::stackPowerObjFunc() {
  FUNC_TRACE_BEG(WEMVA_FD:stackPowerObjFunc);
  //Since we have the image in Img0_hx
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  float costFunc = 0.0f;

  bool b3D = this->waveOp()->acqPars()->b3D();
  cubeData<float> img_hx0;
  if (!b3D) { //if it is 2D, then everything is incore
    //Assume the initial images are ready
    VERIFY(pWEMVAparam->Img0_hx.domain().get_n123() > 0);
    
    //Get the zero offset image.
    hypercube offsetImgShape = pWEMVAparam->Img0_hx.domain();
    hypercube zeroOffImgShape = offsetImgShape.return_compatible(1,2);
    
    //Should not try to take the short cut to calculate the zero offset image. Rather I should stack over all angles. !!!!
    img_hx0.alloc(zeroOffImgShape);
    windParam wind(3,windowAxis());
    axis ax_h = offsetImgShape.get_axis(3);
    wind[2].f = ax_h.get_offset(0.0f); wind[2].n = 1;
    pWEMVAparam->Img0_hx.grab2d(&img_hx0,wind);
    
  } else { //the image is in the file name below,
    //the axis (z,x,y,hx,hy), thus take the hx=hy=0 slice.
    string fn = pWEMVAparam->fn_img_off;
    PRV(pWEMVAparam->fn_img_off);
    dmRsf_oc<float> dm_img(fn,dm::IN,"input off img",true);
    dm_img.init(NULL,5);
    windParam wd(5,windowAxis());
    axis ax_hx,ax_hy;
    ax_hx = dm_img.domain().get_axis(4); ax_hy = dm_img.domain().get_axis(5);
    wd[3].f = ax_hx.get_offset(0.0f);wd[3].n=1;
    wd[4].f = ax_hy.get_offset(0.0f);wd[4].n=1;
    img_hx0.alloc(dm_img.domain().return_compatible(1,3));
    dm_img.grabWindow(&img_hx0,wd);
    //dm_img.unInit(); //no need to call, the destructor will call it automatically.
  }
  img_hx0.dump("dump1");
//if (pWEMVAparam->bCompensateImageZ) {
//  //scale the image by depth, to compensate the amplitude decay with depth.
//  axis ax_z = img_hx0.domain().get_axis(1);
//  float zrange = ax_z.n * ax_z.d;
//  cubeData<float> cubz(hypercube(ax_z.n));
//  for (int i=0; i<ax_z.n; i++) {
//    cubz(i) = ax_z.get_coord(i)-ax_z.o + 0.2*zrange;  //The shallowest depth will have a scaling coef of 0.2*zrange
//  }
//  cubz.dump("dump4");
//  multiplySlice(img_hx0,img_hx0,cubz);
//}
  
  //since this is maximization, therefore scale with -1.
  costFunc = -1.0 * dotProd(img_hx0,img_hx0);
  FUNC_TRACE_END(costFunc);
}

//compute (I^2(h=0))/(sum_{h}I^2(h));
float WEMVA_FD::NormStackPowerObjFunc() {
  FUNC_TRACE_BEG("WEMVA_FD::NormStackPowerObjFunc");
  //Since we have the image in Img0_hx
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  float costFunc = 0.0f;

  bool b3D = this->waveOp()->acqPars()->b3D();
  cubeData<float> img_hx0,img_hsum,img_hx0_bal;

  if (!b3D) { //if it is 2D, then everything is incore
    //Assume the initial images are ready
    VERIFY(pWEMVAparam->Img0_hx.domain().get_n123() > 0);
    //Get the zero offset image.
    hypercube offsetImgShape = pWEMVAparam->Img0_hx.domain();
    hypercube zeroOffImgShape = offsetImgShape.return_compatible(1,2);
    //Should not try to take the short cut to calculate the zero offset image. Rather I should stack over all angles. !!
    img_hx0.alloc(zeroOffImgShape); img_hsum.alloc(zeroOffImgShape); img_hx0_bal.alloc(zeroOffImgShape);
    windParam wind(3,windowAxis());
    axis ax_h = offsetImgShape.get_axis(3);
    wind[2].f = ax_h.get_offset(0.0f); wind[2].n = 1;
    pWEMVAparam->Img0_hx.grab2d(&img_hx0,wind);
    img_hx0.multiply(img_hx0); //compute I^2(h=0)

    pWEMVAparam->Img0_hx.copyTo(pWEMVAparam->Img_hx_res);
    pWEMVAparam->Img_hx_res.multiply(pWEMVAparam->Img_hx_res); //compute I^2(h)
    //Stack for all h
    multAndStackAlongAxis(img_hsum, pWEMVAparam->Img_hx_res, PosInt(3));
    
    //now divide, first retrieve the window information
    seplib::AGC_Param agcP(*(pWEMVAparam->imgBalAgcPar_)); agcP.normPow_ = 1;
    seplib::AGCdivide(img_hx0,img_hsum,m_pWEMVAparams->imgh0_mask.get(),img_hx0_bal,&agcP);

    dmRsf_ocDump<float> imgDump("dump17","imgbal for normalized stack power");
    imgDump.init(&zeroOffImgShape);
    imgDump.dumpFrame(img_hx0); imgDump.dumpFrame(img_hsum); imgDump.dumpFrame(img_hx0_bal);

    costFunc = img_hx0_bal.sum();
  } else { //the image is in the file name below,
    //the axis (z,x,y,hx,hy), thus take the hx=hy=0 slice.
    //dm_img.unInit(); //no need to call, the destructor will call it automatically.
    VERIFY(0); //Not implemented yet.
  }

//if (pWEMVAparam->bCompensateImageZ) {
//  //scale the image by depth, to compensate the amplitude decay with depth.
//  axis ax_z = img_hx0.domain().get_axis(1);
//  float zrange = ax_z.n * ax_z.d;
//  cubeData<float> cubz(hypercube(ax_z.n));
//  for (int i=0; i<ax_z.n; i++) {
//    cubz(i) = ax_z.get_coord(i)-ax_z.o + 0.2*zrange;  //The shallowest depth will have a scaling coef of 0.2*zrange
//  }
//  cubz.dump("dump4");
//  multiplySlice(img_hx0,img_hx0,cubz);
//}

  //since this is maximization, therefore scale with -1.
  costFunc = -1.0 * costFunc;
  FUNC_TRACE_END(costFunc);
}

//FWI like WEMVA backprojection method.
void  WEMVA_max0offset::calcBackProjData() {
  FUNC_TRACE_BEG(WEMVA_maxAngleStack::calcBackProjData);
  
  //this->postCalcData(); //don't need to calculate Angle gather at this step
  
  waveOps& waveOp = (*this->waveOp());
  bool b3D = waveOp.acqPars()->b3D();
  
  cubeData<float> img_hx0;
  //The image to back project is just the 0 offset image itself.  
  if (b3D) {
    for (int i=0; i<5; i++) {
    PRV(i);
    dmRsf_oc<float> dm_img_off(m_pWEMVAparams->fn_img_off,    dm::IN    ,"img_off",true);
    PRV(m_pWEMVAparams->fn_bpjImg_off);
    dmRsf_oc<float> dm_pbjImg_off(m_pWEMVAparams->fn_bpjImg_off, dm::OUT,"bpjimg_off",true);
    dm_img_off.init(NULL,5);
    const hypercube& shpOffimg = dm_img_off.domain();
    LOG("cc2"); shpOffimg.info("shpoffimg");
    dm_pbjImg_off.init(&shpOffimg);
    LOG("cc3");
    //shpOffimg.info("shpOffimg");
    //Take the hx=hy=0 slice only, the axis order is (z,x,y,hx,hy);
    windParam wd(5,windowAxis());
    axis ax_hx,ax_hy;
    ax_hx = dm_img_off.domain().get_axis(4); ax_hy = dm_img_off.domain().get_axis(5);
    wd[3].f = ax_hx.get_offset(0.0f);wd[3].n=1;
    wd[4].f = ax_hy.get_offset(0.0f);wd[4].n=1;
    img_hx0.alloc(shpOffimg.return_compatible(1,3));
    //dm_img_off.grabWindow(&img_hx0,wd); 
    dm_img_off.seq_grab(&img_hx0);
    //then copy the img back
    //PRV(wd.size()); dm_pbjImg_off.domain().info("dm_pbjImg_off_domain");
    LOG("cc4");
    //dm_pbjImg_off.update3d(&img_hx0,wd);
    dm_pbjImg_off.seq_update(&img_hx0);
    auxclose(m_pWEMVAparams->fn_bpjImg_off.c_str());
    LOG("cc5");
    LOGC('b3D=1 in WEMVA_max0offset::calculateBackProjData');
    dm_pbjImg_off.unInit();
    }
  } else { //2D case, the image is z,x,hx
    //Get the zero offset image.
    hypercube offsetImgShape = m_pWEMVAparams->Img0_hx.domain();
    hypercube zeroOffImgShape = offsetImgShape.return_compatible(1,2);
    img_hx0.alloc(zeroOffImgShape);
    windParam wind(3,windowAxis());
    axis ax_h = offsetImgShape.get_axis(3);
    wind[2].f = ax_h.get_offset(0.0f); wind[2].n = 1;
    m_pWEMVAparams->Img0_hx.grab2d(&img_hx0,wind);
    
    m_pWEMVAparams->backProjImg_hx.alloc(offsetImgShape);
    m_pWEMVAparams->backProjImg_hx.update2d(&img_hx0,wind);
  }
  LOGC('WEMVA_max0offset::calcBackProjData');
  FUNC_TRACE_END();
}

//velGradCube is been allocated outside.
void WEMVA_FD::calcSlowGradient( cubeData<float> & slowGradCube) {
  FUNC_TRACE_BEG(WEMVA_FD::calcSlowGradient);
  WEMVA_Param* pWEMVAparam  = m_pWEMVAparams;
  waveOps*      pModeling   = this->waveOp().get();
  const waveOpParams* TomoParams  = pModeling->waveOpPars().get();
  
  //Init the operator, doing TomAdj this time
  pModeling->initOperatorEx(TomoParams, waveOps::acTomAdj);
  
  //Verify the size of velGradCube
  slowGradCube.zero();
  //doing TomAdj this time to calculate gradient
  //The input is in &(pWEMVAparam->backProjImg_hx
  //pWEMVAparam->backProjImg_hx.dump("dump5");
  bool b3D = pModeling->acqPars()->b3D();
  
  if (!b3D) { //doing 2D case.
    pModeling->TomAdj_op(&(pWEMVAparam->backProjImg_hx) ,&slowGradCube);
  } else { //doing 3D case
    pModeling->TomAdj3d_op(pWEMVAparam->fn_bpjImg_off   ,slowGradCube);
    //string fn_slowGrad;
    //ostringstream oss;
    //oss << "slgd" << rand()%65536 << ".H";
    //fn_slowGrad = oss.str();
    //  //Then we read the file back into memory
    //  dmRsf_oc<float> dm_slow(fn_slowGrad,dm::IN,"",true);
    //  dm_slow.init(NULL);
    //  assert(slowGradCube.domain().sameShape(dm_slow.domain()));
    //  dm_slow.seq_grabAll(slowGradCube);
  }
  
  //Don't need to scale the slowness grad by s0, it is included in the TomAdj Operator
  //slowGradCube.dump("dump3");
  FUNC_TRACE_END();
}

void WEMVA_FD::flipGradient(cubeData<float>& gradCube) {
   //Because the magic -1 phenomenon, flip the sign
   //gradCube.scale(-1.0f);
}

void WEMVA_FD_DSO_off::flipGradient(cubeData<float>& gradCube) {
   //gradCube.scale(-1.0f);
}

void WEMVA_FD::gen_h_term_cube(cubeData<float>& hTermCube, const hypercube& shapeOffImage, int h_order) {
  FUNC_TRACE_BEG(WEMVA_FD_DSO_off::gen_h_term_cube);
   int ndim = shapeOffImage.get_ndim();
   hypercube shape_h_term = shapeOffImage.return_axes(3, ndim);
   hTermCube.alloc(shape_h_term);
   //Just do 2D in this case
   assert(ndim==3);
   axis ax_hx = shape_h_term.get_axis(1);
   //Set the h cube
   float fh,h;
   for (int i=0; i<ax_hx.n; i++) {
	  h = i*ax_hx.d + ax_hx.o; fh=1.0f;
	  for (int j=0; j<h_order; j++) {
		 fh = fh*h;
	  }
	  hTermCube[i] = fh;
   }
   FUNC_TRACE_END();
}

//virtual function
float WEMVA_FD::offDSO_ObjFunc() {
  FUNC_TRACE_BEG(WEMVA_FD_DSO_off::calcObjFunc);
  
   //need to calcualte the initial images first somewhere before
   //Check the initial Image is not NULL
   WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
   VERIFY(pWEMVAparam->Img0_hx.domain().get_ndim() >= 2);
   //then calc the cost function, which is the || h*I(x,h) ||^2
   //Since we have the image in Img0_hx
   const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
   cubeData<float> hTermCube;
   this->gen_h_term_cube(hTermCube, shapeOffImage,1);
   pWEMVAparam->Img_hx_res.alloc(shapeOffImage);

   const axis& ax_z = shapeOffImage.get_axis(1);
   const axis& ax_x = shapeOffImage.get_axis(2);
   const axis& ax_h = shapeOffImage.get_axis(3);
   float costFunc;

   if (pWEMVAparam->dso_bal_type_ == WEMVA_Param::none_bl) {
     scaleSlice(pWEMVAparam->Img_hx_res, pWEMVAparam->Img0_hx,hTermCube);
     costFunc = dotProd(pWEMVAparam->Img_hx_res,pWEMVAparam->Img_hx_res);
   } else if (pWEMVAparam->dso_bal_type_ == WEMVA_Param::x_bal) {
     costFunc = 0.0f;
     //balance along x axis, using Tang's thesis formula
     scaleSlice(pWEMVAparam->Img_hx_res, pWEMVAparam->Img0_hx,hTermCube);
     //now stack the image along z and h axis
     int iz,ix,ih; float f1,f2;
     //for bp model
     //int ixmin = ax_x.get_offset(0.0f); int ixmax = ax_x.get_offset(19.0f);
     int ixmin = ax_x.get_offset(gParamReader->get_float("imgval_xmin")); 
     int ixmax = ax_x.get_offset(gParamReader->get_float("imgval_xmax"));
     PRV2(ixmin,ixmax);
     for (ix=ixmin; ix<=ixmax; ix++) {
       f1=f2=0.0;
       for (ih=0; ih<ax_h.n; ih++) {
       for (iz=0; iz<ax_z.n; iz++) {
         f1 += SQ(pWEMVAparam->Img_hx_res(ih,ix,iz));
         f2 += SQ(pWEMVAparam->Img0_hx(ih,ix,iz));
       }}
       costFunc += f1/f2;
     }
   } else {
     VERIFY(0);
   }
   PRT_LOC; prt("finish calc off_DSO Obj Func, = %f\n", costFunc);
   FUNC_TRACE_END(costFunc);
}

//Calculate Gradient, this is a virtual function
void WEMVA_FD_DSO_off::calcBackProjData() {
   //The Back Proj data is the Image -(I(x,h)h^2)
   //The Image0 is been calculated already.
   WEMVA_Param* pWEMVAparam = m_pWEMVAparams;

   const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
   cubeData<float> hTermCube;
   this->gen_h_term_cube(hTermCube, shapeOffImage,2);	//Need h^2
   pWEMVAparam->backProjImg_hx.alloc(shapeOffImage);

   const axis& ax_z = shapeOffImage.get_axis(1);
   const axis& ax_x = shapeOffImage.get_axis(2);
   const axis& ax_h = shapeOffImage.get_axis(3);
   if (pWEMVAparam->dso_bal_type_ == WEMVA_Param::none_bl) {
     scaleSlice(pWEMVAparam->backProjImg_hx, pWEMVAparam->Img0_hx,hTermCube);
   } else if (pWEMVAparam->dso_bal_type_ == WEMVA_Param::x_bal) {
     int ix,iz,ih; float f1,f2;
     //balance the amplitude difference along x direction
     int ixmin = ax_x.get_offset(gParamReader->get_float("imgval_xmin")); 
     int ixmax = ax_x.get_offset(gParamReader->get_float("imgval_xmax"));
     for (ix=ixmin; ix<=ixmax; ix++) {
       f1=f2=0.0;
       for (ih=0; ih<ax_h.n; ih++) {
       for (iz=0; iz<ax_z.n; iz++) {
         f1 += SQ(pWEMVAparam->Img_hx_res(ih,ix,iz));
         f2 += SQ(pWEMVAparam->Img0_hx(ih,ix,iz));
       }}
       for (ih=0; ih<ax_h.n; ih++) {
       for (iz=0; iz<ax_z.n; iz++) {
       pWEMVAparam->backProjImg_hx(ih,ix,iz) = (f2*hTermCube(ih)-f1)/(f2*f2)
                                                *pWEMVAparam->Img0_hx(ih,ix,iz);
       }}
     }
   } else {
     VERIFY(0);
   }

   //If only calc sensitivity kernel, then only back project on image point of the angle gather.
   //i.e Mask out all the values in the image gathers but the center one.
   if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
     axis ax_x = pWEMVAparam->backProjImg_hx.domain().get_axis(2);
     hypercube shp1(ax_x);  cubeData<float> mask_x(shp1);
     mask_x = 0.0f;
     
     //Put six spikes into the domain
     int dnw = ax_x.n/8;
     for (int i=0; i<7; i++) {
       mask_x(ax_x.n/2+(i-3)*dnw) = 1.0f;
     }
     
     multiplySubCube(pWEMVAparam->backProjImg_hx,pWEMVAparam->backProjImg_hx,mask_x,2,2);
   }
   
   //There is a negative sign 
   pWEMVAparam->backProjImg_hx.scale(-1.0f);
}

//For WEMVA objFunc to update status in a iterated solver
void WEMVA_FD::callBackAtIter(int iIter) {
  velAnaParam* pVelAnaParam = this->velAnaPar();
  //update my status, currently just change the dJdu param
  bool bUseSembForDJDU = pVelAnaParam->bUseSemblanceToCalcDJdb;
  if (bUseSembForDJDU) {
    const axis &ax_semb = pVelAnaParam->ax_semb;
    float fMinSembWnd   = ax_semb.d * 6.0f; 
    float fSembWindow   = pVelAnaParam->fSembSmoothWindow;
    pVelAnaParam->fSembSmoothWindow = MAX(fSembWindow/1.3f, fMinSembWnd);
  }
  return;
}

/////////////////// WEMVA objective function Interfaces///////////////////////////////////////////
//model is velocity, simply wrap around the function interface
float WEMVA_FD::evaluate(const gVecComponent& model) {
  FUNC_TRACE_BEG(WEMVA_FD::evaluate);
  //set the model to the FWI object
  //convert slowness to velocity
  cubeData<float> vel;
  //LOG(mm1);
  model.allocMemoryTo(vel);
  reciprocal(vel,model);
  //LOG(mm2);
  this->setVelocity(&vel);
  
  //LOG(calcualte initial image);
  //First calculate the data
  this->calcData();
  
  //LOG(mm4);
  //LOG(calc obj func in wemva_evaluate);
  float ret = this->calcObjFunc();
  if (std::isnan(ret)) {
    prt("\n the cost func is nan ! ");
  }
  FUNC_TRACE_END(ret);
}

void  WEMVA_FD::calcSearchDir( gVecComponent &searchDir, const gVecComponent& curModel) {
  FUNC_TRACE_BEG(WEMVA_FD::calcSearchDir);
  //Here we have alrceady calculated the data, since we evaluated the obj function already.
  this->calcBackProjData();
  m_pWEMVAparams->backProjImg_hx.dump("dimg.H",true);
  //calculate gradient this time
  //It is actually SLOWNESS gradient
  this->calcSlowGradient(searchDir);
  
  this->flipGradient(searchDir);
  
  //not masking before smoothing yet.
  if (NULL != velAnaPar()->gradMaskPreSmooth_.get()) {
    searchDir.multiply(*velAnaPar()->gradMaskPreSmooth_);
  }
  //m_pWEMVA->m_pWEMVAparams->Img0_hx.dump("dump6");
  //searchDir.dump("dump7");
  FUNC_TRACE_END();
}

