#ifndef __WEMVA_H_INCL__
#define __WEMVA_H_INCL__

// TODO: your code here
#include "velAna.h"
#include "objFunc.h"

using boost::shared_ptr;

namespace Yang {

  //Inherit the objFunc interface as well 
class WEMVA_FD:public velAna_FD, public objFunc {
public:
  // Constructor
  WEMVA_FD(velAnaParam* pVelAnaParam, const shared_ptr<waveOps>& pWave, WEMVA_Param* pWEMVApar ):
                  velAna_FD(pVelAnaParam,pWave),m_pWEMVAparams(pWEMVApar), m_bCalcAngleGather(true), m_bAngleGatherCalculated(false){}
  // Destructor
  virtual ~WEMVA_FD();
  
  virtual void calcData();
  void postCalcData();
  virtual void calcBackProjData() = 0;
  //calculate objective function, used in fn evaluate();
  virtual float calcObjFunc();

  virtual const   cubeDataF& getCalcData();
  virtual string  getCalcData_fn();
  virtual int setCalcData(const cubeData<float>& data);
  virtual int setCalcData_fn(const string& fn_data);
  
  //The interfaces from the objFunc class
  float evaluate(const gVecComponent& model);
  virtual void  calcSearchDir( gVecComponent& searchDir, const gVecComponent& curModel); //means gradient*(-1)
  virtual void  callBackAtIter(int iIter); //call back for iterative solver
  
private:
  void calcSlowGradient(cubeData<float>& velGradCube);
  
protected:
  virtual void flipGradient(cubeData<float>& gradCube);
  void gen_h_term_cube(cubeData<float>& hTermCube, const hypercube& shapeOffImage, int h_order);

  float offDSO_ObjFunc();
  float stackPowerObjFunc();
  float NormStackPowerObjFunc();
  float cheatTrueModObjFunc();
  
  void spreadWtFromCoarseToFineGrid(cubeDataF& wtCube,const cubeDataF& coarseCube);
  void findSmoothDJDUfromSemb(cubeDataF& dJdu, cubeDataF& sembStack, int iSembAxis);
  void convertAngleOffImage(cubeData<float>& img_ax_t, cubeData<float>& img_hx, bool bOff2Ang);
  
  virtual void onImageUpdated(); //if the migrated image is just updated, set some flags;
  
protected:
  WEMVA_Param *m_pWEMVAparams;
  
  //intermediate memory objects;
  cubeData<float> mImg0_hx_t;
  cubeData<float> m_bpjImg_ax_t;

protected:
  bool m_bCalcAngleGather;
  bool m_bAngleGatherCalculated;
};

// Destructor implementation
inline WEMVA_FD::~WEMVA_FD()
{
}

//offset domain implementation
class WEMVA_FD_DSO_off : public WEMVA_FD {
public:
  WEMVA_FD_DSO_off(velAnaParam* pVelAnaParam, const shared_ptr<waveOps>& pModel, WEMVA_Param * pPar): WEMVA_FD(pVelAnaParam, pModel, pPar) 
  {m_bCalcAngleGather=false;}
   
  virtual void  calcBackProjData();
protected:
	virtual void flipGradient(cubeData<float>& gradCube);
};

class WEMVA_max0offset: public WEMVA_FD{
public:
  WEMVA_max0offset(velAnaParam* pVelAnaParam,
       const shared_ptr<waveOps>& pModel, WEMVA_Param * pPar): 
  WEMVA_FD(pVelAnaParam,pModel,pPar) {}
  
  virtual void calcBackProjData();
  //calculate objective function, used in solver
};

class WEMVA_FD_eventBased : public WEMVA_FD {
public:
  WEMVA_FD_eventBased(velAnaParam* pVelAnaParam,
                   const shared_ptr<waveOps>& pModel, WEMVA_Param * pPar):
  WEMVA_FD(pVelAnaParam,pModel,pPar) {}
   
  virtual void calcBackProjData();
  
private:
  virtual void calcBackProjData2d() {}
  virtual void calcBackProjData3d() {}
  virtual float calcObjFunc() = 0;
protected:
};

//RMO style, tan(\gamma)^2 moveout function or (two params (plus a shift) for each image pt.
class WEMVA_FD_beamFocus : public WEMVA_FD_eventBased {
public:
   WEMVA_FD_beamFocus(velAnaParam* pVelAnaParam,
			  const shared_ptr<waveOps>& pModel, WEMVA_Param * pPar): 
	 WEMVA_FD_eventBased(pVelAnaParam,pModel,pPar) {}
   
   virtual float calcObjFunc();
   //void calcBackProjData();

   virtual void getfrCube(cubeData<float>& frCube);
   
private:
   virtual void calcDJDU2d();
   virtual void calcDJDU3d();
   
   virtual void calcBackProjData2d();
   virtual void calcBackProjData3d();
protected:
  void calcDJDU();
  virtual void flipGradient(cubeData<float>& gradCube);
  void calcBackProjData3d_mJacob();
  void calcBackProjData3d_2Jacob();
};

//Use curvature for RMO, but the deltaS is linked to the the shifts of individual traces
class WEMVA_FD_beamFocus_m: public WEMVA_FD_beamFocus {
public:
   WEMVA_FD_beamFocus_m(velAnaParam* pVelAnaParam,
			  const shared_ptr<waveOps>& pModel, WEMVA_Param * pPar): 
	 WEMVA_FD_beamFocus(pVelAnaParam,pModel,pPar) {}
   
   //The only difference to beamFocus is how to calculate du/ds
   virtual void calcBackProjData2d();
protected:
	 virtual void flipGradient(cubeData<float>& gradCube) {}
};

} //end namespace
#endif // __WEMVA_H_INCL__

