#include "stdafx.h"
#include "omp.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "WEMVA.h"
#include "waveOps.h"
#include "dampingFunc.h"
#include "misc.h"
#include "cubeAlgo2.h"
#include "point2d.hpp"
#include "commonOps.h"

#include "off2ang3d.h"
#include "LUsolver.h"
//#include <mkl.h>
using namespace Yang;

void WEMVA_FD_beamFocus::calcDJDU3d() {
  FUNC_TRACE_BEG(WEMVA_FD_beamFocus::calcDJDU3d);
  assert(m_bCalcAngleGather);
  
  velAnaParam* pVelAnaParam 	=	velAnaPar();
  //The computed angle gather image should be ready. (z,gamma,azim,mx,my)
  dmRsf_oc<float> dm_img_ang(m_pWEMVAparams->fn_img_ang,    dm::IN    ,"img_ang",true);
  dm_img_ang.init(NULL,5);  //should have 5 axes
  hypercube shape_img_ang = dm_img_ang.domain(); //should be of (z,gamma,azim,mx,my). we take every 3D slice of (z,gamma,azim)
  
  axis ax_azim = shape_img_ang.get_axis(3);
  axis ax_x = shape_img_ang.get_axis(4); axis ax_y = shape_img_ang.get_axis(5);
  axis ax_z = shape_img_ang.get_axis(1);
  axis axis_angle = shape_img_ang.get_axis(2);
  axis ax_beam_z; float hfBeam_z_window=0.0f;
  hypercube shapeSgGather = shape_img_ang.return_compatible(1,4);
  //convert the angle axis for deg to radian.
  Yang::tagMoveOutType moType = Yang::tangentSquare; //not quadratic in this case.
  if (moType == Yang::tangentSquare) { //need convert deg to radian.
    VERIFY(axis_angle.d > 0.5f); //this is likely to be in degree
    deg2radianAxis(axis_angle,axis_angle,true);
    shapeSgGather.set_axis(2,axis_angle);
  }
  
  cubeData<float>   img_ax_t(shapeSgGather);              //(z,gamma,azim,x)
  cubeData<float>&  dJduCube = pVelAnaParam->dJdbCube;  //the dJ/du term
  
  hypercube shapeDJDU, shapeDJDU_oc;

  if (m_pWEMVAparams->bSingleEvent) {
    shapeDJDU = shape_img_ang.return_compatible(3,4); //one for every azi and x
    dJduCube.alloc(shapeDJDU); //one for each x
    shapeDJDU_oc = shape_img_ang.return_compatible(3,5);
  } else {
    ax_beam_z       = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
    hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
    vecAxis axes; axes.push_back(ax_beam_z); axes.push_back(ax_azim); axes.push_back(ax_x);
    shapeDJDU.set_axes(axes);
    dJduCube.alloc(shapeDJDU); //one for each (x_b,azim,z_b);
    axes.push_back(ax_y);    
    shapeDJDU_oc.set_axes(axes);
  }
  
  shapeDJDU_oc.info("shapeDJDU_oc");
  PRV(m_pWEMVAparams->fn_dJdu);
  
  dmRsf_oc<float> dm_dJdu(m_pWEMVAparams->fn_dJdu,dm::OUT,"djdu",true);
  dm_dJdu.init(&shapeDJDU_oc);
  
  if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
    if (m_pWEMVAparams->bSingleEvent) {
      //Put 6 spikes spike for sensitivity kernel
      /**
      int dnw = axis_x.n/8;
      for (int i=0; i<7; i++) {
        dJduCube(axis_x.n/2+(i-3)*dnw) = 1.0f;
      }
      **/
      for (int iy=0; iy<ax_y.n; iy++) {
        if (iy%(ax_y.n/6)==0) {
          dJduCube(ax_x.n/2,0) = 1.0f;
        }
        dm_dJdu.seq_update(&dJduCube);
      }
    } else { //Do multiple events. Now the DJDU term is of (z_b,azim,x_b,y_b), this grid might be coarser than the image grid.      
      int dnw   = ax_x.n/4;
      int dnwb  = ax_beam_z.n/6;
      //Put a series of spikes in the center x for sensitivity kernel
      for (int iy=0; iy<ax_y.n; iy++) {
        if (iy%(ax_y.n/6)==0) {
          for (int i=0; i<3; i++) {
            for (int izb=dnwb/2; izb<ax_beam_z.n; izb+=dnwb) {
              dJduCube(ax_x.n/2+(i-1)*dnw,0,izb) = 1.0f;
            }
          }
        }
        dm_dJdu.seq_update(&dJduCube);
      }
    } //end if single events
  } else { //if not onlycalcsensKernel
    VERIFY(pVelAnaParam->bUseSemblanceToCalcDJdb); //the only option is to use semblance.
  }
  
  //If use semblance panel to determine the moveout parameters
  if (!m_pWEMVAparams->bOnlyCalcSensitivityKernel && pVelAnaParam->bUseSemblanceToCalcDJdb)   {
    LOG('create semblance panel');
    //set the semblance panel size.
    cubeDataF& semb = pVelAnaParam->semblanceCube;
    cubeDataF envSemb;
    vecAxis axes= img_ax_t.domain().get_axes();
    axis ax_mu  = pVelAnaParam->ax_semb;
    axes[1] = ax_mu; //switch the second axis from angle to rho
    semb.alloc(hypercube(axes)); //semb is of (z,rho,azim,x)
    semb.allocMemoryTo(envSemb);
    float fAntiAlias=1.0f;
    
    semb.domain().info("semb");
    
    hypercube shapeSembZstack;
    if (m_pWEMVAparams->bSingleEvent) {
      shapeSembZstack = envSemb.domain().return_compatible(2,4); //(rho,azim,x)
    } else {
      vecAxis axes2;
      axes2.resize(4); axes2[0]=ax_mu; axes2[1]=ax_beam_z; axes2[2]=ax_azim; axes2[3]=ax_x;
      shapeSembZstack.set_axes(axes2);  //(bz,mu,azim,x)  
    }
    cubeDataF sembZstack; sembZstack.alloc(shapeSembZstack);
    
    int iy;
    dmRsf_oc<float> dm_envSemb(m_pWEMVAparams->fn_envSemb,dm::OUT,"semb",true),dm_img_ang(m_pWEMVAparams->fn_img_ang,dm::IN,"img_ang",true);
    hypercube shape_semb_oc; axes.push_back(ax_y); shape_semb_oc.set_axes(axes);
    
    dm_envSemb.init(&shape_semb_oc); dm_img_ang.init(NULL,5);
    if (m_pWEMVAparams->fn_envSemb.empty()) {
      //get a random name for fn_envSemb.
      std::ostringstream oss; oss<<"envSemb"<<rand()%65536; m_pWEMVAparams->fn_envSemb = oss.str();
    }
    //boost::shared_ptr<dmRsf_oc<float> > pdm_envSemb;
    bool bNormalizeSemb   = pVelAnaParam->bNormalizeSemb; //Divide by the sum of squares over angles, not square of sums over angles.
    cubeDataF img_ax_t_sq, semb2; //would be allocated later by copyTo function.
    
    bool bPickMoveoutPar  = pVelAnaParam->bPickMoveoutPar; 
    PRV(bPickMoveoutPar);
    
    //out of core, once for each crossline pt
    for (iy=0;iy<ax_y.n;iy++) {
      dm_img_ang.seq_grab(&img_ax_t); //read in the ang image, this should be parallelized with MPI
      
      if (iy%10==0) {prt("Starting semblance calc iy=%d \n,",iy);}
      
      //First Calculate the semblance panel, based on the formula given in SimonLuo's abstract, basically stack over parabolas then smooth along t.
      calcMoveoutSemblance_omp(semb, img_ax_t, moType,fAntiAlias, true); //quadratic or tan_quadratic ???!!! Quadratic
      //calcMoveoutSemblance(semb, img_ax_t, Yang::Quadratic,fAntiAlias, true); 
      //seplib::callSepTool(envSemb,semb,"Envelope","",true);
      
      //instead, just square the semblance
      envSemb.zero(); envSemb.multiply(semb,semb);
      seplib::callSepTool(envSemb,envSemb,"Smooth"," repeat=2 rect1=3 rect2=1 ",false); 
      
      //envSemb.dump("dump7");
      if (bNormalizeSemb) {
        //LOG(calc the normalized semblance);
        //might need to pull it out of the loop
        //Calculate the denominator
        img_ax_t.copyTo(img_ax_t_sq); img_ax_t_sq.multiply(img_ax_t);
        semb.copyTo(semb2);
        
        //This way to calc denom ??? is it the best way ???????
        //Quadratic or tanQuadratic ??? !!!
        calcMoveoutSemblance_omp(semb2, img_ax_t_sq, moType, fAntiAlias, true);
        //seplib::callSepTool(semb2,semb2,"Envelope","",true); //since we applied envelope above
        //Smooth the panel and add epsilon
        seplib::callSepTool(semb2,semb2,"Smooth"," repeat=2 rect1=3 rect2=1 ",false);
        float fRMS = semb2.RMS_NZ(1e-4); //PRV(fRMS);
        semb2.add(fRMS*0.05+1e-3); //in case it is absolute zero
        //envSemb.dump("dump9");  img_ax_t_sq.dump("dump10");
        envSemb.divide(semb2); envSemb.scale(1.0f/(pVelAnaParam->ax_semb.n));
      }
            
      //Then calc the derivative, and put it to dJ over dU cube
      if (m_pWEMVAparams->bSingleEvent) { //if single events
        //First stack the semblance along the z window, the shape is (mu,azim,x)
        sembZstack.zero();
        sumAlongTimeTrc(sembZstack, envSemb);
        if (!bPickMoveoutPar) { //Take derivative
          this->findSmoothDJDUfromSemb(dJduCube,sembZstack,1);        
        } else {
          findMaxValTimePerTrc(dJduCube,sembZstack);
        }
      } else { //multiple events.
        int izwbeg,izwend;
        //Here the way I implement is summing the values in the smoothed envelope as in a beam window and then take derivatives or find maximum values
        //determine the stacked cube shape the shape is (mu,z_b,azim,x)
        sembZstack.zero();
        int nz = ax_z.n;
        for (int izb=0; izb<ax_beam_z.n; izb++) {
          izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
          izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
          assert (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1);
          //Sum along the small window along z
          for (int ix=0; ix<ax_x.n; ix++) {
          for (int iazim=0; iazim<ax_azim.n; iazim++) {
          for (int imu=0; imu<ax_mu.n; imu++) {
              for (int iiz=izwbeg; iiz<=izwend; iiz++) {
                sembZstack(ix,iazim,izb,imu) += envSemb(ix,iazim,imu,iiz);
              }
          }}}
        }
        //sembZstack.dump("dump5");
        if (!bPickMoveoutPar) { //Take derivative
          this->findSmoothDJDUfromSemb(dJduCube,sembZstack,1);
        } else {
          findMaxValTimePerTrc(dJduCube,sembZstack);
        }
      }
      
      //IO
      dm_envSemb.domain().info("dm_envSemb"); envSemb.domain().info("envSemb");
      dm_dJdu.domain().info("dm_dJdu"); dJduCube.domain().info("dJduCube");
      dm_envSemb.seq_update(&envSemb);
      dm_dJdu.seq_update(&dJduCube);
    } //for iy
  } //end if use semblance
  LOGC('finish calc DJDU 3d');
  FUNC_TRACE_END();
}


void WEMVA_FD_beamFocus::calcBackProjData3d() {
   WEMVA_Param::dRho_dsOpType opType = m_pWEMVAparams->m_dRho_dsOpType;
  if (opType==WEMVA_Param::drds_Jacob_m) {
    this->calcBackProjData3d_mJacob();
  } else if (opType==WEMVA_Param::drds_Jacob_2) {
    this->calcBackProjData3d_2Jacob();
  } else {
    VERIFY(0);
  }
}

// calc du/ds (u is the curvature) by inverting the 2X2 hessian matrix with the curvature and static shift moveout parameters (the direct operator).
void WEMVA_FD_beamFocus::calcBackProjData3d_2Jacob() {
  FUNC_TRACE_BEG(WEMVA_FD_beamFocus::calcBackProjData3d_2Jacob);
  
  //The resulting angle image is already calculated already.
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  
  //We should have the angle image in the par->fn_img_ang file.
  dmRsf_oc<float> dm_img_ang(m_pWEMVAparams->fn_img_ang,    dm::IN    ,"img_ang",true);
  dmRsf_oc<float> dm_pbjImg_ang(m_pWEMVAparams->fn_bpjImg_ang, dm::OUT,"bpjimg_ang",true);
  dm_img_ang.init(NULL); 
  hypercube shape_img_ang = dm_img_ang.domain(); //should be of (z,gamma,azim,x,y). we take every 3D slice of (z,gamma,azim) and then update the bpjimg.
  dm_pbjImg_ang.init(&shape_img_ang);
  axis ax_azim = shape_img_ang.get_axis(3);
  
  hypercube shapeSgGather = shape_img_ang.return_compatible(1,4);
  axis ax_x = shape_img_ang.get_axis(4); axis ax_y = shape_img_ang.get_axis(5);
  axis ax_z = shape_img_ang.get_axis(1); int nz=ax_z.n;
  axis axis_angle = shape_img_ang.get_axis(2);
  //convert angle axis from deg to radian.
  VERIFY (axis_angle.d > 0.5f); //is degree
  deg2radianAxis(axis_angle,axis_angle,true);

  cubeData<float> img_ax_t(shapeSgGather); //(z,gamma,azim,x)
  LOGC("jj1");
  //Calculate the dJ/du term first, then the DJDU cube should be in some files as well
  this->calcDJDU3d();
  
  LOGC("hh2");
  dmRsf_oc<float> dm_dJdu(m_pWEMVAparams->fn_dJdu, dm::IN,"dJdu",true);
  dm_dJdu.init(NULL);
  
  LOGC("hh3");
  hypercube frShape(vecAxis(1,axis_angle));  cubeData<float> frCube,grCube;
  frCube.alloc(frShape); grCube.alloc(frShape); //grCube is the constant shift amount.
  cubeData<float> fr2cube(frShape),gr2cube(frShape),fgrcube(frShape);
  this->getfrCube(frCube); //get fr
  velAna_FD::setMoveoutDeriv(grCube,Yang::constant);
  fr2cube.multiply(frCube,frCube); fgrcube.multiply(frCube,grCube); gr2cube.multiply(grCube,grCube);
  
  float vth = pWEMVAparam->fEmaskValThres; VERIFY(vth>1e-6f);
  
  LOGC("hh4");
  int iy; //ix
  if (pWEMVAparam->bSingleEvent) {
    if (pWEMVAparam->bUseSingleMoveoutParam) {
      VERIFY(0);
    } else { //Use two parameter fitting
      verbPrt("Use two moveout parameters");
      //In this case we need to do the inversion one by one (x,y) location.
      //First calculate the E matrix.
      cubeData<float>& EmatCube = pWEMVAparam->Ecube; //the divider
      vecAxis axes = shape_img_ang.return_axes(3,4);
      hypercube shapeEslice(axes); //function of (azimuth,x)
      axis ax_tt(3); axes.push_back(ax_tt); //Three planes for E11,E12 and E22.
      hypercube shapeEcube(axes);
  
      EmatCube.alloc(shapeEcube); //one for each (azim,x)
      cubeData<float> EmatSlice; EmatSlice.becomeProxyCube(shapeEslice);
      cubeData<float> dJduCube(shapeEslice);

      cubeData<float> Fcube; EmatCube.allocMemoryTo(Fcube);
      Fcube.allocMemoryTo(pWEMVAparam->finalWtCube); // (azim,tt)
      
      cubeData<float> Img0_dz(img_ax_t.domain()),Img0_dz2(img_ax_t.domain());              
      stdVecInt iAxes(3,int()); iAxes[0]=1; iAxes[1]=3; iAxes[2]=4;
      hypercube shapeAcube = shape_img_ang.return_arbAxesShape(iAxes); //(z,azim,x)
      cubeData<float> Acube(shapeAcube);

      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      cubeData<float> img_ax_t2(m_bpjImg_ax_t.domain());
      
      for (iy=0;iy<ax_y.n;iy++) {
      //for (ix=0;ix<ax_x.n;ix++) {
        dm_img_ang.seq_grab(&img_ax_t); //fetch a (z,gamma,azim,x) cube
        dm_dJdu.seq_grab(&dJduCube);
        //Get first/2nd derivative, //Take derivative along z direction
        timeDeriv1st(Img0_dz,img_ax_t,true); timeDeriv2nd(Img0_dz2,img_ax_t,true);
        //LOGC(Calc E matrix (denominator));
        Img0_dz2.multiply(img_ax_t);
      
        //calc E11,E12,E22, E cube is of size (azim,x,3);
        windParam windE(3,windowAxis());
        windE[2].n = 1; 

        //LOG(calc E22);
        windE[2].f = 2;
        EmatCube.setProxyXd(&EmatSlice,windE,2);
        Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&gr2cube);
        sumAlongTimeTrc(EmatSlice,Acube);

        //build a mask based on EmatSlice value.
        EmatSlice.allocMemoryTo(pWEMVAparam->thresMaskCube);
        buildThresholdMask(pWEMVAparam->thresMaskCube,EmatSlice,vth,true);
        pWEMVAparam->thresMaskCube.dump("Emask"); PRV(vth); EmatSlice.dump("dump2");

        //LOG(calc E12);
        windE[2].f = 1;
        EmatCube.setProxyXd(&EmatSlice,windE,2);
        Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&fgrcube);
        sumAlongTimeTrc(EmatSlice,Acube);
        //LOG(calc E11);
        windE[2].f = 0;
        EmatCube.setProxyXd(&EmatSlice,windE,2);
        Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&fr2cube); //sum over the gamma axis.
        sumAlongTimeTrc(EmatSlice,Acube); //single event, sum along z.

        //EmatCube.dump("Ecube");frCube.dump("dump1"); 
        //LOG(invert E mat);
        //Now do the back projection part
        point2dF pt2d; float ptmod;
        VERIFY(!pWEMVAparam->bBalanceEnumeratorWt);
        for (int ix = 0; ix<ax_x.n; ix++) {
        for (int iazim = 0; iazim<ax_azim.n; iazim++) {
          if (pWEMVAparam->thresMaskCube(ix,iazim)<0.5f) {
            continue;
          } else {
            //First solve the F matrix.
            float tmpF[4]; float tmpE[4]; float determ, cond;
            tmpE[0] = EmatCube(0,ix,iazim); tmpE[1]=tmpE[2]=EmatCube(1,ix,iazim); tmpE[3] = EmatCube(2,ix,iazim);
            cond = invert2X2(tmpF,tmpE,determ);
            if (cond > 1e3) {
              verbPrt("\n cond number is very big %f in beamFocus2_3d!! E=%.3e,%.3e,%.3e,%.3e", 
                      cond, tmpE[0], tmpE[1], tmpE[1], tmpE[3]);
            }
            pt2d.x = tmpF[0]; pt2d.y = tmpF[1];
            ptmod = 1.0f;

            if (pWEMVAparam->bBalanceEnumeratorWt) {
              ptmod = 1.0f/pt2d.module();
              pt2d *= ptmod;
            }
            Fcube(0,ix,iazim) = pt2d.x; Fcube(1,ix,iazim) = pt2d.y; Fcube(2,ix,iazim) = tmpF[3]*ptmod; //will not use the 3rd element.
          }
        } }
        pWEMVAparam->finalWtCube.zero();
        multiplySlice(pWEMVAparam->finalWtCube,Fcube,dJduCube);
      
        Fcube.dump("Fcube"); pWEMVAparam->finalWtCube.dump("finalWt");
        pWEMVAparam->finalWtCube.normalize();
        
        //LOG(assemble the bpj angle image gather);
        //Take F11 slice
        windE[2].f = 0; pWEMVAparam->finalWtCube.setProxyXd(&EmatSlice,windE,2);
        m_bpjImg_ax_t.zero();
        multiplySubCube(m_bpjImg_ax_t,Img0_dz,frCube,2,2);
        scaleSlice(m_bpjImg_ax_t,m_bpjImg_ax_t,EmatSlice); //(z,gamma,azim,x)
        //Take F12 slice
        windE[2].f = 1; pWEMVAparam->finalWtCube.setProxyXd(&EmatSlice,windE,2);
        multiplySubCube(img_ax_t2,Img0_dz,grCube,2,2);
        scaleSlice(img_ax_t2,img_ax_t2,EmatSlice);
        m_bpjImg_ax_t.add(img_ax_t2);
        
        dm_pbjImg_ang.seq_update(&m_bpjImg_ax_t);
      } //end for iy
    }
  } else { //If Multiple events
    axis ax_beam_z = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
    float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
    if (pWEMVAparam->bUseSingleMoveoutParam) {
      VERIFY(0);
    } else { //Use two parameter fitting
      //verbPrt("\nUse two moveout parameters");
      
      //In this case we need to do the inversion one by one on (z_b,azim,x) location.
      //First calculate the E matrix.
      cubeData<float>& EmatCube = pWEMVAparam->Ecube; //the divider
      vecAxis axes; axes.push_back(ax_beam_z);axes.push_back(ax_azim);axes.push_back(ax_x);
      hypercube shapeEslice(axes);
      axis ax_tt(3); axes.push_back(ax_tt); //Three planes for E11,E12 and E22.
      hypercube shapeEcube(axes);

      EmatCube.alloc(shapeEcube);
      cubeData<float> EmatSlice; EmatSlice.becomeProxyCube(shapeEslice);
      cubeData<float> dJduCube(shapeEslice);
      
      //Get first/2nd derivative
      cubeData<float> Img0_dz(img_ax_t.domain()),Img0_dz2(img_ax_t.domain());

      stdVecInt iAxes(4,int()); iAxes[0]=1; iAxes[1]=3; iAxes[2]=4;
      hypercube shapeAcube = shape_img_ang.return_arbAxesShape(iAxes); //(z,azim,x)
      cubeData<float> Acube(shapeAcube),Bcube(shapeAcube),Ccube(shapeAcube);
      
      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      cubeDataF scaleCube; //size of (z,ax,azim,x), a multiplier at each image location
      img_ax_t.allocMemoryTo(scaleCube);
      
      for (iy=0;iy<ax_y.n;iy++) {
        dm_img_ang.seq_grab(&img_ax_t); //fetch a (z,gamma,azim,x) cube
        dm_dJdu.seq_grab(&dJduCube);
        //Take derivative along z direction
        timeDeriv1st(Img0_dz,img_ax_t,true);timeDeriv2nd(Img0_dz2,img_ax_t,true);

        LOGC("Calc E matrix (denominator) in multiple events case");
        Img0_dz2.multiply(img_ax_t);
        
        //stack over angles.
        Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&fr2cube);
        Bcube.zero(); multAndStackAlongAxis2(Bcube,Img0_dz2,&fgrcube);
        Ccube.zero(); multAndStackAlongAxis2(Ccube,Img0_dz2,&gr2cube);
              
        //LOG(calc E matrix);
        int izwbeg,izwend;
        for (int ix=0; ix<ax_x.n; ix++) {
        for (int iazim=0;iazim<ax_azim.n;iazim++ ) {
          for (int izb=0; izb<ax_beam_z.n; izb++) {
            izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
            izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
            if (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1) {
              //Sum along the small window along z
              for (int iiz=izwbeg; iiz<=izwend; iiz++) {
                EmatCube(0,ix,iazim,izb) +=  Acube(ix,iazim,iiz);
                EmatCube(1,ix,iazim,izb) +=  Bcube(ix,iazim,iiz);
                EmatCube(2,ix,iazim,izb) +=  Ccube(ix,iazim,iiz);
              }
            }
          }
        } }
        //take E11 for mask building.
        windParam windE(4,windowAxis());
        windE[3].n = 1; windE[3].f = 0;
        EmatCube.setProxyXd(&EmatSlice,windE,3);
        
        //build a mask based on EmatSlice value, !!Maybe we should use different threshold value for different the events.
        
        //PRV(pWEMVAparam->thresMaxValPercent);
        EmatSlice.allocMemoryTo(pWEMVAparam->thresMaskCube);
        buildThresholdMask(pWEMVAparam->thresMaskCube,EmatSlice,vth,true);
  
        EmatCube.dump("Ecube");
        VERIFY(!pWEMVAparam->bBalanceEnumeratorWt);
        
        LOG(invert E mat in multiple event case);
        cubeData<float> Fcube; EmatCube.allocMemoryTo(Fcube);
        //Now do the back projection part
        //First solve the F matrix.
        point2dF pt2d; float ptmod;
        float tmpF[4]; float tmpE[4]; float determ, cond;
        for (int ix = 0; ix<ax_x.n; ix++) { 
        for (int iazim = 0; iazim<ax_azim.n; iazim++) { 
          for (int izb=0; izb<ax_beam_z.n; izb++) {
            if (pWEMVAparam->thresMaskCube(ix,iazim,izb)<0.5f) { //don't operate on mask=0
              continue;
            } else {
              tmpE[0] = EmatCube(0,ix,iazim,izb); tmpE[1]=tmpE[2]=EmatCube(1,ix,iazim,izb); tmpE[3] = EmatCube(2,ix,iazim,izb);
              cond = invert2X2(tmpF,tmpE,determ);
              if (cond > 1000.0f) {
                verbPrt("\n cond number is very big %f in beamFocus2!! E=%.3e,%.3e,%.3e,%.3e F=%.3e,%.3e,%.3e", 
                        cond, tmpE[0], tmpE[1], tmpE[2], tmpE[3], tmpF[0],tmpF[1],tmpF[3]);
              }
              pt2d.x = tmpF[0]; pt2d.y = tmpF[1];
              ptmod = 1.0f;
              if (pWEMVAparam->bBalanceEnumeratorWt) {
                ptmod = 1.0f/pt2d.module();
                pt2d *= ptmod;
              }
              //will not use the 3rd element.
              Fcube(0,ix,iazim,izb) = pt2d.x; Fcube(1,ix,iazim,izb) = pt2d.y; Fcube(2,ix,iazim,izb) = tmpF[3]*ptmod; 
            }
          }
        }}
        
        
        //I_z(z,ax,azim,x), need to add two terms., define wt(ax,zb,azim,x);
        cubeData<float> wt2[2];
        axes.resize(4); axes[0] = axis_angle; axes[1] = ax_beam_z; axes[2]=ax_azim; axes[3]=ax_x;
        hypercube wtShape(axes);
        wt2[0].alloc(wtShape); wt2[1].alloc(wtShape);
        
        LOG(comp the F11* + F12* term);
        //(Take F11(z_b,x) * fr(ax) + F12(z_b ,x) * gr(ax)) * Iz(zw+z_b,ax;x,z_b);
        //wt2[0].info("wt2[0]"); EmatSlice.info("EmatSliceinfo");
        windE[3].f = 0;
        Fcube.setProxyXd(&EmatSlice,windE,2);
        duplicateAlongAxis(wt2[0],EmatSlice,1); multiplySlice(wt2[0],wt2[0],frCube);
  
        windE[3].f = 1;
        Fcube.setProxyXd(&EmatSlice,windE,2);
        duplicateAlongAxis(wt2[1],EmatSlice,1); multiplySlice(wt2[1],wt2[1],grCube);
        wt2[0].add(wt2[1]);
        
              
        //Fcube.dump("Fcube"); //wt2[0].dump("dump5"); wt2[1].dump("dump6");
        wt2[0].allocMemoryTo(pWEMVAparam->finalWtCube);
        scaleSlice(pWEMVAparam->finalWtCube,wt2[0],dJduCube);
        pWEMVAparam->thresMaskCube.dump("Emask"); pWEMVAparam->finalWtCube.dump("finalWt");
  
        m_bpjImg_ax_t.zero(); scaleCube.zero();
        spreadWtFromCoarseToFineGrid(scaleCube, pWEMVAparam->finalWtCube);
        //scaleCube.dump("wtAll");
        m_bpjImg_ax_t.multiply(Img0_dz,scaleCube);
        dm_pbjImg_ang.seq_update(&m_bpjImg_ax_t);
        //m_bpjImg_ax_t.dump("angCubeBp");
      } //for iy
    } //end if use single moveout param
  } //end if bSingle event
  
  LOGC("gg1");
  //Then finish all the IO
  dm_dJdu.unInit(); dm_img_ang.unInit();
  dm_pbjImg_ang.unInit(); usleep(100000);
  
  LOGC("gg2");
  //LOG(Convert resulted angle image back to offset image);
  ang2off3d_file(*(pWEMVAparam->par_a2o3d), pWEMVAparam->fn_bpjImg_ang, pWEMVAparam->fn_bpjImg_off);
  LOGC("gg3");
  FUNC_TRACE_END();
}

// calc du/ds (u is the curvature) by inverting the m+1 by m+1 Jacobian hessian matrix with the curvature and static shift moveout parameters (the direct operator).
void WEMVA_FD_beamFocus::calcBackProjData3d_mJacob() {
  FUNC_TRACE_BEG(WEMVA_FD_beamFocus::calcBackProjData3d_mJacob);
  
  //The resulting angle image is already calculated already.
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  
  //We should have the angle image in the par->fn_img_ang file.
  dmRsf_oc<float> dm_img_ang(m_pWEMVAparams->fn_img_ang,    dm::IN    ,"img_ang",true);
  dmRsf_oc<float> dm_pbjImg_ang(m_pWEMVAparams->fn_bpjImg_ang, dm::OUT,"bpjimg_ang",true);
  dm_img_ang.init(NULL,5); 
  hypercube shape_img_ang = dm_img_ang.domain(); //should be of (z,gamma,azim,x,y). we take every 3D slice of (z,gamma,azim) and then update the bpjimg.
  dm_pbjImg_ang.init(&shape_img_ang);
  axis ax_azim = shape_img_ang.get_axis(3);
  
  hypercube shapeSgGather = shape_img_ang.return_compatible(1,4);
  axis ax_x = shape_img_ang.get_axis(4); axis ax_y = shape_img_ang.get_axis(5);
  axis ax_z = shape_img_ang.get_axis(1); const int nz=ax_z.n; const int nx=ax_x.n;
  axis axis_angle = shape_img_ang.get_axis(2);
  //convert angle axis from deg to radian.
  VERIFY (axis_angle.d > 0.5f); //is degree
  deg2radianAxis(axis_angle,axis_angle,true);
  
  cubeData<float> img_ax_t(shapeSgGather); //(z,gamma,azim,x)
//  LOGC("jj1");
  //Calculate the dJ/du term first, then the DJDU cube should be in some files as well
  
  //TEMPORARY !!!!
  bool bDJDU3dCalculated = false;
  if (!bDJDU3dCalculated) {
    this->calcDJDU3d();
  }
  
//  LOGC("hh2");
  dmRsf_oc<float> dm_dJdu(m_pWEMVAparams->fn_dJdu, dm::IN,"dJdu",true);
  dm_dJdu.init(NULL);
  
  PRV4(m_pWEMVAparams->fn_img_ang,m_pWEMVAparams->fn_bpjImg_off,m_pWEMVAparams->fn_dJdu,1);
  
//  LOGC("hh3");
  hypercube frShape(vecAxis(1,axis_angle));  cubeData<float> frCube,grCube;
  frCube.alloc(frShape); grCube.alloc(frShape); //grCube is the constant shift amount.
  cubeData<float> fr2cube(frShape),gr2cube(frShape),fgrcube(frShape);
  this->getfrCube(frCube); //get fr
  velAna_FD::setMoveoutDeriv(grCube,Yang::constant);
  fr2cube.multiply(frCube,frCube); fgrcube.multiply(frCube,grCube); gr2cube.multiply(grCube,grCube);
  float frIntegralScalar = axis_angle.d;
    
  float vth = pWEMVAparam->fEmaskValThres; VERIFY(vth>1e-6f);
  cubeData<float>& Emask = pWEMVAparam->thresMaskCube;
//  LOGC("hh4");
  int mP1 = ax_azim.n + 1; //m+1 
  
  LOGC('Start calculating the bpj ang image in calcBackPrjData3d_mJacob');
  int iy,ix;
  if (pWEMVAparam->bSingleEvent) {
    if (pWEMVAparam->bUseSingleMoveoutParam) {
      VERIFY(0);
    } else { //Use two parameter fitting
      verbPrt("Use two moveout parameters");
      //In this case we need to do the inversion one by one (x,y) location.
      //First calculate the E matrix.
      cubeData<float>& EmatCube = pWEMVAparam->Ecube; //the divider, contain the Ematrix for each x location.
      hypercube shapeJacobMat(mP1,mP1);
      vecAxis axes = shapeJacobMat.get_axes();

      axes.push_back(ax_x); //add x axis
      hypercube shapeEcube(axes);
  
      EmatCube.alloc(shapeEcube); //(m+1 by m+1 by n_x)
      
      cubeData<float> dJduCube(hypercube(ax_azim.n,nx)); // (m by n_x)

      cubeData<float> Img0_dz(img_ax_t.domain()),Img0_dz2(img_ax_t.domain()); //first and 2nd z derivative of the image.        
      cubeData<float> Img0_energy; img_ax_t.allocMemoryTo(Img0_energy);
      //A cube, for calculating the Jacob(E) matrix
      stdVecInt iAxes(3,int()); iAxes[0]=1; iAxes[1]=3; iAxes[2]=4;
      hypercube shapeAcube = shape_img_ang.return_arbAxesShape(iAxes); //(z,azim,x)
      cubeData<float> Acube(shapeAcube);

      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      Emask.alloc(hypercube(ax_azim.n,nx));
      cubeDataF maskEnergy; Emask.allocMemoryTo(maskEnergy);
      PRV(ax_y.n);
      for (iy=0;iy<ax_y.n;iy++) {
        if (iy%10==0) {PRV(iy);}
        
        dm_img_ang.seq_grab(&img_ax_t); //fetch a (z,gamma,azim,x) cube
        dm_dJdu.seq_grab(&dJduCube);
        //IO finished
        
        m_bpjImg_ax_t.zero(); //zero the output
        //Get first/2nd derivative, //Take derivative along z direction
        timeDeriv1st(Img0_dz,img_ax_t,false); timeDeriv2nd(Img0_dz2,img_ax_t,false);
        //LOGC(Calc E matrix (denominator));
        Img0_dz2.multiply(img_ax_t);  //I mult I''.

        //build a mask based on the image energy for each (azim,x), sum over z and gamma
        cubeAbs(img_ax_t, Img0_energy); //take the abs value of the original image as the image.
        Acube.zero(); multAndStackAlongAxis2(Acube,Img0_energy,NULL); //z,gamma,azim,x
        sumAlongTimeTrc(maskEnergy,Acube);

        if (iy==60) {
          maskEnergy.dump("dump7"); 
        }
        buildThresholdMask(Emask,maskEnergy,vth,true);
        if (iy==60) {
          Emask.dump("Emask"); PRV(vth);
        }

        //LOG('ss1');

        
        //calc Ematrix
        EmatCube.zero();
        
        std::valarray<float> sumii_ar(ax_azim.n),sumib_ar(ax_azim.n); float sumbb;
        for (ix=0; ix<nx; ix++) { //Emat are (m+1,m+1) for each x
          sumbb=0.0f;
          for (int iaz=0; iaz<ax_azim.n; iaz++) {
            sumii_ar = 0.0f; sumib_ar = 0.0f; 
            for (int ig=0; ig<axis_angle.n; ig++) {
              for (int iz=0; iz<nz; iz++) {
                sumii_ar[iaz] += Img0_dz2(ix,iaz,ig,iz) * fr2cube(ig);
                sumib_ar[iaz] += Img0_dz2(ix,iaz,ig,iz) * fgrcube(ig);
                sumbb += Img0_dz2(ix,iaz,ig,iz);
              }
            } //end ig gamma
            EmatCube(ix,iaz,iaz) = sumii_ar[iaz];
            EmatCube(ix,iaz,mP1-1) = sumib_ar[iaz];
            EmatCube(ix,mP1-1,iaz) = sumib_ar[iaz];
          }// end iaz azimuth
          EmatCube(ix,mP1-1,mP1-1) = sumbb;
        } //end for x
        
        EmatCube.scale(frIntegralScalar);
        
        if (iy==ax_y.n/2) {
          fr2cube.dump("dump1");
          fgrcube.dump("dump2");
          EmatCube.dump("Emat");
        }
        
        
        //LOG('ss2');
        
        //Now invert the E matrix with the mask, first find how many nonzero elements.
        cubeData<float> EsumMat(hypercube(mP1,mP1));
        std::valarray<float> rhs(mP1);
        stdVecInt       indSel; indSel.reserve(ax_azim.n); //storing the indices of the nonzero location in the mask.
        int cnt = 0, icnt; int ret; float G_i,G_mp1,tan2gm, Gi_tan2g;
        bool bMatrixEmpty = false;
        
        int ind = 0;
        for (ix=0; ix<nx; ix++) {
          bMatrixEmpty = false;
          indSel.clear(); cnt = 0; 
          for (int iaz=0; iaz<ax_azim.n; iaz++) {
            if(Emask(ix,iaz)>0.5f) {
              indSel.push_back(iaz); 
              EsumMat(cnt,cnt) = EmatCube(ix,iaz,iaz);
              cnt++;
            } else {
            }
          }
          if (cnt>0) { bMatrixEmpty = false;}
          if (!bMatrixEmpty) { //invert for the coefficients.
            //first fill in the rest m+1_th row/col elements
            EsumMat = 0.0f; rhs = 0.0f;
            for (icnt=0; icnt < cnt; icnt++) {
              ind = indSel[icnt];
              EsumMat(icnt,icnt) = EmatCube(ix,ind,ind);
              EsumMat(cnt,icnt) = EmatCube(ix,mP1-1,ind);
              EsumMat(icnt,cnt) = EmatCube(ix,mP1-1,ind);
              //assemble the right hand side.
              rhs[icnt] = dJduCube(ix,indSel[icnt]);
            }
            EsumMat(cnt,cnt) = EmatCube(ix,mP1-1,mP1-1);
            PRVEC(EsumMat,(cnt+1)*(cnt+1));
            PRVEC(rhs,cnt+1);
            //Invert The subsetted matrix
            ret = luSolve(EsumMat,cnt+1,va_ptr(rhs));
            VERIFY(ret!=-1);
            PRVEC(rhs,cnt+1);
            //now the G vectors are in rhs vector, calculate the backprojected image here.
            G_mp1 = rhs[cnt];
            for (icnt=0; icnt < cnt; icnt++) { //for azim
              G_i = rhs[icnt];
              for (int ig=0; ig < axis_angle.n; ig++) { //for angle
                tan2gm = frCube[ig]; 
                Gi_tan2g = G_i * tan2gm; Gi_tan2g += G_mp1;
                
                for (int iz=0; iz < nz; iz++) { //for z
                  m_bpjImg_ax_t(ix,indSel[icnt],ig,iz) = Gi_tan2g*Img0_dz(ix,indSel[icnt],ig,iz);          
                  }
              }
            }
          } //end if !bMatrixEmpty
        } //end ix
        
        //LOG('ss3');
        //Sequentially update to file for this iy
        dm_pbjImg_ang.seq_update(&m_bpjImg_ax_t);
      } //end for iy
    }
  } else { //If Multiple events
    axis ax_beam_z = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
    int n_bz = ax_beam_z.n;
    float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
    if (pWEMVAparam->bUseSingleMoveoutParam) {
      VERIFY(0);
    } else { //Use two parameter fitting
      //verbPrt("\nUse two moveout parameters");
      
      //In this case we need to do the inversion one by one on (z_b,x,y) location.     
      cubeData<float>& EmatCube = pWEMVAparam->Ecube;
      hypercube shapeJacobMat(mP1,mP1);
      vecAxis axes = shapeJacobMat.get_axes();

      axes.push_back(ax_beam_z); axes.push_back(ax_x); //add bz, and x axis
      hypercube shapeEcube(axes);
  
      EmatCube.alloc(shapeEcube); //(m+1 by m+1 by beam_z by n_x)
      
      cubeData<float> dJduCube(hypercube(n_bz,ax_azim.n,nx)); // (n_bz by m by n_x)

      cubeData<float> Img0_dz(img_ax_t.domain()),Img0_dz2(img_ax_t.domain()); //first and 2nd z derivative of the image.        
      cubeData<float> Img0_energy; img_ax_t.allocMemoryTo(Img0_energy);

      //A cube, for calculating the Jacob(E) matrix
      stdVecInt iAxes(3,int()); iAxes[0]=1; iAxes[1]=3; iAxes[2]=4;
      hypercube shapeAcube = shape_img_ang.return_arbAxesShape(iAxes); //(z,azim,x)
      cubeData<float> Acube(shapeAcube);
      
      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      Emask.alloc(dJduCube.domain()); //(n_bz by m by n_x)
      
      for (iy=0;iy<ax_y.n;iy++) {
        //start IO, read in the input cubes.
        dm_img_ang.seq_grab(&img_ax_t); //fetch a (z,gamma,azim,x) cube
        dm_dJdu.seq_grab(&dJduCube);
        //IO finished
        m_bpjImg_ax_t.zero(); //zero the output
        //Get first/2nd derivative, //Take derivative along z direction
        timeDeriv1st(Img0_dz,img_ax_t,true); timeDeriv2nd(Img0_dz2,img_ax_t,true);
        //LOGC(Calc E matrix (denominator));
        Img0_dz2.multiply(img_ax_t);  //I mult I''.
        //build a mask based on the image energy for each (azim,x), sum over z and gamma
        cubeAbs(img_ax_t, Img0_energy); //take the abs value of the original image as the image.
        Acube.zero(); multAndStackAlongAxis2(Acube,Img0_energy,NULL); //stack over angles.
        
        //define the beg&end for each z window
        stdVecInt iz_wnd_beg_vec(n_bz),iz_wnd_end_vec(n_bz); //store the iz index for the start of each z_w window.
        int izwbeg,izwend,iiz;
        for (int izb=0; izb<n_bz; izb++) {
          izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
          izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
          VERIFY(izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1);
          iz_wnd_beg_vec[izb] = izwbeg; iz_wnd_end_vec[izb] = izwend;
        }

        //sumAlongTimeTrc(Emask,Acube);
        for (int izb=0; izb<ax_beam_z.n; izb++) { //here I swap the loop order
          izwbeg = iz_wnd_beg_vec[izb];
          izwend = iz_wnd_end_vec[izb];
          for (ix=0;ix<nx;ix++) {
            for (int iaz=0; iaz<ax_azim.n;iaz++) {
              for (iiz=izwbeg; iiz<=izwend; iiz++) {
                Emask(ix,iaz,izb) += Acube(ix,iaz,iiz);
              }
            }
          }
        } // end for izb
        
        if (iy==(ax_y.n/2)) {
          Emask.dump("Energy"); PRV(vth);
        }
        //mark where the image energy is too low
        buildThresholdMask(Emask,Emask,vth,true);
        if (iy==(ax_y.n/2)) {
          Emask.dump("Emask"); PRV(vth);
        }
        
        //Calculate the Jacobian matrices.
        EmatCube.zero();
                
        std::valarray<float> sumii_ar(ax_azim.n),sumib_ar(ax_azim.n); float sumbb;
        for (ix=0; ix<nx; ix++) { //Emat are (m+1,m+1) for each x
          for (int izb =0; izb<n_bz; izb++) {
            izwbeg = iz_wnd_beg_vec[izb]; izwend = iz_wnd_end_vec[izb];
            sumbb=0.0f;
            for (int iaz=0; iaz<ax_azim.n; iaz++) {
              sumii_ar = 0.0f; sumib_ar = 0.0f; 
              for (int ig=0; ig<axis_angle.n; ig++) {
                for (int iz=izwbeg; iz<=izwend; iz++) {
                  sumii_ar[iaz] += Img0_dz2(ix,iaz,ig,iz) * gr2cube(ig);
                  sumib_ar[iaz] += Img0_dz2(ix,iaz,ig,iz) * fgrcube(ig);
                  sumbb += Img0_dz2(ix,iaz,ig,iz);
                }
              } //end ig gamma
              EmatCube(ix,izb,iaz,iaz) = sumii_ar[iaz];
              EmatCube(ix,izb,iaz,mP1-1) = sumib_ar[iaz];
              EmatCube(ix,izb,mP1-1,iaz) = sumib_ar[iaz];
            }// end iaz azimuth
            EmatCube(ix,izb,mP1-1,mP1-1) = sumbb;
          }//end for izb
        } //end for x

        //Now invert the E matrix with the mask, first find how many nonzero elements.
        cubeData<float> EsumMat(hypercube(mP1,mP1));
        std::valarray<float> rhs(mP1);
        stdVecInt       indSel; indSel.reserve(ax_azim.n); //storing the indices of the nonzero location in the mask.
        int cnt = 0, icnt; int ret; float G_i,G_mp1,tan2gm, Gi_tan2g;
        bool bMatrixEmpty = false;
        for (ix=0; ix<nx; ix++) {
        for (int izb =0; izb<n_bz; izb++) {
          //for each (ix,izb) find the unilluminated azimuth
          bMatrixEmpty = false;
          indSel.clear(); cnt = 0; 
          for (int iaz=0; iaz<ax_azim.n; iaz++) {
            if(Emask(ix,izb,iaz)>0.5f) {
              indSel.push_back(iaz); 
              EsumMat(cnt,cnt) = EmatCube(ix,izb,iaz,iaz);
              cnt++;
            } else {
            }
          }
          
          izwbeg = iz_wnd_beg_vec[izb]; izwend = iz_wnd_end_vec[izb];
          if (cnt>0) { bMatrixEmpty = false;}
          if (!bMatrixEmpty) { //invert for the coefficients.
            //first fill in the rest m+1_th row/col elements
            EsumMat = 0.0f; rhs = 0.0f;
            for (icnt=0; icnt < cnt; icnt++) {
              EsumMat(cnt,icnt) = EmatCube(ix,izb,mP1-1,indSel[icnt]);
              EsumMat(icnt,cnt) = EsumMat(cnt,icnt);
              //assemble the right hand side.
              rhs[icnt] = dJduCube(ix,indSel[icnt],izb);
            }
            EsumMat(cnt,cnt) = EmatCube(ix,izb,mP1-1,mP1-1);
            //Invert The subsetted matrix
            ret = luSolve(EsumMat,cnt+1,va_ptr(rhs));
            //now the G vectors are in rhs vector, calculate the backprojected image here.
            G_mp1 = rhs[cnt];
            for (icnt=0; icnt < cnt; icnt++) { //for azim
              G_i = rhs[icnt];
              for (int ig=0; ig < axis_angle.n; ig++) { //for angle
                tan2gm = frCube[ig]; 
                Gi_tan2g = G_i * tan2gm; Gi_tan2g += G_mp1;
                
                for (int iz=izwbeg; iz<=izwend; iz++) { //for z
                  m_bpjImg_ax_t(ix,indSel[icnt],ig,iz) += Gi_tan2g*Img0_dz(ix,indSel[icnt],ig,iz);
                }
              }
            }
          } //end if !bMatrixEmpty
        }} //end izb, ix

        //Sequentially update to file for this iy
        dm_pbjImg_ang.seq_update(&m_bpjImg_ax_t);
      } //end if iy
    } //end if using two parameter fitting
  } //end if not bSingle_event
  
  LOGC("gg1");
  //Then finish all the IO
  dm_dJdu.unInit(); dm_img_ang.unInit();
  dm_pbjImg_ang.unInit(); usleep(100000);
  
  LOGC("gg2");
  //LOG(Contvert resulted angle image back to offset image);
  ang2off3d_file(*(pWEMVAparam->par_a2o3d), pWEMVAparam->fn_bpjImg_ang, pWEMVAparam->fn_bpjImg_off);
  LOGC("gg3");
  FUNC_TRACE_END();
}


