#include "stdafx.h"
#include "cubeData.hpp"
#include "waveOps.h"
#include "WEMVA_RMO_Imit.h"
#include "ADCIG_Recon.h"
#include "point2d.hpp"
#include "util_seplib.h"
#include "misc.h"
#include "param_func.h"

#undef NDEBUG
#include <assert.h>

#define DEBUG
#include "myDebug.h"

using namespace Yang;
using namespace std;
using boost::shared_ptr;

WEMVA_RMO_Imit::WEMVA_RMO_Imit(velAnaParam* pVelAnaParam, const shared_ptr<waveOps>& pWave, WEMVA_Param* pWEMVApar, const shared_ptr<RMO_Measurer>& RMO_measure,const shared_ptr<ADCIG_Recon>& ADCIG_Recon):
    WEMVA_FD_eventBased(pVelAnaParam,pWave,pWEMVApar), RMO_measure_(RMO_measure), ADCIG_Recon_(ADCIG_Recon){
  m_bCalcAngleGather = false; m_bAngleGatherCalculated = false;
  bUpdateRMOneeded_ = true;
  //TRACE(("Yeah,%d\n",4));
}

//Utility function
struct __0off2angType{
  enum _0off2angType{uniformly_spread, regular_o2a};
};

void convert_0offImg_to_adcig_2d(const cubeDataF& img0, cubeDataF& img_ax_t, __0off2angType::_0off2angType tp = __0off2angType::uniformly_spread) {
  //Here we just assume the reflectivity at all angles are the same.
  const hypercube& shp = img_ax_t.domain();
  const hypercube& shp_zx = img0.domain();
  const axis& ax_gamma  = shp.get_axis(2);
  const axis& ax_z      = shp.get_axis(1);
  const axis& ax_x      = shp.get_axis(3);
  VERIFY( ax_x.same_axis(shp_zx.get_axis(2)));
  
  int iz,ix,ig;
  if (tp == __0off2angType::uniformly_spread) {
    for (ig=0;  ig<ax_gamma.n;ig++) {
      for (ix=0;ix<ax_x.n;    ix++) {
      for (iz=0;iz<ax_z.n;    iz++) {
        img_ax_t(ix,ig,iz) = img0(ix,iz);
      }}
    }
  } else if (tp == __0off2angType::regular_o2a) {
    VERIFY(0);
  } else {
    VERIFY(0);
  }
  return;
}

float WEMVA_RMO_Imit::RMO_objFunc() {
  RMO_upToDate(); //make sure that RMO info is up to date.
  
  const cubeDataF& Moveout  = RMO_measure_->RMO();
  const cubeDataF& MO_pts   = RMO_measure_->MoveoutPts();
  const cubeDataF& MO_mask  = RMO_measure_->RMOmask();
  const std::valarray<int>& candPtsSel = RMO_measure_->candPtsSel();
  //Then calculate the object function, for 2D case only, the difference between the stack angle and each reflection angle.
  int nPts  = MO_pts.domain().get_n23();
  int nValidPts = 0;
  const axis& ax_g = Moveout.domain().get_n(1);
  float costFunc = 0.0f;

  for (int ip=0; ip<nPts; ip++) {
    if (candPtsSel[ip]>0) {
      nValidPts++;
      for (int ig=0; ig<ax_g.n; ig++) { //for each angle
        costFunc += SQ(Moveout(ip,ig));
      }
    }
  }
  PRV(nValidPts);
  costFunc /= float(nValidPts); //averaging by the total number of anchor points found.
  return costFunc;
}

//virtual function, compute the objective function
float WEMVA_RMO_Imit::calcObjFunc() { //Temporary, should be minimizing the moveout parameters.
  float costFunc;
  //costFunc = this->RMO_objFunc();
  //costFunc = this->stackPowerObjFunc();
  costFunc = this->WEMVA_FD::calcObjFunc();
  if (equalf(costFunc,-1234.0f)) {
    //need to process objFunc here
    if (m_pWEMVAparams->objFuncType_==WEMVA_Param::RMOobjFunc) {
      costFunc = RMO_objFunc();
    } else {
      VERIFY(0);
    }
  }
  prt("finish calc Obj Func in WEMVA_RMO_Imit, obj = %f\n", costFunc);
  return costFunc;
}

//Make sure the RMO info is up to date
void WEMVA_RMO_Imit::RMO_upToDate() {
  //measure the Residual moveout fist.
  if (bUpdateRMOneeded_) { //if the image is updated, then need to recompute the RMO information, otherwise no need to redo.
    //First ask to reconstruct the ADCIG
    bool b3D = this->waveOp()->acqPars()->b3D();
    VERIFY(!b3D);
    //first transpose the result into hx_t cube
    m_pWEMVAparams->Img0_hx_t.info("img0_hx_t");
    ADCIG_Recon_->inputGatherIs(m_pWEMVAparams->Img0_hx_t,m_pWEMVAparams->sampMask_hx_t.get());
    ADCIG_Recon_->recon_adcig().copyTo(m_pWEMVAparams->Img0_ax_t);
    m_pWEMVAparams->Img0_ax_t.dump("dump1");
    //Now the image has been reconstructed, measure the moveout.
    //LOG("d1");
    RMO_measure_->inputADCIG_is(m_pWEMVAparams->Img0_ax_t,m_pWEMVAparams->Img0);
  }
  bUpdateRMOneeded_=false; //mark it.
}


void WEMVA_RMO_Imit::calcSlowGradient(cubeData<float>& slowGradCube) {
  //the back projected dimg is in m_pWEMVAparams->backProjImg_hx;
  FUNC_TRACE_BEG(WEMVA_RMO_Imit::calcSlowGradient);
  WEMVA_Param* pWEMVAparam  = m_pWEMVAparams;
  waveOps*     pModeling   = this->waveOp().get();
  const waveOpParams* TomoParams  = pModeling->waveOpPars().get();

  //Init the operator, doing TomAdj this time
  pModeling->initOperatorEx(TomoParams, waveOps::acTomImit);

  //Verify the size of velGradCube
  slowGradCube.zero();
  //doing TomAdj this time to calculate gradient
  //The input is in &(pWEMVAparam->backProjImg_hx
  //pWEMVAparam->backProjImg_hx.dump("dump5");
  bool b3D = pModeling->acqPars()->b3D();

  pWEMVAparam->Img0.dump("img0.H",true);
  if (!b3D) { //doing 2D case.
    pModeling->TomoImit_op(&(pWEMVAparam->Img0), &(pWEMVAparam->backProjImg_hx) ,&slowGradCube);
  } else { //doing 3D case
    //pModeling->TomAdj3d_op(pWEMVAparam->fn_bpjImg_off   ,slowGradCube);
    VERIFY(0);
  }
  FUNC_TRACE_END();
}

void WEMVA_RMO_Imit::flipGradient(cubeData<float>& gradCube) {
  //gradCube.scale(-1.f);
  return;
}

void WEMVA_RMO_Imit::calcBackProjData2d() {
  FUNC_TRACE_BEG(WEMVA_RMO_Imit::calcBackProjData2d);
  RMO_upToDate();
  
  const hypercube& finalOffImgShp = this->waveOp()->imgPars()->finalSubOffImageShp();
  finalOffImgShp.info("finalOffImgShp");
  
  const cubeDataF& Img0_ax_t= ADCIG_Recon_->recon_adcig();
  const cubeDataF& Moveout  = RMO_measure_->RMO();
  const cubeDataF& RMOmask  = RMO_measure_->RMOmask();
  const cubeDataF& MO_pts   = RMO_measure_->MoveoutPts();
  const valarray<int>& candPtsSelMask = RMO_measure_->candPtsSel();
  const cubeDataF& ptWeights= RMO_measure_->ptWeights();
  //Assume the calculated moveout is organized in this format (z,x) and (ang0,ang1,ang2,..ang{N-1}), in which ang0->angN is defined by the ax_gamma axis in the par_off2ang3d format, and this cube is organized as (N) by M 2D cube, MO_pts contains the coordinates.
  
  LOG("g1");
  //Use the RMO info to modulate the ADCIG.
  const hypercube& final_adcig_shp = Img0_ax_t.domain();
  PRT_LOC; final_adcig_shp.info("final_adcig_shp");
  //final_adcig_shp could come from other sources.
  const axis& ax_z     = final_adcig_shp.get_axis(1);
  const axis& ax_gamma = final_adcig_shp.get_axis(2);
  const axis& ax_x     = final_adcig_shp.get_axis(3);
  
  PRV2(this->RMO_measure_,RMO_measure_->RMO_Params());
  
  int n_hf_zw = int(this->RMO_measure_->RMO_Params()->fEventWndLength/ax_z.d+0.5f)/2;
  PRV(n_hf_zw); VERIFY(n_hf_zw>3);
  
  LOG("g2");
  VERIFY(ax_gamma.n == (Moveout.domain().get_n(1)));
  //build the angle gather to be modulated first
  cubeDataF img_ax_t2; img_ax_t2.alloc(final_adcig_shp);
  cubeDataF img0_z1d; m_pWEMVAparams->Img0.copyTo(img0_z1d);

  //Do an AGC on the zero offset image.
  const AGC_Param& agcPar = *(m_pWEMVAparams->imgBalAgcPar_);
  SEEV2(agcPar.nw1_,agcPar.nw2_);
  NnegInt divideTimes(1); //use 2 if wants to boost further up the weak reflectors
  seplib::AGC(m_pWEMVAparams->Img0,m_pWEMVAparams->Img0_AGC, &agcPar, divideTimes);
  m_pWEMVAparams->Img0_AGC.dump("imgAgc.H",true);

  timeDeriv1st(img0_z1d,m_pWEMVAparams->Img0_AGC,false); //Take first derivative along z;
  convert_0offImg_to_adcig_2d(img0_z1d,img_ax_t2);
  //m_pWEMVAparams->Img0.dump("dump1");img_ax_t2.dump("dump2"); Moveout.dump("dump5"); m_pWEMVAparams->Img0_AGC.dump("dump6");
  
  cubeDataF& bpjImg_ax_t = m_pWEMVAparams->backProjImg_ax;
  bpjImg_ax_t.alloc(final_adcig_shp);
  LOG("g3");
  
  bool bWeightUpZ = this->velAnaPar()->bWeightUpZ_;
  
  const Param_RMO_Measurer* RMO_par = RMO_measure_->RMO_Params();
  int ig,ip,ix,iz,izw; int ig0 = ax_gamma.get_offset(0.0f); int ng=ax_gamma.n;
  int np = Moveout.domain().get_n(2); valarray<float> modulateCoef(ng);

  float nearAngThres(20.5f); //smaller than that is called near angle.
  cubeDataF stkWtsRef,stkWtsPerAngle;
  float f1_10thWidthGauss = 2*nearAngThres;
  getGaussfuncWithMaxWidth(PosFloat(0.1f),stkWtsRef,ax_gamma, 2*nearAngThres, 0.0f,0);
  stkWtsRef.dump("gs2.H",true); stkWtsRef.copyTo(stkWtsPerAngle);

  for (ip=0;ip<np;ip++) {//for each point where RMO is measured
    if (candPtsSelMask[ip]<1) continue; //skip if such point is not measured.
    iz = ax_z.get_offset(MO_pts(ip,0)); ix = ax_x.get_offset(MO_pts(ip,1));
    assert(iz+n_hf_zw < ax_z.n);  assert(iz-n_hf_zw >= 0);  //otherwise out of range
    
    modulateCoef = 0.0f;
    int nValidAngles=0;
    //Use Gaussian weights for diffNearAngStack methods.
    stkWtsPerAngle = stkWtsRef;

    float fSumNearAng = 0.0f, fTotalNearAngWt=0.0f; int nNearAngles = 0;
    Param_RMO_Measurer::bpRMOdiffMethodType method = RMO_par->bpRMOdiffMethod_;
    for (ig=0;ig<ax_gamma.n;ig++) {
      if (RMOmask(ip,ig)>0.5f) {
        nValidAngles++;
        if (method == Param_RMO_Measurer::subtractAngStk) {
          modulateCoef[ig] = Moveout(ip,ig);
        } else if (method == Param_RMO_Measurer::subtract0ang) {
          modulateCoef[ig] = Moveout(ip,ig)-Moveout(ip,ig0);
        } else if (method == Param_RMO_Measurer::diffNeighbor) {
          modulateCoef[ig] = Moveout(ip,ig); VERIFY(0);
        } else if (method == Param_RMO_Measurer::diff0ang) {
          modulateCoef[ig] = Moveout(ip,ig)-Moveout(ip,ig0);
        } else if (method == Param_RMO_Measurer::diffAngStack || Param_RMO_Measurer::diffFullVaraince) {
          modulateCoef[ig] = Moveout(ip,ig);
        } else if (method == Param_RMO_Measurer::diffNearAngStack) {
          modulateCoef[ig] = Moveout(ip,ig);
          if (fabs(ax_gamma.get_coord(ig)) < nearAngThres) {
            nNearAngles++; fSumNearAng += Moveout(ip,ig)*stkWtsPerAngle(ig);
            fTotalNearAngWt += stkWtsPerAngle(ig);
          }
        } else {
          VERIFY(0);
        }
      } //end if RMO mask.
    }
    
    //PRVEC(modulateCoef,ng);
    float avgAngStk = modulateCoef.sum() / float(nValidAngles);
    float fsum = 0.0f;
    //modify in some cases.
    if (method == Param_RMO_Measurer::diffNeighbor) {
      //modulateCoef[ig] = Moveout(ip,ig); VERIFY(0);
    } else if (method == Param_RMO_Measurer::diff0ang) {
      if (RMOmask(ip,ig0)<0.5f) { //if 0-angle is invalid
        modulateCoef = 0.0f;
      } else {
        fsum = modulateCoef.sum();
        modulateCoef[ig0] = -fsum; //put negative offset on zero-angle.
      }
    } else if (method == Param_RMO_Measurer::diffAngStack) {
      modulateCoef -= 1.0f*avgAngStk;
      modulateCoef *= (1-1.0f/float(nValidAngles));
    } else if (method == Param_RMO_Measurer::diffFullVaraince) {
      modulateCoef -= 1.0f*avgAngStk;
      modulateCoef *= float(nValidAngles);
    } else if (method == Param_RMO_Measurer::diffNearAngStack) {
      if (nNearAngles <= 0) continue; //move to next candidate point.
      
      float avgNearAngStk = fSumNearAng/fTotalNearAngWt, fSumAngDif = 0.0f;
      for (ig=0; ig<ax_gamma.n; ig++) {
        if (RMOmask(ip,ig) > 0.5f) {
          modulateCoef[ig] -= 1.0f*avgNearAngStk;
          fSumAngDif += modulateCoef[ig];
        }
      }
      fsum = 0.0f;
      for (ig=0; ig<ax_gamma.n; ig++) {
        if (RMOmask(ip,ig)>0.5f && fabs(ax_gamma.get_coord(ig))<nearAngThres) {
          modulateCoef[ig] -= stkWtsPerAngle(ig)/fTotalNearAngWt*fSumAngDif;
        }
      }
    } else {}
    
    //double check the implementation, should sum to zero
    //if (fabsf(modulateCoef.sum())>1e-5*modulateCoef.max()) {dumpStdArray(modulateCoef,"dump2",true); PRVEC(modulateCoef,ng); PRV(modulateCoef.sum()); PRV(modulateCoef.max()); VERIFY(0);}

    if (bWeightUpZ) {modulateCoef = modulateCoef*SQ(float(iz)); }//weight up the deeper reflectors

    if (ptWeights(ip)>0.0f) { //if the measurement is consistent
      modulateCoef *= ptWeights(ip); //multiply by the weights
      //Use the modulating function to modulate all events.
      for (ig=0; ig<ax_gamma.n; ig++) {
        if (RMOmask(ip,ig) > 0.5f) {
          for (izw=-n_hf_zw; izw<=n_hf_zw; izw++) {
            bpjImg_ax_t(ix,ig,izw+iz) += (img_ax_t2(ix,ig,izw+iz)*modulateCoef[ig]);
          }
        }
      }
    } else {} //do nothing
  }//for each ip
  
  //PRV(modulateCoef.sum()); PRVEC(modulateCoef,ng);
  const Param_o2a* par_o2a = this->ADCIG_Recon_->o2a_param();
  //do 1/cos compensation.
  if (par_o2a->bOnewayCompensate_) { //divide the angle gather by cos^2(gamma) to balance the amplitude.
    hypercube shp1(ax_gamma);
    cubeData<float> cosr(shp1);
    float fang,f2;
    VERIFY(ax_gamma.d>0.5f);
    for (int ia=0; ia<ax_gamma.n; ia++) {
      fang = deg2rad(ax_gamma.get_coord(ia));
      f2 = (1.0/cosf(fang));
      cosr(ia) = SQ(f2)*CB(f2); //do less high angle compensation
    }
    //cosr.dump("dump10"); cosr.info("cosr");
    multiplySubCube(bpjImg_ax_t,bpjImg_ax_t,cosr,2,2); //required, in order to boost up the far angles.
  }
  LOG("g4");
  //Now convert the bpjImg_ax_t back to offset gather, !make sure that the angle parameters are correct.
  //notice that we don't need as many subsurf offset as the initial case
  const axis& ax_h = finalOffImgShp.get_axis(3);
  
  int hf_nh_short=gParamReader->get_int("hfnh",5); PRV(hf_nh_short); //!!This parameter should be fetched from outside
  
  axis ax_h_short(2*hf_nh_short+1,-hf_nh_short*ax_h.d,ax_h.d,"hx");
  cubeDataF& img_hx_t = m_pWEMVAparams->backProjImg_hx_t;
  off2Ang2d_new(img_hx_t,bpjImg_ax_t,true, par_o2a->off2angType_,ax_h_short);
  
  LOG("g5");
  //this->convertAngleOffImage(bpjImg_ax_t,m_pWEMVAparams->backProjImg_hx,false);
  seplib::callSepTool(m_pWEMVAparams->backProjImg_hx,img_hx_t,"Transp","plane=23",true);
  LOG("g6");
  bpjImg_ax_t.dump("bpjAngImg.H",true);  m_pWEMVAparams->backProjImg_hx.dump("dump10",true);
  FUNC_TRACE_END();
}

void WEMVA_RMO_Imit::calcBackProjData3d() {
  VERIFY(0);
}

