#include "stdafx.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "WEMVA.h"
#include "waveOps.h"
#include "misc.h"
#include "cubeAlgo2.h"
#include "point2d.hpp"
#include "commonOps.h"

using namespace Yang;

//return the minimum position, run a linear filter (z-z0)^2
int matchFilterScan(stdVecFloat& tempVec, cubeData<float>& trc, int iSearchStart, int iSearchEnd, int halfFilterLen) {
  axis ax = trc.domain().get_axis(1);
  VERIFY(iSearchEnd < ax.n);
  int nS = iSearchEnd - iSearchStart + 1;
  tempVec.resize(nS);
  
  float fsum, ftmp;
  for (int iz=iSearchStart; iz<=iSearchEnd; iz++) {
    fsum = 0.0f;
    for (int j=MAX(iz-halfFilterLen,0);j<MIN(ax.n,iz+halfFilterLen+1); j++) {
      ftmp = (j-iz)*trc[j];
      fsum += (ftmp*ftmp);
    }
    tempVec[iz-iSearchStart] = fsum;
  }
  int ret = 0; float fminval = tempVec[0];

  for (int i=1; i<nS; i++) {
    if (fminval > tempVec[i]) {
      ret = i;
      fminval = tempVec[i];
    }
  }
  //float* pMin = std::find(&tempVec[0], &tempVec[nS], val);
  return (ret+iSearchStart);
}


//Calculate Gradient, This is just for single event
//use the zero-offset image as the reference depth trace (equal to use the angle stack image).
void WEMVA_FD_angleMoveout::calcBackProjData2d() {
  FUNC_TRACE_BEG(WEMVA_FD_angleMoveout::calcBackProjData)
  
  //The Image0 is been calculated already.
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  
  const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
  cubeData<float>& img_ax_t = pWEMVAparam->Img0_ax_t;
  //convert offset image to angle domain, has been calculated in postCalcData
  //convertAngleOffImage(pWEMVAparam->Img0_ax_t, pWEMVAparam->Img0_hx,true);
  
  LOG(WEMVA_FD_angleMoveout:m4);
  //now axis is z,ax,x
  assert(!img_ax_t.domain().isEmpty());
  pWEMVAparam->Img0_ax_t.dump("angCube");
  
  //Calculate the terms
  cubeData<float> Acube, Ccube;
  axis ax_x = shapeOffImage.get_axis(2);
  hypercube shape_ax_gather = img_ax_t.domain().return_compatible(1,2);
  axis ax_ang = shape_ax_gather.get_axis(2);
  axis ax_z   = shape_ax_gather.get_axis(1);
  
  cubeData<float>& bpjImg_ax_t = m_bpjImg_ax_t;
  img_ax_t.allocMemoryTo(bpjImg_ax_t);

  //For single event
  //performe calculation for each x location
  cubeData<float> Icube, IzCube, IzzCube, tIcube,IcubeZsq;
  cubeData<float> gammaCube, trcCube, trcZsqCube;
  hypercube shapeGamma_x = img_ax_t.domain().return_compatible(2,3);
  Icube.becomeProxyCube(shape_ax_gather);
  IzCube.alloc(shape_ax_gather); IzzCube.alloc(shape_ax_gather); IcubeZsq.alloc(shape_ax_gather);
  tIcube.alloc(shape_ax_gather);
  gammaCube.alloc(ax_ang.n);
  
  windParam windPerX, windPerX2; //X2 is only a window in the (ax,x) domain;
  windPerX.resize(3); windPerX2.resize(2);
  windPerX[2].n = 1;  windPerX2[1].n = 1;

  cubeData<float> refZcube;
  LOG(mm42);
  //First calc the Acube, the shift param cube
  if (pWEMVAparam->bOnlyCalcSensitivityKernel) {
    //Put a spike on some angle at the center
    if (pWEMVAparam->bSingleEvent) {
      int ixc = ax_x.n/2;
      //int iang = ax_ang.get_offset(tanf(30.0f/180.0f*PI)); //Get zero degree
      int iang = ax_ang.get_offset(tanf(10.0f/180.0f*PI)); //Get zero degree
      //A cube, one value for each x,gamma(angle) position.
      hypercube shapeGamma_x = img_ax_t.domain().return_compatible(2,3);
      Acube.alloc(shapeGamma_x); Ccube.alloc(shapeGamma_x);
      int dnw = ax_x.n/8;
      for (int i=0; i<7; i++) {
        Acube(ixc+(i-3)*dnw,iang) = 1.0f;
      }
    } else { //multiple event
      VERIFY(0);
    } //end if single events
  } else { //Not calc the sensitivity kernel
     bool bFindShiftUseCorr = m_pWEMVAparams->bFindShiftUserCorr;
     //Get the zero subsurface offset image first
     windParam wind0offset(3);
     axis ax_hx  = pWEMVAparam->Img0_hx.domain().get_axis(3);
     int zeroOffsetIndex = ax_hx.get_offset(0.0f);
     wind0offset[2].n = 1; wind0offset[2].f = zeroOffsetIndex;
     cubeDataF imgAngleStack; imgAngleStack.becomeProxyCube(shapeOffImage.return_compatible(1,2));
     pWEMVAparam->Img0_hx.setProxyXd(&imgAngleStack,wind0offset,2);
     cubeDataF imgAngleFlat, imgAngleFlatSlice;
     if (bFindShiftUseCorr) { //create a flat gather use this trace
       LOG(generate flat angle gather trace using angle stack);
       cubeDataF img_hx0off(shapeOffImage);
       img_hx0off.update2d(&imgAngleStack,wind0offset);
       convertAngleOffImage(imgAngleFlat,img_hx0off, true );
       //img_hx0off.dump("dump5"); imgAngleFlat.dump("dump6");
     }
     
     if (pWEMVAparam->bSingleEvent) {
       //A cube, one value for each x,gamma(angle) position.
       Acube.alloc(shapeGamma_x); 
       hypercube shapeX(vecAxis(1,ax_x)), shapeZ(vecAxis(1,ax_z));
       refZcube.alloc(shapeX);
       //For single event
       //performe calculation for each x location
       imgAngleFlatSlice.becomeProxyCube(shape_ax_gather);
       trcCube.alloc(shapeZ); trcZsqCube.alloc(shapeZ);       
       windParam windTrc; windTrc.resize(2); windTrc[1].n = 1;
       stdVecFloat tempVec; int iSearchStart,iSearchEnd; int halfFilterLen;
       cubeDataF corrCube, maxCorrPosCube;
       maxCorrPosCube.alloc(ax_ang.n);
       for (int ix=0; ix<ax_x.n; ix++) {
         windPerX[2].f = ix; windPerX2[1].f = ix;
         //Take a slice
         img_ax_t.setProxyXd(&Icube,windPerX,2);
         timeDeriv1st(IzCube,Icube,true);
         timeDeriv2nd(IzzCube,Icube,true);
         
         //Find the reference depth on the zero angle. basically the maximum value
         //int zeroAngOffset = ax_ang.get_offset(0.0f);
         ////Take the zero angle trc, should take zero offset
         //Icube.grab1d(&trcCube,windTrc);
         windTrc[1].f = ix; 
         imgAngleStack.grab1d(&trcCube,windTrc);
         
         
         if (!bFindShiftUseCorr) {
           //Find the reference using only the maximum value does not suffice. We need to run a filter through the trace
           //and pick the maximum.
           int izMax = 0;
           iSearchStart = ax_z.get_offset(500.0f);
           iSearchEnd = ax_z.get_offset(1150.0f);
           halfFilterLen = iSearchEnd - iSearchStart;
           izMax = matchFilterScan(tempVec, trcCube, iSearchStart , iSearchEnd, halfFilterLen);
           
           refZcube(ix) = ax_z.get_coord(izMax);
           //Based on reference depth, build trace zw^2
           float zw = 0.0f;
           trcZsqCube.zero();
           for (int iz=MAX(0,izMax-halfFilterLen); iz<MIN(ax_z.n,izMax+halfFilterLen+1); iz++) {
             zw = (iz-izMax) * ax_z.d;
               trcZsqCube[iz] = zw*zw;
           }
           spreadTrc(IcubeZsq,trcZsqCube);
                  
           //calculate A cube, which tells the shift direction
           tIcube.zero(); tIcube.multiply(Icube,IzCube); tIcube.multiply(IcubeZsq);
           sumAlongTimeTrc(gammaCube,tIcube);
           Acube.update1d(&gammaCube,windPerX2);
         } else { //Use the method of correlations to determine the shift.
           //Do correlation with the zero-offset trace, 
           //this is not very accurate since there is a waveform stretch at far angles.
           //what's better is to use this zero-offset trace and generate the full angle gather.
           //Compare with this full angle gather with other terms.
           
           //this is use single trace as referece
           // crossCorr1dGeneral(corrCube,Icube,trcCube,m_pWEMVAparams->iStartLag, m_pWEMVAparams->iEndLag);
  
           //Now we convert the angle stack trace to a flat gather, which will bear the stretched waveform as well
           imgAngleFlat.setProxyXd(&imgAngleFlatSlice,windPerX,2);
           crossCorr1dGeneral(corrCube,imgAngleFlatSlice,Icube,m_pWEMVAparams->iStartLag, m_pWEMVAparams->iEndLag);
           //find the peak, tick out the unreliables
           findMaxValTimePerTrc(maxCorrPosCube,corrCube);
           //Do some qc
           float fmaxval = corrCube.maxvalAbs();
           int off1;
           for (int i=0; i<ax_ang.n;i++) {
             off1 = ax_z.get_offset(maxCorrPosCube(i));
             if (corrCube(i,off1) < fmaxval*1e-2) {
               maxCorrPosCube(i) = 0.0f;
             }
           }
           //Save this peak to A cube.
           Acube.update1d(&maxCorrPosCube,windPerX2);
         } //end if use correlation
       } //end for
       //refZcube.dump("dump5");
     } else { 
       this->calcBackProjDataMultipleEvents(imgAngleFlat);
     } //end if single event
   }//end if calc sens kernel
   
  LOG(mm5);
  //Then determine the backProjImage
  if (pWEMVAparam->bSingleEvent) {
    tIcube.alloc(shape_ax_gather);
    Ccube.alloc(shapeGamma_x);
    for (int ix=0; ix<ax_x.n; ix++) {
      windPerX[2].f = ix; windPerX2[1].f = ix;
      //Take a slice
      img_ax_t.setProxyXd(&Icube,windPerX,2);
      timeDeriv1st(IzCube,Icube,true);
      timeDeriv2nd(IzzCube,Icube,true);

      //Calc C cube, the energy term that will be divided
      tIcube.zero(); tIcube.multiply(Icube,IzzCube);
      gammaCube.zero(); sumAlongTimeTrc(gammaCube,tIcube);
      Ccube.update1d(&gammaCube,windPerX2);
    }
    
    //Since we need to divide Ccube. It is necessary to build a mask based on Ccube's value
    float vThres = Ccube.maxvalAbs() * pWEMVAparam->thresMaxValPercent;
    prt("\n The max absVal of Ccube and the threshold val for beamfocus: %f,%f", Ccube.maxvalAbs(), vThres);
    
    pWEMVAparam->thresMaskCube.alloc(Ccube.domain());
    cubeData<float> ACcube; Acube.allocMemoryTo(ACcube);
    ACcube.divideByMask(Acube,Ccube,pWEMVAparam->thresMaskCube); //!!This division has divide zero risks.
    
    float qt_mask = pWEMVAparam->thresMaxValPercent;
    Ccube.dump("dump10");
    divideWithCaveat(ACcube,Acube,Ccube,qt_mask,qt_mask,
                     m_pVelAnaParam->bDivideEnergyTermUseMask,
                     &(pWEMVAparam->thresMaskCube),true);
    pWEMVAparam->thresMaskCube.dump("Emask");
    
    //Or we add a small number to cCube;
    //float veps = Ccube.RMS_NZ(1e-3) * 1e-2;
    //Ccube.add(veps);
    //PRV(Ccube.maxvalAbs()); PRV(Ccube.minvalAbs());
    Acube.dump("Acube"); Ccube.dump("Bcube"); ACcube.dump("ABcube");
    Acube.dump("dJdu");
    ACcube.normalize();
    
    pWEMVAparam->finalWtCube.alloc(bpjImg_ax_t.domain());
    pWEMVAparam->finalWtCube = 1.0f;
    scaleSlice(pWEMVAparam->finalWtCube,pWEMVAparam->finalWtCube,ACcube);
    timeDeriv1st(bpjImg_ax_t,img_ax_t,true);
    bpjImg_ax_t.multiply(pWEMVAparam->finalWtCube);
  } else { //b Multiple event
  }
  
  LOG(m6);
  
  bpjImg_ax_t.scale(-1.0f); //There is a negative sign in db/dv term.
  bpjImg_ax_t.dump("angCubeBp");
  //convert it back to offset image
  pWEMVAparam->backProjImg_hx.alloc(shapeOffImage);
  this->convertAngleOffImage(bpjImg_ax_t,pWEMVAparam->backProjImg_hx,false);
  FUNC_TRACE_END()
}

void WEMVA_FD_angleMoveout::calcBackProjDataMultipleEvents(cubeData<float>& imgAngleFlat) {
  assert(!m_pWEMVAparams->bSingleEvent);

  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
  cubeData<float>& img_ax_t = pWEMVAparam->Img0_ax_t;
  //convert offset image to angle domain, has been calculated in postCalcData
  //convertAngleOffImage(pWEMVAparam->Img0_ax_t, pWEMVAparam->Img0_hx,true);

  LOG(WEMVA_FD_angleMoveout:calcBackProjDataMultipleEvents:m4);
  
  //now axis is z,ax,x
  assert(!img_ax_t.domain().isEmpty());

  //Calculate the terms
  axis ax_x = shapeOffImage.get_axis(2);
  hypercube shape_ax_gather = img_ax_t.domain().return_compatible(1,2);
  axis ax_ang = shape_ax_gather.get_axis(2);
  axis ax_z = shapeOffImage.get_axis(1); int nz = ax_z.n;
  cubeData<float> imgAngleFlatSlice; imgAngleFlatSlice.becomeProxyCube(shape_ax_gather);
  cubeData<float>& bpjImg_ax_t = m_bpjImg_ax_t;
  img_ax_t.allocMemoryTo(bpjImg_ax_t);
  
  cubeData<float> imgz, imgzz, imgzmzz;
  img_ax_t.allocMemoryTo(imgz); img_ax_t.allocMemoryTo(imgzz); img_ax_t.allocMemoryTo(imgzmzz);
  timeDeriv1st(imgz,img_ax_t,true); timeDeriv2nd(imgzz,img_ax_t,true);
  imgzmzz.multiply(img_ax_t,imgzz);
  
  cubeDataF Icube,IzCube,IzzCube, IzmzzCube;
  Icube.becomeProxyCube(shape_ax_gather); IzCube.becomeProxyCube(shape_ax_gather); IzzCube.becomeProxyCube(shape_ax_gather); IzmzzCube.becomeProxyCube(shape_ax_gather);
  
  windParam windPerX; //X is  a window in the (z,ax,x) domain;
  windPerX.resize(3); windPerX[2].n = 1;

  //Do multiple events, break the correlation into small windows, and (maybe smooth along it to get the derviative)
   axis ax_beam_z = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
   float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
   //int icorbeg = -nzw/2,icorend = nzw/2; //Set the correlation window to be half zw
   axis ax_zw(0,0,0,"zb_window",""); 
   ax_zw.o = -hfBeam_z_window; ax_zw.d = ax_z.d; ax_zw.n = lroundf(2.0*hfBeam_z_window/ax_z.d)+1;
   vecAxis axes; axes.push_back(ax_zw); axes.push_back(ax_ang);
   hypercube shape_dat_zw(axes);
   cubeDataF I_zw,Iz_zw,Izz_zw,I_angleFlat_zw;
   I_zw.alloc(shape_dat_zw); Iz_zw.alloc(shape_dat_zw); Izz_zw.alloc(shape_dat_zw);
   I_angleFlat_zw.alloc(shape_dat_zw);
   
   windParam wind_dat_zw(2); wind_dat_zw[1].n = ax_ang.n;
   wind_dat_zw[0].n = ax_zw.n;
   
   LOG(mm13);
   
   axes.resize(3); axes[0]=ax_ang; axes[1]=ax_beam_z; axes[2]=ax_x;
   hypercube shapebeamVariable(axes);
   //shapebeamVariable.info("shapebeamVariable");
   cubeData<float>  deltaZcube;
   cubeData<float>& Ccube = pWEMVAparam->Ecube;
   //the deltaTau cube, one value for each (x,z_b,gamma(angle)) position.Note the 
   //axis order
   deltaZcube.alloc(shapebeamVariable); Ccube.alloc(shapebeamVariable); 
   
   hypercube shapeZ(vecAxis(1,ax_z));

   cubeDataF corrCube;
   cubeDataF corrCubeTotal; //this cube is of size (corrCubeshape,z_b,x); it stores all
   //the correlation output we computed, find out the size first
   hypercube shapeCorrCube;
   crossCorr1d(corrCube,I_zw,I_angleFlat_zw,m_pWEMVAparams->iStartLag, m_pWEMVAparams->iEndLag,&shapeCorrCube);
   axes = shapeCorrCube.get_axes(); axes.push_back(ax_beam_z); axes.push_back(ax_x);
   hypercube shapeCorrCubeTotal(axes);
   //shapeCorrCubeTotal.info("shapeCorrCubeTotal");
   corrCubeTotal.alloc(shapeCorrCubeTotal);
   windParam windCorrCube(4); windCorrCube[3].n=1;windCorrCube[2].n=1;
   
   LOG(mm23);
   int ix,izb;
   int izwbeg,izwend;
   //performe calculation for each x location
   for (ix=0; ix<ax_x.n; ix++) {
     windPerX[2].f = ix;
     windCorrCube[3].f=ix;
     //Take a slice
     img_ax_t.setProxyXd(&Icube,windPerX,2);
     imgz.setProxyXd(&IzCube,windPerX,2);
     imgzz.setProxyXd(&IzzCube,windPerX,2);
     imgzmzz.setProxyXd(&IzmzzCube,windPerX,2);
     imgAngleFlat.setProxyXd(&imgAngleFlatSlice,windPerX,2);

     //LOG(mm24);
     
     VERIFY(pWEMVAparam->bFindShiftUserCorr); //require using correlation based method.
     for (izb=0; izb<ax_beam_z.n; izb++) {
       //Take the beam window zw from the image gather.
       izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
       izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
       assert(izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1);
       //LOG(mm25);
       wind_dat_zw[0].f = izwbeg;
       Icube.grabWindow(    &I_zw,  wind_dat_zw);
       IzCube.grabWindow(   &Iz_zw, wind_dat_zw);
       IzzCube.grabWindow(  &Izz_zw,wind_dat_zw);
       imgAngleFlatSlice.grabWindow(&I_angleFlat_zw,wind_dat_zw);
       //LOG(mm26);
       
       //Use the method of correlations to determine the shift.
       crossCorr1dGeneral(corrCube,I_angleFlat_zw,I_zw,m_pWEMVAparams->iStartLag, m_pWEMVAparams->iEndLag);
       windCorrCube[2].f=izb;
       corrCubeTotal.update2d(&corrCube,windCorrCube);
       
       //calc the energy term, the C cube
       for (int ia=0; ia<ax_ang.n; ia++) {
         for (int izw=izwbeg; izw<=izwend; izw++) {
           Ccube(ix,izb,ia) += IzmzzCube(ia,izw);
         }
       }
     } //end for izb
   } //end for ix
   
   corrCubeTotal.dump("corr");
   axis ax_lag = corrCubeTotal.domain().get_axis(1);
   if (m_pVelAnaParam->maxCorrUsage == velAnaParam::pickMaxCorr) {
     //find the peak, Save this peak to the deltaZ cube, (ax,z_b,x); the corrCubeTotal to be (lag,ax,z_b,x)
     findMaxValTimePerTrc(deltaZcube,corrCubeTotal);
     //Do some qc on the picked correlation cube.
     float fmaxval = corrCubeTotal.maxvalAbs(); float fTh = fmaxval*1e-3;
     int off1;
     for (ix=0; ix<ax_x.n; ix++) {
       for (izb=0; izb<ax_beam_z.n;izb++) {
         for (int ia=0; ia<ax_ang.n; ia++)  {
           off1 = ax_lag.get_offset(deltaZcube(ix,izb,ia));
           if (corrCubeTotal(ix,izb,ia,off1) < fTh) {
             deltaZcube(ix,izb,ia) = 0.0f;
           }
         }
       }
     }
   } else if (m_pVelAnaParam->maxCorrUsage == velAnaParam::smoothCorrDeriv0) { //Do a smooth on the corrCube then we can just take the derivative at angle=0
     char smoothPar[128]; cubeDataF corrCubeSm,corrCubeEnv;
     cubeDataF& corrCubeDeriv0 = deltaZcube;
     //First need to take envelope
     seplib::callSepTool(corrCubeEnv,corrCubeTotal,"Envelope","",true);
     sprintf(smoothPar,"rect1=%d repeat=%d",ax_zw.n/3,2); //Smooth the 1/4th window size.
     seplib::callSepTool(corrCubeSm,corrCubeEnv,"Smooth",string(smoothPar),true);
     int indexLag = ax_lag.get_offset(0.0f);
     PRV(indexLag); corrCubeEnv.dump("dump11"); corrCubeSm.dump("dump12"); 
     //Take derivative at zero lag
     deriv1stAtIndex(corrCubeDeriv0,corrCubeSm,1,indexLag,true);
   } else if (m_pVelAnaParam->maxCorrUsage == velAnaParam::GaussianWeightCorr) { //Use a gaussian function multiplied on the correlation to find the gradients.
     //Get the proper Gaussian function
     cubeDataF gaussfunc;
     getGaussfunc(gaussfunc,ax_lag,ax_lag.n*ax_lag.d/6.0f);
     gaussfunc.dump("dump9");
     
     cubeDataF deltaZcubeSlice; deltaZcubeSlice.becomeProxyCube(ax_ang);
     cubeDataF corrCube1(corrCube.domain());
     
     windParam wind(3); // (ax,z_b,x);
     wind[2].n = 1; wind[1].n = 1;
     for (ix=0; ix<ax_x.n; ix++) {
       windCorrCube[3].f  = ix;
       wind[2].f          = ix;
       for (izb=0; izb<ax_beam_z.n;izb++) {
         wind[1].f = izb;
         windCorrCube[2].f=izb;
         corrCubeTotal.grab2d(&corrCube,windCorrCube);
         timeDeriv1st(corrCube1,corrCube,true);
         corrCube1.multiply(corrCube);
         multiplySlice(corrCube1,corrCube1,gaussfunc);
         deltaZcube.setProxyXd(&deltaZcubeSlice,wind,1);
         sumAlongTimeTrc(deltaZcubeSlice,corrCube1);
       }
     }
   } else { //there should be no the 4th way to use correlations
     VERIFY(0);
   } //end calculating deltaZcube
  
  cubeDataF& wtCube = pWEMVAparam->finalWtCube;
  deltaZcube.allocMemoryTo(wtCube); 
  //Do the division of deltaZcube and the Ccube.
  float qt_mask = pWEMVAparam->thresMaxValPercent;
  bool bUseMask = true; bool bMaskResultToo = true;
  //Ccube should be always negative, anything above zero is invalid
  float fClipRat = clipCubeVal(Ccube,0.0f,-1000,true,false);
  PRV(fClipRat);
  
  divideWithCaveat(wtCube, deltaZcube, Ccube,
                   qt_mask, 0.0f, bUseMask, &(pWEMVAparam->thresMaskCube),bMaskResultToo);

  deltaZcube.dump("dJdu"); 
  wtCube.dump("dump10");
  
  //Now we have the forces in the wtCube cube. we need to put this value to the back projected image cube
  img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
  cubeDataF scaleCube(img_ax_t.domain());
  spreadWtFromCoarseToFineGrid(scaleCube,wtCube);
  m_bpjImg_ax_t.multiply(imgz,scaleCube);
  
  Ccube.dump("Bcube"); scaleCube.dump("finalWt");
  pWEMVAparam->thresMaskCube.dump("Emask");   
  //Ang to Off will be calculated using m_bpjImg_ax_t in the above function.
}

float WEMVA_FD_angleMoveout::calcObjFunc() {
  FUNC_TRACE_BEG(WEMVA_FD_angleMoveout::calcObjFunc);
  float costFunc = this->stackPowerObjFunc();
  prt("\n finish calc Obj Func in WEMVA_FD_angleMoveout, obj = %f\t", costFunc);
  FUNC_TRACE_END(costFunc);
}


///Time shift image gathers, DOESNOT work. Junk Idea!! Basically the focusing of the time gather
///too poor, and also the assumption that deltaS only makes the events shift laterally in time but
///keep invariant in (z,x) is not valid at all.
float WEMVA_FD_beamFocusTime::calcObjFunc() {
  return 1.0f;
}

//Calculate bakProjImg, timeShiftImage
//Input image is arranged as z,x,\tau
void WEMVA_FD_beamFocusTime::calcBackProjData2d() {
  FUNC_TRACE_BEG(WEMVA_FD_beamFocusTime::calcBackProjData)
  
  //The Image0 is been calculated already.
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  
  const hypercube& shapeTauImage = pWEMVAparam->Img0_hx.domain();
  //First transpose
  char tmpPar[1024]; char tmpCmd[1024];
  //Assume 2D tranpose from z,x,hx -> z,hx,x, thus plane=23
  VERIFY(shapeTauImage.get_ndim()==3);
  sprintf(tmpCmd,"Transp"); sprintf(tmpPar,"plane=23");
  
  vecAxis axes; axes.push_back(shapeTauImage.get_axis(1));
  axis tauAxis = shapeTauImage.get_axis(3);
  axes.push_back(tauAxis); axes.push_back(shapeTauImage.get_axis(2));
  
  hypercube transpShape(axes), transpShapeAng;
  mImg0_hx_t.alloc(transpShape);
  seplib::callSepTool(mImg0_hx_t,pWEMVAparam->Img0_hx,string(tmpCmd),string(tmpPar));
  
  //now axis is z,tau,x
  mImg0_hx_t.dump("tauCube");
  
  //Calculate the terms
  cubeData<float>& img_t = mImg0_hx_t;
  cubeData<float> Acube,Ccube;
  
  //A cube, one value for each x,z position.
  hypercube shapeXZ = shapeTauImage.return_compatible(1,2);
  Acube.alloc(shapeXZ); Ccube.alloc(shapeXZ);
  
  
  //performe calculation for each x and z location
  
  cubeData<float> I1cube, I2cube, tIcube;
  axis ax_tau = shapeTauImage.get_axis(3);
  axis ax_x = shapeTauImage.get_axis(2);
  axis ax_z = shapeTauImage.get_axis(1);
  I1cube.alloc(transpShape); I2cube.alloc(transpShape); 
  
  //Take deriv on tau axis
  deriv1st( I1cube, img_t, 2, true);
  deriv2nd( I2cube, img_t, 2, true);
  
  //calc E cube
  img_t.copyTo(tIcube); tIcube.multiply(I2cube);
  multAndStackAlongAxis2(Ccube,tIcube,NULL);
  //A cube
  img_t.copyTo(tIcube); tIcube.multiply(I2cube);
  cubeData<float> tau2Cube; tau2Cube.alloc(hypercube(ax_tau));
  for (int i=0; i<ax_tau.n; i++) {
   tau2Cube[i] = SQ(ax_tau.get_coord(i));
  }
  img_t.copyTo(tIcube); tIcube.multiply(I1cube);
  I1cube.dump("dump2");
  I2cube.dump("dump3");
  tau2Cube.dump("dump4");
  tIcube.dump("dump5");
  multiplySubCube(tIcube,tIcube,tau2Cube,2,2);
  tIcube.dump("dump6");
  multAndStackAlongAxis2(Acube,tIcube,&tau2Cube);
   
  //Since we need to divide Ccube. It is necessary to build a mask based on Ccube's value
  float vThres = Ccube.maxvalAbs() * pWEMVAparam->thresMaxValPercent;
  prt("\n The max absVal of Ccube and the threshold val for beamfocus: %f,%f", Ccube.maxvalAbs(), vThres);
  bool bUseAbsVal = true;
   
  pWEMVAparam->thresMaskCube.alloc(Ccube.domain());
  buildThresholdMask(pWEMVAparam->thresMaskCube,Ccube,vThres,bUseAbsVal);
  pWEMVAparam->thresMaskCube.dump("Emask");
   
  //Or we add a small number to cCube;
  //float veps = Ccube.RMS_NZ(1e-3) * 1e-2;
  //Ccube.add(veps);
  //PRV(Ccube.maxvalAbs()); PRV(Ccube.minvalAbs());
   
  cubeData<float> ACcube;
  Acube.allocMemoryTo(ACcube);
  ACcube.divideByMask(Acube,Ccube,pWEMVAparam->thresMaskCube); //!!This division has divide zero risks.
   
  Acube.dump("Acube"); Ccube.dump("Ccube"); ACcube.dump("ACcube");
  ACcube.normalize();
   
  cubeData<float> bpjImg_t;
  I1cube.copyTo(bpjImg_t);
  //Tranp it back to z,x,tau
  pWEMVAparam->backProjImg_hx.alloc(shapeTauImage);
  seplib::callSepTool(pWEMVAparam->backProjImg_hx,bpjImg_t,"Transp",string("plane=23"));

  //scale use ACcube
  multiplySlice(pWEMVAparam->backProjImg_hx,pWEMVAparam->backProjImg_hx,ACcube);

  FUNC_TRACE_END()
}

//angle domain DSO implementation. Taper on the edge
void WEMVA_FD_DSO_ang::flipGradient(cubeData<float>& gradCube) {
   //gradCube.scale(-1.0f);
}

float WEMVA_FD_DSO_ang::calcObjFunc() {

   //need to calcualte the initial images first somewhere before
   //Check the initial Image is not NULL
   
   WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
   VERIFY(pWEMVAparam->Img0_hx.domain().get_ndim() >= 2);

   //then calc the cost function, First tranform it to angle domain, and then take derivative
   //along the angle axis, and then taper.
   //const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
   //!!! Wait for implement
   float costFunc = 0.0f;
   //prt("finish calc Obj Func in WEMVA_FD_DSO_ang, obj = %f\n", costFunc);
   return costFunc;
}


void WEMVA_FD_DSO_ang::calcBackProjData() {
  FUNC_TRACE_BEG(WEMVA_FD_DSO_ang::calcBackProjData);
   //The Back Proj data is the Image -(dI^2(x,a)/da^2). 
   //The Image0 is been calculated already, Second derivative along the angle axis.
   VERIFY(m_bAngleGatherCalculated);
   
   const hypercube& shapeOffImage = m_pWEMVAparams->Img0_hx.domain();
   axis ax_off = shapeOffImage.get_axis(3);
   float hmax = -1.0f*ax_off.o; //assume the offset axis is from -hmax -> +hmax
   
   //transpose from z,x,hx --> z,hx,x and then perform off2ang tranform.
   cubeData<float>& Img0_hx_t = mImg0_hx_t;
   seplib::callSepTool(Img0_hx_t,m_pWEMVAparams->Img0_hx,"Transp","plane=23");
   
   //Transform to angle domain.
   Yang::off2Ang2d(m_pWEMVAparams->Img0_ax_t,Img0_hx_t,false,m_pWEMVAparams->off2angType,m_pWEMVAparams->amax);
   
   //Then take the 2nd derivative on angle axis.
   m_bpjImg_ax_t.alloc(m_pWEMVAparams->Img0_ax_t.domain());
   deriv2nd(m_bpjImg_ax_t,m_pWEMVAparams->Img0_ax_t,2,true);
   
   //m_pWEMVAparams->Img0_ax_t.dump("dump1");
   //m_pWEMVAparams->backProjImg_ax_t.dump("dump2");
   //Might need tapering here.
   cosDampingFunc dampFunc;
   stdVecFloat coefVec; int length = int(ax_off.n/2);
   dampFunc.getCoefEx(coefVec,length,1.0f);
   cubeData<float> TaperCube(ax_off.n);

   assert(ax_off.n%2 == 1);
   TaperCube[length] = 1.0f;
   for (int i=0; i<length; i++) {
     TaperCube[length+1+i] = coefVec[i];
     TaperCube[length-1-i] = coefVec[i];
   }
   
   //scale along the 2nd(ax) axis.
   if (m_pWEMVAparams->bTaperAngleGather) {
     multiplySubCube(m_bpjImg_ax_t,m_bpjImg_ax_t,TaperCube,2,2);
   }
   
   cubeData<float> backProjImg_hx_t;
   //Then tranform back to offset domain
   Yang::off2Ang2d(backProjImg_hx_t,m_bpjImg_ax_t,true,
                   m_pWEMVAparams->off2angType,hmax);
                                                                                         
   //Transpose back
   seplib::callSepTool(m_pWEMVAparams->backProjImg_hx, backProjImg_hx_t,"Transp","plane=23");
   
   //There is a negative sign 
   m_pWEMVAparams->backProjImg_hx.scale(-1.0f);
  FUNC_TRACE_END();
}

//FWI like WEMVA backprojection method, 3D case not finished yet!~~
void  WEMVA_maxAngleStack::calcBackProjData() {
  FUNC_TRACE_BEG(WEMVA_maxAngleStack::calcBackProjData);
  
  this->postCalcData(); //calc the angle gather at this step
  assert(m_bCalcAngleGather);
  VERIFY(m_bAngleGatherCalculated);
  
  cubeDataF& imgaxt = m_pWEMVAparams->Img0_ax_t;
  assert(!imgaxt.domain().isEmpty());
  //Stack the (z,ax,x) along the angle(ax) axis, the backProj is the spreading of the angle averaged trace.
  axis ax_x      = imgaxt.domain().get_axis(3);
  axis ax_angle  = imgaxt.domain().get_axis(2);
  axis ax_z      = imgaxt.domain().get_axis(1);
  
  cubeDataF&  bpjImgaxt = m_pWEMVAparams->backProjImg_ax;
  imgaxt.allocMemoryTo(bpjImgaxt);
  vecAxis axes; axes.push_back(ax_z); axes.push_back(ax_angle);
  hypercube sliceShape(axes);
  cubeData<float> imgSlice,imgSliceOut;
  imgSlice.becomeProxyCube(sliceShape);imgSliceOut.becomeProxyCube(sliceShape);
  windParam wind(3); wind[2].n=1;
  
  cubeDataF trcStacked;
  trcStacked.alloc(hypercube(ax_z));
  
  int ibeg=0, iend=ax_x.n;
  //If only calc sensitivity kernel, then only back project on image point of the angle gather.
  if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
    ibeg = ax_x.n/2; iend = ibeg+1;
  }
  for (int ix=ibeg; ix<iend; ix++) {
    wind[2].f = ix;
    imgaxt.setProxyXd(&imgSlice,wind,2);
    bpjImgaxt.setProxyXd(&imgSliceOut,wind,2);
    //Stack the angle
    trcStacked.zero(); stackTrc(trcStacked,imgSlice);
    spreadTrc(imgSliceOut,trcStacked);
  }
  
  if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
    cubeData<float> mask_x(ax_x);
    mask_x = 0.0f;
    //Put six spikes into the domain
    //int dnw = ax_x.n/8;
    //for (int i=0; i<7; i++) {
    //  mask_x(ax_x.n/2+(i-3)*dnw) = 1.0f;
    //}
    mask_x(ax_x.n/2) = 1.0f;
    scaleSlice(bpjImgaxt,bpjImgaxt,mask_x);
    LOG(calc only sens kernel in maxStackPowerBackProjImg);
  }
  
  imgaxt.dump("angCube"); bpjImgaxt.dump("angCubeBp");
  m_pWEMVAparams->Img0_hx.allocMemoryTo(m_pWEMVAparams->backProjImg_hx);
  this->convertAngleOffImage(bpjImgaxt,m_pWEMVAparams->backProjImg_hx,false);
  //Need to convert back to the offset image
  FUNC_TRACE_END();
}

float WEMVA_maxAngleStack::calcObjFunc() {
  return (this->stackPowerObjFunc());
}

