/* This is 2D GPU RTM with subsurface offset generation */
#include "gpumodel2d.cu.h"

//#define DEBUG
#if defined (DEBUG)
	#define _DBG printf("DBG: %d \n", __LINE__);
#else
	#define _DBG
#endif

//Macros
//#define IK4X_XBLKSIZE 16
//#define IK4X_ZBLKSIZE 32

//zero lag imaging kernel, p1 is wavefield 1, p2 is wavefield 2
//Remember the kernel domain now shrink to nwfld-FAT-pad
extern "C" __global__ void img_kernel( float* img, float* p1, float* p2) {
  int nImgBnd = FAT + npad_gpu;
	int i=blockIdx.x*blockDim.x+threadIdx.x + nImgBnd;
	int j=blockIdx.y*blockDim.y+threadIdx.y + nImgBnd;
	if (i >= (n1gpu-nImgBnd) || j>=(n2gpu-nImgBnd) ) return; //if nImgBnd>BLKSIZE, there should be zero probability. But Just in Case.
	
	img[gloc(i-nImgBnd,j-nImgBnd,n1gpu-2*nImgBnd)] += p1[dloc(i,j)]*p2[dloc(i,j)];
}

//Imaging gathers on subsurface offset(only horizontal), the h axis starts from h0, dh=number of grid points per point along h axis, 
//nh=number of h slices need to be calculated
//I(x,h) += p1(x-h) p2(x+h)
//Not really try to make it work

/**
extern "C" __global__ void ximg_kernel4(int h0, int dh, int nh, float *p1, float *p2){
	__shared__ float s_p1[IK4X_ZBLKSIZE][2*IK4X_XBLKSIZE];
	__shared__ float s_p2[IK4X_ZBLKSIZE][2*IK4X_XBLKSIZE];

    int i=blockIdx.x*blockDim.x+threadIdx.x;  	//z-coordinate
    int j=hmax_gpu+blockIdx.y*blockDim.y+threadIdx.y; //x-coordinate
	int hmin=h0-hmax_gpu; 
    int hmax=hmin+IK4X_XBLKSIZE;
	
	s_p1[i][threadIdx.y]=((j-hmax>=0 && j-hmax<n2gpu)?p1[dloc(i, j-hmax)]:0);
	s_p2[i][threadIdx.y]=((j+hmin>=0 && j+hmin<n2gpu)?p2[dloc(i, j +hmin)]:0);
	s_p1[i][IK4X_XBLKSIZE+threadIdx.y]=((j-hmin>=0 && j-hmin<n2gpu)?p1[dloc(i, j-hmin)]:0);
	s_p2[i][IK4X_XBLKSIZE+threadIdx.y]=((j+hmax>=0 && j+hmax<n2gpu)?p2[dloc(i, j+hmax)]:0);
	
	__syncthreads();
	
	//base points to the wavefield slice at h=h0
	int base=dloc(i,j)+h0*n1gpu*n2gpu;
	int stride=n1gpu*n2gpu;

	for(int lh=0; lh<nh; lh++){
		ximg_gpu0[base+(lh*dh)*stride]+=
			s_p1[i][IK4X_XBLKSIZE+threadIdx.y-lh]*s_p2[i][threadIdx.y+lh];
	}

**/

void copy_src_recv_slice_to_gpu(int *src_iz_cpu, int src_iz_len, float *src_cpu,
							    int *rec_iz_cpu, int rec_iz_len, float *rec_cpu,
								int nt,int nx) 
{
   
	assert(src_iz_len == nx - 2*FAT); 	assert(rec_iz_len == nx - 2*FAT);
	int ntrace = nx - 2*FAT;
	cuda_call(cudaMemcpy(recv_gpu,   rec_cpu,   nt*ntrace*sizeof(float),   cudaMemcpyHostToDevice));
	cuda_call(cudaMemcpy(source_gpu, src_cpu, 	nt*ntrace*sizeof(float),   cudaMemcpyHostToDevice));
	cuda_call(cudaMemcpy(recv_iz_gpu,rec_iz_cpu,    ntrace*sizeof(int),   cudaMemcpyHostToDevice));
	cuda_call(cudaMemcpy(src_iz_gpu, src_iz_cpu, 	ntrace*sizeof(int),   cudaMemcpyHostToDevice));
}




//Imaging gathers on subsurface offset(only horizontal), the h axis starts from h0, dh=number of grid points per point along h axis, 
//nh=number of h slices need to be calculated
//I(x,h) += p1(x-h) p2(x+h)
//Divide the (x,z) planes into 1D stripes along (x) direction, and let the length of each stripe H (which is the block size)
#define MAX_1D_BLOCK_SIZE 512
//define the allowed maximum (n-1)dh, if unfortunately the actualy H+(nh-1)*dh > sum{Two macros}
//then we divide nh into several sections, each section the h offset axis ns then statisfies H+(nh-1)*dh
#define MAX_NDH_SIZE 512
//For simplicity, Computing grid should cover n1gpu*n2gpu
//here h0,dh,nh are based on integer type grid point coordinates, not in physical system
extern "C" __global__ void ximg_kernelY1(float* pImg_gpu, int h0, int dh, int nh, float *p1, float *p2){
   //Share memory should be bigger than H+(nh-1)*dh
	__shared__ float s_p1[MAX_1D_BLOCK_SIZE+MAX_NDH_SIZE];
	__shared__ float s_p2[MAX_1D_BLOCK_SIZE+MAX_NDH_SIZE];

	int ig=blockIdx.x*blockDim.x+threadIdx.x;       /* global z-coordinate */
	int jg=blockIdx.y*blockDim.y+threadIdx.y; 		/* global x-coordinate */
	
	//int il = threadIdx.x;	//In this case it will always be one.
	int jl = threadIdx.y;  
	int hmax=h0+(nh-1)*dh;
	
	//There is blockDim.x+(nh-1)*dh float data needs to be read
	//while there are blockDim.x threads
	int nTotalShareMemSize = blockDim.y+(nh-1)*dh;
	int nReads = nTotalShareMemSize / (blockDim.y);
	
	int j1,j2;
	for (int i=0; i<nReads; i++) {
	   //fetch the data if not out of domain
	   j1 = jg-hmax+blockDim.y*i;
	   s_p1[jl+blockDim.y*i]=((j1 >=0 && j1<n2gpu)? p1[dloc(ig,j1)]:0);
	  j2 = jg+hmax+blockDim.y*i;
	   s_p2[jl+blockDim.y*i]=((j2 >=0 && j2<n2gpu)? p2[dloc(ig,j2)]:0);
	}
	
	//Read the last fraction
	int last_index = jl+nReads*blockDim.y;
	if ( last_index < nTotalShareMemSize) {
	   j1 = jg - hmax + blockDim.y*nReads;
	   j2 = jg + hmax + blockDim.y*nReads;
	   s_p1[last_index] = ((j1 >=0 && j1<n2gpu)? p1[dloc(ig,j1)]:0);
	   s_p2[last_index] = ((j1 >=0 && j1<n2gpu)? p2[dloc(ig,j2)]:0);
	}
	
	__syncthreads();
	
	//Now write the I(x,h) cube from h0 to hmax
	//base points to the wavefield slice at h=h0
	//Notice that address pImg_gpu is at h=h0 slice.
	int base=dloc(ig,jg);
	int stride=n1gpu*n2gpu;

	//I(x,h) = p1(x-h) * p2(x+h)
	//int xShift1 = jg-lh*dh-(jg-hmax-jl); //from global to shared memory
	//int xShift2 = jg+lh*dh-(jg+hmax-jl); //from global to shared memory
	for(int ih=0; ih<nh; ih++){
		pImg_gpu[base+(ih*dh)*stride]+=
			s_p1[jl-ih*dh+hmax]*s_p2[jl+ih*dh-hmax];
	}
	return;
}

//Fast algorithm for imaging subsurface offset gathers
//First check all the preconditions. Then invoke the kernel, return 1 if succeed.
int ximg_Y1(float* pImg_gpu, int h0, int dh, int nh, float *p1, float *p2, int desired1DblockSize)
{
   int ret = 0;
   
   VERIFYC(desired1DblockSize+(nh-1)*dh <= MAX_1D_BLOCK_SIZE+MAX_NDH_SIZE);
   //if not we need to subdivide nh into smaller sections....
   
   int wfd_nz, wfd_nx;
   cuda_call(cudaMemcpyFromSymbol(&wfd_nz, n1gpu, sizeof(int)));
   cuda_call(cudaMemcpyFromSymbol(&wfd_nx, n2gpu, sizeof(int)));
   
   dim3 dimGrid(wfd_nz,(wfd_nx-1)/desired1DblockSize+1);
   dim3 dimBlock(1,desired1DblockSize);
   kernel_exec(ximg_kernelY1<<<dimGrid,dimBlock>>>(pImg_gpu,h0,dh,nh,p1,p2));
   
   ret = 1; return ret;
}

//straightforward way of calculating horizontal offset gathers, not using shared memory
//For simplicity, Computing grid should cover n1gpu*n2gpu
//here h0,dh,nh are based on integer type grid point coordinates, not in physical system
//Remember the kernel domain now SHRINK to nwfld-FAT-pad
extern "C" __global__ void ximg_kernelY0(float* pImg_gpu, int h0, int dh, int nh, float *p1, float *p2){
  int nImgBnd = FAT + npad_gpu;
	int ig=blockIdx.x*blockDim.x+threadIdx.x + nImgBnd;       /* global z-coordinate */
	int jg=blockIdx.y*blockDim.y+threadIdx.y + nImgBnd; 		 /* global x-coordinate*/
	
	if (ig >= n1gpu-nImgBnd || jg >= n2gpu-nImgBnd) return;
	
	//Now write the I(x,h) cube from h0 to hmax
	//base points to the wavefield slice at h=h0
	//Notice that address pImg_gpu is at h=h0 slice.
	int base=gloc(ig-nImgBnd,jg-nImgBnd,n1gpu-2*nImgBnd);
	int stride=(n1gpu-2*nImgBnd)*(n2gpu-2*nImgBnd);

	int h;
	//I(x,h) = p1(x-h) * p2(x+h)
	//Notice that h could be negative
	for(int ih=0; ih<nh; ih++)
	{
	   h=h0+ih*dh;
	   if(jg-h>=nImgBnd && jg-h<(n2gpu-nImgBnd) && jg+h>=nImgBnd && jg+h<(n2gpu-nImgBnd)) 
	   {
		   pImg_gpu[base+(ih)*stride]+=
			  p1[dloc(ig,jg-h)]*p2[dloc(ig,jg+h)];
	   }
	}
	return;
}

//Common interface for prestack imaging condition
//output = p1(x-h) * p2(x+h)
int ximgY(float* pImg_gpu, int h0, int dh, int nh, float *p1, float *p2, int type, grid_layouts *p_gdlo) {
   
	if (type==0) { //dummy method
	   kernel_exec(ximg_kernelY0<<<p_gdlo->dimGridImage,p_gdlo->dimBlock>>>(
				   pImg_gpu,h0,dh,nh,p1,p2));
	}
	if (type==1) { //fast algorithm
	   ximg_Y1(pImg_gpu,h0,dh,nh, p1,p2,p_gdlo->desired1DblockSize);
	}
	return 1;
}
	
//For Adjoint RTM, multiply the refl model with the wavefield pIn, and add it to pOut
//n1gpu X n2gpu pOut(x) = pIn(x) * refl(x);
//refl should be only of size image
__global__ void scatter_kernel(float *pOut_gpu, float *pIn_gpu, float * pRefl)
{
  int nImgBnd = FAT + npad_gpu;
  int ig=blockIdx.x*blockDim.x+threadIdx.x+nImgBnd; 	// global z coordinates
  int jg=blockIdx.y*blockDim.y+threadIdx.y+nImgBnd; 	// global x coordinates

  if (ig >= n1gpu-nImgBnd || jg>=n2gpu-nImgBnd) return;

  //find the index of local grid point (ig,jg) in global coordinates.
  int addr=dloc(ig,jg);	  		  			// linear memroy address
  pOut_gpu[addr] += pIn_gpu[addr] * pRefl[gloc(ig-nImgBnd,jg-nImgBnd,n1gpu-2*nImgBnd)];
  return;
}

//For Adjoint RTM, multiply the refl model with the wavefield pIn, and add it to pOut
//n1gpu X n2gpu, Consider SUBSURFACE offset in this case
//pOut(x+h) += pIn(x-h) * refl(x,h), sum along all h if bSignPos=true
//pOut(x-h) += pIn(x+h) * refl(x,h), sum along all h if bSignPos=false
__global__ void scatter_subsurf_off_kernel(float *pOut,float *pIn,bool bSignPositive, 
										   float* pImgGather_gpu, int h0, int dh, int nh)
{
  int nImgBnd = FAT + npad_gpu;
  int ig=blockIdx.x*blockDim.x+threadIdx.x+nImgBnd;       /* global z-coordinate */
  int jg=blockIdx.y*blockDim.y+threadIdx.y+nImgBnd; 		/* global x-coordinate */

  if (ig >= n1gpu-nImgBnd || jg>=n2gpu-nImgBnd ) return;

	//Now write the I(x,h) cube from h0 to hmax
	//base points to the wavefield slice at h=h0
	//Notice that address pImgGather_gpu is at h=h0 slice.
  int base=gloc(ig-nImgBnd,jg-nImgBnd,n1gpu-2*nImgBnd);
  int stride=(n1gpu-2*nImgBnd)*(n2gpu-2*nImgBnd);

	int h,h2;
    /** !This implementation is wrong. Because the threads will have
        write conflict, notice each threads modify the memory from
    (jg-hmax)->(jg+hmax) and therefore different threads writes to 
    memory that is overlapped. 
     
    if (bSignPositive) { 
       for(int ih=0; ih<nh; ih++) {
		  h=h0+ih*dh;
		  if(jg-h>=0 && jg-h<n2gpu && jg+h>=0 && jg+h<n2gpu) 
		  {
			 pOut[dloc(ig,jg+h)] += pIn[dloc(ig,jg-h)]* pImgGather_gpu[base+(ih)*stride];
		  }
	   }
	}else {
		  for(int ih=0; ih<nh; ih++)
		  {
			 h=h0+ih*dh;
		  if(jg-h>=0 && jg-h<n2gpu && jg+h>=0 && jg+h<n2gpu)
			 {
				pOut[dloc(ig,jg-h)] += pIn[dloc(ig,jg+h)]* pImgGather_gpu[base+(ih)*stride];		  		
			}
		  }
	} **/
  //Each thread must writes to only one memory location, in this case, it is (ig,jg).
  //only the wfld inside the image region should be considered in the calculation
  if (bSignPositive) { 
    for(int ih=0; ih<nh; ih++) {
      h=h0+ih*dh;
      //since jg corresponds to (x+h). therefore jg-h <->x; and jg-2h <->x-h
      h2 = jg-2*h; base=gloc(ig-nImgBnd,jg-nImgBnd-h,n1gpu-2*nImgBnd);
      if(jg-h>=nImgBnd && jg-h<(n2gpu-nImgBnd) && h2>=nImgBnd && h2<(n2gpu-nImgBnd)) 
      {
        pOut[dloc(ig,jg)] += pIn[dloc(ig,h2)]* pImgGather_gpu[base+(ih)*stride];
      }
    }
  }else {
    for(int ih=0; ih<nh; ih++)
    {
      h=h0+ih*dh;
      //jg corresponds to (x-h), thus jg+h<->x; and jg+2*h<->x+h
      h2 = jg+2*h; base=gloc(ig-nImgBnd,jg-nImgBnd+h,n1gpu-2*nImgBnd);
      if(jg+h>=nImgBnd && jg+h<(n2gpu-nImgBnd) && h2>=nImgBnd && h2<(n2gpu-nImgBnd))
      {
        pOut[dloc(ig,jg)] += pIn[dloc(ig,h2)]* pImgGather_gpu[base+(ih)*stride];		  		
      }
    }
  } 
	return;
}

extern "C" void gpu2dRTM_init(int *wfld_j, int *gpu_id) {
   gpu2dInit(wfld_j, gpu_id);
   //extra initialization for RTM stuff
   init_wfLists(&wfList_gb);
}

extern "C" void gpu2dAdjRTM_init(int *wfld_j, int* record_j, int *gpu_id) {
   gpu2dModeling_init(wfld_j,record_j,gpu_id);
   pU0_gpu = NULL; pU1_gpu=NULL;
   recv_gpu = NULL; recv_iz_gpu = NULL;
   jImaging_gb = 0;
}

//srcIntense and recvIntense will be the same size as the image size.
extern "C" bool gpu2dRTM_setVelocity(float *cpu_space, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int *jImaging, cube_c* pImageCube_cpu,
                      float *srcIntense, float* recvIntense)	{

   bool bSizeChanged = gpu2dSetVelocity(cpu_space,new_nz,new_nx,nt,dt,dz,dx,dampCoefArray,dampArrayLength,npad);

   //Assume the nImaging parameter never changes
   jImaging_gb = *jImaging;
   int nImaging = int( (*nt)-1 / jImaging_gb) + 1;
   
   int wfnx = *new_nx + FAT*2;
   int wfnz = *new_nz + FAT*2;
   int imgx = (*new_nx-2*npad_cpu); int imgz = (*new_nz-2*npad_cpu);
   //needs to initialize recv_gpu, plist_global
   if (bSizeChanged) {
	  cuda_call(cudaFree(recv_gpu)); cuda_call(cudaMalloc((void**)&recv_gpu, (*nt)*(*new_nx)*sizeof(float)));
	  cuda_call(cudaFree(recv_iz_gpu)); cuda_call(cudaMalloc((void**)&recv_iz_gpu, (*new_nx)*sizeof(int)));
	  //re construct the plist memory
	  alloc_wfLists(&wfList_gb,wfnx,wfnz,nImaging);
    
    //Here the image_gpu size now changed to
	  cuda_call(cudaFree(image_gpu));
	  cuda_call(cudaMalloc((void**)(&image_gpu), imgx*imgz*sizeof(float)));
    cuda_call(cudaFree(dpDt2_gpu)); cuda_call(cudaMalloc(    (void**) &dpDt2_gpu,  wfnx*wfnz*sizeof(float)));
   }
   
   //deal intensity   
   if (NULL!=srcIntense) {
     bDoSrcIntens_cpu = true;
     //see if need reallocate memory
     if (srcIntense2d_gpu.nx != wfnx || srcIntense2d_gpu.nz!= wfnz) {
       reAlloc_slice2d_gpu(&srcIntense2d_gpu,wfnz,wfnx);
     }
     srcIntense2d_gpu.p_cpu = srcIntense;
   } else {
     bDoSrcIntens_cpu = false;
   }
   
   if (NULL!=recvIntense) {
     bDoRecvIntens_cpu = true;
     //see if need reallocate memory
     if (recvIntense2d_gpu.nx != wfnx || recvIntense2d_gpu.nz!=wfnz) {
       reAlloc_slice2d_gpu(&recvIntense2d_gpu,wfnz,wfnx);
     }
     recvIntense2d_gpu.p_cpu = recvIntense;
   } else {
     bDoRecvIntens_cpu = false;
   }

  bDoImageGather_gb = (NULL != pImageCube_cpu);
  ///Check the current gpu cube size
  if (bDoImageGather_gb) {
    //check if size changed, relocate memory if necessary
    if (!sameSize_cube_c(&imgCube_cpu_gb,pImageCube_cpu)) {
     copy_cube_c_info(&imgCube_cpu_gb, pImageCube_cpu);
     //Need to reallocate the gpu memory for gathers, set the memory pointer at NULL
     copy_cube_c_info(&imgGatherHorzOffset_gb, pImageCube_cpu);
     alloc_cube_c(&imgGatherHorzOffset_gb);
    }
    assert(imgGatherHorzOffset_gb.nx == (*new_nx-2*npad_cpu));
    //save the CPU ptr for later output
    imgCube_cpu_gb.p = pImageCube_cpu->p;
  }
  
  //Alloc the wfld_separation
  FFT1dWfldSeparate_alloc(&wfld_sep_gb,imgz,imgx);
  
  //prt("\n bSizeChanged=%d", bSizeChanged);
  return bSizeChanged;
}

//All cpu images should have a dimension of the wavefield, not velocity grid
//Can generate a image or image gathers, depending on whether the final pointer argument is null or not
//should keep the source intensity and receiver intensity functions.
//Keep in mind that Now the CPU memory for the image and off image are of size vel_nx-npad
extern "C" void gpu2dRTM_imaging(int *src_iz_cpu, int src_iz_len, float *src_cpu,		//source inject
								  int *rec_iz_cpu, int rec_iz_len, float *rec_cpu, 	 	//receiver inject
								  float *wfld_src, float *wfld_recv, //wavfield snapshot is wfld_j
								  float *cpu_image,bool bTake2ndDeriv, ////Stack image, offset gather image is decided in setVelocity function
                  bool bDoUpDownSeparation, float* cpu_image_refl, float* cpu_image_trans) //If do updown separation using Faqi Liu's method
                  
{
  //printf("\nGPUMIG2D: Beginning RTM on GPU\n");
  /*********************************************************************************************/
  /** ACTUAL WAVE PROPAGATION & IMAGING                                                       **/
  int nt, nz, nx, wfld_j;
	//retrieve the nt,nz,nx parameters from GPU
  get_ntzx_from_gpu(&nt,&nz,&nx,&wfld_j);
  int imgx = nx - 2*FAT - 2*npad_cpu;
  int imgz = nz - 2*FAT - 2*npad_cpu;
  
	//Copy the data from GPU to CPU
	copy_src_recv_slice_to_gpu( src_iz_cpu, src_iz_len, src_cpu,
							    rec_iz_cpu, rec_iz_len, rec_cpu,
								nt,nx);
	grid_layouts gdlo;
	gen_grid_layout_for_kernels(&gdlo,nt,nz,nx);

	int jSnapShot = wfld_j;
	int indexSnap = 0; int indexImaging = 0;
  float *ptemp = NULL;
	//reset the wavefield.
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
  float *pSrcInt=NULL, *pRecvInt=NULL;
  if (bDoSrcIntens_cpu) {
    pSrcInt = srcIntense2d_gpu.p_gpu;
    cuda_call(cudaMemset(pSrcInt,  0,nz*nx*sizeof(float)));
  }
  if (bDoRecvIntens_cpu) {
    pRecvInt = recvIntense2d_gpu.p_gpu;
    cuda_call(cudaMemset(pRecvInt, 0,nz*nx*sizeof(float)));
  }

  //prt("\nzzz,%d,%d,%d,  %d,%d,  %d,%d,%d  \n",nz,nx,npad_cpu,wfList_gb.nz,wfList_gb.nx, imgz,imgx,bTake2ndDeriv);
	zero_wfLists(&wfList_gb); //!IMPROTANT
	
  //LOGC(grtmi1);
	/** Propagate Data forward and save src wavefield on GPU **/
	//at time it, calcuate wfld at time it+2, for the sake of memory pointer switch efficiency
	//The wavfiled p0->it, p1->it+1
	//LOGC(enter RTM gpu loop);
	 for(int it=0; it<nt; it++){
		//verbPrt("\n it=%d", it);
		// Inject src cube, src at it should be injected to wfld at it+1
		  kernel_exec(inject_kernel<<<gdlo.inject_dimGrid,gdlo.inject_dimBlock>>>(src_iz_gpu, it, source_gpu, p1_gpu));
		  //update boundary at it+1
		  kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		  bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);	
		  //kernel_exec(bnd1d_kernel<<<dimGrid_bnd,dimBlock>>>(p1_gpu,p0_gpu));
		  
		 //Take snap shot first
		 if(it%jSnapShot == 0) {
			 cuda_call(cudaMemcpy(&wfld_src[nz*nx*indexSnap], p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));
			 indexSnap++;
		  }
		 
		 //Do propagation and modeling
		 if(it%jImaging_gb != 0){ 
		     // Propagate without saving for most steps
			 kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,NULL,pSrcInt));
			 ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
		 }else{ 
			 // At imaging timesteps, save wavefield for imaging
			 // there is a link-list trick.
			 ptemp=wfList_gb.pLists[indexImaging];
			 kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, ptemp,NULL,pSrcInt));
			 //swtich the pointers
			 wfList_gb.pLists[indexImaging]=p0_gpu;
			 p0_gpu=p1_gpu;p1_gpu=ptemp;
			 indexImaging++;
		}
	 }
  
  //LOGC(grtmi3); 
	//Now we've get through the source part
	//Propagate receiver backward, imaging correlation with source wavefields 
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float))); 
  cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
  //Now the image have different size.
	cuda_call(cudaMemset(image_gpu, 0,imgz*imgx*sizeof(float)));

  //Clean the offset Image as well
	if (bDoImageGather_gb) {
	   zero_cube_c(&imgGatherHorzOffset_gb);
	}
	
  //LOGC("do backward prop in gpu");
	indexImaging = 0;
	//Calculate it-2 from the wavslice at it and it-1
	//p0->it; p1->it-1; the receiver at time it should be injected to slice it-1.
	//LOGC("enter the back prop gpu loop")
	//verbPrt("\njImaging_gb=%d",jImaging_gb);
	for (int it=nt-1; it>=0; it--){
      //now p0->it, p1->it-1		 
      //Inject receiver wavefield at it to wav slice at it-1
      kernel_exec(inject_kernel<<<gdlo.inject_dimGrid,gdlo.inject_dimBlock>>>(recv_iz_gpu,it,recv_gpu, p1_gpu));
      //update boundary at it-1
      //kernel_exec(bnd1d_kernel<<<dimGrid_bnd,dimBlock>>>(p1_gpu,p0_gpu));
      kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
      bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
      //LOGC("grtmi34");
      //use p0->it as the wavefield at time it
      if(it%jImaging_gb==0){
         indexImaging = it / jImaging_gb; 
         //Call the imaging kernel
         if (!bTake2ndDeriv) {
            kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],p0_gpu));
         } else { //Take 2nd deriv on wfld
            kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],dpDt2_gpu));
         }
         //Do subsurface offset image gathers if required
         if(bDoImageGather_gb) {
            if (!bTake2ndDeriv) {
              kernel_exec(ximg_kernelY0<<<gdlo.dimGridImage,gdlo.dimBlock>>>(
                 imgGatherHorzOffset_gb.p,imgCube_cpu_gb.o,imgCube_cpu_gb.d,imgCube_cpu_gb.nframe,
                 p0_gpu,wfList_gb.pLists[indexImaging])); 
               //Fast algorithm, but not tested
               //ximg_Y1(imgGatherHorzOffset_gb.p,imgCube_cpu_gb.o,imgCube_cpu_gb.d,imgCube_cpu_gb.n,p0_gpu,wfList_gb.pLists[indexImg],MAX_1D_BLOCK_SIZE/2);
            } else {
               kernel_exec(ximg_kernelY0<<<gdlo.dimGridImage,gdlo.dimBlock>>>(
                 imgGatherHorzOffset_gb.p,imgCube_cpu_gb.o,imgCube_cpu_gb.d,imgCube_cpu_gb.nframe,
                 dpDt2_gpu,wfList_gb.pLists[indexImaging]));
            }
         }
      }
     //LOGC("grtmi35");
		 
		 //dump the receiver wavefield snapshot
		if(it%jSnapShot == 0) {
			indexSnap = it / jSnapShot;
      //prt("\n grtmi36, it=%d, jSnap=%d,%d,%d,%d",it,jSnapShot,nz,nx,indexSnap);
			//verbPrt("\n indexSnap = %d", indexSnap);
			cuda_call(cudaMemcpy(&(wfld_recv[nz*nx*indexSnap]), p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));
    }
    //Propagate the wavefield, afterthen p0->it-2, p1->it-1
    kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,dpDt2_gpu,pRecvInt));
    ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp; //switch pointer, then p0->it-1, p1->it-2
	}
  //LOGC(grtmi4);
	//Copy the image back to cpu space
  //The size is different now
	cuda_call(cudaMemcpy(cpu_image,image_gpu,imgx*imgz*sizeof(float),cudaMemcpyDeviceToHost));
  if (bDoImageGather_gb) {
	   VERIFYC(imgCube_cpu_gb.nx == imgx && imgCube_cpu_gb.nz == imgz);
	   long long sizeCopy = imgx*imgz*imgCube_cpu_gb.nframe;
	   cuda_call(cudaMemcpy(imgCube_cpu_gb.p,imgGatherHorzOffset_gb.p,sizeCopy*sizeof(float),cudaMemcpyDeviceToHost));
	}
  //Copy the src and recv intensity, here the gpu memory has different size than cpu memory. thus
  //nz*nx on gpu and imgx and imgz on cpu
  if (bDoSrcIntens_cpu) {
    myCudaMemcpy2d(srcIntense2d_gpu.p_cpu,imgz, srcIntense2d_gpu.p_gpu,nz, imgz,imgx, cudaMemcpyDeviceToHost,
                   0,0,FAT+npad_cpu,FAT+npad_cpu);
    //cuda_call(cudaMemcpy(srcIntense2d_gpu.p_cpu,srcIntense2d_gpu.p_gpu,nz*nx*sizeof(float),cudaMemcpyDeviceToHost));
  }
  if (bDoRecvIntens_cpu) {
    myCudaMemcpy2d(recvIntense2d_gpu.p_cpu,imgz, recvIntense2d_gpu.p_gpu,nz, imgz,imgx, cudaMemcpyDeviceToHost,
                   0,0,FAT+npad_cpu,FAT+npad_cpu);
  }
	return;
}
                 
extern "C" void gpuRTM2d_uninit() {
	uninit_wfLists(&wfList_gb);
	if (bDoImageGather_gb) {
	   	uninit_cube_c(&imgGatherHorzOffset_gb);
		bDoImageGather_gb = false;
	}
	gpuModeling2d_uninit();
}

extern "C" void gpu2dAdjRTM_uninit() {
   cuda_call(cudaFree(pU0_gpu));
   cuda_call(cudaFree(pU1_gpu));
   pU0_gpu = pU1_gpu = NULL;
   gpuRTM2d_uninit();
}

//Here remember the image(refl) might be I(x,h) with subsurface offset.
//Init only alloc the memory, does not copy data from cpu. This is in order to sepearte
//memory allocation to memcpy, to avoid unnecessary memcpies.
//Here the SIZE of refl and refl_off is ONLY wfnx-npad-FAT

extern "C" bool gpu2dAdjRTM_setVelocityAndRefl(float* vel_cpu, float* refl_cpu, cube_c* refl_off_gather_cpu, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int* jImaging, int nRecv, int ntRec, int* ix_recv, int* iz_recv)
{
   	bool bSizeChanged = gpu2dModeling_setVelocity(vel_cpu,new_nz,new_nx,nt,dt,dz,dx,dampCoefArray,dampArrayLength,npad, nRecv, ntRec, ix_recv, iz_recv);
	jImaging_gb = *jImaging;
	
	//Needs to allocate memory for reflctivity
	int new_wf_n1 = *new_nz + 2*FAT; int new_wf_n2 = *new_nx + 2*FAT;
	int imgz,imgx;
  imgz = *new_nz - 2*npad_cpu; imgx = *new_nx - 2*npad_cpu;
  
	bDoImageGather_gb = (NULL != refl_off_gather_cpu);
	VERIFYC( (bDoImageGather_gb&&NULL==refl_cpu) || (!bDoImageGather_gb&&NULL!=refl_cpu));
	int n12 = new_wf_n1*new_wf_n2;
  int n12Img = (imgz*imgx);
	if (bSizeChanged) {
	   cuda_call(cudaFree(reflGpu));
     //Here the size is different
	   cuda_call(cudaMalloc((void**)&reflGpu, n12Img*sizeof(float)));
	   //allocate the scattered wavefields
	   cuda_call(cudaFree(pU0_gpu)); cuda_call(cudaFree(pU1_gpu)); 
	   cuda_call(cudaMalloc(    (void**) &pU0_gpu,  n12*sizeof(float)));
	   cuda_call(cudaMalloc(    (void**) &pU1_gpu,  n12*sizeof(float)));
     cuda_call(cudaFree(dpDt2_gpu)); cuda_call(cudaMalloc(    (void**) &dpDt2_gpu,  n12*sizeof(float)));
	}

	
  if (bDoImageGather_gb) {
    assert(refl_off_gather_cpu->nz == imgz);
	  if (!sameSize_cube_c(refl_off_gather_cpu,&imgGatherHorzOffset_gb)) {
		 //reallocate the memory to save the img gather in GPU
		 //Need to reallocate the gpu memory for gathers
		 copy_cube_c_info(&imgGatherHorzOffset_gb, refl_off_gather_cpu);
		 alloc_cube_c(&imgGatherHorzOffset_gb);
	  }
   }
   
   //Memory Copy
   if (!bDoImageGather_gb) {
		cuda_call(cudaMemcpy(reflGpu, refl_cpu, n12Img*sizeof(float), cudaMemcpyHostToDevice) );
   } else {
	  //Copy the imgGather to GPU memory
	   int nframe = refl_off_gather_cpu->nframe;
	   cuda_call(cudaMemcpy(imgGatherHorzOffset_gb.p, refl_off_gather_cpu->p, 
					  n12Img*nframe*sizeof(float), cudaMemcpyHostToDevice));
   }
   return bSizeChanged;
}

//iz_cpu the inject source locations:iz along the ix grid axis
extern "C" void gpu2dAdjRTM(int *iz_cpu, int iz_cpu_len, float *source_cpu, float *wfldSrc, float *wfldScatter,
							float *recordBuffer) //output
{
    int nt, nz, nx, wfld_j;
	int nRecv, record_j, itRec, ntRec;

	get_ntzx_from_gpu(&nt,&nz,&nx,&wfld_j);
	
	int indexSnap = 0;
  float *ptemp = NULL;

	record_j = record_j_cpu;
	nRecv =  nRecordRecvCpu;
	ntRec =  ntRecCpu;
	
	//The source inject slice copy to GPU, need to revise.
  cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*(nx-2*FAT)*sizeof(float), cudaMemcpyHostToDevice));
	//printf("iz_cup_len=%d, nx=%d\n",iz_cpu_len,nx);
	assert(iz_cpu_len == nx - 2*FAT);
  cuda_call(cudaMemcpy(src_iz_gpu, iz_cpu, (nx-2*FAT)*sizeof(int), cudaMemcpyHostToDevice));
    
	//clean up the wavfield memory
  cuda_call(cudaMemset(p0_gpu, 0, nz*nx*sizeof(float)));
  cuda_call(cudaMemset(p1_gpu, 0, nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
    
	//For data recording
	grid_layouts gdlo;
	gen_grid_layout_for_kernels(&gdlo,nt,nz,nx);
	int block1D = gdlo.inject_dimBlock;
	
	/* The ACTUAL WAVE PROPAGATION */
    indexSnap = 0; 
	//Here at time it, calculate the wavfield values at time it+1
    //p0-->it-1, p1-->it, p0-->it+1
	//pU0-->it-1,pU1-->it,pU0-->it+1
	//meanwhile, inject p1-->it * reflModel into pU1_gpu-->it+1
    for (int it=0; it<nt; it++)
    {
	   //Prop the src wavefield
        kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu, dpDt2_gpu));
        //source at time 'it' should be applied to wavefield at time 'it+1'
        kernel_exec(inject_kernel<<<gdlo.inject_dimGrid, gdlo.inject_dimBlock>>>(src_iz_gpu, it, source_gpu, p0_gpu));
		
		//Absorbing boundary condition on wavefield at it+1, using wavefield at it(p1_gpu)
		//kernel_exec(bnd1d_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(p0_gpu,p1_gpu));
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p0_gpu,p1_gpu));
		bnd_abs_ord1(p0_gpu, p1_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);	

		
		kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(pU0_gpu, pU1_gpu, pU0_gpu,NULL));
		//Inject the scatter wavefield
		if (it%jImaging_gb == 0) {
		   if (!bDoImageGather_gb) {
         //inject 2nd time deriv, not the src wfld itself
				 kernel_exec(scatter_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(pU0_gpu,dpDt2_gpu,reflGpu));
		   } else { //Need to scatter using the offset gather
			//(float *pOut,float *pIn,bool bSignPositive, float* pImgGather_gpu, int h0, int dh, int nh)
			  LOGC(mark adjRTM gpu XX);
				 kernel_exec(scatter_subsurf_off_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(
					pU0_gpu,dpDt2_gpu,true, imgGatherHorzOffset_gb.p, imgGatherHorzOffset_gb.o,
					imgGatherHorzOffset_gb.d, imgGatherHorzOffset_gb.nframe));
		   }
		}
		//kernel_exec(bnd1d_kernel<<<dimGrid_bnd,dimBlock>>>(pU0_gpu,pU1_gpu));
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(pU0_gpu,pU1_gpu));
		bnd_abs_ord1(pU0_gpu, pU1_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);

    //repeatively copy the wavefield memory out to host memory for snapShot
	  if(it%wfld_j==0)
	  {
  		cuda_call(cudaMemcpy(&wfldSrc[nz*nx*indexSnap], p1_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));
  		cuda_call(cudaMemcpy(&wfldScatter[nz*nx*indexSnap], pU1_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));
  		indexSnap++;
	  }
		//Harvest the data at time it;
		if(it%record_j==0) {
		   itRec = it / record_j;
		   kernel_exec(
			 record_kernel<<<(nRecv-1)/block1D+1,block1D>>>(recordRecvIzGpu,recordRecvIxGpu,nRecv,pU1_gpu,ntRec,itRec,recordDataGpu));
		}
		//switch the wavefield memory pointer
    ptemp=p0_gpu;  p0_gpu=p1_gpu;   p1_gpu=ptemp;
		ptemp=pU0_gpu; pU0_gpu=pU1_gpu; pU1_gpu=ptemp;
    }
	
	//Copy the recorded data back to the host memory
	cuda_call(cudaMemcpy(recordBuffer,recordDataGpu,ntRec*nRecv*sizeof(float),cudaMemcpyDeviceToHost));
	return;
}

//////////////
