#include "gpumig2d.cu.h"

/***************************************************************/
extern "C" void gpu2dTomAdj_init(int *wfld_j, int *gpu_id) {
   int record_j_dummy = 0;
   gpu2dAdjRTM_init(wfld_j,&record_j_dummy, gpu_id);
   image_gpu = NULL;
}

extern "C" void gpu2dTomAdj_uninit() {
   gpu2dAdjRTM_uninit();
}

//Input background slowness/velocity, and input Image perturbation, can be postStack image or prestack image.
//Input image perturbation, this image has been cropped due to the aperture.
extern "C" bool gpu2dTomAdj_setVelocity(float *vel_cpu, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int *jImaging, 
										float* pImg_cpu, cube_c* img_off_gather_cpu)	
{
   bool bSizeChanged = gpu2dSetVelocity(vel_cpu,new_nz,new_nx,nt,dt,dz,dx,dampCoefArray,dampArrayLength,npad);

   //Assume the nImaging parameter never changes
   jImaging_gb = *jImaging;
   int nImaging = int( (*nt)-1 / jImaging_gb) + 1;
   
   int wfnx = *new_nx + FAT*2;
   int wfnz = *new_nz + FAT*2;
   int imgx = *new_nx - 2*npad_cpu;
   int imgz = *new_nz - 2*npad_cpu;
   //LOGC(tsgp1);
   //needs to initialize recv_gpu, plist_global, RTM part
   if (bSizeChanged) {
	  cuda_call(cudaFree(recv_gpu)); cuda_call(cudaMalloc((void**)&recv_gpu, (*nt)*(*new_nx)*sizeof(float)));
	  cuda_call(cudaFree(recv_iz_gpu)); cuda_call(cudaMalloc((void**)&recv_iz_gpu, (*new_nx)*sizeof(int)));
	  //re construct the plist memory
	  alloc_wfLists(&wfList_gb,wfnx,wfnz,nImaging);
	  cuda_call(cudaFree(image_gpu)); cuda_call(cudaMalloc((void**)(&image_gpu), imgx*imgz*sizeof(float)));
    cuda_call(cudaFree(dpDt2_gpu)); cuda_call(cudaMalloc((void**)(&dpDt2_gpu), wfnx*wfnz*sizeof(float)));
   }

   
  //Needs to allocate memory for reflctivity, adjoint RTM part, here we assume the npad+FAT never changes. In this case
  //bSizeChanged will also tell me whether the size of refl imgx*imgz changed or not.
  //LOGC(tsgp2);
	bDoImageGather_gb = (NULL != img_off_gather_cpu);
	VERIFYC( (bDoImageGather_gb&&NULL==pImg_cpu) || (!bDoImageGather_gb&&NULL!=pImg_cpu))
	int n12 = wfnx*wfnz;
	if (bSizeChanged) {
	  cuda_call(cudaFree(reflGpu));
	  cuda_call(cudaMalloc((void**)&reflGpu, imgx*imgz*sizeof(float)));
	
    //LOGC(tsgp22);   	   
	  //allocate the scattered wavefields
	  cuda_call(cudaFree(pU0_gpu)); cuda_call(cudaFree(pU1_gpu));
	  cuda_call(cudaMalloc(    (void**) &pU0_gpu,  n12*sizeof(float)));
	  cuda_call(cudaMalloc(    (void**) &pU1_gpu,  n12*sizeof(float)));
	}
	//LOGC(tsgp3);
  
  if (bDoImageGather_gb) {
	  if (!sameSize_cube_c(img_off_gather_cpu,&imgGatherHorzOffset_gb)) {
		  //reallocate the memory to save the img gather in GPU
		  //Need to reallocate the gpu memory for gathers
		  copy_cube_c_info(&imgGatherHorzOffset_gb, img_off_gather_cpu);
		  alloc_cube_c(&imgGatherHorzOffset_gb);
	  }
    assert(img_off_gather_cpu->nz == imgz);
  }

  //LOG(memcpy in gpu tomAdj setvel);
  if (!bDoImageGather_gb) {
	   cuda_call(cudaMemcpy(reflGpu, pImg_cpu, imgx*imgz*sizeof(float), cudaMemcpyHostToDevice) );
	} else {
	   //Copy the imgGather to GPU memory
	   int nframe = imgGatherHorzOffset_gb.nframe;
     //LOGC(tsgp5);
     //prt("\n%d,%d,%d,%d,%d\n",imgx,imgz,nframe,imgGatherHorzOffset_gb.nz,img_off_gather_cpu->p);
	   cuda_call(cudaMemcpy(imgGatherHorzOffset_gb.p, img_off_gather_cpu->p, 
							  imgx*imgz*nframe*sizeof(float), cudaMemcpyHostToDevice));
	}
  //LOGC(tsgp6);
  return bSizeChanged;   
}


////// For tomographic operators
//Refer to Ali and Tang (2010) , SEP 142

//adjoint tomographic operator, T' delta(I(x,h)) = delta(S(y)), Per Shot
///INPUT: 	src, recv information, BE AWARE, here RECV data is without second deriv, thus needs to take 2nd wfld inside
///         at least once.
///		  	  image pertubation I(x,h) of size imgx*imgz,i.e. (n1gpu-FAT-npad)  //cropped
///OUTPUT: 	slowness pertubation delta(S(x)) of same size imgz*imgx

extern "C" void gpu2dTomAdj(int *src_iz_cpu, int src_iz_len, float *src_cpu,		//source inject
								  int *rec_iz_cpu, int rec_iz_len, float *rec_cpu, 	 	//receiver inject
								  float *wfld_src, float *wfld_recv, float *wfld_src2, float *wfld_recv2,
								  float* dSs_cpu, float* dSr_cpu, bool bTake4thDeriv) //slowness pertubation for output
{
    //printf("\nGPU2dTomAdj: Beginning 2d Tom Adj on GPU\n");
    /*********************************************************************************************/
	/** Use 2 way acoustic WAVE PROPAGATION  **/
  int nt, nz, nx, wfld_j;
	//retrieve the nt,nz,nx parameters from GPU
  get_ntzx_from_gpu(&nt,&nz,&nx,&wfld_j);
	
	//Copy the data from GPU to CPU
	copy_src_recv_slice_to_gpu(src_iz_cpu, src_iz_len, src_cpu,
                             rec_iz_cpu, rec_iz_len, rec_cpu,
                             nt,nx);
	grid_layouts gdlo;
	gen_grid_layout_for_kernels(&gdlo,nt,nz,nx);
  
  int imgx = nx - 2*FAT - 2*npad_cpu;
  int imgz = nz - 2*FAT - 2*npad_cpu;
  
	int indexSnap = 0; int indexImaging = 0;
  float *ptemp = NULL;
	//reset the wavefield.
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU1_gpu, 0,nz*nx*sizeof(float)));
  cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
  zero_wfLists(&wfList_gb); //!IMPROTANT    
	
	bool bSnapSrc1 = (NULL != wfld_src);
	bool bSnapSrc2 = (NULL != wfld_src2);
	bool bSnapRecv1 = (NULL != wfld_recv);
	bool bSnapRecv2 = (NULL != wfld_recv2);
	//The extra part for the Tomo operator
	//need two images memory for deltaS_r and deltaS_s, thus add image2_gpu
	
	//First calculate deltaS_s: 
	//1. Do adjoint RTM modeling using src and the image gathers, remember to save the wavfield delta(wav_SRC) to the list
	//2. Back propagate wavefield and then cross correlate with the wavefield saved at step I to result in the final Image. Needs to scale S_0(x) in CPU code.
	
	/** Propagate Data forward and save src wavefield on GPU **/
	//at time it, calcuate wfld at time it+2, for the sake of memory pointer switch efficiency
	//The wavfiled p0->it, p1->it+1; same as pU0 and pU1, in order to do the link-list trick, has to arrange this way.
	//LOGC(enter RTM gpu loop);
	 for(int it=0; it<nt; it++){
		//source at time 'it' should be applied to wavefield at time 'it+1'
    kernel_exec(inject_kernel<<<gdlo.inject_dimGrid, gdlo.inject_dimBlock>>>(src_iz_gpu, it, source_gpu, p1_gpu));
		
		//Absorbing boundary condition on wavefield at it+1, using wavefield at it(p0_gpu)
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);	
		
		//Inject the scatter wavefield to it+1, inject the 2nd time deriv of bg wfld
		if (!bDoImageGather_gb) {
			  kernel_exec(scatter_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(pU1_gpu,dpDt2_gpu,reflGpu));
              //kernel_exec(scatter_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(pU1_gpu,p0_gpu,reflGpu));
		} else { //Need to scatter using the offset gather
		 //(float *pOut,float *pIn,bool bSignPositive, float* pImgGather_gpu, int h0, int dh, int nh)
		      kernel_exec(scatter_subsurf_off_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(
				 pU1_gpu,dpDt2_gpu,true, imgGatherHorzOffset_gb.p, imgGatherHorzOffset_gb.o,
				 imgGatherHorzOffset_gb.d, imgGatherHorzOffset_gb.nframe));
		}
		
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(pU1_gpu,pU0_gpu));
		bnd_abs_ord1(pU1_gpu, pU0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		
		//repeatively copy the wavefield memory out to host memory for snapShot
		if(it%wfld_j==0)
		{
		   if (bSnapSrc1) {cuda_call(cudaMemcpy(&wfld_src[nz*nx*indexSnap], p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));  }
       //if (bSnapSrc1) {cuda_call(cudaMemcpy(&wfld_src[nz*nx*indexSnap], dpDt2_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));  }
		   if (bSnapSrc2) {cuda_call(cudaMemcpy(&wfld_src2[nz*nx*indexSnap], pU0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));}
		   indexSnap++;
		}

		//Prop the src wavefield from it,it+1 -> it+1,it+2
    kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,dpDt2_gpu));
		ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
		
		//progagate the scattered source wavefield from it,it+1, -> it+2,it
		//Do propagation, same link-list trick, save all the source scattered wavefield
		if(it%jImaging_gb != 0){ 
			  /* Propagate without saving for most steps */
			  kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(pU0_gpu, pU1_gpu, pU0_gpu));			  
			  ptemp=pU0_gpu; pU0_gpu=pU1_gpu;pU1_gpu=ptemp;
		 }else{ 
			  // At imaging timesteps, save wavefield
			  ptemp=wfList_gb.pLists[indexImaging];
			  kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(pU0_gpu, pU1_gpu,ptemp)); //ptemp point to it+2, then switch these three pointers.
			  //swtich the pointers
			  wfList_gb.pLists[indexImaging]=pU0_gpu;
			  pU0_gpu=pU1_gpu;pU1_gpu=ptemp;
			  indexImaging++;
		}		 		
	} //for it=0->nt-1 loop
	
	//Now we've get through the source part, the scattered source wfld is stored in wfList_gb.
	 
	// Propagate receiver backward imaging correlation against
	// the scattered source wavefields 
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(image_gpu, 0,imgz*imgx*sizeof(float)));
	cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
    
	indexImaging = 0; indexSnap=0; ptemp=NULL;
	
	//Calculate it-2 from the wavslice at it and it-1
	//p0->it; p1->it-1; the receiver at time it should be injected to slice it-1.
	//LOGC("enter the back prop gpu loop")
	//verbPrt("\njImaging_gb=%d",jImaging_gb);
	for (int it=nt-1; it>=0; it--){
		 //now p0->it, p1->it-1		 
		 //Inject receiver wavefield at it to wav slice at it-1
		 kernel_exec(inject_kernel<<<gdlo.inject_dimGrid,gdlo.inject_dimBlock>>>(recv_iz_gpu,it,recv_gpu, p1_gpu));
		 //update boundary at it-1
		 kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		 bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		 //use p0->it as the wavefield at time it, cross-correlate with the wavefield we save
		 if(it%jImaging_gb==0){
			indexImaging = it / jImaging_gb; 
			//Call the imaging kernel, zero lag
            if (!bTake4thDeriv) {
               kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],p0_gpu));			
            } else {
               kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],dpDt2_gpu));
            }
		 }
		 
		 //dump the receiver wavefield snapshot
		 if(it%wfld_j== 0) {
			indexSnap = it / wfld_j;
			//verbPrt("\n indexSnap = %d", indexSnap);
			if (bSnapRecv1) { cuda_call(cudaMemcpy(&(wfld_recv[nz*nx*indexSnap]), p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost)); }
		}
		
		//Propagate the wavefield, afterthen p0->it-2, p1->it-1
         if (!bTake4thDeriv) {
           kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
         } else {
            kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu, dpDt2_gpu));
         }
		 ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp; //switch pointer, then p0->it-1, p1->it-2
	}
	
	//Copy the image back to cpu space
	cuda_call(cudaMemcpy(dSs_cpu,image_gpu,imgx*imgz*sizeof(float),cudaMemcpyDeviceToHost));
	
	////SECOND STEP, CALCULATE deltaS_{r}
	indexSnap = 0; indexImaging = 0; ptemp = NULL;
	//reset the wavefield.
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU1_gpu, 0,nz*nx*sizeof(float)));
  cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
    
	zero_wfLists(&wfList_gb);
	//3. forward propagate src wavefield and save to the wavefield list
	//4. Do adjoint RTM modeling backward in time using receiver data and the image gathers, then cross correlate with the wavefield saved at step 3 to result in the final Image. Needs to scale S_0(x) in CPU code.
	
	/** Propagate Data forward and save src wavefield on GPU **/
	//at time it, calcuate wfld at time it+2, for the sake of memory pointer switch efficiency
	//The wavfiled p0->it, p1->it+1; in order to do the link-list trick, has to arrange this way.
	for(int it=0; it<nt; it++){
		//source at time 'it' should be applied to wavefield at time 'it+1'
    kernel_exec(inject_kernel<<<gdlo.inject_dimGrid, gdlo.inject_dimBlock>>>(src_iz_gpu, it, source_gpu, p1_gpu));
		
		//Absorbing boundary condition on wavefield at it+1, using wavefield at it(p0_gpu)
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);	
		
		//Do propagation, same link-list trick, save all the source scattered wavefield
		if(it%jImaging_gb != 0){
			  /* Propagate without saving for most steps */
			  kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,dpDt2_gpu));
			  ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
		 }else{ 
            if (!bTake4thDeriv) {
                 // At imaging timesteps, save wavefield
                 ptemp=wfList_gb.pLists[indexImaging];
                 kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu,ptemp)); //ptemp point to it+2, then switch these three pointers.
                 //swtich the pointers
                 wfList_gb.pLists[indexImaging]=p0_gpu;
                 p0_gpu=p1_gpu;p1_gpu=ptemp;
            } else {
                 kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
                 cuda_call(cudaMemcpy(wfList_gb.pLists[indexImaging],dpDt2_gpu,nx*nz*sizeof(float),cudaMemcpyDeviceToDevice));
                 ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
            }
			indexImaging++;
		}		 		
	} //for it=0->nt-1 loop
	 
	//Now we've get through the source part, source wfld is stored in wfList_gb.
	 
	// Propagate receiver backward imaging correlation against
	// the scattered source wavefields 
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU1_gpu, 0,nz*nx*sizeof(float)));
  cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(image_gpu, 0,imgz*imgx*sizeof(float)));
	indexSnap = 0; indexImaging = 0; ptemp = NULL;

	
	///step 4: Calculate it-2 from the wavslice at it and it-1, mean while do born modeling backwards.
	//p0->it; p1->it-1; the receiver at time it should be injected to slice it-1.
	for (int it=nt-1; it>=0; it--){
		 //Inject receiver wavefield at it to wav slice at it-1
		 kernel_exec(inject_kernel<<<gdlo.inject_dimGrid,gdlo.inject_dimBlock>>>(recv_iz_gpu,it,recv_gpu, p1_gpu));
		 //update boundary at it-1
		 kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		 bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		 
		//Inject to the back scattered wavefield at it-1, 2nd order deriv in time
		if (it%jImaging_gb==0) {
		   if (!bDoImageGather_gb) {
				 kernel_exec(scatter_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(pU1_gpu,dpDt2_gpu,reflGpu));
		   } else { //Need to scatter using the offset gather
			//(float *pOut,float *pIn,bool bSignPositive, float* pImgGather_gpu, int h0, int dh, int nh)
				 kernel_exec(scatter_subsurf_off_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(
					pU1_gpu,dpDt2_gpu,false,imgGatherHorzOffset_gb.p, imgGatherHorzOffset_gb.o,
					imgGatherHorzOffset_gb.d, imgGatherHorzOffset_gb.nframe));
		   }
		}
		
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(pU1_gpu,pU0_gpu));
		bnd_abs_ord1(pU1_gpu, pU0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		
		
		//use p0->it as the wavefield at time it, APPLY ZERO LAG IMAGING CONDITION cross-correlate with the wavefield we save
		 if(it%jImaging_gb==0){
			indexImaging = it / jImaging_gb; 
			//Call the imaging kernel, zero lag
			kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],pU0_gpu));			
		}

		//dump the receiver wavefield snapshot
		   if(it%wfld_j== 0) {
			  indexSnap = it / wfld_j;
			  //cuda_call(cudaMemcpy(&(wfld_src[nz*nx*indexSnap]), p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));
			  if (bSnapRecv2) {cuda_call(cudaMemcpy(&(wfld_recv2[nz*nx*indexSnap]), pU0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));}
		}
		
		//back Prop the recv wavefield from it,it-1 -> it-1,it-2
    kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,dpDt2_gpu));
		ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
		
		//back Propagate the scattered recv wavefield.
		kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(pU0_gpu, pU1_gpu, pU0_gpu));
		ptemp=pU0_gpu; pU0_gpu=pU1_gpu;pU1_gpu=ptemp; 
		//switch pointer, then pU0->it-1, pU1->it-2
	}

	//Finish all steps, 	//Copy the delta(S)_r image back to cpu space
	cuda_call(cudaMemcpy(dSr_cpu,image_gpu,imgx*imgz*sizeof(float),cudaMemcpyDeviceToHost));
	return;
}


///Tomographic operator, T delta(S(y)) = delta(I(x,h)) or just the h=0 slice, in which case output is I(x), 
///the prestack image output is optional
///INPUT: 	slowness pertubation delta(S(x)) of size n1gpu*n2gpu, 
///			src, recv information, BE AWARE, here RECV data is processed with 2nd derviative
///		  	
///OUTPUT: 	image pertubation I(x,h) of size n1gpu*n2gpu, Two image cubes for source & receiver wfld respectively
///  		or I(x)
extern "C" void gpu2dTomFwd_init(int *wfld_j, int *gpu_id) {
   int record_j_dummy = 0;
   gpu2dAdjRTM_init(wfld_j,&record_j_dummy,gpu_id);
}

extern "C" void gpu2dTomFwd_uninit() {
   gpu2dAdjRTM_uninit();
}

//Input base slowness/velocity, and slowness Image perturbation, can be postStack image or prestack image.
//Here the output image is already cropped size n1gpu-2*FAT-2*npad
extern "C" bool gpu2dTomFwd_setVelocity(float *vel_cpu, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int *jImaging, 
										  cube_c* img_off_gather_cpu)	
{
   bool bSizeChanged = gpu2dSetVelocity(vel_cpu,new_nz,new_nx,nt,dt,dz,dx,dampCoefArray,dampArrayLength,npad);

   //Assume the nImaging parameter never changes
   jImaging_gb = *jImaging;
   int nImaging = int( (*nt)-1 / jImaging_gb) + 1;
   
   int wfnx = *new_nx + FAT*2;
   int wfnz = *new_nz + FAT*2;
   int imgx = *new_nx - npad_cpu*2;
   int imgz = *new_nz - npad_cpu*2;
   
   //needs to initialize recv_gpu, plist_global, RTM part
   if (bSizeChanged) {
	  cuda_call(cudaFree(recv_gpu)); cuda_call(cudaMalloc((void**)&recv_gpu, (*nt)*(*new_nx)*sizeof(float)));
	  cuda_call(cudaFree(recv_iz_gpu)); cuda_call(cudaMalloc((void**)&recv_iz_gpu, (*new_nx)*sizeof(int)));
	  //re construct the plist memory
	  alloc_wfLists(&wfList_gb,wfnx,wfnz,nImaging);
	  
	  cuda_call(cudaFree(image_gpu)); cuda_call(cudaMalloc((void**)(&image_gpu), imgx*imgz*sizeof(float)));
    cuda_call(cudaFree(dpDt2_gpu)); cuda_call(cudaMalloc((void**)(&dpDt2_gpu), wfnx*wfnz*sizeof(float)));
   }

  //Needs to allocate memory for reflctivity, adjoint RTM part
	bDoImageGather_gb = (NULL != img_off_gather_cpu);
	int n12 = wfnx*wfnz;
	if (bSizeChanged) {
	   cuda_call(cudaFree(reflGpu));cuda_call(cudaMalloc((void**)&reflGpu, imgz*imgx*sizeof(float)));
	   	   
	  //allocate the scattered wavefields
	  cuda_call(cudaFree(pU0_gpu)); cuda_call(cudaFree(pU1_gpu));
	  cuda_call(cudaMalloc(    (void**) &pU0_gpu,  n12*sizeof(float)));
	  cuda_call(cudaMalloc(    (void**) &pU1_gpu,  n12*sizeof(float)));
	}
	
  if (bDoImageGather_gb) {
    assert(img_off_gather_cpu->nx == imgx);
    //If size is different, relocate memory
    if (!sameSize_cube_c(img_off_gather_cpu,&imgCube_cpu_gb)) {
      //reallocate the memory to save the img gather in GPU
      copy_cube_c_info(&imgCube_cpu_gb, img_off_gather_cpu);
      //Need to reallocate the gpu memory for gathers
      copy_cube_c_info(&imgGatherHorzOffset_gb, img_off_gather_cpu);
      alloc_cube_c(&imgGatherHorzOffset_gb,false); //Save memory, only allocate one cube in GPU memory
    }
    //save the memory address for output anyway
    imgCube_cpu_gb.p = img_off_gather_cpu->p;   //Save the memory pointer on CPU as well, for deltaI_(source)
	  imgCube_cpu_gb.p2 = img_off_gather_cpu->p2; //Save the memory pointer on CPU as well, for deltaI_(receiver)
  }
  return bSizeChanged;   
}

//If bTake4thDeriv, then there are w^4 in the operator
extern "C" void gpu2dTomFwd(int *src_iz_cpu, int src_iz_len, float *src_cpu,		//source inject
								  int *rec_iz_cpu, int rec_iz_len, float *rec_cpu, 	 	//receiver inject
								  int prestack_imaging_kernel_type,
								  float *wfld_src, float *wfld_recv, float *wfld_src2, float *wfld_recv2,
								  float *pDeltaS, //Input velocity/slowness perturbation
								  float* pImgS_cpu,float *pImgR_cpu, bool bTake4thDeriv)
{
    //printf("\nGPU2dTomAdj: Beginning 2d Tom Adj on GPU\n");
    /*********************************************************************************************/
	/** Use 2 way acoustic WAVE PROPAGATION  **/
    int nt, nz, nx, wfld_j;
	//retrieve the nt,nz,nx parameters from GPU
  get_ntzx_from_gpu(&nt,&nz,&nx,&wfld_j);
  int imgx = nx - 2*FAT - 2*npad_cpu;
  int imgz = nz - 2*FAT - 2*npad_cpu;
  
	//Copy the data from GPU to CPU
	copy_src_recv_slice_to_gpu( src_iz_cpu, src_iz_len, src_cpu,
							    rec_iz_cpu, rec_iz_len, rec_cpu,
                              nt,nx);
	grid_layouts gdlo;
	gen_grid_layout_for_kernels(&gdlo,nt,nz,nx);

	int indexSnap = 0; int indexImaging = 0;
  float *ptemp = NULL;
	//reset the wavefield.
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU1_gpu, 0,nz*nx*sizeof(float)));
  cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
	zero_wfLists(&wfList_gb); //!IMPROTANT
	
	//memcpy, the input delta S
	cuda_call(cudaMemcpy(reflGpu, pDeltaS, nz*nx*sizeof(float), cudaMemcpyHostToDevice) );
    
  bool bSnapSrc1 = (NULL != wfld_src);
  bool bSnapSrc2 = (NULL != wfld_src2);
  bool bSnapRecv1 = (NULL != wfld_recv);
  bool bSnapRecv2 = (NULL != wfld_recv2);

	
	//!!Assume the refl_gpu and the offsetGather_gb is ready.
	//The extra part for the Tomo operator
	
	//First calculate deltaI_s: 
	//1. Do born modeling using src and the image gathers, remember to save the scattered wavfield delta(wav_SRC) to the list
	//2. Back propagate wavefield and then cross correlate (with HORIZONTAL SHIFT) with the wavefield saved at step I to result in the final Image. Needs to scale S_0(x) in CPU code.
	
	/** Propagate Data forward and save src wavefield on GPU **/
	//at time it, calcuate wfld at time it+2, for the sake of memory pointer switch efficiency
	//The wavfiled p0->it, p1->it+1; same as pU0 and pU1, in order to do the link-list trick, has to arrange this way.
	 for(int it=0; it<nt; it++){
		//source at time 'it' should be applied to wavefield at time 'it+1'
        kernel_exec(inject_kernel<<<gdlo.inject_dimGrid, gdlo.inject_dimBlock>>>(src_iz_gpu, it, source_gpu, p1_gpu));
		
		//Absorbing boundary condition on wavefield at it+1, using wavefield at it(p0_gpu)
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);	
		
		if (0 == it%jImaging_gb) {
		   //Inject the scatter wavefield to it+1, using the slowness perturbation delta(S(x)), eq is
		   //G0(y,x_s,w) deltaS(y) w^2 , 2nd derivative in time
		   kernel_exec(scatter_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(pU1_gpu,dpDt2_gpu,reflGpu));
		}
		
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(pU1_gpu,pU0_gpu));
		bnd_abs_ord1(pU1_gpu, pU0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		
		//repeatively copy the wavefield memory out to host memory for snapShot
		if(it%wfld_j==0)
		{
		   if (bSnapSrc1) {cuda_call(cudaMemcpy(&wfld_src[nz*nx*indexSnap], p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));}
		   if (bSnapSrc2) {cuda_call(cudaMemcpy(&wfld_src2[nz*nx*indexSnap], pU0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));}
		   indexSnap++;
		}
		
		//Prop the src wavefield from it,it+1 -> it+1,it+2
    kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu, dpDt2_gpu));
		ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
		
		//progagate the scattered source wavefield from it,it+1, -> it+2,it
		//Do propagation, same link-list trick, save all the source scattered wavefield
		if(it%jImaging_gb != 0){ 
			  /* Propagate without saving for most steps */
			  kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(pU0_gpu, pU1_gpu, pU0_gpu));			  
			  ptemp=pU0_gpu; pU0_gpu=pU1_gpu;pU1_gpu=ptemp;
		 }else{ 
			  // At imaging timesteps, save wavefield
			  ptemp=wfList_gb.pLists[indexImaging];
			  kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(pU0_gpu, pU1_gpu,ptemp)); //ptemp point to it+2, then switch these three pointers.
			  //swtich the pointers
			  wfList_gb.pLists[indexImaging]=pU0_gpu;
			  pU0_gpu=pU1_gpu;pU1_gpu=ptemp;
			  indexImaging++;
		}		 		
	} //for it=0->nt-1 loop
	//Now we've get through the source part, the scattered source wfld is stored in wfList_gb.
	 
	// Propagate receiver backward imaging correlation against
	// the scattered source wavefields 
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(image_gpu, 0,imgz*imgx*sizeof(float)));
	cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
    
	if (bDoImageGather_gb) { 
	   zero_cube_c(&imgGatherHorzOffset_gb);
	}
	
	indexImaging = 0; indexSnap=0; ptemp=NULL;
	//Calculate it-2 from the wavslice at it and it-1
	//p0->it; p1->it-1; the receiver at time it should be injected to slice it-1.
	for (int it=nt-1; it>=0; it--){
		 //now p0->it, p1->it-1		 
		 //Inject receiver wavefield at it to wav slice at it-1
		 kernel_exec(inject_kernel<<<gdlo.inject_dimGrid,gdlo.inject_dimBlock>>>(recv_iz_gpu,it,recv_gpu, p1_gpu));
		 //update boundary at it-1
		 kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		 bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		 //use p0->it as the wavefield at time it, cross-correlate with the wavefield we save

		 if(it%jImaging_gb==0){
			indexImaging = it / jImaging_gb;
			//Call the imaging kernel, zero lag
            if (!bTake4thDeriv) {
               kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],p0_gpu));
            } else {
               kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],dpDt2_gpu));
            }
			//further create the prestack image, wfld_src(x-h)wfld_recv(x+h)
			if (bDoImageGather_gb) {
			  assert(prestack_imaging_kernel_type == 0);
          if (!bTake4thDeriv) {
            ximgY(imgGatherHorzOffset_gb.p, imgGatherHorzOffset_gb.o, 
                  imgGatherHorzOffset_gb.d, imgGatherHorzOffset_gb.nframe, 
                  wfList_gb.pLists[indexImaging], p0_gpu,prestack_imaging_kernel_type,
                  &gdlo);
          } else {
            ximgY(imgGatherHorzOffset_gb.p, imgGatherHorzOffset_gb.o, 
                  imgGatherHorzOffset_gb.d, imgGatherHorzOffset_gb.nframe, 
                  wfList_gb.pLists[indexImaging], dpDt2_gpu,prestack_imaging_kernel_type,
                  &gdlo);
          }
			  }
		  }
		 
		 //dump the receiver wavefield snapshot
		 if(it%wfld_j== 0) {
			indexSnap = it / wfld_j;
			if (bSnapRecv1) {cuda_call(cudaMemcpy(&(wfld_recv[nz*nx*indexSnap]), p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));}
		}
		
		//back Propagate the receiver wavefield, afterthen p0->it-2, p1->it-1
    if (!bTake4thDeriv) {
       kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
    } else {
       kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,dpDt2_gpu));
    }
		ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp; //switch pointer, then p0->it-1, p1->it-2
	}
	
	//Copy the prestack image back to cpu space
	int nf = imgGatherHorzOffset_gb.nframe;
	cuda_call(cudaMemcpy(pImgS_cpu,image_gpu,imgx*imgz*sizeof(float),cudaMemcpyDeviceToHost));
	cuda_call(cudaMemcpy(imgCube_cpu_gb.p,imgGatherHorzOffset_gb.p,
						 nf*imgx*imgz*sizeof(float),cudaMemcpyDeviceToHost));
	
	////SECOND STEP, CALCULATE deltaI_r // receiver side image pertubation
	indexSnap = 0; indexImaging = 0; ptemp = NULL;
	
	//reset the wavefield.
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
    
	zero_wfLists(&wfList_gb);
	//3. forward propagate src wavefield and save to the wavefield list
	//4. Do adjoint RTM modeling backward in time using receiver data and the image gathers, then cross correlate with the wavefield saved at step 3 to result in the final Image. Needs to scale S_0(x) in CPU code.
	/** Propagate Data forward and save src wavefield on GPU **/
	//at time it, calcuate wfld at time it+2, for the sake of memory pointer switch efficiency
	//The wavfiled p0->it, p1->it+1; in order to do the link-list trick, has to arrange this way.
	for(int it=0; it<nt; it++){

		//source at time 'it' should be applied to wavefield at time 'it+1'
        kernel_exec(inject_kernel<<<gdlo.inject_dimGrid, gdlo.inject_dimBlock>>>(src_iz_gpu, it, source_gpu, p1_gpu));
		
		//Absorbing boundary condition on wavefield at it+1, using wavefield at it(p0_gpu)
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);	
		
		//Do propagation, same link-list trick, save all the source scattered wavefield
		if(it%jImaging_gb != 0){
			  /* Propagate without saving for most steps */
			  kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,dpDt2_gpu));			  
			  ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
		 }else{ 
			  // At imaging timesteps, save wavefield
			  indexImaging = it/jImaging_gb;
              if (!bTake4thDeriv) {
                 ptemp=wfList_gb.pLists[indexImaging];
                 kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu,ptemp)); //ptemp point to it+2, then switch these three pointers.
                 //swtich the pointers
                 wfList_gb.pLists[indexImaging]=p0_gpu;
                 p0_gpu=p1_gpu;p1_gpu=ptemp;
              } else {
                 kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu,p0_gpu));
                 //Save the dp/dt^2 from last time into the wfld lists
                 cuda_call(cudaMemcpy(wfList_gb.pLists[indexImaging],dpDt2_gpu,sizeof(float)*nx*nz,cudaMemcpyDeviceToDevice));
                 //swtich the pointers
                 ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
              }
         }		 		
	} //for it=0->nt-1 loop
	 
	//Now we've get through the source part, source wfld is stored in wfList_gb.
	 
	// Propagate receiver backward imaging correlation against
	// the scattered source wavefields 
	cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU0_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(pU1_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(image_gpu, 0,imgz*imgx*sizeof(float)));
  cuda_call(cudaMemset(dpDt2_gpu, 0,nz*nx*sizeof(float)));
	cuda_call(cudaMemset(imgGatherHorzOffset_gb.p,0,nf*nx*nz*sizeof(float)));
	
	indexSnap = 0; indexImaging = 0; ptemp = NULL;

	
	///step 4: Calculate it-2 from the wavslice at it and it-1, mean while do born modeling backwards.
	//p0->it; p1->it-1; the receiver at time it should be injected to slice it-1.
	for (int it=nt-1; it>=0; it--){
		 //Inject receiver wavefield at it to wav slice at it-1
		 kernel_exec(inject_kernel<<<gdlo.inject_dimGrid,gdlo.inject_dimBlock>>>(recv_iz_gpu,it,recv_gpu, p1_gpu));
		 //update boundary at it-1
		 kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		 bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		
		//Inject to the back scattered wavefield at it-1 to back propagate, using slowness perturbation. 2nd deriv in time
		if (0 == it%jImaging_gb) { 
		   kernel_exec(scatter_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(pU1_gpu,dpDt2_gpu,reflGpu));
		}
		
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(pU1_gpu,pU0_gpu));
		bnd_abs_ord1(pU1_gpu, pU0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);
		//use p0->it as the wavefield at time it, APPLY IMAGING CONDITION cross-correlate with the wavefield we save
		 if(it%jImaging_gb==0){
			indexImaging = it / jImaging_gb; 
			//Call the imaging kernel, zero lag
			kernel_exec(img_kernel<<<gdlo.dimGridImage,gdlo.dimBlock>>>(image_gpu,wfList_gb.pLists[indexImaging],pU0_gpu));			
			//further create the prestack image, wfld_src(x-h)wfld_recv(x+h)
			if (bDoImageGather_gb) {
			   ximgY(imgGatherHorzOffset_gb.p, imgGatherHorzOffset_gb.o, 
					 imgGatherHorzOffset_gb.d, imgGatherHorzOffset_gb.nframe, 
					 wfList_gb.pLists[indexImaging], pU0_gpu,prestack_imaging_kernel_type,
					 &gdlo);
			}
		}
		//dump the receiver wavefield snapshot
		   if(it%wfld_j== 0) {
			  indexSnap = it / wfld_j;
			  if (bSnapRecv2) {cuda_call(cudaMemcpy(&(wfld_recv2[nz*nx*indexSnap]), pU0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));}
		}
		
		//back Prop the recv wavefield from it,it-1 -> it-1,it-2
        kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu, dpDt2_gpu));
		ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
		
		//back Propagate the scattered recv wavefield.
		kernel_exec(wave_kernel<<<gdlo.dimGrid, gdlo.dimBlock>>>(pU0_gpu, pU1_gpu, pU0_gpu));
		ptemp=pU0_gpu; pU0_gpu=pU1_gpu;pU1_gpu=ptemp; 
		//switch pointer, then pU0->it-1, pU1->it-2
	}
	
	//Finish all steps, 	//Copy the image perturbation back to cpu space
	cuda_call(cudaMemcpy(pImgR_cpu,image_gpu,imgx*imgz*sizeof(float),cudaMemcpyDeviceToHost));
	cuda_call(cudaMemcpy(imgCube_cpu_gb.p2,imgGatherHorzOffset_gb.p,
						  nf*imgx*imgz*sizeof(float),cudaMemcpyDeviceToHost));
	return;
}

