program main
  use sep
  use base_wavelt_mod
  implicit none

  integer                             :: nx, ny, nz
  real                                :: dx, dy, dz
  real                                :: ox, oy, oz
  real                                :: x11, x12, x13
  real                                :: x21, x22, x23
  real                                :: x, y, z, depth, dplnz, zref
  real                                :: v0, alpha1,alpha2, beta, decay1, decay2, fm 
  real, dimension(:,:,:), allocatable :: vmod, refl, bvel
  integer                             :: ix, iy, iz, nref, iref
  integer, dimension(:) , allocatable :: plnz


  real 	 :: slope, xc, wt1, wt2
	integer:: ixc

  real :: width1, height1, width2, height2
  integer::	nxwid
  real :: decaySpeed

  call initpar()

  call from_param("nx", nx, 1); call from_param("dx", dx, 1.); call from_param("ox", ox, 0.)
  !call from_param("ny", ny, 1); call from_param("dy", dy, 1.); call from_param("oy", oy, 0.)
  call from_param("nz", nz, 1); call from_param("dz", dz, 1.); call from_param("oz", oz, 0.)

  call from_param("x11", x11, 0.)      ! the x coord of 1st anomaly
  !call from_param("x12", x12, 0.)
  call from_param("x13", x13, 1000.)   ! the z coord of 1st anomaly

  call from_param("width1", width1, 0.5)
  call from_param("width2", width2, width1)
  call from_param("height1", height1,0.5)
  call from_param("height2", height2,height1)

  call from_param("x21", x21, 0.)
  !call from_param("x22", x22, 0.)
  call from_param("x23", x23, 1000.)

  call from_param("alpha1", alpha1, 0.)
  call from_param("alpha2", alpha2, alpha1)

  call from_param("v0", v0, 1000.)
  
	call from_param("xcref",xc,ox+(nx/2)*dx)
  call from_param("zref", zref, 0.)
  call from_param("slope",slope,0.) !refl slope, tan(\alpha)
  call from_param("fm", fm, 0.5/dz)

  ny = 1
  allocate(vmod(nz,nx,ny))
  allocate(refl(nz,nx,ny))
  allocate(bvel(nz,nx,ny))
  !allocate(plnz(nref))

	vmod = v0
  !decaySpeed=2.0;
  decaySpeed=1.0;
  do iz=1, nz; z = oz+(iz-1)*dz
  !do iy=1, ny; y = oy+(iy-1)*dy
  do ix=1, nx; x = ox+(ix-1)*dx
     !decay1 = -36*((x-x11)**2 + (y-x12)**2 + (z-x13)**2)/1000000.
     !decay2 = -36*((x-x21)**2 + (y-x22)**2 + (z-x23)**2)/1000000.
	  decay1 = -( ((x-x11)*2.0/width1)**2  + ((z-x13)*2.0/height1)**2)
	  decay2 = -( ((x-x21)*2.0/width2)**2  + ((z-x23)*2.0/height2)**2)
    
		vmod(iz,ix,1) = v0 * (1.0 - alpha1*exp(decay1) - alpha2*exp(decay2))
  end do
  !end do
  end do

  bvel = v0
  write(0,*) "xc",xc, abs(slope)
  if (abs(slope) .gt. 0.) then
	  nxwid = (oz+(nz-1)*dz-zref) / abs(slope) * 0.8 / dx;
	  write(0,*) "nxwid",nxwid
  else
	  nxwid = nx/2;
  endif

	nxwid = nxwid-1
	ixc = (xc-ox)/dx+1
  do ix=ixc-nxwid, ixc+nxwid
  x = ox + (ix-1)*dx
!  do iz=1, nz
!     z = oz+(iz-1)*dz
!     refl(iz,ix,1) = refl(iz,ix,1) + wavelet(z-zref+(x-xc)*slope, fm, 1)
!	
!  end do
	z 	= ((zref-(x-xc)*slope));
  iz	= (z-oz)/dz + 1; !iz is lower bound
	wt1 = (z-oz)/dz + 1 - iz;
	wt2 = 1.0 - wt1;
	!put two spikes at correct location
	refl(iz,ix,1) = refl(iz,ix,1) + wt2
	refl(iz+1,ix,1) = refl(iz+1,ix,1) + wt1
  end do

  call to_history("n1", nz); call to_history("d1", dz); call to_history("o1", oz)
  call to_history("n2", nx); call to_history("d2", dx); call to_history("o2", ox)
  !call to_history("n3", ny); call to_history("d3", dy); call to_history("o3", oy)

  call to_history("n1", nz, "refl"); call to_history("d1", dz, "refl"); call to_history("o1", oz, "refl")
  call to_history("n2", nx, "refl"); call to_history("d2", dx, "refl"); call to_history("o2", ox, "refl")
 ! call to_history("n3", ny, "refl"); call to_history("d3", dy, "refl"); call to_history("o3", oy, "refl")

  call to_history("n1", nz, "bvel"); call to_history("d1", dz, "bvel"); call to_history("o1", oz, "bvel")
  call to_history("n2", nx, "bvel"); call to_history("d2", dx, "bvel"); call to_history("o2", ox, "bvel")
 ! call to_history("n3", ny, "bvel"); call to_history("d3", dy, "bvel"); call to_history("o3", oy, "bvel")

  call sep_write(vmod)
  call sep_write(refl, "refl")
  call sep_write(bvel, "bvel")
end program
