#include "stdafx.h"

#include "sep_params.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "SR_manager.h"
#include "modelingFDgpu.h"
#include "cudaModeling2d.h"

#include "interp.h"
#include "dampingFunc.h"
#include "cubeAlgo2.h"
using namespace Yang;

//Do Image space WETomo operators
//
//Do the WETom forward, from deltaS -> delta I
void modelingFDgpu2d_op::TomFwd()
{
  FUNC_TRACE_BEG(FDgpu2d_op::TomFwd);
  VERIFY(this->mAction == Yang::acTomFwd);
  gpu2dTomFwd_init(&(m_pars->timeParam.jSnap),&m_gpuID);
  
  prt("\n");
  for (int ishot=0; ishot < m_pars->m_nShots; ishot++)  { //For each shot, it might be super shot
	  prt("\r Start Tomo GPU iShot %d...",ishot);
	  this->initPerShot(ishot);
    //shot or simple point shot. 
	  m_pSR_manager->initPerShot(ishot,Yang::acTomFwd);
	  
	  //Need to interpolate the Receiver record first
    m_pSR_manager->interpRecvRec(ishot);
 
	  this->TomFwd_perShot(ishot);
	  
	  //Image has the size of current vel model, need to be assembled to a final image later.
	  //Save the RTM images to m_pars structures.
	  this->genPerShotImage(ishot);
  } //end loop iShot
  
  //The final image is now at m_pars->pFinalImage, multiply -1
  bool bDoOffset= m_pars->mOffImgParams.bDoOffsetGather;
  if (bDoOffset) {
    SAFE_PTR_CALL(m_pars->pFinalImage,scale(-1.0f));
    m_pars->pFinalOffImage->scale(-1.0f);
  } else {
    m_pars->pFinalImage->scale(-1.0f);
  }
  
  gpu2dTomFwd_uninit();
  FUNC_TRACE_END();
} 

//generate src recv info and return snapshot cube shape, auxilary member functions.
void modelingFDgpu2d_op::setVelocityPtrForGPU(cubeData<float>& vdtSqCube, float dt) {
   m_pars->pVel->getVelCube().copyTo(vdtSqCube);
   VERIFY(vdtSqCube.domain().sameSize(m_pars->modelGridShape)); //Check the size   
   //Get the velocity memory and set to GPU
   vdtSqCube.scale(dt); vdtSqCube.multiply(vdtSqCube);
}

//For each shot, do Tomography Fwd operator
void modelingFDgpu2d_op::TomFwd_perShot(int ishot) {
   FUNC_TRACE_BEG(FDgpu2d_op::TomFwd_pershot)
   mReflCubePerShot.alloc(m_pars->mImageShape); //store the cropped slowness perturbation
   mS0cube.alloc(m_pars->mImageShape);

   //Crop and align the deltaS into mReflCubePerShot for this shot, notice here what we want to supply is actually deltaS X S0
   cropAlign(mReflCubePerShot, *(m_pars->pFinalDeltaS));
   cropAlign(mS0cube, *(m_pars->pVel));
   mReflCubePerShot.multiply(mS0cube); //ds/v0 = ds*s0;
   
   //first set background velocity
   const hypercube& velShape = m_pars->modelGridShape;
   int vel_nx,vel_nz;
   vel_nz = velShape.get_n(1); vel_nx = velShape.get_n(2);

   float dt = m_pars->timeParam.dtComp;
   int nt = m_pars->timeParam.Nt;
   //The velocity cube we sent to GPU is actually v*v*dt*dt
   cubeData<float> vTemp;
   setVelocityPtrForGPU(vTemp,dt);
   //Get the velocity memory and set to GPU
   float* pf_vel = vTemp.dataPtr();
   
   float dx,dz; 
   dz = velShape.get_d(1); dx = velShape.get_d(2);
      
   int fullLen = MAX(vel_nx+2*GPU_STENCIL_WIDTH,vel_nz+2*GPU_STENCIL_WIDTH)/2 + 1;
   stdVecFloat coefVecFull;
   this->getDampCoefVec(coefVecFull, fullLen);
   
   //allocate for delta_Is, delta_Ir   
   mRTM_imageCubePerShot.alloc(m_pars->mShapeWavfield);  
   mRTM_imageCubePerShot2.alloc(m_pars->mShapeWavfield);     

   bool bDoOffSetGather = m_pars->mOffImgParams.bDoOffsetGather;
   cube_c *pImgOffGather = NULL;
   cube_c img_off_gather_cpu;
   
   int imgz = m_pars->mImageShape.get_n(1); int imgx = m_pars->mImageShape.get_n(2);
   //Assume do only on x direction, i,e hx
   if (bDoOffSetGather) {
	  //Then find the correct size of off_img_gather for current shot
	  hypercube offImgShape;
	  get_prestack_img_cube_shape(offImgShape, m_pars->mOffImgParams, m_pars->mImageShape);
	  mRTM_offsetImgCubePerShot.alloc(offImgShape);
	  mRTM_offsetImgCubePerShot2.alloc(offImgShape);
	  
	  img_off_gather_cpu.p = mRTM_offsetImgCubePerShot.dataPtr();
	  img_off_gather_cpu.bUse_p2 = true;
	  img_off_gather_cpu.p2= mRTM_offsetImgCubePerShot2.dataPtr();
	  img_off_gather_cpu.nframe = m_pars->mOffImgParams.nhx;
	  img_off_gather_cpu.o  = m_pars->mOffImgParams.ohx;
	  img_off_gather_cpu.d  = m_pars->mOffImgParams.dhx;
	  img_off_gather_cpu.nz = imgz;
	  img_off_gather_cpu.nx = imgx;
	  pImgOffGather = &img_off_gather_cpu;
   }

   dt = m_pars->timeParam.dtComp;
   gpu2dTomFwd_setVelocity(pf_vel, &vel_nz, &vel_nx, &nt, &dt, &dz, &dx, &coefVecFull[0], &fullLen,&(m_pars->m_nPad), &(m_pars->timeParam.jImaging),pImgOffGather);
   
   //generate src recv info and return snapshot cube shape
   cubeData<float> srcSliceCube, recvSliceCube;
   cubeData<int> srcIzVec, recvIzVec; //notice it is integer elements.
   hypercube shapeWfldSnapshot;
   //prepare the injected source, it is a cubeslice with the size nt by vel_nz.
   m_pSR_manager->genSrcSlice(ishot,srcSliceCube,srcIzVec);
   m_pSR_manager->genRecvSlice(ishot,recvSliceCube,recvIzVec);

   //take 2nd derivative is not as good as take 2nd derivative on the src wfld
   //cubeData<float> recv2SliceCube; recv2SliceCube.alloc(recvSliceCube.domain());
   //timeDeriv2nd(recv2SliceCube,recvSliceCube,true);
   
   //if (m_pars->pSrcSnapshot) {
     this->getSnapshotShapePerShot(shapeWfldSnapshot);   
     mSrcSnapshot.alloc(shapeWfldSnapshot); mRecvSnapshot.alloc(shapeWfldSnapshot);
     mAdjRTM_srcSnapshot.alloc(shapeWfldSnapshot); mAdjRTM_recvSnapshot.alloc(shapeWfldSnapshot);
   //}
   //Call the Tomography Forward operator
   gpu2dTomFwd(srcIzVec.dataPtr(),  vel_nx, srcSliceCube.dataPtr(),recvIzVec.dataPtr(), vel_nx, recvSliceCube.dataPtr(),
			   m_pars->offset_imaging_gpu_kernel_type,
			   mSrcSnapshot.dataPtr(), mRecvSnapshot.dataPtr(), mAdjRTM_srcSnapshot.dataPtr(), mAdjRTM_recvSnapshot.dataPtr(), //wavfield snapshot
			   mReflCubePerShot.dataPtr(), //input deltaS is here
			   mRTM_imageCubePerShot.dataPtr(),mRTM_imageCubePerShot2.dataPtr(),m_pars->bTake2ndTimeDerivForRTM);
   
   //Add the src and recv image part together
   mRTM_imageCubePerShot.add(mRTM_imageCubePerShot2);
   if (bDoOffSetGather) {
     mRTM_offsetImgCubePerShot.add(mRTM_offsetImgCubePerShot2);
   }

   recordSnapshotGpu(&mSrcSnapshot, &mRecvSnapshot,ishot);   
   recordSnapshotGpu2(&mAdjRTM_srcSnapshot, &mAdjRTM_recvSnapshot,ishot);   

   //seperr("stop here\n");
   FUNC_TRACE_END();
}

//Do the Adjoint WETomo operator, Input deltaI/deltaIh, output deltaS
void modelingFDgpu2d_op::TomAdj()
{
   FUNC_TRACE_BEG(FDgpu2d_op::TomAdj);
   VERIFY(this->mAction == Yang::acTomAdj);
   
   gpu2dTomAdj_init(&(m_pars->timeParam.jSnap),&m_gpuID);
   
   //calculate deltaS for each shot
   prt("\n Start GPU tomo Adj (vel backProj) operator \n");
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++) {
	  prt("\r Start Tomo Adj iShot = %d", ishot);
	  this->initPerShot(ishot);
      m_pSR_manager->initPerShot(ishot,Yang::acTomAdj);
	  //Need to interpolate the Receiver record first
      m_pSR_manager->interpRecvRec(ishot);
	  //m_pSR_manager->mInterpSeisCube is still not the size of vel_nx*vel_nz

	  this->TomAdj_perShot(ishot);

	  this->genPerShotImage(ishot);
   }
   
   //Then multiplies s0 then we get deltaS
   m_pars->pFinalDeltaS->divide(*(m_pars->pVel));

   gpu2dTomAdj_uninit();
   FUNC_TRACE_END();
}

//For each shot, do Tomography Fwd operator, from deltaI --> deltaS
void modelingFDgpu2d_op::TomAdj_perShot(int ishot) {
   FUNC_TRACE_BEG(FDgpu2d_op::TomAdj_pershot);
   //first set background velocity
   const hypercube& velShape = m_pars->modelGridShape;
   int vel_nx,vel_nz;
   vel_nz = velShape.get_n(1); vel_nx = velShape.get_n(2);
   float dt = m_pars->timeParam.dtComp; int nt = m_pars->timeParam.Nt;
   //The velocity cube we sent to GPU is actually v*v*dt*dt
   cubeData<float> vTemp;
   setVelocityPtrForGPU(vTemp,dt);
   //Get the velocity memory and set to GPU
   float* pf_vel = vTemp.dataPtr();
   
   float dx,dz; 
   dz = velShape.get_d(1); dx = velShape.get_d(2);
      
   int fullLen = MAX(vel_nx+2*GPU_STENCIL_WIDTH,vel_nz+2*GPU_STENCIL_WIDTH)/2 + 1;
   stdVecFloat coefVecFull;
   this->getDampCoefVec(coefVecFull, fullLen);
   
   //allocate for deltaSs, deltaSr  
   mRTM_imageCubePerShot.alloc(m_pars->mImageShape);  
   mRTM_imageCubePerShot2.alloc(m_pars->mImageShape);     

   int nx,nz;
   nz = m_pars->mShapeWavfield.get_n(1); nx = m_pars->mShapeWavfield.get_n(2);
  
   //Crop and align the deltaI(deltaIh) into mReflCubePerShot/mReflOffCubePerShot for this shot
   bool bDoOffImg = m_pars->mOffImgParams.bDoOffsetGather;
   if (!bDoOffImg) {
	  mReflCubePerShot.alloc(m_pars->mImageShape);
    cropAlign(mReflCubePerShot,*(m_pars->pFinalImage)); //These two function does same thing
	  //alignCropToSliceShapeGrid(mReflCubePerShot, *(m_pars->pFinalImage), mReflCubePerShot.domain());
    //multiply -1, it is in the operator
    mReflCubePerShot.scale(-1.0f);
   } else {
	  hypercube offImgShape;
	  get_prestack_img_cube_shape(offImgShape, m_pars->mOffImgParams, m_pars->mImageShape);
	  mReflOffCubePerShot.alloc(offImgShape);
	  alignCropToSliceShapeGrid(mReflOffCubePerShot, *(m_pars->pFinalOffImage), m_pars->mImageShape);
    mReflOffCubePerShot.scale(-1.0f);
   }
   
  cube_c *pDeltaIoffGather = NULL; float* pDeltaI = NULL;
  cube_c img_off_gather_cpu;
  int imgx,imgz;
  imgz = m_pars->mImageShape.get_n(1); imgx = m_pars->mImageShape.get_n(2);
  
  //Assume do only on x direction, i,e hx
  if (bDoOffImg) {
    img_off_gather_cpu.p = mReflOffCubePerShot.dataPtr();
    img_off_gather_cpu.bUse_p2 = false;
    img_off_gather_cpu.nframe = m_pars->mOffImgParams.nhx;
    img_off_gather_cpu.o  = m_pars->mOffImgParams.ohx;
    img_off_gather_cpu.d  = m_pars->mOffImgParams.dhx;
    img_off_gather_cpu.nz = imgz;
    img_off_gather_cpu.nx = imgx;
    pDeltaIoffGather = &img_off_gather_cpu;
  } else {
    pDeltaI = mReflCubePerShot.dataPtr();
  }
  
  //mReflOffCubePerShot.info("input Delta Ih"); mReflOffCubePerShot.dump("dump10");
  
  //LOG(begin gpu2dTomAdj_setVel);
  gpu2dTomAdj_setVelocity(pf_vel, &vel_nz, &vel_nx, &nt, &dt, &dz, &dx, &coefVecFull[0], &fullLen,&(m_pars->m_nPad), &(m_pars->timeParam.jImaging),
						   pDeltaI, pDeltaIoffGather);
   
   //generate src recv info and return snapshot cube shape
   cubeData<float> srcSliceCube, recvSliceCube;
   cubeData<int> srcIzVec, recvIzVec; //notice it is integer elements.
   //prepare the injected source, it is a cubeslice with the size nt by vel_nz.
   m_pSR_manager->genSrcSlice(ishot,srcSliceCube,srcIzVec);
   m_pSR_manager->genRecvSlice(ishot,recvSliceCube,recvIzVec);

   //Not good take 2nd time derive here.
   //cubeData<float> recv2SliceCube; recv2SliceCube.alloc(recvSliceCube.domain());
   //timeDeriv2nd(recv2SliceCube,recvSliceCube,true);

   float *p1,*p2,*p3,*p4;
   p1 = p2 = p3 = p4 = NULL;
   hypercube shapeWfldSnapshot;
   if (m_pars->bSaveWfldSnap) {
     this->getSnapshotShapePerShot(shapeWfldSnapshot);
     mSrcSnapshot.alloc(shapeWfldSnapshot); mRecvSnapshot.alloc(shapeWfldSnapshot);
     mAdjRTM_srcSnapshot.alloc(shapeWfldSnapshot); mAdjRTM_recvSnapshot.alloc(shapeWfldSnapshot);
     p1 = mSrcSnapshot.dataPtr();
     p2 = mRecvSnapshot.dataPtr();
     p3 = mAdjRTM_srcSnapshot.dataPtr();
     p4 = mAdjRTM_recvSnapshot.dataPtr();
   }

   //LOG(begin gpu2d Tom Adj);
   //Call the Tomography Forward operator
   gpu2dTomAdj(srcIzVec.dataPtr(),  vel_nx, srcSliceCube.dataPtr(),
			   recvIzVec.dataPtr(), vel_nx, recvSliceCube.dataPtr(),
			   p1,p2,p3,p4,
			   mRTM_imageCubePerShot.dataPtr(),mRTM_imageCubePerShot2.dataPtr(),m_pars->bTake2ndTimeDerivForRTM);

   //LOGC(Add the src and recv image part together);
   mRTM_imageCubePerShot.add(mRTM_imageCubePerShot2);
   //mReflOffCubePerShot.dump("dump5"); mRTM_imageCubePerShot.dump("dump6");
   //mAdjRTM_srcSnapshot.dump("dump6");mAdjRTM_recvSnapshot.dump("dump7");

   //Then save the src/recv snapshots we have here.
   if (m_pars->bSaveWfldSnap) {
     recordSnapshotGpu(&mSrcSnapshot, &mRecvSnapshot,ishot);   
     recordSnapshotGpu2(&mAdjRTM_srcSnapshot, &mAdjRTM_recvSnapshot,ishot);   
   }
   FUNC_TRACE_END();
}

