#include "stdafx.h"

#include "sep_params.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "SR_manager.h"
#include "modelingFDgpu.h"
#include "cudaModeling2d.h"

#include "interp.h"
#include "dampingFunc.h"

#include "dmRsfDump.h"

#include "cubeAlgo2.h"

using namespace Yang;

void modelingFDgpu_op::initOperator(waveOpParams* params, action_type ac_type) {
   FUNC_TRACE_BEG(FDgpu_op::initOp)
   modelingFDgen_op::initOperator(params,ac_type);
   VERIFY(params->offx == GPU_STENCIL_WIDTH);    VERIFY(params->offz == GPU_STENCIL_WIDTH);
   FUNC_TRACE_END()
}

void modelingFDgpu_op::unInitOperator() {
   FUNC_TRACE_BEG(modelingFDgpu_op::unInit)
   
   FUNC_TRACE_END()
}

void modelingFDgpu_op::setGPU_ID(int id) {
   //should Test if there is actually enough devices, !!!
   m_gpuID = id;
   VERIFYC(id<=3 && id>=0);
}

/// The damp coef is arranged in this manner:
/// starts from the most significant negative value which corresponds to the damping coef at the leftmost grid pt, 
/// then slightly changing to near zero. 

/*
void modelingFDgpu_op::getDampCoefVec(stdVecFloat& coefVecFull, int fullLen) {
   //Get the damping function

   //cosDampingFunc dampFunc;
   lineDampingFunc dampFunc;
   float extent = 0.1f; int length = m_pars->m_nPad;
   stdVecFloat coefVec;
   dampFunc.getCoef(coefVec, length, extent);
   std::reverse(coefVec.begin(),coefVec.end());
   
   coefVecFull.clear(); coefVecFull.resize(fullLen,-0.0f);
   
   for (size_t i=0; i<coefVec.size();i++) {
	  //coefVec[i] = logf(coefVec[i]) * 0.2f;
	  coefVec[i] += -1.0f; coefVec[i] *= m_dampingCoef;
	  coefVecFull[i] = coefVec[i];
   }
   return;
}
***/

//Not very useful
void modelingFDgpu_op::getDampCoefVec(stdVecFloat& coefVecFull, int fullLen) {
   //Get the damping function

   //cosDampingFunc dampFunc;
   lineDampingFunc dampFunc;
   float extent = 0.1f; int length = m_pars->m_nPad/2;
   stdVecFloat coefVec;
   dampFunc.getCoef(coefVec, length, extent);
   std::reverse(coefVec.begin(),coefVec.end());
   
   coefVecFull.clear(); coefVecFull.resize(fullLen,-0.0f);
   
   for (size_t i=0; i<coefVec.size();i++) {
	  //coefVec[i] = logf(coefVec[i]) * 0.2f;
	  coefVec[i] += -1.0f; coefVec[i] *= m_dampingCoef;
	  coefVecFull[i] = coefVec[i];
	  coefVecFull[2*length-1-i] = coefVec[i];
   }
   return;
}

//For each shot, it might be super shot or simple point shot. 
void modelingFDgpu2d_op::modelingPerShot(int ishot) {
   FUNC_TRACE_BEG(FDgpu2_op::modelingPershot)
   //first set velocity
   const hypercube& velShape = m_pars->modelGridShape;
   int vel_nx,vel_nz;
   vel_nz = velShape.get_n(1); vel_nx = velShape.get_n(2);
   
   //Get the velocity memory and set to GPU
   //The velocity cube we sent to GPU is actually v*v*dt*dt
   cubeData<float> vTemp;
   m_pars->pVel->getVelCube().copyTo(vTemp);
   VERIFY(vTemp.domain().sameSize(m_pars->modelGridShape)); //Check the size   
	  
   int nt = m_pars->timeParam.Nt;
   float dt = m_pars->timeParam.dtComp;
   vTemp.scale(dt); vTemp.multiply(vTemp);

   float dx,dz;
   dz = velShape.get_d(1); dx = velShape.get_d(2);
   
   int fullLen = MAX(vel_nx,vel_nz)/2 + 1 + GPU_STENCIL_WIDTH;
   stdVecFloat coefVecFull;
   this->getDampCoefVec(coefVecFull, fullLen);
   //PRVEC(coefVecFull,40);
   
   int nRecv, ntRec;
   ntRec = m_pars->timeParam.nRec;
   stdVecInt ix_recv_vec, iz_recv_vec;
   
   SR_managerFD2d* pSR_manager2D = dynamic_cast<SR_managerFD2d*>(m_pSR_manager);
   assert(pSR_manager2D!=NULL);
   
   cubeData<float> cubeRecordBuffer;
   pSR_manager2D->genRecGeometryInfo(ishot, nRecv, iz_recv_vec, ix_recv_vec, cubeRecordBuffer);
   
   //char p=','; cerr<<endl<<vel_nz<<p<<vel_nx<<p<<nt<<p<<dt<<p<<dz<<p<<dx<<p<<fullLen<<p<<nRecv<<p<<ntRec<<endl;
   gpu2dModeling_setVelocity(vTemp.dataPtr(), &vel_nz, &vel_nx, &nt, &dt, &dz, &dx, &coefVecFull[0], &fullLen, &(m_pars->m_nPad), 
							 nRecv, ntRec, &ix_recv_vec[0],&iz_recv_vec[0] );
   
   cubeData<float> srcSliceCube; hypercube sliceShape(nt,vel_nx);
   srcSliceCube.alloc(sliceShape);
   
   cubeData<int> izVec; hypercube shapeIzVec(vel_nx); //notice it is integer elements.
   izVec.alloc(shapeIzVec);
   
   //prepare the injected source, it is a cubeslice with the size nt by vel_nz.
   m_pSR_manager->genSrcSlice(ishot, srcSliceCube, izVec);
   
   vector<axis> shapeSnapshotAxes = (m_pars->mShapeWavfield).get_axes();
   float o = 0.0; float d = m_pars->timeParam.jSnap*dt; int n = m_pars->timeParam.nSnap;
   axis axisT(n,o,d,"time","sec");
   shapeSnapshotAxes.push_back(axisT);
   mSrcSnapshot.alloc(hypercube(shapeSnapshotAxes));
   //Call the wave propagation program.
   gpu2d_wave_prop(izVec.dataPtr(),vel_nx,srcSliceCube.dataPtr(), mSrcSnapshot.dataPtr(), 
                   cubeRecordBuffer.dataPtr());

   pSR_manager2D->recordFromGpuBuffer(ishot,cubeRecordBuffer);

   recordSnapshotGpu(&mSrcSnapshot,NULL,ishot);

   //Output one shot record to file, here Recv geometry might be irregular
   m_pSR_manager->finishOneShot(ishot);
   FUNC_TRACE_END();
}

//Do the wave-equation modeling, non-linear, output to pRsfSeisRec
void modelingFDgpu2d_op::modeling()
{
   FUNC_TRACE_BEG(FDgpu2d_op::modeling)   
   //Initialize GPU modeling
   verbPrt("\nStart GPU modeling..");
   VERIFY(this->mAction == Yang::acModeling);
   
   gpu2dModeling_init(&(m_pars->timeParam.jSnap),&(m_pars->timeParam.jRec), &m_gpuID);
   
   //do the modeling using GPU 
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++) {
	   this->initPerShot(ishot);
     m_pSR_manager->initPerShot(ishot,Yang::acModeling);

	   verbPrt("\r GPU modeling ishot=%d...", ishot);
	   this->modelingPerShot(ishot);
   }
   
   mSrcSnapshot.dealloc();
   //The data and snap shot should be ready by now.
   gpuModeling2d_uninit();
   FUNC_TRACE_END()
}

//Do RTM per shot
//The operator take seisRecord as input, output a Image
void modelingFDgpu2d_op::RTM() {
   FUNC_TRACE_BEG(FDgpu2d_op::RTM)
   verbPrt("\n Start GPU RTM ...");
   VERIFY(this->mAction == Yang::acMigration);
   gpu2dRTM_init(&(m_pars->timeParam.jSnap),&m_gpuID);

   for (int ishot=0; ishot < m_pars->m_nShots; ishot++)  { //For each shot, it might be super shot
	  prt("\r Start RTM iShot %d...",ishot);
	  
	  this->initPerShot(ishot);
    //shot or simple point shot. 
	  m_pSR_manager->initPerShot(ishot,Yang::acMigration);
	  
	  //Need to interpolate the Receiver record first
    m_pSR_manager->interpRecvRec(ishot);
	  //Note after interpolation, m_pSR_manager->mInterpSeisCube is still not the size of vel_nx*vel_nz
		 
	  this->RTM_perShot(ishot);
    
	  //Image has the size of current vel model, need to be assembled to a final image later.
	  //Save the RTM images
	  this->genPerShotImage(ishot);
   } //end loop iShot
   //The final image is now at m_pars->pFinalImage, same size as wavefield.
   
   gpuRTM2d_uninit();
   FUNC_TRACE_END()
} 

//For each shot, do RTM //Use GPU
void modelingFDgpu2d_op::RTM_perShot(int ishot) {
   FUNC_TRACE_BEG(FDgpu2d_op::RTM-pershot);

   //first set velocity
   const hypercube& velShape = m_pars->modelGridShape;
   int vel_nx,vel_nz;
   vel_nz = velShape.get_n(1); vel_nx = velShape.get_n(2);
   int nx,nz;
   nx = m_pars->mShapeWavfield.get_n(2);  nz = m_pars->mShapeWavfield.get_n(1);
   float dt = m_pars->timeParam.dtComp;
   int nt = m_pars->timeParam.Nt;
   //The velocity cube we sent to GPU is actually v*v*dt*dt
   cubeData<float> vTemp;
   setVelocityPtrForGPU(vTemp,dt);
   //Get the velocity memory and set to GPU
   float* pf_vel = vTemp.dataPtr();
   
   float dx,dz; dz = velShape.get_d(1); dx = velShape.get_d(2);
   
   int fullLen = MAX(vel_nx+2*GPU_STENCIL_WIDTH,vel_nz+2*GPU_STENCIL_WIDTH)/2 + 1;
   stdVecFloat coefVecFull;
   this->getDampCoefVec(coefVecFull, fullLen);
   
   mRTM_imageCubePerShot.alloc(m_pars->mImageShape);
   bool bDoImgGather = m_pars->mOffImgParams.bDoOffsetGather;
   cube_c *pImageOffGather = NULL; cube_c img_off_gather_cpu;
   if (bDoImgGather) { 
	  hypercube offImgShape;
	  get_prestack_img_cube_shape(offImgShape,m_pars->mOffImgParams, m_pars->mImageShape);
	  mRTM_offsetImgCubePerShot.alloc(offImgShape);
	  img_off_gather_cpu.p = mRTM_offsetImgCubePerShot.dataPtr();
	  img_off_gather_cpu.bUse_p2 = false;
	  img_off_gather_cpu.nframe = m_pars->mOffImgParams.nhx;
	  img_off_gather_cpu.o  = m_pars->mOffImgParams.ohx;
	  img_off_gather_cpu.d  = m_pars->mOffImgParams.dhx;
	  img_off_gather_cpu.nz = m_pars->mImageShape.get_n(1);
	  img_off_gather_cpu.nx = m_pars->mImageShape.get_n(2);
	  pImageOffGather = &img_off_gather_cpu;
   }
   
   //PRV(bDoImgGather);
   //mRTM_offsetImgCubePerShot.info("offestImgPershot");
   
   //Set src/recv intensity
   float *pSrcIntense = NULL, *pRecvIntense = NULL;
   if (m_pars->bDoSrcInt) {
     mSrcIntensePerShot.alloc(m_pars->mImageShape);
     pSrcIntense = mSrcIntensePerShot.dataPtr();
   }
   if (m_pars->bDoRecvInt) {
     mRecvIntensePerShot.alloc(m_pars->mImageShape);
     pRecvIntense = mRecvIntensePerShot.dataPtr();
   }
   
   //m_pars->pVel->dump("dump3");
   //vTemp.info("vdt");
   //prt("\n %x,%d,%d,%d,%f,%f,%f, %d,%d,%d", pf_vel,vel_nz,vel_nx,nt,dt,dz,dx, fullLen, m_pars->m_nPad, (m_pars->timeParam.jImaging));
   //LOG(call gpu RTM setVel);
   gpu2dRTM_setVelocity(pf_vel, &vel_nz, &vel_nx, &nt, &dt, &dz, &dx, &coefVecFull[0], &fullLen,
						&(m_pars->m_nPad), &(m_pars->timeParam.jImaging), pImageOffGather, pSrcIntense,pRecvIntense);
      
   cubeData<float> srcSliceCube,recvSliceCube;
   cubeData<int> srcIzVec,recvIzVec; //notice it is integer elements.
   hypercube shapeWfldSnapshot;
   //prepare the injected source, it is a cubeslice with the size nt by vel_nz.
   m_pSR_manager->genSrcSlice(ishot,srcSliceCube,srcIzVec);
   m_pSR_manager->genRecvSlice(ishot,recvSliceCube,recvIzVec);
   this->getSnapshotShapePerShot(shapeWfldSnapshot);
   mSrcSnapshot.alloc(shapeWfldSnapshot);mRecvSnapshot.alloc(shapeWfldSnapshot);
   
   //shapeWfldSnapshot.info("shapewfldSnapshot");
   //srcIzVec.info("srcIz"); recvIzVec.info("recvIzVec");
   //srcSliceCube.info("ss"); recvSliceCube.info("rs");
   //srcSliceCube.dump("dump3");  recvSliceCube.dump("dump4");

   //LOG(call gpu RTM Imaging); 
   gpu2dRTM_imaging(srcIzVec.dataPtr(),  vel_nx, srcSliceCube.dataPtr(),		//source inject
						 recvIzVec.dataPtr(), vel_nx, recvSliceCube.dataPtr(), 	 	//receiver inject
						 mSrcSnapshot.dataPtr(), mRecvSnapshot.dataPtr(),
						 mRTM_imageCubePerShot.dataPtr(),m_pars->bTake2ndTimeDerivForRTM);
   
   //mSrcIntensePerShot.dump("dump3"); m_pSR_manager->mInterpSeisCube.dump("dump6");
   //mRecvSnapshot.dump("dump8"); 
   //mRTM_imageCubePerShot.dump("dump7"); mRTM_offsetImgCubePerShot.dump("dump6");
   //LOG(finish gpu RTM imaging);
   
   //Balance the gradient amplitude using the src Intensity
   float qt_mask = 0.05;   
   cubeData<float> mask;
   if (m_pars->bDoSrcInt) {
     //Take square root
     sqrtCube(mSrcIntensePerShot,mSrcIntensePerShot);
     divideWithCaveat(mRTM_imageCubePerShot,mRTM_imageCubePerShot,mSrcIntensePerShot,qt_mask,0,true,&mask);
   }
   if (m_pars->bDoRecvInt) {
     sqrtCube(mRecvIntensePerShot,mRecvIntensePerShot);
     divideWithCaveat(mRTM_imageCubePerShot,mRTM_imageCubePerShot,mRecvIntensePerShot,qt_mask,0,true,&mask);
   }

   //mSrcIntensePerShot.dump("dump4");mRecvIntensePerShot.dump("dump3");
   //mask.dump("dump8");mRTM_imageCubePerShot.dump("dump9");
   
   //Then save the src/recv snapshots we have here.
   this->recordSnapshotGpu(&mSrcSnapshot, &mRecvSnapshot, ishot);
   FUNC_TRACE_END()
}

void modelingFDgpu2d_op::recordSnapshotGpu(cubeDataF* pSrcSnap,cubeDataF* pRecvSnap, int ishot) {
   FUNC_TRACE_BEG(FDgpu2d_op::recordSnapshotGpu)
   cubeData<float> wavfieldCube(m_pars->mShapeWavfield);
   bool bRecordSnap = shouldRecordSnap(ishot);
   if (bRecordSnap) {
	  for (int iFrame=0; iFrame < m_pars->timeParam.nSnap; iFrame++)
	  {
      if (m_pars->pSrcSnapshot) {
        pSrcSnap->seq_grab(&wavfieldCube);
        m_pars->pSrcSnapshot->seq_update(&wavfieldCube);
      }
		  if (NULL != pRecvSnap && NULL != m_pars->pRecvSnapshot) {
        pRecvSnap->seq_grab(&wavfieldCube);
        m_pars->pRecvSnapshot->seq_update(&wavfieldCube);
      }
	  }
  }
  FUNC_TRACE_END()
}

void modelingFDgpu2d_op::recordSnapshotGpu2(cubeDataF* pSrc2Snap,cubeDataF* pRecv2Snap, int ishot) {
   cubeData<float> wavfieldCube(m_pars->mShapeWavfield);
   bool bRecordSnap = shouldRecordSnap(ishot);
   if (bRecordSnap) {
	  for (int iFrame=0; iFrame < m_pars->timeParam.nSnap; iFrame++)
	  {
      if (m_pars->pSrcSnapshot2) {
        pSrc2Snap->seq_grab(&wavfieldCube);
        m_pars->pSrcSnapshot2->seq_update(&wavfieldCube);
      }
      if (NULL != pRecv2Snap && NULL != m_pars->pRecvSnapshot2) {
			pRecv2Snap->seq_grab(&wavfieldCube);
			m_pars->pRecvSnapshot2->seq_update(&wavfieldCube);
		 }
	  }
   }
}

//Linear Born Modeling
//Do the wave-equation modeling, non-linear, output to pRsfSeisRec
void modelingFDgpu2d_op::adjRTM()
{
   FUNC_TRACE_BEG(FDgpu2d_op::adjRTM)
   //set the domain for SeisRec and output domain info to file, already done in initOperator

   //Initialize GPU modeling
   verbPrt("\nStart GPU Adj RTM modeling..");
   VERIFY(this->mAction == Yang::acBornModeling);
   
   gpu2dAdjRTM_init(&(m_pars->timeParam.jSnap),&(m_pars->timeParam.jRec),&m_gpuID);
   

   //do the modeling using GPU
   for (int ishot=0; ishot < m_pars->m_nShots; ishot++) {
	  this->initPerShot(ishot);
      m_pSR_manager->initPerShot(ishot,Yang::acBornModeling);
	  
	  verbPrt("\r GPU modeling ishot=%d...", ishot);
	  this->adjRTM_perShot(ishot);
   }
   
   mSrcSnapshot.dealloc();
   mAdjRTM_srcSnapshot.dealloc();
   
   //The data and snap shot should be ready by now.
   gpu2dAdjRTM_uninit();
   FUNC_TRACE_END()
}

//Do the AdjRTM
//For each shot, it might be super shot or simple point shot. 
void modelingFDgpu2d_op::adjRTM_perShot(int ishot) {
   FUNC_TRACE_BEG(FDgpu2d_op::adjRTM_pershot)
   //first set velocity
   const hypercube& velShape = m_pars->modelGridShape;
   int vel_nx,vel_nz;
   vel_nz = velShape.get_n(1); vel_nx = velShape.get_n(2);
   int imgx,imgz;
   imgx = m_pars->mImageShape.get_n(2); imgz = m_pars->mImageShape.get_n(1);
   
   //Get the velocity memory and set to GPU
   float dt = m_pars->timeParam.dtComp;
   int nt = m_pars->timeParam.Nt;
   //The velocity cube we sent to GPU is actually v*v*dt*dt
   cubeData<float> vTemp;
   setVelocityPtrForGPU(vTemp,dt);
   //Get the velocity memory and set to GPU
   float* pf_vel = vTemp.dataPtr();
   float dx,dz;
   dz = velShape.get_d(1); dx = velShape.get_d(2);
   
   int fullLen = MAX(vel_nx,vel_nz)/2 + 1 + GPU_STENCIL_WIDTH;
   stdVecFloat coefVecFull;
   this->getDampCoefVec(coefVecFull, fullLen);
   //PRVEC(coefVecFull,40);

   //Align the global refl model to the current imageShape
   bool bDoImgGather = m_pars->mOffImgParams.bDoOffsetGather;
  if (!bDoImgGather) { //Post stack image
	  mReflCubePerShot.alloc(m_pars->mImageShape);
	  cropAlign(mReflCubePerShot,*(m_pars->pFinalImage));
    
    //mReflCubePerShot.dump("dump3"); mReflCubePerShot.info("mReflCubePerShot"); m_pars->pFinalImage->dump("dump4");
  } else { //prestack image
	  hypercube offImgShape;
	  get_prestack_img_cube_shape(offImgShape, m_pars->mOffImgParams, m_pars->mImageShape);
	  mReflOffCubePerShot.alloc(offImgShape);
	  alignCropToSliceShapeGrid(mReflOffCubePerShot,*(m_pars->pFinalImage),m_pars->mImageShape);
  }
   cube_c *pReflOff = NULL; cube_c reflOffModel;
   float* pRefl = NULL;
   //Assume do only on x direction, i,e hx
   if (bDoImgGather) {
	  reflOffModel.p = mReflOffCubePerShot.dataPtr();
	  reflOffModel.bUse_p2 = false;
	  reflOffModel.nframe = m_pars->mOffImgParams.nhx;
	  reflOffModel.o  = m_pars->mOffImgParams.ohx;
	  reflOffModel.d  = m_pars->mOffImgParams.dhx;
	  reflOffModel.nz = imgz;
	  reflOffModel.nx = imgx;
	  pReflOff = &reflOffModel;
   } else {
	  pRefl = mReflCubePerShot.dataPtr();   
   }

   int nRecv, ntRec;
   ntRec = m_pars->timeParam.nRec;
   stdVecInt ix_recv_vec, iz_recv_vec;
   SR_managerFD2d* pSR_manager2D = dynamic_cast<SR_managerFD2d*>(m_pSR_manager);
   assert(pSR_manager2D!=NULL);
   
   cubeData<float> cubeRecordBuffer;
   pSR_manager2D->genRecGeometryInfo(ishot, nRecv, iz_recv_vec, ix_recv_vec, cubeRecordBuffer);
   
   gpu2dAdjRTM_setVelocityAndRefl(pf_vel, pRefl, pReflOff, 
								  &vel_nz, &vel_nx, &nt, &dt, &dz, &dx, 
								  &coefVecFull[0], &fullLen, &(m_pars->m_nPad),&(m_pars->timeParam.jImaging),
								  nRecv, ntRec, &ix_recv_vec[0],&iz_recv_vec[0] );
   
   //prepare the injected source, it is a cubeslice with the size nt by vel_nz.
   cubeData<float> srcSliceCube; cubeData<int> izVec;
   m_pSR_manager->genSrcSlice(ishot, srcSliceCube, izVec);
   //PRV(izVec[120]);
   hypercube shapeSnapShot;
   this->getSnapshotShapePerShot(shapeSnapShot);   
   mSrcSnapshot.alloc(shapeSnapShot); mAdjRTM_srcSnapshot.alloc(shapeSnapShot);
   
   //Call the wave propagation program.
   gpu2dAdjRTM(izVec.dataPtr(),vel_nx,srcSliceCube.dataPtr(), mSrcSnapshot.dataPtr(), mAdjRTM_srcSnapshot.dataPtr(),
				cubeRecordBuffer.dataPtr());
   
   pSR_manager2D->recordFromGpuBuffer(ishot,cubeRecordBuffer);
   
   //take snapshot, this part should always combined with propagator
   this->recordSnapshotGpu(&mSrcSnapshot,NULL,ishot);
   this->recordSnapshotGpu2(&mAdjRTM_srcSnapshot, NULL, ishot);
   
   //Output one shot record to file, here Recv geometry might be irregular
   m_pSR_manager->finishOneShot(ishot);
   FUNC_TRACE_END()
}

