#ifndef __MODELINGFDGPU_H_INCL__
#define __MODELINGFDGPU_H_INCL__

#include "op.h"
#include "hypercube.h"
#include "cubeData.hpp"
#include "waveOpParams.h"
#include "waveOps.h"

namespace Yang {
class vel;
//Forward declaration of template class
//template <typename T>
//class dmRsf_oc;
class modelingFDgpu_op : public modelingFDgen_op
{
public:
   modelingFDgpu_op(): modelingFDgen_op(), m_gpuID(0) {}    
   virtual ~modelingFDgpu_op()	{}
   
   //Get the parameters for wave propagation operator. e.g. dt,dx,freq(center freq of wavlet),
   virtual void initOperator(waveOpParams* params, action_type ac_type);
   virtual void unInitOperator();
   virtual void modeling() {}
   virtual void RTM() {}
   virtual void adjRTM(){}
   
   virtual void modelingPerShot(int ishot) = 0;
   virtual void RTM_perShot(int ishot) = 0;
   virtual void adjRTM_perShot(int ishot) = 0;

   void setGPU_ID(int id);
protected:
   void getDampCoefVec(stdVecFloat& coefVec, int fullLen);
   
public:
   //Per shot
   cubeData<float> mSrcSnapshot; 
   cubeData<float> mRecvSnapshot; //Per shot
   cubeData<float> mAdjRTM_srcSnapshot;
   cubeData<float> mAdjRTM_recvSnapshot;
   
public:
   float m_dampingCoef;
   
protected:
   int m_gpuID;
   
};

class modelingFDgpu2d_op: public modelingFDgpu_op {
public:
   virtual void modeling();
   virtual void RTM();
   virtual void adjRTM();
   
   virtual void modelingPerShot(int ishot);
   virtual void RTM_perShot(int ishot);
   virtual void adjRTM_perShot(int ishot);
   
   virtual void TomAdj();
   virtual void TomFwd();
   virtual void TomAdj_perShot(int ishot);
   virtual void TomFwd_perShot(int ishot);
   
protected:
   void setVelocityPtrForGPU(cubeData<float>& vdtSqCube, float dtComp);
   void recordSnapshotGpu (cubeData<float>* snapSrcPershot, cubeData<float>* snapRecvPershot, int ishot);
   void recordSnapshotGpu2(cubeData<float>* snapSrc2Pershot, cubeData<float>* snapRecv2Pershot, int ishot);
};

} //end of namespace

#endif // __MODELINGGPU_H_INCL__

