#include "stdafx.h"
#include "sep_params.h"

#include "srcWav.h"
#include "velFD2d.h"

#include "point3d.hpp"

#include "../param_entry.h"

#include "modelingFD2d_op.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"
#include "myOMP.h"
#include "stackTraceC.h"

using namespace Yang;
using namespace std;

//Doing Adjoing RTM
int main (int argc, char *argv[])
{
   //Init openMP, use 7 cores
   init_omp(1);
   initCallStackTrace();
   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"velocity"));

   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, just input a Guassian wavelet, beacause it will take 2nd 
   //derivate to create the ricker wavelet.
   waveletObj->loadWavelet();
   velObj->loadVel();

   velObj->domain().info("vel");
   
   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFD2d_op> modelingObj(new modelingFD2d_op());
   //Create the padding and boundary processor
   auto_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
   
   
   //chain two padding method together
   float dampingExtent = pars->get_float(csz_damping,0.05f);
   PRV(dampingExtent)
   paddingFD2dDamping* ppaddingDamp= new paddingFD2dDamping(NULL,dampingExtent);
   //auto_ptr<paddingFD2d> paddingObj (new paddingFD2d(NULL));
   cosDampingFunc dampFunc;
   ppaddingDamp->m_pDampingFunc = &dampFunc;
   
   //Value should be around 0.15
   float spongeCoef = pars->get_float(csz_spongeCoef,0.15);
      
   auto_ptr<paddingFD2dSponge1> paddingObj (new paddingFD2dSponge1(ppaddingDamp,spongeCoef));
   
   //boundary condition
   auto_ptr<boundaryFD2dAbsorb> boundaryObj(new boundaryFD2dAbsorb());

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());

   //read and set the parameters
   FDModelingParams params;
   
   float maxFreq,vmax,vmin;
   params.pVel = velObj.get();
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   
   vmax = velObj->maxval();
   vmin = velObj->minval();
   params.velModelShape = velObj->domain();
   params.modelGridShape = params.velModelShape;

   
   //also get the time param from the params
   params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
   params.stencilWidth = pars->get_int(csz_stenWidth);
   params.bUse4thOrder= pars->get_bool(csz_use4thOrderInTime);
   
   params.ndim = params.mShapeWavfield.get_ndim();
   params.offx = params.stencilWidth;params.offz = params.stencilWidth;
   
   params.m_nShots 	= pars->get_int(csz_nShots);
   params.mShot1_ix = pars->get_int(csz_shot1_ix);
   params.mShot1_iz = pars->get_int(csz_shot1_iz);
   params.mShotIncX = pars->get_int(csz_shotIncX);

   //Recording Snapshot params
   params.f_shot_snap = pars->get_int(csz_snap_f,0);
   params.j_shot_snap = pars->get_int(csz_snap_j);
   params.n_shot_snap = pars->get_int(csz_snap_n,0);
   checkSnapRecordParams(&params);

   params.mTrace1ix = pars->get_int(csz_trace1_ix);
   params.mTrace1iz = pars->get_int(csz_trace1_iz);
   params.mTraceIncX = pars->get_int(csz_traceIncX);
   params.m_nTraces = pars->get_int(csz_nTraces);
   
   params.m_nPad = pars->get_int(csz_npad);

   //Read in Absorption type
   int itype = pars->get_int(csz_bdType);
   switch (itype)
   {
      case 1:
         params.m_bdType = abs_ord1;
         break;
      case 2:
         params.m_bdType = abs_ord2;
         break;
      case 3:
         params.m_bdType = abs_Dave;
         break;
	  case 0:
	     params.m_bdType = abs_null;
         break;
	  case 4:
		 params.m_bdType = abs_ord1par; 
		 break;
      default:
         seperr("!unknown boundary condition type!\n");
   }
   //calc time parameters
   params.timeParam.totalTime = pars->get_float(csz_totalTime);
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtRecord = pars->get_float(csz_dtRec);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);
   
   params.mOffImgParams.bDoOffsetGather = false;
   
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);
   
   LOG(mark1);
   params.pSrcSnapshot 	= new cubeData<float>(csz_tag_srcSnapshot,dm::OUT,"src snapshot");
   params.pSrcSnapshot2	= new cubeData<float>(csz_tag_srcSnapshot2,dm::OUT,"scattered snapshot");
   params.pRsfSeisRec	= new cubeData<float>(csz_tag_seisRec,dm::OUT,"seis-record");
   
   params.pFinalImage	= new cubeData<float>(csz_tag_reflModel,dm::IN,"Refl Model");
   //Load refl model
   params.pFinalImage->init(NULL,true);
	
      //Init the output   
   cubeData<float> rsfSeisRec(csz_tag_seisRec,dm::OUT,"seis-record");
   //We need to figure out the size of recorded data ourself
   hypercube seisRecShape = SR_managerObj->calcModelingDataSize(&params);
   rsfSeisRec.init(&seisRecShape);
   params.pRsfSeisRec = &rsfSeisRec;
   
   //Init the operator and do the modeling
   modelingObj->initOperatorEx(&params, acAdjRTM);
   modelingObj->adjRTM();

   //then save the necessary wavefields
   //After modeling, save the snapshot and seisrecord
   params.pSrcSnapshot->unInit();
   params.pSrcSnapshot2->unInit();
   params.pRsfSeisRec->unInit();
   
   srcWav* pWav = waveletObj.release();
   //pWav->getWavelet()->dump("dump");
   delete pWav;
}
