#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "modelingFDgpu.h"

#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"
#include "imaging.h"
#include "stackTraceC.h"

#include "FWI_FD.h"
#include "FWI_FD_beamFocus.h"
#include "FWI_FD_DSO.h"
#include "WEMVA.h"
///Read parameters
#include "nonLinSolver.param.h"
#include "wavePropagate.param.h"
#include "velAna.param.h"

//#include "../misc.h"

/**
 * For Wave Equation Tomography, gradient calculation
 * Calculate the image perturbation (back projected image) based 
 * on the initial input images 
 */

using namespace Yang;
using namespace std;
   

int main (int argc, char *argv[])
{
  initCallStackTrace();
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  //Read in the parameters for the modeling Operator
  waveOpParams params;
  velAnaParam velAnaParams;
  WEMVA_Param WEMVAparam;
  
  //-Just Load the inital Image to get image size.
  dmRsf_oc<float> Img0_dm("in",dm::IN,"Initial RTM Image");
  Img0_dm.init(NULL,true);
  WEMVAparam.Img0_hx.alloc(Img0_dm.domain());
  Img0_dm.seq_grab(&WEMVAparam.Img0_hx);
  Img0_dm.unInit();
  //--------------------
  
  params.velModelShape = Img0_dm.domain().return_compatible(1,2);
  read_migration_param(&params,pars);
  params.bUseOnewayPropagator = true;

  read_velAna_param(&velAnaParams,pars);
  
  WEMVAparam.imageShape    = WEMVAparam.Img0_hx.domain().return_compatible(1,2);
  WEMVAparam.off_img_param = params.mOffImgParams;
  read_WEMVA_param(&WEMVAparam,pars);
  
  //Just a dummy modeling Obj
  modelingFD2d_op modelingFD;
  //create WEMVA FWIObj
  WEMVA_FD* pWEMVAobj = NULL;
  
  if (!WEMVAparam.bUseSingleMoveoutParam) {
    prt("\n!Use two params fitting\n");
  }
  if (WEMVAparam.bOnlyCalcSensitivityKernel) {
    prt("\n!Only calculate sensitivity kernel\n");
  }
  pWEMVAobj = create_WEMVA_obj(&velAnaParams, &params, &modelingFD, &WEMVAparam,pars);
  
  pWEMVAobj->postCalcData();
  pWEMVAobj->calcBackProjData();
  
  //Image Residue
  WEMVAparam.backProjImg_hx.dump("output");
  SAFE_DELETE(pWEMVAobj);
  
  //We use auto_ptr now, no need to delete those ptrs.
  return 0;
}

