#include "stdafx.h"
#include "axis.h"
#include "dmRsf.hpp"
#include "dmRsfDump.h"

#include "sep_params.h"
#include "stackTraceC.h"
#include "commonOps.h"

#include "algo.h"
#include "cubeAlgo2.h"
/*
 * Input point source, generate planewave sources.
 */
using boost::shared_ptr;
using namespace Yang;
using namespace std;

/// Use sparse sampled ODCIG to reconstructed ADCIG
int main (int argc, char *argv[]) {
  initCallStackTrace();
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> pars(new sep_params(argc,argv));
  
  shared_ptr<cubeData<cmplxf> > csou(new cubeData<cmplxf>());
  csou->load("in",5); //read in the point source function
  axis ax_rad = csou->domain().get_axis(3);
  const hypercube& shp = csou->domain();

  LOG("c1");
  shp.info("shp");
  cubeDataF vmod; vmod.load("vel",3);
  const hypercube& vmodshp = vmod.domain();
  
  bool b3D = false;
  if (vmodshp.get_n(3)>1) {
    b3D = true;
  }
  LOG("c1.5");
  
  int ix = vmodshp.get_n(2)/2;
  int iy = vmodshp.get_n(3)/2;
  float vsurf = vmod(iy,ix,0);
  SEEV(vsurf);

  LOG("c2");
  
  axis ax_px,ax_py;
  //The taking off angles of source x direction
  fwk::degree angleMinX(pars->get_float("xAngleMin"));
  fwk::degree angleMaxX(pars->get_float("xAngleMax"));
  
  int nAngX = pars->get_int("nxAngle");
  
  //the ray-parameter p = sin(theta)/v_surf;
  ax_px.n = nAngX;  ax_px.o = sinf(deg2rad(angleMinX))/vsurf;
  if (ax_px.n==1) { 
    ax_px.d = 1.0f;
  } else {
    ax_px.d = (sinf(deg2rad(angleMaxX))/vsurf-ax_px.o)/(nAngX-1);
  }
  
  LOG("c3");
  
  fwk::degree angleMinY(pars->get_float("yAngleMin",0.0f));
  fwk::degree angleMaxY(pars->get_float("yAngleMax",0.0f));
  int nAngY = pars->get_int("nyAngle",1);
  ax_py.n = nAngY;  ax_py.o = sinf(deg2rad(angleMinY))/vsurf;
  if (ax_py.n==1) { 
    ax_py.d = 1.0f;
  } else {
    ax_py.d = (sinf(deg2rad(angleMaxY))/vsurf - ax_py.o)/(nAngY-1);
  }
  
  vecAxis axes(5);
  axes[2] = ax_rad;
  axes[3] = ax_px; axes[4] = ax_py;
  
  axis ax_sx,ax_sy; //figuring out the geometry.
  
  ax_sx = vmodshp.get_axis(2); axes[0] = ax_sx; axes[0].label = "sx";
  ax_sy = vmodshp.get_axis(3); axes[1] = ax_sy; axes[1].label = "sy";
  axes[3].label = "px"; axes[4].label= "py";
    
  hypercube shapeO(axes);
  cubeData<cmplxf> cPlaneSrc(shapeO);
  
  LOG("c4");
  
  //Now composite the planewave sources
  int ipx,ipy,isx,isy;
  int iw; float w;
  float origx,origy; 
  origx = pars->get_float("ox",0.0f); origy = pars->get_float("oy",0.0f);
  
  float phiy,phix, px,py;
  cmplxf coef;
  for (iw=0; iw<ax_rad.n; iw++) {
    w = ax_rad.get_coord(iw);
    for (ipy=0; ipy<ax_py.n; ipy++) {
    for (isy=0; isy<ax_sy.n; isy++) {
      py    = ax_py.get_coord(ipy);
      phiy  = w*py*(ax_sy.get_coord(isy)-origy);
      for (ipx=0; ipx<ax_px.n; ipx++) {
      for (isx=0; isx<ax_sx.n; isx++) {  
        px    = ax_px.get_coord(ipx);
        phix  = w*px*(ax_sx.get_coord(isx)-origx);
        coef = exp(cmplxf(0,-(phiy+phix)));
        cPlaneSrc(ipy,ipx,iw,isy,isx) = coef*(*csou)[iw];
      }}
    }}
  }
  cPlaneSrc.dump("out");
  return 0;
}

