#include "stdafx.h"
#include "axis.h"
#include "dmRsf.hpp"
#include "dmRsfDump.h"

#include "sep_params.h"

#include "stackTraceC.h"
#include "commonOps.h"

#include "algo.h"
#include "cubeAlgo2.h"

/*
 * Input salt velocity model, gen mask which mask salt and area above salt.
 */

using boost::shared_ptr;
using namespace Yang;
using namespace std;
using namespace seplib;

/// Use sparse sampled ODCIG to reconstructed ADCIG
int main (int argc, char *argv[]) {
  initCallStackTrace();
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  shared_ptr<cubeDataF> vmod(new cubeDataF());
  
  vmod->load("in");
  float vthres = pars->get_float("vthres"); //above which the velocity is considered salt.
    
  const hypercube& shp = vmod->domain();
  cubeDataF mask(shp);

  //build mask based.
  buildThresholdMask(mask,*vmod,vthres,true,false);
  mask.dump("dump1.H",true);
  int n23 = shp.get_n23();
  int i,ii,j;
  int nz = shp.get_n(1);
  bool bFound;
  for (j=0;j<n23;j++) { //do not check the boundaries.
    bFound = false;
    for (i=nz-1;i>=0;i--) {
      if (mask(j,i)<0.5f) {
        bFound = true;
        break;
      }
    }
    if (bFound) {
      for (ii=i;ii>=0;ii--) {
        mask(j,ii) = 0.0f;
      }
    }
  }
  mask.dump("out");
}

