#include "stdafx.h"
#include "sep_params.h"
#include "cubeData.hpp"

using namespace Yang;
using namespace std;

void calc2dGradAmp(const cubeData<float>& in, cubeData<float>& out)
{
   const hypercube& shape = in.domain();
   assert( shape.sameSize(out.domain()));
   
   int nx,nz;
   nx = shape.get_n(2);
   nz = shape.get_n(1);
   int i,j;
   
   float gradX,gradZ;   
   for (j=0;j<nx-1;j++)
   {
	  for(i=0;i<nz-1;i++)
	  {
		 gradX = in(j+1,i)-in(j,i);
		 gradZ = in(j,i+1)-in(j,i);
		 out(j,i) = sqrtf(SQ(gradX)+SQ(gradZ));
	  }
   }
}

int main( int argc, char** argv){
   
   //Read params
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   
   //in is the model
   //out is the gradiant
   
   cubeDataF model("in",dm::IN,"model");
   cubeDataF data("out",dm::OUT,"data");
   
   model.init(NULL);
   data.init(&(model.domain()));
   
   //Take the 2d Gradiant operator, just the amplitude, not the directon. So it is a non-linear op.
   calc2dGradAmp(model,data);
   
   model.unInit();
   data.unInit();
}

