#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "waveOpParams.h"
#include "waveOpOneway.h"

#include "param_entry.h"
#include "WEMVA.h"

#include "stackTraceC.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "dipEstimator.h"
#include "ADCIG_Recon.h"
#include "WEMVA_RMO_Imit.h"
#include "util_seplib.h"
#include "rayTracer.h"
#include "RayTomoOp.h"

///Read parameters
#include "nonLinSolver.param.h"
#include "velAna.param.h"
#include "misc.h"

#include "my_mpi.h"
#include "waveOpMPI.h"
#include "my_timer.h"

#include <mpi.h>
//Need to build binary library files, annoying.
//#include <boost/mpi/environment.hpp>
//#include <boost/mpi/communicator.hpp>
//#include <boost/mpi/collectives.hpp>

/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search 
 */

//Reflection case

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;
//Doing beam wave-equation tomography for reflection case.

#include <unistd.h>
//Do born modeling
int main (int argc, char *argv[])
{
  
  mpi::environment MPI_Init(&argc,&argv);
  mpi::communicator world;
  
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  int rank,nw;
  rank  = world.rank();
  nw    = world.size();

  if (rank==2) {
    LOGC("in process 2");
    PRVEC(argv,argc);
  }
  
  //important, fix the stdin/stdout if no is provided.
  if (rank==0) mpi_sep_send_args(nw,10,0);
  else mpi_sep_receive_args();
  
  //if (rank==0) {
  if (0) {
    for (int ir=1; ir<nw; ir++) {
      string sz("sleep 10");
      world.send(ir,WRKTAG,sz);
      prt("Sending msg %s to node %d\n", sz.c_str(), ir);
    }
    
    char sz_cwd[512];
    getcwd(sz_cwd,511);
    PRV(sz_cwd);
    PRVEC(argv,argc);
    cubeDataF tool2Cube;
    strcpy(sz_cwd, "/net/server/homes/sep/yang/Research/beamWET2/bp/sepToolTemp7537.H");
    //strcpy(sz_cwd, "vel");
    tool2Cube.load(sz_cwd,3);
    //tool2Cube.load("s1",3);
    tool2Cube.info("t2c");
    //tool2Cube.dump("tt159.H",true);
  }
  
  //if (0) { //master node, execute all the main logic
  if (world.rank()==0) { //master node, execute all the main logic
    Yang::timer clk;
    LOGC("in root process");
    //PRVEC(argv,argc);
    //PRV(world.size());
    initCallStackTrace(); 
    init_omp(0);
    //Read params, params need contain geometry info that is exact the same as the input data. In
    //synthetic this is easy to do.
    //shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
    //sep_params* pars = autoPars.get();
    
    shared_ptr<vel2d>  velObj     (new vel2d(csz_tag_vel,dm::IN,"vel"));
    shared_ptr<srcWav> waveletObj (new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
    
    
    //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
    waveletObj->loadWavelet();
    velObj->loadVel(2); velObj->info("Initial velocity model");
    
    shared_ptr<ImgPar> imagingPar(new ImgPar());  ImgPar& imgPar = *imagingPar;
    //load acquisition parameters
    shared_ptr<AcqPar> acqPar(new AcqPar());
    acqPar->readParam(pars,velObj->domain());
    imgPar.readParam(pars,*acqPar); //Load imaging parameters
    LOG("A2");
    shared_ptr<waveOpOneway> modelingObj;
    //Read in the parameters for the modeling Operator
    shared_ptr<waveOpParams>  waveOpPar(new waveOpParams()); waveOpParams& params = *waveOpPar;
    LOG("A3");
    params.readParam(pars);
    bool bUseOneWay = params.bOneway();
    LOG("A3.5");
    //split the jobs
    shared_ptr<workScheduler1d> ws(new workScheduler1d(acqPar->nShots(),world.size()-1,workScheduler1d::chunk));
    LOG("A4");
    
    //Here we just focus on use one-way
    if (bUseOneWay) {
      modelingObj.reset(new waveOpOnewayMPI(waveOpPar,acqPar,imagingPar,*velObj,pars,ws));
      params.read_oneway_prop_param(pars);

      //split the csou files
      int iax = 4; //sx is axis 4 for oneway program
      LOG("B2");
      if (acqPar->sourceType()==AcqPar::STpoint) {
        dup_file_MPI(params.onewayPar()->csouFileName, params.suf_MPI_,*ws,""); //duplicate single file
      } else { //split source files
        split_file_MPI(params.onewayPar()->csouFileName, params.suf_MPI_,4,*ws,"");
      }
      LOG("After spliting the csou file"); cerr.flush();
    } else {
      params.read_twoway_prop_param(pars);
      //calc time parameters
      float maxFreq,vmax,vmin;
      //Read in the peak/center freq of the seismic wavelet
      maxFreq = waveletObj->get_float("peakFreq");
      waveletObj->setPeakFreq(maxFreq);
      vmax = velObj->maxval();  vmin = velObj->minval();
      params.calcTimeParamsFor2wayModeling(maxFreq, vmax,vmin, velObj->domain());
    }
    
    bool bBornModeling = pars->get_bool("bornFwd"); //true means modeling, otherwise imaging.
    if (!bBornModeling) { //migration
      string fn_crec = params.onewayPar()->crecFileName;
      //Need to split the data
      split_file_MPI(fn_crec,params.suf_MPI_,4,*ws,"");

      const hypercube& shpOffimg= imgPar.finalSubOffImageShp(); axis ax_h = shpOffimg.get_axis(3);
      const hypercube& shp0img  = imgPar.finalImageShp();
      shpOffimg.info("shpO");
      //born modeling
      cubeDataF img(shpOffimg);
      modelingObj->migration_op(fn_crec,&img);
      img.dump("imgh");
    } else { //born modeling
      bool b3D = acqPar->b3D();
      cubeDataF img;
      if (!b3D) {
        img.load("refl",3);
      } else {
        VERIFY(0);
      }
      string fn_crec = params.onewayPar()->crecFileName;
      modelingObj->bornModeling_op(&img,fn_crec);
    }
    
    //release the slave modes, send exit cmd.
    for (int ir=1; ir<world.size(); ir++) {
      world.send(ir,EXITAG,string(""));
    }
    clk.check("Migration finish");
  }
  
  mpi::slaveModeMPI slaveMode;
  if (world.rank() > 0) { //the slaver node, just accept cmds, execute it, and wait for another cmd. Keep in mind that the master process is also a slave node.
    slaveMode.on();
    prt("exit slaveMode, irank=%d \n",world.rank());
  }
  ////finally synchronize, then exit.
  world.barrier();
  return 0;
}

