#include "stdafx.h"
#include "axis.h"
#include "dmRsf.hpp"
#include "dmRsfDump.h"
#include "sep_params.h"
#include "stackTraceC.h"
#include "commonOps.h"

///Read parameters
#include "algo.h"
#include "cubeAlgo2.h"

/*
 * Input migrated offset gather(might be random subsampled,
 then compute the objective function and calculate the back projected data.
 */

//Reflection case
using boost::shared_ptr;
using namespace Yang;
using namespace std;
using namespace seplib;

/// Use sparse sampled ODCIG to reconstructed ADCIG
int main (int argc, char *argv[]) {
  initCallStackTrace();
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  shared_ptr<cubeDataF> dat(new cubeDataF());
  
  dat->load("in");
  float* p = dat->dataPtr(), fMaxVal = dat->maxvalAbs();
  long n123 = dat->domain().get_n123(); long i; //cnt = 0;
//float thresRatio = pars->get_float("thrRatio",1e-3);
//for (i=0; i<n123; i++) {
//  if (fabsf(p[i])>fMaxVal*thresRatio) {
//    cnt++;
//  }
//}
//prt("Maxval is %e, Spareness ratio is %2.3f %%\n",fMaxVal,float(cnt)/float(n123)*100.f);
  
  //Rather than use Maxval, we use computing energy, and count the energy of the largest x% data points.
  dat->multiply(*dat);
  float energyRatio = pars->get_float("energyRatio",0.99f);
  std::sort(p,&p[n123-1]+1,std::greater<float>()); //sort in descending order
  double fE = 0.0; double fThr = 0.0f;
  for (i=0; i<n123; i++) {fThr += p[i];}
  fThr *= energyRatio;
  for (i=0; i<n123; i++) {
    fE += p[i];
    if (fE>=fThr) break;
  }
  prt("Maxval is %e, Spareness ratio is %2.3f %% at %2.3f of total energy\n",fMaxVal, float(i)/n123*100.0f, energyRatio);
}

