#include "stdafx.h"
#include "sep_params.h"

#include "myOMP.h"
#include "stackTraceC.h"

#include "cubeData.hpp"

using namespace Yang;
using namespace std;

//Doing Adjoing RTM
int main (int argc, char *argv[])
{

  int nx, ny, nz;
  float dx, dy, dz;
  float ox, oy, oz;
  float x11, x12, x13;
  float x21, x22, x23;
  float x, y, z, depth, dplnz, zref;
  float v0, alpha1,alpha2, beta, decay1, decay2, fm ;
  int ix, iy, iz, nref, iref;
  cubeData<float> vmod, refl, bvel;
  cubeData<int> plnz;


  float slope, xc, yc, wt1, wt2;
	int ixc,iyc;

  float width1, height1, thick1, width2, height2, thick2;
  int   nxwid,nywid;

  sep_params* pars=new sep_params(argc,argv);

  nx = pars->get_int("nx",1); dx = pars->get_float("dx", 1.); ox = pars->get_float("ox", 0.);
  ny = pars->get_int("ny",1); dy = pars->get_float("dy", 1.); oy = pars->get_float("oy", 0.);
  nz = pars->get_int("nz",1); dz = pars->get_float("dz", 1.); oz = pars->get_float("oz", 0.);
  axis ax_x,ax_y,ax_z;
  ax_x.set(nx,ox,dx,"x");
  ax_y.set(ny,oy,dy,"y");
  ax_z.set(nz,oz,dz,"z");
  
  x11 = pars->get_float("x11", 0.);    // the x coord of 1st anomaly
  x12 = pars->get_float("x12", 0.);
  x13 = pars->get_float("x13", 0.);   // the z coord of 1st anomaly

  width1 = pars->get_float("width1", 0.5f);
  width2 = pars->get_float("width2", width1);
  height1 = pars->get_float("height1", 0.5f);
  height2 = pars->get_float("height2", height1);
  thick1 = pars->get_float("thick1", 0.5f);
  thick2 = pars->get_float("thick2", thick1); //in y dimension

  x21 = pars->get_float("x21", 0.);
  x22 = pars->get_float("x22", 0.);
  x23 = pars->get_float("x23", 0.);

  alpha1 = pars->get_float("alpha1", 0.);
  alpha2 = pars->get_float("alpha2", alpha1);

  v0 = pars->get_float("v0", 1.f);
  
	xc    = pars->get_float("xcref",ox+(nx/2)*dx);
  yc    = pars->get_float("ycref",oy+(ny/2)*dy);
  zref  = pars->get_float("zref", 0.);
  slope = pars->get_float("slope",0.); //reflector slope, tan(\alpha)
  fm    = pars->get_float("fm", 0.5/dz);

  hypercube shape(nz,nx,ny);
  shape.set_axis(1,ax_z);
  shape.set_axis(2,ax_x);
  shape.set_axis(3,ax_y);
  shape.info("shape");
  vmod.alloc(shape);
  refl.alloc(shape);
  bvel.alloc(shape);
  //plnz.alloc(hypercube(nref));

	vmod = v0;
  for (iz=1; iz<=nz; iz++) {
    z = oz+(iz-1)*dz;
    for (iy=1; iy<=ny; iy++) {
      y = oy+(iy-1)*dy;
      for (ix=1; ix<=nx; ix++) {
        x = ox+(ix-1)*dx;
        decay1 = -( SQ((x-x11)*2.0/width1)  + SQ((z-x13)*2.0/height1) + SQ((y-x12)*2.0/thick1));
        decay2 = -( SQ((x-x21)*2.0/width2)  + SQ((z-x23)*2.0/height2) + SQ((y-x22)*2.0/thick2));
        vmod(ix-1,iy-1,iz-1) = v0 * (1.0 - alpha1*exp(decay1) - alpha2*exp(decay2));
      }
    }
  }

  bvel = v0;
  PRV2(xc,slope);
  nxwid = 0;
  if (fabs(slope) > epsilon) {
	  // nxwid = (oz+(nz-1)*dz-zref); // abs(slope) * 0.8 / dx; need further change !!!
	  PRV(nxwid);
    VERIFY(0);
  } else {
	  nxwid = nx/2;
  }
  nywid = ny/2;
  
	nxwid = nxwid-1;
	ixc = (xc-ox)/dx+1;
  nywid = nywid-1;
  iyc = (yc-oy)/dy+1;
  
    
  for (ix=ixc-nxwid; ix<=ixc+nxwid; ix++) {
  for (iy=iyc-nywid; iy<=iyc+nywid; iy++) {
    x = ox + (ix-1)*dx;
  //  do iz=1, nz
  //     z = oz+(iz-1)*dz
  //     refl(iz,ix,1) = refl(iz,ix,1) + wavelet(z-zref+(x-xc)*slope, fm, 1)
  //	
  //  end do
  	z 	= ((zref-(x-xc)*slope));
    iz	= int((z-oz)/dz) + 1; //iz is lower bound
  	wt1 = (z-oz)/dz + 1.0f - iz;
  	wt2 = 1.0f - wt1;
  	//put two spikes at correct location
  	refl(iy-1,ix-1,iz-1) += wt2;
  	refl(iy-1,ix-1,iz) += wt1;
  }}
  
  vmod.dump("out",3);
  refl.dump("refl",3);
  bvel.dump("bvel",3);  
}
