#ifndef __VELANA_H_INCL__
#define __VELANA_H_INCL__

#include "cubeData.hpp"
#include "cubeAlgo.h"
#include "waveOpParams.h"
#include "waveOps.h"
#include "imaging.h"
#include "cubeAlgo2.h"
#include "velAna.param.h"
#include "util_seplib.h"

using boost::shared_ptr;
// TODO: your code here
namespace Yang {
class param_off2ang3d;
//Memory objects and params used in WEMVA Tomograpy case (reflection case)
struct WEMVA_Param {
  typedef enum tagObjFuncType {
    classic,
    YiLuo,
    beamFocusing1,  //single param fitting
    beamFocusing2,  //double param fitting
    DSO_off, //offset domain DSO
    DSO_ang,  //angle domain DSO
    RMO
  } WEMVAmethodType; //method type.
  WEMVAmethodType mWEMVAtype; 
  int WEMVAmethod_index_;
  
  typedef enum __optObjFunc {
    StackPowerObjFunc, //simply zero-offset image maximization.
    DSO_off_objFunc,
    cheatTrueModObjFunc, //use true model as object, this is cheat
    RMOobjFunc, //Measured residual moveout objective function.
    NormalizedStackPowerOF
  } WEMVAobjFuncType; //object function type.
  WEMVAobjFuncType objFuncType_;
  bool bDSO_balanceX; //Balance the amplitude along X direciton in DSO objective function. 
     
  enum dRho_dsOpType {drds_Jacob_2, drds_Jacob_m}; //drho/ds operator type, 1) invert 2X2 Jacobian 2) invert m+1 by m+1 Jacobian
  dRho_dsOpType m_dRho_dsOpType;
  
  //if for 3D, we need out-of-core access.
  param_off2ang3d *par_o2a3d,*par_a2o3d;
  string fn_img_off,fn_bpjImg_off;  //(z,x,y,hx,hy)
  string fn_img_ang,fn_bpjImg_ang;  //(z,gamma,azim,x,y)
  
  string fn_dJdu,fn_semb,fn_envSemb;
  //dmRsf<float>  *initImg_off_oc, *initImg_ang_oc, *backProjImg_off_oc, *backProjImg_ang_oc;
  
  
  //Initial images.
  cubeData<float> Img0,Img0_AGC;
  shared_ptr<cubeDataF> croppedImg0_; //crop the outer region of Img0, the shp of it should be specified in main().
  cubeData<float> Img0_hx,Img0_hx_t; //offset image
  shared_ptr<cubeDataF> imgh0_mask; //The imaging mask for h=0;
  shared_ptr<cubeDataF> sampMask_hx_t; //for CSensing
  
  shared_ptr<cubeDataF> vmodTrue;
  //shared_ptr<cubeDataF> layerStripMask_;
  
  cubeData<float> backProjImg; //zero-subsurf-offset Img;
  cubeData<float> backProjImg_hx, backProjImg_hx_t; //offset Image for back propagation.
  cubeData<float> backProjImg_ax; //angle domain bpj image
  cubeData<float> img_ax_t2; //angle domain bpj image, tmp
  //For angle gather offset transform
  enumOff2angType off2angType;
  int ax_ang_n; //The number of points on angle axis
  
  //For DSO    
  cubeData<float> Img_hx_res; //Img residue. 
  enum dso_balance_type {none_bl,x_bal,xz_bal};
  dso_balance_type dso_bal_type_;

  //For YiLuo 1991 method
  
  //For beam Focus Biondi,sep142
  //These images axises are transposed. e.g (z,ax,x) Not (z,ax,x)
  cubeData<float> Img0_ax_t; //The offset/angle domain Image, transposed
  // cubeData<float> dJduCube; //calc the derivative of J over the moveout params. (z,x), already defined in its father class
  cubeData<float> Ecube,finalWtCube; //(z,x), the denumerator
  
  float thresMaxValPercent; //from 0~1
  float fEmaskValThres;  //the alternative, an arbitrary value.
  cubeData<float> thresMaskCube;
  
  bool bSingleEvent; //If there is only one event
  bool bUseSingleMoveoutParam; //If use only one moveout params.
  bool bOnlyCalcSensitivityKernel; //if then just set dJ/du = spike.
  float amax; //max angle for angle transform
  
  //Taper angle gathers for far angle, used in DSO.
  bool bTaperAngleGather;
  
  axis ax_beam_z; //For dividing windows along z directions
  float beam_z_window;
  bool bBalanceEnumeratorWt; //Will normalize the amplitude of the F matrix in beamfocus method.
  
  //For correlation based shift parameter method.
  bool bFindShiftUserCorr;
  int iStartLag,iEndLag; //[istart,iend]
  //For the method using LS fitting to determine curvature change to shift at each angle
  bool bUseWeightedLSfit;
  bool bNoLSfit;
   
  bool bCompensateImageZ; //compensate amplitude along depth direction of migrated image.
  shared_ptr<seplib::AGC_Param> imgBalAgcPar_;

  //constructor
  WEMVA_Param();
};


//This is the base class for Time Finite difference velocity analysis method.
class velAna_FD {
private:
  //Data objects for vel Ana
  velAnaParam* m_pVelAnaParam;
  shared_ptr<waveOps> waveOp_;
public:
  // Constructor
  velAna_FD(velAnaParam* pVelAnaParam, const shared_ptr<waveOps>& pWave): m_pVelAnaParam(pVelAnaParam), waveOp_(pWave) {}
  static void setMoveoutDeriv(cubeData<float>& cube, tagMoveOutType tp);
  
  const shared_ptr<waveOps>& waveOp() const {return waveOp_;}
  velAnaParam*         velAnaPar() const {return m_pVelAnaParam;} //no const pointer.
  
  // Destructor
  virtual ~velAna_FD();
  
  void  setVelocity(const cubeDataF* pVelCube);
  
  virtual void loadDataObs(); //should not be inherited anymore
  virtual void calcData() = 0;
  virtual void calcBackProjData() = 0;
  virtual void calcSlowGradient(cubeData<float>& velGradCube) = 0;
  
  //calculate objective function, used in solver
  virtual float calcObjFunc() = 0;

  virtual void callBackAtIter(int iIter) {} //Call back function for a nonlin solver
protected:
	virtual void flipGradient(cubeData<float>& gradCube) = 0;
};

// Destructor implementation
inline velAna_FD::~velAna_FD()
{
}

}

#endif // __VELANAPARAMS_H_INCL__

