#include "stdafx.h"
#include "param_func.h"

#include "decl_fwd.h"
#include "velAna.param.h"
#include "velAna.h"
#include "WEMVA.h"
#include "param_entry.h"

#include "misc.h"
#include "dipEstimator.h"
#include "ADCIG_Recon.h"
#include "WEMVA_RMO_Imit.h"
#include "util_seplib.h"
#include "rayTracer.h"
#include "RayTomoOp.h"

// TODO: your code here
using namespace seplib;
using namespace Yang;
using namespace std;
using boost::shared_ptr;

WEMVA_Param::WEMVA_Param():par_o2a3d(NULL), thresMaxValPercent(0.1), bSingleEvent(true), bUseSingleMoveoutParam(true),
     bOnlyCalcSensitivityKernel(false),
     amax(75.0f), bTaperAngleGather(false),bFindShiftUserCorr(false),iStartLag(0),iEndLag(0),
     m_dRho_dsOpType(drds_Jacob_m),bCompensateImageZ(true),imgBalAgcPar_(new AGC_Param())
     //initImg_off_oc(NULL),initImg_ang_oc(NULL),backProjImg_off_oc(NULL),backProjImg_ang_oc(NULL)
{
  dso_bal_type_ = x_bal;
}

void Yang::read_velAna_param(velAnaParam* pParam, param_func* pars) {
  velAnaParam& velAnaParams = *pParam;
  
  //weight more on deeper reflectors
  velAnaParams.bWeightUpZ_  = pars->get_bool("bWeightUpZ",false);
  velAnaParams.bIllumBal_   = pars->get_bool("bIllumBal",false); 
  PRV2(velAnaParams.bWeightUpZ_,velAnaParams.bIllumBal_);
  LOG("v1");
  //beam focus related
  velAnaParams.beamSize = pars->get_int(csz_subGroupSize,-1); //e.g. beam size;

  int moveOutFuncType = pars->get_int("moveOutType",2);
  switch(moveOutFuncType) {
    case 1: 
      velAnaParams.moveOutType = Linear;
      break;
    case 2:
      velAnaParams.moveOutType = Quadratic;
    break;
    case 3:
      velAnaParams.moveOutType = tangentSquare;
    break;
    default:
      seperr("wrong moveOutType for beam focusing");
    break;
  }

  LOG("v2");
  velAnaParams.bUseSemblanceToCalcDJdb = pars->get_bool("bUseSemblance",false);
  if (velAnaParams.bUseSemblanceToCalcDJdb) { //read the axis for semblance scan
    velAnaParams.ax_semb.d = pars->get_float("ax_semb_d");
    velAnaParams.ax_semb.n = pars->get_int("ax_semb_n");
    float ax_semb_o = (velAnaParams.ax_semb.n/2)*-1.0f*velAnaParams.ax_semb.d;
    velAnaParams.ax_semb.o = pars->get_float("ax_semb_o",ax_semb_o);
    velAnaParams.ax_semb.print_info("ax_semb");
    velAnaParams.bUseGuassianDerivForDJDU   = pars->get_bool("bUseGaussianDerivForDJDU", false);
    PRV(velAnaParams.bUseGuassianDerivForDJDU);
    velAnaParams.fSembSmoothWindow          = pars->get_float("fSembSmoothWindow",  velAnaParams.ax_semb.n*velAnaParams.ax_semb.d/2.0f);
    velAnaParams.bNormalizeSemb             = pars->get_bool("bNormSemb",false); 
    PRV(velAnaParams.bNormalizeSemb);
  }
  
  LOG("v3");
  int indexCorrUsage = pars->get_int("corrUsage",0);
  switch(indexCorrUsage) {
  case 0:
    velAnaParams.maxCorrUsage = velAnaParam::pickMaxCorr;       break;
  case 1:
    velAnaParams.maxCorrUsage = velAnaParam::smoothCorrDeriv0;  break;
  case 2:
    velAnaParams.maxCorrUsage = velAnaParam::GaussianWeightCorr;break;
  default:
    VERIFY(0);
  }
  velAnaParams.bPickMoveoutPar= pars->get_bool("bPickMoveoutPar",false);
  if (velAnaParams.bPickMoveoutPar) {
    PRV(velAnaParams.bPickMoveoutPar);
    VERIFY(velAnaParams.bUseSemblanceToCalcDJdb);
  }
  velAnaParams.bDivideEnergyTermUseMask = pars->get_bool("bDivideUseMask",true);
  
  LOG("v4");
  //Read the model Mask
  velAnaParams.gradMaskPreSmooth_.reset();
  LOG("v8");
  if (tagExist("gradMaskPreSmooth")) {
    velAnaParams.gradMaskPreSmooth_.reset(new cubeDataF());
    bool b3d = pars->get_bool(csz_b3D,false);
    if (!b3d) {
      velAnaParams.gradMaskPreSmooth_->load("gradMaskPreSmooth",2);
    } else {
      velAnaParams.gradMaskPreSmooth_->load("gradMaskPreSmooth",3);
    }
  }
  LOG("v9");
  return;
}

void Yang::read_WEMVA_param(WEMVA_Param* pParam, param_func* pars, const ImgPar& imgPar) {
  FUNC_TRACE_BEG(read_WEMVA_param);
  WEMVA_Param& WEMVAparam = *pParam;
  WEMVAparam.bSingleEvent = pars->get_bool(csz_bSingleEvent,true);
  WEMVAparam.bTaperAngleGather = pars->get_bool(csz_bTaperAngGather,false);

  WEMVAparam.bUseSingleMoveoutParam = pars->get_bool("bUseSingleMoveoutParam",false);
  WEMVAparam.bOnlyCalcSensitivityKernel = pars->get_bool("bSensKernel",false);
  int off2angTypeIndex = pars->get_int(csz_off2angType,0);
  
  const hypercube& fImgShp = imgPar.finalImageShp();
  bool b3D = fImgShp.get_ndim()==3;
  
  switch (off2angTypeIndex) {
    case 0:
      WEMVAparam.off2angType = Yang::slantstackYangAA; VERIFY(!b3D);
      break;
    case 1:
      WEMVAparam.off2angType = Yang::slantstackYang;  VERIFY(!b3D);
      break;
    case 2:
      WEMVAparam.off2angType = Yang::seplibModified;  //VERIFY(!b3D);
      break;
    case 3:
      WEMVAparam.off2angType = Yang::biondi3d_fd; VERIFY(b3D);
      break;
    default:
      VERIFY(0);
  }
  PRV(off2angTypeIndex);
  
  //The two below are alternative to each other.
  WEMVAparam.thresMaxValPercent = pars->get_float(csz_thresMax,0.05f);
  WEMVAparam.fEmaskValThres     = pars->get_float(csz_thresMax,0.01f);

  //Param for offset->angle transform
  const offsetImagingParams& offParam = imgPar.offImgPars();
  int ax_ang_n = pars->get_int("ax_ang_n", offParam.nhx); //default the number of pts is the same as in offset axis
  //PRV(ax_ang_n);
  WEMVAparam.ax_ang_n = ax_ang_n;
  WEMVAparam.amax     = pars->get_float(csz_amax,50.0f); //if use two way prop, then the amax could be higher.

  //Read params for multiple event beam focusing
  VERIFY(!fImgShp.isEmpty());
  axis ax_imgz = fImgShp.get_axis(1);
  float dz = ax_imgz.d;
  
  if (!WEMVAparam.bSingleEvent) {
    float fBeam_z_window      = pars->get_float("beam_z_window");
    int   iBeam_z_window      = lroundf(fBeam_z_window/dz);
    fBeam_z_window            = iBeam_z_window*dz;
    WEMVAparam.beam_z_window  = fBeam_z_window;

    float fBeam_z_interv  = pars->get_float("beam_z_interv",dz);
    int   iBeam_z_interv  = lroundf(fBeam_z_interv/dz);
    fBeam_z_interv        = iBeam_z_interv * dz;
    float fBeam_z_o       = pars->get_float("beam_z_o",  ax_imgz.o-0.5f*dz+fBeam_z_interv*0.5f);
    //put the origin to closest 0.5dz grid
    fBeam_z_o = ax_imgz.o + 0.5f*dz*lroundf((fBeam_z_o-ax_imgz.o)/(0.5f*dz));
    
    int   iBeam_z_n       = pars->get_int("beam_z_n",    int(ax_imgz.n/iBeam_z_interv));
    WEMVAparam.ax_beam_z.label = "beam_z";
    WEMVAparam.ax_beam_z.o= fBeam_z_o;
    WEMVAparam.ax_beam_z.d= fBeam_z_interv;
    WEMVAparam.ax_beam_z.n= iBeam_z_n;
    WEMVAparam.ax_beam_z.print_info("ax_beam_z");
    
    //VERIFY the axis info 
    //fBeam_z_o either sits on grid or sits in 50% between two grids, corresponding to fBeam_z_window being odd or even 
    float   coordBeam_z_o       = ax_imgz.get_offsetf(fBeam_z_o);
    bool bOddGrid = false;
    int     iCoordBeam_z_o      = int(coordBeam_z_o);
    if (fabs(coordBeam_z_o - iCoordBeam_z_o) < 0.1f) {
      bOddGrid = false;
      VERIFY(iBeam_z_window %2 == 1);
    } else if (fabs(coordBeam_z_o/dz - iCoordBeam_z_o - 0.5f) < epsilon) {
      bOddGrid = true;
      VERIFY(iBeam_z_window %2 == 0); //beam window should have even size if beam_o sits 50% grid
    }
    prt("\n iBeam_z_window=%d, iBeam_z_interv=%d ", iBeam_z_window, iBeam_z_interv);
    if (iBeam_z_window % iBeam_z_interv != 0) {
      prt("\n WARN: iBeam_z_window mod iBeam_z_interv !!== 0, the beam window does not cover the image point uniformly.");
    }
    
    //Check the beam window, should not be out of range.
    float max_z; ax_imgz.getRange(NULL,&max_z);
    PRV4(fBeam_z_o,ax_imgz.o,fBeam_z_window,dz);
    VERIFY(fBeam_z_o - ax_imgz.o - fBeam_z_window*0.5f + 0.5f*dz >= 0.0f);
    PRV2(fBeam_z_o + (iBeam_z_n-1)*fBeam_z_interv + (fBeam_z_window-dz)*0.5f,max_z);
    VERIFY(fBeam_z_o + (iBeam_z_n-1)*fBeam_z_interv + (fBeam_z_window-dz)*0.5f <= max_z);
    
    fBeam_z_window -= dz; //subtract dz for the convenience of calculating the start/end grid point
    WEMVAparam.beam_z_window = fBeam_z_window; 
    PRV(WEMVAparam.beam_z_window);
  }
  
  WEMVAparam.bBalanceEnumeratorWt = pars->get_bool("bBalanceEnum",false);
  //For angle stack reference shift method
  WEMVAparam.bFindShiftUserCorr = pars->get_bool("bUseCorr",true);
  WEMVAparam.iStartLag          = pars->get_int("startLag",-50);
  WEMVAparam.iEndLag            = pars->get_int("endLag",50);
  
  WEMVAparam.bUseWeightedLSfit  = pars->get_bool("bUseWeightedLSfit",true);
  WEMVAparam.bNoLSfit           = pars->get_bool("bNoLSfit",false);
  
  if (b3D) { //we should have all file names for the out-of-core results.
    WEMVAparam.fn_img_off     = pars->get_string("fImgOff");
    WEMVAparam.fn_img_ang     = pars->get_string("fImgAng");
    WEMVAparam.fn_bpjImg_off  = pars->get_string("fBpjImgOff");
    WEMVAparam.fn_bpjImg_ang  = pars->get_string("fBpjImgAng");
    
    WEMVAparam.fn_dJdu        = pars->get_string("fDJDU");
    WEMVAparam.fn_semb        = pars->get_string("fSemb");
    WEMVAparam.fn_envSemb     = pars->get_string("fEnvSemb");
  }
  
  //What objective function to use.
  static pair<int,WEMVA_Param::WEMVAobjFuncType> OF_match_table[] = {make_pair(0,WEMVA_Param::StackPowerObjFunc), make_pair(1,WEMVA_Param::DSO_off_objFunc), make_pair(2,WEMVA_Param::cheatTrueModObjFunc), make_pair(3,WEMVA_Param::RMOobjFunc), make_pair(4,WEMVA_Param::NormalizedStackPowerOF)};
  WEMVAparam.objFuncType_ = findEnumItemFromPar(pars,OF_match_table,NELEM(OF_match_table),"OF",2); //default is TrueModObjFunc
  PRV(WEMVAparam.objFuncType_);
  
  WEMVAparam.WEMVAmethod_index_ = pars->get_int("WEMVA_type",11); //default is RMO method.
  WEMVAparam.imgBalAgcPar_->readParam(pars,fImgShp,"imgBal_");

  FUNC_TRACE_END();
}

shared_ptr<WEMVA_FD> Yang::create_WEMVA_obj(velAnaParam* pVApar, const shared_ptr<waveOps>& pModelingObj,
                                 WEMVA_Param* pWEMVApar, param_func* pars) {
  WEMVA_FD*    pWEMVAobj    = NULL;
  WEMVA_Param& WEMVAparam   = *pWEMVApar;
  
  //create WEMVA FWIObj
  switch(WEMVAparam.WEMVAmethod_index_) {
  case 1: 
    prt("\n using maximize angle stack obj function");
    WEMVAparam.mWEMVAtype = WEMVA_Param::classic; 
    //pWEMVAobj = new WEMVA_maxAngleStack(pVApar, pModelPar, pModelingObj, &WEMVAparam);
    pWEMVAobj = new WEMVA_max0offset(pVApar, pModelingObj, &WEMVAparam);
    break;
  case 4:  
    prt("\nusing minimization objective function with shift moveout params, with respect"
       " to the shift refered to the angle-stack trc");
    WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1; 
    //pWEMVAobj = new WEMVA_FD_angleMoveout(pVApar, pModelingObj, &WEMVAparam);
    break;
  case 5:
    prt("\nusing img-space beam focusing");
    WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1;      
    pWEMVAobj = new WEMVA_FD_beamFocus(pVApar, pModelingObj, &WEMVAparam);
    break;
  case 55:
    prt("\nusing beamfocus_m method");
    WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1; 
    pWEMVAobj = new WEMVA_FD_beamFocus_m(pVApar, pModelingObj, &WEMVAparam);
    break;
  case 61:
  case 62:
    prt("using DSO Refl obj Function, offset domain");
    WEMVAparam.mWEMVAtype = WEMVA_Param::DSO_off; 
    pWEMVAobj = new WEMVA_FD_DSO_off(pVApar, pModelingObj, &WEMVAparam);
    if (WEMVAparam.WEMVAmethod_index_ == 61) {
      WEMVAparam.dso_bal_type_ = WEMVA_Param::none_bl;
    } else if (WEMVAparam.WEMVAmethod_index_ == 62) {
      WEMVAparam.dso_bal_type_ = WEMVA_Param::x_bal;
    }
    break;
  case 7: 
    prt("using DSO Refl obj Function, angle domain");
    WEMVAparam.mWEMVAtype = WEMVA_Param::DSO_ang; 
    //pWEMVAobj = new WEMVA_FD_DSO_ang(pVApar, pModelingObj, &WEMVAparam);
    break;
  case 101:
    prt("\n using time-shift image bfocusing");
    WEMVAparam.bSingleEvent = false;
    WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1; 
    //pWEMVAobj = new WEMVA_FD_beamFocusTime(pVApar, pModelingObj, &WEMVAparam);
    break;
   //case 9: break;
   //case 10: break
  case 11: //use RMO Imit, need to measure RMO
    prt("using RMO Imit WEMVA, angle domain\n");
    WEMVAparam.mWEMVAtype = WEMVA_Param::RMO; 
    pWEMVAobj = NULL; //let WEMVA_Imit function to create the WEMVA object
    break;
  default:
    seperr("!!Wrong  WEMVAmethod_index_ !!\n");
  }
  PRV(WEMVAparam.mWEMVAtype);
  return shared_ptr<WEMVA_FD>(pWEMVAobj);
}

//Create the RMO_imit operator
shared_ptr<WEMVA_FD> Yang::create_WEMVA_imit(const param_func* pars, const axis& ax_h, const hypercube& shp0img, velAnaParam& velAnaParams, const shared_ptr<waveOps>& modelingObj, WEMVA_Param& WEMVAparam, const shared_ptr<AcqPar>& acqPar, bool bOneway, const cubeDataF& vel) {
  LOG("a1");
  bool b3D = acqPar->b3D();
  //read the axis information of the angle domain.
  axis ax; read_axis_from_par(ax,pars,string("ang")); axis_deg ax_gamma(ax);
  shared_ptr<Param_o2a> par_o2a(new Param_o2a(ax_gamma, ax_h,slantstackYangAA, b3D, bOneway));
  
  LOG("a2");
  shared_ptr<ADCIG_Recon> adcig_recon;
  //adcig_recon.reset(new ADCIG_ReconStomp(par_o2a,true,pars));
  adcig_recon.reset(new ADCIG_ReconFull(par_o2a));
  
  //Load the sub-samp mask for the hx
  //WEMVAparam.sampMask_hx_t.reset(new cubeDataF());WEMVAparam.sampMask_hx_t->load("subSampMask",3);
  LOG("a3");
  shared_ptr<StructTensorParam> par_ST(new StructTensorParam(string("")));par_ST->readParam(pars);
  shared_ptr<dipEstimator> est(new dipEstimator(par_ST));
  
  LOG("a4");
  shared_ptr<AGC_Param>           agcPar(new AGC_Param());            agcPar->readParam(pars,shp0img);
  shared_ptr<Param_RMO_Measurer>  par_RMO(new Param_RMO_Measurer());  par_RMO->readParam(pars,shp0img);
  shared_ptr<RMO_Measurer>        RMO_measure(new RMO_Measurer(est,par_RMO,agcPar));
  LOG("a5");
  
////Use a 1D rayTracer to give better RMO measurement
//shared_ptr<rayTracer> rayTrace( new rayTracerZ1d());
//shared_ptr<cubeDataF> velSm(new cubeDataF());
//vel.allocMemoryTo(*velSm);  seplib::callSepTool(*velSm,vel,"Smooth","rect1=10 rect2=20");
//rayTrace->velIs(velSm);
//shared_ptr<RayTomoOp> rayTomoOp(new RayTomoOp(acqPar,rayTrace));
//RMO_measure->RayTomoOpIs(rayTomoOp);
  
  
  shared_ptr<WEMVA_FD> pWEMVAobj(new WEMVA_RMO_Imit(&velAnaParams, modelingObj, &WEMVAparam,RMO_measure,adcig_recon));
  return pWEMVAobj;
}

