#ifndef __VELANA_PARAM_H_INCL__
#define __VELANA_PARAM_H_INCL__

#include "decl_fwd.h"
#include "imaging.h"

using boost::shared_ptr;
// TODO: your code here
namespace Yang {
  struct velAnaParam {
    cubeDataF velGradCube;
    shared_ptr<cubeDataF> velModelMask; //Mask the part of vel model that can be resolved by survey geometry, some part of the model should be fixed, useful for masking out the alt body or a layer stripping scheme,the fixed part is 0, and the variable part is 1.0
    shared_ptr<cubeDataF> gradMaskPreSmooth_;
      
    string dObsFilename;
    dmRsf<float>* 	  pDataObs;
    cubeData<float>   dObsCube; //in core observed data, not always in place
    
    //For Beamfocusing
    int beamSize;
    //Currently quadratic form, theta= u*(offset)^2
    //assume input is 1d, axis is along offset
    tagMoveOutType moveOutType;   
        
  	//For Yang's Beam focusing
    bool bIgnoreBfY2ndBpjTerm;
  	cubeData<float> pickedCurvatureCube; //a param
  	cubeData<float> pickedSlopeCube; // b parameter 
    //Subsampled version, because picking is kind of expensive.
    cubeData<float> pickedCurvatureCubeSubsamp; 
  	cubeData<float> pickedSlopeCubeSubsamp; 
  	axis ax_a, ax_b; //axes for a,b parameter scaning.
  
    cubeData<float> dJdbCube; //The derivative of obj function over moveout params
    bool  bUseSemblanceToCalcDJdb;
    cubeData<float> semblanceCube;
    axis ax_semb;
    bool  bNormalizeSemb;
    bool bUseGuassianDerivForDJDU; float fSembSmoothWindow;
    
    typedef enum tagMaxCorrUsageType {
      pickMaxCorr,
      smoothCorrDeriv0,
      GaussianWeightCorr
    } maxCorrUsageType;
    maxCorrUsageType maxCorrUsage;
    
    bool bPickMoveoutPar; //if false, use derivative at zero
    bool bDivideEnergyTermUseMask;
    
    //Params helping to form better gradient
    bool bWeightUpZ_; //give more weights to deeper reflectors
    bool bIllumBal_;  //use src/recv illumination to balance the model gradient
    
    //default param values
  	velAnaParam(): beamSize(0), bIgnoreBfY2ndBpjTerm(false), bUseSemblanceToCalcDJdb(false),bDivideEnergyTermUseMask(false),
      bWeightUpZ_(false),bIllumBal_(false) {}
  };
  
  //Memory objects and params used in FWI or Transmission (Data space inversion)
  //public velAna_FD_Param 
  struct FWI_Param {
    typedef enum tagObjFuncType {
      classic,
      YiLuo,
      YiLuo2,
      beamFocusing1,  //single param fitting
      beamFocusing2,  //double param fitting
      beamFocusing21, //double param, but with shift and slope
      beamFocusingY1,	//require picking, Just measuring the slope
      beamFocusingY2,	//require picking, 2 parameter picking, but we can also use the slope only, 
                      //by setting the a cube equal to zero, as the one above
      DSO //analogy to DSO
    } FWIobjFuncType;
  
    FWIobjFuncType mFWItype;
    
     cubeData<float> dCalCube;
     cubeData<float> backProjCube;   
     //For correlation based method
     cubeData<float> C_cube;
     cubeData<float> deltaTauCube; //shift for Maximum correlations
     int iStartLag;
     int iEndLag;
     
     //For YiLuo 1991 method
     cubeData<float> dObs2_shiftedCube;
     cubeData<float> dObs_shiftedCube;
     cubeData<float> E_cube; //scalar for each xs,xg;
     float widthGaussDeriv;
  };
  
  void read_velAna_param(velAnaParam* pParam, param_func* pars);  
  void read_WEMVA_param(WEMVA_Param* pParam, param_func* pars,const ImgPar& imgPar);
  void read_FWI_param(FWI_Param *pFWIparam, param_func* pars);
  shared_ptr<WEMVA_FD> create_WEMVA_obj(velAnaParam* pVApar, const shared_ptr<waveOps>& pModelingObj, WEMVA_Param* pWEMVApar, param_func* pars);
  FWI_FD*   create_FWI_obj(velAnaParam* pVApar,   const shared_ptr<waveOps>& pModelingObj, FWI_Param* pFWIpar, param_func* pars);  
  shared_ptr<WEMVA_FD> create_WEMVA_imit(const param_func* pars, const axis& ax_h, const hypercube& shp0img, velAnaParam& velAnaParams, const shared_ptr<waveOps>& modelingObj, WEMVA_Param& WEMVAparam, const shared_ptr<AcqPar>& acqPar, bool bOneway, const cubeDataF& vel);
}//end namepsace
#endif // __VELANA_PARAM_H_INCL__

